<?php
if(!defined('sugarEntry') || !sugarEntry) die('Not A Valid Entry Point');
/*********************************************************************************
 * The contents of this file are subject to the SugarCRM Public License Version
 * 1.1.3 ("License"); You may not use this file except in compliance with the
 * License. You may obtain a copy of the License at http://www.sugarcrm.com/SPL
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * All copies of the Covered Code must include on each user interface screen:
 *    (i) the "Powered by SugarCRM" logo and
 *    (ii) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.  See full license for
 * requirements.
 *
 * The Original Code is: SugarCRM Open Source
 * The Initial Developer of the Original Code is SugarCRM, Inc.
 * Portions created by SugarCRM are Copyright (C) 2004-2006 SugarCRM, Inc.;
 * All Rights Reserved.
 * Contributor(s): ______________________________________.
 ********************************************************************************/
/*********************************************************************************

 * Description:  Updates entries for the Last Viewed functionality tracking the 
 * last viewed records on a per user basis.
 * Portions created by SugarCRM are Copyright (C) SugarCRM, Inc.
 * All Rights Reserved.
 * Contributor(s): ______________________________________..
 ********************************************************************************/


/** This class is used to track the recently viewed items on a per user basis.
 * It is intended to be called by each module when rendering the detail form.
 * Portions created by SugarCRM are Copyright (C) SugarCRM, Inc.
 * All Rights Reserved.
 * Contributor(s): ______________________________________..
*/
class Tracker {

    private function __construct() {}
    
    static function table_name() {
		return AppConfig::setting('model.detail.tracker.table_name', 'tracker');
    }
    
    static function history_max() {
    	return AppConfig::setting('layout.history_max_viewed', 50);
    }

    /**
     * Add this new item to the tracker table.  If there are too many items (global config for now)
     * then remove the oldest item.  If there is more than one extra item, log an error.
     * If the new item is the same as the most recent item then do not change the list
 * Portions created by SugarCRM are Copyright (C) SugarCRM, Inc.
 * All Rights Reserved.
 * Contributor(s): ______________________________________..
     */

    /* longreach - added $owner_id parameter
        function track_view($user_id, $module_name, $item_id, $item_summary)
     */

    static function track_view($user_id, $module_name, $model_name, $item_id, $item_summary, $owner_id = '')
	{
    	global $db, $log;

		if (AppConfig::module_primary_bean($module_name) != $model_name)
			return;
		if (AppConfig::setting("model.detail.$model_name.disable_tracker"))
			return;
		if (AppConfig::setting("model.detail.$model_name.type") != 'bean')
			return;

        self::delete_history($user_id, $module_name, $item_id);

		// longreach - added
		if ($item_summary === null || $db->isReadOnly()) return;
		
        // Add a new item to the user's list
		$datetime=gmdate("Y-m-d H:i:s");		
		$id = create_guid();
		$model = self::table_name();

        $upd = RowUpdate::blank_for_model($model);
        $upd->set([
            'id' => $id,
            'user_id' => $user_id, 
            'module_name' => $module_name,
            'item_id' => $item_id,
            'item_summary' => $item_summary,
            'date_modified' => $datetime,
            'owner_id' => $owner_id
        ]);
        $upd->save();
		self::prune_history($user_id);
    }

    /**
     * param $user_id - The id of the user to retrive the history for
     * param $module_name - Filter the history to only return records from the specified module.  If not specified all records are returned
     * return - return the array of result set rows from the query.  All of the table fields are included
 * Portions created by SugarCRM are Copyright (C) SugarCRM, Inc.
 * All Rights Reserved.
 * Contributor(s): ______________________________________..
     */
    static function get_recently_viewed($user_id, $module_name=null, $exclude=null)
    {
    	global $db, $log;
        if(! $user_id)
            return array();
        $cache_key = "Tracker.$user_id";
        $ret = sugar_cache_get($cache_key, $found, true);
        if($found)
        	return $ret;
        $table = self::table_name();
        $where = "user_id='$user_id'";
        if(is_array($exclude) && count($exclude)) {
        	$where .= ' AND item_id NOT IN (\''. implode("','", array_map([$db, 'quote'], $exclude)) . '\')';
        }
        $query = "SELECT * FROM `$table` WHERE $where ORDER BY date_modified DESC";
        $result = $db->limitQuery($query, 0, self::history_max(), true);

        $list = array();
        while($row = $db->fetchByAssoc($result, -1, false))
        {
            
            // longreach - added
            if (! ACLController::checkAccess($row['module_name'], 'view', $row['owner_id'])) continue;
            
            // If the module was not specified or the module matches the module of the row, add the row to the list
            if($module_name == "" || $row['module_name'] == $module_name)
            {
            	$list[] = $row;
            }
        }
        sugar_cache_put($cache_key, $list, true);

        return $list;
    }

    // longreach - added
    /**
     * param $user_id - The id of the user to retrive the history for
     * param $module_name - Filter the history to only return records from the specified module.  If not specified all records are returned
     * param $limit - Limit for result
     * return - return the array of result set rows from the query.  ID field included
     */
    static function get_for_search($user_id, $module_name=null, $limit = 10)
    {
    	global $db;
        if(! $user_id)
            return array();
        if($module_name && ! is_array($module_name)) {
			$cache_key = "Tracker.$user_id.search.$module_name";
			$ret = sugar_cache_get($cache_key, $found);
			if($found)
				return $ret;
		}
        $table = self::table_name();
        $ext_sel = '';
        $mod_filter = '';
		if(is_array($module_name)) {
			if($module_name) {
				$names = array_map(array($db, 'quote'), $module_name);
				$mod_filter = "AND module_name IN ('" . implode("','", $names) . "')";
			}
			$ext_sel = ', module_name, id, date_modified';
		} else if($module_name) {
			$nm = $db->quote($module_name);
			$mod_filter = "AND module_name='$nm'";
        }
        $query = "SELECT item_id{$ext_sel} FROM `$table` WHERE user_id='$user_id' $mod_filter ORDER BY date_modified DESC";
        $result = $db->limitQuery($query, 0, $limit, true);

        $list = Array();
        while($row = $db->fetchByAssoc($result, -1, false)) {
        	if($ext_sel)
        		$list[] = $row;
        	else
				$list[] = $row['item_id'];
        }
        if(isset($cache_key))
			sugar_cache_put($cache_key, $list, true);

        return $list;
    }

    static function clear_cache($user_id, $module=null) {
        sugar_cache_clear("Tracker.$user_id");
        if($module)
			sugar_cache_clear("Tracker.$user_id.search.$module");
    }

    /**
     * INTERNAL -- This method cleans out any entry for a record for a user.
     * It is used to remove old occurances of previously viewed items.
 * Portions created by SugarCRM are Copyright (C) SugarCRM, Inc.
 * All Rights Reserved.
 * Contributor(s): ______________________________________..
     */
    static function delete_history( $user_id, $module, $item_id)
    {
        global $db;
    	if($db->isReadOnly()) return;
        $lq = new ListQuery(self::table_name());
        $lq->addSimpleFilter('user_id', $user_id);
        $lq->addSimpleFilter('module_name', $module);
        $lq->addSimpleFilter('item_id', $item_id);
        $db->query($lq->getDeleteSql(), true);
        self::clear_cache($user_id, $module);
    }
    
    /**
     * INTERNAL -- This method cleans out any entry for a record.  This is intended to delete
     * all occurances of an item that is being deleted from the system.
 * Portions created by SugarCRM are Copyright (C) SugarCRM, Inc.
 * All Rights Reserved.
 * Contributor(s): ______________________________________..
     */
    static function delete_item_history($module, $item_id)
    {
    	global $db;
    	if($db->isReadOnly() || ! $item_id) return;
        $lq = new ListQuery(self::table_name());
        $lq->addSimpleFilter('module_name', $module);
        $lq->addSimpleFilter('item_id', $item_id);
        $db->query($lq->getDeleteSql(), true);
        self::clear_cache(AppConfig::current_user_id(), $module);
    }
    
    /**
     * INTERNAL -- This function will clean out old history records for this user if necessary.
 * Portions created by SugarCRM are Copyright (C) SugarCRM, Inc.
 * All Rights Reserved.
 * Contributor(s): ______________________________________..
     */
    static function prune_history($user_id)
    {
    	global $db, $log;
    	if($db->isReadOnly()) return;
        // Check to see if the number of items in the list is now greater than the config max.
		$table = self::table_name();

        $lq = new ListQuery($table);
        $lq->addSimpleFilter('user_id', $user_id);
        $count = $lq->fetchResultCount();
        $history_max = self::history_max();
		
		if($count > $history_max) {
			$log->debug("prune history (current $count, max $history_max)");
			$del_count = $count - $history_max;
            $lq = new ListQuery($table);
            $lq->addSimpleFilter('user_id', $user_id);
            $lq->setOrderBy('date_modified');
			$db->query($lq->getDeleteSql(false, $del_count), true, "Error pruning tracker history");
		}
    }
    
	static function add_tracker_query_join(ListQuery $lq, $name=null, $filter=true, $sort=false) {
		if($lq->isUnionQuery()) {
			$src = $lq->getUnionQueries();
			$trk = array(
				'name' => '_track_date',
				'type' => 'datetime',
				'union_source' => array(),
				'no_query' => true,
			);
			if($src) foreach($src as $idx => &$s) {
				self::add_tracker_query_join($s, $name, $filter, false);
				$trk['union_source'][$idx] = '_track_date';
				if($sort)
					$s->addField($name . '.date_modified', '_track_date');
			}
			$lq->addField($trk);
			if($sort)
				$lq->setOrderBy('_track_date DESC');
			return;
		}
	
		if(! isset($name)) $name = 'trk';
		$trk = new ModelDef('tracker');
		$spec = array(
			'type' => 'link',
			'name' => $name,
			'required' => 1,
			'link_type' => 'joined',
			'relationship_spec' => array(
				'left' => array(
					'model' => $lq->getModelName(),
					'key' => 'id',
					'count' => 'many',
				),
				'right' => array(
					'model' => 'tracker',
					'key' => 'item_id',
					'role_field' => 'module_name',
					'role_value' => $lq->getModuleDir(),
					'count' => 'many',
				),
			),
		);
		$lq->addCustomLink($trk, $name, $spec);
		$lq->addSimpleFilter($name . '.user_id', AppConfig::current_user_id());
		if($sort)
			$lq->setOrderBy($name . '.date_modified DESC');
	}
}

?>
