<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo,
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */

require_once 'include/upload_file.php';
require_once 'include/GoogleSyncClient.php';

abstract class GoogleSync
{
	const MAX_TIME = 300;

	protected $user;
	protected $options;
	protected $startTime;
	protected $client;
	protected $client_wrapper;
	protected $service;

	protected $params = array(
		'max_items' => 50,
	);

	abstract protected function getConfigKey();
	abstract protected function getUpdated($item);
	abstract protected function getItems($query);
	abstract protected function getQuery();
	abstract protected function getService();
	abstract protected function createIAHItem($row = null);
	abstract protected function setIAHItemFields($iahItem, $item);
	abstract protected function saveItem($item, $iahItem);
	abstract protected function deleteItem($item);
	abstract protected function mapLegacyGoogleId($googleId);
	abstract protected function getGoogleIdClause($googleId);

	/**
	 * Get SQL query to retrieve IDs of items modified since $lastSyncTime.
	 * Should return IDs of deleted objects if deletion is supported.
	 */
	abstract protected function getSQLQuery($lastSyncTime);

	abstract protected function realCreateGoogleItem($iahItem);
	abstract protected function realUpdateGoogleItem($item, $iahItem);
	abstract protected function getListEntry($googleId);


	protected function getPreference($name)
	{
		$key = $this->getConfigKey();
		return $this->user->getPreference($key . '_' . $name);
	}

	protected function setPreference($name, $value)
	{
		$key = $this->getConfigKey();
		$this->user->setPreference($key . '_' . $name, $value, 0, 'global', $this->user);
	}

	protected function savePreferences()
	{
		$this->user->savePreferencesToDB();
	}

	protected function getDirection()
	{
		return $this->getPreference('direction');
	}

	public function __construct($user = null, $options = array())
	{
		if (!$user) {
			$user = $GLOBALS['current_user'];
		}
		if (!($user instanceof User)) {
			return;
		}
		$this->user = $user;
		$this->options = $options;

		$cl = GoogleSyncClient::create($this->user);
		$client = $cl->getClient();

		$this->client = $client;
		$this->client_wrapper = $cl;

		$this_reset = $this->getPreference('reset_time');
		$global_reset = $this->user->getPreference('google_sync_reset_time');
		if ($global_reset > $this_reset)
			$this_reset = $global_reset;
		$check_reset = $this->getPreference('check_reset_time');
		if ($this_reset > $check_reset) {
			$reset = array('last_sync_time', 'last_sync_time2', 'sync_window_current', 'sync_window_start', 'last_sync_token');
			foreach ($reset as $k)
				$this->setPreference($k, '');
			$this->setPreference('check_reset_time', $this_reset);
		}
		$this->startTime = time();
	}

	public function sync()
	{
		if (!($this->user instanceof User)) {
			return false;
		}
		$direction = $this->getDirection();
		if (!$direction) {
			return true;
		}
		if (!$this->createService()) {
			return false;
		}

		$ret = false;
		try {
			switch ($direction) {
				case 'to_google':
					$ret = $this->syncToGoogle();
					break;
				case 'to_iah':
					$ret = $this->syncToIAH();
					break;
				case 'both':
					$ret = $this->syncBoth();
					break;
			}
		} catch (Exception $e) {
			$this->logException($e);
			$ret = false;
		}
		$this->savePreferences();
		return $ret;
	}

	protected function createService()
	{
		try {
			if(!$this->client_wrapper->assignStoredToken()) {
				return false;
			}
			$this->service = $this->getService();
		} catch (Exception $e) {
			$this->logException($e);
			return false;
		}
		return true;
	}

	protected function syncBoth()
	{
		try {
			$this->syncToGoogle();
			$this->syncToIAH();
		} catch (Exception $e) {
			$this->logException($e);
		}
	}

	protected function syncToIAH()
	{
		$lastSyncToken = $this->getPreference('last_sync_token');
		$lastPageToken = $this->getPreference('last_page_token');
		$lastSyncTime = $this->getPreference('last_sync_time');

		$query = $this->getQuery();
		if (($lastSyncToken || $lastPageToken) && $this->supportsSyncToken()) {
			unset($query['showDeleted']);
			unset($query['orderBy']);
		}
		if ($lastSyncToken && $this->supportsSyncToken()) {
			$query = ['syncToken' => $lastSyncToken];
		} elseif ($lastPageToken && $this->supportsSyncToken()) {
			$query = ['pageToken' => $lastPageToken];
		} else {
			$this->setQueryUpdatedMin($query, $lastSyncTime);
		}
		$max = (int)array_get_default($this->params, 'max_items', 50);
		$query['maxResults'] = $max;
		$processed = 0;
		for (;;) {
			try {
				$items = $this->backoff(function() use ($query) {
					 return $this->getItems($query);
				});
			} catch (Google_Service_Exception $e) {
				if ($e->getCode() == 410 || $e->getCode() == 400) {
					$query = $this->getQuery();
					$query['showDeleted'] = true;
					$items = $this->backoff(function() use ($query) {
						return $this->getItems($query);
					});
				} else {
					throw $e;
				}
			}
			$lastSyncToken = $lastPageToken = null;
			foreach ($items as $item) {
				if (!$item)
					continue;
				$processed++;
				$this->syncOneToIAH($item);
				$updated = $this->getUpdated($item);
				if ($updated > $lastSyncTime)
					$lastSyncTime = $updated;
			}
			$token = null;
			if ($this->supportsSyncToken() && $items && method_exists($items, 'getNextSyncToken')) {
				$token = $items->getNextSyncToken();
			}
			if ($token) {
				$lastSyncToken = $token;
				$lastPageToken = null;
				break;
			} else {
				$page = null;
				if ($items && method_exists($items, 'getNextPageToken'))
					$page = $items->getNextPageToken();

				$lastSyncToken = null;
				$lastPageToken = $page;
				break;
			}
		}
		$this->setPreference('last_sync_time', $lastSyncTime);
		$this->setPreference('last_sync_token', $lastSyncToken);
		$this->setPreference('last_page_token', $lastPageToken);
		return true;
	}

	protected function canCreateInIAH($item, $iahItem)
	{
		return true;
	}

	protected function canSyncToGoogle($iahItem)
	{
		return true;
	}

	protected function syncOneToIAH($item)
	{
		$iahItem = $this->loadIahItemForGoogleItem($item);
		$deleted = $this->isItemDeleted($item);
		
		if ($iahItem->new_record && (!$this->canCreateInIAH($item, $iahItem) || $deleted) ) {
			return;
		}
		
		$this->loadSyncFields($iahItem);
		
		if (!$deleted && strtotime($iahItem->_google_fields['google_timestamp']) >= strtotime($this->getUpdated($item))) {
			return;
		}
		
		$iahItem->_google_fields['google_id'] = $this->getItemId($item);

		$actions = [];	
		if (!$deleted) {
			$actions = $this->setIAHItemFields($iahItem, $item);
		}
		
		if ($iahItem->new_record) {
			$iahItem->set('created_by', $this->user->id);
			$iahItem->set('date_entered', gmdate('Y-m-d H:i:s', strtotime($this->getUpdated($item))));
		}
		$iahItem->set('date_modified', $iahItem->_google_fields['google_last_sync'] = gmdate('Y-m-d H:i:s'));
		$iahItem->_google_fields['google_timestamp'] = $this->getUpdated($item);
		if (!is_array($actions) || empty($actions)) {
			$actions = [
				[
					'perform' => 'save',
					'item' => $iahItem,
				]
			];
			if ($deleted) {
				$actions[] = [
					'perform' => 'delete',
					'item' => $iahItem,
				];
			}
		}

		$this->executePostSyncActions($actions);
	}

	protected function executePostSyncActions(array $actions) {
		global $db;
		foreach ($actions as $action) {
			$item = array_get_default($action, 'item');
			switch (array_get_default($action, 'perform')) {
			case 'save':
				try {
					$item->setRelatedData('break_sequence', 'none');
					$item->save();
					$this->saveSyncFields($item);
				} catch (Exception $e) {
					$this->logException($e);
				}
				break;
			case 'delete':
				$item->setRelatedData('break_sequence', 'none');
				$item->markDeleted(true);
				$this->saveSyncFields($item);
				break;
			case 'delete_sync':
				$query = sprintf(
					"DELETE FROM google_sync WHERE related_type='%s' AND related_id='%s'",
					$db->quote($item->getModuleDir()),
					$db->quote($item->getPrimaryKeyValue())
				);
				$db->query($query);
				break;
			}
		}
	}

	protected function syncToGoogle()
	{
		global $db;

		$lastSyncTime = $this->getPreference('last_sync_time');

		$query = $this->getSQLQuery($lastSyncTime);
		$rows = array();
		$res = $db->query($query, true);
		while ($row = $db->fetchByAssoc($res)) {
			$rows[] = $row;
		}
		foreach ($rows as $row) {
			$iahItem = $this->createIAHItem($row);
			if(!$iahItem->getField('id')) {
				continue;
			}
			$this->loadSyncFields($iahItem);

			$this->backoff(function() use ($iahItem) {
				$this->syncOneToGoogle($iahItem);
			});
		}
		return true;
	}

	function syncOneToGoogle($iahItem)
	{
		if ($this->canSyncToGoogle($iahItem)) {
			try {
				if (empty($iahItem->_google_fields['google_id'])) {
					$doc = $this->createGoogleItem($iahItem);
				} else {
					$doc = $this->updateGoogleItem($iahItem);
				}
			} catch (Exception $e) {
				$this->logException($e);
				return false;
			}
		} else {
			$iahItem->_google_fields['google_last_sync'] = $iahItem->getField('date_modified');
			$this->saveSyncFields($iahItem);
		}
	}

	function createGoogleItem($iahItem)
	{
		$iahItem->_google_fields['google_sync_error'] = 0;
		$item = $this->realCreateGoogleItem($iahItem);
		$last_modified = $iahItem->getField('date_modified');
		if (!$item || $item === true) {
			$iahItem->_google_fields['google_last_sync'] = $iahItem->getField('date_modified');
		} elseif ($this->getItemId($item)) {
			$iahItem->_google_fields['google_id'] = $this->getItemId($item);
			$iahItem->_google_fields['google_timestamp'] = $this->getUpdated($item);
			$iahItem->_google_fields['google_last_sync'] = $iahItem->getField('date_modified');
		}
		$this->saveSyncFields($iahItem);
		return $item;
	}

	function updateGoogleItem($iahItem)
	{
		$doUpdate = false;
		$saveSync = false;
		$item = null;
		try {
			$item =  $this->getListEntry($iahItem->_google_fields['google_id']);
			try {
				if ($iahItem->getField('deleted')) {
					$this->deleteItem($item);
					$saveSync = true;
				} else {
					$doUpdate = !$this->googleIsNewer($item, $iahItem);
				}
				$last_modified = $iahItem->getField('date_modified');
				if ($doUpdate) {
					$doUpdate = $this->realUpdateGoogleItem($item, $iahItem);
					if ($doUpdate) {
						$iahItem->_google_fields['google_sync_error'] = 0;
						$this->saveItem($item, $iahItem);
					}
				} else {
					$doUpdate = true;
				}
			} catch (Google_Service_Exception $e) {
				if ($e->getCode() === 410) {
					$iahItem->setRelatedData('break_sequence', 'none');
					$iahItem->markDeleted(true);
					$doUpdate = true;
				} else {
					throw $e;
				}
			}
		} catch (Google_Service_Exception $e) {
			$GLOBALS['log']->fatal($e->getMessage());
			if ($e->getCode() === 404) {
				$doUpdate = true;
			} else {
				throw $e;
			}
		}
		if ($doUpdate || $saveSync) {
			$iahItem->_google_fields['google_timestamp'] = $item ? $this->getUpdated($item) : '';
			$iahItem->_google_fields['google_last_sync'] =  $iahItem->getField('date_modified');
			$this->saveSyncFields($iahItem);
		}
		return $item;
	}

	protected function logException($e)
	{
		if (!defined('SCHEDULER_EXTERNAL_TRIGGER')) {
			pr2(get_class($e));
			pr2($e->getMessage());
		}
		$GLOBALS['log']->fatal($e->getMessage());
	}

	protected function loadSyncFields($iahItem)
	{
		global $db;
		if (!$iahItem->new_record) {
			$query = sprintf(
				"SELECT * FROM google_sync 
					WHERE user_id='%s' AND related_type='%s' AND related_id='%s'",
				$this->user->id, $iahItem->getModuleDir(), $iahItem->getPrimaryKeyValue()
			);
			$res = $db->query($query, true);
			if ($row = $db->fetchByAssoc($res)) {
				$row['google_id'] = $this->mapLegacyGoogleId($row['google_id']);
				$iahItem->_google_fields = $row;
			}
		}
		if (empty($iahItem->_google_fields)) {
			$iahItem->_google_fields = array(
				'google_id' => '',
				'google_timestamp' => '',
				'google_last_sync' => '',
				'google_sync_error' => 0,
			);
		}
	}

	protected function saveSyncFields($iahItem)
	{
		global $db;
		if (empty($iahItem->_google_fields['google_last_sync']))
			$iahItem->_google_fields['google_last_sync'] = '1970-01-03 00:00:00';
		$query = sprintf(
			"REPLACE INTO google_sync 
				SET user_id='%s', related_type='%s', related_id='%s',
				google_id='%s', google_timestamp='%s',
				google_last_sync='%s', google_sync_error=%d
			",
			$this->user->id, $iahItem->getModuleDir(), $iahItem->getPrimaryKeyValue(),
			$iahItem->_google_fields['google_id'],
			$iahItem->_google_fields['google_timestamp'],
			$iahItem->_google_fields['google_last_sync'],
			@$iahItem->_google_fields['google_sync_error']
		);
		$db->query($query, true);
	}

	protected function loadIahItemForGoogleItem($item)
	{
		$googleId = $this->getItemId($item);
		$clause = $this->getGoogleIdClause($googleId);
		$lq = new ListQuery('google_sync', array('related_type', 'related_id', 'google_id'));
		$lq->addFilterClause($clause);
		$result = $lq->runQuerySingle();
		$data = array();
		if ($result && !$result->failed) {
			$data = array(
				'module' => $result->getField('related_type'),
				'id' => $result->getField('related_id'),
			);
		}
		return $this->createIAHItem($data);
	}

	protected function googleIsNewer($item, $iahItem)
	{
		$googleTime = strtotime($this->getUpdated($item));
		$iahTime = strtotime($iahItem->_google_fields['google_timestamp']);
		return $googleTime > $iahTime;
	}

	protected function getItemId($item)
	{
		return $item->getId();
	}

	protected function isItemDeleted($item)
	{
		return false;
	}

	protected function supportsSyncToken() {
		return true;
	}

	protected function setQueryUpdatedMin(&$query, $min) {
	}

	protected function backoff($action) {
		$delay = 1000000;
		$retries = 0;
		for (;;) {
			if (time() > $this->startTime + self::MAX_TIME) throw new Exception("timeout");
			try {
				return $action();
			} catch (\Exception $e) {
				$rethrow = true;
				if ($e instanceof Google_Service_Exception) {
					if ($e->getCode() == 429 || $e->getCode() == 403) {
						usleep($delay);
						$delay *= 2;
						$rethrow = false;
					}
				}
				if ($rethrow) {
					throw $e;
				}
			}
		}
	}
}

