<?php
require_once('include/SimplePie/simplepie.inc');

class IAH_FeedParser extends SimplePie {
	var $cache_duration = 3600;
	var $defer_callback;
	var $force_refresh;
	var $loaded = false;
	
	function __construct($feed_url = null) {
		parent::__construct();
		$this->registry->register('Misc', 'SimplePie_Misc_NoError', true);
		$cache_dir = CacheManager::get_location('feeds');
		$this->set_cache_location($cache_dir);
		$this->set_feed_url($feed_url);
		$this->set_image_handler(false);
	}
	
	function setDeferCallback($callback) {
		if(! $callback) return;
		$this->defer_callback = $callback;
		$this->set_file_class('IAH_DeferredFeed');
	}
	
	function setForceRefresh($flag=true) {
		$this->force_refresh = (bool)$flag;
	}
	
	function init() {
		$ret = parent::init();
		if($this->loaded) {
			if($this->defer_callback) {
				call_user_func($this->defer_callback);
			}
			$this->defer_callback = null;
		} else if($ret)
			$ret = 'download';
		return $ret;
	}
		
	function fetch_data(&$cache) {
		// first the cache is checked by the parent class
		// if that fails then an IAH_DeferredFeed is created, which will throw an exception
		try {
			if($this->force_refresh && $this->cache && $cache) {
				$this->force_refresh = false;
				$cache->unlink();
			}
			$ret = parent::fetch_data($cache);
			if($ret) {
				$this->loaded = true;
			}
		} catch(IAHDeferredFeedException $e) {
			$this->set_file_class('SimplePie_File'); // for future if-modified-since request
			$headers = array(
				'Accept' => 'application/atom+xml, application/rss+xml, application/rdf+xml;q=0.9, application/xml;q=0.8, text/xml;q=0.8, text/html;q=0.7, unknown/unknown;q=0.1, application/unknown;q=0.1, */*;q=0.1',
			);
			$this->file = new IAH_DLMgrFeed($this->feed_url, $this->timeout, 5, $headers, $this->useragent, $this->force_fsockopen);
			$this->file->startDownload($this);
			return true;
		}
		return $ret;
	}
	
	function finishDownload(&$file) {
		if($file->success) {
			if($this->cache && $file->is_cache_req && $file->status_code === 304) {
				$cache = $this->registry->call('Cache', 'get_handler', array($this->cache_location, call_user_func($this->cache_name_function, $this->feed_url), 'spc'));
				$cache->touch();
				$this->loaded = true;
			}
		}
		$cb = $this->defer_callback;
		$this->defer_callback = null;
		if(! $this->loaded) {
			//$this->cache = true;
			$this->file =& $file;
			$this->loaded = $this->init();
		}
		if($cb)
			call_user_func($cb);
	}
	
	function get_favicon() {
		if ($return = $this->get_channel_tags(SIMPLEPIE_NAMESPACE_ATOM_10, 'icon')) {
			return $this->sanitize($return[0]['data'], SIMPLEPIE_CONSTRUCT_IRI, $this->get_base($return[0]));
		}
		elseif (($url = $this->get_link()) !== null && preg_match('/^http(s)?:\/\//i', $url)) {
			$favicon = SimplePie_Misc::absolutize_url('/favicon.ico', $url);
			return $this->sanitize($favicon, SIMPLEPIE_CONSTRUCT_IRI);
		}
		return false;
	}
}

class IAHDeferredFeedException extends Exception {}

class IAH_DeferredFeed extends SimplePie_File {
	function __construct($url, $timeout = 1, $redirects = 5, $headers = null, $useragent = null, $force_fsockopen = false) {
		throw new IAHDeferredFeedException();
	}
}

class IAH_DLMgrFeed extends SimplePie_File {
	var $dl_params;
	var $parent;
	var $is_cache_req = false;
	
	function __construct($url, $timeout = 1, $redirects = 5, $headers = null, $useragent = null, $force_fsockopen = false) {
		$this->dl_params = array(
			'url' => $url,
			'timeout' => $timeout * 10,
			'max_redirects' => $redirects,
			'send_headers' => $headers,
			'user_agent' => $useragent,
			'callback' => array(&$this, 'finishDownload'),
		);
		$this->method = SIMPLEPIE_FILE_SOURCE_REMOTE | SIMPLEPIE_FILE_SOURCE_CURL;
		$this->url = $url;
		if(isset($headers['if-modified-since']))
			$this->is_cache_req = true;
	}
	
	function startDownload($parent) {
		$this->parent = $parent;
		DLManager::add_download($this->dl_params);
	}
	
	function finishDownload(&$dl) {
		$this->status_code = $dl->status_code;
		if(! $dl->failed) {
			$this->success = true;
			$this->headers = $dl->getHeaders();
			$this->body = $dl->getResponse();
		} else
			$this->success = false;
		if($this->parent)
			$this->parent->finishDownload($this);
	}
}

class SimplePie_Misc_NoError extends SimplePie_Misc {
	public static function error($message, $level, $file, $line) {
		// skip
	}
}

class IAH_DeferCache extends SimplePie_Cache {
	static function get_handler($location, $filename, $extension) {
		return new IAH_DeferCache_File($location, $filename, $extension);
	}
}

class IAH_DeferCache_File extends SimplePie_Cache_File {
	function __construct($location, $filename, $extension) {
		parent::__construct($location, $filename, $extension);
	}
	
	function load() {
		// real cache has already been checked. make init() use SimplePie->file instead.
		return false;
	}
}

?>
