<?php

namespace OneCRM\MultiFA;


use Katzgrau\KLogger\Logger;

class MultiFactorAuthLogin {

	/**
	 * Is display authentication code on login or not
	 *
	 * @return bool
	 */
	public function isDisplayCodeVerification() {
		return ! empty($_SESSION['mfa_display_login_code']);
	}

	/**
	 * Display Authentication code verification on user login
	 *
	 * @param \User $user
	 * @param Logger $logger
	 *
	 * @return bool
	 */
	public function setLoginCodeVerification(\User $user, Logger $logger) {
		$multiFA = $this->initAuthenticator($user, $logger);
		$result = false;

		if ($multiFA->isEnabled()) {
			if(session_status() == PHP_SESSION_ACTIVE) {
				$_SESSION['mfa_display_login_code'] = 1;
				$result = true;
			}
		}

		return $result;
	}

	/**
	 * Hide Authentication code verification on user login
	 */
	public function hideLoginCodeVerification() {
		if(session_status() == PHP_SESSION_ACTIVE && isset($_SESSION['mfa_display_login_code'])) {
			unset($_SESSION['mfa_display_login_code']);

			if (isset($_SESSION['mfa_login_details']))
				unset($_SESSION['mfa_login_details']);
		}
	}

	/**
	 * Store login details for futher authentication
	 *
	 * @param array $details
	 */
	public function storeLoginDetails($details) {
		if (is_array($details) && ! empty($details))
			$_SESSION['mfa_login_details'] = json_encode($details);
	}

	/**
	 * Get stored login details for authentication
	 *
	 * @return array
	 */
	public function getLoginDetails() {
		$result = [];

		if (isset($_SESSION['mfa_login_details'])) {
			$result = json_decode($_SESSION['mfa_login_details']);

			if (is_object($result)) {
				$result = (array) $result;
				$properties = ['params', 'settings', 'redirect_params'];

				foreach ($properties as $property) {
					if (isset($result[$property]) && is_object($result[$property]))
						$result[$property] = (array)$result[$property];
				}
			}
		}

		if (! is_array($result))
			$result = [];

		return $result;
	}

	/**
	 * @param string $login_message - login form text
	 *
	 * @return string
	 */
	public function getCodeVerificationForm($login_message) {
		require_once('include/MultiFA/MultiFactorAuthFormRenderer.php');
		return  MultiFactorAuthFormRenderer::render($login_message);
	}

	/**
	 * @param \User $user
	 * @param Logger $logger
	 * @param string $code
	 *
	 * @return bool
	 */
	public function verifyCode(\User $user, Logger $logger, $code) {
		$multiFA = $this->initAuthenticator($user, $logger);
		return $multiFA->isCodeValid($code);
	}

	/**
	 * @param \User $user
	 * @param Logger $logger
	 *
	 * @return MultiFactorAuthenticator
	 */
	private function initAuthenticator(\User $user, Logger $logger) {
		require_once('include/MultiFA/MultiFactorAuthenticator.php');
		return new MultiFactorAuthenticator($user, $logger);
	}
}