<?php

namespace OneCRM\MultiFA;

include_once('modules/UserPreferences/UserPreference.php');

use Katzgrau\KLogger\Logger;
use RobThree\Auth\TwoFactorAuth;
use RobThree\Auth\TwoFactorAuthException;

class MultiFactorAuthenticator {

	const NUMBER_OF_RECOVERY_CODES = 5;

	/**
	 * @var TwoFactorAuth
	 */
	private $object;

	/**
	 * @var Logger
	 */
	private $logger;

	/**
	 * The number of seconds a code will be valid
	 *
	 * @var int
	 */
	private $period;

	/**
	 * @var \User
	 */
	private $user;

	/**
	 * Set the number of seconds a code will be valid
	 *
	 * @param int $period
	 */
	public function setPeriod($period) {
		$this->period = $period;
	}

	/**
	 * Authenticator constructor.
	 *
	 * @param \User $user
	 * @param Logger $logger
	 */
	public function __construct(\User $user, Logger $logger) {
		$this->logger = $logger;
		$this->user = $user;
		$this->period = 30;

		try {
			$this->object = new TwoFactorAuth('1CRM Systems Corp.', 6, $this->period);
		} catch (TwoFactorAuthException $e) {
			$this->logError($e->getMessage());
		}
	}

	/**
	 * Is multi factor auth enabled or not
	 *
	 * @return bool
	 */
	public function isEnabled() {
		return (bool)$this->user->getPreference('multi_fa_enabled', 'global');
	}

	/**
	 * Enable multi factor authentication
	 */
	public function enable() {
		$this->user->setPreference('multi_fa_enabled', true);
		$this->user->savePreferencesToDB();
	}

	/**
	 * Disable multi factor authentication
	 */
	public function disable() {
		$this->user->setPreference('multi_fa_enabled', false);
		$this->user->setPreference('mfa_secret', '');
		$this->user->setPreference('mfa_recovery_codes', '');
		$this->user->savePreferencesToDB();
	}

	/**
	 * Get secret (generate new one if not exists)
	 *
	 * @param bool $recreate - recreate secret code
	 *
	 * @return string
	 */
	public function getSecret($recreate = false) {
		$secret = $this->user->getPreference('mfa_secret');

		if (! $secret || $recreate) {
			$secret = $this->generateSecret();

			$this->user->setPreference('mfa_secret', $secret);
			$this->user->savePreferencesToDB();
		}

		return $secret;
	}

	/**
	 * Calculate the code with given secret
	 *
	 * @return string
	 */
	public function getCode() {
		return $this->object->getCode($this->getSecret());
	}

	/**
	 * Get URI of Auth QRCode
	 *
	 * @return null|string
	 */
	public function getQRCodeUri() {
		try {
			return $this->object->getQRCodeImageAsDataUri('1CRM Systems Corp.', $this->getSecret());
		} catch (TwoFactorAuthException $e) {
			$this->logError($e->getMessage());

			return null;
		}
	}

	/**
	 * Check if the code is correct
	 *
	 * @param string $code
	 *
	 * @return bool
	 */
	public function isCodeValid($code) {
		if (is_numeric($code)) {
			return $this->object->verifyCode($this->getSecret(), $code);
		} else {
			$recovery_codes = $this->getRecoveryCodes();

			return in_array($code, $recovery_codes);
		}
	}

	/**
	 * Recovery codes are used to access user's account in the event
	 * user cannot receive two-factor authentication codes
	 *
	 * @return array
	 */
	public function generateRecoveryCodes() {
		$codes = [];
		$this->refreshSecret();

		try {

			for ($i = 0; $i < self::NUMBER_OF_RECOVERY_CODES; $i++) {
				$bytes = $this->object->getRngprovider()->getRandomBytes(4);
				$codes[] = bin2hex($bytes);
			}

			$this->storeRecoveryCodes($codes);

		} catch (TwoFactorAuthException $e) {
			$this->logError($e->getMessage());
		}

		return $codes;
	}

	/**
	 * @return array
	 */
	public function getRecoveryCodes() {
		$codes = $this->user->getPreference('mfa_recovery_codes', 'global');
		$decoded = base64_decode($codes);
		$decoded = unserialize($decoded);

		$result = null;
		$hash = $this->getHash($this->getSecret());

		if (is_array($decoded) && isset($decoded[$hash])) {
			$result = $decoded[$hash];
		}

		if (! is_array($result))
			$result = [];

		return $result;
	}

	/**
	 * @param array $codes
	 */
	private function storeRecoveryCodes($codes) {
		$hash = $this->getHash($this->getSecret());
		$list[$hash] = $codes;
		$encoded = base64_encode(serialize($list));

		$this->user->setPreference('mfa_recovery_codes', $encoded);
		$this->user->savePreferencesToDB();
	}

	/**
	 * Generate a string hash
	 *
	 * @param string $string
	 *
	 * @return string
	 */
	private function getHash($string) {
		return hash_hmac('sha256', $string, $this->getSecret());
	}

	/**
	 * Generate secret for the auth code
	 *
	 * @return string
	 */
	private function generateSecret() {
		$secret = '';

		try {
			$secret = $this->object->createSecret();
		} catch (TwoFactorAuthException $e) {
			$this->logError($e->getMessage());
		}

		return $secret;
	}

	/**
	 * Refresh secret
	 */
	private function refreshSecret() {
		$this->getSecret(true);
	}

	/**
	 * Logging error message
	 *
	 * @param string $message
	 */
	private function logError($message) {
		$this->logger->error('Two-Factor Auth error: ' . $message);
	}
}
