<?php

namespace OneCRM\OAuth;

use League\OAuth2\Client\Provider\AbstractProvider;
use League\OAuth2\Client\Provider\Exception\IdentityProviderException;
use League\OAuth2\Client\Token\AccessToken;
use Psr\Http\Message\ResponseInterface;

use \AppConfig;

require_once 'include/OAuth/OAuthTokenOwner.php';

class OAuthProvider extends AbstractProvider {

	protected $tokenOwner = null;
	protected $service;

	public function __construct(array $options = [], array $collaborators = []) {
		$redirectUri = AppConfig::site_url() . 'oauth_redirect.php';
		switch ($options['service']) {
		case 'o365':
			$clientId = AppConfig::setting('company.oauth_microsoft_id');
			$clientSecret = AppConfig::setting('company.oauth_microsoft_secret');
			break;
		case 'gmail':
			$clientId = AppConfig::setting('company.oauth_gmail_id');
			$clientSecret = AppConfig::setting('company.oauth_gmail_secret');
			break;
		case 'google_sync':
			$clientId = AppConfig::setting('company.oauth_google_id');
			$clientSecret = AppConfig::setting('company.oauth_google_secret');
			break;
		}
		$options = array_merge(
			[
            	'clientId' => $clientId,
				'clientSecret' => $clientSecret,
				'redirectUri' => $redirectUri,
			],
			$options
		);
		parent::__construct($options);
		if (isset($options['tokenOwner'])) {
			$this->tokenOwner  = new OAuthTokenOwner($options['tokenOwner']);
		}
	}

	public function getBaseAuthorizationUrl() {
		switch ($this->service) {
		case "o365":
			$tenantId = $this->microsoftTenantId();
			return 'https://login.microsoftonline.com/' . $tenantId  .'/oauth2/v2.0/authorize';
		case "gmail":
		case "google_sync":
			return "https://accounts.google.com/o/oauth2/auth?access_type=offline";
		}
	
	}

	public function getBaseAccessTokenUrl(array $params) {
		switch ($this->service) {
		case "o365":
			$tenantId = $this->microsoftTenantId();
			return 'https://login.microsoftonline.com/' .$tenantId . '/oauth2/v2.0/token';
		case "gmail":
		case "google_sync":
			return "https://accounts.google.com/o/oauth2/token?access_type=offline";
		}
	}

	public function getResourceOwnerDetailsUrl(AccessToken $token) {
		return '';
	}

    protected function getAuthorizationParameters(array $options) {
		$options = parent::getAuthorizationParameters($options);
		if ($this->service == "gmail") {
			$options["approval_prompt"] = "force";
		}

		return $options;
    }

	protected function microsoftTenantId() {
		$id = AppConfig::setting('company.oauth_microsoft_tenant_id');
		if (!$id) $id = 'common';
		return $id;
	}

	protected function getDefaultScopes() {
		switch ($this->service) {
		case "o365":
			return 'offline_access https://outlook.office.com/SMTP.Send https://outlook.office.com/IMAP.AccessAsUser.All';
		case "gmail":
			return "https://www.googleapis.com/auth/gmail.send https://mail.google.com/";
		case "google_sync":
			return join(" ", [
				\Google\Service\Calendar::CALENDAR,
				\Google\Service\PeopleService::CONTACTS,
				\Google\Service\Drive::DRIVE,
				\Google\Service\Sheets::SPREADSHEETS,
			]);
		}
	}

	protected function checkResponse(ResponseInterface $response, $data) {
		if (empty($data['error_description'])) {
			return;
		}
		$error = $data['error_description'];
		throw new IdentityProviderException($error, 0, $data);
	}

	protected function createResourceOwner(array $response, AccessToken $token) {
	}

	public function getAccessToken($grant, array $options = [], $persist = false) {
		$token = parent::getAccessToken($grant, $options);
		$t = $token->jsonSerialize();

		// Google liraries and League\OAuth2\Client\Token have different logic for
		// token expiration detection. Google considers the token expired if
		// less than 30 secons left to expiration time. League\OAuth2\Client\Token
		// uses exact time
		
		// Add threshold for League
		$t['expires'] -= 30;

		// For Google, shift expiration 5 secends forward, so we can detect expiration
		// before Google library kicks in and tries to refresh the token
		$t['created'] = time();
		$t['expires_in'] = $t['expires'] - $t['created'] + 5; 
		if ($persist)
			$this->persistToken($t);
		return $t;
	}

	private function persistToken(array $token) {
		if ($this->tokenOwner) {
			$this->tokenOwner->persist($token);
		}
	}
}

