<?php

require_once 'include/OnlineMeetings/GoToMeeting/Adapter.php';
require_once 'include/OnlineMeetings/OnlineMeeting.php';

class GoToMeetingException extends Exception{}
class GoToMeetingExceptionInvalidBody extends GoToMeetingException {}

class GoToMeeting extends OnlineMeeting
{
	private $adapter;

	public static function getRedirectUri($for_company = false) {
		$return = AppConfig::site_url() . '/async.php?module=Meetings&action=GTMAuth';
		if ($for_company) 
			$return .= '&system=1';
		$redirect = 'https://info.longreach.net/gtm_auth.php?redirect=' . urlencode($return);
		return $redirect;
	}

	/**
	 * Creates an instance of GoToMeeting class.
	 * Login and password are used to obtain an access token
	 * if it is not passed in, invalid or expired
	 */

	public function __construct($token = null)
	{
		$this->adapter = GoToMeeting_Adapter::create();
		$this->appName = 'gtm';
		parent::__construct($token);
	}

	/**
	 * Obtains a new access token.
	 * GoToMeetingException is thrown if not able to login,
	 * or network error occures.
	 *
	 * @returns a new access token
	 * @throws GoToMeetingException
	 *
	 */
	public function obtainToken($code, $for_company = false)
	{
		$params = array(
			'grant_type' => 'authorization_code',
			'code' => $code,
			'redirect_uri' => self::getRedirectUri($for_company)
		);
		$url = 'https://api.getgo.com/oauth/v2/token';
		$resp = $this->adapter->post($url, $this->getHeaders(false), http_build_query($params));
		$resp = $this->parseResponse($resp);
		$this->setRefreshTimeInfo($resp);
		$this->accessToken = $resp;
		return $resp;
	}

	public function refreshToken() {
		if (!$this->accessToken) {
			throw new GoToMeetingException('No access token set');
		}
		if ($this->accessToken['refresh_before'] < time()) {
			throw new GoToMeetingException('Refresh token expired');
		}
		$params = array(
			'grant_type' => 'refresh_token',
			'refresh_token' => $this->accessToken['refresh_token']
		);
		$url = 'https://api.getgo.com/oauth/v2/token';
		$resp = $this->adapter->post($url, $this->getHeaders(false), http_build_query($params));
		$resp = $this->parseResponse($resp);

		$this->setRefreshTimeInfo($resp);
		$this->accessToken = $resp;
		return $resp;
	}

	public function obtainOrganizerInfo() {
		$url = 'https://api.getgo.com/G2M/rest/organizers/' . $this->accessToken['organizer_key'];
		$resp = $this->adapter->get($url, $this->getHeaders());
		$resp = $this->parseResponse($resp);
		return $resp;
	}

	/**
	 * Creates a meeting.
	 * $params is an associative array with following elements:
	 *   - subject Meeting subject. Required
	 *   - starttime ISO8601 UTC string. Required
	 *   - endtime ISO8601 UTC string. Required
	 *   - meetingtype Immediate, Scheduled or Recurring. Defaults to Scheduled
	 *
	 *   @returns meeting ID on success
	 *   @throws GoToMeetingException
	 */
	public function schedule($params)
	{
		return $this->createOrUpdate(null, $params);
	}

	public function update($id, $params)
	{
		return $this->createOrUpdate($id, $params);
	}

	public function delete($id)
	{
		$url = 'https://api.getgo.com/G2M/rest/meetings/' . $id;
		$resp = $this->adapter->delete($url, $this->getHeaders());
		$resp = $this->parseResponse($resp);
		return $resp;
	}

	private function createOrUpdate($id, $params)
	{
		$json = getJSONObj();
		$url = 'https://api.getgo.com/G2M/rest/meetings';
		if ($id)
			$url .= '/' . $id;

		if (!isset($params['meetingtype']))
			$params['meetingtype'] = 'Scheduled';
		$params['passwordrequired'] = 'false';
		$params['conferencecallinfo'] = 'Free';
		$params['timezonekey'] = '';
		$params['conferencecallinfo'] = 'voip';
		if (!$id)
			$resp = $this->adapter->post($url, $this->getHeaders(), $json->encode($params));
		else
			$resp = $this->adapter->put($url, $this->getHeaders(), $json->encode($params));
		$resp = $this->parseResponse($resp);
		return $id? $id : $resp[0]['meetingid'];
	}

	/**
	 * Get join to meeting URL
	 *
	 * @param array $params
	 *
	 * @return string
	 */
	public function getJoinUrl($params) {
		$gtm_id = $params['gtm_id'];
		$url = '';

		if($gtm_id && AppConfig::setting('site.features.gotomeeting.enabled')) {
			$base_url = AppConfig::setting('site.features.gotomeeting.meeting_url');

			if($base_url)
				$url = str_replace('{ID}', $gtm_id, $base_url);
		}

		return $url;
	}

	/**
	 * Get online meeting join URL label
	 *
	 * @return string
	 */
	public function getJoinUrlLabel()
	{
		return 'LBL_GTM_JOIN_URL';
	}

	/**
	 * Get create online meeting checkbox label
	 *
	 * @return string
	 */
	public function getCreateLabel()
	{
		return translate('LBL_GTM_CREATE_SHORT', 'Meetings');
	}

	public function setAccessToken($token)
	{
		if (is_string($token)) {
			$token = unserialize($token);
		}
		$this->accessToken = $token;
	}

	public static function getApiKey()
	{
		return AppConfig::setting('company.gtm_api_key');
	}

	/**
	 * Returns access token, either set in constructor,
	 * or created by obtainToken
	 */
	public function getAccessToken()
	{
		return $this->accessToken;
	}

	/**
	 * Set refresh time for API access token
	 *
	 * @param array $token
	 */
	protected function setRefreshTimeInfo(&$token) {
		$now = time();
		// Access token is normally valid for 1 hour.
		// To be safe, we consider that refresh is required
		// in 50 minutes.
		$token['refresh_after'] = $now + $token['expires_in'] - 10 * 60;
		// Refresh token is valid for 30 days.
		// To be safe, we consider it valid for one hour less.
		$token['refresh_before'] = $now + 30 * 24 * 3600 - 3600;
	}

	/**
	 * Returns an array of HTTP headers to be sent in request
	 */
	private function getHeaders($includeToken = true)
	{
		$CLIENT_ID = 'vnfNX7sCYgMVSEbSwJouyMA1PUYZHbiB';
		$SECRET = 'ObsGA8nlE81G0GDb';
		$headers = array(
			"HTTP/1.1",
			"Accept: application/json",
		);
		if ($includeToken) {
			$headers[] = "Content-type: application/json";
		}
		if ($includeToken && $this->accessToken) {
			$headers[]  = "Authorization: OAuth oauth_token=". $this->accessToken['access_token'];
		} elseif (!$includeToken) {
			$headers[]  = "Authorization: Basic " . base64_encode($CLIENT_ID . ':' . $SECRET);
		}
		return $headers;
	}

	private function makeQueryString(array $params)
	{
		$parts = array();
		foreach ($params as $name => $value) {
			$parts[] = urlencode($name) . '=' . urlencode($value);
		}
		return join('&', $parts);
	}

	private function parseResponse($resp)
	{
		if ($resp == 'Invalid request body') {
			throw new GoToMeetingExceptionInvalidBody($resp);
		}
		$json = getJSONObj();
		$respData = $json->decode($resp);
		if (isset($respData['int_err_code'])) {
			throw new GoToMeetingException($respData['msg']);
		}
		if (isset($respData['err'])) {
			throw new GoToMeetingException($respData['message']);
		}
		if (isset($respData['error'])) {
			throw new GoToMeetingException($respData['error']);
		}
		return $respData;
	}
}

