<?php

use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Client;
use Microsoft\Graph\Graph;
use Microsoft\Graph\Model\OnlineMeeting as MeetingModel;

require_once 'include/OnlineMeetings/OnlineMeeting.php';

class MsTeams extends OnlineMeeting {

	const LOGIN_URL = 'https://login.microsoftonline.com/common/oauth2/v2.0/';

	/**
	 * OnlineMeeting constructor.
	 *
	 * @param array|null $token
	 */
	public function __construct($token = null) {
		$this->appName = 'ms_teams';
		parent::__construct($token);
	}

	/**
	 * Get Redirect URI for access token obtaining
	 *
	 * @param bool $for_company
	 *
	 * @return string
	 */
	public static function getRedirectUri($for_company = false) {
		$url = AppConfig::site_url() . 'OnlineMeetingAuth.php';
		//$url = 'http://localhost/1crm/temp.php';

		return $url;
	}

	/**
	 * Get URL for requesting authorization code
	 *
	 * @param bool $for_company
	 *
	 * @return string
	 */
	public function getAuthCodeRequestUrl($for_company = false) {
		$self = new self();
		$client_id = $self->getClientId();
		$url = '';

		if ($client_id) {
			$return = self::getRedirectUri();
			$url = $self::LOGIN_URL . 'authorize';
			$url .= "?client_id={$client_id}&response_type=code&redirect_uri={$return}&response_mode=query&scope=offline_access%20OnlineMeetings.ReadWrite";

			if ($for_company)
				$url .= '&state=company';
		}

		return $url;
	}

	/**
	 * Schedule a meeting
	 *
	 * @param array $params
	 *
	 * @throws OnlineMeetingException
	 * @return mixed
	 */
	public function schedule($params) {
		if (! isset($this->accessToken['access_token'])) {
			$GLOBALS['log']->error('Empty MS Teams token');
			return null;
		}

		$body = [
			'startDateTime' => gmdate('Y-m-d\TH:i:s', $params['start']),
			'endDateTime' => gmdate('Y-m-d\TH:i:s', $params['end']),
			'externalId' => $params['crm_id'],
			'subject' => $params['subject']
		];

		$graph = new Graph;
		$graph->setAccessToken($this->accessToken['access_token']);

		try {
			$result = $graph->createRequest("POST", "/me/onlineMeetings/createOrGet")
				->setReturnType(MeetingModel::class)
				->attachBody($body)
				->execute();
		} catch (Exception $e) {
			$GLOBALS['log']->error('Error during MS Teams meeting scheduling: ' . $e->getMessage());
			throw new OnlineMeetingException($e->getMessage());
		}

		return $result;
	}

	/**
	 * Update existent meeting
	 *
	 * @param string $id
	 * @param array $params
	 *
	 * @throws OnlineMeetingException
	 * @return mixed
	 */
	public function update($id, $params) {
		try {
			$this->delete($id);
		} catch (OnlineMeetingException $e) {
			$GLOBALS['log']->error($e->getMessage());
		}

		return $this->schedule($params);
	}

	/**
	 * Delete meeting
	 *
	 * @param string $id
	 *
	 * @throws OnlineMeetingException
	 * @return bool
	 */
	public function delete($id) {
		if (! isset($this->accessToken['access_token'])) {
			$GLOBALS['log']->error('Empty MS Teams token');
			return false;
		}

		$graph = new Graph();
		$graph->setAccessToken($this->accessToken['access_token']);

		try {
			$graph->createRequest("DELETE", "/me/onlineMeetings/{$id}")
				->execute();
		} catch (Exception $e) {
			$GLOBALS['log']->error('Error during MS Teams meeting deleting: ' . $e->getMessage());
			throw new OnlineMeetingException($e->getMessage());
		}

		return true;
	}

	/**
	 * Get join to meeting URL
	 *
	 * @param array $params
	 *
	 * @return string
	 */
	public function getJoinUrl($params) {
		$url = '';

		if (isset($params['join_url']))
			$url = $params['join_url'];

		return $url;
	}

	/**
	 * Get online meeting join URL label
	 *
	 * @return string
	 */
	public function getJoinUrlLabel() {
		return 'LBL_MS_TEAMS_JOIN_URL';
	}

	/**
	 * Get create online meeting checkbox label
	 *
	 * @return string
	 */
	public function getCreateLabel() {
		return translate('LBL_MS_TEAMS_CREATE', 'Meetings');
	}

	/**
	 * Obtain API access token
	 *
	 * @param string $code
	 * @param bool $for_company
	 * @throws OnlineMeetingException
	 *
	 * @return array
	 */
	public function obtainToken($code, $for_company = false) {
		$guzzle = new Client;

		try {
			$token = json_decode($guzzle->post(self::LOGIN_URL . 'token', [
				'form_params' => [
					'client_id' => $this->getClientId(),
					'scope' => 'OnlineMeetings.ReadWrite offline_access',
					'code' => $code,
					'redirect_uri' => self::getRedirectUri(),
					'client_secret' => $this->getClientSecret(),
					'grant_type' => 'authorization_code'
				],
			])->getBody()->getContents());

		} catch (Exception $e) {
			throw new OnlineMeetingException($this->getExceptionMessage($e));
		}

		$token = (array)$token;
		$this->setRefreshTimeInfo($token);

		return $token;
	}

	/**
	 * Refresh API access token
	 *
	 * @return array
	 * @throws OnlineMeetingException
	 */
	public function refreshToken() {
		if (! $this->accessToken) {
			throw new OnlineMeetingException('No MS Graph access token set');
		}
		if ($this->accessToken['refresh_before'] < time()) {
			throw new OnlineMeetingException('MS Graph refresh token expired');
		}

		$guzzle = new Client();

		try {
			$token = json_decode($guzzle->post(self::LOGIN_URL . 'token', [
				'form_params' => [
					'grant_type' => 'refresh_token',
					'client_id' => $this->getClientId(),
					'client_secret' => $this->getClientSecret(),
					'refresh_token' => $this->accessToken['refresh_token']
				],
			])->getBody()->getContents());

		} catch (Exception $e) {
			throw new OnlineMeetingException($this->getExceptionMessage($e));
		}

		$token = (array)$token;
		$this->setRefreshTimeInfo($token);

		return $token;
	}

	protected function getErrorResponseMessageField() {
		return 'error_description';
	}

	protected function getUnknowErrorMessage() {
		return 'Unknown error during MS Graph request';
	}


	/**
	 * Get Azure App Client ID
	 *
	 * @return string
	 */
	private function getClientId() {
		return AppConfig::setting('company.ms_teams_client_id');
	}

	/**
	 * Get Azure App Client Secret
	 *
	 * @return string
	 */
	private function getClientSecret() {
		return AppConfig::setting('company.ms_teams_client_secret');
	}
}
