<?php

use GuzzleHttp\Exception\RequestException;

abstract class OnlineMeeting {

	/**
	 * @var array
	 */
	protected $accessToken;

	/**
	 * @var string
	 */
	protected $appName;

	/**
	 * @return array
	 */
	public function getAccessToken() {
		return $this->accessToken;
	}

	/**
	 * @param array $accessToken
	 */
	public function setAccessToken($accessToken) {
		$this->accessToken = $accessToken;
	}

	/**
	 * @return string
	 */
	public function getAppName() {
		return $this->appName;
	}

	/**
	 * OnlineMeeting constructor.
	 *
	 * @param array|null $token
	 */
	public function __construct($token = null) {
		$this->setAccessToken($token);
	}

	/**
	 * Get Redirect URI for access token obtaining
	 *
	 * @param bool $for_company
	 *
	 * @return string
	 */
	public abstract static function getRedirectUri($for_company = false);

	/**
	 * Schedule a meeting
	 *
	 * @param array $params
	 *
	 * @throws OnlineMeetingException
	 * @return mixed
	 */
	public abstract function schedule($params);

	/**
	 * Update existent meeting
	 *
	 * @param string $id
	 * @param array $params
	 *
	 * @throws OnlineMeetingException
	 * @return mixed
	 */
	public abstract function update($id, $params);

	/**
	 * Delete meeting
	 *
	 * @param string $id
	 *
	 * @throws OnlineMeetingException
	 * @return mixed
	 */
	public abstract function delete($id);

	/**
	 * Get join to meeting URL
	 *
	 * @param array $params
	 *
	 * @return string
	 */
	public abstract function getJoinUrl($params);

	/**
	 * Get online meeting join URL label
	 *
	 * @return string
	 */
	public abstract function getJoinUrlLabel();

	/**
	 * Get create online meeting checkbox label
	 *
	 * @return string
	 */
	public abstract function getCreateLabel();

	/**
	 * Obtain API access token
	 *
	 * @param string $code
	 * @param bool $for_company
	 * @throws GoToMeetingException|OnlineMeetingException
	 *
	 * @return array
	 */
	public abstract function obtainToken($code, $for_company = false);

	/**
	 * Refresh API access token
	 *
	 * @return array
	 */
	public abstract function refreshToken();

	/**
	 * Load previously stored API access token
	 *
	 * @param bool $for_company
	 * @param string $user_id
	 *
	 * @return array
	 */
	public function loadStoredAccessToken($for_company = false, $user_id = null) {
		if ($for_company) {
			$accessToken = @unserialize(AppConfig::setting("company.{$this->appName}_auth_token"));
		} else {
			require_once 'modules/UserPreferences/UserPreference.php';
			$accessToken = UserPreference::getPreference("{$this->appName}_token", 'global', $user_id);
		}

		return $accessToken;
	}

	/**
	 * Save API access token
	 *
	 * @param array|string $token
	 * @param bool $for_company
	 * @param string|null $user_id
	 */
	public function saveAccessToken($token, $for_company = false, $user_id = null) {
		if ($for_company) {
			$value = $token;

			if (is_array($value))
				$value = serialize($value);

			AppConfig::set_local("company.{$this->appName}_auth_token", $value, true);
			AppConfig::save_local('company');
		} else {
			require_once 'modules/UserPreferences/UserPreference.php';
			$user = null;

			if ($user_id) {
				require_bean('User');
				$user = new User();
				$user->id = $user_id;
			}

			UserPreference::setPreference("{$this->appName}_token", $token, 0, 'global', $user);
			UserPreference::savePreferencesToDB();
		}
	}

	/**
	 * Execute this method when saving API/Auth settings
	 *
	 * @param $params
	 * @throws OnlineMeetingException
	 *
	 * @return bool
	 */
	public function onSaveSettingsEvent($params) {
		return false;
	}

	/**
	 * Reset API access token
	 *
	 * @param bool $for_company
	 * @param string|null $user_id
	 */
	public function resetAccessToken($for_company = false, $user_id = null) {
		$this->saveAccessToken('', $for_company, $user_id);
	}

	/**
	 * Decode stored meeting params
	 *
	 * @param string $encoded_params - JSON encoded meeting params
	 *
	 * @return array
	 */
	public function decodeStoredParams($encoded_params) {
		$app = AppConfig::setting('company.online_meeting_app');
		$params = json_decode($encoded_params);

		if (is_object($params))
			$params = (array)$params;

		if (isset($params['extra_data']))
			$params['extra_data'] = (array)$params['extra_data'];

		if (! is_array($params))
			$params = [];

		$saved_for_app = $params['app'] ?? 'ms_teams';

		if ($app == $saved_for_app) {
			return $params;
		} else {
			return [];
		}
	}

	/**
	 * Set refresh time for API access token
	 *
	 * @param array $token
	 */
	protected function setRefreshTimeInfo(&$token) {
		$now = time();
		// Access token is normally valid for 1 hour.
		// To be safe, we consider that refresh is required
		// in 50 minutes.
		$token['refresh_after'] = $token['expires_in'] - 10 * 60;
		// Refresh token is valid for 30 days.
		// To be safe, we consider it valid for one hour less.
		$token['refresh_before'] = $now + 30 * 24 * 3600 - 3600;
	}

	protected function getErrorResponseMessageField() {
		return '';
	}

	protected function getUnknowErrorMessage() {
		return '';
	}
	
	protected function getExceptionMessage(Exception $e) {
		if ($e instanceof RequestException && $e->hasResponse()) {
			$error = $e->getResponse()->getBody()->getContents();
			$error = json_decode($error, true);
			$field = $this->getErrorResponseMessageField();
			if (is_array($error) && isset($error[$field])) {
				$message = $error[$field];
			} else {
				$message = $this->getUnknowErrorMessage();
			}
		} else {
			$message = $e->getMessage();
		}
		return $message;
	}


}