<?php
require_once('OnlineMeetingsFactory.php');

class OnlineMeetingProcess {

	/**
	 * @var OnlineMeeting
	 */
	private $onlineMeeting;

	/**
	 * OnlineMeetingProcess constructor.
	 */
	public function __construct() {
		try {
			$this->onlineMeeting = OnlineMeetingsFactory::getInstance();
		} catch (OnlineMeetingException $e) {}
	}

	/**
	 * Prepare parameters:
	 *
	 *	delete existing online meeting? This happens when assigned
	 *	user has changed, and existing  online meeting was created with
	 *	that users online meeting account, not system account.
	 *	Also, delete online meeting if user unchecked "create" checkbox
	 *
	 *	create new online meeting? This happens when online meeting was
	 *	not created earlier, or it was deleted because of assigned
	 *	user change
	 *
	 * @param RowUpdate $upd - meeting update
	 *
	 * @return bool
	 * @throws IAHActionAbort
	 */
	public function run(RowUpdate &$upd) {
		if (! $this->onlineMeeting)
			return false;

		$do_schedule = $upd->getRelatedData('online_meeting_create');
		$prev_user_id = $upd->getField('assigned_user_id', null, true);
		$user_id = $upd->getField('assigned_user_id');
		$user_changed = $prev_user_id && ($prev_user_id != $user_id);

		$system_token = $this->onlineMeeting->loadStoredAccessToken(true);

		if ($this->onlineMeeting->getAppName() == 'gtm') {
			$this->processGoToMeeting($upd, $system_token, $user_id, $prev_user_id, $user_changed, $do_schedule);
		} else {
			$prev_params = $this->onlineMeeting->decodeStoredParams($upd->getField('online_meeting_params', null, true));
			$do_delete = ! empty($prev_params) && ( ! $do_schedule || $user_changed);

			if ($do_delete) {

				if (isset($prev_params['type']) && $prev_params['type'] == 'system') {
					$this->onlineMeeting->setAccessToken($system_token);
				} else {
					$user_token = $this->onlineMeeting->loadStoredAccessToken(false, $prev_user_id);
					$this->onlineMeeting->setAccessToken($user_token);
				}

				if ($this->onlineMeeting->getAccessToken()) {
					try {
						$this->onlineMeeting->delete($prev_params['id']);
						$upd->set('online_meeting_params', '');
					} catch (OnlineMeetingException $e) {
						$GLOBALS['log']->error($e->getMessage());
						add_flash_message(translate('LBL_ONLINE_MEETING_DELETE_ERROR', 'Meetings') . '. ' . $e->getMessage(), 'warning');
					}
				}
			}

			if ($do_schedule) {
				$user_token = $this->onlineMeeting->loadStoredAccessToken(false, $user_id);
				$system = false;

				if ( ! empty($user_token)) {
					$token = $user_token;
				} else {
					$token = $system_token;
					$system = true;
				}

				if ($token) {
					$this->onlineMeeting->setAccessToken($token);

					try {
						$start = strtotime($upd->getField('date_start') . ' GMT');


						if ($upd->getField('is_daylong')) {
							$duration = 60;
						} else {
							$duration = $upd->getField('duration');
						}

						$end = $start + $duration * 60;

						$hash = $upd->getField('name') . $upd->getField('assigned_user_id');
						$hash .= $start . $end . gmdate('Y-m-d');
						$hash = md5($hash);

						$params = [
							'subject' => $upd->getField('name'),
							'start' => $start,
							'end' => $end,
							'duration' => $duration,
							'crm_id' => $hash
						];

						if (empty($prev_params) || $do_delete) {
							$meeting = $this->onlineMeeting->schedule($params);
							$update_params = true;
							$show_error = true;
						} else {
							$meeting = $this->onlineMeeting->update($prev_params['id'], $params);
							$update_params = is_object($meeting);

							if (is_bool($meeting) && $meeting === true) {
								$show_error = false;
							} else {
								$show_error = true;
							}
						}

						if ($show_error) {
							if ( ! $meeting || ! is_object($meeting) || ! method_exists($meeting, 'getId')) {
								throw new IAHActionAbort(translate('LBL_ONLINE_MEETING_ERROR_INVALID_BODY', 'Meetings'));
							}
						}

						if ($update_params) {
							if (method_exists($meeting, 'getJoinUrl')) {
								$joinUrl = $meeting->getJoinUrl();
							} elseif (method_exists($meeting, 'getJoinWebUrl')) {
								$joinUrl = $meeting->getJoinWebUrl();
							} else {
								throw new IAHActionAbort(translate('LBL_ONLINE_MEETING_ERROR_INVALID_BODY', 'Meetings'));
							}

							$meeting_params = [
								'type' => $system ? 'system' : 'user',
								'app' => $this->onlineMeeting->getAppName(),
								'id' => $meeting->getId(),
								'join_url' => $joinUrl
							];

							if (method_exists($meeting, 'getExtraData')) {
								$meeting_params['extra_data'] = $meeting->getExtraData();
							}

							$upd->set('online_meeting_params', json_encode($meeting_params));
						}

					} catch (OnlineMeetingException $e) {
						$GLOBALS['log']->error($e->getMessage());
						throw new IAHActionAbort(translate('LBL_ONLINE_MEETING_ERROR', 'Meetings') . '. '
							. $e->getMessage());
					}
				} else {
					add_flash_message(translate('LBL_ONLINE_MEETING__MISSING_ACCOUNT', 'Meetings'), 'warning');
				}
			}

		}

		return true;
	}

	/**
	 * @param RowUpdate $upd
	 * @param array $system_token
	 * @param string $user_id
	 * @param string $prev_user_id
	 * @param bool $user_changed
	 * @param bool $do_schedule
	 *
	 * @return bool
	 * @throws IAHActionAbort
	 */
	private function processGoToMeeting(RowUpdate &$upd, $system_token, $user_id, $prev_user_id, $user_changed, $do_schedule) {
		if (! AppConfig::setting('site.features.gotomeeting.enabled'))
			return false;

		$system_gtm_org = array_get_default($system_token, 'organizer_key');
		$prev_gtm_org = $upd->getField('gtm_org_id', null, true);
		$do_delete = $prev_gtm_org && (!$do_schedule || ($user_changed && $prev_gtm_org != $system_gtm_org));

		if ($do_delete) {

			if ($prev_gtm_org == $system_gtm_org) {
				$this->onlineMeeting->setAccessToken($system_token);
			} else {
				$user_token = $this->onlineMeeting->loadStoredAccessToken(false, $prev_user_id);
				$this->onlineMeeting->setAccessToken($user_token);
			}

			try {
				$this->onlineMeeting->delete($upd->getField('gtm_id', null, true));

				$upd->set('gtm_id', '');
				$upd->set('gtm_email', '');
				$upd->set('gtm_org_id', '');
			} catch (GoToMeetingException $e) {
				$GLOBALS['log']->error($e->getMessage());
				add_flash_message(translate('LBL_GTM_DELETE_ERROR', 'Meetings') . '. ' . $e->getMessage(), 'warning');
			}
		}

		if ($do_schedule) {
			$user_token = $this->onlineMeeting->loadStoredAccessToken(false, $user_id);

			if (! empty($user_token)) {
				$token = $user_token;
			} else {
				$token = $system_token;
			}

			if ($token) {
				$this->onlineMeeting->setAccessToken($token);

				try {
					$start = strtotime($upd->getField('date_start') . ' GMT');

					if ($upd->getField('is_daylong')) {
						$end = $start + 60 * 60;
					} else {
						$end = $start + $upd->getField('duration') * 60;
					}

					$params = array(
						'subject' => $upd->getField('name'),
						'starttime' => gmdate('Y-m-d\TH:i:s', $start),
						'endtime' => gmdate('Y-m-d\TH:i:s', $end),
					);

					if (!$upd->getField('gtm_id', null, true)) {
						$gtm_id = $this->onlineMeeting->schedule($params);
					} else {
						$gtm_id = $upd->getField('gtm_id', null, true);
						$this->onlineMeeting->update($gtm_id, $params);
					}

					$upd->set('gtm_id', $gtm_id);
					$upd->set('gtm_email', $token['organizer_email']);
					$upd->set('gtm_org_id', $token['organizer_key']);

				} catch (GoToMeetingException $e) {
					$GLOBALS['log']->error($e->getMessage());
					if ($e instanceof GoToMeetingExceptionInvalidBody) {
						throw new IAHActionAbort(translate('LBL_ONLINE_MEETING_ERROR_INVALID_BODY', 'Meetings'));
					}
					throw new IAHActionAbort(translate('LBL_GTM_ERROR', 'Meetings') . '. ' . $e->getMessage());
				}
			} else {
				add_flash_message(translate('LBL_GTM_MISSING_ACCOUNT', 'Meetings'), 'warning');
			}
		}

		return true;
	}
}
