<?php

use GuzzleHttp\Client;

require_once 'include/OnlineMeetings/OnlineMeeting.php';
//Use libs from Microsoft Graph or avoiding duplicates
//require_once 'include/OnlineMeetings/MicrosoftGraph/vendor/autoload.php';

class Zoom extends OnlineMeeting {

	const BASE_API_URI = 'https://api.zoom.us/v2/';

	/**
	 * OnlineMeeting constructor.
	 *
	 * @param array|null $token
	 */
	public function __construct($token = null) {
		$this->appName = 'zoom';
		parent::__construct($token);
	}

	/**
	 * Get Redirect URI for access token obtaining
	 *
	 * @param bool $for_company
	 *
	 * @return string
	 */
	public static function getRedirectUri($for_company = false) {
		return '';
	}

	/**
	 * Schedule a meeting
	 *
	 * @param array $params
	 *
	 * @throws OnlineMeetingException
	 * @return mixed
	 */
	public function schedule($params) {
		if (! isset($this->accessToken['access_token'])) {
			$GLOBALS['log']->error('Empty Zoom token');
			return null;
		}

		$guzzle = new Client(['base_uri' => self::BASE_API_URI]);

		$body = [
			'topic' => $params['subject'],
			'type' => 2,
			'start_time' => gmdate('Y-m-d\TH:i:s', $params['start']) . 'Z',
			'duration' => $params['duration']
		];

		try {
			$result = json_decode($guzzle->request('POST', "users/me/meetings", [
				"json" => $body,
				'headers' => [
					'User-Agent' => 'Zoom-api-StoS-Request',
					"Authorization" => "Bearer " . $this->accessToken['access_token'],
					"Content-type" => "application/json"
				]
			])->getBody()->getContents());


		} catch (Exception $e) {
			$error = $this->getExceptionMessage($e);
			$GLOBALS['log']->error('Error during Zoom meeting scheduling: ' . $error);
			throw new OnlineMeetingException(translate('LBL_ZOOM_API_CALL_ERROR', 'Configurator') .': '. $error);
		}

		return $this->wrapResult($result);
	}

	/**
	 * Update existent meeting
	 *
	 * @param string $id
	 * @param array $params
	 *
	 * @throws OnlineMeetingException
	 * @return mixed
	 */
	public function update($id, $params) {
		if (! isset($this->accessToken['access_token'])) {
			$GLOBALS['log']->error('Empty Zoom token');
			return null;
		}

		$guzzle = new Client(['base_uri' => self::BASE_API_URI]);

		$body = [
			'topic' => $params['subject'],
			'type' => 2,
			'start_time' => gmdate('Y-m-d\TH:i:s', $params['start']) . 'Z',
			'duration' => $params['duration']
		];

		try {
			$result = $guzzle->request('PATCH', "meetings/{$id}", [
				"json" => $body,
				'headers' => [
					'User-Agent' => 'Zoom-api-StoS-Request',
					"Authorization" => "Bearer " . $this->accessToken['access_token'],
					"Content-type" => "application/json"
				]
			])->getStatusCode();


		} catch (Exception $e) {
			$error = $this->getExceptionMessage($e);
			$GLOBALS['log']->error('Error during Zoom meeting scheduling: ' . $error);
			throw new OnlineMeetingException(translate('LBL_ZOOM_API_CALL_ERROR', 'Configurator') .': '. $error);
		}

		if ($result == 204) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Delete meeting
	 *
	 * @param string $id
	 *
	 * @throws OnlineMeetingException
	 * @return bool
	 */
	public function delete($id) {
		if (! isset($this->accessToken['access_token'])) {
			$GLOBALS['log']->error('Empty Zoom token');
			return null;
		}

		$guzzle = new Client(['base_uri' => self::BASE_API_URI]);

		try {
			$result = $guzzle->request('DELETE', "meetings/{$id}?schedule_for_reminder=true", [
				'headers' => [
					'User-Agent' => 'Zoom-api-StoS-Request',
					"Authorization" => "Bearer " . $this->accessToken['access_token'],
					"Content-type" => "application/json"
				]
			])->getStatusCode();

		} catch (Exception $e) {
			$error = $this->getExceptionMessage($e);
			$GLOBALS['log']->error('Error during Zoom meeting removing: ' . $error);
			throw new OnlineMeetingException(translate('LBL_ZOOM_API_CALL_ERROR', 'Configurator') .': '. $error);
		}

		if ($result == 204) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Get join to meeting URL
	 *
	 * @param array $params
	 *
	 * @return string
	 */
	public function getJoinUrl($params) {
		$url = '';

		if (isset($params['join_url'])) {
			$url = $params['join_url'];
		}

		return $url;
	}

	/**
	 * Get online meeting join URL label
	 *
	 * @return string
	 */
	public function getJoinUrlLabel() {
		return 'LBL_ZOOM_JOIN_URL';
	}

	/**
	 * Get create online meeting checkbox label
	 *
	 * @return string
	 */
	public function getCreateLabel() {
		return translate('LBL_ZOOM_CREATE', 'Meetings');
	}

	/**
	 * Obtain API access token
	 *
	 * @param string $code
	 * @param bool $for_company
	 *
	 * @throws OnlineMeetingException
	 *
	 * @return array
	 */
	public function obtainToken($code, $for_company = false) {
		$api_key = $this->getApiKey($for_company);
		$api_secret = $this->getApiSecret($for_company);
		$account_id = $this->getAccountId($for_company);

		if (empty($api_key) || empty($api_secret)) {
			throw new OnlineMeetingException('Zoom token obtaining error: empty API Key or Secret');
		}

		return $this->generateStoSToken($api_key, $api_secret, $for_company, $account_id);
	}

	/**
	 *
	 * Refresh API access token
	 *
	 * @return array
	 * @throws OnlineMeetingException
	 */
	public function refreshToken() {
		if (! $this->accessToken) {
			throw new OnlineMeetingException('No MS Graph access token set');
		}
		if ($this->accessToken['refresh_before'] < time()) {
			throw new OnlineMeetingException('MS Graph refresh token expired');
		}

		$for_company = $this->accessToken['type'] == 'system' ? true : false;

		return $this->obtainToken('', $for_company);
	}

	/**
	 * Execute this method when saving API/Auth settings
	 *
	 * @param $params
	 * @throws OnlineMeetingException
	 *
	 * @return bool
	 */
	public function onSaveSettingsEvent($params) {
		if ($params['for_company']) {
			$user_id = null;
			$api_key = $params['company.zoom_api_key'] ?? null;
			$api_secret = $params['company.zoom_api_secret'] ?? null;
			$api_secret = $params['company.zoom_api_secret'] ?? null;
			$account_id = $params['company.zoom_account_id'] ?? null;
		} else {
			$user_id = $params['user_id'] ?? null;
			$api_key = $params['zoom_api_key'] ?? null;
			$api_secret = $params['zoom_api_secret'] ?? null;
			$account_id = $params['zoom_account_id'] ?? null;
		}

		if ($api_key && $api_secret) {
			$token = $this->generateStoSToken($api_key, $api_secret, $params['for_company'], $account_id);

			try {

				$connection_result = $this->testConnection($token['access_token']);

				if (is_object($connection_result) && $connection_result->id) {
					$this->saveAccessToken($token, $params['for_company'], $user_id);
				} else {
					$this->resetAccessToken($params['for_company'], $user_id);
					throw new OnlineMeetingException(translate('LBL_ZOOM_API_RESULT_ERROR', 'Configurator'));
				}

			} catch (OnlineMeetingException $e) {
				$this->resetAccessToken($params['for_company'], $user_id);
				throw new OnlineMeetingException($e->getMessage());
			}

			return true;
		} else {
			$this->resetAccessToken($params['for_company'], $user_id);
		}

		return false;
	}

	/**
	 * Test connection to Zoom JWT Application using token
	 *
	 * @param string $token
	 *
	 * @return mixed
	 * @throws OnlineMeetingException
	 */
	private function testConnection($token) {
		$guzzle = new Client(['base_uri' => self::BASE_API_URI]);

		try {
			$result = json_decode($guzzle->request('GET', "users/me", [
				'headers' => [
					'User-Agent' => 'Zoom-api-StoS-Request',
					"Authorization" => "Bearer " . $token,
					"Content-type" => "application/json"
				]
			])->getBody()->getContents());

		} catch (Exception $e) {
			$error = $this->getExceptionMessage($e);
			$GLOBALS['log']->error('Can not connect to Zoom JWT App: ' . $error);
			throw new OnlineMeetingException(translate('LBL_ZOOM_API_CALL_ERROR', 'Configurator') .': '. $error);
		}

		return $result;
	}

	/**
	 * Generate JSON Web Token
	 *
	 * @param string $api_key
	 * @param string $api_secret
	 * @param bool $for_company
	 *
	 * @return array
	 */
	private function generateStoSToken($api_key, $api_secret, $for_company, $account_id) {
		$guzzle = new Client(['base_uri' => 'https://zoom.us/']);
		$result = json_decode($guzzle->request('POST', "oauth/token?grant_type=account_credentials&account_id=$account_id", [
				'headers' => [
					'User-Agent' => 'Zoom-api-S2S-Request',
					"Authorization" => "Basic " . base64_encode($api_key.':'.$api_secret),
					"client_id" => $api_key,
					"client_secret" => $api_secret,
					"Content-type" => "application/json"
				]
			])->getBody()->getContents());
		$jwt = $result->access_token;

		// Create token payload as a JSON string
		$expires_in = time() + 60 * 60;

		$token = [
			'access_token' => $jwt,
			'expires_in' => $expires_in,
			'refresh_token' => $jwt,
			'type' => $for_company ? 'system' : 'user'
		];

		$this->setRefreshTimeInfo($token);

		return $token;
	}

	/**
	 * Wrap API call result
	 *
	 * @param stdClass $result
	 *
	 * @return null|__anonymous
	 */
	private function wrapResult($result) {
		if (! is_object($result))
			return null;

		$wrapped_object = new class($result) {
			private $wrapped;

			public function __construct(\stdClass $wrapped) {
				$this->wrapped = $wrapped;
			}

			public function getId() {
				return $this->wrapped->id;
			}

			public function getJoinUrl() {
				return $this->wrapped->join_url;
			}

			public function getExtraData() {
				return ['password' => $this->wrapped->password];
			}
		};

		return $wrapped_object;
	}

	/**
	 * Get Zoom API Key
	 *
	 * @param bool $for_company
	 * @param string|null $user_id
	 *
	 * @return string
	 */
	private function getApiKey($for_company = false, $user_id = null) {
		if ($for_company) {
			return AppConfig::setting('company.zoom_api_key');
		} else {
			require_once 'modules/UserPreferences/UserPreference.php';
			return UserPreference::getPreference("zoom_api_key", 'global', $user_id);
		}
	}

	/**
	 * Get Zoom API Key
	 *
	 * @param bool $for_company
	 * @param string|null $user_id
	 *
	 * @return string
	 */
	private function getApiSecret($for_company = false, $user_id = null) {
		if ($for_company) {
			return AppConfig::setting('company.zoom_api_secret');
		} else {
			require_once 'modules/UserPreferences/UserPreference.php';
			return UserPreference::getPreference("zoom_api_secret", 'global', $user_id);
		}
	}

	/**
	 * Get Zoom Account ID
	 *
	 * @param bool $for_company
	 * @param string|null $user_id
	 *
	 * @return string
	 */
	private function getAccountId($for_company = false, $user_id = null) {
		if ($for_company) {
			return AppConfig::setting('company.zoom_account_id');
		} else {
			require_once 'modules/UserPreferences/UserPreference.php';
			return UserPreference::getPreference("zoom_account_id", 'global', $user_id);
		}
	}

	protected function getErrorResponseMessageField() {
		return 'message';
	}

	protected function getUnknowErrorMessage() {
		return 'Unknown error during Zoom request';
	}

}
