<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo, 
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */
class ProgressIndicator
{
	private $fileName;
	private $file;
	private $cli;
	private $silent;

	public function __construct($fileName, $remove = false, $cli = false, $silent=false)
	{
		// absolutize filename for usage within shutdown function, when working directory may have changed
		$dir = realpath('.');
		$this->fileName = $dir . '/' . $fileName;
		
		$this->cli = $cli;
		if ($cli) {
			if(defined('INSTALLER_CLI_OUTPUT'))
				$this->silent = constant('INSTALLER_CLI_OUTPUT') !== 'human';
			else
				$this->silent = $silent;
		}

		if ($remove) {
			@unlink($fileName);
		}
		if ($cli && !$this->silent) {
			echo "\n";
		}
	}
	
	public function getFileName()
	{
		return $this->fileName;
	}

	private function readData()
	{
		clearstatcache(true, $this->fileName);
		$data = null;
		if (file_exists($this->fileName)) {
			$this->file = @fopen($this->fileName, 'r+');
			if($this->file) {
				flock($this->file,  LOCK_EX);
				$sz = filesize($this->fileName);
				if($sz) $data = unserialize(fread($this->file, $sz));
			}
		} else {
			$this->file = @fopen($this->fileName, 'w');
			if($this->file)
				flock($this->file,  LOCK_EX);
		}
		if(! is_array($data)) {
			$data = [
				'count' => 0, 'timestamp' => microtime(true),
				'messages' => [], 'finished' => 0,
				'status' => '', 'status_message' => '',
			];
			$this->writeData($data, false);
		}
		return $data;
	}

	private function writeData($data, $close=true)
	{
		if(! $this->file) return;
		rewind($this->file);
		ftruncate($this->file, 0);
		fwrite($this->file, serialize($data));
		fflush($this->file);
		if($close) {
			flock($this->file,  LOCK_UN);
			fclose($this->file);
			$this->file = null;
		}
	}

	public function logProgress($processName, $progress, $overall = null)
	{
		$data = $this->readData();
		$newProcess = false;
		$progressPC = $progress;
		if (!is_numeric($progress)) {
			$progressPC = 100;
		}
		if (isset($data['current_process'])) {
			if ($data['current_process'] != $processName) {
				$newProcess = true;
				$data['history'][] = [
					'name' => $data['current_process'],
					'progress' => is_numeric($data['current_progress']) ? 100 : $data['current_progress'],
					'count' => $data['count'],
					'messages' => $data['messages'],
				];
				unset($data['current_process'], $data['current_progress']);
				$data['count'] = 0;
				$data['messages'] = [];
			}
		}

		$data['current_process'] = $processName;
		$data['current_progress'] = $progress;
		$data['timestamp'] = microtime(true);
		$data['count']++;

		if (!is_null($overall)) {
			$data['overall_progress'] = $overall;
		}
		if ($this->cli && !$this->silent) {
			if ($newProcess) {
				echo "\n";
			}
			echo "\r\x1B[s\x1B[K";
			if (isset($data['overall_progress'])) {
				printf("%3d%% ---- %3d%% %s",
					$data['overall_progress'],
					$progressPC,
					$data['current_process']
				);
			} else {
				printf("          %3d%% %s",
					$progressPC,
					$data['current_process']
				);
			}
			echo "\x1B[u";
		}
		$this->writeData($data);
	}
	
	public function logMessage($message, $severity='info') {
		$data = $this->readData();
		$data['messages'][] = compact('message', 'severity');
		$data['timestamp'] = microtime(true);
		if ($this->cli && !$this->silent) {
			echo "\n";
			if($severity) echo strtoupper($severity);
			echo "> $message\n";
		}
		$this->writeData($data);
	}

	public function finish($status='success', $message='')
	{
		$data = $this->readData();
		if (!empty($data['current_process'])) {
			if($status != 'success') {
				$data['status'] = $status;
				$data['messages'][] = ['message' => $message, 'severity' => $status];
				$message = '';
			}
			$data['history'][] = [
				'name' => $data['current_process'],
				'progress' => is_numeric($data['current_progress']) ? 100 : $data['current_progress'],
				'count' => $data['count'],
				'messages' => $data['messages'],
			];
			unset($data['current_process']);
			$data['count'] = 0;
			$data['messages'] = [];
		}

		$data['finished'] = true;
		$data['status'] = $status;
		$data['status_message'] = $message;
		$data['timestamp'] = microtime(true);

		if ($this->cli && !$this->silent) {
			echo "\n";
			if($message) echo $message . "\n";
		}
		$this->writeData($data);
	}

	public function requestHistory()
	{
		$ret = $data = $this->readData();
		if(isset($data['history'])) {
			unset($data['history']);
			$this->writeData($data);
		} else if($this->file) {
			// skip updating file - no new progress
			flock($this->file,  LOCK_UN);
			fclose($this->file);
			$this->file = null;
		}
		return $ret;
	}
	
	public function pollHistory($timestamp, $timeout=null)
	{
		$timestamp = ''.$timestamp;
		$start = microtime(true);
		$wait = 50000;
		$count = 0;
		while(1) {
			$count ++;
			$hist = $this->requestHistory();
			$timed_out = (microtime(true) - $start) > ($timeout - $wait*2/1000000);
			if(! $hist || empty($hist['timestamp']) || ''.$hist['timestamp'] != $timestamp || $timed_out) {
				return $hist;
			}
			usleep($wait);
		}
	}
	
	public function cleanup() {
		if($this->fileName)
			@unlink($this->fileName);
	}
}

