<?php

class RemoteFSException extends Exception {}
class RemoteFSExceptionNotImp extends RemoteFSException {}

interface RemoteFS {
	/**
	 * Connect to remote host.
	 *
	 * Some backends, SSH for instance, do not create an actual 
	 * connection to be reused by other methods,  but create a
	 * new connection for each method call.
	 * Other backends, FTP for instance, create a connection that
	 * can be reused by other methods.
	 * In any case, this method can be used to test the connection.
	 *
	 * @throws RemoteFSException
	 *
	 */
	public function connect();

	/**
	 * Return directory listing. Directory name is relative to
	 * base directory specified in backend config.
	 * Each entry in returned array mus be a 2-element array
	 * with name and file size. If file size is unknown, is is set
	 * to -1 (for directories). Directory names end with slash
	 *
	 * @return array Files list
	 * @param string $dir Remote directory
	 * @throws RemoteFSException
	 *
	 */
	public function ls($dir);

	/**
	 * Remove a file. File name is relative to
	 * base directory specified in backend config.
	 *
	 * @param string $path Remote file path
	 * @throws RemoteFSException
	 *
	 */
	public function rm($path);

	/**
	 * Remove a file or directory. File name is relative to
	 * base directory specified in backend config. If specified file
	 * is a directory, it is removed recursively. No error is reported
	 * if directory does not exist or user has no access rights.
	 *
	 * @param string $dir Remote directory
	 * @throws RemoteFSException
	 *
	 */
	public function rmdir($dir);

	/**
	 * Create a directory. Directory name is relative to
	 * base directory specified in backend config. 
	 * If $create_parents is true, the whole directory
	 * hierarchy is created.
	 *
	 * @param string $dir Remote directory
	 * @param bool $create_parents
	 * @throws RemoteFSException
	 *
	 */
	public function mkdir($dir, $create_parents = true);

	/**
	 * Copy a file from local filesystem to remote host.
	 * Remote file name is relative to
	 * base directory specified in backend config. 
	 *
	 * @param string $src Local file name
	 * @param string $dst Remote file name
	 * @throws RemoteFSException
	 *
	 */
	public function copyToRemote($src, $dst);

	/**
	 * Copy a file from remote host to local filesystem.
	 * Remote file name is relative to
	 * base directory specified in backend config. 
	 *
	 * @param string $src Remote file name
	 * @param string $dst Local file name
	 * @throws RemoteFSException
	 *
	 */
	public function copyFromRemote($src, $dst);

	/**
	 * Return contents of a  file from remote host as string.
	 *
	 * Remote file name is relative to
	 * base directory specified in backend config. 
	 *
	 * @param string $src Remote file name
	 * @return string File contents
	 * @throws RemoteFSException
	 *
	 */
	public function readRemote($src);

	/**
	 * Write data from string to a  file on remote host.
	 *
	 * Remote file name is relative to
	 * base directory specified in backend config. 
	 *
	 * @param string $dst Remote file name
	 * @param string $data File contents
	 * @throws RemoteFSException
	 *
	 */
	public function writeRemote($dst, $data);

	/**
	 * Disconnect from remote host.
	 *
	 * @throws RemoteFSException
	 */
	public function disconnect();
}

