<?php

require_once 'include/RemoteFS/RemoteFS.php';
require_once 'include/RemoteFS/RemoteFSCmdLine.php';

class RemoteFSSsh implements RemoteFS {

	private $options;

	public function __construct($options) {
		$this->options = $options;
	}

	public function connect() {
		$client = $this->createClient('echo 123');
		if (!$client)
			throw new RemoteFSException;
		$result = '';
		while (!$client->stdOutEof())
			$result .= $client->readFromStdOut(1024);
		if ($client->close())
			throw new RemoteFSException;

		$res = trim($result) === '123';
		return $res;
	}

	public function ls($dir) {
		$dir = $this->options['ssh_dir'] . '/' . $dir;
		$client = $this->createClient('ls -l ' . escapeshellarg($dir));
		if (!$client)
			throw new RemoteFSException;
		$result = '';
		while (!$client->stdOutEof())
			$result .= $client->readFromStdOut(1024);

		if ($client->close())
			throw new RemoteFSException;
		$files = array();
		$lines = preg_split('~\\r?\\n~', $result);
		foreach ($lines as $line) {
			$parts = preg_split('~\\s+~', $line, 9);
			if (is_countable($parts) && count($parts) < 9)
				continue;
			$flags = $parts[0];
			$suffix = '';
			$type = substr($flags, 0, 1);
			$size = $parts[4];
			if ($type === '-') {
			} elseif ($type === 'd') {
				$size = -1;
				$suffix = '/';
			} else {
				continue;
			}
			if (substr($size, -1) === ',') {
				continue;
			}
			$name = end($parts);
			if ($suffix && (substr($name, -1) !== $suffix))
				$name .= $suffix;
			$files[] = array($name, $size);
		}
		return $files;
	}

	public function rm($path) {
		$path = $this->options['ssh_dir'] . '/' . $path;
		$client = $this->createClient('rm -i ' . escapeshellarg($path));
		if (!$client)
			throw new RemoteFSException;
		$client->writeToStdIn("y\n");
		$result = '';
		while (!$client->stdErrEof())
			$result .= $client->readFromStdErr(1024);

		if ($client->close())
			throw new RemoteFSException($result);
	}

	public function rmdir($path) {
		$path = $this->options['ssh_dir'] . '/' . $path;
		$client = $this->createClient('rm -rf ' . escapeshellarg($path));
		if (!$client)
			throw new RemoteFSException;
		$result = '';
		while (!$client->stdErrEof())
			$result .= $client->readFromStdErr(1024);

		if ($client->close())
			throw new RemoteFSException($result);
	}

	public function mkdir($dir, $create_parents = true) {
		$path = $this->options['ssh_dir'] . '/' . $dir;
		$client = $this->createClient('mkdir ' . ($create_parents ? '-p ' : '') . escapeshellarg($path));
		if (!$client)
			throw new RemoteFSException;
		$result = '';
		while (!$client->stdErrEof())
			$result .= $client->readFromStdErr(1024);

		if ($client->close())
			throw new RemoteFSException($result);
	}

	public function copyToRemote($src, $dst) {
		$f = fopen($src, 'rb');
		if (!is_resource($f))
			throw new RemoteFSException("File not found : $src");
		$dst = $this->options['ssh_dir'] . '/' . $dst;
		$client = $this->createClient('cat > ' . escapeshellarg($dst));
		if (!$client)
			throw new RemoteFSException;
		$err = '';
		while (!feof($f)) {
			$client->writeToStdIn(fread($f, 1024));
		}
		$client->closeStdIn();
		while (!$client->stdErrEof())
			$err .= $client->readFromStdErr(1024);
		if ($client->close())
			throw new RemoteFSException($err);
	}

	public function copyFromRemote($src, $dst) {
		$src = $this->options['ssh_dir'] . '/' . $src;
		$client = $this->createClient('cat ' . escapeshellarg($src));
		if (!$client)
			throw new RemoteFSException;
		$f = fopen($dst, 'wb');
		if (!is_resource($f)) {
			$client->close();
			throw new RemoteFSException("Cannot open file for writing : $dst");
		}
		$err = '';
		while (!$client->stdOutEof()) {
			fwrite($f, $client->readFromStdOut(1024));
		}
		$client->closeStdIn();
		while (!$client->stdErrEof())
			$err .= $client->readFromStdErr(1024);
		if ($client->close())
			throw new RemoteFSException($err);
	}

	public function readRemote($src) {
		$src = $this->options['ssh_dir'] . '/' . $src;
		$client = $this->createClient('cat ' . escapeshellarg($src));
		if (!$client)
			throw new RemoteFSException;
		$err = '';
		$ret = '';
		while (!$client->stdOutEof()) {
			$ret .= $client->readFromStdOut(1024);
		}
		$client->closeStdIn();
		while (!$client->stdErrEof())
			$err .= $client->readFromStdErr(1024);
		if ($client->close())
			throw new RemoteFSException($err);
		return $ret;
	}

	public function writeRemote($dst, $data) {
		$dst = $this->options['ssh_dir'] . '/' . $dst;
		$client = $this->createClient('cat > ' . escapeshellarg($dst));
		if (!$client)
			throw new RemoteFSException;
		$err = '';
		$towrite = strlen($data);
		$written = 0;
		while ($towrite) {
			$bytes = $client->writeToStdIn(substr($data, $written, 1024));
			if ($bytes === false) {
				$client->close();
				throw new RemoteFSException("Error writing to $dst");
			}
			$written += $bytes;
			$towrite -= $bytes;
		}
		$client->closeStdIn();
		while (!$client->stdErrEof())
			$err .= $client->readFromStdErr(1024);
		if ($client->close())
			throw new RemoteFSException($err);
	}
	
	public function disconnect() {}

	protected function createClient($command) {
		$args = array(
			'-l' . $this->options['ssh_username'],
			$this->options['ssh_host'],
			$command,
		);
		$client = new RemoteFSCmdLine('ssh', $args);
		return $client;
	}
}

