<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo, 
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */

abstract class SVGChartValueAxis
{
	const HORIZONTAL = 1;
	const VERTICAL = 2;

	protected $layout, $size, $data;
	protected $orientation;
	protected $min, $max, $step;
	protected $additive;
	protected $padding = 8;
	protected $ticklength = 5;

	abstract protected function addLabel($doc, $x, $label, $font, $fontSize, $offset);
	abstract public function getMinHeight();

	protected function __construct($size, $data, $layout, $additive, $orientation)
	{
		$this->layout = $layout;
		$this->data   = $data;
		$this->size   = $size;
		$this->orientation = $orientation;
		$this->additive = $additive;
	}
	
	public function setSize($size)
	{
		$this->size = $size;
	}

	public function render($doc, $grid = false)
	{


		$font = $this->layout->getValueAxisFont();
		$fontSize = $this->layout->getValueAxisFontSize();
		$this->calcTicks($font, $fontSize);
		$nLabels = ($this->max - $this->min) / $this->step + 1;
		$step = 0;
		$x = 0;
		if ($nLabels > 1) {
			$step = $this->size / ($nLabels - 1);
		}

		$g = $doc->createElement('g');

		$offset = $this->getOtherSize();
		for ($value=$this->min, $x = 0; $value <= $this->max; $value += $this->step, $x += $step) {
			if ($grid) {
				$g->appendChild($this->addGridLine($doc, $x, $offset, $grid));
			}
			$g->appendChild($this->addLabel($doc, $x, (string)$value, $font, $fontSize, $offset));
		}
		
		return $g;
	}

	public function getMin()
	{
		return $this->min;
	}

	public function getScale()
	{
		$range = max($this->max - $this->min, 1);
		return $this->size / $range;
	}

	protected function calcTicks($font, $fontSize)
	{
		$max = null;
		$min = 0;

		if ($this->additive) {
			foreach ($this->data->getLabels() as $label => $col) {
				$sum = 0;
				foreach ($this->data->getSeries() as $name => $series) {
					foreach ($series['points'] as $vlabel => $value) {
						if ($vlabel == $label) {
							$sum += $value['n'];
							break;
						}
					}
				}
				if ($min === null || $min > $sum) {
					$min = $sum;
				}
				if ($max === null || $max < $sum) {
					$max = $sum;
				}
			}
		} else {
			foreach ($this->data->getSeries() as $name => $series) {
				foreach ($series['points'] as $value) {
					if ($min === null || $min > $value['n']) {
						$min = $value['n'];
					}
					if ($max === null || $max < $value['n']) {
						$max = $value['n'];
					}
				}
			}
		}
		$min = floor($min);
		$max = ceil($max);

		$metricsKey = ($this->orientation == SVGChartValueAxis::HORIZONTAL) ? 'w' : 'h';

		$tickStep = 1;
		while (($max - $min) / $tickStep * $fontSize > $this->size) {
			if (substr($tickStep, 0, 1) == '2') {
				$tickStep *= 2.5;
			} else {
				$tickStep *= 2;
			}
		}
		while (true) {
			if($metricsKey == 'w') {
				$max_dim = 0;
				for ($value = $min; $value <= $max ; $value += $tickStep) {
					$metrics = $font->measure((string)$value . ' ', $fontSize);
					if($metrics[$metricsKey] > $max_dim)
						$max_dim = $metrics[$metricsKey];
				}
				$size = $max_dim * ($max - $min) / $tickStep;
			} else {
				$size = $fontSize * 1.2 * ($max - $min) / $tickStep;
			}
			if ($size <= $this->size) {
				break;
			}
			if (substr($tickStep, 0, 1) == '2') {
				$tickStep *= 2.5;
			} else {
				$tickStep *= 2;
			}
		}
		$this->min = $min;
		$this->max = $max;
		$this->step = $tickStep;
	}

	public function getOtherSize()
	{
		$font = $this->layout->getValueAxisFont();
		$fontSize = $this->layout->getValueAxisFontSize();
		$this->calcTicks($font, $fontSize);
		
		$metricsKey = ($this->orientation == SVGChartValueAxis::VERTICAL) ? 'w' : 'h';
		$s = 0;

		for ($value=$this->min; $value <= $this->max; $value += $this->step) {
			$metrics = $font->measure((string)$value, $fontSize);
			if ($s < $metrics[$metricsKey]) $s = $metrics[$metricsKey];
		}
		return $s + $this->padding;
	}

	abstract protected function addGridLine($doc, $x, $offset, $s);

}

class SVGChartValueAxisH extends SVGChartValueAxis
{
	public function __construct($size, $data, $layout, $additive)
	{
		parent::__construct($size, $data, $layout, $additive, SVGChartValueAxis::HORIZONTAL);
	}

	protected function addLabel($doc, $x, $label, $font, $fontSize, $offset)
	{
		$metrics = $font->measure($label, $fontSize);
		$text = $doc->createElement('text');
		$text->setAttribute('font-family', $font->getName());
		$text->setAttribute('font-size', $fontSize);
		$text->setAttribute('x', $x);
		$text->setAttribute('y', 0);
		$text->setAttribute('text-anchor', 'middle');
		$text->setAttribute('pointer-events', 'none');
		$text->setAttribute('fill', $this->layout->getTextColor());
		$text->appendChild($doc->createTextNode($label));
		return $text;
	}

	protected function addGridLine($doc, $x, $offset, $s)
	{
		$line = $doc->createElement('line');
		$line->setAttribute('x1', $x);
		$line->setAttribute('x2', $x);
		$line->setAttribute('y1', -$offset + $this->ticklength);
		$line->setAttribute('y2', -$s - $offset);
		$line->setAttribute('stroke', '#CCC');
		return $line;
	}
	
	public function getMinHeight()
	{
		return 0;
	}
}	

class SVGChartValueAxisV extends SVGChartValueAxis
{
	public function __construct($size, $data, $layout, $additive)
	{
		parent::__construct($size, $data, $layout, $additive, SVGChartValueAxis::VERTICAL);
	}
	
	protected function addLabel($doc, $x, $label, $font, $fontSize, $offset)
	{
		$metrics = $font->measure($label, $fontSize);
		$text = $doc->createElement('text');
		$text->setAttribute('font-family', $font->getName());
		$text->setAttribute('font-size', $fontSize);
		$text->setAttribute('x', $offset - $this->padding);
		$text->setAttribute('y', $this->size - ( $x - $metrics['h'] *0.4));
		$text->setAttribute('text-anchor', 'end');
		$text->setAttribute('pointer-events', 'none');
		$text->setAttribute('fill', $this->layout->getTextColor());
		$text->appendChild($doc->createTextNode($label));
		//$text->setAttribute('alignment-baseline', 'middle');
		return $text;
	}

	protected function addGridLine($doc, $x, $offset, $s)
	{
		$line = $doc->createElement('line');
		$line->setAttribute('y1', $this->size - $x);
		$line->setAttribute('y2', $this->size - $x);
		$line->setAttribute('x1', $offset - $this->ticklength);
		$line->setAttribute('x2', $offset + $s);
		$line->setAttribute('stroke', '#CCC');
		return $line;
	}

	public function getMinHeight()
	{
		return 200;
	}
}

