<?php


namespace OneCRM\CalDAV;

use Sabre\CalDAV\Backend\AbstractBackend;
use Sabre\CalDAV\Backend\SyncSupport;
use Sabre\CalDAV\Backend\SubscriptionSupport;
use Sabre\CalDAV\Plugin;
use Sabre\VObject\Reader;
use Sabre\CalDAV\Xml\Property\SupportedCalendarComponentSet;
use Sabre\CalDAV\Xml\Property\ScheduleCalendarTransp;
use Sabre\VObject\Property\ICalendar\DateTime as ICalDateTime;
use Sabre\DAV\PropPatch;
use Sabre\VObject\DateTimeParser;
use Sabre\VObject\Component\VCalendar;

require_once 'include/utils/vcal_utils.php';

class Backend extends AbstractBackend implements SyncSupport, SubscriptionSupport {

    /**
     * We need to specify a max date, because we need to stop *somewhere*
     *
     * On 32 bit system the maximum for a signed integer is 2147483647, so
     * MAX_DATE cannot be higher than date('Y-m-d', 2147483647) which results
     * in 2038-01-19 to avoid problems when the date is converted
     * to a unix timestamp.
     */
    const MAX_DATE = '2038-01-01';

    public function __construct() {}
   
	public function getCalendarsForUser($principalUri) {
		$ret = [];

		$meta = [
			'meetings' => ['Meetings', 'VEVENT', false],
			'calls' => ['Calls', 'VEVENT', false],
			'tasks' => ['Tasks', 'VTODO', true],
		];

		foreach ($meta as $cid => $m) {
            $components = [$m[1]];
			$token = $this->checkPrefUpdates($this->getSyncToken($cid), $cid);

            $ret[] = [
                'id' => $cid,
                'uri' => $cid,
                'principaluri' => $principalUri,
                '{' . Plugin::NS_CALENDARSERVER . '}getctag' => 'http://sabredav.org/ns/sync/' . $token,
                '{DAV:}sync-token' => $token,
                '{' . Plugin::NS_CALDAV . '}supported-calendar-component-set' => new SupportedCalendarComponentSet($components),
				'{' . Plugin::NS_CALDAV . '}schedule-calendar-transp' => new ScheduleCalendarTransp($m[2] ? 'transparent':'opaque'),
				'{DAV:}displayname' => $m[0],
			];
		}

        return $ret;
    }

	public function createCalendar($principalUri, $calendarUri, array $properties) {
		return false;
    }

    public function updateCalendar($calendarId, PropPatch $propPatch) {
        return false;
    }

    public function deleteCalendar($calendarId) {
		return false;
    }

    public function getCalendarObjects($calendarId) {
		return $this->getCalendarObjectsList($calendarId);
    }

    public function getCalendarObject($calendarId,$objectUri) {

		switch ($calendarId) {
			case 'calls':
				$models = ['Call'];
				break;
			case 'meetings':
				$models = ['Meeting', 'RecurrenceRule'];
				break;
			case 'tasks':
				$models = ['Task'];
				break;
			default:
				return null;
		}

		$id = preg_replace('~\\.ics$~', '', $objectUri);

		foreach ($models as $modelName) {
			$row = \ListQuery::quick_fetch_key($modelName, 'vcal_uri', $objectUri, null, ['acl_checks' => 'view']);

			if(! $row) {
				$row = \ListQuery::quick_fetch($modelName, $id, null, ['acl_checks' => 'view']);
			}

			if(! $row) {
				continue;
			}
			if(!$row->getAclAllowed('view')) {
				return null;
			}

			$cdata = $row->getField('vcal');

			if (!$cdata)
				$cdata = \vCalUtils::vcal_update($row);

			return [
				'id'            => $id,
				'uri'           => $objectUri,
				'lastmodified'  => strtotime($row->getField('date_modified') . ' GMT'),
				'etag'          => '"' . md5($cdata) . '"',
				'calendarid'    => $calendarId,
				'size'          => strlen($cdata),
				'calendardata'  => $cdata,
				'component'     => $calendarId == 'tasks' ? 'VTODO' : 'VEVENT',
			];
		}
    }

	private function getCalendarObjectsList($calendarId, $uris = null, $callback = null) {
		$ret = [];
		switch ($calendarId) {
			case 'calls':
				$models = ['Call'];
				break;
			case 'meetings':
				$models = ['Meeting', 'RecurrenceRule'];
				break;
			case 'tasks':
				$models = ['Task'];
				break;
			default:
				return null;
		}

		foreach ($models as $modelName) {
			$lq = new \ListQuery($modelName);
			$lq->addColumnFields();
			$alias = $lq->getTableAlias(null, true);
			$lq->addFieldLiteral('some_vcal_uri', "IFNULL(" . $alias . ".vcal_uri,CONCAT(" . $alias . ".id, '.ics'))");
			$lq->addSimpleFilter('recurrence_rule_id', '', 'not');

			if (is_array($uris) && !empty($uris)) {
				$lq->addSimpleFilter('some_vcal_uri', $uris, '=');
			}

			$lq->addAclFilter('list');
			$this->generateListWhereClause($lq, $calendarId, false, null);

			
			/** @var \Closure $callback */
			if ($callback)
				$callback($lq);

			$result = $lq->runQuery(0, null, false, null, 100);

			$it = $result->getIterator();
			foreach ($it as $row) {
				if ($modelName == 'RecurrenceRule') {
					$template = @json_decode($row->getField('template'), true);
					if (!is_array($template))
						continue;
					if ( ($template['assigned_user_id'] ?? '') != \AppConfig::current_user_id())
						continue;
				}
				$cdata = $row->getField('vcal');

				if (!$cdata)
					$cdata = \vCalUtils::vcal_update($row);
				if (! $cdata)
					continue;

				$ret[] =  [
					'id'            => preg_replace('~\\.ics$~', '', $row->getField('some_vcal_uri')),
					'uri'           => $row->getField('some_vcal_uri'),
					'lastmodified'  => strtotime($row->getField('date_modified') . ' GMT'),
					'etag'          => '"' . md5($cdata) . '"',
					'calendarid'    => $calendarId,
					'size'          => strlen($cdata),
					'calendardata'  => $cdata,
					'component'     => $calendarId == 'tasks' ? 'VTODO' : 'VEVENT',
				 ];
			}
		}

		$GLOBALS['log']->debug("CalDAV Sync: return " . sizeof($ret) ." ". $calendarId);

		return $ret;
	}

	public function getMultipleCalendarObjects($calendarId, array $uris) {
		return $this->getCalendarObjectsList($calendarId, $uris);
    }

	public function createCalendarObject($calendarId, $objectUri, $calendarData) {
		global $db;
		$id = create_guid();
		$evs = $this->extractEvents($calendarId, $calendarData);
		if (empty($evs)) return null;

		switch ($calendarId) {
			case 'calls':
				$modelName = 'Call';
				break;
			case 'meetings':
				$modelName =  empty($evs[0]->RRULE) ? 'Meeting' : 'RecurrenceRule';
				break;
			case 'tasks':
				$modelName = 'Task';
				break;
			default:
				return null;
		}

		$row = \RowUpdate::blank_for_model($modelName, \AppConfig::current_user_id());
		$row->set('id', $id);
		$row->set('assigned_user_id', \AppConfig::current_user_id());

		if (!\vCalUtils::vcal_import($calendarData, $row))
			return null;

		$row->set('vcal', $calendarData);
		$row->set('vcal_uri', $objectUri);
		$row->set('vcal_size', strlen($calendarData));
		$row->setRelatedData('skip_first_instance', false);
		$row->save();
    }

    public function updateCalendarObject($calendarId, $objectUri, $calendarData) {
		$gmtTZ = new \DateTimeZone('GMT');
		$evs = $this->extractEvents($calendarId, $calendarData);
		if (empty($evs)) return null;
		$result = $this->findCalendarObject($calendarId, $objectUri, $calendarData);
		if(! $result || ! $result->getAclAllowed('edit')) {
			return null;
		}
		$upd = \RowUpdate::for_result($result);
		$modelName = $result->getModelName();
		$wasRecurring = $modelName == 'RecurrenceRule';
		$isRecurring = !empty($evs[0]->RRULE);

		// if recurring object was converted to non-recurring or vice versa, 
		// need to delete existing record and create a new one
		if ($isRecurring != $wasRecurring) {
			$upd->set([
				'vcal_uri' => '',
				'vcal' => '',
				'vcal_size' => 0,
			]);
			$upd->markDeleted();
			return $this->createCalendarObject($calendarId, $objectUri, $calendarData);
		}

		$data = $this->getFieldsValues($calendarId, $evs[0]);

		if (!\vCalUtils::vcal_import($calendarData, $upd)) {
			return null;
		}

		$modDate = '';
		foreach ($evs as $e) {
			$mod = $e->select('LAST-MODIFIED')[0]->getDateTime()->setTimeZone($gmtTZ)->format("Y-m-d H:i:s");
			if ($mod > $modDate)
				$modDate = $mod;
		}
		if ($modDate) {
			$upd->update_date_modified = false;
			$upd->set('date_modified', $modDate);
		}
		$upd->set('vcal', $calendarData);
		$upd->set('vcal_size', strlen($calendarData));
		$upd->setRelatedData('skip_first_instance', false);
		$upd->save();

		$calendarData = $upd->getField('vcal');

		if ($isRecurring) {
			$this->removeExceptions($upd->getField('id'), $evs[0]);
			$this->updateExceptions($upd->getField('id'), $evs);
			\vCalUtils::rebuild_recur_vcal($upd->getField('id'));
			$updated = \ListQuery::quick_fetch('RecurrenceRule', $upd->getField('id'), ['vcal']);
			if ($updated)
				$calendarData = $updated->getField('vcal');
		}

		return '"' . md5($calendarData) . '"';
	}


	// finds an existing record. For meetings, Meetings and Recurrence are searched,
	// and the order depends on wheter the event coming from the client is recurring.
	private function findCalendarObject($calendarId, $objectUri, $event) {
		switch ($calendarId) {
			case 'calls':
				$models = ['Call'];
				break;
			case 'meetings':
				if (!empty($event->RRULE)) 
					$models = ['RecurrenceRule', 'Meeting'];
				else
					$models = ['Meeting', 'RecurrenceRule'];
				break;
			case 'tasks':
				$modelName = ['Task'];
				break;
			default:
				return null;
		}
		$result = null;
		$id = preg_replace('~\\.ics$~', '', $objectUri);
		foreach ($models as $modelName) {
			$result = \ListQuery::quick_fetch_key($modelName, 'vcal_uri', $objectUri, null, ['acl_checks' => 'edit']);
			if(! $result) {
				$result = \ListQuery::quick_fetch($modelName, $id, null, ['acl_checks' => 'edit']);
			}
			if ($result) break;
		}
		return $result;
	}

    public function deleteCalendarObject($calendarId,$objectUri) {
		switch ($calendarId) {
			case 'calls':
				$models = ['Call'];
				break;
			case 'meetings':
				$models = ['Meeting', 'RecurrenceRule'];
				break;
			case 'tasks':
				$models = ['Task'];
				break;
			default:
				return false;
		}

		$id = preg_replace('~\\.ics$~', '', $objectUri);

		$acl_checks = ['acl_checks' => 'delete'];

		foreach ($models as $modelName) {
			$do_acl_checks = $modelName != 'RecurrenceRule';

			$result = \ListQuery::quick_fetch_key($modelName, 'vcal_uri', $objectUri, false, $do_acl_checks ? ['acl_checks' => 'delete'] : []);

			if(! $result) {
				$result = \ListQuery::quick_fetch($modelName, $id, false, $do_acl_checks ? ['acl_checks' => 'delete'] : []);
			}

			if(! $result || ($do_acl_checks && !$result->getAclAllowed('delete'))) {
				continue;
			}
			$ev = \RowUpdate::for_result($result);
			$ev->markDeleted(true);
		}

		return true;
    }

	public function calendarQuery($calendarId, array $filters) {
		$callback = function($lq) use ($filters) {

			if (!empty($filters['comp-filters'])) {

				foreach ($filters['comp-filters'] as $f) {
					if (!isset($f['time-range']))
						continue;

					if (isset($f['time-range']['start'])) {
						if ($f['time-range']['start'] instanceof \DateTime) {
							$start = gmdate('Y-m-d H:i:s', strtotime($f['time-range']['start']->format('Y-m-d H:i:s e')));
							$lq->addSimpleFilter('date_start', $start, '>=');
						}
					}

					if (isset($f['time-range']['end'])) {
						if ($f['time-range']['end'] instanceof \DateTime) {
							$end = gmdate('Y-m-d H:i:s', strtotime($f['time-range']['end']->format('Y-m-d H:i:s e')));
							$lq->addSimpleFilter('date_start', $end, '<');
						}
					}
					
				}

			}

		};

		$list = $this->getCalendarObjectsList($calendarId, null, $callback);
		$ret = [];

		if (!empty($list)) {
			foreach ($list as $o)
				$ret[] = $o['uri'];
		}

		return $ret;
    }

	private function generateListWhereClause(\ListQuery $lq, $calendarId, $all, $syncTokenDate) {
		$user_id = \AppConfig::current_user_id();
		$type = \UserPreference::getPreference('caldav_' . $calendarId);

		$date_clause = ['operator' => 'OR', 'multiple' => [] ];
		$lq->addSimpleFilter('recurrence_rule_id', '', 'not');
		
		if ($type == 'invited') {
			$lq->addLink('users~join');
			$lq->addField('users~join.deleted');
			$lq->addField('users~join.date_modified');
			$lq->addSimpleFilter('users~join.user_id', $user_id);

			if ($syncTokenDate && !$all) {
				$date_clause['multiple'][] = array(
					'value' => $syncTokenDate,
					'field' => 'users~join.date_modified',
					'operator' => '>=',
				);
			}
		}

		if ($syncTokenDate && !$all) {
			$date_clause['multiple'][] = array(
				'value' => $syncTokenDate,
				'field' => 'date_modified',
				'operator' => '>=',
			);
		}

		if (!empty($date_clause['multiple']))
			$lq->addFilterClause($date_clause);

		if (!$all) {
			$lq->addAclFilter('list');

			switch ($type) {
				case 'invited':
				case 'visible':
					break;
				default:
					$lq->addSimpleFilter('assigned_user_id', $user_id, '=');
					break;
			}
		}
		
	}

	public function getChangesForCalendar($calendarId, $syncToken, $syncLevel, $limit = null) {
		$user_id = \AppConfig::current_user_id();

		$result = [
            'added'     => [],
            'modified'  => [],
            'deleted'   => [],
        ];

		switch ($calendarId) {
			case 'calls':
				$models = ['Call'];
				break;
			case 'meetings':
				$models = ['Meeting', 'RecurrenceRule'];
				break;
			case 'tasks':
				$models = ['Task'];
				break;
			default:
				return $result;
		}

		$currentToken = $this->getSyncToken($calendarId);
		$prevSyncDateTime = $this->getSyncDate($syncToken);

		if (!$currentToken || !$prevSyncDateTime)
			return null;

		$result['syncToken'] = $currentToken;

		$acl_updated = \AppConfig::setting('misc.acl_updated');
		$type_updated = \UserPreference::getPreference('caldav_pref_updated');
		$type = \UserPreference::getPreference('caldav_' . $calendarId);
		$globalResyncStatus = $this->getResyncStatus();

		if ($globalResyncStatus) {
			$resync = true;
			$this->setResynced();
		} else {
			$resync = $syncToken && ($acl_updated >= $prevSyncDateTime || $type_updated >= $prevSyncDateTime);
		}

		foreach ($models as $modelName) {
			$lq = new \ListQuery($modelName, null, ['acl_checks' => 'list', 'filter_deleted' => false]);
			$this->generateListWhereClause($lq, $calendarId, $resync, $prevSyncDateTime);

			if (!$limit)
				$limit = null;

			$lq->setOrderBy('date_modified');
			$lq->addField('id');
			$lq->addField('assigned_user_id');
			$lq->addField('date_entered');
			$lq->addField('date_modified');
			$lq->addField('deleted');
			$lq->addField('template');
			$alias = $lq->getTableAlias(null, true);
			$lq->addFieldLiteral('some_vcal_uri', "IFNULL(" . $alias . ".vcal_uri,CONCAT(" . $alias . ".id, '.ics'))");

			$res = $lq->runQuery(0, $limit, false, null, 100);
			$it = $res->getIterator();
			foreach ($it as $row) {
				$deleted = $row->getField('deleted') || !$row->getAclAllowed('list');
				if ($modelName != 'RecurrenceRule') {
					switch ($type) {
					case 'invited':
						if ($row->getField('users~join.deleted') || !$row->getField('users~join.date_modified') || $row->getField('users~join.user_id') != $user_id)
							$deleted = true;
						break;
					case 'visible':
						break;
					default:
						if ($row->getField('assigned_user_id') != \AppConfig::current_user_id())
							$deleted = true;
					}
				} else {
					$template = @json_decode($row->getField('template', true));
					if (!is_array($template)) {
						$deleted = true;
					} elseif (($template['assigned_user_id'] ?? '') != \AppConfig::current_user_id()) {
						$deleted = true;
					}
				}
				if ($deleted)
					$result['deleted'][] = $row->getField('some_vcal_uri');
				else if ($row->getField('date_entered') == $row->getField('date_modified'))
					$result['added'][] = $row->getField('some_vcal_uri');
				else
					$result['modified'][] = $row->getField('some_vcal_uri');
			}
		}

		$mCount = is_countable($result["modified"]) ? sizeof($result["modified"]) : 0;
		$aCount = is_countable($result["added"]) ? sizeof($result["added"]) : 0;
		$dCount = is_countable($result["deleted"]) ? sizeof($result["deleted"]) : 0;
		$GLOBALS['log']->debug("CalDAV Sync: Modified: ". $mCount . " Added: " . $aCount . " Deleted: " . $dCount);

		return $result;
    }

    public function getSubscriptionsForUser($principalUri) {
		return [];
    }

    public function createSubscription($principalUri, $uri, array $properties) {
        return false;
    }

    public function updateSubscription($subscriptionId, PropPatch $propPatch) {
		return false;
	}

    public function deleteSubscription($subscriptionId) {
		return false;
    }

	/**
	 * If CardDAV preferences or ACL were updated the method will update sync token and return it.
	 * Returns given sync token otherwise.
	 *
	 * @param int $currentToken
	 * @param string $calendarId: calls, meetings, tasks
	 *
	 * @return int|null
	 */
	private function checkPrefUpdates($currentToken, $calendarId) {
		$aclUpdated = \AppConfig::setting('misc.acl_updated');
		$typeUpdated = \UserPreference::getPreference('carddav_pref_updated');
		$currentSyncDateTime = $this->getSyncDate($currentToken);

		$result = $currentToken;

		if ($currentToken && ($aclUpdated >= $currentSyncDateTime || $typeUpdated >= $currentSyncDateTime)) {
			$this->updateSyncToken($calendarId);
			$result = $this->getSyncToken($calendarId);
		}

		return $result;
	}

	/**
	 * @param string $type: calls, meetings, tasks
	 * @return int|null
	 */
	private function getSyncToken($type) {
		return \vCalUtils::get_sync_token($type);
	}

	/**
	 * @param int|string $token
	 *
	 * @return false|string
	 */
	private function getSyncDate($token) {
		return \vCalUtils::convert_sync_token_to_date($token);
	}

	/**
	 * @param string $type: calls, meetings, tasks
	 */
	private function updateSyncToken($type) {
		\vCalUtils::update_sync_token($type);
	}

	/**
	 * @return bool
	 */
	private function getResyncStatus() {
		return \vCalUtils::get_resync_value();
	}

	private function setResynced() {
		require_bean("User");
		$user = new \User();
		$user->id = \AppConfig::current_user_id();

		\vCalUtils::set_resync_value(false, $user);
	}


	// extract events from VCALENDAR. Recurrence is ignored unless $calendarId
	// is 'meetings'. 'Root' recurring event is returned first, exception instances
	// follow.
	private function extractEvents($calendarId, $vcal_text) {
		try {
			$vcal = Reader::read($vcal_text,  Reader::OPTION_FORGIVING | Reader::OPTION_IGNORE_INVALID_LINES);
		} catch (\Exception $e) {
			return false;
		}
		$events = $vcal->select($calendarId == 'tasks' ? 'VTODO' : 'VEVENT');
		$events = array_filter($events, function($ev) use ($calendarId) {
			return $calendarId == 'meetings' || empty($ev->RRULE);
		});
		if (empty($events)) return false;
		// usort prior to PHP 8.0 is not stable, but in this case it does not matter
		usort($events, function($a, $b) {
			$recA = !empty($a->RRULE);
			$recB = !empty($b->RRULE);
			if ($recA && !$recB) {
				return -1;
			}
			if (!$recA && $recB) {
				return 1;
			}
			return 0;
		});
		return $events;
	}

	private function updateExceptions($ruleId, $events) {
		global $timedate;
		$gmtTZ = new \DateTimeZone('GMT');
		foreach ($events as $e) {
			if (!empty($e->RRULE)) continue;
			$rids = $e->select('RECURRENCE-ID');
			if (empty($rids)) return 0;
			$rid = $rids[0];
			$lq = new \ListQuery('Meeting');
			$lq->addSimpleFilter('recurrence_rule_id', $ruleId);
			$allDay = !$rid->isFloating();
			if (!$allDay) {
				$originalStart = $rid->getDateTime('GMT')->format("Y-m-d H:i:s");
				$lq->addSimpleFilter('recurrence_start_time', $originalStart);
			} else {
				// relying on fact that getDateTime() returns \DateTimeImmutable
				$originalStart = $rid->getDateTime($timedate->getUserTimeZone());
				$lowBound = $originalStart->setTime(0, 0)->setTimeZone($gmtTZ)->format("Y-m-d H:i:s");
				$highBound = $originalStart->setTime(23, 59, 59)->setTimeZone($gmtTZ)->format("Y-m-d H:i:s");
				$lq->addSimpleFilter('recurrence_start_time', $lowBound, '>=');
				$lq->addSimpleFilter('recurrence_start_time', $highBound, '<=');
			}
			$row = $lq->runQuerySingle();
			if ($row && !$row->failed) {
				$upd = \RowUpdate::for_result($row);
				$cal = new VCalendar;
				$cal->add($e);
				if (\vCalUtils::vcal_import($cal->serialize(), $upd)) {
					$upd->set('recurrence_exception', 1);
					$upd->setRelatedData('disable_vcal_rebuild', true);
					$upd->save();
				}
			}
		}
	}

	private function removeExceptions($ruleId, $event) {
		global $timedate;
		$gmtTZ = new \DateTimeZone('GMT');
		$exceptions = $event->select('EXDATE');
		$dtStart = $event->select('DTSTART');
		if (empty($dtStart)) return;
		$dtStart = $dtStart[0];
		$allDay = !$dtStart->isFloating();
		foreach ($exceptions as $e) {
			$lq = new \ListQuery('Meeting');
			$lq->addSimpleFilter('recurrence_rule_id', $ruleId);
			if (!$allDay) {
				$originalStart = $e->getDateTime('GMT')->format("Y-m-d H:i:s");
				$lq->addSimpleFilter('recurrence_start_time', $originalStart);
			} else {
				// relying on fact that getDateTime() returns \DateTimeImmutable
				$originalStart = $e->getDateTime($timedate->getUserTimeZone());
				$lowBound = $originalStart->setTime(0, 0)->setTimeZone($gmtTZ)->format("Y-m-d H:i:s");
				$highBound = $originalStart->setTime(23, 59, 59)->setTimeZone($gmtTZ)->format("Y-m-d H:i:s");
				$lq->addSimpleFilter('recurrence_start_time', $lowBound, '>=');
				$lq->addSimpleFilter('recurrence_start_time', $highBound, '<=');
			}
			$result = $lq->runQuerySingle();
			if ($result && !$result->failed) {
				$upd = \RowUpdate::for_result($result);
				$upd->set('recurrence_exception', 1);
				$upd->setRelatedData('disable_vcal_rebuild', true);
				$d = $upd->markDeleted();
			}
		}
	}

	// returns values that can be used as template for recurring events
	// or as field values for normal events
	private function getFieldsValues($calendarId, $ev) {
		$isTask = $calendarId == 'tasks';
		$data = [];
		$data['name'] = (string)$ev->SUMMARY;
		if (!strlen($data['name']))
			$data['name ']= translate('LBL_EMPTY_MEETING_TITLE', 'Calendar');
		$data['description'] = (string)$ev->DESCRIPTION;
		if (isset($ev->CLASS))
			$data['is_private'] = (string)$ev->CLASS !== 'PUBLIC' ? 1 : 0;
		if (isset($ev->LOCATION))
			$data['location'] = (string)$ev->LOCATION;
		if (isset($ev->STATUS)) {
			if ($isTask) {
				$status_map = ['In Progress' => 'IN-PROCESS', 'Completed' => 'COMPLETED', 'Pending Input' => 'NEEDS-ACTION', 'Deferred' => 'CANCELLED'];
			} else {
				$status_map = ['Planned' => 'TENTATIVE', 'Held' => 'CONFIRMED', 'Not Held' => 'CANCELLED'];
			}
			$status_map = array_flip($status_map);
			$s = (string)$ev->STATUS;
			if (isset($status_map[$s]))
				$data['status'] = $status_map[$s];
		}
		if (isset($ev->PRIORITY)) {
			$p = (int)(string)$ev->priority;
			if ($p > 6)
				$p = 'P0';
			else if ($p > 3)
				$p = 'P1';
			else
				$p = 'P2';

		} else {
			$p = '';
		}
		$data['priority'] = $p;

		if (@$ev->DTSTART instanceof ICalDateTime) {
			if ($ev->DTSTART->getValueType() == 'DATE') {
				$data['is_daylong'] = 1;
				if(!$isTask)
					$data['date_start'] = gmdate('Y-m-d 12:00:00', $ev->DTSTART->getDateTime()->getTimestamp());
			} else {
				$data['is_daylong'] = 0;
				if(!$isTask)
					$data['date_start'] = gmdate('Y-m-d H:i:s', $ev->DTSTART->getDateTime()->getTimestamp());
				if (isset($ev->DURATION)) {
					$range = DateTimeParser::parseDuration($ev->DURATION, true);

					$seconds = strtotime($ev->DTSTART->getDateTime()->format('Y-m-d H:i:s e') . ' ' . $range) - $ev->DTSTART->getDateTime()->getTimestamp();
					$idata['duration'] = (int)($seconds / 60);
				} else if (@$ev->DTEND instanceof ICalDateTime) {
					$diff = $ev->DTEND->getDateTime()->diff($ev->DTSTART->getDateTime());
					$data['duration'] = $diff->format('%a') * 24*60 + $diff->format('%h') * 60  + $diff->format('%i');
				}
			}
		}
		if (@$ev->DUE instanceof ICalDateTime) {
			$data['date_due'] = gmdate('Y-m-d H:i:s', $ev->DUE->getDateTime()->getTimestamp());
		}
		return $data;
	}

	// get values of RecurrenceRule fields
	private function getRecurrenceFields($ev) {
		return $ev->RRULE->getParts();
	}

}

