<?php


namespace OneCRM\CardDAV;

require_once 'include/utils/vcard_utils.php';
require_once 'modules/Favorites/Favorite.php';

/**
 * 1CRM CardDAV backend
 *
 */
class Backend extends \Sabre\CardDAV\Backend\AbstractBackend implements \Sabre\CardDAV\Backend\SyncSupport {

    public function __construct() {}

    /**
     * Returns the list of addressbooks for a specific user.
     *
     * @param string $principalUri
     * @return array
     */
    public function getAddressBooksForUser($principalUri) {
	    $token = $this->checkPrefUpdates($this->getSyncToken());

		return [
			[
                'id'  => 'contacts',
                'uri' => 'contacts',
                'principaluri' => $principalUri,
                '{DAV:}displayname' => 'Contacts',
                '{' . \Sabre\CardDAV\Plugin::NS_CARDDAV . '}supported-address-data' => new \Sabre\CardDAV\Xml\Property\SupportedAddressData(),
				'{http://calendarserver.org/ns/}getctag' => $token,
                '{DAV:}sync-token' => $token
			]
		];
    }

    /**
     * Updates an addressbook's properties
     *
     * See Sabre\DAV\IProperties for a description of the mutations array, as
     * well as the return value.
     *
     * @param mixed $addressBookId
     * @see \Sabre\DAV\IProperties::updateProperties
     * @return bool|array
     */
    public function updateAddressBook($addressBookId, \Sabre\DAV\PropPatch $propPatch) {
		return false;
    }

    /**
     * Creates a new address book
     *
     * @param string $principalUri
     * @param string $url Just the 'basename' of the url.
     * @param array $properties
     * @return void
     */
    public function createAddressBook($principalUri, $url, array $properties) {}

    /**
     * Deletes an entire addressbook and all its contents
     *
     * @param int $addressBookId
     * @return void
     */
    public function deleteAddressBook($addressBookId) {}

	private function getCardsList($addressbookId, $uris = null) {
		$lq = new \ListQuery('Contact');
		$lq->addColumnFields();
		$lq->addFieldLiteral('some_vcard_uri', "IFNULL(vcard_uri,CONCAT(" . $lq->getTableAlias(null, true) . ".id, '.vcf'))");
		$lq->addAclFilter('list');
		if (is_array($uris) && !empty($uris))
			$lq->addSimpleFilter('some_vcard_uri', $uris, '=');
		$type = \UserPreference::getPreference('carddav_contacts_which');

		switch ($type) {
			case 'fav':
				\Favorite::add_join($lq);
				$lq->addSimpleFilter('favorites.deleted', 0, '=');
				break;
			case 'visible':
				break;
			default:
				$lq->addSimpleFilter('assigned_user_id', \AppConfig::current_user_id(), '=');
				break;
		}

		$result = $lq->runQuery(0, null, false, null, 100);

		$ret = [];
		while(! $result->failed) {
			foreach ($result->getRowIndexes() as $idx) {
				$row = $result->getRowResult($idx);
				$cdata = $row->getField('vcard');
				if (!$cdata)
					$cdata = \vCardUtils::vcard_update($row);
				if (! $cdata)
					continue;
				$ret[] = [
					'id' => preg_replace('~\\.vcf$~', '', $row->getField('some_vcard_uri')),
					'uri' => $row->getField('some_vcard_uri'),
					'lastmodified' => strtotime($row->getField('date_modified') . ' GMT'),
					'carddata' => $cdata,
					'etag' => '"'.md5($cdata).'"',
				];
			}
			if($result->page_finished)
				break;
			$lq->pageResult($result);
		}

		return $ret;
	}

    /**
     * Returns all cards for a specific addressbook id.
     *
     * This method should return the following properties for each card:
     *   * carddata - raw vcard data
     *   * uri - Some unique url
     *   * lastmodified - A unix timestamp
     *
     * It's recommended to also return the following properties:
     *   * etag - A unique etag. This must change every time the card changes.
     *   * size - The size of the card in bytes.
     *
     * If these last two properties are provided, less time will be spent
     * calculating them. If they are specified, you can also ommit carddata.
     * This may speed up certain requests, especially with large cards.
     *
     * @param mixed $addressbookId
     * @return array
     */
	public function getCards($addressbookId) {
		return $this->getCardsList($addressbookId);
    }

    /**
     * Returns a specfic card.
     *
     * The same set of properties must be returned as with getCards. The only
     * exception is that 'carddata' is absolutely required.
     *
     * If the card does not exist, you must return false.
     *
     * @param mixed $addressBookId
     * @param string $cardUri
     * @return array
     */
	public function getCard($addressBookId, $cardUri) {
		$id = preg_replace('~\\.vcf$~', '', $cardUri);
		$row = \ListQuery::quick_fetch_key('Contact', 'vcard_uri', $cardUri, true, ['acl_checks' => 'view']);
		if(! $row) {
			$row = \ListQuery::quick_fetch('Contact', $id, true, ['acl_checks' => 'view']);
		}
		if(! $row || ! $row->getAclAllowed('view'))
			return null;
		$cdata = $row->getField('vcard');
		if (!$cdata)
			$cdata = \vCardUtils::vcard_update($row);
		$ret = [
			'id' => $id,
			'uri' => $cardUri,
			'lastmodified' => strtotime($row->getField('date_modified') . ' GMT'),
			'carddata' => $cdata,
			'etag' => '"'.md5($cdata).'"',
		];
		return $ret;

    }

    /**
     * Returns a list of cards.
     *
     * This method should work identical to getCard, but instead return all the
     * cards in the list as an array.
     *
     * If the backend supports this, it may allow for some speed-ups.
     *
     * @param mixed $addressBookId
     * @param array $uris
     * @return array
     */
    public function getMultipleCards($addressBookId, array $uris) {
		return $this->getCardsList($addressBookId, $uris);
    }

    /**
     * Creates a new card.
     *
     * The addressbook id will be passed as the first argument. This is the
     * same id as it is returned from the getAddressbooksForUser method.
     *
     * The cardUri is a base uri, and doesn't include the full path. The
     * cardData argument is the vcard body, and is passed as a string.
     *
     * It is possible to return an ETag from this method. This ETag is for the
     * newly created resource, and must be enclosed with double quotes (that
     * is, the string itself must contain the double quotes).
     *
     * You should only return the ETag if you store the carddata as-is. If a
     * subsequent GET request on the same card does not have the same body,
     * byte-by-byte and you did return an ETag here, clients tend to get
     * confused.
     *
     * If you don't return an ETag, you can just return null.
     *
     * @param mixed $addressBookId
     * @param string $cardUri
     * @param string $cardData
     * @return string|null
     */

	/*
		BEGIN:VCARD
		VERSION:3.0
		PRODID:-//Apple Inc.//Mac OS X 10.8.5//EN
		N:aa;aa;;aa;
		FN:aa aa aa
		ORG:Company;Department
		TITLE:Title
		BDAY:1976-04-30
		EMAIL;type=INTERNET;type=HOME;type=pref:home@example.com
		EMAIL;type=INTERNET;type=WORK:work@example.com
		TEL;type=CELL;type=VOICE;type=pref:123456
		TEL;type=IPHONE;type=CELL;type=VOICE:7891011
		TEL;type=HOME;type=VOICE:121314
		TEL;type=WORK;type=VOICE:151617
		TEL;type=MAIN:181920
		TEL;type=HOME;type=FAX:212223
		TEL;type=WORK;type=FAX:242526
		TEL;type=OTHER;type=FAX:272829
		TEL;type=PAGER:303132
		TEL;type=OTHER;type=VOICE:333435
		ADR;type=HOME;type=pref:;;Home Street;city;region;zip;country
		ADR;type=WORK:;;Work street;city;state;zip;country
		X-SOCIALPROFILE;type=twitter:http://twitter.com/twitter
		X-SOCIALPROFILE;type=facebook:http://facebook.com/facebook2
		X-SOCIALPROFILE;type=flickr:http://www.flickr.com/photos/flickr
		X-SOCIALPROFILE;type=linkedin:http://www.linkedin.com/in/linkedin
		X-SOCIALPROFILE;type=myspace:http://www.myspace.com/mysqpace
		NOTE:Notes!!!!!!!!!!!!!!!
		X-AIM;type=HOME;type=pref:aim
		IMPP;X-SERVICE-TYPE=AIM;type=HOME;type=pref:aim:aim
		IMPP;X-SERVICE-TYPE=Facebook;type=HOME:xmpp:facebook
		REV:2014-04-14T07:26:54Z
		UID:7e68ff0b-d009-499d-a4c8-45f634f0c7bd
		END:VCARD
	 */

	public function createCard($addressBookId, $cardUri, $cardData) {
		$id = create_guid();
		$contact = \RowUpdate::blank_for_model('Contact', \AppConfig::current_user_id());
		$contact->set('assigned_user_id', \AppConfig::current_user_id());

		$ret = \vCardUtils::vcard_import($cardData, $contact);

		if (!$ret)
			return null;

		if ($ret instanceof \RowUpdate || $ret instanceof \RowResult) {
			$contact->set('primary_account_id', $ret->getField('id'));
			if ($ret instanceof \RowUpdate)
				$ret->save();
		}
		$contact->set('id', $id);
		$contact->set('vcard_uri', $cardUri);
		$contact->set('vcard', $cardData);
		$contact->set('vcard_size', strlen($cardData));
		$contact->save();

		$type = \UserPreference::getPreference('carddav_contacts_which');

		if ($type == 'fav') {
	        $favorite = new \Favorite;
    	    $favorite->addOrRemoveNew($contact->getField('id'), $contact->getModuleDir());
		}

		$eTag = '"'.md5($cardData).'"';

		return $eTag;
	}

    /**
     * Updates a card.
     *
     * The addressbook id will be passed as the first argument. This is the
     * same id as it is returned from the getAddressbooksForUser method.
     *
     * The cardUri is a base uri, and doesn't include the full path. The
     * cardData argument is the vcard body, and is passed as a string.
     *
     * It is possible to return an ETag from this method. This ETag should
     * match that of the updated resource, and must be enclosed with double
     * quotes (that is: the string itself must contain the actual quotes).
     *
     * You should only return the ETag if you store the carddata as-is. If a
     * subsequent GET request on the same card does not have the same body,
     * byte-by-byte and you did return an ETag here, clients tend to get
     * confused.
     *
     * If you don't return an ETag, you can just return null.
     *
     * @param mixed $addressBookId
     * @param string $cardUri
     * @param string $cardData
     * @return string|null
     */
	public function updateCard($addressBookId, $cardUri, $cardData) {
		$id = preg_replace('~\\.vcf$~', '', $cardUri);
		$result = \ListQuery::quick_fetch_key('Contact', 'vcard_uri', $cardUri, true, ['acl_checks' => 'edit']);

		if(! $result) {
			$result = \ListQuery::quick_fetch('Contact', $id, true, ['acl_checks' => 'edit']);
		}

		if(! $result || ! $result->getAclAllowed('edit')) {
			return null;
		}

		$contact = \RowUpdate::for_result($result);
		$ret = \vCardUtils::vcard_import($cardData, $contact);

		if (!$ret)
			return null;

		if ($ret instanceof \RowUpdate || $ret instanceof \RowResult) {
			$contact->set('primary_account_id', $ret->getField('id'));
			if ($ret instanceof \RowUpdate)
				$ret->save();
		}

		$contact->set('vcard', $cardData);
		$contact->set('vcard_size', strlen($cardData));
		$contact->save();

		$eTag = '"'.md5($cardData).'"';

		return $eTag;
    }

	function hasProperty($elt, $name, $value = null, $case = false) {
		$props = $elt[$name];
		if (!$props)
			return false;
		if (is_null($value))
			return true;
		if (!is_array($props))
			$props = [$props];
		foreach ($props as $prop) {
			$val = (string)$prop;
			if (!$case) {
				$val = strtoupper($val);
				$value = strtoupper($value);
			}
			if ($val === $value)
				return true;
		}
		return false;
	}

    /**
     * Deletes a card
     *
     * @param mixed $addressBookId
     * @param string $cardUri
     * @return bool
     */
	public function deleteCard($addressBookId, $cardUri) {
		$id = preg_replace('~\\.vcf$~', '', $cardUri);
		$result = \ListQuery::quick_fetch_key('Contact', 'vcard_uri', $cardUri, ['id', 'deleted'], ['acl_checks' => 'delete']);

		if(! $result) {
			$result = \ListQuery::quick_fetch('Contact', $id, ['id', 'deleted'], ['acl_checks' => 'delete']);
		}

		if(! $result || ! $result->getAclAllowed('delete'))
			return false;

		$contact = \RowUpdate::for_result($result);
		$ret = !!$contact->markDeleted(true);

		return $ret;
    }

    /**
     * The getChanges method returns all the changes that have happened, since
     * the specified syncToken in the specified address book.
     *
     * This function should return an array, such as the following:
     *
     * [
     *   'syncToken' => 'The current synctoken',
     *   'added'   => [
     *      'new.txt',
     *   ],
     *   'modified'   => [
     *      'updated.txt',
     *   ],
     *   'deleted' => [
     *      'foo.php.bak',
     *      'old.txt'
     *   ]
     * ];
     *
     * The returned syncToken property should reflect the *current* syncToken
     * of the addressbook, as reported in the {DAV:}sync-token property This is
     * needed here too, to ensure the operation is atomic.
     *
     * If the $syncToken argument is specified as null, this is an initial
     * sync, and all members should be reported.
     *
     * The modified property is an array of nodenames that have changed since
     * the last token.
     *
     * The deleted property is an array with nodenames, that have been deleted
     * from collection.
     *
     * The $syncLevel argument is basically the 'depth' of the report. If it's
     * 1, you only have to report changes that happened only directly in
     * immediate descendants. If it's 2, it should also include changes from
     * the nodes below the child collections. (grandchildren)
     *
     * The $limit argument allows a client to specify how many results should
     * be returned at most. If the limit is not specified, it should be treated
     * as infinite.
     *
     * If the limit (infinite or not) is higher than you're willing to return,
     * you should throw a Sabre\DAV\Exception\TooMuchMatches() exception.
     *
     * If the syncToken is expired (due to data cleanup) or unknown, you must
     * return null.
     *
     * The limit is 'suggestive'. You are free to ignore it.
     *
     * @param string $addressBookId
     * @param string $syncToken
     * @param int $syncLevel
     * @param int $limit
     * @return array
     * @throws \IAHListQueryError
     */
    public function getChangesForAddressBook($addressBookId, $syncToken, $syncLevel, $limit = null) {
		$currentToken = $this->getSyncToken();
	    $prevSyncDateTime = $this->getSyncDate($syncToken);

	    if (!$currentToken || !$prevSyncDateTime)
	    	return null;

	    $result = [
            'added'     => [],
            'modified'  => [],
            'deleted'   => [],
        ];
		
		$acl_updated = \AppConfig::setting('misc.acl_updated');
		$type_updated = \UserPreference::getPreference('carddav_pref_updated');
		$type = \UserPreference::getPreference('carddav_contacts_which');

		$globalResyncStatus = $this->getResyncStatus();

	    if ($globalResyncStatus) {
			$resync = true;
			$this->setResynced();
		} else {
			$resync = $syncToken && ($acl_updated >= $prevSyncDateTime || $type_updated >= $prevSyncDateTime);
		}

		$date_clause = array('operator' => 'OR', 'multiple' => array());

		$lq = new \ListQuery('Contact', null, ['acl_checks' => 'list', 'filter_deleted' => false]);

		if ($syncToken && !$resync) {
			$date_clause['multiple'][] = array(
				'value' => $prevSyncDateTime,
				'field' => 'date_modified',
				'operator' => '>=',
			);
		}

		if (!$limit)
			$limit = null;

		$lq->setOrderBy('date_modified');

		$lq->addField('id');
		$lq->addField('assigned_user_id');
		$lq->addField('date_entered');
		$lq->addField('date_modified');
		$lq->addField('deleted');
		$lq->addFieldLiteral('some_vcard_uri', "IFNULL(vcard_uri,CONCAT(" . $lq->getTableAlias(null, true) . ".id, '.vcf'))");

		if ($type == 'fav') {
			\Favorite::add_join($lq);
			$lq->addField('favorites.module_record_id');

			if ($syncToken && !$resync) {
				$date_clause['multiple'][] = array(
					'value' => $prevSyncDateTime,
					'field' => 'favorites.date_modified',
					'operator' => '>=',
				);
			}
		}

		if (!empty($date_clause['multiple']))
			$lq->addFilterClause($date_clause);

		if (!$resync) {
			$lq->addAclFilter('list');
			switch ($type) {
			case 'fav':
			case 'visible':
				break;
			default:
				$lq->addSimpleFilter('assigned_user_id', \AppConfig::current_user_id(), '=');
				break;
			}
		}

		$res = $lq->runQuery(0, $limit, false, null, 100);

		while (!$res->failed) {
			foreach ($res->getRowIndexes() as $idx) {
				$row = $res->getRowResult($idx);
				$deleted = $row->getField('deleted') || !$row->getAclAllowed('list');
				switch ($type) {
				case 'fav':
					if ($row->getField('favorite') || !$row->getField('favorites.module_record_id'))
						$deleted = true;
					break;
				case 'visible':
					break;
				default:
					if ($row->getField('assigned_user_id') != \AppConfig::current_user_id())
						$deleted = true;
				}
				if ($deleted)
					$result['deleted'][] = $row->getField('some_vcard_uri');
				else if ($row->getField('date_entered') == $row->getField('date_modified'))
					$result['added'][] = $row->getField('some_vcard_uri');
				else
					$result['modified'][] = $row->getField('some_vcard_uri');
			}

			if ($res->page_finished)
				break;

			$lq->pageResult($res);
		}
		
		$mCount = is_countable($result["modified"]) ? sizeof($result["modified"]) : 0;
		$aCount = is_countable($result["added"]) ? sizeof($result["added"]) : 0;
		$dCount = is_countable($result["deleted"]) ? sizeof($result["deleted"]) : 0;
		$GLOBALS['log']->debug("CardDAV Sync: Modified: ". $mCount . " Added: " . $aCount . " Deleted: " . $dCount);

		$result['syncToken'] = $currentToken;

        return $result;
    }

	/**
	 * If CardDAV preferences or ACL were updated the method will update sync token and return it.
	 * Returns given sync token otherwise.
	 *
	 * @param int|null $currentToken
	 *
	 * @return int|null
	 */
	private function checkPrefUpdates($currentToken) {
		$aclUpdated = \AppConfig::setting('misc.acl_updated');
		$typeUpdated = \UserPreference::getPreference('carddav_pref_updated');
		$currentSyncDateTime = $this->getSyncDate($currentToken);

		$result = $currentToken;

		if ($currentToken && ($aclUpdated >= $currentSyncDateTime || $typeUpdated >= $currentSyncDateTime)) {
			$this->updateSyncToken();
			$result = $this->getSyncToken();
		}

		return $result;
	}

	/**
	 * @return int|null
	 */
    private function getSyncToken() {
    	return \vCardUtils::get_sync_token();
    }

	/**
	 * @param int|string $token
	 *
	 * @return false|string
	 */
    private function getSyncDate($token) {
	    return \vCardUtils::convert_sync_token_to_date($token);
    }

	private function updateSyncToken() {
		\vCardUtils::update_sync_token();
	}

	/**
	 * @return bool
	 */
	private function getResyncStatus() {
		return \vCardUtils::get_resync_value();
	}

	private function setResynced() {
		require_bean("User");
		$user = new \User();
		$user->id = \AppConfig::current_user_id();

		\vCardUtils::set_resync_value(false, $user);
	}
}
