// vim: set filetype=markdown tw=80:
## Extending 1CRM API

This article describes how developers of 1CRM extension modules can extend the 
API. Let's suppose that you created a custom module named `Echo`.

### Registering an API group

Each endpoint in 1CRM API belongs to a group. Grouping does not affect the API
functionality, but it is useful for API documentation formatting. We start by
creating a file named `modules/Echo/ext/config/standard/api_config.php`:

```
groups
	myutils
		title: "My Utils"
		docs: "Miscellanous useful utilities"

```

### Registering endpoints

Next step is to register an endpoint. 

```
groups
	myutils
		title: My Utils
		docs: "Miscellanous useful utilities"
endpoints
	--
		file: modules/Echo/API.php
		class: EchoEndpoint
		group: myutils

```
This new config file section adds an endpoint to `myutils` group. The endpoint
will be defined in `EchoEndpoint` class, located in `modules/Echo/API.php`
file.

### Endpoint class

Endpoint class must extend `OneCRM\API\Endpoint`:

``` php
<?php

use OneCRM\API\Endpoint;

class EchoEndpoint extends Endpoint {
}
```

#### Required methods

First, we need to define several required methods:

``` php
<?php

use OneCRM\API\Endpoint;

class EchoEndpoint extends Endpoint {

	public function allowedMethods() {
		return ['post'];
	}

	public function isAuthEndpoint() {
		return false;
	}

	public function requiredScopes($method) {
		return ['read'];
	}

	public function descriptor($method) {
		if ($method == 'post')
			return $this->descriptorPost();
	}

	public function endpoint() {
		return '/ext/echo';
	}

}

```

This endpoint uses `/ext/echo` path.
Each endpoint can support multiple HTTP request methods (GET,POST, PUT, etc.),
so we have to return all method it supports from allowedMethods(). In this case,
the only supported HTTP method is POST.

Authentication scope required by this endpoint is `read`. You may also specify
`write` and `profile` if needed.

`isAuthEndpoint()` should always return false.

#### Endpoint descriptor

`descriptor()` method must return endpoint descriptor describing endpoint
parameters and returned value. In our case, it will return a descriptor only for
HTTP POST method, because that is the only one supported.

We start with a minimal descriptor containing only a title and documentation.
Notice that here we reference documentation placed into a separate file —
`modules/Echo/docs/docs.md`.

``` php
<?php

use OneCRM\API\Endpoint;

class EchoEndpoint extends Endpoint {
	// ..... skipped ....

	private function descriptorPost() {
		return [
			'title' => 'Echo',
			'docs' => '@+modules/Echo/docs/docs', // docs in a .md file, notice that extension is omited
		];
	}
}

```

#### Parameters

An endpoint can accept a number of parameters that can be passed when making an
API call

``` php
<?php

use OneCRM\API\Endpoint;

class EchoEndpoint extends Endpoint {
	// ..... skipped ....

	private function descriptorPost() {
		return [
			'title' => 'Echo',
			'docs' => '@+modules/Echo/docs/docs', // docs in a .md file, notice that extension is omited
			'params' => [
				'data' => [
					'location' => 'body',
					'type' => 'Object',
					'docs' => 'Data to echo',
					'required' => true,
				],
				'format' => [
					'location' => 'query',
					'type' => 'Enum',
					'options' => ['json', 'xml'],
					'docs' => 'Output format',
					'required' => false,
				],
			],
			'return' => [
				'' => [
					'type' => 'String',
					'docs' => 'Data echoed back',
					'required' => true,
				],
			],
		];
	}
}

```

Here we declared a required parameter `data` which must be passed in request
body, and an optional parameter `format` which can be passed in query string.
These parameters will be checked when serving API calls, and an error will be
returned if passed parameters do not match the specification.

You can also describe the data returned by the endpoint, but this serves for
documentation purposes only.

#### HTTP handler

Now it is time to create a method that will process requests and return
responses. The method name must match corresponding HTTP method.

``` php
<?php

use OneCRM\API\Endpoint;

class EchoEndpoint extends Endpoint {
	// ..... skipped ....
	protected function post($input, $res, $req) {

		if ($input['format'] != 'xml') {
            $res->getBody()->write(json_encode($input['data']));
		    return $res->withHeader('Content-Type', 'application/json');
        }
        
        $body = $res->getBody();
        $body->rewind();

		$xml = new SimpleXMLElement('<echo/>');
		$body->write($this->arrayToXML($input['data'], $xml, 'element'));

		return $res->withHeader('Content-type', 'text/xml; charset=UTF-8')->withBody($body);
    }

	public function arrayToXML($array, SimpleXMLElement $xml, $child_name)
	{
    	foreach ($array as $k => $v) {
    	    if(is_array($v)) {
        	    (is_int($k)) ? $this->arrayToXML($v, $xml->addChild($child_name), $v) : $this->arrayToXML($v, $xml->addChild(strtolower($k)), $child_name);
	        } else {
    	        (is_int($k)) ? $xml->addChild($child_name, $v) : $xml->addChild(strtolower($k), $v);
        	}
	    }

    	return $xml->asXML();
	}

}

```

This is just an example, normally you want to always use:
``` php 
$response->getBody()->write(json_encode($data));
return $response->withHeader('Content-Type', 'application/json');
```
1CRM API uses 
<a href="https://www.slimframework.com" target="_blank">Slim framework</a>.
Refer to its documentation to learn how to work with Request and Response
objects.

#### Returning errors

If during a request processing an error occurs, you can throw one of
`BadRequest`, `InternalError`, `NotFound` from `OneCRM\API\Errors` namespace.

``` php
<?php

use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class EchoEndpoint extends Endpoint {
	// ..... skipped ....
	protected function post($input, $res, $req) {

		if (!empty($input['data']['prohibited'])) 
			throw new Errors\BadRequest("prohibited parameter passed");
		if ($input['format'] != 'xml') {
            $res->getBody()->write(json_encode($input['data']));
		    return $res->withHeader('Content-Type', 'application/json');
        }

        $body = $res->getBody();
        $body->rewind();

		$xml = new SimpleXMLElement('<echo/>');
		$body->write($this->arrayToXML($input['data'], $xml, 'element'));

		return $res->withHeader('Content-type', 'text/xml; charset=UTF-8')->withBody($body);
    }
	// ..... skipped ....
}

```
