// vim: set filetype=markdown tw=80:

## Creating webhooks

This page provides some examples of webhook filters construction. For the
examples, `Contact` model is used.


### Basic filter structure

Webhook filter consists of 2 elements: `conditions` and `glue`. Conditions are
an array of conditions to check, and `glue` tells how the conditions are to be
combined.

`glue` can be either `and` or `or`.  If `glue` is `and`, and __all __conditions
are satisfied, the webhook triggers. If `glue` is `or`, and __at
least one__ condition is satisfied, the webhook triggers.

Basic filter structure:

<pre>
{
	glue: "or", // "or", "and"
	conditions: [
		{
			field: "first_name",
			filter: "eq",
			value: "John",
			when: "current_value"
		}
	]
}
</pre>

This simple filter triggers the webhook if after update `first_name` equals to "John". Note
that in this case `glue` can be either `or` or `and` - it does not matter when
there is only one condition.

### Combining multiple conditions

In previous example, the webhook would trigger if contact's first name is
"John". Now imagine you want to trigger a webhook for every update to a contact
whose first name is John, but only if their last name was changed to Smith. In
this case the filter will look like this:

<pre>
{
	glue: "and",
	conditions: [
		{
			field: "first_name",
			filter: "eq",
			value: "John",
			when: "current_value"
		},
		{
			field: "last_name",
			filter: "eq",
			value: "Smith",
			when: "changed_to"
		}
	]
}
</pre>

You can add as many conditions as you need:

<pre>
{
	glue: "or",
	conditions: [
		{
			field: "last_name",
			filter: "eq",
			value: "Smith",
			when: "changed_to"
		},
		{
			field: "last_name",
			filter: "eq",
			value: "Johnson",
			when: "changed_to"
		},
		{
			field: "last_name",
			filter: "eq",
			value: "Black",
			when: "changed_to"
		}
	]
}
</pre>

### Choosing values to check

As you must have noticed, we used different values for `when` in the conditions.
This field tells which value you want to check: the old value (before the update)
or the new value (after the update).

Let's see how to use different `when` values.


<pre>
{
	glue: "or",
	conditions: [
		{
			field: "last_name",
			filter: "eq",
			value: "Smith",
			when: "changed_to"
		},
	]
}
</pre>

With this filter, the webhook will trigger when last name changed to "Smith".

<pre>
{
	glue: "or",
	conditions: [
		{
			field: "last_name",
			filter: "ne",
			value: "Smith",
			when: "changed_to"
		},
	]
}
</pre>

With this filter, the webhook will trigger when last name changed to anything
other than "Smith". Note - if the name was not equal to "Smith" and it was not
changed - the condition is not satisfied!

<pre>
{
	glue: "or",
	conditions: [
		{
			field: "last_name",
			filter: "ne",
			value: "Smith",
			when: "current_value"
		},
	]
}
</pre>

With this filter, the webhook will trigger when last name after the update is
not equal to "Smith". It does not matter if the last name was updated or not -
the value after the update is tested.

<pre>
{
	glue: "or",
	conditions: [
		{
			field: "last_name",
			filter: "eq",
			value: "Smith",
			when: "previous_value"
		},
	]
}
</pre>

With this filter, the webhook will trigger when last name was changed, and after
the update it is equal to "Smith".

<pre>
{
	glue: "or",
	conditions: [
		{
			field: "last_name",
			when: "changed"
		},
	]
}
</pre>

With this filter, the webhook will trigger when last name was changed. The
actual new value does not matter, and the condition does not need `filter` and
`value`.

<pre>
{
	glue: "or",
	conditions: [
		{
			field: "last_name",
			when: "not_changed"
		},
	]
}
</pre>

With this filter, the webhook will trigger when last name was not changed. This
 condition does not need `filter` and `value`.

### Operators

All conditions except `changed` and `not_changed` need an operator, passed in
`filter`. Applicable operators depend on field type. 

See [NumberFilter](?#type_NumberFilter),
[DatetimeFilter](?#type_DatetimeFilter),
[EnumFilter](?#type_EnumFilter),
[BoolFilterr](?#type_BoolFilter),
[StringFilter](?#type_StringFilter)
for operators available for different field types.

### Nested filters

As you can see, pretty sophisticated filters can be constructed. But there are
cases when simply combining multiple conditions with `and` or `or` is not
enough. In such cases, nested filters can be used. Let's continue with our
Contact examples. Imagine you need a webhook triggered after contact's name is
changed to either "John Smith" or "Jack Brown". In this case you construct your
filter as follows:

<pre>
{
	glue: "and",
	conditions: [
		{
			glue: "or",
			conditions: [
				{
					field: "last_name",
					when: "changed"
				},
				{
					field: "first_name",
					when: "changed"
				}
			]
		},
		{
			glue: "or",
			conditions: [
				{
					glue: "and",
					conditions: [
						{
							field: "last_name",
							when: "current_value",
							filter: "eq",
							value: "Smith"
						},
						{
							field: "first_name",
							when: "current_value",
							filter: "eq",
							value: "John"
						}
					]
				},
				{
					glue: "and",
					conditions: [
						{
							field: "last_name",
							when: "current_value",
							filter: "eq",
							value: "Brown"
						},
						{
							field: "first_name",
							when: "current_value",
							filter: "eq",
							value: "Jack"
						}
					]
				}
			]
		},
	]
}

