// vim: set filetype=markdown tw=80:
## Using OAuth 2.0

### Application registration

In order to make calls to the 1CRM API using OAuth 2.0, you need to register an
application, or API client. Registered clients are assigned a unique Client ID
(client\_id) and a unique Client Secret (client\_secret). Make sure to store the
Client Secret securely.

1CRM suports 2 types of API clients: public and private. 

Private API clients can be registered by 1CRM administrator, and are intended for
use by your organization only. Public API clients are managed by 1CRM Corp.,
and represent 3rd party applications. If you are developing an application that
can be potentially useful for all 1CRM customers, you should apply for a public
API client.

Client ID and Client Secret for private clients are available from 1CRM admin
interface. For public clients, 1CRM will provide Client ID and Client Secret
after creating your API CLient.

### Resource owners

One of roles defined by OAuth 2.0 Authorization Framework is Resource Owner:

> An entity capable of granting access to a protected resource.
> When the resource owner is a person, it is referred to as an
> end-user. 

Naturally, in 1CRM, its users are resource owners. API client can be granted
access on behalf of a 1CRM user. But there are many cases, when it is also
desirable to allow Contacts to authenticate themselves: using 1CRM as single
sign-on server, customer portals, etc.

1CRM defines two types of Resource Owners: User and Contact. Depending on type
of resource owner you want to authenticate, you use different endpoints. Note
that Contact access token only provides access to Contact's own information such
as name, email address and telephone number.

### Scopes

When requesting access to protected resources, API client can specify the scope
of request. 1CRM supports 3 scopes:

* `read` client requests access to read information from 1CRM
* `write` client requests access to write information to 1CRM
* `profile` client requests access to Resource Owner's account information

For Resource Owner of Contact type, only `profile` scope is allowed. `read` and
`write` scopes are subject to limitations set by 1CRM administrator using 1CRM
ACL system.

If you do not specify a scope in authorization request, `profile` will be used
by default.

### Authentication workflow

1CRM allows you to use OAuth 2.0 via one of five flows:

* Authorization Code Grant - use this flow if you want to authorize a web
  application.
* Implicit Grant - use this flow if you want to authorize a user-agent, a desktop
  or a mobile application.
* Resource Owner Password Credentials Grant - use this flow if you want to
  authenticate directly via 1CRM API using resource owner’s password credentials.
* Client Credentials Grant - use this flow to authorize an application on behalf
  of predefined 1CRM user. The actual user that gains access in this flow is
  decided by 1CRM administrator.
* Refresh Token Grant - use this flow to renew an access token when the current
  one expires.

#### Authorization Code grant

Authorization Code grant is most suitable for web server applications or native
mobile applications. The flow should be very familiar if you have ever signed
into a web app using your Facebook or Google account.

##### Step one

The client will redirect the user's web browser to `/auth/user/authorize` or 
`/auth/contact/authorize` with the following parameters in the query string:

* `response_type` with the value `code`
* `client_id` with the client identifier
* `redirect_uri` with the client redirect URI. This parameter is optional,
  and defaults to first URI from the list configured by 1CRM administrator. If
  specified, this parameter must match one of redirect URIs configured by 1CRM
  administrator.
* `state` optional CSRF token. This parameter is optional but highly recommended. You
  should store the value of the CSRF token in the user’s session to be validated
  when they return to your application

The user will be asked to login and approve the client. If the user approves the
client, they will be redirected to redirect URI, with the following parameters
in the query string:


* `code` authorization code
* `state` state parameter sent in the original request. You should compare
  this value with the value stored in the user’s session to ensure the
  authorization code obtained is in response to requests made by this client
  rather than another client application.

##### Step two

The client now sends a POST request to `/auth/user/access_token` or 
`/auth/contact/access_token` with the following parameters in request body:

* `grant_type` with the value of `authorization_code`
* `client_id` with the client identifier
* `scope` a space delimited list of scopes
* `client_secret` with the client secret
* `redirect_uri` with the same redirect URI the user was redirect back to
* `code` with the authorization code from the query string

Instead of specifying `client_id` and `client_secret` in request body, you can
send HTTP Basic Authorization header, using the client ID as username and client
secret as password.

The server will respond with a JSON object containing the
following properties:

* `token_type` with the value `Bearer`
* `expires_in` with an integer representing the TTL of the access token
* `access_token` access token - a JWT signed with the server’s private key
* `refresh_token` an encrypted payload that can be used to refresh the access token
  when it expires.

Access token can be now used to make calls to protected endpoints.

#### Implicit Grant

The implicit grant is similar to the authorization code grant with two distinct
differences.

It is intended to be used for user-agent-based clients (e.g. single page web
apps) that can’t keep a client secret because all of the application code and
storage is easily accessible.

Secondly instead of the authorization server returning an authorization code
which is exchanged for an access token, the authorization server returns an
access token.

##### Flow

The client will redirect the user's web browser to `/auth/user/authorize` or 
`/auth/contact/authorize` with the following parameters in the query string:

* `response_type` with the value `token`
* `client_id` with the client identifier
* `scope` a space delimited list of scopes
* `redirect_uri` with the client redirect URI. This parameter is optional,
  and defaults to first URI from the list configured by 1CRM administrator. If
  specified, this parameter must match one of redirect URIs configured by 1CRM
  administrator.
* `state` optional CSRF token. This parameter is optional but highly recommended. You
  should store the value of the CSRF token in the user’s session to be validated
  when they return to your application

The user will be asked to login and approve the client. If the user approves the
client, they will be redirected to redirect URI, with the following parameters
in the query string:


* `token_type` with the value `Bearer`
* `expires_in` with an integer representing the TTL of the access token
* `access_token` access token - a JWT signed with the authorization server’s private key
* `state` with the state parameter sent in the original request. You should compare
  this value with the value stored in the user’s session to ensure the
  authorization code obtained is in response to requests made by this client
  rather than another client application.

Note : this grant does not return refresh token

#### Resource Owner Password Credentials Grant

This grant provides great user experience for web applications and native mobile
applications, because it does not require the user to be redirected to 1CRM for
password entry. This grant is only available for private clients.

##### Flow

The client will send a a POST request to `/auth/user/access_token` or 
`/auth/contact/access_token` with the following parameters in request body:

* `grant_type` with the value of `password`
* `client_id` with the client identifier
* `scope` a space delimited list of scopes
* `client_secret` with the client secret
* `username`
* `password`

Instead of specifying `client_id` and `client_secret` in request body, you can
send HTTP Basic Authorization header, using the client ID as username and client
secret as password.

The server will respond with a JSON object containing the
following properties:

* `token_type` with the value `Bearer`
* `expires_in` with an integer representing the TTL of the access token
* `access_token` access token - a JWT signed with the server’s private key
* `refresh_token` an encrypted payload that can be used to refresh the access token
  when it expires.

Access token can be now used to make calls to protected endpoints.

#### Client Credentials Grant

To enable this grant, 1CRM administrator assigns a user to API client using 1CRM
admin interface. After obtaining access token, API client can make requests to
the API on behalf of that user. This grant is only available for Resource Owner
of User type.

##### Flow

The client will senda a POST request to `/auth/user/access_token` or 
with the following parameters in request body:

* `grant_type` with the value of `client_credentials`
* `client_id` with the client identifier
* `scope` a space delimited list of scopes
* `client_secret` with the client secret

Instead of specifying `client_id` and `client_secret` in request body, you can
send HTTP Basic Authorization header, using the client ID as username and client
secret as password.

The server will respond with a JSON object containing the
following properties:

* `token_type` with the value `Bearer`
* `expires_in` with an integer representing the TTL of the access token
* `access_token` access token - a JWT signed with the server’s private key

Access token can be now used to make calls to protected endpoints.

#### Refresh Token Grant

Access tokens eventually expire; however some grants respond with a refresh
token which enables the client to refresh the access token

##### Flow

The client will senda a POST request to `/auth/user/access_token` or 
`/auth/contact/access_token` with the following parameters in request body:

* `grant_type` with the value of `refresh_token`
* `refresh_token` with the refresh token
* `client_id` with the client identifier
* `scope` a space delimited list of scopes
* `client_secret` with the client secret

Instead of specifying `client_id` and `client_secret` in request body, you can
send HTTP Basic Authorization header, using the client ID as username and client
secret as password.

The server will respond with a JSON object containing the
following properties:

* `token_type` with the value `Bearer`
* `expires_in` with an integer representing the TTL of the access token
* `access_token` access token - a JWT signed with the server’s private key
* `refresh_token` an encrypted payload that can be used to refresh the access token
  when it expires.

