<?php

namespace OneCRM\API;

use JetBrains\PhpStorm\NoReturn;
use Psr\Container\ContainerInterface;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;

use Slim\App;
use Slim\Routing\RouteContext;

use Psr\Container\ContainerExceptionInterface as ContainerException;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;

require_once dirname(__FILE__) . '/Parsedown.php';

class DocsGenerator {

	private $container;
	private $docsDir;

	public function __construct(ContainerInterface $container)
	{
		$this->container = $container;
	}

	public function __invoke(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface {
		$this->docsDir = $request->getAttribute('docsDir');
		$this->generateHeader($request, $response);
		$this->generateNavigation($request, $response);
		$this->generateContent($request, $response);
		$this->generateFooter($request, $response);

		return $response;
	}

	private function generateNavigation(Request $request, Response $response) {
		$params = $request->getQueryParams();
		if (isset($params['page'])) {
			$routeContext = RouteContext::fromRequest($request);
			$path = $routeContext->getRouteParser()->urlFor('docs');

			$html = '<ul class="sidenav">';
			$html .= '<li class="item"><a href="' . $path . '">&laquo; Back to documentation</a></li>';
			$html .= '</ul>';
			$response->getBody()->write($html);
			return;
		}

		$registry = $this->container->get('registry')->getRegistry();

		$html = '<ul class="sidenav">';

		foreach ($registry['groups'] as $gid => $g) {
			if (!empty($g['hidden']) && empty($_GET['hidden']))
				continue;
			$html .= '<li class="group"><a href="#group_' . $gid . '">' . $g['title'] . '</a></li>';

			if (isset($g['articles'])) {
				foreach ($g['articles'] as $aid => $a) {
					if (!empty($a['page']))
						$html .= '<li class="item"><a href="?page=' . $a['page'] . '">' . $a['title'] . '</a></li>';
					else
						$html .= '<li class="item"><a href="#article_' . $gid . '_' . $aid . '">' . $a['title'] . '</a></li>';
				}
			}

			if ($gid == 'types') {
				$types = $this->container->get('registry')->getTypes();
				foreach ($types as $type) {
					$name = $type->getName();
					$html .= '<li class="item"><a href="#type_' . $name . '">' . $name . '</a></li>';
				}
			}

			foreach ($registry['endpoints'] as $epName => $epMeta) {
				if (!empty($epMeta['hidden']) && empty($_GET['hidden']))
					continue;
				if ($epMeta['group'] == $gid) {
					$ep = $this->container->get('registry')->getEndpointByName($epName);
					if ($ep) {
						foreach ($ep->allowedMethods() as $method) {
							$desc =  $ep->descriptor($method);
							if (!empty($desc['hidden']) && empty($_GET['hidden']))
								continue;
							$html .= '<li class="item"><a href="#endpoint_' . $epName . '_' . $method . '">' . $desc['title'] . '</a></li>';
						}
					} else {
					}
				}
			}

		}

		$html .= '</ul>';
		$response->getBody()->write($html);
	}

	private function generateContent(Request $request, Response $response) {
		$params = $request->getQueryParams();
		if(isset($params['page'])) {
			$page = preg_replace('~[^a-z0-9_-]~', '', $params['page']);
			if (strlen($page)) {
				return $this->generatePageContent($request, $response, $page);
			}
		}

		return $this->generateMainContent($request, $response);
	}


	private function generatePageContent(Request $request, Response $response, $page) {
		$html = '<div id="content">';
		$html .= '<h1 style="text-align:center">1CRM REST API</h1>';
		$html .= $this->parseDocs('@pages/' . $page);
		$html .= '</div>';
		$response->getBody()->write($html);
	}

	private function generateMainContent(Request $request, Response $response) {
		$registry = $this->container->get('registry')->getRegistry();

		$html = '<div id="content">';
		$html .= '<h1 style="text-align:center">1CRM REST API</h1>';

		foreach ($registry['groups'] as $gid => $g) {
			if (!empty($g['hidden']) && empty($_GET['hidden'])) continue;
			$html .= '<a name="group_' . $gid . '"></a><h2>' . $g['title'] . '</h2>';

			if (isset($g['docs']))
				$html .= $this->parseDocs($g['docs']);

			if (isset($g['articles'])) {
				foreach ($g['articles'] as $aid => $a) {
					$html .= '<a name="article_' . $gid . '_' . $aid . '"></a><h3>' . $a['title'] . '</h3>';
					if (isset($a['docs']))
						$html .= $this->parseDocs($a['docs']);
				}
			}

			if ($gid == 'types') {
				$types = $this->container->get('registry')->getTypes();
				foreach ($types as $type) {
					$name = $type->getName();
					$html .= '<a name="type_' . $name . '"></a><h3>' . $name . '</h3>';
					if (!$type->isBasic()) {
						$parentType = get_parent_class($type);
						$parentName = str_replace('OneCRM\\API\\Types\\', '', $parentType);
						$parentName = preg_replace('~Type$~', '', $parentName);
						$def = ['type' => $parentName];
						$type->expandComplexDef($def);
						$html .= '<p>Inherits ' . $this->formatType($def);
					}
					$html .= $this->parseDocs($type->getDocs());
					if (!$type->isBasic()) {
						if(isset($def['schema']))
						$html .= $this->formatParameters($def['schema'], 'Schema');
					}
				}
			}

			foreach ($registry['endpoints'] as $epName => $epMeta) {
				if ($epMeta['group'] == $gid) {
					$ep = $this->container->get('registry')->getEndpointByName($epName);
					if ($ep) {
						foreach ($ep->allowedMethods() as $method) {
							$desc =  $ep->descriptor($method);
							if (!empty($desc['hidden']) && empty($_GET['hidden']))
								continue;
							$auth = $ep->requiresAuthentication();
							$html .= '<a name="endpoint_' . $epName . '_' . $method . '"></a><h3>' . $desc['title'] . '</h3>';
							$html .= '<pre class="http-method" data-method="' . $method . '">';
							if ($auth)
								$html .= '<i class="glyphicon glyphicon-lock"></i>&nbsp;';
							$html .= $ep->endpoint() . '</pre>';
							if (isset($desc['docs']))
								$html .= $this->parseDocs($desc['docs']);
							$html .= $this->formatParameters($desc['params'], 'Parameters');
							if (isset($desc['return']))
								$html .= $this->formatParameters($desc['return'], 'Return');
						}
					}
				}
			}
		}

		$html .= '</div>';
		$response->getBody()->write($html);
	}

	private function generateHeader(Request $request, Response $response) {
		$docsDir = $request->getAttribute('docsDir');
		$templateDir = $request->getAttribute('templateDir');
		$style = file_get_contents($templateDir . '/style.css');
		$bs_css = file_get_contents($templateDir . '/bootstrap.min.css');
		$bs_css = str_replace('../fonts/', \AppConfig::site_url() . '/include/api/template/', $bs_css);
		$html = <<<EOF
<!DOCTYPE html>
<html>
<head>
	<title>1CRM REST API documentation</title>
	<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
	<style type="text/css">$bs_css</style>
	<style type="text/css">$style</style>
</head>
<body>
<div id="container">
EOF;
		$response->getBody()->write($html);
	}

	private function generateFooter(Request $request, Response $response) {
		$year = date('Y');
		$html = <<<EOF


</div>
<div id="copyright">
<div>&copy; 2004-$year <a href="http://www.1crm.com/" target="_blank">1CRM Systems Corp.</a> All Rights Reserved.</div>
</div>
</body>
</html>

EOF;
		$response->getBody()->write($html);
	}

	private function formatParameters($params, $title) {
		if (empty($params))
			return '';
		$html = <<<EOT
<div class="panel panel-default">
  <div class="panel-heading"><b>$title</b></div>

<table class="table table-striped">
<thead>
<tr>
	<th>Name</th>
	<th>Type</th>
	<th>Description</th>
	<th style="width:80px"></th>
</tr>
</thead>
<tbody>
EOT;
		foreach ($params as $name => $def) {
			if (!empty($def['hidden']) && empty($_GET['hidden']))
				continue;
			$this->formatParameter($name, $def, $html);
		}
		$html .= '</tbody></table></div>';
		return $html;
	}

	private function formatParameter($name, $def, &$html, $level = 0) {
		$marker = isset($def['location']) ? '<span data-marker="' . $def['location']. '"></span>' : '';
		$marker .= (!empty($def['required']) ? '<span data-marker="required"></span>' : '');
		$docs = '';
		if (isset($def['docs']))
			$docs = $this->parseDocs($def['docs']);
		$type = $this->formatType($def);
		$html .= "<tr><td>$name</td><td style=\"width:45%\">$type</td><td style=\"width:45%\">$docs</td><td>$marker</td></tr>";

		if (isset($def['type']['element_type']))
			$this->formatParameter('<span style="visibility:hidden">' . str_repeat('↳', $level) . '</span>↳[n]', $def['type']['element_type'], $html, $level + 1);
		if (isset($def['schema'])) {
			$schema = $def['schema'];
			foreach ($schema as $sk => $sv) {
				$this->formatParameter('<span style="visibility:hidden">' . str_repeat('↳', $level) . "</span>↳[{$sk}]", $sv, $html, $level + 1);
			}
		}

	}

	private function formatType($def) {
		$typeName = (isset($def['type']) && is_array($def['type'])) ? $def['type']['type'] : '';

		$index = array_get_default($def, 'min', array_get_default($def, 'min_len', ''));
		if (strlen($index))
			$index .= ':';
		$max = array_get_default($def, 'max', array_get_default($def, 'max_len', ''));
		if (strlen($max)) {
			if (!strlen($index))
				$index = ':';
			$index = $index . $max;
		}
		if (strlen($index))
			$index = " [$index]";
		if (isset($def['default']))
			$index .= " ({$def['default']})";

		$vt = '';
		if (isset($def['value_type'])) {
			$valueType = $def['value_type'];
			if (is_array($valueType))
				$valueType = $valueType['type'];
			$vt = '&nbsp;{' . $this->formatType(['type' => $valueType]). '}';
		}
		if (isset($def['options'])) {
			$vt = '&nbsp;{' . join(', ', $def['options']) . '}';
		}

		if (isset($def['type']['choices'])) $choices = $def['type']['choices'];
		else $choices = array_get_default($def, 'choices');
		if (!empty($choices)) {
			$vt = '&nbsp;(' . join(', ', array_map([$this, 'formatType'], $choices)) . ')';
		}

		return '<a href="#type_' . $typeName . '">' . $typeName . '</a>' . $index . $vt;
	}

	private function parseDocs($md) {
		if (substr($md, 0, 1) === '@') {
			if (substr($md, 0, 2) === '@+') {
				$filename = substr($md, 2) . '.md';
			} else {
				$filename = $this->docsDir . '/' . substr($md, 1) . '.md';
			}
			if (is_file($filename))
				$md = file_get_contents($filename);
			else $md = '';
		}
		$md = preg_replace('~^//[^\r\n]*\r?\n~', '', $md);
		$base_uri = $_SERVER['SCRIPT_NAME'];
		$md = str_replace('{BASEURI}', $base_uri, $md);
		$md = str_replace('{HOST}', $_SERVER['HTTP_HOST'], $md);
		return \Parsedown::instance()->text($md);
	}


}

