<?php


namespace OneCRM\API;
use League\OAuth2\Server\ResourceServer;
use OneCRM\API\Errors;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;

abstract class Endpoint {

	static protected $disabled_modules = [
		'ACL',
		'ACLActions',
		'ACLRoles',
		'APClients',
		'Activities',
		'ActivityLog',
		'Administration',
		'Audit',
		'BookableObjects',
		'Calendar',
		'CampaignLog',
		'CampaignTrackers',
		'Charts',
		'CompanyAddress',
		'Configurator',
		'Dashboard',
		'DataPrivacy',
		'Directory',
		'DynFields',
		'EmailFolders',
		'EmailMan',
		'EmailPOP3',
		'Favorites',
		'Forecasts',
		'HR',
		'History',
		'Home',
		'ImportDB',
		'LayoutTest',
		'LicenseInfo',
		'MailMerge',
		'MergeRecords',
		'ModuleDesigner',
		'NewStudio',
		'OptimisticLock',
		'PDFPrint',
		'PDFTemplates',
		'PaymentsOut',
		'Recurrence',
		'Relationships',
		'ReportData',
		'Reports',
		'SavedSearch',
		'Scheduler',
		'SecurityGroups',
		'SocialAccounts',
		'SystemMessages',
		'UWizard',
		'UserPreferences',
		'Users',
		'Vacations',
		'Workflow',
		'iFrames',
		'vCals',
		'APIClients',
		'Webhooks',
	];

	protected $params = [];
	protected $app;

	public function __construct($app) {
		$this->app = $app;
	}

	public function getRouteName($method) {
	}

	public function __invoke(Request $request, Response $response, array $args): Response {
		list($request, $params) = Util::serveEndpoint($this, $request, $response, $args);
		$this->params=$params;
		$method = strtolower($request->getMethod());
		$request = $this->checkAuthentication($request);
		$scopes = $request->getAttribute('oauth_scopes', []);
		if (!$this->isAuthEndpoint()) {
			foreach ($this->requiredScopes($method) as $scope) {
				if (!in_array($scope, $scopes))
					throw new Errors\Unauthorized(translate('API_ERROR_WRONG_CREDENTIALS', 'app'));
			}
		}
        if ($this->providesQueryLog()){
            $logEntry=$this->queryLog($params,$request);
        }
        try {
            $result=$this->$method($params, $response, $request);
            if ($this->providesQueryLog()){
                $this->queryLog($params, $request, $logEntry, $result);
			}
			return $result;
        } catch (Error $e) {
			throw $e;
		} catch (\Exception $e) {
			throw new Errors\InternalError($e->getMessage());
		}
	}

	abstract public function endpoint();

	abstract public function descriptor($method);
	abstract public function allowedMethods();
	abstract public function requiredScopes($method);

	public function addMiddleware($app, $route) {
	}

	public function requiresAuthentication() {
		return true;
	}

	public function isAuthEndpoint() {
		return false;
	}

	protected function checkAuthentication($request) {
		if ($this->providesCustomAuthMethod()) {
			$result = $this->checkCustomAuthentication($request);
		} else {
			if (\AppConfig::setting('site.security.api_basic_auth')) {
				$result = $this->checkBasicAuthentication($request);
			} else {
				$result = false;
			}
			if (!$result) {
				$result = $this->checkOauthAuthentication($request);
			}
			if (!$result && \AppConfig::setting('site.security.api_session_auth')) {
				$result = $this->checkSessionAuthentication($request);
			}
		}
		if ($result) {
			$request = $result->withAttribute('isAuthenticated', true);
		} else if ($this->requiresAuthentication()) {
			throw new Errors\Unauthorized(translate('API_ERROR_WRONG_CREDENTIALS', 'app'));
		}
		return $request;
	}

	protected function checkOauthAuthentication($request) {
		$container = $this->app->getContainer();
		$rs = $container->get(ResourceServer::class);
		try {
			$request = $rs->validateAuthenticatedRequest($request);
		} catch (\Exception $e) {
			return false;
		}
		$client_id = $request->getAttribute('oauth_client_id');
		$clientRowResult = \ListQuery::quick_fetch('ClientEntity', $client_id);
		if ($clientRowResult) {
			$user_type = $user_id = null;
			$user_type_id = $request->getAttribute('oauth_user_id');
			if ($user_type_id) {
				list($user_type, $user_id) = explode(':', $user_type_id, 2);
			}
			if (!$user_id || $user_type != 'Users') {
				$user_type = 'Users';
				$user_id = $clientRowResult->getField('user_id');
			}
			if ($user_id) {
				$_SESSION['authenticated_user_id'] = $user_id;
				$authController = new \AuthenticationController();
				if($authController->sessionAuthenticate()) {
					return $request;
				}
				unset($_SESSION['authenticated_user_id']);
			}
		}
	}

	protected function checkBasicAuthentication($request) {
		if (isset($_SERVER['PHP_AUTH_USER']) && isset($_SERVER['PHP_AUTH_PW'])) {
			$authController = new \AuthenticationController();
			//$authController->authController = new AuthenticateUser();
			$userName = $_SERVER['PHP_AUTH_USER'];
			$password = $_SERVER['PHP_AUTH_PW'];
			$params = ['interface' => 'api'];
			if(preg_match('~^[0-9a-f]{32}$~', $password))
				$params['encoded_password'] = true;
			$result = $authController->login($userName, $password, $params);
			if ($result) {
				return $request
					->withAttribute('oauth_user_id', 'Users:' . $_SESSION['authenticated_user_id'])
					->withAttribute('oauth_scopes', ['read', 'write', 'profile'])
				;
			}
		}
	}

	protected function checkSessionAuthentication($request) {
		if (\AppConfig::session_start() && isset($_SESSION['authenticated_user_id'])) {
			$authController = new \AuthenticationController();
			$result = $authController->sessionAuthenticate();
			if ($result) {
				return $request
					->withAttribute('oauth_user_id', 'Users:' . $_SESSION['authenticated_user_id'])
					->withAttribute('oauth_scopes', ['read', 'write', 'profile'])
					;
			}
		}
	}

	protected function get($input, Response $response, Request $request) {
		throw new Errors\NotAllowed;
	}

	protected function post($input, Response $res, Request $req) {
		throw new Errors\NotAllowed;
	}

	protected function put($input, $res, $req) {
		throw new Errors\NotAllowed;
	}

	protected function delete($input, $res, $req) {
		throw new Errors\NotAllowed;
	}

	protected function patch($input, $res, $req) {
		throw new Errors\NotAllowed;
	}

	public function supportsMethod($method) {
		return in_array($method, $this->allowedMethods());
	}

	public function servesPath($path) {
		$pattern = '~^' . preg_replace('/:[a-zA-Z0-9]+/', '[^\/]+', $this->endpoint()) . '$~';
		return preg_match($pattern, $path);
	}


	protected static function check_module_access($module) {
		if (array_search($module, self::$disabled_modules) !== false) {
			throw new Errors\Forbidden;
		}
	}

	public final function validateParams(Request $req, $args) {
		$registry = $this->app->getContainer()->get('registry');
		$params = array();
		$method = strtolower($req->getMethod());
		$desc = $this->descriptor($method);
		$desc['params']['from_1crm_mobile'] = [
			'location' => 'query',
			'type' => 'Bool',
			'docs' => 'Indicates that request is coming from 1CRM Mobile',
			'hidden' => true,
		];
		$desc['params']['X_ONECRM_APPLICATION'] = [
			'location' => 'header',
			'type' => 'String',
			'docs' => 'Application ID',
			'hidden' => true,
		];
		foreach ($desc['params'] as $paramName => $paramDesc) {
			switch(array_get_default($paramDesc, 'location')) {
			case 'path' :
				$source = $args;
				break;
			case 'body' :
				$source = $req->getParsedBody();
				break;
			case 'header' :
				$headers = $req->getHeaders();
				$source = [];
				foreach ($headers as $k => $v) {
					if (is_array($v) && !empty($v)) {
						$k = strtoupper($k);
						if (strpos($k, 'HTTP_') === 0)
							$k = substr($k, 5);
						$source[strtoupper($k)] = $v[0];
					}
				}
				break;
			default:
				$source = $req->getQueryParams();
			}
			if (!empty($paramDesc['required']) && !isset($source[$paramName]))
				throw new Errors\BadRequest("Param {$paramName} is required");
			if (isset($source[$paramName])) {
				$typeName = $paramDesc['type'];
				if (is_array($typeName)) {
					$typeParams = $typeName;
					$typeName = $typeName['type'];
				} else {
					$typeParams = $paramDesc;
				}
				$type = $registry->getType($typeName);
				if (!$type)
					throw new Errors\InternalError('No definition for type ' . $typeName);
				$params[$paramName] = $type->parseValueEx($paramName, $source[$paramName], $typeParams);
			} elseif (isset($paramDesc['default'])) {
				$params[$paramName] = $paramDesc['default'];
			}
		}
		return $params;
	}

	public function expectsJSONBody() {
		return true;
	}

	public function expectsParsedBody() {
		return true;
	}

	protected function disabledForPortal(Request $request) {
		$uid = $request->getAttribute('oauth_user_id');
		if (!$uid) $uid = 'Users:' . \AppConfig::current_user_id();
		list($utype, $uid) = explode(':', $uid);
		if ($utype !== 'Contacts') {
			return;
		}
		throw new Errors\Forbidden('Not available to Contacts');
	}

	protected function applyPortalAcl(\ListQuery $lq, Request $request) {
		$container = $this->app->getContainer();
		$uid = $request->getAttribute('oauth_user_id');
		if (!$uid) $uid = 'Users:' . \AppConfig::current_user_id();
		list($utype, $uid) = explode(':', $uid);
		if ($utype !== 'Contacts') {
			return;
		}
		$model = $lq->getModelName();
		require_bean($model);
		$filter = \AppConfig::setting("model.portal.$model.filter");
		if (!$filter || !\AppConfig::setting("model.portal.$model.enabled")) {
			$lq->addFilterClause(['operator' => 'literal', 'value' => '0']);
			return;
		}
		$row = \ListQuery::quick_fetch_row('Contact', $uid);
		call_hook_function($filter, $model, [ $lq, $row['id'], $row['primary_account_id'] ]);
	}

	protected function providesCustomAuthMethod() {
		return false;
	}

	// should return an instance of Request on success
	// (just return the $req argument back in simplest case)
	protected function checkCustomAuthentication(Request $req) {
		throw new Errors\InternalError("Not implemented");
	}

    public function providesQueryLog()
    {
        return false;
    }

    /*
     * @param array $params     : includes path params
     * @param Request $request
     * @param null $logEntry    : if not null this is a target row id to update
     * @param null $result      : if not null this the API response
     * @return bool|int         : if a new entry was created return the row id
     */
    public function queryLog($params, Request $request, $logEntry = null, $result = null)
    {
        throw new Errors\InternalError("Not implemented");
    }

	/**
	 * @param Response $response
	 * @param array $data
	 * @return Response
	 */
	public function withJson(Response $response, array $data = []): Response
	{
		$response->getBody()->write(json_encode($data));
		return $response
			->withHeader('Content-Type', 'application/json');
	}
}
