<?php

namespace OneCRM\API\Endpoints\Auth;

use OneCRM\API\Endpoint;
use Psr\Http\Message\ServerRequestInterface as Request;
use Psr\Http\Server\RequestHandlerInterface as RequestHandler;
use League\OAuth2\Server\AuthorizationServer;
use OneCRM\API\Middleware\AuthorizationServerMiddleware;

abstract class AccessTokenBaseBase extends Endpoint {
	use SessionTrait;
	abstract public function resourceOwnerType();
}

abstract class AccessTokenBase extends AccessTokenBaseBase {

	public function descriptor($method) {
		if ($method == 'post')
			return $this->descriptorPost();
	}

	public function requiredScopes($method) {
		return [];
	}

	public function allowedMethods() {
		return ['post'];
	}

	public function isAuthEndpoint() {
		return true;
	}

	public function requiresAuthentication() {
		return false;
	}

	private function descriptorPost() {
		static $desc = [
			'title' => 'Authorization grant',
			'docs' => 'See [OAuth 2.0](?page=oauth)',
			'params' => [
				'grant_type' => [
					'type' => 'Enum',
					'options' => ['client_credentials', 'password', 'authorization_code', 'refresh_token'],
					'location' => 'body',
					'docs' => 'Grant type',
				],
				'client_id' => [
					'docs' => 'API client identifier',
					'type' => 'String',
					'location' => 'body',
				],
				'client_secret' => [
					'docs' => 'API client secret',
					'type' => 'String',
					'location' => 'body',
				],
				'refresh_token' => [
					'docs' => 'Refresh token',
					'type' => 'String',
					'location' => 'body',
				],
				'redirect_uri' => [
					'type' => 'String',
					'required' => false,
					'docs' => 'This parameter is optional, if not specified, the user will be redirected to a pre-registered redirect URI',
					'location' => 'body',
				],
				'code' => [
					'type' => 'String',
					'required' => false,
					'docs' => 'Authorization code',
					'location' => 'body',
				],
			],
			'return' => [
				'token_type' => [
					'type' => 'String',
					'required' => true,
					'docs' => 'Access token type. Always `Bearer`',
				],
				'expires_in' => [
					'type' => 'Int',
					'required' => true,
					'docs' => 'An integer representing the TTL of the access token',
				],
				'access_token' => [
					'type' => 'String',
					'required' => true,
					'docs' => 'A JWT signed with the authorization server’s private key',
				],
				'refresh_token' => [
					'type' => 'String',
					'docs' => 'An encrypted payload that can be used to refresh the access token when it expires',
				],
				'state' => [
					'type' => 'String',
					'docs' => 'State parameter sent in the original request. You should compare this value with the value stored in the user’s session to ensure the authorization code obtained is in response to requests made by this client rather than another client application',
				],

			],
		];
		return $desc;
	}

	protected function post($input, $response, $request) {
		return $response;
	}

	public function addMiddleware($app, $route) {
		$route->add(new AuthorizationServerMiddleware($app->getContainer()->get(AuthorizationServer::class)));
		$self = $this;

		$route->add(function (Request $request, RequestHandler $handler) use ($app, $self)  {
			$body = json_decode($request->getBody(), true);
			$request = $request->withParsedBody($body);
			$app->getContainer()->get('UserRepository')->setUserType($self->resourceOwnerType());
			return $handler->handle($request);
		});

		parent::addMiddleware($app, $route);
	}
}

