<?php

namespace OneCRM\API\Endpoints\Auth;

use OneCRM\API\Endpoint;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use League\OAuth2\Server\AuthorizationServer;
use v4\oAuth\Entities\UserEntity;
use League\OAuth2\Server\Exception\OAuthServerException;
use Slim\Routing\RouteContext;

abstract class AuthorizeBase extends Endpoint {

	use SessionTrait;

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
	}

	public function requiredScopes($method) {
		return [];
	}

	public function allowedMethods() {
		return ['get'];
	}

	public function isAuthEndpoint() {
		return true;
	}

	public function requiresAuthentication() {
		return false;
	}

	protected function keepSession() {
		\SessionManager::keepSession();
	}

	protected function forgetSession() {
		\SessionManager::doNotKeepSession();
	}

	abstract protected function resourceOwnerType();
	abstract protected function loginRoutePath($routeParser);
	abstract protected function grantRoutePath($routeParser);
	abstract protected function isValidScope($id);

	protected function descriptorGet() {
		static $desc = [
			'title' => 'Authorization request',
			'docs' => 'See [OAuth 2.0](?page=oauth)',
			'params' => [
				'response_type' => [
					'type' => 'Enum',
					'options' => ['code', 'token'],
					'location' => 'query',
					'docs' => 'Expected response type',
				],
				'client_id' => [
					'docs' => 'API client identifier',
					'type' => 'String',
					'location' => 'query',
				],
				'redirect_uri' => [
					'type' => 'String',
					'required' => false,
					'docs' => 'This parameter is optional, if not specified, the user will be redirected to a pre-registered redirect URI',
					'location' => 'query',
				],
				'scope' => [
					'type' => 'String',
					'required' => false,
					'docs' => 'A space delimited list of scopes',
					'location' => 'query',
				],
				'state' => [
					'type' => 'String',
					'required' => false,
					'docs' => 'CSRF token. This parameter is optional but highly recommended. You should store the value of the CSRF token in the user’s session to be validated when they return',
					'location' => 'query',
				],
			],
		];
		return $desc;
	}

	protected function get($input, Response $response, Request $request) {

		$this->keepSession();
		$container = $this->app->getContainer();

		/* @var \League\OAuth2\Server\AuthorizationServer $server */
		$server = $container->get(AuthorizationServer::class);	
		try {
			$params = $request->getQueryParams();
			$random = isset($params['c']) ? $params['c'] : false;
			$randomMatched = $random && $random === $this->sessionGet('random');
			$grantResult = $this->sessionGet('grant_result');

			//1. Auth-Request validieren (setzt voraus, dass die login-Seite nur mit Auth request aufgerufen wird
			if ($this->sessionIsSet('auth_request') && $randomMatched) {
				$authRequest = $this->sessionGet('auth_request');
			} else {
				$scope = isset($params['scope']) ? $params['scope'] : '';
				if (!strlen(trim($scope))) {
					$params['scope'] = 'profile';
					$request = $request->withQueryParams($params);
				}
				$authRequest = $server->validateAuthorizationRequest($request);
			}

			// kommt nur hier an, wenn response_type=code&client_id=[vorhandene/richtige client_id]&scope=basic

			// ist der Benutzer bereits angemeldet?

			if ($this->sessionIsSet('user_id') && $randomMatched) {
				$this->sessionUnset('auth_request');
				/** @var UserEntity $user */
				$user = $container->get('UserRepository')->getUserEntityById($this->sessionGet('user_id'));
				$authRequest->setUser($user);
			} else {
				$this->sessionSet('auth_request', $authRequest);
				$this->sessionSet('auth_request_client_name', $authRequest->getClient()->getName());
				$this->sessionSet('auth_request_scopes', array_map(function($scope) {
					return $scope->getIdentifier();
				}, $authRequest->getScopes()));
				$routeParser = RouteContext::fromRequest($request)->getRouteParser();

				if ($this->sessionIsSet('user_id')) {
					$path = $this->grantRoutePath($routeParser);
				} else {
					$path = $this->loginRoutePath($routeParser);
				}

				return $response
					->withHeader('Location', $path)
					->withStatus(301);
			}

			$userEntity = $authRequest->getUser();
			// wie prüfen ob der Nutzer authorisiert ist? reicht es, wenn es einen gibt?
			if ($userEntity) {
				// Once the user has approved or denied the client update the status
				// (true = approved, false = denied)
				$authRequest->setAuthorizationApproved($grantResult);

				$this->sessionUnset('auth_request');
				$this->sessionUnset('random');
				$this->sessionUnset('user_id');
				$this->forgetSession();
				return $server->completeAuthorizationRequest($authRequest, $response);
			}
			// Auth Request in der Session speichern
			$this->sessionSet('auth_request', $authRequest);
		} catch (OAuthServerException $exception) {
			return $exception->generateHttpResponse($response);
		} catch (\Exception $exception) {
			$response->getBody()->write($exception->getMessage());
			return $response->withStatus(500);
		}
		return $response->withStatus(500)->getBody()->write('This should not have happened');

	}

}

