<?php

namespace OneCRM\API\Endpoints\Auth;

use OneCRM\API\Endpoint;
use OneCRM\API\Errors;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Slim\Routing\RouteContext;

abstract class GrantBase extends Endpoint {

	use SessionTrait, TemplateTrait;

	public function descriptor($method) {
		return $this->descriptorPost();
	}

	public function requiredScopes($method) {
		return [];
	}

	public function allowedMethods() {
		return ['get', 'post'];
	}

	public function isAuthEndpoint() {
		return true;
	}

	public function requiresAuthentication() {
		return false;
	}

	public function expectsJSONBody() {
		return false;
	}

	abstract protected function resourceOwnerType();
	abstract protected function authRoutePath($routeParser, $random);

	protected function descriptorPost() {
		static $desc = [
			'title' => 'Login',
			'hidden' => true,
			'docs' => '',
			'params' => [

			],
		];
		return $desc;
	}

	protected function get($input, Response $response, Request $request) {
		return $this->processRequest($request, $response);
	}

	protected function post($input, Response $response, Request $request) {
		return $this->processRequest($request, $response);
	}

	protected function keepSession() {
		\SessionManager::keepSession();
	}

	protected function processRequest (Request $request, Response $response) {
		$this->keepSession();
		$container = $this->app->getContainer();
		$authRequest = $this->sessionGet('auth_request');

		if (!$authRequest) {
			throw new Errors\BadRequest('Missing auth request');
		}


		$errors = [];
		$username = '';
		if ($this->sessionIsSet('username')) $username = $this->sessionGet('username');

		$user = $container->get('UserRepository')->getUserEntityById($this->sessionGet('user_id'));


		// post überschreibt auch bestehende Benutzerdaten
		$params = array_merge($request->getQueryParams(), $request->getParsedBody() ?? []);
		$redirParam = $params['redirect'] ?? '';
		$routeParser = RouteContext::fromRequest($request)->getRouteParser();
		if ($request->getMethod() == 'POST') {
			//CSFR Check
			if (false === $request->getAttribute('csrf_status')) {
				$container->get('flash')->addMessage('error', 'Validationsfehler, bitte noch einmal!');
				return $response
					->withHeader('Location', $routeParser->urlFor($this->getRouteName('get')))
					->withStatus(301);
			}

			if (isset($params['cancel'])) {
				$this->sessionSet('grant_result', false);
			} else {
				$this->sessionSet('grant_result', true);
			}

			$random = bin2hex(random_bytes(20));
			$this->sessionSet('random', $random);
			return $response
				->withHeader('Location', $this->authRoutePath($routeParser, $random))
				->withStatus(301);
		}

		/*
		 * Formularausgabe
		 */

		// CSRF token name and value
		$csfr = [];
		$csfr['nameKey'] = $container->get('csrf')->getTokenNameKey();
		$csfr['valueKey'] = $container->get('csrf')->getTokenValueKey();
		$csfr['name'] = $request->getAttribute($csfr['nameKey']);
		$csfr['value'] = $request->getAttribute($csfr['valueKey']);

		$messages = $container->get('flash')->getMessages();
		if (isset($messages['error']))
			$errors = array_merge($errors, $messages['error']);


		$scopes = array_map(function($scope) {
			return translate_dd('oauth_scopes_desc_dom', $scope->getIdentifier(), $scope->getIdentifier());
		}, $authRequest->getScopes());
		$data = [
			'username' => $username,
			'client' => $authRequest->getClient(),
			'scopes' => $scopes,
			'csfr' => $csfr,
			'errors' => $errors,
			'router' => $routeParser,
			'redirect' => $redirParam,
			'action' => $routeParser->urlFor($this->getRouteName('get')),
			'first_name' => $user->getFirstName(),
			'last_name' => $user->getLastName(),
		];
		$this->setupTemplate($container);
		$response = $container->get('view')->render($response, 'grant.php', $data);
		return $response;
	}
}


