<?php

namespace OneCRM\API\Endpoints\Auth;

use OneCRM\API\Endpoint;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use v4\oAuth\Entities\UserEntity;
use v4\oAuth\Entities\ClientEntity;
use Slim\Routing\RouteContext;

abstract class LoginBase extends Endpoint {

	use SessionTrait, TemplateTrait;

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
		else
			return $this->descriptorPost();
	}

	public function requiredScopes($method) {
		return [];
	}

	public function allowedMethods() {
		return ['get', 'post'];
	}

	public function isAuthEndpoint() {
		return true;
	}

	public function requiresAuthentication() {
		return false;
	}

	public function expectsJSONBody() {
		return false;
	}

	abstract protected function resourceOwnerType();
	abstract protected function authRouteName();

	protected function descriptorGet() {
		static $desc = [
			'title' => 'Login',
			'hidden' => true,
			'docs' => '',
			'params' => [

			],
		];
		return $desc;
	}

	protected function descriptorPost() {
		static $desc = [
			'title' => 'Login',
			'hidden' => true,
			'docs' => '',
			'params' => [

			],
		];
		return $desc;
	}

	protected function get($input, Response $response, Request $request) {
		return $this->processRequest($request, $response);
	}

	protected function post($input, Response $response, Request $request) {
		return $this->processRequest($request, $response);
	}

	protected function keepSession() {
		\SessionManager::keepSession();
	}

	protected function processRequest (Request $request, Response $response) {
		$container = $this->app->getContainer();
		$routeParser = RouteContext::fromRequest($request)->getRouteParser();
		$errors = [];
		$username = '';
		if ($this->sessionIsSet('username')) $username = $this->sessionGet('username');
		// post überschreibt auch bestehende Benutzerdaten
		$params = array_merge($request->getQueryParams(), $request->getParsedBody() ?? []);
		$redirParam = $params['redirect'] ?? 'Home';
		$this->keepSession();
		if ($request->getMethod() == 'POST') {
			//CSFR Check
			if (false === $request->getAttribute('csrf_status')) {
				$container->get('flash')->addMessage('error', translate('OAUTH_WRONG_USERNAME', 'app'));
				return $response
					->withHeader('Location', $routeParser->urlFor($this->getRouteName('get')))
					->withStatus(301);
			}

			//check Login / Benutzer

			// ?? erstmal leer, da nur Login - Prüfung bei /authorize
			$scopes = '';

			$container->get('UserRepository')->setUserType($this->resourceOwnerType());

			$bodyParams = $request->getParsedBody();
			$username = isset($bodyParams['username']) ? $bodyParams['username'] : null;
			$password = isset($bodyParams['password']) ? $bodyParams['password'] : null;
			if ($username) $this->sessionSet('username', $username);
			/** @var UserEntity $userEntity */
			$userEntity = $container->get('UserRepository')->getUserEntityByUserCredentials(
				$username,
				$password,
				$scopes,
				new ClientEntity());
			if ($userEntity) {
				// State Logged in
				// save user ID in session
				$this->sessionSet('user_id', $userEntity->getIdentifier());

				$redirect = $routeParser->urlFor($redirParam);

				$random = bin2hex(random_bytes(20));
				$this->sessionSet('random', $random);

				return $response
					->withHeader('Location', $redirect . '?c=' . $random)
					->withStatus(301);
			}
		}

		/*
		 * Formularausgabe
		 */

		// CSRF token name and value
		$csfr = [];
		$csfr['nameKey'] = $container->get('csrf')->getTokenNameKey();
		$csfr['valueKey'] = $container->get('csrf')->getTokenValueKey();
		$csfr['name'] = $request->getAttribute($csfr['nameKey']);
		$csfr['value'] = $request->getAttribute($csfr['valueKey']);

		$messages = $container->get('flash')->getMessages();
		if (isset($messages['error']))
			$errors = array_merge($errors, $messages['error']);


		$data = [
			'username' => $username,
			'csfr' => $csfr,
			'errors' => $errors,
			'router' => $routeParser,
			'redirect' => $redirParam,
            'action' => $routeParser->urlFor($this->getRouteName('get')),
		];

		$this->setupTemplate($container);

		$response = $container->get('view')->render($response, 'login.php', $data);
		return $response;
	}


}


