<?php

namespace OneCRM\API\Endpoints\Calendar;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class Events extends Endpoint {

	public function endpoint() {
		return '/calendar/events';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
	}

	public function requiredScopes($method) {
		return ['read'];
	}

	public function allowedMethods() {
		return ['get'];
	}

	private function descriptorGet() {
		static $desc = [
			'title' => 'Get list of events',
			'docs' => '@endpoint/calendar/events_get',
			'params' => [
				'start_date' => [
					'type' => 'DateTime',
					'location' => 'query',
					'docs' => 'Lower date bound',
					'required' => true,
				],
				'end_date' => [
					'type' => 'DateTime',
					'location' => 'query',
					'docs' => 'Upper date bound',
					'required' => true,
				],
				'types' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Enum',
							'options' => ['Call', 'Meeting', 'Task', 'ProjectTask']
						]
					],
					'location' => 'query',
					'docs' => 'When present, this parameter limits returned events to specified types',
				],
				/* TODO
				'users' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'String',
							'min_len' => 1,
							'docs' => 'User ID',
						]
					],
					'location' => 'query',
					'docs' => 'When specified, this parameter limits returned events to specified users. By default, only authenticated user\'s events are returned',
				],
				*/
			],
			'return' => [
				'records' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
							'schema' => [
								'id' => [
									'type' => 'String',
									'docs' => 'Record ID',
									'required' => true,
								],
								'start_date' => [
									'type' => 'DateTime',
									'docs' => 'Event start date/time'
								],
								'due_date' => [
									'type' => 'DateTime',
									'docs' => 'Event due date/time'
								],
								'name' => [
									'type' => 'String',
									'docs' => 'Event name',
									'required' => true,
								],
								'type' => [
									'type' => 'Enum',
									'docs' => 'Event type',
									'options' => ['Call', 'Meeting', 'Task', 'ProjectTask'],
									'required' => true,
								],
								'location' => [
									'type' => 'String',
									'docs' => 'Event location, if applicable',
								],
							]
						]
					],
					'docs' => 'Array of retrieved records',
					'required' => true,
				],
			]
		];
		return $desc;
	}

	protected function get($input, $res, $req) {
		$this->disabledForPortal($req);

		$meta = [
			'Call' => [],
			'Meeting' => [],
			'Task' => ['has_due' => true],
			'ProjectTask' => ['has_due' => true],
		];
		$types = isset($input['types']) ? $input['types'] : '';

		$data = [];
		foreach ($meta as $model => $m) {
			if (!empty($types) && !in_array($model, $types)) continue;
			$lq = new \ListQuery($model);
			$lq->addFields(['name', 'date_start', 'location', 'date_due']);
			$lq->addSimpleFilter('assigned_user_id', \AppConfig::current_user_id());
			$clause = [
				'operator' => 'OR',
				'multiple' => [
					[
						'field' => 'date_start',
						'operator' => 'between',
						'value' => $input['start_date'],
						'end' => $input['end_date'],
					],
					[
						'field' => 'date_due',
						'operator' => 'between',
						'value' => $input['start_date'],
						'end' => $input['end_date'],
					],
				]
			];
			$lq->addFilterClause($clause);
			if (!empty($m['has_due'])) {
				$order = ['literal' => 'IFNULL(date_start, date_due)'];
			} else {
				$order = 'date_start';
			}
			$result = $lq->runQuery(0, 200, false, $order, 200);
			if (!$result || $result->failed) {
				throw new Errors\InternalError;
			}
			foreach ($result->getRowIndexes() as $i) {
				$row = $result->getRowResult($i);
				$data[] = [
					'name' => $row->getField('name'),
					'date_start' => $row->getField('date_start'),
					'date_due' => $row->getField('date_due'),
					'location' => $row->getField('location'),
					'id' => $row->getField('id'),
					'type' => $model,
				];
			}

		}

		$ret = [
			'records' => $data,
		];

		return $this->withJson($res, $ret);
	}

}


