<?php

namespace OneCRM\API\Endpoints\Data;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class ErasePersonal extends Endpoint {

	public function endpoint() {
		return '/data/erase_personal/{model}/{id}';
	}

	public function descriptor($method) {
		if ($method == 'patch')
			return $this->descriptorPatch();
	}

	public function requiredScopes($method) {
		return ['write'];
	}

	public function allowedMethods() {
		return ['patch'];
	}

	private function descriptorPatch() {
		static $desc = [
			'title' => 'Erase Personal Data',
			'docs' => '@endpoint/data/erase_personal',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Record ID',
					'required' => true,
				],
				'data' => [
					'location' => 'body',
					'type' => [
						'type' => 'Object',
					],
					'docs' => 'An object with `fields` property with the string of comma separated fields for erase. Fields not added to this string are not erased.',
					'required' => true,
				],
			],
			'return' => [
				'result' => [
					'type' => 'Bool',
					'docs' => 'Always true',
					'required' => true,
				],
			]
		];
		return $desc;
	}

	protected function patch($input, $res, $req) {
		$this->disabledForPortal($req);
		$model = $input['model'];
		$module = \AppConfig::module_for_model($model);

		if (!$module || !\AppConfig::setting("model.detail.{$model}.personal_data"))
			throw new Errors\NotFound("Model $model not found");

		Endpoint::check_module_access($module);

		if (!\ACLController::checkAccess($module, 'manage_personal'))
			throw new Errors\Forbidden('This module disabled for update.');

		$record = \ListQuery::quick_fetch($model, $input['id'], true, ['acl_filters' => 'edit']);

		if (! $record)
			throw new Errors\NotFound("Record not found");

		$fields = explode(',', $input['data']['fields']);

		$upd = \RowUpdate::for_result($record);
		foreach ($fields as $f) {
			$upd->eraseField($f);
		}
		$upd->save();

		$ret = [
			'result' => true,
			'data' => $record
		];

		return $this->withJson($res, $ret);
	}

}


