<?php

namespace OneCRM\API\Endpoints\Data;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class LinkedRecords extends Endpoint {

	public function endpoint() {
		return '/data/link/{model}/linked-records/{link}';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
	}

	public function requiredScopes($method) {
		return ['read'];
	}

	public function allowedMethods() {
		return ['get'];
	}

	private function descriptorGet() {
		$max =(int) \AppConfig::setting('api_config.max_list_items');
		$max = max($max, 200);
		$max = min($max, 500000);

		$desc = [
			'title' => 'Get list of all related records',
			'docs' => '@endpoint/data/linked_records_get',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'link' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => 1,
					'docs' => 'Link name',
					'required' => true,
				],
				'fields' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'String',
						]
					],
					'location' => 'query',
					'docs' => 'Array of field names to fetch. When omitted, a limited number of fields are returned, depending on model. Record ID is guaranteed to be returned, and for most models, `name` and/or `_display` fields.',
				],
				'order' => [
					'location' => 'query',
					'type' => 'String',
					'docs' => 'Sort order',
				],
				'offset' => [
					'location' => 'query',
					'type' => 'Int',
					'min' => 0,
					'default' => 0,
					'docs' => 'Offset in the list to start retrieval from',
				],
				'limit' => [
					'location' => 'query',
					'type' => 'Int',
					'min' => 1,
					'max' => $max,
					'default' => 20,
					'docs' => 'Limits number of records returned',
				],

			],
			'return' => [
				'records' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'Array of retrieved records',
					'required' => true,
				],
				'total_results' => [
					'type' => 'Int',
					'docs' => 'Total number of results, without taking offset and limit in account',
					'required' => true,
				]
			]
		];
		return $desc;
	}

	protected function get($input, $res, $req) {
		$model = $input['model'];
		$module = \AppConfig::module_for_model($model);

		if (!$module)
			throw new Errors\NotFound("Model $model not found");

		require_once('include/ListView/ListFormatter.php');
		$fmt = new \ListFormatter($model, $input['link'], null, 'listview');
		$lq =& $fmt->getQuery();

		$filter =& $fmt->getFilterForm();
		$filter->loadFilterLayout('Standard');

		if (!empty($input['fields']))
			$lq->addFields($input['fields']);

		$lq->addField('name');
		$lq->addField('_display');

		if (! empty($input['order']))
			$lq->setOrderBy($input['order']);

		$this->applyPortalAcl($lq, $req);

		$result = $fmt->getQueryResult($input['offset'], $input['limit']);

		if (!$result || $result->failed)
			throw new Errors\InternalError;

		ListEndpoint::filterExposedFields($result->rows, $result->getModelName());
		ListEndpoint::expandImageFields($result->rows, $result->getModelName());

		$ret = [
			'records' => array_values($result->rows),
			'total_results' => $result->total_count,
		];

		return $this->withJson($res, $ret);
	}
}
