<?php

namespace OneCRM\API\Endpoints\Data;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class ListEndpoint extends Endpoint {

	public function endpoint() {
		return '/data/{model}';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
		elseif ($method == 'post')
			return $this->descriptorPost();
	}

	public function requiredScopes($method) {
		if ($method == 'get')
			return ['read'];
		return ['write'];
	}

	public function allowedMethods() {
		return ['get', 'post'];
	}

	private function descriptorGet() {
		$max =(int) \AppConfig::setting('api_config.max_list_items');
		$max = max($max, 200);
		$max = min($max, 500000);
		$desc = [
			'title' => 'Get list of records',
			'docs' => '@endpoint/data/list_get',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'fields' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'String',
						]
					],
					'location' => 'query',
					'docs' => 'Array of field names to fetch. When omitted, a limited number of fields are returned, depending on model. Record ID is guaranteed to be returned, and for most models, `name` and/or `_display` fields.',
				],
				'query_favorite' => [
					'location' => 'query',
					'type' => 'Bool',
					'required' => false,
					'docs' => 'Query favorites. If set, `favorite` field will be added to each result row, indicating whether the current user added the item to Favorites. Note that item is in favorites if and only if the value of `favorite` field is zero ',
				],
				'filter_text' => [
					'location' => 'query',
					'type' => 'String',
					'docs' => 'Generic search string. Fields involved in search depend on model',
				],
				'filters' => [
					'location' => 'query',
					'type' => 'Object',
					'docs' => 'Filters to apply. To get list of available filters for a model, use [metadata](#group_meta)',
					'value_type' => [
						'type' => 'String',
					]
				],
				'order' => [
					'location' => 'query',
					'type' => 'String',
					'docs' => 'Sort order',
				],
				'offset' => [
					'location' => 'query',
					'type' => 'Int',
					'min' => 0,
					'default' => 0,
					'docs' => 'Offset in the list to start retrieval from',
				],
				'limit' => [
					'location' => 'query',
					'type' => 'Int',
					'min' => 1,
					'max' => $max,
					'default' => 20,
					'docs' => 'Limits number of records returned',
				],

			],
			'return' => [
				'records' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'Array of retrieved records',
					'required' => true,
				],
				'total_results' => [
					'type' => 'Int',
					'docs' => 'Total number of results, without taking offset and limit in account',
					'required' => true,
				]
			]
		];
		return $desc;
	}

	private function descriptorPost() {
		static $desc = [
			'title' => 'Create records',
			'docs' => 'Create a record of specified `model`',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model',
					'required' => true,
				],
				'data' => [
					'location' => 'body',
					'type' => [
						'type' => 'Object',
					],
					'docs' => 'An object with keys matching field names to update. Fields not defined in this object are not modified',
					'required' => true,
				],
			],
			'return' => [
				'id' => [
					'type' => 'String',
					'required' => true,
					'docs' => 'ID of created record',
				],
			],
		];
		return $desc;
	}

	protected function get($input, $res, $req) {

		// we do not apply ACL filters to query explicitly,
		// because that is done dy ListFormatter

		$model = $input['model'];
		$module = \AppConfig::module_for_model($model);
		if (!$module)
			throw new Errors\NotFound("Model $model not found");
		Endpoint::check_module_access($module);

		require_once('include/ListView/ListFormatter.php');
		$fmt = new \ListFormatter($model);
		$lq =& $fmt->getQuery();

		$filter =& $fmt->getFilterForm();
		$filter->loadFilterLayout('Standard');


		$filter_data = array_get_default($input, 'filters', array());
		if (isset($input['filter_text']))
			$filter_data['filter_text'] = $input['filter_text'];

		if (!empty($input['from_1crm_mobile'])) {
			$this->applyOverrideFilters($model, $filter);
		}

		$fmt->loadFilter($filter_data);


		if (!empty($input['fields']))
			$lq->addFields($input['fields']);
		$lq->addField('name');
		$lq->addField('_display');
		if (!empty($input['query_favorite'])) {
			require_once 'modules/Favorites/Favorite.php';
			\Favorite::add_join($lq);
		}
		if (isset($input['order']))
			$lq->setOrderBy($input['order']);
		$this->applyPortalAcl($lq, $req);
		$result = $fmt->getQueryResult($input['offset'], $input['limit']);
		if (!$result || $result->failed)
			throw new Errors\InternalError;

		self::filterExposedFields($result->rows, $model);
		self::expandImageFields($result->rows, $model);

		$data = array_values($result->rows);

		require_once('modules/TaxCodes/TaxCode.php');
		$seedcode = new \TaxCode();

		$ordering_modules = [
			'Quotes', 'Invoice', 'Bills', 'SalesOrders', 'PurchaseOrders', 'CreditNotes'
		];

		$adj_field = [
			'Quotes' => 'quote_id',
			'Invoice' => 'invoice_id',
			'Bills' => 'bills_id',
			'SalesOrders' => 'sales_orders_id',
			'PurchaseOrders' => 'purchase_orders_id',
			'CreditNotes' => 'credit_id',
		];

		if($module == 'Payments') {
			foreach ($data as &$r) {
				$relinv_lq = new \ListQuery('invoices_payments');
				$relinv_lq->addSimpleFilter('payment_id', $r['id']);
				$r['related_invoices'] = $relinv_lq->fetchAllRows();
				foreach($r['related_invoices'] as &$pmt_inv) {
                    $invoice_data = \ListQuery::quick_fetch_row('Invoice', $pmt_inv['invoice_id']);
					if($invoice_data && isset($invoice_data['integrated_id'])) {
						$pmt_inv['integrated_id'] = $invoice_data['integrated_id'];
					}
				}
			}
		}

		if (in_array($module, $ordering_modules) && \AppConfig::setting("model.detail.{$model}Adjustment")) {
			require_once('include/ListView/ListFormatter.php');
			foreach ($data as &$r) {
				$lq_tax = new \ListQuery($model.'Adjustment');
				$lq_tax->addSimpleFilter($adj_field[$module], $r['id'], '=');
				$lq_tax->addSimpleFilter('related_type', 'TaxRates', '=');
				$tax_result = $lq_tax->fetchAllRows();
				$r['total_tax'] = 0;
				foreach($tax_result as $tax_row) {
					$r['total_tax'] += $tax_row['amount'];
				}

				$lq_tax = new \ListQuery($model.'Adjustment');
				$lq_tax->addSimpleFilter($adj_field[$module], $r['id'], '=');
				$lq_tax->addSimpleFilter('related_type', 'Discounts', '=');
				$tax_result = $lq_tax->fetchAllRows();
				$r['total_discount'] = 0;
				foreach($tax_result as $tax_row) {
					$r['total_discount'] += $tax_row['amount'];
				}

				$lq_tax = new \ListQuery($model.'Adjustment');
				$lq_tax->addSimpleFilter($adj_field[$module], $r['id'], '=');
				$lq_tax->addSimpleFilter('related_type', 'ShippingProviders', '=');
				$tax_result = $lq_tax->fetchAllRows();
				$r['total_shipping'] = 0;
				foreach($tax_result as $tax_row) {
					$r['total_shipping'] += $tax_row['amount'];
				}

				require_once('include/ListView/ListFormatter.php');
				$rel_fmt = new \ListFormatter($model, 'lines_link', null, 'listview', ['parent_key' => $r['id']]);
				$rel_lq =& $rel_fmt->getQuery();
				$rel_lq->addFields(array('id', 'line_group_id', 'name', 'quantity', 'mfr_part_no', 'serial_no', 'cost_price', 'list_price', 'unit_price', 'ext_price', 'tax_class_id'));
				$rel_result = $rel_fmt->getQueryResult(0, 999);
				ListEndpoint::filterExposedFields($rel_result->rows, $rel_result->getModelName());
				if(!isset($_GET['skip_line_items']) || $_GET['skip_line_items'] == 0) {
					$r['line_items'] = array_values($rel_result->rows);
					foreach($r['line_items'] as &$line) {
						$rates = $seedcode->get_tax_rates($line['tax_class_id'], false);
						$rates_arr = [];
						foreach($rates as $k => $v) {
							$rates_arr[] = $v;
						}
						$lq_tax = new \ListQuery($model.'Adjustment');
						$lq_tax->addSimpleFilter('line_id', $line['id'], '=');
						$lq_tax->addSimpleFilter('type', 'PercentDiscount', '=');
						$disc_result = $lq_tax->fetchAllRows();
						$line['line_discount_amount'] = 0;
						$line['line_discount_perc'] = 0;
						foreach($disc_result as $disc_row) {
							$line['line_discount_amount'] += number_format(($line['ext_price']*$disc_row['rate'])/100, 2, '.', '');
							$line['line_discount_perc'] = $disc_row['rate'];
							$r['total_discount'] += $line['line_discount_amount'];
						}
						if(isset($rates_arr[0])) {
							$line['line_tax_perc'] = $rates_arr[0]['rate'];
							$line['line_tax_amount'] = number_format(($line['ext_price']*$rates_arr[0]['rate'])/100, 2, '.', '');
						}
					}
				}
			}
		}



		$ret = [
			'records' => $data,
			'total_results' => $result->total_count,
		];
		$hooks = \AppConfig::setting("model.hooks.$model.api_decorate", []);
		foreach ($hooks as $hook) {
			foreach ($ret['records'] as &$rec) {
				call_hook_function($hook, null, array($model, &$rec));
			}
			unset($rec);
		}
		if (\AppConfig::module_primary_bean($module) === $model) {
			$ret['module'] = $module;
		}
		
		return $this->withJson($res, $ret);
	}

	protected function post($input, $res, $req) {
		$this->disabledForPortal($req);
		$model = $input['model'];
		$module = \AppConfig::module_for_model($model);
		if (!$module)
			throw new Errors\NotFound("Model $model not found");
		Endpoint::check_module_access($module);
		$allowed = \ACLController::checkAccess($module, 'edit');
		if (!$allowed) {
			throw new Errors\Forbidden("Access denied");
		}

		$_REQUEST = $input['data'];

		if (($model === 'Invoice' || $model === 'Quote' || $model === 'SalesOrder' || $model === 'Bill') && ((isset($_REQUEST['lines_link']) && isset($_REQUEST['quantity']) && isset($_REQUEST['unit_price'])) || isset($_REQUEST['line_items']))) {
			if(is_array($_REQUEST['line_items']))
				$r = $this->createModuleItemWithLineItemsArray($model, $_REQUEST);
			else
				$r = $this->createModuleItemWithLineItems($model, $_REQUEST);

			$ret = [
				'id' => $r->getPrimaryKeyValue()
			];
	    } else {
			$update = \RowUpdate::blank_for_model($model);
			$update->loadRequest('', false);
			foreach ($input['data'] as $k => $v) {
				if (strpos($k, "_social_") === 0) {
					$update->set($k, $v);
				}
			}
			if (isset($input['X_ONECRM_APPLICATION'])) {
				$update->setRelatedData('X_ONECRM_APPLICATION', $input['X_ONECRM_APPLICATION']);
			}
			$update->save();
			$ret = [
				'id' => $update->getPrimaryKeyValue(),
			];
		}
		
		return $this->withJson($res, $ret);
	}

	private function getItemIdByFieldValue($model, $field, $value) {
		$lq = new \ListQuery($model);
		$lq->addSimpleFilter($field, $value, '=');
		$result = $lq->runQuery(0, 200);
		if (!$result || $result->failed) {
			return false;
		}
		if(count($lq->getResultIds($result)) == 0)
			return false;
		return $lq->getResultIds($result)[0];
	}

	private function createModuleItemWithLineItemsArray($model, $req_data) {
		require_once 'modules/TaxRates/TaxRate.php';
		require_once 'modules/Accounts/Account.php';
		require_once 'modules/ProductCatalog/Product.php';
		require_once 'modules/ProductCategories/ProductCategory.php';
		require_once 'modules/ProductTypes/ProductType.php';
		$mgr = \RowUpdate::blank_for_model($model);
		require_bean($model);

		$account = new \Account;
		if (!$account->retrieve($req_data['billing_account_id']) || $account->retrieve($req_data['billing_account_id']) == '') {
			$new_account = \RowUpdate::blank_for_model('Account');
			$currency_id = '-99';
			$new_account->set(array(
				'name' => $req_data['billing_account'],
				'phone_office' => $req_data['billing_phone'],
				'email1' => $req_data['billing_email'],
				'billing_address_street' => $req_data['billing_address_street'],
				'billing_address_city' => $req_data['billing_address_city'],
				'billing_address_state' => $req_data['billing_address_state'],
				'assigned_user_id' => \AppConfig::current_user_id(),
				'billing_address_postalcode' => $req_data['billing_address_postalcode'],
				'billing_address_country' => $req_data['billing_address_country'],
				'shipping_address_street' => $req_data['shipping_address_street'],
				'shipping_address_city' => $req_data['shipping_address_city'],
				'shipping_address_state' => $req_data['shipping_address_state'],
				'shipping_address_postalcode' => $req_data['shipping_address_postalcode'],
				'shipping_address_country' => $req_data['shipping_address_country'],
				'industry' => '',
				'account_type' => "Customer",
				'currency_id' => $currency_id,
				'exchange_rate' => 1,
				'invalid_email' => 0,
				'email_opt_out' => 0,
			));
			$new_account->save();
			$account_id = $new_account->getPrimaryKeyValue();
		} else {
			$account_id = $account->id;
			$currency_id = $account->currency_id;
		}
		$data = call_user_func(array($model, 'init_from_account'), $account_id);
		if($data)
			$mgr->set($data);

		$mgr->set('currency_id', $currency_id);
		$mgr->set('assigned_user_id', \AppConfig::current_user_id());

		foreach($req_data as $k => $v) {
			$mgr->set($k, $v);
		}

		$new_groups = array(
			'group~0' => array(
				'status' => 'Confirmed',
				'lines' => array() ,
				'adjusts' => array() ,
			) ,
		);

		// creating products / line groups / line items
		$amount = 0;
		foreach ($req_data['line_items'] as $k => $line) {
			$line_id = "line~$k";

			$line['ext_price'] = $line['quantity'] * $line['unit_price'];
			$line['list_price'] = $line['unit_price'];
			$adjust_id = '';

			if(isset($line['line_discount_perc']) || isset($line['line_discount_amount'])) {
				if(isset($line['line_discount_perc'])) {
					$line['line_discount_amount'] = number_format((($line['unit_price']*$line['quantity'])*$line['line_discount_perc'])/100, 2);
				} else {
					$dc_rate = round(($line['line_discount_amount'] / ($line['quantity'] * $line['unit_price'])), 3)*100;
					$line['line_discount_perc'] = $dc_rate;
				}
				$adjust_id = 'adjdisc~$k';
				$new_groups['group~0']['adjusts'][$adjust_id] = array(
					'related_id' => '',
					'related_type' => '',
					'line_id' => $line_id,
					'type' => 'PercentDiscount',
					'rate' => $line['line_discount_perc'],
					'name' => '',
					'position' => 2,
					'amount' => 0.0,
				);

				$line['ext_price'] = $line['ext_price']-$line['line_discount_amount'];
				$line['unit_price'] = $line['unit_price']-(($line['unit_price']*$line['line_discount_perc'])/100);
			}

			$product_id = $this->getItemIdByFieldValue('Product', 'name', $line['name']);
			$product = \ListQuery::quick_fetch_row('Product', $product_id);

			$cost = 0.0;
			if($product)
				$cost = $product['cost'];


			// checking if tax rate already exist and if not create both tax rate and code
			$tax_code_id = false;
			if(isset($line['line_tax_perc'])) {
				$line['line_tax_amount'] = ($line['line_tax_perc']*($line['ext_price']))/100;
			}
			if (isset($line['line_tax_amount']) && $line['line_tax_amount'] !== '' && $line['line_tax_amount'] !== 0) {
				$tax_rate = round(($line['line_tax_amount'] / ($line['ext_price'])), 3)*100;
				$tax_rate_id = $this->getItemIdByFieldValue('TaxRate', 'rate', $tax_rate);
				if (!$tax_rate_id || $tax_rate_id === null) {
					$tax = \RowUpdate::blank_for_model('TaxRate');
					$tax->set(array(
						'name' => 'Tax Rate '.$tax_rate,
						'rate' => $tax_rate,
						'compounding' => '',
						'status' => 'Active',
					));
					$tax->save();
					$tax_rate_id = $tax->getPrimaryKeyValue();
					$tax_code_sys = \RowUpdate::blank_for_model('TaxCode');
					$tax_code_sys->set('name', 'Tax Code');
					$tax_code_sys->set('code', 'TAX_'.$tax_rate);
					$tax_code_sys->set('position', 1);
					$tax_code_sys->save();
					$tax_code_id = $tax_code_sys->getPrimaryKeyValue();
					// link tax_rate and tax_code
					$tax->addUpdateLink('taxcodes', array(
						'id' => $tax_code_id,
						'position' => 1,
					));
				} else {
					// getting the related TaxCode id from TaxRate id
					require_once('include/ListView/ListFormatter.php');
					$rate = \ListQuery::quick_fetch('TaxRate', $tax_rate_id);
					$codes = $rate->getLinkResults('taxcodes');
					$rows = $codes->getRows();
					$tax_code_id = null;

					foreach ($rows as $row) {
						$tax_code_id = $row['id'];
						break;
					}
				}
				$tax_rate_sys = new \TaxRate;
				$tax_sys = $tax_rate_sys->retrieve($tax_rate_id);

				$new_groups['group~0']['adjusts']['adjtax~$k'] = array(
					'related_id' => $tax_rate_id,
					'related_type' => 'TaxRates',
					'type' => 'StandardTax',
					'rate' => $tax_sys->rate,
					'name' => $tax_sys->name,
					'amount' => ($line['ext_price']*$tax_sys->rate)/100,
				);
			}

			$mfr_part_no = isset($line['mfr_part_no']) ? $line['mfr_part_no'] : '';
			$new_groups['group~0']['lines'][$line_id] = array(
				'name' => $line['name'],
				'related_id' => $product_id,
				'related_type' => 'ProductCatalog',
				'quantity' => $line['quantity'],
				'mfr_part_no' => $mfr_part_no,
				'cost_price' => $cost,
				'list_price' => $line['list_price'],
				'unit_price' => $line['unit_price'],
				'std_unit_price' => $line['unit_price'],
				'tax_class_id' => ($tax_code_id ? $tax_code_id : null),
			);

			if($adjust_id != '')
				$new_groups['group~0']['lines'][$line_id]['pricing_adjust_id'] = $adjust_id;

			$last = $k;
			$amount += $line['ext_price'];
		}

		if(isset($req_data['total_discount'])) {
			$last = $last+1;
			$adjust_id = 'adjdisc~'.$last;
			$rate = round(($req_data['total_discount'] / ($amount)), 3)*100;
			$new_groups['group~0']['adjusts'][$adjust_id] = array(
				'related_id' => '',
				'related_type' => 'Discounts',
				'type' => 'StandardDiscount',
				'rate' => $rate,
				'name' => 'Discount',
				'position' => 0,
				'amount' => $req_data['total_discount'],
			);
		}

		if(isset($req_data['total_shipping'])) {
			$last = $last+1;
			$adjust_id = 'adjship~'.$last;
			$new_groups['group~0']['adjusts'][$adjust_id] = array(
				'related_id' => '',
				'related_type' => 'Shipping',
				'type' => 'UntaxedShipping',
				'name' => 'Shipping',
				'position' => 0,
				'amount' => $req_data['total_shipping'],
			);
		}

		$mgr->replaceGroups($new_groups);
		$mgr->save();
		return $mgr;
	}

	private function createModuleItemWithLineItems($model, $req_data) {
		require_once 'modules/TaxRates/TaxRate.php';
		require_once 'modules/Accounts/Account.php';
		require_once 'modules/ProductCatalog/Product.php';
		require_once 'modules/ProductCategories/ProductCategory.php';
		require_once 'modules/ProductTypes/ProductType.php';
		$mgr = \RowUpdate::blank_for_model($model);
		require_bean($model);

		$account = new \Account;
		if (!$account->retrieve($req_data['billing_account_id'])) {
			$new_account = \RowUpdate::blank_for_model('Account');
			$currency_id = '-99';
			$new_account->set(array(
				'name' => $req_data['billing_account_id'],
				'phone_office' => $req_data['billing_phone'],
				'email1' => $req_data['billing_email'],
				'billing_address_street' => $req_data['billing_address_street'],
				'billing_address_city' => $req_data['billing_address_city'],
				'billing_address_state' => $req_data['billing_address_state'],
				'assigned_user_id' => \AppConfig::current_user_id(),
				'billing_address_postalcode' => $req_data['billing_address_postalcode'],
				'billing_address_country' => $req_data['billing_address_country'],
				'shipping_address_street' => $req_data['shipping_address_street'],
				'shipping_address_city' => $req_data['shipping_address_city'],
				'shipping_address_state' => $req_data['shipping_address_state'],
				'shipping_address_postalcode' => $req_data['shipping_address_postalcode'],
				'shipping_address_country' => $req_data['shipping_address_country'],
				'industry' => 'Tech',
				'account_type' => "Customer",
				'currency_id' => $currency_id,
				'exchange_rate' => 1,
				'invalid_email' => 0,
				'email_opt_out' => 0,
			));
			$new_account->save();
			$account_id = $new_account->getPrimaryKeyValue();
		} else {
			$account_id = $account->id;
			$currency_id = $account->currency_id;
		}
		$data = call_user_func(array($model, 'init_from_account'), $account_id);
		if($data)
			$mgr->set($data);

		$mgr->set('currency_id', $currency_id);
		$mgr->set('assigned_user_id', \AppConfig::current_user_id());

		$due_date = isset($req_data['due_date']) ? $req_data['due_date'] : date('Y-m-d H:i:s', time());
		$mgr->set('due_date', $due_date);
		if($model == 'Quote')
			$mgr->set('valid_until', $due_date);
		if($model == 'Bill')
			$mgr->set('bill_date', $due_date);

		$mgr->set('quote_stage', $req_data['stage']);
		$mgr->set('terms', $req_data['terms']);
		$new_groups = array(
			'group~0' => array(
				'status' => 'Confirmed',
				'lines' => array() ,
				'adjusts' => array() ,
			) ,
		);
		$lines_link = explode(',', $req_data['lines_link']);
		$quantity = explode(',', $req_data['quantity']);
		$mfr_part_no = explode(',', $req_data['mfr_part_no']);
		$unit_price = explode(',', $req_data['unit_price']);
		$line_item_tax = explode(',', $req_data['line_item_tax']);
		$category_id = $this->getItemIdByFieldValue('ProductCategory', 'name', 'Category for products from Zapier');
		$product_type_id = $this->getItemIdByFieldValue('ProductType', 'name', 'Product type from Zapier');
		if (!$category_id || $category_id === null) {
			$prod_cat = $this->addBean('ProductCategory', 'Category for products from Zapier');
			$category_id = $prod_cat->id;
		}
		if (!$product_type_id || $product_type_id === null) {
			$prod_type = $this->addBean('ProductType', 'Product type from Zapier', array(
				'category_id' => $category_id
			));
			$product_type_id = $prod_type->id;
		}
		$mgr->set('name', $req_data['name']);
		// creating products / line groups / line items
		foreach ($lines_link as $k => $line) {
			$info = array(
				'product_category_id' => $category_id,
				'product_type_id' => $product_type_id,
				'list_price' => sprintf('%0.2f',  $unit_price[$k]),
				'cost' => sprintf('%0.2f',  $unit_price[$k]),
				'purchase_price' => sprintf('%0.2f',  $unit_price[$k]),
				'is_available' => 'yes',
				'date_available' => date('Y-m-d H:i:s', time()),
				'is_taxable' => 'yes',
				'manufacturers_part_no' => (isset($mfr_part_no[$k]) ? $mfr_part_no[$k] : ''),
				'pricing_formula' => 'Fixed Price',
				'in_stock' => 99,
				'description' => 'Products imported via Zapier',
			);
			$product = $this->addBean('Product', $line, $info);

			// checking if tax rate already exist and if not create both tax rate and code
			$tax_code_id = false;
			if (isset($line_item_tax[$k]) && $line_item_tax[$k] !== '' && $line_item_tax[$k] !== 0) {
				$tax_rate = round(($line_item_tax[$k] / ((floatval($quantity[$k]) * floatval($unit_price[$k])) - $line_item_tax[$k])), 3)*100;
				$tax_rate_id = $this->getItemIdByFieldValue('TaxRate', 'rate', $tax_rate);
				if (!$tax_rate_id || $tax_rate_id === null) {
					$tax = \RowUpdate::blank_for_model('TaxRate');
					$tax->set(array(
						'name' => 'Zapier Rate '.$tax_rate,
						'rate' => $tax_rate,
						'compounding' => '',
						'status' => 'Active',
					));
					$tax->save();
					$tax_rate_id = $tax->getPrimaryKeyValue();
					$tax_code_sys = \RowUpdate::blank_for_model('TaxCode');
					$tax_code_sys->set('name', 'Zapier Code');
					$tax_code_sys->set('code', 'ZAP_'.$tax_rate);
					$tax_code_sys->set('position', 1);
					$tax_code_sys->save();
					$tax_code_id = $tax_code_sys->getPrimaryKeyValue();
					// link tax_rate and tax_code
					$tax->addUpdateLink('taxcodes', array(
						'id' => $tax_code_id,
						'position' => 1,
					));
				} else {
					// getting the related TaxCode id from TaxRate id
					require_once('include/ListView/ListFormatter.php');
					$fmt = new \ListFormatter('TaxRate', 'taxcodes', null, 'listview', ['parent_key' => $tax_rate_id]);
					$result = $fmt->getQueryResult();
					$tax_code_id = $result->rows[0]->id;
				}
				$tax_rate_sys = new \TaxRate;
				$tax_sys = $tax_rate_sys->retrieve($tax_rate_id);

				$new_groups['group~0']['adjusts']['adjtax~$k'] = array(
					'related_id' => $tax_rate_id,
					'related_type' => 'TaxRates',
					'type' => 'StandardTax',
					'rate' => $tax_sys->rate,
					'name' => $tax_sys->name,
					'amount' => 0.0,
				);
			}

			$line_id = "line~$k";
			$new_groups['group~0']['lines'][$line_id] = array(
				'name' => $product->name,
				'related_id' => $product->id,
				'related_type' => 'ProductCatalog',
				'quantity' => $quantity[$k],
				'mfr_part_no' => $product->manufacturers_part_no,
				'cost_price' => $product->cost,
				'list_price' => $product->list_price,
				'unit_price' => $product->purchase_price,
				'std_unit_price' => $product->purchase_price,
				'tax_class_id' => ($tax_code_id ? $tax_code_id : null),
			);
		}

		$mgr->replaceGroups($new_groups);
		$mgr->save();
		return $mgr;
	}

	private function addBean($class, $name, $extra = array())
	{
		static $map = array();
		if (!isset($map[$class][$name])) {
			$bean = new $class;
			$bean->name = $name;
			foreach ($extra as $k => $v) {
				$bean->$k = $v;
			}
			if (!empty($bean->id))
				$bean->new_with_id = true;
			$bean->save();
			$bean->retrieve($bean->id, false);
			$map[$class][$name] =& $bean;
		}
		return $map[$class][$name];
	}

	private function applyOverrideFilters($model, $filter) {
		if ($model == 'Note') {
			$filter->override_filters = ['email_attachment' => ['field' => 'email_attachment', 'value' => 0]];
		}
	}

	public static function expandImageFields(&$rows, $model) {
		$modelDef = new \ModelDef($model);
		foreach ($rows as &$row) {
			foreach ($row as $k => $v) {
				$def = $modelDef->getFieldDefinition($k);
				if (isset($def['type']) && $def['type'] == 'image') {
					require_once 'include/upload_file.php';
					$upload_dir = \UploadFile::replace_placeholders($def['upload_dir']);
					$filename_field = $def['filename'];
					$filename = array_get_default($row, $filename_field);
					if (!$filename && isset($def['alt_filename'])) {
						$filename = array_get_default($row, $def['alt_filename']);
					}
					if ($filename) {
						$value = \UploadFile::get_path($filename, '', $upload_dir);
						$row[$filename_field] = $value;
						$row['__full_' . $filename_field] = \AppConfig::site_url() . $value;
					}
				}
			}
		}
	}

	public static function filterExposedFields(&$rows, $model) {
		$modelDef = new \ModelDef($model);
		foreach ($rows as &$row) {
			foreach ($row as $k => $v) {
				$f = $modelDef->getFieldDefinition($k);
				if (!array_get_default($f, 'api_exposed', true))
					unset($row[$k]);
			}
		}
	}
}


