<?php

namespace OneCRM\API\Endpoints\Data;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class Record extends Endpoint {

	public function endpoint() {
		return '/data/{model}/{id}';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
		elseif ($method == 'patch')
			return $this->descriptorPatch();
		elseif ($method == 'delete')
			return $this->descriptorDelete();
	}

	public function requiredScopes($method) {
		if ($method == 'get')
			return ['read'];
		return ['write'];
	}

	public function allowedMethods() {
		return ['get', 'patch', 'delete'];
	}

	private function descriptorGet() {
		static $desc = [
			'title' => 'Get single record',
			'docs' => '@endpoint/data/record_get',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Record ID',
					'required' => true,
				],
				'fields' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'String',
						]
					],
					'location' => 'query',
					'docs' => 'Array of field names to fetch. When omitted, all available fields are returned.',
				],

			],
			'return' => [
				'record' => [
					'type' => [
						'type' => 'Object',
					],
					'docs' => 'Retrieved record',
					'required' => true,
				],
			]
		];
		return $desc;
	}

	private function descriptorPatch() {
		static $desc = [
			'title' => 'Update a record',
			'docs' => 'Update record belonging to specified `model`, identified by `id`',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Record ID',
					'required' => true,
				],
				'data' => [
					'location' => 'body',
					'type' => [
						'type' => 'Object',
					],
					'docs' => 'An object with keys matching field names to update. Fields not defined in this object are not modified',
					'required' => true,
				],
				'create' => [
					'location' => 'body',
					'type' => [
						'type' => 'Bool',
					],
					'docs' => 'If true, record will be created if it does not exist',
					'required' => false,
				],

			],
			'return' => [
				'result' => [
					'type' => 'Bool',
					'docs' => 'Always true',
					'required' => true,
				],
			]
		];
		return $desc;
	}

	private function descriptorDelete() {
		static $desc = [
			'title' => 'Delete a record',
			'docs' => 'Delete record belonging to specified `model`, identified by `id`',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model',
					'required' => true,
				],
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Record ID',
					'required' => true,
				],

			],
			'return' => [
				'result' => [
					'type' => 'Bool',
					'docs' => 'True if deleted successfully',
					'required' => true,
				],
			]
		];
		return $desc;
	}

	protected function delete($input, $res, $req) {
		$this->disabledForPortal($req);
		$model = $input['model'];
		$id = $input['id'];
		$module = \AppConfig::module_for_model($model);
		if (!$module)
			throw new Errors\NotFound("Model $model not found");
		Endpoint::check_module_access($module);
		$record = \ListQuery::quick_fetch($model, $input['id'], true, ['acl_filters' => 'delete']);
		if (!$record)
			throw new Errors\NotFound;
		$update = \RowUpdate::for_result($record);
		$result = $update->markDeleted();

		return $this->withJson($res, ['result' => !!$result]);
	}

	protected function get($input, $res, $req) {
		require_once 'include/upload_file.php';
		$model = $input['model'];
		$module = \AppConfig::module_for_model($model);
		$id = $input['id'];
		if (!$module)
			throw new Errors\NotFound("Model $model not found");
		Endpoint::check_module_access($module);
		$fields = true;
		if (!empty($input['fields'])) {
			$fields = array();
			foreach ($input['fields'] as $f) {
				$fields[$f] = $f;
			}
		}
		$data = null;

		$params = ['process_results' => false, 'acl_filters' => 'view'];
		$lq = new \ListQuery($model, $fields, $params);
		if (is_array($fields) && in_array("favorite", $fields)) {
			require_once 'modules/Favorites/Favorite.php';
			\Favorite::add_join($lq);
		}
		$lq->addAclFilter('view');
		$this->applyPortalAcl($lq, $req);
		$result = $lq->queryRecord($id);
		if ($result && !$result->failed)
			$data = $result->row;

		if (!$data)
			throw new Errors\NotFound;

		$modelDef = new \ModelDef($model);
		if (is_array($fields) && array_search('_social_accounts', $fields) !== false) {
			require_once 'modules/SocialAccounts/SocialAccount.php';
			$accounts = \SocialAccount::getList($module, $id);
			if (!empty($accounts)) {
				foreach ($accounts as &$acc) {
					if (isset($acc['params'])) {
						$params = unserialize($acc['params']);
						if ($params && isset($params['entered_url'])) {
							$acc['entered_url'] = $params['entered_url'];
						}
					}
				}
				$data['_social_accounts'] = array_values($accounts);
			}
		}

		require_once('modules/TaxCodes/TaxCode.php');
		$seedcode = new \TaxCode();

		$ordering_modules = [
			'Quotes', 'Invoice', 'Bills', 'SalesOrders', 'PurchaseOrders', 'CreditNotes'
		];

		$adj_field = [
			'Quotes' => 'quote_id',
			'Invoice' => 'invoice_id',
			'Bills' => 'bills_id',
			'SalesOrders' => 'sales_orders_id',
			'PurchaseOrders' => 'purchase_orders_id',
			'CreditNotes' => 'credit_id',
		];

		if($module == 'Payments') {
			$relinv_lq = new \ListQuery('invoices_payments');
			$relinv_lq->addSimpleFilter('payment_id', $data['id']);
			$data['related_invoices'] = $relinv_lq->fetchAllRows();
			foreach($data['related_invoices'] as &$pmt_inv) {
				$invoice_data = \ListQuery::quick_fetch_row('Invoice', $pmt_inv['invoice_id']);
				$data['integrated_id'] = $invoice_data['integrated_id'];
			}
		}

		if (in_array($module, $ordering_modules) && \AppConfig::setting("model.detail.{$model}Adjustment")) {
			$lq_tax = new \ListQuery($model.'Adjustment');
			$lq_tax->addSimpleFilter($adj_field[$module], $data['id'], '=');
			$lq_tax->addSimpleFilter('related_type', 'TaxRates', '=');
			$tax_result = $lq_tax->fetchAllRows();
			$data['total_tax'] = 0;
			foreach($tax_result as $tax_row) {
				$data['total_tax'] += $tax_row['amount'];
			}

			$lq_tax = new \ListQuery($model.'Adjustment');
			$lq_tax->addSimpleFilter($adj_field[$module], $data['id'], '=');
			$lq_tax->addSimpleFilter('related_type', 'Discounts', '=');
			$tax_result = $lq_tax->fetchAllRows();
			$data['total_discount'] = 0;
			foreach($tax_result as $tax_row) {
				$data['total_discount'] += $tax_row['amount'];
			}

			$lq_tax = new \ListQuery($model.'Adjustment');
			$lq_tax->addSimpleFilter($adj_field[$module], $data['id'], '=');
			$lq_tax->addSimpleFilter('related_type', 'ShippingProviders', '=');
			$tax_result = $lq_tax->fetchAllRows();
			$data['total_shipping'] = 0;
			foreach($tax_result as $tax_row) {
				$data['total_shipping'] += $tax_row['amount'];
			}
			require_once('include/ListView/ListFormatter.php');
			$rel_fmt = new \ListFormatter($model, 'lines_link', null, 'listview', ['parent_key' => $data['id']]);
			$rel_lq =& $rel_fmt->getQuery();
			$rel_lq->addFields(array('id', 'line_group_id', 'name', 'quantity', 'mfr_part_no', 'serial_no', 'cost_price', 'list_price', 'unit_price', 'ext_price', 'tax_class_id'));
			$rel_result = $rel_fmt->getQueryResult(0, 999);
			ListEndpoint::filterExposedFields($rel_result->rows, $rel_result->getModelName());
			$data['line_items'] = array_values($rel_result->rows);
			foreach($data['line_items'] as &$line) {
				$rates = $seedcode->get_tax_rates($line['tax_class_id'], false);
				$rates_arr = [];
				foreach($rates as $k => $v) {
					$rates_arr[] = $v;
				}
				$lq_tax = new \ListQuery($model.'Adjustment');
				$lq_tax->addSimpleFilter('line_id', $line['id'], '=');
				$lq_tax->addSimpleFilter('type', 'PercentDiscount', '=');
				$disc_result = $lq_tax->fetchAllRows();
				$line['line_discount_amount'] = 0;
				$line['line_discount_perc'] = 0;
				foreach($disc_result as $disc_row) {
					$line['line_discount_amount'] += number_format(($line['ext_price']*$disc_row['rate'])/100, 2);
					$line['line_discount_perc'] = $disc_row['rate'];
				}
				if(isset($rates_arr[0]) && $rates_arr[0]) {
					$line['line_tax_perc'] = $rates_arr[0]['rate'];
					$line['line_tax_amount'] = number_format(($line['ext_price']*$rates_arr[0]['rate'])/100, 2);
				}
			}
		}

		$ret = [
			'record' => $data
		];
		ListEndpoint::filterExposedFields($ret, $model);
		ListEndpoint::expandImageFields($ret, $model);
		$module = \AppConfig::module_for_model($model);

		$hooks = \AppConfig::setting("model.hooks.$model.api_decorate", []);
		foreach ($hooks as $hook) {
			call_hook_function($hook, null, array($model, &$ret['record']));
		}

		if (\AppConfig::module_primary_bean($module) === $model) {
			$ret['module'] = $module;
		}

		return $this->withJson($res, $ret);
	}

	protected function patch($input, $res, $req) {
		$this->disabledForPortal($req);
		$model = $input['model'];
		$module = \AppConfig::module_for_model($model);
		if (!$module)
			throw new Errors\NotFound("Model $model not found");
		Endpoint::check_module_access($module);
		$record = \ListQuery::quick_fetch($model, $input['id'], true, ['acl_filters' => 'edit']);
		if (!$record && empty($input['create']))
			throw new Errors\NotFound;
		if ($record) {
			$update = \RowUpdate::for_result($record);
		} else {
			$update = \RowUpdate::blank_for_model($model);
			$update->set('id', $input['id']);
		}
		$_REQUEST = $input['data'];
		$update->loadRequest('', false);
		foreach ($input['data'] as $k => $v) {
			if (strpos($k, "_social_") === 0) {
				$update->set($k, $v);
			}
		}
		if (isset($input['X_ONECRM_APPLICATION'])) {
			$update->setRelatedData('X_ONECRM_APPLICATION', $input['X_ONECRM_APPLICATION']);
		}
		$update->save();
		$ret = [
			'result' => true,
		];

		return $this->withJson($res, $ret);
	}

}


