<?php

namespace OneCRM\API\Endpoints\Data;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class RelatedList extends Endpoint {

	public function endpoint() {
		return '/data/{model}/{id}/{link}';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
		if ($method == 'post')
			return $this->descriptorPost();
	}

	public function requiredScopes($method) {
		if ($method == 'get')
			return ['read'];
		return ['write'];
	}

	public function allowedMethods() {
		return ['get', 'post'];
	}

	private function descriptorGet() {
		$max =(int) \AppConfig::setting('api_config.max_list_items');
		$max = max($max, 200);
		$max = min($max, 500000);
		$desc = [
			'title' => 'Get list of related records',
			'docs' => '@endpoint/data/relatedlist_get',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => 1,
					'docs' => 'Record ID',
					'required' => true,
				],
				'link' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => 1,
					'docs' => 'Link name',
					'required' => true,
				],
				'fields' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'String',
						]
					],
					'location' => 'query',
					'docs' => 'Array of field names to fetch. When omitted, a limited number of fields are returned, depending on model. Record ID is guaranteed to be returned, and for most models, `name` and/or `_display` fields.',
				],
				'filter_text' => [
					'location' => 'query',
					'type' => 'String',
					'docs' => 'Generic search string. Fields involved in search depend on model',
				],
				'filters' => [
					'location' => 'query',
					'type' => 'Object',
					'docs' => 'Filters to apply. To get list of available filters for a model, use [metadata](#group_meta)',
					'value_type' => [
						'type' => 'String',
					]
				],
				'order' => [
					'location' => 'query',
					'type' => 'String',
					'docs' => 'Sort order',
				],
				'offset' => [
					'location' => 'query',
					'type' => 'Int',
					'min' => 0,
					'default' => 0,
					'docs' => 'Offset in the list to start retrieval from',
				],
				'limit' => [
					'location' => 'query',
					'type' => 'Int',
					'min' => 1,
					'max' => $max,
					'default' => 20,
					'docs' => 'Limits number of records returned',
				],

			],
			'return' => [
				'records' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'Array of retrieved records',
					'required' => true,
				],
				'total_results' => [
					'type' => 'Int',
					'docs' => 'Total number of results, without taking offset and limit in account',
					'required' => true,
				]
			]
		];
		return $desc;
	}

	private function descriptorPost() {
		static $desc = [
			'title' => 'Add related records',
			'docs' => '@endpoint/data/relatedlist_post',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Parent model',
					'required' => true,
				],
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => 1,
					'docs' => 'Parent ID',
					'required' => true,
				],
				'link' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => 1,
					'docs' => 'Link name',
					'required' => true,
				],
				'records' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'String',
							'min_len' => 1,
						]
					],
					'location' => 'body',
					'docs' => 'Array of related record IDs to be added to specified link',
				],
				'records_with_data' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						],
					],
					'location' => 'body',
					'docs' => 'Array of objects, each representing additional data to be inserted into join table',
				],

			],
		];
		return $desc;
	}



	protected function post($input, $res, $req) {
		$this->disabledForPortal($req);
		$model = $input['model'];
		$id = $input['id'];
		$link = $input['link'];
		$records = array_get_default($input, 'records', []);
		$records_with_data = array_get_default($input, 'records_with_data', []);
		require_once('include/ListView/ListFormatter.php');
		foreach($records_with_data as &$inv_records) {
			foreach($inv_records as $k => &$v) {
				if($k == 'invoice_id') {
					$invoice_row = \ListQuery::quick_fetch('Invoice', $v);
					if(!$invoice_row) {
						$result = \ListQuery::quick_fetch_key('Invoice', 'integrated_id', $v);
						if($result)
							$v = $result->getPrimaryKeyValue();
					}
				}
			}
		}
		$module = \AppConfig::module_for_model($model);
		if (!$module)
			throw new Errors\NotFound("Model $model not found");
		Endpoint::check_module_access($module);
		$row = \ListQuery::quick_fetch($model, $id, true, ['acl_filters' => 'edit']);
		if (!$row || $row->failed)
			throw new Errors\NotFound("Record not found");
		$upd = \RowUpdate::for_result($row);
		$upd->addUpdateLinks($link, $records);
		$upd->addUpdateLinks($link, $records_with_data);
		
		return $this->withJson($res, ['result' => true]);
	}

	protected function get($input, $res, $req) {
		$model = $input['model'];
		$module = \AppConfig::module_for_model($model);
		if (!$module)
			throw new Errors\NotFound("Model $model not found");


		require_once('include/ListView/ListFormatter.php');
		$fmt = new \ListFormatter($model, $input['link'], null, 'listview', ['parent_key' => $input['id']]);
		$lq =& $fmt->getQuery();

		$filter =& $fmt->getFilterForm();
		$filter->loadFilterLayout('Standard');

		$filter_data = array_get_default($input, 'filters', array());
		if (isset($input['filter_text']))
			$filter_data['filter_text'] = $input['filter_text'];


		$fmt->loadFilter($filter_data);


		if (!empty($input['fields']))
			$lq->addFields($input['fields']);
		$lq->addField('name');
		$lq->addField('_display');
		if (isset($input['order']))
			$lq->setOrderBy($input['order']);
		$this->applyPortalAcl($lq, $req);
		$result = $fmt->getQueryResult($input['offset'], $input['limit']);
		if (!$result || $result->failed)
			throw new Errors\InternalError;

		ListEndpoint::filterExposedFields($result->rows, $result->getModelName());
		ListEndpoint::expandImageFields($result->rows, $result->getModelName());
		$ret = [
			'records' => array_values($result->rows),
			'total_results' => $result->total_count,
		];
		$hooks = \AppConfig::setting("model.hooks." . $result->getModelName() .  ".api_decorate", []);
		foreach ($hooks as $hook) {
			foreach ($ret['records'] as &$rec) {
				call_hook_function($hook, null, array($result->getModelName(), &$rec));
			}
			unset($rec);
		}

		return $this->withJson($res, $ret);
	}

}


