<?php

namespace OneCRM\API\Endpoints\Files;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class Metadata {

	private $model;
	private $id;
	private $unchecked;

	public function __construct($model, $id, $unchecked = false) {
		$this->model = $model;
		$this->id = $id;
		$this->unchecked = $unchecked;
	}

	public function getMetadata() {
		switch ($this->model) {
			case 'Note':
				return $this->getNoteMetadata($this->id);
			case 'DocumentRevision':
				return $this->getRevisionMetadata($this->id);
			case 'Document':
				return $this->getDocumentMetadata($this->id);
		}
	}

	private function getObjectMetadata($model, $id) {
		$params = ['process_results' => false];
		if (!$this->unchecked)
			$params['acl_filters'] = 'view';
		$lq = new \ListQuery($model, true, $params);
		if (!$this->unchecked)
			$lq->addAclFilter('view');
		$result = $lq->queryRecord($id);
		if (!$result || $result->failed)
			throw new Errors\NotFound;
		$spec = $result->getFieldDefinition('filename');
		require_once 'include/upload_file.php';
		$upload = new \UploadFile($spec);
		$filename = $result->getField('filename');
		$path = $filename ? $upload->get_file_path($filename) : '';
		if (!file_exists($path))
			throw new Errors\NotFound;
		$size = filesize($path);
		return [
			'path' => $path,
			'name' => basename($filename),
			'size' => $size,
			'mimetype' => $result->getField('file_mime_type'),
			'modified' => filemtime($path),
		];
	}

	private function getNoteMetadata($id) {
		return $this->getObjectMetadata('Note', $id);
	}

	private function getRevisionMetadata($id) {
		return $this->getObjectMetadata('DocumentRevision', $id);
	}

	private function getDocumentMetadata($id) {
		$params = ['process_results' => false, 'acl_filters'];
		if (!$this->unchecked)
			$params['acl_filters'] = 'view';
		$lq = new \ListQuery('Document', true, $params);
		if (!$this->unchecked)
			$lq->addAclFilter('view');
		$result = $lq->queryRecord($id);
		if (!$result || $result->failed)
			throw new Errors\NotFound;
		return $this->getRevisionMetadata($result->getField('document_revision_id'));
	}

}


class Info extends Endpoint {

	public function endpoint() {
		return '/files/info/{model}/{id}';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
	}

	public function requiredScopes($method) {
		return ['read'];
	}

	public function allowedMethods() {
		return ['get'];
	}

	private function descriptorGet() {
		static $desc = [
			'title' => 'Get information about a file',
			'docs' => '@endpoint/files/info',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => [
						'type' => 'Enum',
						'options' => ['Note', 'Document', 'DocumentRevision'],
					],
					'docs' => 'Specifies model',
					'required' => true,
				],
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => 1,
					'docs' => 'Specifies ID',
					'required' => true,
				],
			],
			'return' => [
				'name' => [
					'type' => 'String',
					'docs' => 'File name',
					'required' => true,
				],
				'mimetype' => [
					'type' => 'String',
					'docs' => 'File MIME type',
					'required' => true,
				],
				'modified' => [
					'type' => 'Int',
					'docs' => 'File modification time, in secons since UNIX epoch',
					'required' => true,
				],
				'size' => [
					'type' => 'Int',
					'docs' => 'File sizee in bytes',
					'required' => true,
				],
				'temp_url' => [
					'type' => 'String',
					'docs' => 'Temporary download URL',
					'required' => false,
				],
			]
		];
		return $desc;
	}

	protected function get($input, $res, $req) {
		$meta = new Metadata($input['model'], $input['id']);
		$ret = $meta->getMetadata();
		unset($ret['path']);

		return $this->withJson($res, $ret);
	}

}


