<?php

namespace OneCRM\API\Endpoints\Layouts;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class ListView extends Endpoint {

	public function endpoint() {
		return '/layouts/list_view/{model}';
	}

	public function descriptor($method) {
		if ($method == 'post')
			return $this->descriptorPost();

		return null;
	}

	public function requiredScopes($method) {
		return ['write'];
	}

	public function allowedMethods() {
		return ['post'];
	}

	private function descriptorPost() {
		$max =(int) \AppConfig::setting('api_config.max_list_items');
		$max = max($max, 200);
		$max = min($max, 500000);

		$desc = [
			'title' => 'Set ListView layout(tab) params for a model',
			'docs' => '@endpoint/layouts/listview_post',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model',
					'required' => true,
				],
				'layout' => [
					'location' => 'body',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'New selected ListView layout(tab)',
					'required' => true,
				],
				'filter_layout' => [
					'location' => 'body',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Filters form layout',
					'required' => false,
				],
				'filter_values' => [
					'location' => 'body',
					'type' => 'Array',
					'docs' => 'Array of list filters values',
					'required' => false,
				],
				'offset' => [
					'location' => 'body',
					'type' => 'Int',
					'min' => 0,
					'default' => 0,
					'docs' => 'ListView offset value',
					'required' => false,
				],
				'limit' => [
					'location' => 'body',
					'type' => 'Int',
					'min' => 1,
					'max' => $max,
					'default' => 20,
					'docs' => 'ListView limit - number of records returned',
					'required' => false,
				],
				'order_by' => [
					'location' => 'body',
					'type' => 'String',
					'docs' => 'Sort order',
					'required' => false,
				],
			],
			'return' => [
				'return' => [
					'result' => [
						'type' => 'Bool',
						'docs' => 'Always true',
						'required' => true,
					],
				]
			]
		];
		return $desc;
	}

	protected function post($input, $res, $req) {
		$this->disabledForPortal($req);
		$model = $input['model'];
		$layout = $input['layout'];
		$module = \AppConfig::module_for_model($model);

		if (!$module)
			throw new Errors\NotFound("Model $model not found");

		if (empty($layout))
			throw new Errors\NotFound("Layout value is empty");

		Endpoint::check_module_access($module);

		if (! \ACLController::checkAccess($module, 'list'))
			throw new Errors\Forbidden("Access denied");

		require_once('include/ListView/ListViewManager.php');
		$manager = \ListViewManager::create('listview', ['is_primary' => true, 'module' => $module]);

		if (! empty($input['filter_values']) && is_array($input['filter_values'])) {
			$filter_values = $input['filter_values'];
		} else {
			$filter_values = [];
		}

		$session = array(
			'layout_name' => $layout,
			'offset' => $input['offset'] ?? 0,
			'limit' => $input['limit'] ?? 20,
			'order_by' => $input['order_by'] ?? '',
			'filter_layout' => $input['filter_layout'] ?? $layout,
			'filter_values' => $filter_values,
		);

		$manager->saveLayoutSession($session);
		$manager->loadLayoutSession();

		return $this->withJson($res, ['result' => true]);
	}
}


