<?php

namespace OneCRM\API\Endpoints\Meta;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;


class DetailView extends Endpoint {

	public function endpoint() {
		return '/meta/detail_view/{model}/{layout}';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
	}

	public function requiredScopes($method) {
		return ['read'];
	}

	public function allowedMethods() {
		return ['get'];
	}

	private function descriptorGet() {
		static $desc = [
			'title' => 'Get DetailView metadata for a model',
			'docs' => '@endpoint/meta/detailview_get',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'layout' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Selected DetailView layout',
					'required' => true,
				],
			],
			'return' => [
				'title' => [
					'type' => 'String',
					'docs' => 'Form title',
					'required' => true,
				],
				'current_layout' => [
					'type' => 'Array',
					'docs' => 'An array with current DetailView layout (tab) params',
					'required' => true,
				],
				'layouts' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'An Array of all DetailView layouts(tabs) for the module',
					'required' => true,
				],
				'buttons' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'An Array of custom module DetailView buttons',
					'required' => false,
				],
				'scripts' => [
					'type' => 'Array',
					'docs' => 'An array of custom scripts which should be included',
					'required' => false,
				],
				'form_hooks' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'An array of custom form hooks',
					'required' => false,
				],
				'summary' => [
					'type' => 'Object',
					'docs' => 'Summary part of the detail page',
					'required' => false,
				],
				'sections' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'An array of detail page sections',
					'required' => true,
				],
			]
		];

		return $desc;
	}

	protected function get($input, $res, $req) {
		$model = $input['model'];
		$layout = array_get_default($input, 'layout', 'Standard');

		$module = \AppConfig::module_for_model($model);
		if (!$module)
			throw new Errors\NotFound("Model $model not found");

		if (! \ACLController::checkAccess($module, 'view'))
			throw new Errors\Forbidden("Access denied");

		require_once 'include/utils/misc_functions.php';
		require_once('include/DetailView/DetailFactory.php');
		require_once('include/ListView/ListFormatter.php');

		/** @var \StandardDetailManager $mgr */
		$mgr = \DetailFactory::getManager('standard', $module);
		$mgr->layout_name = $layout;
		$mgr->format = 'html';
		$mgr->setModel($model);

		try {
			$mgr->standardInit();
		} catch(\Exception $e) {
			throw new Errors\InternalError("Can't load Detail manager for {$model}: {$e->getMessage()}");
		}

		$form_def = $mgr->form_def;
		$form_hooks = $form_def['form_hooks'] ?? [];

		$scripts = $this->formatScripts($form_def['scripts'] ?? []);
		$this->addDisplayScripts($scripts, $model);

		$result = [
			'title' => $form_def['title'],
			'current_layout' => [
				'name' => $mgr->layout->getName(),
				'type' => $mgr->layout->type
			],
			'layouts' => $mgr->layout->getTabs(),
			'buttons' => $this->convertToStdArray($form_def['form_buttons']),
			'scripts' => $scripts,
			'form_hooks' => $this->formatHooks($form_hooks),
			'summary' => $form_def['summary'] ?? [],
			'sections' => $form_def['sections']
		];

		return $this->withJson($res, $result);
	}

	private function convertToStdArray($list) {
		$result = [];

		array_walk($list, function ($item, $key) use (&$result) {
			if (! isset($item['name']))
				$item['name'] = $key;

			$result[] = $item;
		});

		return $result;
	}

	private function formatHooks($list) {
		$result = [];

		array_walk($list, function ($item, $key) use (&$result) {
			$result[] = ['name' => $key, 'hook' => $item];
		});

		return $result;
	}

	private function formatScripts($list) {
		$result = [];

		array_walk($list, function ($item, $key) use (&$result) {
			if (! empty($item['file']))
				$result[] = $item['file'];
		});

		return $result;
	}

	private function addDisplayScripts(&$scripts, $model) {
		$types = ['edit', 'view'];

		foreach ($types as $type) {
			$type_scripts = \AppConfig::setting("display.{$type}.{$model}.scripts");

			if (is_array($type_scripts)) {
				$scripts = array_merge($scripts, $this->formatScripts($type_scripts));
			}
		}

		$scripts = array_unique($scripts);
	}
}
