<?php

namespace OneCRM\API\Endpoints\Meta;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;


class ListView extends Endpoint {

	public function endpoint() {
		return '/meta/list_view/{model}';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
	}

	public function requiredScopes($method) {
		return ['read'];
	}

	public function allowedMethods() {
		return ['get'];
	}

	private function descriptorGet() {
		static $desc = [
			'title' => 'Get ListView metadata for a model',
			'docs' => '@endpoint/meta/listview_get',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'layout' => [
					'location' => 'body',
					'type' => [
						'type' => 'String',
					],
					'docs' => 'Get metadata for this layout',
					'required' => false,
				],
			],
			'return' => [
				'current_layout' => [
					'type' => 'Array',
					'docs' => 'An array with current ListView layout (tab) params',
					'required' => true,
				],
				'layouts' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'Array of all ListView layouts for the module',
					'required' => true,
				],
				'settings' => [
					'type' => 'Object',
					'docs' => 'An array with Listview layout settings',
					'required' => true,
				],
				'hooks' => [
					'type' => 'Array',
					'docs' => 'An array with ListView hooks',
					'required' => false,
				],
				'columns' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'Array of ListView columns',
					'required' => true,
				],
				'filters' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'Array of filters for the current layout',
					'required' => true,
				],
				'hidden_fields' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'Array of hidden fields',
					'required' => false,
				],
				'mass_update' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Array',
							'schema' => [
								'buttons' => [
									'type' => 'Array',
									'docs' => 'ListView mass update buttons(actions)',
									'element_type' => [
										'type' => 'Object'
									]
								],
								'fields' => [
									'type' => 'Array',
									'docs' => 'Mass update form fields',
									'element_type' => [
										'type' => 'Object'
									]
								],
							]
						]
					],
					'docs' => 'Mass Update details - buttons and fields',
					'required' => false,
				],
			]
		];

		return $desc;
	}

	public function get($input, $res, $req) {
		$model = $input['model'];

		$module = \AppConfig::module_for_model($model);
		if (!$module)
			throw new Errors\NotFound("Model $model not found");

		if (! \ACLController::checkAccess($module, 'list'))
			throw new Errors\Forbidden("Access denied");

		if (\AppConfig::setting("display.list.{$model}.disabled"))
			throw new Errors\Forbidden("ListView disabled for {$module}");

		require_once('include/ListView/ListViewManager.php');
		$manager = \ListViewManager::create('listview', ['is_primary' => true, 'module' => $module]);
		$manager->loadLayoutSession();

		if (! empty($input['layout'])) {
			$layout = $input['layout'];
			$current_layout = ['layout_name' => $input['layout']];
		} elseif (! empty($manager->layout_session) && ! empty($manager->layout_session['layout_name'])) {
			$layout = $manager->layout_session['layout_name'];
			$current_layout = $manager->layout_session;
		} else {
			$layout = 'Browse';
			$current_layout = ['layout_name' => $layout];
		}

		if (! $manager->initModuleView($module, $layout))
			throw new Errors\NotFound(translate('LBL_VIEW_NOT_FOUND', 'app'));

		$fmt =& $manager->getFormatter();

		if (! $manager->initFormatter($fmt))
			throw new Errors\NotFound(translate('LBL_VIEW_NOT_FOUND', 'app'));

		$layouts = $manager->layout_options;

		$filter =& $fmt->getFilterForm();
		$filter->loadFilterLayout($layout);

		$result = [
			'current_layout' => $current_layout,
			'layouts' => $this->formatLayoutsList($layouts),
			'settings' => $this->fillSettings($manager),
			'hooks' => $manager->hooks,
			'columns' => $this->formatColumns($fmt->getColumns()),
			'filters' => $this->convertToStdArray($filter->getFilterOptions()),
			'hidden_fields' => $this->formatHiddenFields($manager->getHiddenFields()),
			'mass_update' => [],
		];

		$mass_update = $fmt->getMassUpdate();
		//can be call in in button hooks
		require_once('include/layout/pages/ModulePage.php');
		$GLOBALS['pageInstance'] = new \ModulePage();

		if ($manager->show_mass_update) {
			$result['mass_update']['buttons'] = $this->convertToStdArray($mass_update->getButtons($model, $module));
			$result['mass_update']['fields'] = $this->convertToStdArray($mass_update->getFields($model));
		}

		return $this->withJson($res, $result);
	}

	private function formatLayoutsList($list) {
		$result = [];

		array_walk($list, function ($item, $key) use (&$result) {
			if (isset($item['vname']))
				unset($item['vname']);

			if (isset($item['view_name']))
				unset($item['view_name']);

			$result[] = $item;
		});

		return $result;
	}

	private function formatColumns($columns) {
		$result = [];

		foreach ($columns as $column) {
			$item = $column;

			if (! empty($item['add_fields'])) {
				foreach ($item['add_fields'] as $index => $field) {
					$this->removeExcessColumnsData($item['add_fields'][$index]);
				}
			}

			if (! empty($item['type']) && $item['type'] == 'widget')
				$item['widget'] = get_class($item['widget']);

			$this->removeExcessColumnsData($item);
			$result[] = $item;
		}

		return $result;
	}

	private function formatHiddenFields($list) {
		$result = [];

		array_walk($list, function ($item, $key) use (&$result) {
			$result[] = ['name' => $key, 'value' => $item];
		});

		return $result;
	}

	private function convertToStdArray($list) {
		$result = [];

		array_walk($list, function ($item, $key) use (&$result) {
			if (! isset($item['name']))
				$item['name'] = $key;

			$result[] = $item;
		});

		return $result;
	}

	private function removeExcessColumnsData(&$column) {
		$excess_fields = ['std_vname', 'dbType', 'alias'];

		foreach ($excess_fields as $field) {
			if (isset($column[$field]))
				unset($column[$field]);
		}
	}

	private function fillSettings(\ListViewManager $manager) {
		$settings = [
			'show_title', 'show_tabs', 'show_filter', 'disable_mass_update_data',
			'show_mass_update', 'show_create_button', 'show_checks', 'show_additional_details',
			'show_favorites', 'show_compact_empty', 'show_help', 'hide_navigation_controls',
			'disable_request_filter'
		];

		$result = [];

		foreach ($settings as $option) {
			if (property_exists($manager, $option)) {
				$result[$option] = $manager->$option;
			}
		}

		return $result;
	}
}
