<?php

namespace OneCRM\API\Endpoints\Meta;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;


class SubPanels extends Endpoint {

	public function endpoint() {
		return '/meta/sub_panels/{model}/{detail_layout}';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
	}

	public function requiredScopes($method) {
		return ['read'];
	}

	public function allowedMethods() {
		return ['get'];
	}

	private function descriptorGet() {
		static $desc = [
			'title' => 'Get Sub Panels metadata for a model',
			'docs' => '@endpoint/meta/subpanels_get',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'detail_layout' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'DetailView layout (tab)',
					'required' => true,
				],
			],
			'return' => [
				'panels' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
							'schema' => [
								'name' => [
									'type' => 'String',
									'docs' => 'SubPanel name',
									'required' => true
								],
								'module' => [
									'type' => 'String',
									'docs' => 'SubPanel module',
									'required' => true
								],
								'bean_name' => [
									'type' => 'String',
									'docs' => 'SubPanel model name',
									'required' => true
								],
								'model_type' => [
									'type' => 'String',
									'docs' => 'SubPanel model type',
									'required' => true
								],
								'managed' => [
									'type' => 'Bool',
									'required' => false
								],
								'removeable' => [
									'type' => 'Bool',
									'required' => false
								],
								'updateable' => [
									'type' => 'Bool',
									'required' => false
								],
								'no_create' => [
									'type' => 'Bool',
									'docs' => 'Hide Create button if true',
									'required' => false
								],
								'icon' => [
									'type' => 'String',
									'docs' => 'Custom icon',
									'required' => false
								],
								'detail_link_url' => [
									'type' => 'String',
									'docs' => 'Custom DetailView URL',
									'required' => false
								],
								'detail_link_action' => [
									'type' => 'String',
									'docs' => 'Custom DetailView Action',
									'required' => false
								],
								'layout' => [
									'type' => 'String',
									'docs' => 'SubPanel layout name',
									'required' => false
								],
								'layout_type' => [
									'type' => 'String',
									'docs' => 'SubPanel layout type',
									'required' => false
								],
								'title' => [
									'type' => 'String',
									'docs' => 'Title lang constant',
									'required' => false
								],
								'target_key' => [
									'type' => 'Bool',
									'docs' => 'Target Key',
									'required' => false
								],
								'source_key' => [
									'type' => 'Bool',
									'docs' => 'Source Key',
									'required' => false
								],
								'custom_buttons' => [
									'type' => 'Array',
									'docs' => 'An array with SubPanel custom buttons',
									'element_type' => [
										'type' => 'Object'
									]
								],
								'columns' => [
									'type' => 'Array',
									'docs' => 'An array with SubPanel columns',
								],
							]
						]
					],
					'docs' => 'A list of SubPanels with metadata',
					'required' => true,
				],

			]
		];

		return $desc;
	}

	protected function get($input, $res, $req) {
		$model = $input['model'];
		$layout = array_get_default($input, 'detail_layout', 'Standard');

		$module = \AppConfig::module_for_model($model);

		if (!$module)
			throw new Errors\NotFound("Model $model not found");

		if (! \ACLController::checkAccess($module, 'view'))
			throw new Errors\Forbidden("Access denied");

		require_once 'include/utils/misc_functions.php';
		require_once('include/DetailView/DetailFactory.php');
		require_once('include/ListView/ListFormatter.php');

		/** @var \StandardDetailManager $mgr */
		$mgr = \DetailFactory::getManager('standard', $module);
		$mgr->layout_name = $layout;
		$mgr->format = 'html';
		$mgr->setModel($model);

		try {
			$mgr->standardInit();
		} catch(\Exception $e) {
			throw new Errors\InternalError("Can't load Detail manager for {$model}: {$e->getMessage()}");
		}

		$panels = $mgr->subpanelMgr->getVisiblePanels();
		$result = [];

		foreach($panels as &$panel) {
			$panel['custom_buttons'] = $mgr->subpanelMgr->getPanelButtons($panel, '');

			$fmt = new \ListFormatter($model, $panel['name'], null, 'subpanel');

			try {
				$layout_loaded = $fmt->loadLayout($panel['layout_name'], $panel['layout_type'], $panel['module']);

				if (! $layout_loaded)
					throw new Errors\InternalError("Can't load {$panel['name']} sub panel layout for {$model}");

			} catch (\Exception $e) {
				throw new Errors\InternalError("Can't load {$panel['name']} sub panel layout for {$model}: {$e->getMessage()}");
			}

			$columns = $fmt->getColumns();
			$this->formatColumns($columns);
			$panel['columns'] = $columns;

			$result[] = $panel;
		}

		return $this->withJson($res, ['panels' => $result]);
	}

	private function formatColumns(& $list) {
		array_walk($list, function (&$item) {
			if (isset($item['widget']) && ! empty($item['widget']))
				$item['widget'] = get_class($item['widget']);
		});
	}
}
