<?php

namespace OneCRM\API\Endpoints\Reports;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class ArchivedRuns extends Endpoint {

	public function endpoint() {
		return '/reports/{model}/{id}';
	}

	public function descriptor($method) {
		if ($method == 'get') {
			return $this->descriptorGet();
		}
	}

	public function requiredScopes($method) {
		return ['read'];
	}

	public function allowedMethods() {
		return ['get'];
	}

	private function descriptorGet() {
		$max =(int) \AppConfig::setting('api_config.max_list_items');
		$max = max($max, 200);
		$max = min($max, 500000);

		$desc = [
			'title' => 'Get list of archived runs',
			'docs' => '@endpoint/reports/archived_runs',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Report ID',
					'required' => true,
				],
				'offset' => [
					'location' => 'query',
					'type' => 'Int',
					'min' => 0,
					'default' => 0,
					'docs' => 'Offset in the list to start retrieval from',
				],
				'limit' => [
					'location' => 'query',
					'type' => 'Int',
					'min' => 1,
					'max' => $max,
					'default' => 20,
					'docs' => 'Limits number of records returned',
				],
			],
			'return' => [
				'records' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
						]
					],
					'docs' => 'Array of retrieved records',
					'required' => true,
				],
				'total_results' => [
					'type' => 'Int',
					'docs' => 'Total number of results, without taking offset and limit',
					'required' => true,
				]
			]
		];
		return $desc;
	}

	protected function get($input, $res, $req) {
		$model = $input['model'];

		$module = \AppConfig::module_for_model($model);

		if (!$module)
			throw new Errors\NotFound("Model $model not found");

		Endpoint::check_module_access($module,"");

		// Access for User only with ACL report all.
		if (!\ACLController::checkAccess($module, "report", false, false))
			throw new Errors\Forbidden("Access denied");

		require_once('include/ListView/ListFormatter.php');
		$fmt = new \ListFormatter('ReportData');
		$lq =& $fmt->getQuery();

		$fields = \AppConfig::setting('model.fields.ReportData');
		$lq->addFields($fields);

		$lq->addSimpleFilter("primary_module", $module);
		$lq->addSimpleFilter("report_id", $input['id']);

		$result = $fmt->getQueryResult($input['offset'], $input['limit']);
		if (!$result || $result->failed)
			throw new Errors\NotFound("Report data for {$model} not found");

		\OneCRM\API\Endpoints\Data\ListEndpoint::filterExposedFields($result->rows, $model);

		$ret = [
			'records' => array_values($result->rows),
			'total_results' => $result->total_count,
		];

		return $this->withJson($res, $ret);
	}
}


