<?php

namespace OneCRM\API\Endpoints\Tally;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class Record extends Endpoint {

	static $SupportedModels = [
		'Quote', 'Invoice', 'Bill',
		'SalesOrder', 'PurchaseOrder', 'CreditNote',
		'Shipping', 'Receiving',
	];

	public function endpoint() {
		return '/tally/{model}/{id}';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
	}

	public function requiredScopes($method) {
		if ($method == 'get')
			return ['read'];
		return ['write'];
	}

	public function allowedMethods() {
		return ['get'];
	}

	private function descriptorGet() {
		static $desc = [
			'title' => 'Get a Tally Object',
			'params' => [
				'model' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model to query',
					'required' => true,
				],
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Record ID',
					'required' => true,
				],
			],
			'return' => [
				'record' => [
					'type' => 'Object',
					'docs' => '"Parent" record as if it was retrieved using `/data/{model}/{id}` endpoint',
				],
				'groups' => [
					'type' => 'Array',
					'docs' => 'Array of groups, each group contains one or more line items',
				],
			],
		];
		return $desc;
	}

	protected function get($input, $res, $req) {
		$model = $input['model'];
		if (!in_array($model, self::$SupportedModels))
			throw new Errors\BadRequest("Unsupported model");
		$module = \AppConfig::module_for_model($model);
		$id = $input['id'];
		if (!$module)
			throw new Errors\NotFound("Model $model not found");
		Endpoint::check_module_access($module);
		$params = ['process_results' => true, 'acl_filters' => 'view'];
		$lq = new \ListQuery($model, true, $params);
		$lq->addAclFilter('view');
		$fs = $lq->base_model->getFieldDefinitions();
		foreach ($fs as $f) {
			if ($f['type'] === 'ref')
				$lq->addField($f['name']);
		}
		//$this->applyPortalAcl($lq, $req);
		$result = $lq->queryRecord($id);
		$data = null;
		if ($result && !$result->failed)
			$data = $result->row;

		$upd = \RowUpdate::for_result($result, null);
		$groups = $upd->getGroups();
		if ($model != 'Shipping' && $model != 'Receiving') {
			$totals = $upd->getTotals();
		}
		$fmt = new GroupFormatter();
		$groups = array_map(function($g) use($fmt, $model) {
			return $fmt->formatGroup($g, $model);
		}, array_values($groups));

		if (!$data)
			throw new Errors\NotFound;

		$ret = [
			'record' => $data,
			'groups' => $groups,
		];
		if ($model != 'Shipping' && $model != 'Receiving') {
			$ret['totals'] = $totals;
		}

		return $this->withJson($res, $ret);
	}
}


