<?php

namespace OneCRM\API\Endpoints\Util;

use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class PushToken extends Endpoint {

	const EXPIRATION_TIME = 2592000; //30 * 24 * 3600;

	public function endpoint() {
		return '/pushtoken/{token}';
	}

	public function descriptor($method) {
		if ($method == 'post')
			return $this->descriptorPost();
		if ($method == 'delete')
			return $this->descriptorDelete();
	}

	public function requiredScopes($method) {
		return [];
	}

	public function allowedMethods() {
		return ['post', 'delete'];
	}

	public function requiresAuthentication() {
		return true;
	}

	private function descriptorPost() {
		static $desc = [
			'hidden' => true,
			'title' => 'Register push notification token',
			'docs' => '@endpoint/util/push_register',
			'params' => [
				'type' => [
					'location' => 'body',
					'type' => 'Enum',
					'options' => ['apns', 'fcm'],
					'docs' => 'Token type. Currently, apns and fcm are supported',
					'required' => true,
				],
				'device_id' => [
					'location' => 'body',
					'type' => 'String',
					'min_len' => 1,
					'docs' => 'Device ID',
					'required' => true,
				],
				'sandbox' => [
					'location' => 'body',
					'type' => 'Bool',
					'docs' => 'True if sandbox environment should be used',
					'required' => false,
					'default' => false,
				],
				'app_name' => [
					'location' => 'body',
					'type' => 'String',
					'min_len' => 1,
					'docs' => 'Application name',
					'required' => false,
					'default' => '1crm_mobile',
				],
				'token' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => 1,
					'docs' => 'Device token',
					'required' => true,
				],
			],
			'return' => [
				'expires' => [
					'type' => 'DateTime',
					'docs' => 'Date/time when token registration will expire',
					'required' => true,
				],
			]
		];
		return $desc;
	}

	private function descriptorDelete() {
		static $desc = [
			'hidden' => true,
			'title' => 'Unregister push notification token',
			'docs' => '@endpoint/util/push_unregister',
			'params' => [
				'token' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => 1,
					'docs' => 'Device token',
					'required' => true,
				],
			],
			'return' => [
				'success' => [
					'type' => 'Bool',
					'docs' => 'True on success',
					'required' => true,
				],
			]
		];
		return $desc;
	}

	protected function post($input, $res, $req) {
		$this->disabledForPortal($req);
		$token = $input['token'];
		$device_id = $input['device_id'];
		$type = $input['type'];
		$app_name = $input['app_name'];
		$sandbox = empty($input['sandbox']) ? 0 : 1;
		$user_id = \AppConfig::current_user_id();
		if (empty($input['from_1crm_mobile'])) {
			throw new Errors\NotFound;
		}
		$expires = gmdate('Y-m-d H:i:s', time() + self::EXPIRATION_TIME);

		$result = \ListQuery::quick_fetch_key('push_tokens', 'device_id', $device_id);

		if ($result)
			$upd = \RowUpdate::for_result($result);
		else
			$upd = \RowUpdate::blank_for_model('push_tokens');

		$upd->set('token', $token);
		$upd->set('device_id', $device_id);
		$upd->set('token_type', $type);
		$upd->set('app_name', $app_name);
		$upd->set('user_id', $user_id);
		$upd->set('expires', $expires);
		$upd->set('sandbox', $sandbox);
		$upd->save();

		$ret = [
			'expires' => $expires,
		];

		return $this->withJson($res, $ret);
	}

	protected function delete($input, $res, $req) {
		$this->disabledForPortal($req);
		$token = $input['token'];
		$user_id = \AppConfig::current_user_id();
		if (empty($input['from_1crm_mobile'])) {
			throw new Errors\NotFound;
		}
		$expires = gmdate('Y-m-d H:i:s', time() + self::EXPIRATION_TIME);

		$result = \ListQuery::quick_fetch_key('push_tokens', 'token', $token);

		if ($result) {
			if ($result->getField('user_id') == $user_id) {
				$upd = \RowUpdate::for_result($result);
				$upd->deleteRow();
			}
		}
		$ret = [
			'success' => true,
		];

		return $this->withJson($res, $ret);
	}

}


