<?php

namespace OneCRM\API\Endpoints\Webhooks;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

require_once 'modules/Webhooks/WebhookFormulaParser.php';

class Webhook extends Endpoint {

	public function endpoint() {
		return '/webhooks/{id}';
	}

	public function descriptor($method) {
		if ($method == 'get')
			return $this->descriptorGet();
		if ($method == 'put')
			return $this->descriptorPut();
		if ($method == 'delete')
			return $this->descriptorDelete();
	}

	public function requiredScopes($method) {
		if ($method == 'get')
			return ['read'];
		return ['write'];
	}

	public function allowedMethods() {
		return ['delete', 'get', 'put'];
	}

	private function descriptorPut() {
		static $desc = [
			'title' => 'Update a webhook',
			'params' => [
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Webhook ID',
					'required' => true,
				],
				'type' => [
					'location' => 'body',
					'type' => 'Enum',
					'options' => ['create', 'update', 'create_update'],
					'docs' => 'Webhook type',
					'required' => true,
				],
				'url' => [
					'location' => 'body',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Webhook URL',
					'required' => true,
				],
				'model' => [
					'location' => 'body',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model that triggers this webhook',
					'required' => true,
				],
				'filters' => [
					'location' => 'body',
					'type' => 	'WebhookFilter',
					'docs' => 'Describes filters to trigger webhook on specific events only. ',
					'required' => false,
				],
			],
			'return' => [
				'id' => [
					'type' => 'String',
					'docs' => 'Webhook ID',
					'required' => true,
				],
				'type' => [
					'type' => 'String',
					'docs' => 'Webhook type',
					'required' => true,
				],
				'url' => [
					'type' => 'String',
					'docs' => 'Webhook URL',
					'required' => true,
				],
				'model' => [
					'type' => 'String',
					'docs' => 'Model that triggers this webhook',
					'required' => true,
				],
				'filters' => [
					'type' => 'WebhookFilter',
					'docs' => 'Model-specific filters',
					'required' => false,
				],
			]
		];
		return $desc;
	}

	private function descriptorDelete() {
		static $desc = [
			'title' => 'Delete a webhook',
			'docs' => 'Delete webhook identified by `id`',
			'params' => [
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Webhook ID',
					'required' => true,
				],

			],
			'return' => [
				'id' => [
					'type' => 'String',
					'docs' => 'Webhook ID',
					'required' => true,
				],
				'type' => [
					'type' => 'String',
					'docs' => 'Webhook type',
					'required' => true,
				],
				'url' => [
					'type' => 'String',
					'docs' => 'Webhook URL',
					'required' => true,
				],
				'model' => [
					'type' => 'String',
					'docs' => 'Model that triggers this webhook',
					'required' => true,
				],
			]
		];
		return $desc;
	}

	private function descriptorGet() {
		static $desc = [
			'title' => 'Get a webhook',
			'docs' => 'Get data for webhook identified by `id`',
			'params' => [
				'id' => [
					'location' => 'path',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Webhook ID',
					'required' => true,
				],

			],
			'return' => [
				'id' => [
					'type' => 'String',
					'docs' => 'Webhook ID',
					'required' => true,
				],
				'type' => [
					'type' => 'String',
					'docs' => 'Webhook type',
					'required' => true,
				],
				'url' => [
					'type' => 'String',
					'docs' => 'Webhook URL',
					'required' => true,
				],
				'model' => [
					'type' => 'String',
					'docs' => 'Model that triggers this webhook',
					'required' => true,
				],
				'filters' => [
					'type' => 'WebhookFilter',
					'docs' => 'Model-specific filters.' ,
					'required' => false,
				],
			]
		];
		return $desc;
	}

	protected function delete($input, $res, $req) {
		$this->disabledForPortal($req);
		$id = $input['id'];
		$record = \ListQuery::quick_fetch('Webhook', $input['id'], true, ['acl_filters' => 'delete']);
		if (!$record)
			throw new Errors\NotFound;
		$update = \RowUpdate::for_result($record);
		$result = $update->markDeleted();

		return $this->withJson($res, [
			'id' => $record->getField('id'),
			'type' => $record->getField('type'),
			'url' => $record->getField('url'),
			'model' => $record->getField('model'),
		]);
	}

	protected function get($input, $res, $req) {
		$this->disabledForPortal($req);
		$json = getJSONobj();
		$id = $input['id'];
		$record = \ListQuery::quick_fetch('Webhook', $input['id'], true, ['acl_filters' => 'view']);
		if (!$record)
			throw new Errors\NotFound;
		$filters = $json->decode($record->getField('filters'));
		
		return $this->withJson($res, [
			'id' => $record->getField('id'),
			'type' => $record->getField('type'),
			'url' => $record->getField('url'),
			'model' => $record->getField('model'),
			'filters' => $filters
		]);
	}

	protected function put($input, $res, $req) {
		$this->disabledForPortal($req);
		$id = $input['id'];
		$record = \ListQuery::quick_fetch('Webhook', $input['id'], true, ['acl_filters' => 'delete']);
		if (!$record)
			throw new Errors\NotFound;
		if (!\AppConfig::setting("model.detail.{$input['model']}"))
			throw new Errors\BadRequest("Unknown model {$input['model']}");
		$json = getJSONobj();
		$filters = array_get_default($input, 'filters');
		if ($filters) {
			if ($filters['glue'] == 'formula' && $filters['formula']) {
				list($success) = \WebhookFormulaParser::parse($filters['formula']);
				if (!$success)
					throw new Errors\BadRequest('Invalid formula');
			}
		}
		$update = \RowUpdate::for_result($record);
		$update->set([
			'url' => $input['url'],
			'type' => $input['type'],
			'model' => $input['model'],
			'filters' => $json->encode($filters),
		]);
		$update->save();
		
		return $this->withJson($res, [
			'id' => $update->getPrimaryKeyValue(),
			'url' => $input['url'],
			'type' => $input['type'],
			'model' => $input['model'],
			'filters' => array_get_default($input, 'filters'),
		]);
	}
}


