<?php

namespace OneCRM\API\Endpoints\Webhooks;
use OneCRM\API\Endpoint;
use OneCRM\API\Errors;

class Webhooks extends Endpoint {

	public function endpoint() {
		return '/webhooks';
	}

	public function descriptor($method) {
		if ($method == 'post')
			return $this->descriptorPost();
		if ($method == 'get')
			return $this->descriptorGet();
	}

	public function requiredScopes($method) {
		if ($method == 'get')
			return ['read'];
		return ['write'];
	}

	public function allowedMethods() {
		return ['get', 'post'];
	}

	private function descriptorPost() {
		static $desc = [
			'title' => 'Create a webhook',
			'params' => [
				'type' => [
					'location' => 'body',
					'type' => 'Enum',
					'options' => ['create', 'update', 'create_update'],
					'docs' => 'Webhook type',
					'required' => true,
				],
				'url' => [
					'location' => 'body',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Webhook URL',
					'required' => true,
				],
				'model' => [
					'location' => 'body',
					'type' => 'String',
					'min_len' => '1',
					'docs' => 'Model that triggers this webhook',
					'required' => true,
				],
				'filters' => [
					'location' => 'body',
					'type' => 'WebhookFilter',
					'docs' => 'Describes filters to trigger webhook on specific events only. ',
					'required' => false,
				],

			],
			'return' => [
				'id' => [
					'type' => 'String',
					'docs' => 'Webhook ID',
					'required' => true,
				],
				'type' => [
					'type' => 'String',
					'docs' => 'Webhook type',
					'required' => true,
				],
				'url' => [
					'type' => 'String',
					'docs' => 'Webhook URL',
					'required' => true,
				],
				'model' => [
					'type' => 'String',
					'docs' => 'Model that triggers this webhook',
					'required' => true,
				],
				'filters' => [
					'type' => 'WebhookFilter',
					'docs' => 'Model-specific filters',
					'required' => false,
				],
			],
		];
		return $desc;
	}

	private function descriptorGet() {
		static $desc = [
			'title' => 'Get List of webhooks',
			'docs' => 'Get List of webhooks',
			'params' => [
				'offset' => [
					'location' => 'query',
					'type' => 'Int',
					'min' => 0,
					'default' => 0,
					'docs' => 'Offset in the list to start retrieval from',
				],
				'limit' => [
					'location' => 'query',
					'type' => 'Int',
					'min' => 1,
					'max' => 200,
					'default' => 20,
					'docs' => 'Limits number of records returned',
				],
			],
			'return' => [
				'total_results' => [
					'type' => 'Int',
					'docs' => 'Total number of results, without taking offset and limit in account',
					'required' => true,
				],
				'records' => [
					'type' => [
						'type' => 'Array',
						'element_type' => [
							'type' => 'Object',
							'schema' => [
								'id' => [
									'type' => 'String',
									'docs' => 'Webhook ID',
									'required' => true,
								],
								'type' => [
								'type' => 'String',
								'docs' => 'Webhook type',
								'required' => true,
								],
								'url' => [
								'type' => 'String',
								'docs' => 'Webhook URL',
								'required' => true,
								],
								'model' => [
								'type' => 'String',
								'docs' => 'Model that triggers this webhook',
								'required' => true,
								],
								'filters' => [
								'type' => 'WebhookFilter',
								'docs' => 'Model-specific filters',
								'required' => false,
								],
								'assigned_user_id' => [
								'type' => 'String',
								'docs' => 'User assigned to the webhook',
								'required' => true,
								],
							],
						],
					],
				],
			]
		];
		return $desc;
	}

	protected function post($input, $res, $req) {
		$this->disabledForPortal($req);
		$json = getJSONobj();
		if (!\AppConfig::setting("model.detail.{$input['model']}"))
			throw new Errors\BadRequest("Unknown model {$input['model']}");
		$allowed = \ACLController::checkAccess('Webhooks', 'edit');
		if (!$allowed)
			throw new Errors\Forbidden("Access denied");
		$filters = array_get_default($input, 'filters');
		if ($filters) {
			if ($filters['glue'] == 'formula' && $filters['formula']) {
				list($success) = \WebhookFormulaParser::parse($filters['formula']);
				if (!$success)
					throw new Errors\BadRequest('Invalid formula');
			}
		}
		$update = \RowUpdate::blank_for_model('Webhook');
		$update->set([
			'url' => $input['url'],
			'type' => $input['type'],
			'model' => $input['model'],
			'assigned_user_id' => \AppConfig::current_user_id(),
			'filters' => $json->encode($filters),
		]);
		$update->save();
		
		return $this->withJson($res, [
			'url' => $input['url'],
			'type' => $input['type'],
			'model' => $input['model'],
			'id' => $update->getPrimaryKeyValue(),
			'filters' => array_get_default($input, 'filters'),
		]);
	}

	protected function get($input, $res, $req) {
		require_once('include/ListView/ListFormatter.php');
		$fmt = new \ListFormatter('Webhook');
		$lq =& $fmt->getQuery();

		$filter =& $fmt->getFilterForm();
		$filter->loadFilterLayout('Standard');

		$filter_data = array_get_default($input, 'filters', array());
		if (isset($input['filter_text']))
			$filter_data['filter_text'] = $input['filter_text'];

		$fmt->loadFilter($filter_data);

		if (isset($input['order']))
			$lq->setOrderBy($input['order']);
		$this->applyPortalAcl($lq, $req);
		$result = $fmt->getQueryResult($input['offset'], $input['limit']);
		if (!$result || $result->failed)
			throw new Errors\InternalError;

		$json = getJSONObj();
		$records = array_map(function($row) use ($json) {
			if (!empty($row['filters'])) {
				$row['filters'] = $json->decode($row['filters']);
			}
			return [
         		'filters' => $row['filters'],
         		'assigned_user_id' => $row['assigned_user_id'],
		        'url' => $row['url'],
		        'type' => $row['type'],
         		'model' => $row['model'],
		        'id' => $row['id'],
			];
		}, array_values($result->rows));

		$ret = [
			'records' => $records,
			'total_results' => $result->total_count,
		];

		return $this->withJson($res, $ret);
	}

}


