<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo, 
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */



require_once('include/config/ConfigBase.php');
require_once('modules/ACLActions/ACLAction.php');

class ACLManager extends ConfigBase {
	var $acl;
	var $acl_standard;
	var $acl_time;
	var $default_role;
	var $session_cache = true;
	
	static $acl_lookup = array(
		'by_user' => array('this', 'loadUserActions'),
		'by_role' => array('this', 'loadRoleActions'),
		'by_team' => array('this', 'loadTeamActions'),
		'user_roles' => array('this', 'loadUserRoles'),
		'user_teams' => array('this', 'loadUserTeams'),
		'allowed_users' => array('ACLAction', 'loadAllowedUsers'),
	);
	static $base_model_actions = array(
		'list', 'view', 'create', 'edit', 'delete',
		'import', 'export', 'report', 'manage_personal'
	); // access, approve
	static $sanity_checks = array(
		// left requires right
		array('list', 'view'),
		array('edit', 'view'),
		array('delete', 'view'),
		//array('import', 'edit'), // has limited values
		array('create', 'edit'),
		array('export', 'list'),
		array('report', 'list'),
		array('approve', 'edit'),
		array('manage_personal', 'edit'),
	);
	static $access_levels = array(
		ACL_ALLOW_ALL => 5, ACL_ALLOW_GROUP => 4, ACL_ALLOW_ROLE => 3, ACL_ALLOW_OWNER => 2, ACL_ALLOW_ADMIN => 1, ACL_ALLOW_NONE => 0,
		ACL_ALLOW_ENABLED => 5, ACL_ALLOW_LIMITED => 2, ACL_ALLOW_DISABLED => 0,
	);
	
	const DEFAULT_ROLE_ID = 'default_-role-0000-0000-000000000001';

	static function acl_cache_file() {
		return CacheManager::get_location('system/') . 'acl_cache.php';
	}
	
	
	function init() {
		parent::init();
		$this->acl = null;
		$this->acl_standard = null;
	}

	function getSetting($name, $default=null, $standard=false) {
		$lastUpd = $this->getACLExpired();
		if(! isset($this->acl) || $this->acl['init_time'] < $lastUpd)
			$this->loadACL($lastUpd);
		if(is_array($name))
			$parts = $name;
		else
			$parts = explode('.', $name);
		$c = count($parts);
		if($c >= 1 && $parts[0] == 'current_user') {
			$cuid = AppConfig::current_user_id();
			if($cuid)
				array_splice($parts, 0, 1, array('by_user', $cuid));
			else
				$parts[0] = 'default_role';
			$c ++;
		}
		if($c >= 1 && $parts[0] == 'default_role') {
			if($standard) {
				array_splice($parts, 0, 1, array('by_role', self::DEFAULT_ROLE_ID));
				$c ++;
			}
			else {
				if(! isset($this->default_role))
					$this->default_role = $this->loadRoleActions(self::DEFAULT_ROLE_ID, false);
				return $this->findSubPath(array_slice($parts, 1), $this->default_role, $default, false);
			}
		}

		if($c >= 2) {
			list($cat, $elt) = $parts;

			if($standard && $cat == 'by_role') {
				return $this->findSubPath(array_slice($parts, 2), $this->acl_standard, $default, false);
			}
			// load user and team ACLs on demand
			if(isset(self::$acl_lookup[$cat]) && ! isset($this->acl[$cat][$elt])) {
				$cache_key = "AppConfig.acl.$cat.$elt";
				$found = false;
				$lookup_result = AppConfig::cache_get_timed($cache_key, $lastUpd, $found);
				if(! $found) {
					$cls = self::$acl_lookup[$cat];
					if($cls[0] == 'this') $cls[0] =& $this;
					$lookup_result = call_user_func($cls, $elt);
					AppConfig::cache_put_timed($cache_key, $lookup_result);
				}
				if($cat == 'by_user' || $cat == 'by_role' || $cat == 'by_team')
					self::post_process_acl($lookup_result);
				$this->acl[$cat][$elt] = $lookup_result;
			}
		}
		if($c == 1 && ($cat = $parts[0]) == 'all_teams') {
			if(! isset($this->acl[$cat]))
				$this->acl[$cat] = self::loadAllTeams();
		}
		
		$ret = $this->findSubPath($parts, $this->acl, $default);
		return $ret;
	}
	
	static function generate_standard_acl() {
		$ret = array('module' => array(), 'access' => array(), 'no_edit' => array());
		$mods = AppConfig::setting("modinfo.index.all");
		$inherit = array();
		
		foreach($mods as $m) {
			$inh = AppConfig::setting("modinfo.by_name.$m.acl.inherit");
			if($inh) $inherit[$m] = $inh;
			else self::add_module_acl($ret, $m);
		}
		foreach($inherit as $m => $parent) {
			self::add_module_acl($ret, $m, $parent);
		}
		
		return $ret;
	}
	
	static function add_module_acl(&$ret, $m, $inherit=null) {
		$actions = array('access' => 'enabled');
		$have_primary = AppConfig::setting("modinfo.by_name.$m.detail.primary_bean", false);
		
		if($inherit && isset($ret['module'][$inherit])) {
			$actions = $ret['module'][$inherit]['actions'];
			$fixed = $ret['module'][$inherit]['fixed'];
		} else {
			$default = AppConfig::setting("modinfo.by_name.$m.acl.defaults");
			if(! is_array($default)) $default = array();
			$fixed = AppConfig::setting("modinfo.by_name.$m.acl.fixed");
			if(! is_array($fixed)) $fixed = array();
			$actions = array_merge($actions, $default, $fixed);

			// change into more standard format
			foreach($actions as $k => $v)
				$actions[$k] = ACLAction::levelStringToInt($v);
		}
		
		$row = array(
			'primary' => $have_primary,
			'actions' => $actions,
			'fixed' => $fixed,
		);
		if(AppConfig::setting("modinfo.by_name.$m.acl.no_disable_access", false))
			$row['no_disable_access'] = true;
		if(! AppConfig::setting("modinfo.by_name.$m.acl.editable", true)) {
			$row['no_edit'] = true;
			$ret['no_edit'][$m] = true;
		}
		$ret['module'][$m] = $row;
	}
	
	static function expand_standard(&$acl) {
		$act_all = ACL_ALLOW_ALL;
		foreach($acl['module'] as $mod => &$row) {
			if(isset($row['primary'])) {
				// previous, expanded cache version did not set 'primary'
				foreach($row['actions'] as $idx => $act) {
					$row['actions'][$idx] = array('id' => '', 'aclaccess' => $act);
				}
				if($row['primary']) {
					foreach(self::$base_model_actions as $a) {
						if(! isset($row['actions'][$a]))
							$row['actions'][$a] = array('id' => '', 'aclaccess' => $a == 'manage_personal' ? ACL_ALLOW_NONE : $act_all);
					}
				}
				$acl['access'][$mod] = $row['actions']['access']['aclaccess'];
			}
		}
	}

	static function load_acl($lastUpdate=0) {
		$cache_key = 'AppConfig.acl';
		$cache_file = self::acl_cache_file();
		$index = AppConfig::cache_file_read($cache_file, $cache_key, $lastUpdate);
		$hash = self::cfg_hash();
		
		if(! is_array($index) || ! isset($index['hash']) || $index['hash'] != $hash) {
			$from = self::config_debug("create acl cache");
			$index = array(
				'hash' => $hash,
				'time' => microtime(true),
				'standard' => self::generate_standard_acl(),
				'system' => array(),
			);
			AppConfig::set_expiry('acl');
			AppConfig::cache_file_write($index, $cache_file, $cache_key);
			self::config_debug("done acl cache", $from);
		}
		
		return $index;
	}
	
	
	static function make_consistent(&$actions) {
		$access = $actions['access']['aclaccess'];
		if($access == ACL_ALLOW_ADMIN || $access == ACL_ALLOW_LIMITED || $access == ACL_ALLOW_DISABLED) {
			foreach($actions as $name => &$act) {
				if($name == 'access' || $act['aclaccess'] == ACL_ALLOW_DISABLED) continue;
				if($access == ACL_ALLOW_LIMITED && ($name == 'view' || $name == 'list'))
					continue;
				$act['aclaccess'] = ACL_ALLOW_ADMIN;
			}
		}
		foreach(self::$sanity_checks as $cmp) {
			list($left, $right) = $cmp;
			if(! isset($actions[$left]))
				continue;
			$l1 = array_get_default(self::$access_levels, $actions[$left]['aclaccess'], '');
			if(! isset($actions[$right]))
				$right = 'access';
			$l2 = array_get_default(self::$access_levels, $actions[$right]['aclaccess'], '');
            $disabled = self::$access_levels[ACL_ALLOW_DISABLED];
            $enabled = self::$access_levels[ACL_ALLOW_DISABLED];
            if($l1 > $l2) {
				$l1 = $actions[$left]['aclaccess'] = $actions[$right]['aclaccess'];
                $disabled = ACL_ALLOW_DISABLED;
                $enabled = ACL_ALLOW_DISABLED;
			}
			if($left == 'approve' && $l1 != $disabled && $l1 != $enabled)
				$l1 = $actions[$left]['aclaccess'] = ACL_ALLOW_ENABLED;
		}
	}
	
	
	function extractRoleActions($role) {
		$ret = array();
		if($role) {
			foreach($role['module'] as $m => $acts) {
				foreach($acts['actions'] as $nm => $act) {
					$act['name'] = $nm;
					$act['category'] = $m;
					$act['acltype'] = 'module';
					$ret[] = $act;
				}
			}
		}
		return $ret;
	}

	
	function loadRoleActions($role_id, $consistent=true) {
		$lq = new ListQuery('ACLAction', array('id', 'name', 'role_id', 'category', 'acltype', 'aclaccess'));
		$lq->addSimpleFilter('role_id', $role_id);
		$lq->addSimpleFilter('acltype', 'module');
		$result = $lq->fetchAll();
		
		$actions = $this->getSetting('default_role', null, true);
		self::merge_actions($actions, $result->rows, null, $consistent);
		return $actions;
	}
	
	
	function loadUserRoles($user_id) {
		if(! $user_id) return array();
		global $db;
		$def_role_id = DEFAULT_ROLE_ID;
		$query = "(SELECT role.name, aru.role_id, '' AS securitygroup_id FROM acl_roles_users aru
			LEFT JOIN acl_roles role ON role.id=aru.role_id
			WHERE aru.user_id='$user_id' AND aru.deleted = 0 AND role.deleted = 0 AND role.id != '$def_role_id')
			UNION (SELECT role.name, sar.role_id, sar.securitygroup_id FROM securitygroups_users su
				INNER JOIN securitygroups_acl_roles sar ON sar.securitygroup_id=su.securitygroup_id AND sar.deleted=0
				LEFT JOIN acl_roles role ON role.id=sar.role_id
				LEFT JOIN securitygroups grp ON su.securitygroup_id=grp.id
				WHERE su.user_id='$user_id' AND grp.deleted = 0 AND su.deleted=0 AND role.deleted = 0 AND role.id != '$def_role_id')";
		$result = $db->query($query);
		$rows = $db->fetchRows($result);
		return $rows;
	}
	
	
	function loadUserActions($user_id, $consistent=true) {
		$hash = self::cfg_hash() . $this->acl_time;
		if($user_id && $user_id == AppConfig::current_user_id()) {
			$use_session = $this->session_cache && $consistent;
			if($use_session && isset($_SESSION['ACL_CACHE'])) {
				$info = $_SESSION['ACL_CACHE'];
				if(isset($info['hash']) && $info['hash'] == $hash) {
					return $info['actions'];
				}
				unset($_SESSION['ACL_CACHE']);
			}
		}
		
		$target = $this->getSetting('default_role', array());
		$roles = $this->getSetting(array('user_roles', $user_id), array());
		
		if($roles) {		
			$role_teams = array();
			foreach($roles as $r) {
				if(! $r['securitygroup_id'] || ! isset($role_teams[$r['role_id']]))
					$role_teams[$r['role_id']] = $r['securitygroup_id'];
			}

			$lq = new ListQuery('ACLAction', array('id', 'name', 'role_id', 'category', 'acltype', 'aclaccess'));
			$lq->addSimpleFilter('role_id', array_keys($role_teams));
			$lq->addSimpleFilter('acltype', 'module');
			$result = $lq->fetchAll();
			$by_role = array();
			foreach($result->rows as $row) {
				$by_role[$row['role_id']][] = $row;
			}

			$additive_roles = AppConfig::setting('aclconfig.additive');
			$user_role_precedence = AppConfig::setting('aclconfig.user_role_precedence');
		
			/*if($user_role_precedence) {
				$direct = $via_team = array();
				foreach($result->rows as $idx => $row) {
					if($role_teams[$row['role_id']]) // inherited via team
						$via_team[$idx] = $row;
					else // directly assigned role
						$direct[$idx] = $row;
				}
				self::merge_actions($target, $direct, $additive_roles);
				self::merge_actions($target, $via_team, $additive_roles);
			}
			else {*/
				$ridx = 0;
				foreach($by_role as $role)
					self::merge_actions($target, $role, $additive_roles, $consistent && (++$ridx) == count($by_role));
			// }
		} else {
			if($consistent)
				foreach($target['module'] as $mod => $_)
					self::make_consistent($target['module'][$mod]['actions']);
		}

		if(! empty($use_session)) {
			$_SESSION['ACL_CACHE'] = array(
				'hash' => $hash,
				'actions' => $target,
			);
		}

		return $target;
	}
	
	
	function loadTeamActions($team_id, $consistent=true) {
		global $db;
		
		$def_role_id = DEFAULT_ROLE_ID;
		$query = "SELECT id, name, role_id, category, acltype, aclaccess
				FROM acl_actions act WHERE role_id IN
					(SELECT sar.role_id FROM securitygroups_acl_roles sar
					LEFT JOIN acl_roles role ON role.id=sar.role_id
					WHERE sar.securitygroup_id='$team_id' AND sar.deleted=0 AND role.deleted = 0 AND role.id != '$def_role_id')
				AND act.acltype='module' AND act.deleted=0";
		$result = $db->query($query);
		$rows = $db->fetchRows($result);

		$additive_roles = AppConfig::setting('aclconfig.additive');
		$target = $this->getSetting('default_role', array());
		$by_role = array();
		foreach($rows as $row)
			$by_role[$row['role_id']][] = $row;

		$ridx = 0;
		foreach($by_role as $role)
			self::merge_actions($target, $role, $additive_roles, $consistent && (++$ridx) == count($by_role));

		return $target;
	}
	
	
	static function merge_actions(&$target, $actions, $additive, $consistent=false, $updated_only=false) {
		$upd = array();
		
		$newacc = array();
		foreach($actions as $row) {
			if($row['acltype'] == 'module' && $row['name'] == 'access') {
				$newacc[$row['category']] = $row['aclaccess'];
			}
		}
		
		foreach($actions as $row) {
			$id = $row['id'];
			$acltype = $row['acltype'];
			$category = $row['category'];
			$action = $row['name'];

			if(! isset($target[$acltype][$category]['actions'][$action]))
				continue;
			if(! empty($target[$acltype][$category]['no_edit']))
				continue;
			$aclaccess = $row['aclaccess'];
			$oldaccess = $target[$acltype][$category]['actions'][$action]['aclaccess'];
			$acllvl = self::$access_levels[$aclaccess];
			$oldlvl = self::$access_levels[$oldaccess];
			$access_override = 0;
			if($additive && $action != 'access' && isset($target['access'][$category]) && isset($newacc[$category])) {
				$oldacc = $target['access'][$category];
				$oldacclvl = self::$access_levels[$oldacc];
				$newacclvl = self::$access_levels[$newacc[$category]];
				if($oldacc == ACL_ALLOW_DISABLED || $oldacc == ACL_ALLOW_ADMIN || ($oldacc == ACL_ALLOW_LIMITED && $action != 'view' && $action != 'list'))
					if($oldacclvl < $newacclvl)
						$access_override = 1;
			}
			if(
				! isset($additive)
				||
				($additive && ($acllvl > $oldlvl || $access_override))
				||
				(! $additive && $acllvl < $oldlvl)
			){
				$target[$acltype][$category]['actions'][$action] = compact('id', 'aclaccess');
				if($acltype == 'module') $upd[$category] = 1;
			}
		}
		
		if(isset($target['module'])) {
			foreach($target['module'] as $mod => $row)
				if(isset($row['actions']['access']['aclaccess']))
					$target['access'][$mod] = $row['actions']['access']['aclaccess'];
		}

		if($consistent) {
			if($updated_only) {
				foreach($upd as $mod => $_)
					self::make_consistent($target['module'][$mod]['actions']);
			} else {
				foreach($target['module'] as $mod => $_)
					self::make_consistent($target['module'][$mod]['actions']);				
			}
		}
	}
	
	
	// apply extra restrictions for read-only mode
	static function post_process_acl(&$target) {
		if(AppConfig::read_only()) {
			$restrict = array('edit', 'delete', 'import', 'report', 'approve');
			foreach($target['module'] as $mod => $modacl) {
				foreach($restrict as $a) {
					if(isset($modacl['actions'][$a])) {
						$target['module'][$mod]['fixed'][$a] = 'none';
						$target['module'][$mod]['actions'][$a]['aclaccess'] = ACL_ALLOW_SYSTEM;
					}
				}
			}
		}
	}
	
	
	function loadUserTeams($user_id) {
		global $db;
		$query = "SELECT sg.id, sg.name, sg.noninheritable FROM securitygroups_users sgu "
				."INNER JOIN securitygroups sg ON sgu.securitygroup_id = sg.id "
				."WHERE sgu.user_id='$user_id' AND sgu.deleted = 0 AND sg.deleted = 0 ORDER BY sg.name";
		$group_array = array();
		$result = $db->query($query);
		while( ($row = $db->fetchByAssoc($result, -1, false)) != null) {
			$group_array[$row['id']] = $row;
		}
		return $group_array;
	}
	
	function loadAllTeams() {
		global $db;
		$query = "SELECT sg.id, sg.name, sg.noninheritable FROM securitygroups sg "
				."WHERE NOT sg.deleted ORDER BY sg.name";
		$group_array = array();
		$result = $db->query($query);
		while( ($row = $db->fetchByAssoc($result, -1, false)) != null) {
			$group_array[$row['id']] = $row;
		}
		return $group_array;
	}
	
	function getACLExpired() {
		return AppConfig::get_expiry('acl');
	}
	
	function loadACL($lastUpdate=0) {
		$index = self::load_acl($lastUpdate);
		$this->acl_standard = $index['standard'];
		self::expand_standard($this->acl_standard);
		$this->acl = array_get_default($index, 'system', array());
		$this->acl_time = array_get_default($index, 'time');
		foreach(self::$acl_lookup as $k => $callback) {
			if (!isset($this->acl[$k])) {
				$this->acl[$k] = array();
			}
		}
		$this->acl['init_time'] = microtime(true);
	}
	
	function invalidateCache($category=null, $limit=null) {
		// setting stored in external cache only
		AppConfig::set_expiry('acl');
		@unlink(self::acl_cache_file());
		$this->init();
		$GLOBALS['log']->debug('ACL cache invalidated');
	}
}

?>
