<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo,
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */



require_once('include/config/format/ConfigParser.php');

abstract class ConfigBase {
	var $qcache;
	var $qcache_enabled = false;
	var $stat_disabled;

	static $debug_level = false;

	function __construct() {
		$this->init();
	}

	function init() {
		$this->qcache = array();
		$this->stat_disabled = false; // disable when debugged
	}

	abstract function getSetting($name, $default=null, $standard=false);
	abstract function invalidateCache($category=null);
	function setLocal($path, $val, $merge=false, $push_db=true) {}
	function saveLocal($category=null, $push_db=true, $out_path=null) {}

	static function check_paths($fs, $append=null) {
		$mt = false;
		$found = 0;
		if(! is_array($fs)) {
			$fs = func_get_args();
			$append = null;
		}
		foreach($fs as $f) {
			if(isset($append)) $f .= $append;
			clearstatcache(false, $f);
			$ft = @filemtime($f);
			if($ft !== false) {
				$mt = max($mt, $ft);
				$found ++;
			}
		}
		return array('count' => $found, 'max_mtime' => $mt);
	}

	static function &load_files($paths, $fname=null) {
		if(! is_array($paths)) {
			$paths = func_get_args();
			$fname = null;
		}
		if(isset($fname))
			foreach($paths as $k => &$v)
				$v .= $fname;
		$cfg = ConfigParser::load_files($paths);
		return $cfg;
	}

	static function expand_paths($paths) {
		$ret = array();
		if(! is_array($paths)) {
			$paths = func_get_args();
		}
		foreach($paths as $p) {
            if(strpos($p, '*') !== false) {
                if(substr($p, -1) == '/') $p = substr($p, 0, -1);
				$ps = glob($p, GLOB_MARK | GLOB_ONLYDIR | GLOB_NOSORT);
                $ps = str_replace('\\','/', $ps);
				if($ps) array_extend($ret, $ps);
			} else if($p) {
				if(substr($p, -1) != '/') $p .= '/';
                $p = str_replace('\\','/', $p);
				$ret[] = $p;
			}
		}
		return $ret;
	}

	static function &cache_load_files($cache_key, $paths, $fname=null, $mtime=true) {
		// mtime may be a number (cache expiry time) or true to perform a stat on the files
		if($mtime === true) {
			$tag = self::check_paths($paths, $fname);
			$mtime = $tag['max_mtime'];
		}
		$data = AppConfig::cache_get_timed($cache_key, $mtime, $found);
		if(! $found) {
			$data = self::load_files($paths, $fname);
			AppConfig::cache_put_timed($cache_key, $data);
		}
		return $data;
	}

	static function merge_values($vals, &$target, $check_upd=false, $prefix='', $recurse = true) {
		$found = array();
		foreach($vals as $k => &$v) {
			if(is_array($v) && isset($target[$k]) && is_array($target[$k]) && $recurse)
				self::merge_values($v, $target[$k], $check_upd, $prefix.$k.'.');
			else {
				if(! $check_upd || array_get_default($target, $k) != $v)
					$found[$prefix.$k] = $v;
				$target[$k] = $v;
			}
		}
		return $found;
	}

	static function cfg_hash() {
		return md5(AppConfig::unique_key() . AppConfig::version() . AppConfig::dev_branch() . AppConfig::site_url());
	}


	function clearQCache($prefix=null) {
		if(! $this->qcache_enabled) return;
		if(! isset($prefix) || $prefix === '') {
			$this->qcache = array();
			return;
		}
		$l = strlen($prefix);
		$ks = array_keys($this->qcache);
		foreach($ks as $k)
			if(substr($k, 0, $l) === $prefix)
				unset($this->qcache[$k]);
	}

	/* this function takes an array of setting names (in various forms)
	   and returns a nested hierarchy of the values
	*/
	function extractSettings($keys, $standard=false, $prefix='') {
		$ret = array();
		foreach($keys as $cat => $val) {
			if(is_array($val)) {
				$pfx = is_int($cat) ? $prefix : $prefix.$cat.'.';
				$subvals = $this->extractSettings($val, $standard, $pfx);
				self::merge_values($subvals, $ret);
			}
			else {
				if(! is_int($cat))
					$val = "$cat.$val";
				$val = $prefix.$val;
				$parts = explode('.', $val);
				unset($targ);
				$targ =& $ret;
				foreach($parts as $k) {
					if(! isset($targ[$k]))
						$targ[$k] = array();
					$targ =& $targ[$k];
				}
				$targ = $this->getSetting($val, null, $standard);
			}
		}
		return $ret;
	}


	function findSubPath($path, &$tree, $default=null, $use_qcache=true) {
		if(is_array($path)) {
			$parts = $path;
			$path = implode('.', $path);
		} else
			$parts = explode('.', $path);
		$src =& $tree;
		unset($tree);
		$i = 0;
		foreach($parts as $p) {
			if(is_array($src) && array_key_exists($p, $src))
				$src =& $src[$p];
			else {
				if($this->qcache_enabled && $use_qcache && $i > 0)
					$this->qcache[$path] = null;
				return $default;
			}
			$i ++;
		}
		if($this->qcache_enabled && $use_qcache && $i > 0 && ! is_array($src))
			$this->qcache[$path] = $src;
		return $src;
	}

	function memoryCacheDisabled() {
		// disable config memory cache if the config files are precompiled
		return ! AppConfig::config_cache_enabled();
	}

	static function config_debug($msg, $from=null) {
		$now = microtime(true);
		if(isset($from)) $msg .= ' ('.round($now - $from, 2).')';
		if(self::$debug_level && isset($GLOBALS['log'])) $GLOBALS['log']->{self::$debug_level}($msg);
		return $now;
	}

}

?>
