<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo, 
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */



require_once('include/config/ConfigBase.php');

class ConfigManager extends ConfigBase {
	var $standard = [];
	var $local = [];
	var $parser;
	var $cached = false;
	var $aclmgr;
	var $modelmgr;
	var $modinfomgr;
	var $prefmgr;
	var $pending_load;
	var $db_pending_load;
	var $db_pending_save;
	var $cache_expiry;
	var $init_time;
	var $total_request_time = 0;
	var $log_all_requests = false;
	var $in_request = false;
	var $log_fp;
	var $have_local_config;
	var $lang;
	var $displaymgr;
	var $viewsmgr;
	var $extmgr;

	static $local_config_dir = 'include/config/';
	static $local_config_name = 'local_config.php';
	static $standard_config_dirs = [
		'include/config/standard/',
		'modules/*/ext/config/standard/',
		'custom/include/config/standard/',
	];
	static $special_config_dirs = [
		'include/config/special/',
		'modules/*/ext/config/special/',
		'custom/include/config/special/',
	];
	static $db_table = 'config';
	static $db_categories = [
		'company' => 1,
		'info' => 1,
		'ldap' => 1,
		'license' => 1,
		'portal' => 1,
		'scheduler' => 1,
		'misc' => 1,
	];
	static $local_only_categories = [
		'cache' => 1,
		'config' => 1,
	];
	static $vt_categories = [
		'acl' => [
			'var' => 'aclmgr',
			'cls' => 'ACLManager',
			'inc' => 'include/config/ACLManager.php',
		],
		'model' => [
			'var' => 'modelmgr',
			'cls' => 'ModelManager',
			'inc' => 'include/config/ModelManager.php',
		],
		'modinfo' => [
			'var' => 'modinfomgr',
			'cls' => 'ModInfoManager',
			'inc' => 'include/config/ModInfoManager.php',
		],
		'userpref' => [
			'var' => 'prefmgr',
			'cls' => 'UserPrefManager',
			'inc' => 'include/config/UserPrefManager.php',
		],
		'lang' => [
			'var' => 'lang',
			'cls' => 'LanguageManager',
			'inc' => 'include/config/LanguageManager.php',
		],
		'display' => [
            'var' => 'displaymgr',
            'cls' => 'DisplayManager',
            'inc' => 'include/config/DisplayManager.php',
        ],
        'views' => [
            'var' => 'viewsmgr',
            'cls' => 'LayoutConfigManager',
            'inc' => 'include/config/LayoutConfigManager.php',
        ],
		'ext' => [
			'var' => 'extmgr',
			'cls' => 'ExtManager',
			'inc' => 'include/config/ExtManager.php',
		],
	];

	static function local_config_path() {
		// allow env variable override?
		return self::$local_config_dir . self::$local_config_name;
	}


	function init() {
		parent::init();
		$this->standard = [];
		$this->local = [];
		$this->pending_load = [];
		$this->db_pending_load = [];
		$this->db_pending_save = [];
		$this->qcache = [];
		$this->cache_expiry = [];
		
		$this->parser = new ConfigParser();
		$this->parser->setTargetArray($this->local);
		$this->parser->getTarget()->setPreloadCallback([$this, 'getStandardCategory']);
		
		foreach(self::$vt_categories as $k => &$v)
			$this->{$v['var']} = null;
	}

	function __destruct() {
		if($this->log_fp) {
			@fwrite($this->log_fp, sprintf('%5.5f  ', $this->total_request_time) . "total time\n");
			@fclose($this->log_fp);
		}
	}

	function localConfigFound() {
		if(! isset($this->have_local_config))
			$this->loadConfig();
		return $this->have_local_config;
	}
	
	function getStandardCategory(ConfigTarget $t, $name) {
		$this->loadStandard($name);
		return $this->standard[$name];
	}
	
	function loadStandard($name, $reload=false, $no_cache=false) {
		if(! $reload && array_key_exists($name, $this->standard))
			return;
		if(empty(self::$local_only_categories[$name]) && preg_match('/^\w+$/', $name)) {
			$paths = self::expand_paths(self::$standard_config_dirs);
			// skip memory cache if configs are precompiled
			if(AppConfig::config_cache_enabled()) $no_cache = true;
			if($no_cache || $reload)
				$vals = self::load_files($paths, "$name.php");
			else {
				$cache_key = "AppConfig.$name";
				$mtime = $this->stat_disabled ? 0 : true;
				$vals = self::cache_load_files($cache_key, $paths, "$name.php", $mtime);
			}
		}
		else {
			$vals = null;
		}
		if(! is_array($vals)) $vals = [];
		$this->standard[$name] = $vals;
	}
	
	function loadCategory($categ, $standard=false) {
		$this->loadStandard($categ);
		if(! $standard) {
			if(! empty($this->pending_load[$categ])) {
				$this->pending_load[$categ] = 0;
				$this->parser->getTarget()->getResultKey($categ);
			} else if(! array_key_exists($categ, $this->local) && isset($this->standard[$categ])) {
				$this->local[$categ] = $this->standard[$categ];
			}
		}
		if(! $standard && ! empty($this->db_pending_load[$categ]))
			$this->loadDBCategory($categ);
	}

	function loadSpecial($name, $no_cache=false) {
		$ret = array();
		$paths = self::expand_paths(self::$special_config_dirs);
		foreach($paths as $p) {
			$p .= "$name.php";
			if(is_file($p) && is_readable($p)) {
				$this->parseLocal($p);
				$ret[] = $p;
			}
		}
		return $ret;
	}

	function parseLocal($path) {
		$stat = $this->parser->parseFile($path, ['no_save_cache' => true]);
		$this->cached = $stat['cached'];
		$this->pending_load = [];
		$pend = $this->parser->getTarget()->getResultKeys();
		if($pend) foreach($pend as $p) {
			if(! array_key_exists($p, $this->local))
				$this->pending_load[$p] = 1;
		}
	}

	function loadConfig($reload_local=false) {
		$this->init();

		/* config load procedure:
			- read local_config.php
			- as categories are encountered during parsing,
			  check for standard definitions and load those (by category)
			- load database settings
			- load special override settings (demo site or other)
		*/

		$this->local = $this->standard;
		$path = self::local_config_path();
		if(! is_file($path) || ! is_readable($path)) {
			$this->have_local_config = false;
		} else {
			$this->parseLocal($path);
			$this->have_local_config = true;
		}

		$this->db_pending_load = self::$db_categories;

		if(is_demo_site())
			$this->loadSpecial('demo_settings');

		if($this->getSetting('config.log_all_requests')) {
			$this->log_all_requests = true;
			$this->log_fp = @fopen('logs/config.log', 'w');
			if($this->log_fp) {
				$path = array_get_default($_SERVER, 'REQUEST_URI');
				@fwrite($this->log_fp, date('Y-m-d H:i:s  ') . $path . ' ' . array_get_default($_REQUEST, 'module') . '  ' . array_get_default($_REQUEST, 'action') . '  ' . array_get_default($_REQUEST, 'record_perform') . "\n");
			}
		}
	}


	// database setting storage

	function loadDBCategories($names=null) {
		global $db;
		if(! isset($db))
			return false;
		if(! isset($names)) {
			$names = self::get_db_categories();
			$use_cache = true;
		}
		if(! count($names))
			return;
		foreach($names as $n) {
			$this->loadStandard($n);
			if(! isset($this->local[$n]) || ! is_array($this->local[$n]))
				$this->local[$n] = $this->standard[$n];
			$this->db_pending_load[$n] = 0;
		}
		$cache_key = 'AppConfig.db_config';
		if($use_cache) {
			$ts = microtime(true) - 5; // cache 5 seconds only
			$rows = AppConfig::cache_get_timed($cache_key, $ts, $found);
		}
		if(! isset($rows)) {
			$names = array_map(array(&$db, 'quote'), $names);
			$sql = "SELECT * FROM ".(self::$db_table)." WHERE category IN ('".implode("','", $names)."')";
			$rows = $db->fetchSQL($sql);
			if($use_cache && isset($rows))
				AppConfig::cache_put_timed($cache_key, $rows, true, true);
		}
		foreach($rows as $row)
			$this->local[$row['category']][$row['name']] = $row['value'];
	}

	static function get_db_categories() {
		return array_keys(array_filter(self::$db_categories));
	}
	
	static function get_all_categories() {
		$cats = array();
		foreach(self::$db_categories as $c => $v) {
			if($v) $cats[$c] = array('type' => 'db');
		}
		foreach(self::$vt_categories as $c => $v) {
			$cats[$c] = array('type' => 'virtual');
		}
		$paths = self::expand_paths(self::$standard_config_dirs);
		foreach($paths as $p) {
			$g = glob($p . '*.php', GLOB_NOSORT);
			if($g) foreach($g as $p2) {
				$name = substr($p2, strrpos($p2, '/')+1, -4);
				if($name) $cats[$name] = array('type' => 'local');
			}
		}
		return $cats;
	}

	function loadDBConfig() {
		$this->loadDBCategories();

		// must reload to override DB settings
		if(is_demo_site())
			$this->loadSpecial('demo_settings');
	}

	function saveDBConfig() {
		global $db;
		$updates = array();
		foreach($this->db_pending_save as $cat => $rows) {
			foreach($rows as $name => $_n) {
				$val = $this->getSetting("$cat.$name");
				if(isset($val))
					$val = "'".$db->quote($val)."'";
				else
					$val = 'null';
				$updates[] = sprintf("('%s', '%s', %s)",
					$db->quote($cat), $db->quote($name), $val);
			}
		}
		if($updates) {
			if(empty($db)) {
				$cats = implode(', ', array_keys($this->db_pending_save));
				$GLOBALS['log']->error("Cannot save local config, database not connected. Categories: $cats");
				return false;
			}
			$sql = "INSERT INTO ".(self::$db_table)." (category,name,value) VALUES "
				. implode(',', $updates) . ' ON DUPLICATE KEY UPDATE value=VALUES(value)';
			$db->query($sql);
			AppConfig::cache_clear('AppConfig.db_config');
		}
		$this->db_pending_save = array();
		return true;
	}

	function getSequencePrefix($prefix_name, $category=null, $format=true, $timestamp = null) {
		if (is_null($timestamp)) {
			$timestamp = time();
		}
		if(! isset($category))
			$category = 'company';
		$prefix = $this->getSetting("$category.$prefix_name", '');
		if($format) {
			$prefix = str_replace(array(
				'%Y', '%y', '%M', '%m', '%W', '%o', '%d',
			), explode(' ', date('Y y M m W o d', $timestamp)), $prefix);
		}
		return $prefix;
	}

	function currentSequenceValue($sequence_name, $category=null) {
		if(! isset($category))
			$category = 'company';
		return $this->getSetting("$category.$sequence_name");
	}

	function nextSequenceValue($sequence_name, $category=null) {
		if(! isset($category))
			$category = 'company';
		global $db;
		$query = "INSERT INTO %s (category,name,value) VALUES "
			. "('%s','%s',(@seqval := 1)+1) "
			. 'ON DUPLICATE KEY UPDATE value=(@seqval := IF(value > 0, value, 1))+1';
		$query = sprintf($query, self::$db_table, $db->quote($category), $db->quote($sequence_name));
		$db->query($query, true, "Error updating sequence value ($sequence_name)");
		return $db->getOne("select @seqval", true, 'Error retrieving sequence value');
	}

	function loadDBCategory($name) {
		// just load all categories
		$this->loadDBConfig();
	}


	// virtual and special categories

	function &getVirtualHandler($categ) {
		if(! isset(self::$vt_categories[$categ])) {
			$n = null;
			return $n;
		}
		$def = self::$vt_categories[$categ];
		if(! isset($this->{$def['var']})) {
			if(! empty($def['inc']))
				require_once($def['inc']);
			$this->{$def['var']} = new $def['cls'];
		}
		return $this->{$def['var']};
	}

	function getVirtualPath($path, $default=null, $standard=false) {
		if(! $path) return;
		$parts = explode('.', $path);
		$categ = array_shift($parts);
		$handler =& $this->getVirtualHandler($categ);
		if(! $handler)
			throw new IAHError("Invalid virtual config category: {$categ}");
		return $handler->getSetting($parts, $default, $standard);
	}

	function setVirtualPath($path, $val, $merge=false, $push_db=true) {
		if(is_array($path))
			$parts = $path;
		else
			$parts = explode('.', $path);
		$categ = array_shift($parts);
		$handler =& $this->getVirtualHandler($categ);
		if(! $handler)
			throw new IAHError("Invalid virtual config category: {$categ}");
		return $handler->setLocal($parts, $val, $merge, $push_db);
	}

	function saveVirtual($path, $push_db=true) {
		if(! is_array($path))
			$path = explode('.', $path);
		if(! isset($path[0]))
			return;
		$categ = array_shift($path);
		$handler =& $this->getVirtualHandler($categ);
		if(! $handler)
			throw new IAHError("Invalid virtual config category: {$categ}");
		return $handler->saveLocal($path, $push_db);
	}

	// accessors and updates

	function getSetting($name, $default=null, $standard=false) {
		if($this->log_all_requests && ! $this->in_request) {
			$this->in_request = true;
			$st = microtime(true);
		}

		if(is_array($name)) {
			foreach($name as $test_name) {
				$val = $this->getSetting($test_name, null, $standard);
				if($val !== null)
					return $val;
			}
			return $default;
		}

		if(! $standard && $this->qcache_enabled && array_key_exists($name, $this->qcache)) {
			$v = $this->qcache[$name];
			return isset($v) ? $v : $default;
		}

		$p = strpos($name, '.');
		$categ = ($p === false) ? $name : substr($name, 0, $p);
		if(! empty(self::$vt_categories[$categ]))
			$ret = $this->getVirtualPath($name, $default, $standard);
		else {
			$this->loadCategory($categ, $standard);
			if($standard)
				$ret = $this->findSubPath($name, $this->standard, $default, false);
			else
				$ret = $this->findSubPath($name, $this->local, $default);
		}

		if($this->log_all_requests && isset($st)) {
			$time = microtime(true) - $st;
			$this->total_request_time += $time;
			if($this->log_fp)
				@fwrite($this->log_fp, sprintf('%5.5f  ', $time) . $name . "\n");
			$this->in_request = false;
		}

		return $ret;
	}

	function getStandard($name, $default=null) {
		return $this->getSetting($name, $default, true);
	}


	function invalidateCache($category=null, $limit=null) {
		if($category) {
			$handler =& $this->getVirtualHandler($category);
			if($handler)
				$handler->invalidateCache(null, $limit);
			$this->clearQCache("$category.");
		}
		$this->cache_expiry = array();
	}


	/* note: give fully-qualified top level attributes to
	   replace specific subtrees when merge=false
	   ie.   updateLocal(array('site.login.disabled' => 1))
	   OTOH, updateLocal(array('site' => array('login' => array('disabled' => 1))))
	   would replace the entire 'site' subtree
	*/
	function updateLocal($settings, $merge=false, $push_db=true) {
		if(! is_array($settings)) return;
		foreach($settings as $k => $v)
			$this->setLocal($k, $v, $merge, false);
		if($push_db)
			$this->saveDBConfig();
		return true;
	}

	function setLocal($key, $val, $merge=false, $push_db=true) {
		$parts = explode('.', $key);
		$cat = $parts[0];
		if(! empty(self::$vt_categories[$cat])) {
			return $this->setVirtualPath($parts, $val, $merge, $push_db);
		}
		$this->loadCategory($cat);
		$last = array_pop($parts);

		$local =& $this->local;
		foreach($parts as $p) {
			if(! isset($local[$p]))
				$local[$p] = array();
			$local =& $local[$p];
		}

		$is_db_cat = ! empty(self::$db_categories[$cat]);

		if( ( ($merge && isset($local[$last])) || $is_db_cat) && is_array($val) ) {
			$changes = self::merge_values($val, $local[$last], $is_db_cat);
			$this->clearQCache($key.'.');
			$differ = !! count($changes);
		} else {
			$differ = (array_get_default($local, $last) !== $val);
			$local[$last] = $val;
			if($this->qcache_enabled && count($parts))
				$this->qcache[$key] = $local[$last];
			$this->clearQCache($key.'.');
		}
		if($differ && $is_db_cat) {
			if(! is_array($val))
				$changes = array($last => null);
			$pfx = implode('.', $parts);
			if($pfx) $pfx .= '.';
			foreach(array_keys($changes) as $node)
				$this->db_pending_save[$cat][$node] = 1;
			if($push_db)
				$this->saveDBConfig();
		}
		return true;
	}

	function getWriter() {
		require_once('include/config/format/ConfigWriter.php');
		$writer = new ConfigWriter;
		$writer->setAppendOnly();
		return $writer;
	}
	
	function completeLoad() {
		// ensure any pending categories are loaded
		$this->parser->getTarget()->loadResultKeys();
	}

	function encodeLocal($mode='string', $out_path=null) {
		$this->completeLoad();

		// load local config, keeping previous key order if possible
		if($this->parser && $this->parser->getTarget())
			$keys = $this->parser->getTarget()->getResultKeys();
		if(empty($keys)) $keys = [];
		foreach(array_keys($this->local) as $k)
			if(! in_array($k, $keys)) $keys[] = $k;
		$loc = $std = [];
		foreach($keys as $k) {
			// remove db and virtual categories - not generally saved to local config
			if(! empty(self::$db_categories[$k]) || ! empty(self::$vt_categories[$k]))
				continue;
			if(isset($this->local[$k])) {
				$loc[$k] = $this->local[$k];
				$std[$k] = array_get_default($this->standard, $k, []);
			}
		}

		$writer = $this->getWriter();
		if($mode == 'file')
			return $writer->writeDiff($out_path, $loc, $std, null);
		$string = ($mode != 'array');
		return $writer->encode($loc, $std, null, $string);
	}

	function saveLocal($category=null, $push_db=true, $out_path=null) {
		if(AppConfig::read_only()) return false;

		if($push_db)
			$this->saveDBConfig();
		if(isset($category)) {
			$parts = explode('.', $category);
			$category = $parts[0];
			if(! empty(self::$vt_categories[$category]))
				return $this->saveVirtual($parts, $push_db);
		}
		
		if(! isset($out_path)) {
			$out_path = self::local_config_path();
			$cache = true;
		}
		if(! $out_path) return;
		if(file_exists($out_path)) {
			if( ($p = strrpos($out_path, '/')) !== false)
				$bkp_path = substr($out_path, 0, $p+1).'_bkp_'.substr($out_path, $p+1);
			else
				$bkp_path = '_bkp_'.$p;
			copy($out_path, $bkp_path);
		}
		$ret = $this->encodeLocal('file', $out_path);
		
		if($ret && ! empty($cache) && defined('CACHE_LOCAL_CONFIG') && constant('CACHE_LOCAL_CONFIG'))
			$this->writeConfigCache();
		
		return $ret;
	}
	
	function writeConfigCache() {
		$mgr = $this->parser->getCacheManager();
		if(! $mgr || ! $mgr->getCacheEnabled())
			return false;

		$path = self::local_config_path();
		$p = new ConfigParser();
		$p->setCacheManager($mgr);
		try {
			$p->parseFile($path, ['no_load_cache' => 1, 'on_demand' => 1]);
		}  catch(IAHConfigFileError $e) {
			$GLOBALS['log']->error('Error caching local config: '.$e->getMessage());
			$mgr->clearCache($path);
			return false;
		}
		return true;
	}
	
	function clearExpiry($category) {
		$this->setExpiry($category, null);
	}
	
	function setExpiry($category, $value=null) {
		if(! $category) return;
		$this->cache_expiry[$category] = $value;
		if(isset($value))
			AppConfig::cache_put("AppConfig.$category.expired", $value, true, true);
		else
			AppConfig::cache_clear("AppConfig.$category.expired");
	}
	
	function getExpiry($category) {
		if(! $category) return 0;
		if(isset($this->cache_expiry[$category]))
			return $this->cache_expiry[$category];
		$val = AppConfig::cache_get("AppConfig.$category.expired", $found);
		if(! $found) $val = 0;
		$this->cache_expiry[$category] = $val;
		return $val;
	}
}

?>
