<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo,
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */



// allow standalone usage (outside IAH)
if(! class_exists('IAHInternalError')) {
	class IAHInternalError extends Exception {}
}
if(! function_exists('array_get_default')) {
	function array_get_default($arr, $key, $default=null) {
		return isset($arr[$key]) ? $arr[$key] : $default;
	}
}
if(! function_exists('array_extend_recursive')) {
	function &array_extend_recursive(array &$gimp, array $dom /*, array ... */) {
		foreach($dom as $k => &$v) {
			if(is_int($k))
				$gimp[] = $v;
			else if(isset($gimp[$k]) && is_array($gimp[$k]) && is_array($v))
				array_extend_recursive($gimp[$k], $v);
			else
				$gimp[$k] = $v;
		}
		$na = func_num_args();
		for($i = 2; $i < $na; $i++)
			array_extend_recursive($gimp, func_get_arg($i));
		return $gimp;
	}
}

class IAHConfigFileError extends IAHInternalError {
	var $no_translate = true; // translation could cause infinite recursion
}

class ConfigEnum {
	var $val;
	function __construct($val) {
		$this->val = $val;
	}
	function __toString() {
		return $this->val;
	}
}

class ConfigParser {

	const DEFAULT_TAB_SIZE = 4;

	const MODE_STRING = 1;
	const MODE_COMMENT = 2;
	const MODE_SIMPLE_VALUE = 3;
	const MODE_ARRAY = 4;
	const MODE_INFO = 6;
	const MODE_PI = 7;

	static $DEBUG = 0;
	static $PREG_OPTIMIZE = 1;

	static $modes = [
		self::MODE_STRING => 'MODE_STRING',
		self::MODE_COMMENT => 'MODE_COMMENT',
		self::MODE_SIMPLE_VALUE => 'MODE_SIMPLE_VALUE',
		self::MODE_ARRAY => 'MODE_ARRAY',
		self::MODE_INFO => 'MODE_INFO',
		self::MODE_PI => 'MODE_PI',
	];

	static $special_values = [
		'true' => true,
		'false' => false,
		'null' => null,
	];

	static $ARRAY_APPEND_SIMPLE;
	static $ARRAY_APPEND_ARRAY;
	static $START_INFO;
	static $all_parsed = [];
	static $time_parsing = false;

	var $results;
	var $target;
	var $warnings = [];
	var $context;
	var $tabsize;
	var $mode, $mode_start;
	var $linepos, $charpos, $linelen;
	var $array_key, $have_key;
	var $info;
	var $depth, $indent;
	var $inline_value;
	var $quoted_value;
	var $ctx_handler, $ctx_handler_max_depth;
	var $cache, $last_cached;
	var $filename;
	var $lastResult;
	var $prev_target;

	static function check_init() {
		if(! isset(self::$ARRAY_APPEND_SIMPLE)) {
			self::$ARRAY_APPEND_SIMPLE = new ConfigEnum('*array_append_simple*');
			self::$ARRAY_APPEND_ARRAY = new ConfigEnum('*array_append_array*');
			self::$START_INFO = new ConfigEnum('*start_info*');
		}
	}

	static function get_all_parsed() {
		return self::$all_parsed;
	}

	static function load_file($file, $options=null, $key=null) {
		$cfg = new self($options);
		$cfg->parseFile($file, $options);
		$ret = $cfg->getResult();
		if(isset($key))
			$ret = isset($ret[$key]) ? $ret[$key] : null;
		return $ret;
	}

	static function load_file_with_target($file, &$targ, $options=null) {
		$cfg = new self($options);
		$cfg->parseFileWithTarget($file, $targ, $options);
		return true;
	}

	static function load_files($files, $options=null) {
		$cfg = new self($options);
		foreach($files as $f) {
			if(file_exists($f)) $to_load = [$f];
			elseif ((strpos($f, '*') !== false) || (strpos($f, '?') !== false)) $to_load = glob($f, GLOB_NOSORT);
			else $to_load = [];
			if($to_load)
				foreach ($to_load as $fname)
					$cfg->parseFile($fname, $options);
		}
		return $cfg->getResult();
	}

	static function load_files_with_target($files, &$targ, $options=null) {
		$cfg = new self($options);
		foreach($files as $f)
			if(file_exists($f)) {
				$cfg->setTargetArray($targ);
				$cfg->parseFile($f, $options);
			}
		return true;
	}

	static function load_string($data, $options=null, $key=null) {
		$cfg = new self($options);
		$cfg->parseStringWithTarget($data, $cfg->target, $options);
		$ret = $cfg->getResult();
		if(isset($key))
			$ret = isset($ret[$key]) ? $ret[$key] : null;
		return $ret;
	}

	///

	function __construct($params=null) {
		self::check_init();
		$this->target = new ConfigArrayTarget();
		if(! is_array($params)) $params = [];
		if (isset($params['extend']))
			$this->target->extend = $params['extend'];
		if(empty($params['no_cache']) && (! defined('CONFIG_CACHE_DISABLED') || ! constant('CONFIG_CACHE_DISABLED'))) {
			$cache_path = defined('CONFIG_CACHE_PATH') ? constant('CONFIG_CACHE_PATH') : 'cache/config';
 			$this->setCacheDirectory($cache_path);
		}
	}

	function getTarget() {
		return $this->target;
	}
	function setTarget(ConfigTarget $targ) {
		$this->target = $targ;
	}

	function setTargetArray(&$targ) {
		if(! is_array($targ))
			$targ = [];
		if(! ($this->target instanceof ConfigArrayTarget))
			$this->target = new ConfigArrayTarget();
		$this->target->setCustomArray($targ);
	}

	// set a function which will generate source data for us to update
	// this is used to load 'standard' config files on-demand, when referenced
	function setContextHandler($ctx_handler, $max_depth=-1) {
		$this->ctx_handler = $ctx_handler;
		$this->ctx_handler_max_depth = $max_depth;
	}

	function getCacheDirectory() {
		return $this->cache ? $this->cache->getCacheDirectory() : null;
	}
	function setCacheDirectory($dir) {
		require_once('include/config/format/ConfigCache.php');
		$this->cache = new ConfigCacheMgr($dir);
	}
	function getCacheManager() {
		return $this->cache;
	}
	function setCacheManager($mgr) {
		$this->cache = $mgr;
	}

	function readCache($path) {
		if($this->cache) return $this->cache->readCache($this->target, $path, $this->getSiteVersion());
	}

	function startCache($path, $options=null) {
		if($this->cache && ! ($this->target instanceof ConfigCacheTarget)) {
			$tg = $this->cache->getCacheTarget($path, $options);
			if($tg) {
				$tg->setTarget($this->target);
				$this->prev_target = $this->target;
				$this->target = $tg;
				return true;
			}
		}
	}

	function endCache() {
		$this->target = $this->prev_target;
	}

	function parseFileWithTarget($file, &$targ, $options=null) {
		$this->setTargetArray($targ);
		return $this->parseFile($file, $options);
	}

	function parseFile($file, $options=null) {
		static $is_windows;
		if (!isset($is_windows))
			$is_windows = substr(strtoupper(PHP_OS), 0, 3) === 'WIN';

		if(! is_array($options)) $options = [];
		if(self::$time_parsing) {
			$st = microtime(true);
		}
		$no_cache = ! empty($options['no_cache']);
		if($no_cache || ! empty($options['no_load_cache']) || ! $this->readCache($file)) {

			if ($is_windows) {
				// on Windows, we rely on file locking to prevent config file corruption.
				// When using locks, it's important to prevent crash between locking and
				// unlocking. In this case, crash can occur only if file size is too big
				// to be read completely into memory (impying there are no bugs in PHP).
				// The following two lines make sure we crash _before_ obtaining
				// a lock if there is not enough memory. memory_get_usage() would not help.
				$contents = str_repeat('1', 100000);
				$contents = null;
			}

			$fp = @fopen($file, "rb");
			if($fp === false)
				throw new IAHConfigFileError("Config file not readable: $file");


			// locking starts here
			if ($is_windows) {
				if (!@flock($fp, LOCK_SH)) {
					fclose($fp);
					throw new IAHConfigFileError("Failed to lock config file for reading: $file");
				}
			}
			$size = filesize($file);
			if ($size)
				$contents = @fread($fp, filesize($file));
			else
				$contents = '';
			if ($is_windows) {
				flock($fp, LOCK_UN);
			}
			fclose($fp);
			// ends here

			if ($contents === false) {
				throw new IAHConfigFileError("Failed to read config file: $file");
			}
			$lines = preg_split('~\r?\n~', $contents);

			if(! $no_cache && empty($options['no_save_cache']))
				$cached = $this->startCache($file, $options) ? 2 : 0;
			else
				$cached = false;
			$this->initParse($file);
			foreach ($lines as $line) {
				$this->parseLine($line . "\n");
			}
			$this->endParse();
			if($cached) $this->endCache();
			$c = '';
		} else {
			$c = ' (c)';
			$cached = 1;
		}
		if(self::$time_parsing) {
			$time = microtime(true) - $st;
			self::$all_parsed[] = sprintf('%5.5f  ', $time) . $file . $c;
		}
		else
			self::$all_parsed[] = $file . $c;
		return compact('cached');
	}

	function parseFiles($files, $options=null) {
		foreach($files as $f)
			$this->parseFile($f, $options);
	}

	function reset() {
		$this->target->reset();
		$this->warnings = [];
	}

	// mainly for testing
	function parseStringWithTarget($data, &$targ, $options=null) {
		$this->setTargetArray($targ);
		$this->initParse('<string>');
		$lines = explode("\n", $data);
		foreach($lines as $l) {
			$this->parseLine($l);
		}
		$this->endParse();
	}

	function getResult() {
		return $this->target->getResult();
	}

	function getInfo() {
		return $this->target->getInfo();
	}

	function initParse($filename = '') {
		$this->context = [];
		$this->tabsize = self::DEFAULT_TAB_SIZE;
		$this->depth = 0;
		$this->indent = 1;
		$this->filename = $filename;
		$this->linepos = 0;
		$this->charpos = 0;
		$this->mode = 0;
		$this->mode_start = [1, 0];
		$this->array_key = '';
		$this->target->init($filename, $this->getSiteVersion());
		$this->have_key = false;
		$this->inline_value = false;
		$this->quoted_value = false;
		//$this->startContext(self::MODE_ARRAY);
	}

	function getSiteVersion() {
		// stub for now. can be used to invalidate cache files
		return [];
	}

	function endParse() {
		if(self::$DEBUG)
			$this->dbg('end parse');
		while(count($this->context)) {
			if($this->mode == self::MODE_COMMENT)
				throw new IAHConfigFileError("Unterminated comment: ".$this->getLinePosition(true, true));
			else if($this->mode == self::MODE_PI)
				throw new IAHConfigFileError("Unterminated processing instruction: ".$this->getLinePosition(true, true));
			$this->endContext();
		}
		$this->target->complete();
	}

	function startContext($mode, $indent=null) {
		if(self::$DEBUG)
			$this->dbg('! '.self::$modes[$mode]);
		$this->context[] = [
			'mode' => $this->mode,
			'mode_start' => $this->mode_start,
			'indent' => $this->indent,
			'array_key' => $this->array_key,
			'have_key' => $this->have_key,
			'inline_value' => $this->inline_value,
		];
		$this->depth ++;
		$this->mode = $mode;
		$this->mode_start = [$this->linepos, $this->charpos];
		if(isset($indent))
			$this->indent = $indent;
		if($this->ctx_handler && (($d = $this->ctx_handler_max_depth) == -1 || $d >= $this->depth))
			call_user_func_array($this->ctx_handler, [&$this, 1]);
	}

	function endContext($noAppend=false) {
		$fromMode = $this->mode;
		$fromDepth = $this->depth;
		$fromInline = $this->inline_value;

		if(self::$DEBUG)
			$this->dbg('/ '.self::$modes[$fromMode]);

		$this->depth --;
		$ctx = array_pop($this->context);
		$this->mode = $ctx['mode'];
		$this->mode_start = $ctx['mode_start'];
		$this->indent = $ctx['indent'];
		$this->array_key = $ctx['array_key'];
		$this->have_key = $ctx['have_key'];
		$this->inline_value = $ctx['inline_value'];
		if($fromMode == self::MODE_SIMPLE_VALUE && ! $noAppend) {
			if($this->quoted_value) {
				$val = $this->lastResult;
				$this->quoted_value = false;
			}
			else
				$val = $this->convertUnquoted($this->lastResult);
			if(self::$DEBUG)
				$this->dbg($val, "Value:", true);
			if($this->array_key === self::$ARRAY_APPEND_SIMPLE)
				$this->target->appendValue($this->depth, $val);
			else
				$this->target->appendValue($this->depth, $val, $this->array_key);
		} else if($fromMode == self::MODE_ARRAY && ! $noAppend) {
			if($this->array_key === self::$ARRAY_APPEND_SIMPLE || $this->array_key === self::$ARRAY_APPEND_ARRAY)
				$this->target->appendResult($this->depth, $fromDepth);
			else
				$this->target->appendResult($this->depth, $fromDepth, $this->array_key);
		} else if($fromMode == self::MODE_INFO && ! $noAppend) {
			$this->target->appendInfo($fromDepth);
		}

		if($this->ctx_handler && (($d = $this->ctx_handler_max_depth) == -1 || $d >= $this->depth))
			call_user_func_array($this->ctx_handler, [&$this, 0]);
		return $fromMode;
	}

	function getLinePosition($filename=true, $mode=false) {
		if($mode)
			$r = $this->mode_start[0] . ':' . ($this->mode_start[1]+1);
		else
			$r = $this->linepos . ':' . ($this->charpos+1);
		if($filename)
			return "{$this->filename}[$r]";
		return $r;
	}

	function parseIndent($text) {
		static $quick = [
			'' => 1,
			"\t" => 2,
			"\t\t" => 3,
			"\t\t\t" => 4,
			"\t\t\t\t" => 5,
		];
		if(isset($quick[$text])) return $quick[$text];
		$c = strlen($text);
		$indent = 1;
		$sp = 0;
		for($i = 0; $i < $c; $i++) {
			if($text[$i] == "\t")
				$indent ++;
			else
				$sp ++;
		}
		if($sp % $this->tabsize)
			return false;
		return $indent + (int)($sp / $this->tabsize);
	}

	function parseSpaces($line, $offset=0, $calc_indent=false) {
		$indent = 1;
		$sp = 0;
		for($i = $offset; ; $i++) {
			if($i >= $this->linelen)
				$c = null;
			else
				$c = $line[$i];
			if($c === ' ') {
				$sp ++;
				if($sp == $this->tabsize) {
					$indent ++;
					$sp = 0;
				}
			}
			else if($c === "\t")
				$indent ++;
			else if($c === "\r")
				continue;
			else if($c === "\n" || $c === '#' || $c === '' || $c === null) {
				$i = false;
				break;
			}
			else if($c == '{' && ($i+1 < $this->linelen) && $line[$i+1] == '#') {
				// multiline comment
				$this->startContext(self::MODE_COMMENT);
				$p = strpos($line, '#}', $i+2);
				if($p !== false) {
					$this->endContext();
					$i = $p + 1;
				}
				else {
					$i = false;
					break;
				}
			}
			else if($c == '<' && ($i+1 < $this->linelen) && $line[$i+1] == '?') {
				// processing instruction
				$this->startContext(self::MODE_PI);
				$p = strpos($line, '?'.'>', $i+2);
				if($p !== false) {
					$this->endContext();
					$i = $p + 1;
				}
				else {
					$i = false;
					break;
				}
			}
			else
				break;
		}
		if($calc_indent)
			return [$indent, $sp, $i];
		else
			return $i;
	}

	function checkIndent($indent, $sp) {
		if($sp)
			throw new IAHConfigFileError("Leading space is not on a tab boundary. Tab size: {$this->tabsize}, ".$this->getLinePosition());
	}

	function convertUnquoted($val, $trimmed=false) {
		if(! $trimmed) $val = rtrim($val);
		if($val === 'null' || $val === 'true' || $val === 'false') //array_key_exists($val, self::$special_values))
			return self::$special_values[$val];
		if(is_numeric($val)) {
			if(strpos($val, '.') !== false)
				return floatval($val);
			return intval($val);
		}
		if(strpos($val, '\\') !== false)
			$val = str_replace(['\#', '\"', '\t', '\r', '\n', '\\\\'], ['#', '"', "\t", "\r", "\n", "\\"], $val);
		return $val;
	}

	function parseValue(&$line, $after_key=false, $new_depth=null) {
		if($this->charpos < $this->linelen)
			$c = $line[$this->charpos];
		else
			return false;

		while($c === '@') {
			if(! ($ret = $this->parseDirective($line, $new_depth)))
				return false;
			if($ret !== true) {
				$this->lastResult = $ret;
				return true;
			}
			$i = $this->parseSpaces($line, $this->charpos);
			if($i === false)
				return false;
			$this->charpos = $i;
			$c = $line[$this->charpos];
		}

		if($c === '"') {
			$ret = $this->parseQuoted($line);
			if(! $after_key) $this->quoted_value = false;
			return $ret;
		}
		else if($c === '[' && $after_key) {
			if($this->mode == self::MODE_SIMPLE_VALUE)
				$this->endContext(true);
			$this->startContext(self::MODE_ARRAY);
			$this->target->initArray($this->depth);
			$this->inline_value = true;
			$this->have_key = true;
			$this->array_key = self::$ARRAY_APPEND_SIMPLE;
			if(self::$DEBUG)
				$this->dbg("start inline array");
			$this->charpos ++;
			return $this->parseInlineValue($line);
		}
		else if($c === '-' && ! $this->have_key && ($this->charpos >= $this->linelen || strpos("- \t\r\n", ($c2 = $line[$this->charpos+1])) !== false)) {
			$this->charpos ++;
			if($c2 == '-') {
				$this->charpos ++;
				$this->lastResult = self::$ARRAY_APPEND_ARRAY;
			}
			else
				$this->lastResult = self::$ARRAY_APPEND_SIMPLE;
			return true;
		}
		else {
			$cmt = strpos($line, '#', $this->charpos);
			if(! $after_key && ($p = strpos($line, ':', $this->charpos)) !== false && ($cmt === false || $cmt > $p)) {
				$this->lastResult = $this->convertUnquoted(substr($line, $this->charpos, $p - $this->charpos));
				$this->charpos = $p;
				return true;
			} else {
				if($this->inline_value) {
					$p1 = strpos($line, ',', $this->charpos);
					$p2 = strpos($line, ']', $this->charpos);
					if($p1 !== false)
						$p = ($p2 !== false ? min($p1, $p2) : $p1);
					else
						$p = ($p2 !== false ? $p2 : $this->linelen);
					$val = substr($line, $this->charpos, $p - $this->charpos);
					$this->charpos = $p;
					if(is_string($this->lastResult))
						$this->lastResult .= $val;
					else
						$this->lastResult = $val;
				} else {
					if($cmt !== false)
						$val = substr($line, $this->charpos, $cmt - $this->charpos);
					else
						$val = substr($line, $this->charpos);
					$this->charpos = $this->linelen;
					$this->lastResult = $after_key ? $val : $this->convertUnquoted($val);
				}
				return true;
			}
		}
	}

	function parseLine($line) {
		$this->linepos ++;
		$this->charpos = 0;
		$this->linelen = strlen($line);

		if(self::$DEBUG) {
			$this->dbg('> '.rtrim($line));
		}

		$indent = $this->indent;

		// previous multiline comment
		if($this->mode == self::MODE_COMMENT) {
			$i = 0;
			$p = strpos($line, '#}', $i);
			if($p !== false) {
				$this->endContext();
				$this->charpos = $p + 2;
			}
			else
				return;
		}
		if($this->mode == self::MODE_PI) {
			$i = 0;
			$p = strpos($line, '?'.'>', $i);
			if($p !== false) {
				$this->endContext();
				$this->charpos = $p + 2;
			}
			else
				return;
		}

		if($this->inline_value) {
			if(! $this->parseInlineValue($line, true))
				return;
			else
				$done = true;
		}
		else {
			if(self::$PREG_OPTIMIZE && (! $this->mode || $this->mode == self::MODE_ARRAY)) {
				if(preg_match('~^([ \t]*)(- +)?([\w\d_\.]+)\s*(?::\s*([\w\d_][^#]*?)\s*)?$~', $line, $m)) {
					$skip_preg = false;
					$indent = $this->parseIndent($m[1]);
					if($indent === false || $indent > $this->indent + 1) {
						throw new IAHConfigFileError("Leading space is not on a tab boundary (tab size={$this->tabsize}): ".$this->getLinePosition());
					}
					while($indent <= $this->indent && $this->mode) {
						//if($this->mode == self::MODE_SIMPLE_VALUE)
						//	$done = true;
						$this->endContext();
					}
					if(! empty($m[2])) { // hyphen
						if(isset($m[4])) {
							// and a value?
							$skip_preg = true;
						} else {
							$this->lastResult = $this->convertUnquoted($m[3], true);
							$this->target->appendValue($this->depth, $this->lastResult);
						}
					} else {
						$this->array_key = $this->convertUnquoted($m[3], true);
						if(isset($m[4])) {
							$this->lastResult = $this->convertUnquoted($m[4], true);
							$this->target->appendValue($this->depth, $this->lastResult, $this->array_key);
						}
						else {
							$this->startContext(self::MODE_ARRAY, $indent);
							$this->target->initArray($this->depth, null, $this->array_key);
						}
					}
					if(! $skip_preg) {
						$this->charpos = $this->linelen;
						return;
					}
				}
			}

			// parse leading spaces and comments
			list($indent, $sp, $i) = $this->parseSpaces($line, $this->charpos, true);
			if($i === false) {
				if(self::$DEBUG)
					$this->dbg('(blank)');
				if($this->mode != self::MODE_SIMPLE_VALUE)
					return;
			}

			if($this->mode == self::MODE_SIMPLE_VALUE) {
				if($indent > $this->indent) {
					// remove indent
					$sp = 0;
					$d = 1;
					for($i = 0; $d <= $this->indent; $i++) {
						$c = $line[$i];
						if($c == ' ') {
							$sp ++;
							if($sp < $this->tabsize)
								continue;
							else
								$sp = 0;
						}
						$d ++;
					}
					$this->lastResult .= rtrim(substr($line, $i), "\r\n") . "\n";
					return;
				} else if($i === false) {
					if (!$this->quoted_value)
						$this->lastResult .= "\n";
					return;
				} else
					$this->endContext();
			}

			$this->charpos = $i;

			if($sp || $indent > $this->indent + 1) {
				throw new IAHConfigFileError("Leading space is not on a tab boundary (tab size={$this->tabsize}): ".$this->getLinePosition());
			}

			if(! $this->parseValue($line, false, $indent))
				return;
			if(self::$DEBUG)
				$this->dbg("Key: '$this->lastResult'");

			while($indent <= $this->indent && $this->mode) {
				if($this->mode == self::MODE_SIMPLE_VALUE)
					$done = true;
				$this->endContext();
			}
		}

		if(! empty($done)) {}
		else if($this->lastResult === self::$START_INFO) {
			$this->startContext(self::MODE_INFO);
			$this->target->initArray($this->depth);
		}
		else {
			$readValue = $appendSimple = false;
			if($this->lastResult === self::$ARRAY_APPEND_SIMPLE) {
				$readValue = true;
				$appendSimple = true;
			} else {
				if($this->charpos < $this->linelen) {
					$c = $line[$this->charpos];
					if($c === ':') {
						$this->charpos ++;
						$readValue = true;
					}
				}
			}
			if($readValue) {
				$i = $this->parseSpaces($line, $this->charpos);
				if($i !== false) {
					$this->charpos = $i;
					$this->array_key = $this->lastResult;
					$this->startContext(self::MODE_SIMPLE_VALUE, $indent);
					$this->have_key = true;
					$this->lastResult = '';
					if(! $this->parseValue($line, true)) {
						//throw new IAHConfigFileError("Unrecognized value: ".$this->getLinePosition());
						$this->have_key = false;
						return; // probably in nested context
					}

					if($this->mode == self::MODE_SIMPLE_VALUE) {
						$this->endContext(); // appends value
					} else if($this->have_key) {
						if(self::$DEBUG)
							$this->dbg($this->lastResult, "VValue:");
						if($appendSimple)
							$this->target->appendValue($this->depth, $this->lastResult);
						else
							$this->target->appendValue($this->depth, $this->lastResult, $this->array_key);
					}
				} else {
					$this->array_key = $this->lastResult;
					$this->startContext(self::MODE_SIMPLE_VALUE, $indent);
					$this->lastResult = '';
				}
			}
			else {
				// if not in update mode
				$this->array_key = $this->lastResult;
				$this->startContext(self::MODE_ARRAY, $indent);
				if(! is_object($this->array_key))
					$this->target->initArray($this->depth, null, $this->array_key);
				else
					$this->target->initArray($this->depth);
			}
		}

		if($this->parseSpaces($line, $this->charpos) !== false)
			throw new IAHConfigFileError("Expected end of line: ".$this->getLinePosition());
	}

	function parseQuoted(&$line) {
		$multiline = ($this->mode == self::MODE_STRING);
		$p = $this->charpos;
		if(! $multiline) {
			$this->quoted_value = true;
			if(substr($line, $p, 3) === '"""') {
				$this->startContext(self::MODE_STRING, '');
				$this->inline_value = true;
				$multiline = true;
				$this->lastResult = '';
				$p += 3;
			} else
				$p ++;
		}
		$val = '';
		while(1) {
			$slashp = strpos($line, '\\', $p);
			$delim = $multiline ? '"""' : '"';
			$qp = strpos($line, $delim, $p);
			if($qp !== false && ($slashp === false || $slashp > $qp)) {
				$val .= substr($line, $p, $qp-$p);
				$this->charpos = $qp + strlen($delim);
				if($this->mode == self::MODE_STRING) {
					$this->lastResult .= $val;
					$this->endContext();
					return true;
				}
				break;
			}
			else if($slashp !== false) {
				$val .= substr($line, $p, $slashp-$p);
				$p = $slashp + 1;
				if($p < $this->linelen) {
					$c = $line[$p];
					if($c == 'n')
						$special = "\n";
					else if($c == 'r')
						$special = "\r";
					else if($c == 't')
						$special = "\t";
					// add '\uNNNN'
					else
						unset($special);
					if(isset($special)) {
						$val .= $special;
						$p ++;
					} else {
						$val .= $c;
						$p ++;
					}
				}
			}
			else if(! $multiline) {
				throw new IAHConfigFileError("Unterminated string: ".$this->getLinePosition());
			}
			else {
				// multiline
				$val .= rtrim(substr($line, $p), "\r\n");
				$this->lastResult .= $val . "\n";
				return false;
			}
		}
		$this->lastResult = $val;
		return true;
	}

	function parseInlineValue(&$line, $continued=false) {
		$d = $this->depth;
		while($this->depth >= $d) {
			$quoted = false;
			if($this->mode == self::MODE_STRING) {
				if(! $this->parseQuoted($line))
					return false;
				if(self::$DEBUG)
					$this->dbg("Multiline string: '".str_replace("\n", "\\n", $this->lastResult)."'");
				if(! $this->inline_value)
					return true;
				$quoted = true;
			}
			else if($this->mode == self::MODE_SIMPLE_VALUE) {
				if(self::$DEBUG)
					$this->dbg("array simple value");
				$p1 = strpos($line, ',', $this->charpos);
				$p2 = strpos($line, ']', $this->charpos);
				if($p1 !== false)
					$p = ($p2 !== false ? min($p1, $p2) : $p1);
				else
					$p = $p2;
				if($p !== false) {
					$this->lastResult .= substr($line, $this->charpos, $p - $this->charpos);
					$this->endContext();
					$this->charpos = $p;
				} else {
					$this->lastResult .= substr($line, $this->charpos);
					return false;
				}
			}
			else {
				$i = $this->parseSpaces($line, $this->charpos);
				if($i === false)
					return false;
				$this->charpos = $i;
				if(! $this->parseValue($line, true))
					return false;
			}
			$c = ($this->charpos < $this->linelen ? $line[$this->charpos] : null);
			if($c === ',' || $c === ']') {
				if($c === ']' && $this->lastResult === '') {
					; // skip blank value
				}
				else if($c === ',' && ! $this->have_key && $this->lastResult === '') {
					$this->have_key = true; // skip blank value
				}
				else {
					if(self::$DEBUG)
						$this->dbg("inline array value: $this->lastResult");
					if($quoted || is_array($this->lastResult))
						$val = $this->lastResult;
					else
						$val = $this->convertUnquoted($this->lastResult);
					$this->target->appendValue($this->depth, $val);
					$this->lastResult = '';
				}
				if($c === ']') {
					if(self::$DEBUG)
						$this->dbg("end inline array");
					while($this->mode && $this->endContext() != self::MODE_ARRAY) ;
					$this->have_key = false; // ignore next comma if nested inline
				}
				$this->charpos ++;
			}
			else {
				if($this->inline_value && $this->mode == self::MODE_ARRAY)
					continue; // continue inline array
				$this->startContext(self::MODE_SIMPLE_VALUE);
				return false;
			}
		}
		return true;
	}

	function convertDirectiveValue($val) {
		$val = trim($val);
		if(preg_match('/^"(.*)"$/', $val, $m))
			return stripslashes($m[1]);
		return $this->convertUnquoted($val);
	}

	function parseDirective($str, $new_depth=null) {
		$direct = substr($str, $this->charpos);
		if(! preg_match('~^@(\w+(?:\-\w+)*)(.*)$~u', $direct, $m))
			throw new IAHConfigFileError("Expected directive: ".$this->getLinePosition());
		$direct = $m[1];
		$args = $m[2];
		if(isset($new_depth)) {
			while($new_depth <= $this->indent && $this->mode) {
				$this->endContext();
			}
		}
		switch($direct) {
			case 'tabsize':
				if(preg_match('~^\s+(\d+)\s*$~', $args, $m))
					$this->tabsize = $m[1];
				else
					throw new IAHConfigFileError("Bad formatting for tab size: ".$this->getLinePosition());
				return false;
			case 'info':
				if($this->charpos != 0)
					throw new IAHConfigFileError("@info must occur at document outline level: ".$this->getLinePosition());
				$this->charpos += 5;
				if(! preg_match('~^\s*$~', $args))
					throw new IAHConfigFileError("Unexpected data after @info: ".$this->getLinePosition());
				return self::$START_INFO;
			case 'cfg-version':
				return false;
			case 'remove':
				if($this->mode != self::MODE_ARRAY && $this->mode != self::MODE_INFO)
					throw new IAHConfigFileError("Unexpected @remove, must be in array context: ".$this->getLinePosition());
				$args = $this->convertDirectiveValue($args);
				$this->target->removeValue($this->depth, $args);
				return false;
			case 'clear':
				if($this->mode != self::MODE_ARRAY && $this->mode != self::MODE_INFO)
					throw new IAHConfigFileError("Unexpected @clear, must be in array context: ".$this->getLinePosition());
				$this->target->clearResult($this->depth);
				return false;
			case 'orig':
				if($this->mode != self::MODE_ARRAY && $this->mode != self::MODE_INFO)
					throw new IAHConfigFileError("Unexpected @orig, must be in array context: ".$this->getLinePosition());
				$args = $this->convertDirectiveValue($args);
				$this->target->appendOrigValue($this->depth, $args);
				return false;
			case 'binary':
			case 'raw':
				$this->charpos += strlen($direct)+1;
				break;
			default:
				throw new IAHConfigFileError("Unrecognized directive '$direct': ".$this->getLinePosition());
		}
		return true;
	}

	function dbg($msg, $label=null, $quote=false) {
		if(is_array($msg) || is_object($msg))
			$msg = print_r($msg, true);
		else if(is_string($msg) && $quote)
			$msg = "'$msg'";
		if(isset($label))
			$msg = $label . ' '. $msg;
		$this->target->log($msg);
	}
}


abstract class ConfigTarget {
	var $info = [];
	var $filename;
	var $parse_log = [];
	var $log_enabled = false;
	var $version;

	// start parsing a new file (may do multiple files in sequence)
	function init($filename='', $version=null) {
		$this->filename = $filename;
		$this->version = $version;
	}
	// finish current file
	function complete() {}
	// check config version (not yet implemented, may throw an exception to stop parsing)
	function checkVersion($info) { }
	// start at a new depth
	abstract function initArray($depth, $value=null, $key=null);
	// add a value to current array
	abstract function appendValue($depth, $value, $key=null);
	// add the current parsed result to the current array
	abstract function appendResult($depth, $from_depth, $key=null);
	// append current stack to info
	abstract function appendInfo($from_depth=1);
	// append original array value
	abstract function appendOrigValue($depth, $key);
	// remove an array key
	abstract function removeValue($depth, $key);
	// return the current parse result
	abstract function getResult($depth=0);
	// check if anything has been added to result
	abstract function isBlank();
	//
	function getInfo() {
		return $this->info;
	}
	// clear the result
	abstract function clearResult($depth=0);
	function reset() {
		$this->info = [];
		$this->parse_log = [];
	}

	function log() {
		if(! $this->log_enabled) return;
		$line = [];
		foreach(func_get_args() as $a) {
			if($a === true) $a = 'true';
			else if($a === false) $a = 'false';
			else if($a === null) $a = 'null';
			else if(is_array($a)) $a = 'array';
			else if($a === '') $a = "''";
			else $a = (string)$a;
			$line[] = $a;
		}
		$this->parse_log[] = implode(' ', $line);
	}
	function getLog() {
		return $this->parse_log;
	}
}

class ConfigArrayTarget extends ConfigTarget {
	var $results = [];
	var $orig = [];
	var $loaded = [];
	var $result_callback;
	var $result_keys = null;
	var $preload_callback;
	var $extend = true;

	function setCustomArray(&$t) {
		$this->results[0] = &$t;
	}

	function isBlank() {
		return (! $this->orig && (! $this->results || ! $this->results[0]));
	}
	function setResultCallback(Callable $cb, $keys=null) {
		$this->result_callback = $cb;
		$this->result_keys = $keys;
	}
	function getResultKeys() {
		return $this->result_keys;
	}
	function setPreloadCallback(Callable $cb) {
		$this->preload_callback = $cb;
	}
	function loadResultKeys() {
		if($this->result_callback && $this->result_keys) {
			$cb = $this->result_callback;
			foreach($this->result_keys as $key) {
				if(! empty($this->loaded[$key])) continue;
				$cb($this, $key);
				$this->loaded[$key] = true;
			}
		}
	}

	function init($filename='', $version=null) {
		parent::init($filename, $version);
		$c = count($this->results);
		if(! $c)
			$this->results[] = [];
		else if($c > 1)
			$this->results = [$this->results[0]];
	}

	function initArray($depth, $value=null, $key=null) {
		for($i = 0; $i < $depth; $i++)
			if(! isset($this->results[$i])) $this->results[$i] = [];
		if($depth == 1 && isset($key) && $this->preload_callback) {
			$plc = $this->preload_callback;
			$this->results[$depth] = $plc($this, $key);
			if(! is_array($this->results[$depth]))
				$this->results[$depth] = [];
		} else if(isset($key) && $depth && isset($this->results[$depth-1][$key]))
			$this->results[$depth] = $this->results[$depth-1][$key];
		else
			$this->results[$depth] = [];
		if(isset($value)) {
			if ($this->extend)
				array_extend_recursive($this->results[$depth], $value);
			else
				array_update_recursive($this->results[$depth], $value);
		}
	}

	function appendValue($depth, $value, $key=null) {
		if(isset($key))
			$this->results[$depth][$key] = $value;
		else
			$this->results[$depth][] = $value;
	}

	function appendResult($depth, $from_depth, $key=null) {
		if(isset($key))
			$this->results[$depth][$key] = $this->getResult($from_depth);
		else
			$this->results[$depth][] = $this->getResult($from_depth);
		$this->results[$from_depth] = [];
	}

	function appendOrigValue($depth, $key) {
		if(! isset($this->orig[$depth])) $this->orig[$depth] = [];
		$this->results[$depth][$key] = array_get_default($this->orig[$depth], $key);
	}

	function appendInfo($from_depth=1) {
		foreach($this->results[$from_depth] as $k => $v)
			$this->info[$k] = $v;
		$this->results[$from_depth] = [];
	}

	function removeValue($depth, $key) {
		unset($this->results[$depth][$key]);
	}

	function getResult($depth=0) {
		if(! $depth && ! isset($this->results[$depth]))
			return; // if parse was never started
		return $this->results[$depth];
	}

	function getResultKey($key) {
		if($this->result_callback && empty($this->loaded[$key])) {
			$cb = $this->result_callback;
			$cb($this, $key);
			$this->loaded[$key] = true;
		}
		if(isset($this->results[0][$key]))
			return $this->results[0][$key];
	}

	function clearResult($depth=0) {
		$this->orig[$depth] = $this->results[$depth];
		$this->results[$depth] = [];
	}

	function reset() {
		parent::reset();
		$this->results = $this->orig = [];
	}
}


// quickly inspect files from the command line
if(php_sapi_name() == 'cli' && empty($_SERVER['REMOTE_ADDR']) && ! empty($_SERVER['argv'])) {
	$args = $_SERVER['argv'];
	$fname = array_shift($args);
	if(basename($fname) === basename(__FILE__)) {
		$mode = 'check';
		$files = [];
		foreach($args as $arg) {
			if($arg === '-d' || $arg === '--debug') $mode = 'debug';
			else if($arg === '-p' || $arg === '--print') $mode = 'print';
			else if($arg[0] != '-') $files[] = $arg;
		}
		foreach($files as $f) {
			try {
				$cfg = ConfigParser::load_file($f);
			} catch(IAHConfigFileError $e) {
				echo $e->getMessage() . "\n";
				continue;
			}
			if($mode == 'debug') {
				print_r($cfg);
			} else if($mode == 'print') {
				if(! class_exists('ConfigWriter')) require_once('ConfigWriter.php');
				$w = new ConfigWriter();
				echo $w->encodeToString($cfg) . "\n";
			} else {
				echo "No syntax errors detected in $f\n";
			}
		}
	}
}


?>
