<?php

class DBCache {
	var $disabled;
	var $row_cache;
	var $obj_cache;
	var $all_fetched;
	var $auto_fetch_all;
	var $cache_objects;
	var $default_row_ids;
	var $stats;
	
	function __construct() {
		$this->init();
	}
	
	function init() {
		$this->setDisabled(! AppConfig::setting('dbcache.enabled', true));
		$this->row_cache = array();
		if($this->obj_cache) {
			// clean up circular references in objects
			foreach($this->obj_cache as $cls => &$rows)
				foreach($rows as &$obj)
					$obj->cleanup();
		}
		$this->stats = array(
			'disabled' => $this->disabled,
			'obj_found' => 0,
			'row_found' => 0,
			'row_retrieved' => 0,
			'all_retrieved' => 0,
			'cls' => array(),
		);
		$this->obj_cache = array();
		$this->all_fetched = array();
		$this->auto_fetch_all = AppConfig::setting('dbcache.auto_fetch_all', array());
		$this->cache_objects = AppConfig::setting('dbcache.cache_objects', array());
	}
	
	function getStats() {
		$stat = $this->stats;
		$stat['disabled'] = $this->disabled;
		return $stat;
	}
	
	function setDisabled($state=true) {
		$state |= AppConfig::in_installer();
		$this->disabled = $state;
	}
	
	function isAutoFetched($bean_name) {
		return ! empty($this->auto_fetch_all[$bean_name]);
	}
	
	function getObject($cls, $id, $autoFetch=true) {
		if(! $id) {
			return null;
		}
		
		$cache_obj = ! $this->disabled && ! empty($this->cache_objects[$cls]);
		if($cache_obj && isset($this->obj_cache[$cls][$id])) {
			if(! isset($this->stats['cls'][$cls])) $this->stats['cls'][$cls] = 0;
			$this->stats['cls'][$cls] ++;
			$this->stats['obj_found'] ++;
			return $this->obj_cache[$cls][$id];
		}
		
		$row = $this->getRow($cls, $id, $autoFetch, true);
		if(! isset($row)) {
			return null;
		}
		require_bean($cls);
		$bean = new $cls;
		$bean->populateFromRow($row);
		$bean->processed_dates_times = array();
		$bean->check_date_relationships_load();
		if($cache_obj)
			$this->obj_cache[$cls][$id] =& $bean;
		return $bean;
	}
	
	
	function getRow($cls, $id, $autoFetch=true, $encode=false) {
		if(! isset($this->stats['cls'][$cls])) $this->stats['cls'][$cls] = 0;
		$this->stats['cls'][$cls] ++;
		if($this->disabled) {
			$row = $this->fetchRow($cls, $id);
			$this->stats['row_retrieved'] ++;
		}
		else if(isset($this->row_cache[$cls][$id])) {
			$row =& $this->row_cache[$cls][$id];
			$this->stats['row_found'] ++;
		}
		else if(! empty($this->all_fetched[$cls])) {
			$row = $this->createStandardRecord($cls, $id);
			if($row === false)
				$row = null;			
		}
		if(! isset($row)) {
			if($autoFetch) {
				if(! empty($this->auto_fetch_all[$cls])) {
					$this->getAllRows($cls);
					$this->stats['all_retrieved'] ++;
					$row =  $this->getRow($cls, $id, false, false);
					if ($row === false)
						$row = null;
				}
				if (!isset($row)) {
					$row = $this->fetchRow($cls, $id);
					if ($row === false)
						$row = null;
					$this->stats['row_retrieved'] ++;
				}
			}
		}

		if(! isset($row)) {
			return null;
		}
		
		if(! $encode)
			return $row;
		return array_map('to_html', $row);
	}
	
	
	function createStandardRecord($cls, $id) {
		global $db;
		$std_ids = AppConfig::setting("model.detail.$cls.standard_records");
		if(! $std_ids || ! is_array($std_ids) || ! in_array($id, $std_ids))
			return false;
		
		if(AppConfig::setting("model.fields.$cls.deleted")) {
			$tbl = AppConfig::setting("model.detail.$cls.table_name");
			$tbl = $db->quoteField($tbl);
			$qid = $db->quote($id);
			$db->query("DELETE FROM $tbl WHERE id='$qid'");
		}
		
		require_once('include/database/RowUpdate.php');
		$upd = RowUpdate::blank_for_model($cls);
		$upd->set('id', $id);
		// model hooks are responsible for populating other fields
		if($upd->save()) {
			return $this->fetchRow($cls, $id, true);
		}
		return false;
	}
	
	function checkStandardRecords($cls, $onlyIfCached=false) {
		if($onlyIfCached && empty($this->auto_fetch_all[$cls]))
			return false;
		$this->getAllRows($cls);
	}
	
	
	function fetchRow($cls, $id, $noStandard=false) {
		global $db;
		$table = AppConfig::setting("model.detail.$cls.table_name");
		
		$idq = $db->quote($id);
		$sql = "SELECT * FROM `{$table}` WHERE id='$idq' AND NOT deleted";
		$result = $db->query($sql);
		if($result) {
			$row = $db->fetchByAssoc($result, -1, false);
		} else if(! $noStandard) {
			$row = $this->createStandardRecord($cls, $id);			
		} else
			$row = false;
		if($row !== false) {
			$dname = AppConfig::setting("model.detail.$cls.display_name");
			if($dname) self::fill_display_name($dname, $row);
			$this->updateRow($cls, $row);
		} else {
			$this->updateRow($cls, $id); // unset cache
			$row = null;
		}
		return $row;
	}
	
	static function fill_display_name($dname, &$row) {
		global $locale;
		// only certain types are currently implemented here (ones needed by default DBCached modules)
		// standard query-based implementation is in ListQuery
		if(! $dname) return;
		if(is_string($dname)) $dname = array($dname);
		$dtype = array_get_default($dname, 'type');
		if($dtype == 'name_rate')
			$dname['value_function'] = array('ListQuery', 'format_name_rate');
		$fields = array_get_default($dname, 'fields', array());
		if($dtype == 'user_name') {
			$use_real_names = isset($locale) ? $locale->getUseRealNames() : null;
			$field = (! $use_real_names || $use_real_names === 'off') ? 'user_name' : 'full_name';
			$display = array_get_default($row, $field, '');
		}
		else if($dtype == 'prefixed') {
			$argvals = array();
			foreach($fields as $f)
				$argvals[] = array_get_default($row, $f);
			$display = implode(': ', $argvals);
		}
		else if(isset($dname['value_function'])) {
			$argvals = array();
			foreach($fields as $f)
				$argvals[] = array_get_default($row, $f);
			$display = call_user_func_array($dname['value_function'], $argvals);
		}
		else {
			$display = array_get_default($row, 'name', '');
		}
		$row['_display'] = $display;
	}
	
	function getAllRows($cls, $autoFetch=true, $encode=false) {
		if(! $this->disabled && ! empty($this->all_fetched[$cls]) && !empty($this->row_cache[$cls]))
			$rows =& $this->row_cache[$cls];
		else if($autoFetch) {
			$found = 0;
			if(! $this->disabled) {
				$rows = AppConfig::cache_get("DBCache.$cls", $found, true);
			}
			if($found) {
				$this->row_cache[$cls] = $rows;
				$this->all_fetched[$cls] = 1;
			} else
				$rows = $this->fetchAllRows($cls);
		}
		if(! isset($rows))
			return null;
		
		if(! $encode)
			return $rows;
		$ret = array();
		foreach($rows as $k => &$row) {
			$ret[$k] = array_map('to_html', $row);
		}
		return $ret;
	}

	function fetchAllRows($cls) {
		global $db;
		
		if($cls == 'DynField') {
			// special behaviour to avoid recursion as custom field definitions are loaded
			$table = 'fields_meta_data';
			$dname = 'name';
			$std_ids = null;
			$ord = null;
		} else {
			$table = AppConfig::setting("model.detail.$cls.table_name");
			$dname = AppConfig::setting("model.detail.$cls.display_name");
			$std_ids = AppConfig::setting("model.detail.$cls.standard_records");
			// FIXME standard sort order is stored in list.php, need to write new ConfigBase to access this quickly		
			// ajrw - order should not matter here?
			if($cls == 'TaxCode')
				$ord = 'position';
			else
				$ord = 'name';
		}
		
		$sql = "SELECT * FROM `{$table}` WHERE NOT deleted" . ($ord ? " ORDER BY $ord" : '');
		$db_rows = $db->fetchSQL($sql, "Error retrieving table: {$table}");
		$rows = array();
		if($db_rows) {
			foreach($db_rows as $k => &$row) {
				if($dname) self::fill_display_name($dname, $row);
				$rows[$row['id']] = $row;
			}
		}
		
		if(is_array($std_ids)) {
			foreach($std_ids as $test_id) {
				if(! isset($rows[$test_id])) {
					$r2 = $this->createStandardRecord($cls, $test_id);
					if($r2) $rows[$test_id] = $r2;
				}
			}
		}
		
		if(! $this->disabled) {
			if(! empty($this->auto_fetch_all[$cls])) {
				AppConfig::cache_put("DBCache.$cls", $rows, true, true);
			}
			$this->row_cache[$cls] =& $rows;
			$this->all_fetched[$cls] = true;
		}
		return $rows;
	}
	
	function getAllObjects($cls, $autoFetch=true) {
		$rows = $this->getAllRows($cls, $autoFetch, true);
		if(! isset($rows))
			return null;
		$ret = array();
		require_bean($cls);
		$seed = new $cls;
		$cache_obj = ! $this->disabled && ! empty($this->cache_objects[$cls]);
		foreach($rows as &$row) {
			if(! isset($row['id']))
				continue;
			$id = $row['id'];
			if($cache_obj && isset($this->obj_cache[$cls][$id]))
				$ret[$id] = $this->obj_cache[$cls][$id];
			else {
				$bean = clone($seed);
				$bean->populateFromRow($row);
				$bean->processed_dates_times = array();
				$bean->check_date_relationships_load();
				if($cache_obj)
					$this->obj_cache[$cls][$id] =& $bean;
				$ret[$id] = $bean;
			}
		}
		$seed->cleanup();
		return $ret;
	}

	function updateRow($cls, $row) {
		if(is_string($row))
			$id = $row;
		else {
			if(! is_array($row) || ! isset($row['id']))
				return;
			$id = $row['id'];
		}
		if(! empty($this->auto_fetch_all[$cls]) && $id)
			AppConfig::cache_clear("DBCache.$cls");
		if(isset($this->obj_cache[$cls][$id])) {
			$this->obj_cache[$cls][$id]->cleanup();
			unset($this->obj_cache[$cls][$id]);
		}
		if(is_array($row))
			$this->row_cache[$cls][$id] = $row;
		else
			unset($this->row_cache[$cls][$id]);
	}
	
	function invalidate($cls, $id=null) {
		if(isset($id)) {
			$this->updateRow($cls, $id);
		} else {
			unset($this->row_cache[$cls]);
			if(isset($this->obj_cache[$cls])) {
				foreach($this->obj_cache[$cls] as &$obj)
					$obj->cleanup();
				unset($this->obj_cache[$cls]);
			}
			if(! empty($this->auto_fetch_all[$cls]))
				AppConfig::cache_clear("DBCache.$cls");
		}
	}

}

?>
