<?php

require_once 'include/utils/basic_utils.php';

class IAHDateRangeError extends IAHInternalError {}
class IAHDateParseError extends IAHDateRangeError {}
class IAHDateEvalError extends IAHDateRangeError {}

class IAHDateRange {
	var $model;
	var $local = true;
	var $context;

	static $db_date_format = 'Y-m-d H:i:s';

	static $offset_intervals = [
		'i' => 'minute', 'h' => 'hour', 'd' => 'day', 'w' => 'week',
		'm' => 'month', 'y' => 'year', 'q' => 'quarter', 'Q' => 'fiscalquarter',
		'F' => 'fiscalyear',
	];


	function __construct($model=null, $context=null) {
		$this->setModel($model);
		$this->setContext($context);
	}

	function getContext() {
		return $this->context;
	}
	function setContext($ctx) {
		$this->context = $ctx;
	}

	function getModel() {
		return $this->model;
	}
	function setModel($model) {
		if(is_string($model)) $model = new ModelDef($model);
		if(is_object($model))
			$this->model = $model;
		else
			$this->model = null;
	}

	function getFieldDefinition($field_name) {
		// fixme - support linked fields
		if(! $this->model) return;
		$fdef = $this->model->getFieldDefinition($field_name);
		return $fdef;
	}

	function getFieldFormat($fdef, $default='datetime') {
		if(is_array($fdef)) {
			$type = array_get_default($fdef, 'type');
			if($type == 'date' || array_get_default($fdef, 'format') == 'date_only')
				return 'date';
			if($type == 'datetime')
				return $type;
		}
		return $default;
	}

	// return information about filter needed by input control
	function expandFilter($field, $filter=null, $default=false, $compact=false) {
		if(is_array($field)) {
			if(empty($field['name']))
				return;
			$fdef = $field;
		} else
			$fdef = $this->getFieldDefinition($field);
		if(! $fdef) return;
		$ret = [
			'name' => $fdef['name'],
			'bean_name' => $this->model ? $this->model->getModelName() : null,
			'module' => $this->model ? $this->model->getModuleDir() : null,
			'context' => $this->context,
		];
		$ret['mode'] = $this->getFieldFormat($fdef);
		if($compact) {
			if(is_string($filter) && $filter !== '' && $filter[0] == '{') {
				$json = getJSONobj();
				$filter = $json->decode($filter);
			}
			if(! is_array($filter))
				$filter = null;
			try {
				if($filter) {
					$encoded = $this->encodeRange($filter);
					$filter = $this->parseRange($encoded);
				}
			} catch(IAHDateEvalError $e) {
				$filter = print_r($filter, true);
				$GLOBALS['log']->warn("Error encoding date filter ($filter): ".$e->getMessage());
				$filter = null;
			}
		}
		else if($filter) {
			try {
				$filter = $this->parseRange($filter);
			} catch(IAHDateParseError $e) {
				if(! $default) throw $e;
				$GLOBALS['log']->warn("Error parsing date filter ($filter): ".$e->getMessage());
				$filter = null;
			}
			if($filter)
				$encoded = $this->encodeRange($filter);
		}
		if(! $filter) {
			$filter = ['operator' => 'any'];
			$encoded = $this->encodeRange($filter);
		}
		$ret['filter'] = $filter;
		$ret['encoded'] = $encoded;
		$ret['text'] = $this->formatRange($filter);
		$ret['label'] = $this->translateFieldLabel($fdef);
		return $ret;
	}

	// parse a formatted range string to produce a tree structure
	function parseRange($str) {
		if(strpos($str, '!!') !== false) {
			$split = array_map('trim', explode('!!', $str, 2));
			$negate = true;
		} else {
			$split = array_map('trim', explode('||', $str, 2));
			$negate = false;
		}
		if(count($split) > 1) {
			// if a range is specified but start or end is empty, default to now/today
			if(! strlen($split[0]))
				$split[0] = 'now';
			if(! strlen($split[1]))
				$split[1] = 'now';
			try {
				$from = $this->parseDate($split[0]);
				$to = $this->parseDate($split[1]);
			} catch(IAHDateParseError $e) {
				if(! isset($from))
					throw new IAHDateParseError("Error parsing from date: ".$e->getMessage());
				else
					throw new IAHDateParseError("Error parsing to date: ".$e->getMessage());
			}
			$range = ['operator' => 'range', 'start' => $from, 'end' => $to];
			if($negate) $range['negate'] = true;
		}
		else {
			$range = $this->parseDate($split[0], true);
		}
		return $range;
	}

	function parseDate($str, $filter=false) {
		$parts = explode('/', $str);
		$ret = [];
		$base = null;
		foreach($parts as $idx => $part) {
			if(preg_match('/^not:\s*(.*)/i', $part, $m)) {
				$negate = true;
				$part = $m[1];
			} else {
				$negate = false;
			}

			if($part == 'any') {
				$base = ['operator' => 'any'];
				$part = '';
			}
			else if(preg_match('/^(\d{4}-\d{1,2}-\d{1,2})\s*(\d{1,2}:?\d{2})?(.*)$/', $part, $m)) {
				if($idx) throw new IAHDateParseError("Expected: date modifier");
				if(! empty($m[2]))
					$base = ['operator' => 'datetime', 'value' => $m[1].' '.$m[2]];
				else
					$base = ['operator' => 'date', 'value' => $m[1]];
				$part = $m[3];
			}
			else if(preg_match('/^(now|today|tomorrow|yesterday|nextbus)\b(.*)$/i', $part, $m)) {
				if($idx) throw new IAHDateParseError("Expected: date modifier");
				$base = ['operator' => 'special', 'value' => strtolower($m[1])];
				$part = $m[2];
			}
			else if(preg_match('/^field:\s*([A-Z_][A-Z0-9_]*)(.*)$/i', $part, $m)) {
				if($idx) throw new IAHDateParseError("Expected: date modifier");
				$base = ['operator' => 'field', 'value' => $m[1]];
				$part = $m[2];
			}
			else if(preg_match('/^(month|year|fiscalyear):\s*([0-9]+)(.*)$/i', $part, $m)) {
				if($idx) throw new IAHDateParseError("Expected: date modifier");
				$base = ['operator' => $m[1], 'value' => $m[2]];
				$part = $m[3];
			}
			if($negate) $base['negate'] = true;
			if(! strlen($part)) continue;
			$base = $this->parseOffset($part, $base, $filter);
		}
		return $base;
	}

	function parseOffset($str, $base=null, $filter=false) {
		$str = trim($str);
		while(strlen($str)) {
			$operator = 'offset';
			$negate = false;
			$ret = null;

			if(preg_match('/^([A-Za-z_]+)(\:?)\s*(.*)$/', $str, $m)) {
				$pfx = strtolower($m[1]);
				if(! empty($m[2])) {
					if($pfx == 'offset') {
						// default
						$str = $m[3];
					}
					else if($pfx == 'date' || $pfx == 'not_date' || $pfx == 'datetime' || $pfx == 'not_datetime'
							|| $pfx == 'special' || $pfx == 'not_special') {
						$operator = str_replace('not_', '', $pfx);
						$negate = ($pfx != $operator);
						$str = $m[3];
					}
					else if($pfx == 'within' || $pfx == 'not_within') {
						$operator = 'within';
						$negate = ($pfx != 'within');
						// now parse offset
						$str = $m[3];
					}
					else if($pfx == 'next' || $pfx == 'previous' || $pfx == 'current') {
						$operator = $pfx;
						$str = $m[3];
					}
					else if($pfx == 'start' || $pfx == 'end') {
						$period = $this->parseOffsetPeriod($m[3]);
						$ret = ['operator' => $pfx, 'period' => $period];
						$str = '';
					}
					else if($pfx == 'before' || $pfx == 'after' || $pfx == 'not_before' || $pfx == 'not_after') {
						// not sure this is needed (before:x is redundant given x/before)
						if(! $filter) throw new IAHDateParseError("Unexpected date filter ({$m[1]})");
						if(substr($pfx, 0, 4) == 'not_') {
							$operator = substr($pfx, 4);
							$negate = true;
						} else
							$operator = $pfx;
						$dt = $this->parseDate($m[3]);
						$ret = ['operator' => $operator, 'value' => $dt];
						$str = '';
					}
					else {
						throw new IAHDateParseError("Unknown date offset: {$pfx}");
					}
				}
				else if($pfx == 'before' || $pfx == 'after' || $pfx == 'not_before' || $pfx == 'not_after') {
					if(! $filter) throw new IAHDateParseError("Unexpected date filter ({$pfx})");
					if(substr($pfx, 0, 4) == 'not_') {
						$operator = substr($pfx, 4);
						$negate = true;
					} else
						$operator = $pfx;
					$ret = ['operator' => $operator];
					$str = $m[3];
				}
				else if($pfx == 'empty' || $pfx == 'not_empty') {
					if(! $filter) throw new IAHDateParseError("Unexpected date filter ({$pfx})");
					if(substr($pfx, 0, 4) == 'not_') {
						$operator = substr($pfx, 4);
						$negate = true;
					} else
						$operator = $pfx;
					$ret = ['operator' => $operator];
					$str = $m[3];
					if(strlen($str)) throw new IAHDateParseError("Expected: blank");
				} else if($pfx == 'after_today') {
					if(! $filter) throw new IAHDateParseError("Unexpected date filter ({$pfx})");
					$ret = ['operator' => 'special', 'value' => 'tomorrow'];
					$str = $m[3];
				} else if($pfx == 'before_today') {
					if(! $filter) throw new IAHDateParseError("Unexpected date filter ({$pfx})");
					$ret = ['operator' => 'special', 'value' => 'yesterday'];
					$str = $m[3];
				}
				else {
					throw new IAHDateParseError("Unknown date offset: {$pfx}");
				}
			}

			if(! $ret) {
				if(preg_match('/^(\+?-?)(\d*\.?\d+)\s*([A-Za-z_]*)(.*)/', $str, $m)) {
					$period = ! empty($m[3]) ? $m[3] : '';
					$period = $this->parseOffsetPeriod($period);
					$count = (float)$m[2];
					if($m[1] == '-') $count = - $count;
					else if($m[1] == '+-') $operator = 'within';
					if($operator == 'within' && ! $filter)
						throw new IAHDateParseError("Unexpected date filter (within)");
					$ret = ['operator' => $operator, 'count' => $count, 'period' => $period];
					$str = $m[4];
				}
				else {
					throw new IAHDateParseError("Unknown date offset");
				}
			}

			if($negate) $ret['negate'] = true;
			if(isset($base)) $ret['value'] = $base;
			$base = $ret;
		}
		return $base;
	}

	function parseOffsetPeriod($period, $default='day') {
		$period = trim($period);
		if(! strlen($period))
			$period = $default;
		else if(isset(self::$offset_intervals[$period]))
			$period = self::$offset_intervals[$period];
		else {
			$idx = array_search($period, self::$offset_intervals);
			if($idx == false && substr($period, -1) == 's') {
				$period = substr($period, 0, -1);
				$idx = array_search($period, self::$offset_intervals);
			}
			if($idx == false)
				throw new IAHDateParseError("Unknown offset period: $period");
		}
		return $period;
	}

	// eval filter on a given row result; used by workflow
	function evalFilter($range, $result, $field, $when='current') {
		if(! $field) return;
		if(is_string($range)) $range = $this->parseRange($range);
		$op = array_get_default($range, 'operator');
		$fval = $result->getField($field, null, ($when == 'previous'));
		$negate = ! empty($range['negate']);
		$range['negate'] = false;
		$fdef = $result->getFieldDefinition($field);
		$format = $this->getFieldFormat($fdef);
		if(substr($op, 0, 4) == 'not_') {
			$negate = true;
			$op = substr($op, 4);
		}
		$chk = null;
		if($op == 'empty') {
			$chk = ! $fval;
		} else if($op == 'month') {
			$mo = array_get_default($range, 'value');
			if(preg_match('/^(\d{4})-(\d{1,2})-(\d{1,2}).*/', $fval, $m)) {
				$chk = ($m[2] == $mo);
			}
		} else {
			$rng = $this->getRangeStartEnd($range, null, false, $format=='date', $result, $when);
			if(! $rng) {
				// throw eval error??
			} else {
				if(isset($rng[0]) && $rng[0] > $fval)
					$chk = false;
				else if(isset($rng[1]) && $rng[1] <= $fval)
					$chk = false;
				else
					$chk = true;
			}
		}
		if($negate)
			$chk = ! $chk;
		return $chk;
	}

	function resolveField($field_name, $result, $when='current') {
		if(! $result) throw new IAHDateEvalError("Missing row result for field value");
		$fdef = $result->getFieldDefinition($field_name);
		$format = $this->getFieldFormat($fdef);
		$val = $result->getField($field_name, null, ($when == 'previous'));
		if(! $val) return;
		if($format == 'date')
			return ['operator' => 'date', 'value' => $val];
		else
			return ['operator' => 'datetime', 'value' => $val];
	}

	function resolveDate($range, $base=null, $seconds=false, $result=null, $when='current') {
		global $timedate;
		switch( ($op = $range['operator']) ) {
			case 'date':
				$dt = sscanf($range['value'], '%4d-%2d-%2d', $y, $m, $d);
				if($dt <= 0) $ret = null;
				else $ret = $this->calcDateStartTime($y, $m, $d, 0, true);
				if(! $seconds) $ret = gmdate(self::$db_date_format, $ret);
				return $ret;
			case 'datetime':
				$ret = $range['value'];
				if($seconds) $ret = strtotime($ret . ' UTC');
				if(! $ret) $ret = null;
				return $ret;
			case 'field':
				$ret = $this->resolveField($range['value'], $result, $when);
				if($ret) $ret = $this->resolveDate($ret, null, $seconds);
				return $ret;
			case 'special':
				switch( ($sp = $range['value']) ) {
					case 'now':
						return $this->currentTime(! $seconds);
					case 'today':
					case 'yesterday':
					case 'tomorrow':
						$offs = ($sp == 'yesterday' ? -1 : ($sp == 'tomorrow' ? 1 : 0));
						$ret = $this->relativeDayStartTime($offs, $base, $seconds);
						return $ret;
					case 'nextbus':
						$range = $this->getPeriodStartEnd(1, 'busday', 1, $base, $seconds);
						return $range[0];
				}
				break;
			case 'range':
				return $this->resolveDate($range['start'], $base, $seconds);
			case 'start':
			case 'end':
			case 'before': // not really
			case 'after':
				$range = $this->getPeriodStartEnd(0, $range['period'], 1, $base, $seconds);
				if($op == 'start' || $op == 'before') return $range[0];
				return $range[1];
			case 'before_today':
			case 'after_today':
				$range = $this->getPeriodStartEnd(0, 'tomorrow', 1, $base, $seconds);
				if($op == 'after_today' || $op == 'before_today') return $range[0];
				return $range[1];
			case 'within':
				if(isset($range['value']))
					return $this->resolveDate($range['value'], $base, $seconds);
				if(! $base)
					return $this->currentTime(! $seconds);
				if(! $seconds) $base = gmdate(self::$db_date_format, $base);
				return $base;
			case 'offset':
				if(isset($range['value']))
					$base = $this->resolveDate($range['value'], $base, true);
				$range = $this->getPeriodStartEnd($range['count'], $range['period'], 0, $base, $seconds);
				return $range[0];
			case 'empty':
			case 'month':
			case 'year':
				return null;
		}
	}

	function toLocalDate($datetime) {
		global $timedate;
		$date = $timedate->handle_offset($datetime, self::$db_date_format, true);
		return $timedate->to_display($date, self::$db_date_format, 'Y-m-d');
	}

	// only applies to certain types of ranges
	function getRangeStartEnd($range, $base=null, $seconds=false, $date_only=false, $result=null, $when='current') {
		global $timedate;
		switch( ($op = $range['operator']) ) {
			case 'date':
				$dt = sscanf($range['value'], '%4d-%2d-%2d', $y, $m, $d);
				$start = $this->calcDateStartTime($y, $m, $d, 0, true);
				$ret = $this->getPeriodStartEnd(0, 'day', 1, $start, $seconds);
				break;
			case 'datetime':
				$ret = $range['value'];
				if($seconds) $ret = strtotime($ret . ' UTC');
				$ret = [$ret, $ret+1];
				break;
			case 'field':
				$ret = $this->resolveField($range['value'], $result, $when);
				if($ret) $ret = $this->getRangeStartEnd($ret, null, $seconds); // ignore date_only: adjusted later
				break;
			case 'special':
				switch( ($sp = $range['value']) ) {
					case 'now':
						$start = $this->currentTime();
						if(! $seconds) $start = gmdate(self::$db_date_format, $start);
						$ret = [$start, $start];
						break;
					case 'today':
					case 'yesterday':
					case 'tomorrow':
						$offs = ($sp == 'yesterday' ? -1 : ($sp == 'tomorrow' ? 1 : 0));
						$ret = $this->getPeriodStartEnd($offs, 'day', 1, $base, $seconds);
						break;
					case 'nextbus':
						$ret = $this->getPeriodStartEnd(1, 'busday', 1, $base, $seconds);
						break;
					default:
						throw new IAHDateEvalError("Unsupported offset type: $sp");
				}
				break;
			case 'range':
				$start = $this->resolveDate($range['start'], $base, $seconds, $result, $when);
				$end = $this->getRangeStartEnd($range['end'], $base, $seconds);
				$ret = [$start, $end[1]];
				break;
			case 'before':
			case 'after':
				if(isset($range['value']))
					$base = $this->resolveDate($range['value'], $base, true, $result, $when);
				else if(! $base)
					$base = $this->currentTime();
				$endp = $seconds ? $base : gmdate(self::$db_date_format, $base);
				$ret = ($op == 'before') ? [null, $endp] : [$endp, null];
				break;
			case 'start':
			case 'end':
				$rng = $this->getPeriodStartEnd(0, $range['period'], 1, $base, true);
				if($op == 'start')
					$ret = $this->getPeriodStartEnd(0, 'day', 1, $rng[0], $seconds);
				else
					$ret = $this->getPeriodStartEnd(0, 'day', -1, $rng[1], $seconds);
				break;
			case 'previous':
			case 'next':
			case 'current':
				$cnt = abs(array_get_default($range, 'count', 1));
				$offs = ($op == 'next' ? 1 : ($op == 'previous' ? -1 : 0));
				if(isset($range['value']))
					$base = $this->resolveDate($range['value'], $base, true, $result, $when);
				$ret = $this->getPeriodStartEnd($offs, $range['period'], $cnt, $base, $seconds);
				break;
			case 'within':
				if(isset($range['value']))
					$rng = $this->getRangeStartEnd($range['value'], $base, true);
				else if($base)
					$rng = [$base, $base];
				else
					$rng = $this->getPeriodStartEnd(0, 'day', 1, null, true);
				$cnt = $range['count'];
				$prd = $range['period'];
				if($prd == 'week') {
					$cnt *= 7;
					$prd = 'day';
				}
				else if($prd == 'month') {
					$cnt *= 30;
					$prd = 'day';
				}
				$start = $this->getPeriodStartEnd(0, $prd, -$cnt, $rng[0], $seconds);
				$end = $this->getPeriodStartEnd(0, $prd, $cnt, $rng[1], $seconds);
				$ret = [$start[0], $end[1]];
				break;
			case 'offset':
				if(isset($range['value']))
					$base = $this->resolveDate($range['value'], $base, true, $result, $when);
				$ret = $this->getPeriodStartEnd(0, $range['period'], $range['count'], $base, $seconds);
				break;
			case 'year':
				if(isset($range['value'])) $year = (int)$range['value'];
				else $year = idate('Y');
				$base = $this->calcDateStartTime($year, 1, 1, 0, true);
				$ret = $this->getPeriodStartEnd(0, 'year', 1, $base, $seconds);
				break;
			default:
				return null; // range not defined
		}
		if(! empty($range['negate']))
			$ret[2] = true;
		if(isset($ret[0]) && isset($ret[1]) && $ret[1] < $ret[0]) {
			$a = $ret[1];
			$ret[1] = $ret[0];
			$ret[0] = $a;
		}
		if($date_only && ! $seconds) {
			if($ret[0]) $ret[0] = $this->toLocalDate($ret[0]);
			if($ret[1]) $ret[1] = $this->toLocalDate($ret[1]);
		}
		return $ret;
	}

	function getTimezoneOffset($base=null, $to=true) {
		global $timedate;
		if(! $this->local) return 0;
		if(! $base) $base = time();
		$secs = $timedate->get_seconds_offset($base, $to);
		return $secs;
	}

	function getPeriodStartEnd($period_offset = 0, $period = 'day', $period_count = 1, $base = null, $seconds = false) {
		global $locale;

		if ($period_count < 0) {
			$period_offset += $period_count;
			$period_count *= -1;
		}
		if(! isset($base)) $base = $this->currentTime();
		$back_days = $forw_days = 0;
		switch ($period) {
			case 'second':
				$start_dt = $base - $period_offset;
				$end_dt = $start_dt + $period_count;
				break;
			case 'minute':
				$start_dt = $base - $period_offset * 60;
				$end_dt = $start_dt + $period_count * 60;
				break;
			case 'hour':
				$start_dt = $base - $period_offset * 3600;
				$end_dt = $start_dt + $period_count * 3600;
				break;
			case 'day':
				$start_dt = $this->relativeDayStartTime($period_offset, $base, $seconds);
				$end_dt = $this->relativeDayStartTime($period_offset + $period_count, $base, $seconds);
				break;
			case 'busday':
				// only next business day currently, holidays not taken into account
				$days = 1;
				$dow = (int)gmdate('w', $base + 24*3600); // start next day
				if($dow == 0) $days ++;
				else if($dow == 6) $days += 2;
				$start_dt = $this->relativeDayStartTime($days, $base, $seconds);
				$end_dt = $this->relativeDayStartTime($days + 1, $base, $seconds);
				break;
			case 'week':
				$first_dow = $locale->getWeekStartDay();
				$dow = (int)gmdate('w', $base);
				$back_days = $first_dow - $dow;
				if ($back_days > 0) $back_days -= 7;
				$back_days += $period_offset * 7;
				$forw_days = $back_days + 7 * $period_count;
				$start_dt = $this->relativeDayStartTime($back_days, $base, $seconds);
				$end_dt = $this->relativeDayStartTime($forw_days, $base, $seconds);
				break;
			case 'month':
				list($cur_year, $cur_month) = explode(' ', date('Y n', $base));
				if ($period_offset) {
					// find beginning of another month (using zero-based cur_month)
					$this->adjustMonthYear($cur_month, $cur_year, $period_offset);
				}
				$start_dt = $this->calcDateStartTime($cur_year, $cur_month, 1, 0, $seconds);
				$this->adjustMonthYear($cur_month, $cur_year, $period_count);
				$end_dt = $this->calcDateStartTime($cur_year, $cur_month, 1, 0, $seconds);
				break;
			case 'quarter':
			case 'fiscalquarter':
				list($cur_year, $cur_month) = explode(' ', date('Y n', $base));
				if ($period == 'fiscalquarter') {
					$fy_start = (int) $locale->getFiscalYearStart();
					if ($fy_start) $fy_start--;
					$fiscal_offset = $fy_start % 3;
					$months_offset = ($cur_month - 1) % 3;
					$months_offset = ($months_offset - $fiscal_offset + 3) % 3;
				} else {
					$cur_month = floor(($cur_month - 1) / 3) * 3 + 1;
					$months_offset = 0;
				}
				if ($period_offset) {
					// find beginning of another month (using zero-based cur_month)
					$this->adjustMonthYear($cur_month, $cur_year, $period_offset * 3 - $months_offset);
				}
				$start_dt = $this->calcDateStartTime($cur_year, $cur_month, 1, 0, $seconds);
				$this->adjustMonthYear($cur_month, $cur_year, $period_count * 3);
				$end_dt = $this->calcDateStartTime($cur_year, $cur_month, 1, 0, $seconds);
				break;
			case 'year':
			case 'fiscalyear':
				if ($period == 'fiscalyear') {
					$fy_start = (int) $locale->getFiscalYearStart();
					if ($fy_start) $start_month = $fy_start;
					list($cur_year, $cur_month) = explode(' ', date('Y n', $base));
					if ($cur_month < $start_month)
						$cur_year--; // fiscal year adjustment
				}
				else {
					$cur_year = idate('Y', $base);
					$start_month = 1;
				}
				$cur_year += $period_offset;
				$start_dt = $this->calcDateStartTime($cur_year, $start_month, 1, 0, $seconds);
				$end_dt = $this->calcDateStartTime($cur_year + $period_count, $start_month, 1, 0, $seconds);
				break;
			default:
				throw new IAHDateEvalError("Unsupported period type: $period");
		}
		if(! is_string($start_dt) && ! $seconds) $start_dt = gmdate(self::$db_date_format, $start_dt);
		if(! is_string($end_dt) && ! $seconds) $end_dt = gmdate(self::$db_date_format, $end_dt);
		return [$start_dt, $end_dt];
	}

	function adjustMonthYear(&$month, &$year, $offset) {
		if (!$offset)
			return;
		// calculate with a zero-based month:
		$month--;
		$month += $offset;
		if ($month >= 12) {
			$year += floor($month / 12);
			$month %= 12;
		}
		if ($month < 0) {
			$year -= floor(abs($month) / 12);
			$month %= 12;
			if ($month < 0) $month += 12;
		}
		// return to one-based
		$month++;
	}

	function relativeDayStartTime($day_offset = 0, $base = null, $in_seconds = false) {
		if(! isset($base)) $base = $this->currentTime();
		$base_info = getdate($base);
		return $this->calcDateStartTime($base_info['year'], $base_info['mon'], $base_info['mday'], $day_offset, $in_seconds);
	}

	function calcDateStartTime($year, $month, $day, $day_offset = 0, $in_seconds = false) {
		global $timedate;
		$base_midnight = gmmktime(0, 0, 0, $month, $day, $year);
		$start = $base_midnight + $this->getTimezoneOffset($base_midnight, false);
		if ($day_offset) $start += $day_offset * 24 * 3600;
		if ($in_seconds) return $start;
		return gmdate(self::$db_date_format, $start);
	}

	function currentTime($fmt=false) {
		global $timedate;
		if($this->local) $ret = $timedate->user_local_time();
		else $ret = time();
		if($fmt) $ret = gmdate(self::$db_date_format, $ret);
		return $ret;
	}

	function defaultBase() {
		return ['operator' => 'special', 'value' => 'now'];
	}

	function formatPeriod($period, $count=null, $embed=false) {
		static $ops = ['today', 'tomorrow', 'yesterday', 'now', 'after_today', 'before_today'];
		if(in_array($period, $ops)) {
			$ret = translate_dd('relative_time_dom', $period, $period);
			if($embed) {
				$ret = strtolower($ret);
				$ret = translate_dd('lower_relative_time_dom', $period, $ret);
			}
			return $ret;
		}
		else if($period == 'nextbus') {
			return $this->formatRange(['operator' => 'next', 'period' => 'busday']);
		}
		$ret = $period.'s';
		$dom = ($count < 2) ? 'interval_unit_dom' : 'interval_units_dom';
		$lc = translate_dd($dom, $ret, $ret);
		if($embed)
			$ret = $lc;
		else
			$ret = translate_dd('upper_'.$dom, $ret, ucfirst($lc));
		if(strlen($count)) $ret = $count . ' ' . $ret;
		return $ret;
	}

	function opLabel($op, $embed=false) {
		$test = [];
		if($embed) {
			$test[] = ['date_range_embed_dom', $op.'_pfx'];
			$test[] = ['date_range_embed_dom', $op];
		}
		$test[] = ['date_range_operators_dom', $op.'_pfx'];
		$test[] = ['date_range_operators_dom', $op];
		foreach($test as $t) {
			$ret = translate_dd($t[0], $t[1], '');
			if($ret) return $ret;
		}
		return $op;
	}

	// natural language
	function formatRange($range, $embed=false) {
		global $timedate;
		$negate = ! empty($range['negate']);
		switch( ($op = $range['operator']) ) {
			case 'any':
			case 'empty':
				if($negate) $op = 'not_'.$op;
				return $this->opLabel($op, $embed);
			case 'before':
			case 'after':
				$ret = $negate ? $this->opLabel('not', $embed) . ' ' : '';
				$ret .= $this->opLabel($op, $embed || $negate);
				$val = isset($range['value']) ? $range['value'] : $this->defaultBase();
				$ret .= ' ' . $this->formatRange($val, true);
				return $ret;
			case 'start':
			case 'end':
				if(isset($range['value'])) {
					$pop = $this->opLabel($op, true);
					$ret = $this->formatRange($range['value'], $embed);
					$ret .= ', ' . $pop . ' ' . $this->formatPeriod($range['period'], null, true);
				} else {
					$pop = $this->opLabel($op, $embed);
					$ret = $pop . ' ' . $this->formatPeriod($range['period'], null, true);
				}
				return $ret;
			case 'special':
				$ret = $this->formatPeriod($range['value'], null, $embed);
				return $ret;
			case 'date':
				$ret = $timedate->to_display_date($range['value'], false);
				if ($negate)
					$ret = $this->opLabel('not', $embed) . ' ' . $ret;
				return $ret;
			case 'datetime':
				$ret = $timedate->to_display_date_time($range['value']);
				if ($negate)
					$ret = $this->opLabel('not', $embed) . ' ' . $ret;
				return $ret;
			case 'field':
				$ret = $this->translateFieldLabel($range['value']);
				if ($negate)
					$ret = $this->opLabel('not', $embed) . ' ' . $ret;
				return $ret;
			case 'within':
				$ret = $negate ? $this->opLabel('not', $embed) . ' ' : '';
				$ret .= $this->opLabel($op, $embed || $negate) . ' ';
				if($embed) $ret = strtolower($ret);
				$c = empty($range['count']) ? 1 : abs($range['count']);
				$ret .= $this->formatPeriod($range['period'], $c, true);
				if(isset($range['value'])) {
					$of = $this->opLabel('of', true);
					$ret .= ' ' . $of . ' ' . $this->formatRange($range['value'], true);
				}
				return $ret;
			case 'offset':
				$c = empty($range['count']) ? 1 : $range['count'];
				if(isset($range['value'])) {
					$ret = $this->formatRange($range['value'], $embed);
					$pop = $this->opLabel($c > 0 ? 'plus' : 'minus', true);
					$ret .= ' ' . $pop . ' ' . $this->formatPeriod($range['period'], abs($c), true);
				} else {
					$ret = $this->opLabel($c > 0 ? 'plus' : 'minus', $embed);
					$ret .= ' ' . $this->formatPeriod($range['period'], abs($c), true);
				}
				return $ret;
			case 'previous':
			case 'next':
				$c = empty($range['count']) ? null : abs($range['count']);
				$pfx = $this->opLabel($op, $embed);
				$ret = $pfx.' '.$this->formatPeriod($range['period'], $c, true);
				if(isset($range['value'])) {
					$pfx = $this->opLabel('from', true);
					$ret .= ' ' . $pfx . ' ' . $this->formatRange($range['value'], true);
				}
				return $ret;
			case 'current':
				if(isset($range['value'])) {
					$pfx = $negate ? $this->opLabel('not', $embed) . ' ' : '';
					$pfx .= $this->opLabel('within', $embed || $negate) . ' ';
					$top = $this->opLabel('current', true);
					$ret = $pfx . $this->formatPeriod($range['period'], $top, true);
					$of = $this->opLabel('of', true);
					$ret .= ' ' . $of . ' ' . $this->formatRange($range['value'], true);
				} else {
					$pfx = $this->opLabel('this', $embed);
					$ret = $pfx.' '.$this->formatPeriod($range['period'], null, true);
				}
				return $ret;
			case 'range':
				$pfx = $negate ? $this->opLabel('not', $embed) . ' ' : '';
				$pfx .= $negate ? $this->opLabel('from', $embed || $negate) . ' ' : '';
				$to = ' ' . $this->opLabel('to', true) . ' ';
				$ret = $pfx . $this->formatRange($range['start'], $negate) . $to . $this->formatRange($range['end'], true);
				return $ret;
			case 'month':
				$m = translate_dd('months_long_dom', $range['value']);
				return ucfirst(translate_dd('interval_unit_dom', $op)) . ': ' . $m;
			case 'year':
			case 'fiscalyear':
				return ucfirst(translate_dd('interval_unit_dom', $op)) . ': ' . $range['value'];
		}
	}

	function translateFieldLabel($field) {
		if(is_string($field))
			$fdef = $this->getFieldDefinition($field);
		else if(is_array($field))
			$fdef = $field;
		if(! $fdef) return $field;
		$vmod = array_get_default(
			$fdef,
			'vname_module',
			$this->model ? $this->model->getModuleDir() : ''
		);
		$label = translate($fdef['vname'], $vmod);
		return $label;
	}

	// convert from tree structure back to string
	function encodeRange($range) {
		$negate = ! empty($range['negate']);
		$op = $range['operator'];
		if(substr($op, 0, 4) == 'not_') {
			$op = substr($op, 4);
			$negate = ! $negate;
		}
		switch($op) {
			case '':
			case 'any':
				return 'any';
			case 'empty':
				return ($negate ? 'not_' : '') . $op;
			case 'before':
			case 'after':
				if($negate) $op = 'not_'.$op;
				if(isset($range['value']))
					return $this->encodeRange($range['value']) . '/' . $op;
				return $op;
			case 'start':
			case 'end':
				$ret = $op . ':' . $range['period'];
				if(isset($range['value']))
					$ret .= $this->encodeRange($range['value']);
				return $ret;
			case 'special':
				if(empty($range['value']))
					throw new IAHDateEvalError("Missing special value");
				if(! in_array($range['value'], ['now','today','tomorrow','yesterday','nextbus','after_today','before_today']))
					throw new IAHDateEvalError("Unrecognized special value");
			case 'on_date':
			case 'date':
			case 'datetime':
				if(empty($range['value']))
					throw new IAHDateEvalError("Missing date");
				$ret = $range['value'];
				if($negate) $ret = "not:$ret";
				return $ret;
			case 'field':
				return 'field:'.$range['value'];
			case 'within':
			case 'offset':
			case 'previous':
			case 'next':
			case 'current':
				$ret = '';
				if(empty($range['count'])) $range['count'] = 1;
				$ret = ($range['count'] < 0 || $op == 'within' ? $range['count'] : '+'.$range['count']);
				$pd = $range['period'];
				$idx = array_search($pd, self::$offset_intervals);
				if($idx !== false) $pd = $idx;
				$ret .= $pd;
				if($op == 'within') $ret = ($negate ? 'not_' : '') . $op . ':' . $ret;
				else if($op != 'offset') $ret = $op . ':' . $ret;
				if(isset($range['value']))
					$ret = $this->encodeRange($range['value']) . '/' . $ret;
				return $ret;
			case 'range':
				$ret = $this->encodeRange($range['start']) . ($negate ? '!!' : '||') . $this->encodeRange($range['end']);
				return $ret;
			case 'month':
			case 'year':
			case 'fiscalyear':
				if(empty($range['value']))
					throw new IAHDateEvalError("Missing month/year value");
				$ret = $op . ':' . $range['value'];
				return $ret;
			default:
				throw new IAHDateEvalError("Unrecognized date range: $op");
		}
	}


	function test() {
		$ts = [
			'today',
			'now',
			'2016-10-15',
			'2012-01-01/before',
			'not_after:2012-01-01',
			'before',
			'2012-01-01 01:01',
			'2012-01-01 01:01||now',
			'today!!yesterday',
			'within:1d',
			'within:1w',
			'within:1m',
			'within:1y',
			'yesterday/within:2weeks',
			'offset:-2w',
			'today-7d',
			'-1day',
			'now+-2i',
			'tomorrow+1w/before',
			'field:date_start-2w',
			'not_empty',
			'nextbus',
			'start:week',
			'end: month',
			'month:1',
			'year:1979',
			'current:week',
			'2011-01-01/next:+2m',
			'2w||end:year',
		];

		$result = new RowResult();
		$result->base_model = 'Task';
		$result->assign('date_start', '2011-01-15 15:01');
		foreach($ts as $t) {
			try {
				$v = $this->parseRange($t);
				$f = $this->formatRange($v);
				$r = json_encode($this->getRangeStartEnd($v, null, false, false, $result));
				pr2("Range: " . $r . "\nParsed: " . json_encode($v) . "\nRe-encoded: ".$this->encodeRange($v), $f . ' -- ' . $t);
			} catch(IAHDateParseError $e) {
				pr2("Parse error: {$e->getMessage()}", $t);
			}
		}
	}
}

if(! empty($_SERVER['PHP_SELF']) && substr($_SERVER['PHP_SELF'], -13) == 'DateRange.php') {
	chdir('../..');
	define('sugarEntry', 1);
	require_once('include/entryPoint.php');
	$current_user = new User();
	$current_user->retrieve(1);
	$dt = new IAHDateRange();
	$dt->test();
}

