<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo,
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */


class IAHListQueryError extends IAHQueryError {
	var $error_name = 'List Query Error';
}

require_once('include/config/ModelDef.php');
require_once('include/ListView/ListFilter.php');
require_once('include/database/ListResult.php');
require_once('include/database/RowResult.php');
require_once('include/database/ListQueryIterator.php');
define('LQ_BATCH_REQUEST_SIZE', 100);

class ListQuery implements IteratorAggregate {
	/** @var $base_model ModelDef */
	var $base_model;
	var $fields;
	var $secondary_calculated;
	var $totals;
	var $links;
	var $order_by;
	var $group_by;
	var $parent_key;
	var $primary_added;
	var $added_field_names;
	var $join_index;
	var $join_link = [];
	/** @var ListFilter */
	var $list_filter;
	var $filter_deleted = true;
	var $filter_deleted_secondary = true;
	var $run_secondary_queries = true;
	var $process_results = true;
	var $no_default_order_by = false;
	var $disable_db_cache = false;
	var $disable_join_filter = false;
	var $disable_literal_types = false;
	var $default_use_key_index = true;
	var $auto_add_fields = true;
	var $join_favorites = false;
	var $join_personal = true;
	var $is_union, $union_queries, $is_union_child;
	var $use_real_names;
	var $skip_acl_groups_filter = false;
	var $acl_checks;
	var $acl_filters;
	var $acl_user_id;
	var $acl_admin_except = true;
	var $acl_owner_field;
	var $acl_query_teams;
	var $fy_start_month;
	var $summation_groups;
	var $formatted;
	var $formatted_context;

	var $secondary_queries;
	var $field_warnings;
	var $unified_search_fields;
    var $fulltext_search;
	var $where;
	var $timezone;
	var $unbuffered = false;

	static $settable_params = array(
		'filter_deleted', 'acl_user_id',
		'run_secondary_queries', 'process_results', 'formatted', 'formatted_context',
		'no_default_order_by', 'is_union', 'use_real_names',
		'where', 'order_by', 'auto_add_fields', 'unbuffered', 'skip_acl_groups_filter',
	);

	static $const_cols = array('module_name', 'bean_name', 'source_name');

	static $valid_totals = array(
		'id' => array('count'),
		'name' => array('count'),
		'int' => array('sum', 'avg', 'min', 'max', 'stddev'),
		'float' => array('sum', 'avg', 'min', 'max', 'stddev'),
		'double' => array('sum', 'avg', 'min', 'max', 'stddev'),
		'currency' => array('sum', 'avg', 'min', 'max'),
		'base_currency' => array('sum', 'avg', 'min', 'max'),
		'raw_currency' => array('sum', 'avg', 'min', 'max'),
		'date' => array('min', 'max'),
		'datetime' => array('min', 'max', 'count'),
		'duration' => array('sum', 'avg', 'min', 'max'),
		'bool' => array('count'),
		'star_rating' => array('min', 'max', 'avg', 'stddev'),
	);

	static $numeric_types = array(
		'bool', 'int', 'tinyint', 'short', 'float', 'double', 'currency', 'base_currency', 'raw_currency',
		'month', 'year', 'percentage', 'star_rating',
	);

	static $ext_search_types = array(
		'name', 'varchar', 'text',
	);


	function __construct($base_model = null, $fields = null, $params = null) {
		$this->init($base_model, $fields, $params);
		$this->loadUserTimeZone();
	}

	// setup
	function init($base_model, $fields = null, $params = null) {
		if (!is_array($params))
			$params = array();

		$alias = array_get_default($params, 'alias');
		if (is_string($base_model))
			$base_model = new ModelDef($base_model, null, $alias);
		else if (isset($alias))
			$base_model->setTableAlias($alias);

		$this->links = array();
		$this->join_index = 0;

		$link_name = array_get_default($params, 'link_name');
		if ($link_name) {
			$src_name = array_get_default($params, 'source_name');
			$this->links['~parent'] = $base_model;
			$new_base = $base_model->getLinkTargetModel($link_name, $src_name, $src_name);
			if(! $new_base)
				throw new IAHListQueryError("Unknown linked field: {$base_model->name}.{$link_name} $src_name");
			if (($join = $new_base->getJoinModel())) {
				$this->_addJoinLink('', '~parent', $join);
			}
			$this->base_model = $new_base;
		} else
			$this->base_model = $base_model;

		if (!empty($params['plain'])) {
			$this->secondary_queries = false;
			$this->process_results = false;
			$this->formatted = null;
			$this->no_default_order_by = true;
		}
		foreach (self::$settable_params as $f)
			if (isset($params[$f]))
				$this->$f = $params[$f];
		if(isset($params['use_key_index']))
			$this->default_use_key_index = $params['use_key_index'];

		$this->clearFields();
		if($this->base_model && $this->base_model->is_union)
			$this->initUnionModel();
		if (isset($fields)) {
			if($fields === true)
				$this->addColumnFields();
			else if($fields)
				$this->addFields($fields);
		}
		if(isset($params['acl_checks']))
			$this->addAclChecks($params['acl_checks']);
		if(isset($params['acl_filters']))
			$this->addAclFilters($params['acl_filters']);
		if(isset($params['filters']))
			$this->addFilterClauses($params['filters']);
		if(isset($params['parent_key']))
			$this->setParentKey($params['parent_key']);
	}


	function initUnionModel() {
		$this->is_union = true;
		$sources = $this->base_model->getSources();
		$lspec = $this->base_model->getLinkSpec();
		if($sources) {
			foreach($sources as $src_name => $src) {
				$mname = $src['model'];
				$params = array();
				if($lspec && isset($lspec['relationship_spec']['sources'])) {
					$rel_srcs = $lspec['relationship_spec']['sources'];
					if(! isset($rel_srcs[$src_name]))
						continue;
					$params['link_name'] = $lspec['name'];
					$params['source_name'] = $src_name;
					$mname = $this->links['~parent']->name;
				} else
					$rel_srcs = null;
				$q = new ListQuery($mname, null, $params);
				if(isset($src['filters'])) {
					$q->addFilterClauses($src['filters']);
				}
				if($rel_srcs && isset($rel_srcs[$src_name]['filters'])) {
					$q->addFilterClauses($rel_srcs[$src_name]['filters']);
				}
				$this->addUnionQuery($q, $src_name);
			}
		}
		$this->addSourceNameField();
	}


	function setTableAlias($alias, $link=null) {
		if(isset($link)) {
			if(isset($this->links[$link]))
				$this->links[$link]->setTableAlias($alias);
		} else if($this->base_model)
			$this->base_model->setTableAlias($alias);
	}


	function clearFields($keep_meta=false) {
		$orig_fs = $this->fields;
		$this->fields = array();
		$this->secondary_calculated = array();
		$this->totals = array();
		$this->primary_added = false;
		$this->added_field_names = array();
		$this->acl_checks = array();
		if($keep_meta) {
			if(isset($orig_fs['query_source']))
				$this->fields['query_source'] = $orig_fs['query_source'];
		}
	}


    /**
     * @param $field            : a fieldspec. see addFields for more details
     * @param null $alias       : ie SQL 'as' for SELECT/JOIN/WHERE, template name for ORDER/GROUP
     * @param bool $replace
     * @param bool $no_query
     * @param bool $allow_dupe
     * @return mixed            : the added (or existing ) field
     */
    function addField($field, $alias = null, $replace = false, $no_query = false, $allow_dupe = false) {
		if (!isset($alias)) $alias = 0;
		$added = $this->addFields(array($alias => $field), $replace, $no_query, $allow_dupe);
		if ($added)
			return $added[0];
	}

    /**
     * @param $fields           : [ alias=>fieldspec, ... ]  see addField*() sources for fieldspec structures
     * @param bool $replace     : true if you want to replace an existing field otherwise it will return the existing one
     * @param bool $no_query    : true if you don't want it in the result set
     * @param bool $allow_dupe  : allow a no_query alias to be used/expanded even if the field name collided
     * @return array            :
     */
    function addFields($fields, $replace = false, $no_query = false, $allow_dupe = false) {
		$warn =& $this->field_warnings;
		$added = array();
		$addLater = array();
		$check_links = $check_refs = $check_currency = $check_duration = $check_sourced = $check_names = $check_image = array();

		foreach ($fields as $alias => $spec) {
			if (is_string($spec)) {
				if (is_int($alias))
					$alias = $spec;
				$spec = array('name' => $spec);
			} else {
				if (!isset($spec['name'])) {
					$warn[] = array('missing field name', null, $alias);
					$added[] = null;
					continue;
				}
				if (is_int($alias)) {
					$alias = $this->makeFieldId($spec, true);
				}
			}

			// return existing field if already registered
			if (! $replace) {
				if (isset($this->fields[$alias]))
					$added_as = $alias;
				else if (isset($this->added_field_names[$alias]) && $this->added_field_names[$alias] != '_display' && $alias != '_display')
					$added_as = $this->added_field_names[$alias];
				else
					$added_as = null;
				if ($added_as) {
					$prev_no_q = array_get_default($this->fields[$added_as], 'no_query', false);
					if (!$prev_no_q || $no_query) {
						if($allow_dupe && $alias != $added_as)
							$this->fields[$alias] = $this->fields[$added_as];
						$added[] = $alias;
						continue;
					}
				}
			}

			if($this->is_union) {
				$ufail = false;
				$utype = null;
				$uvname = null;
				$const_col = false;
				if(isset($spec['source']) && in_array($spec['source']['type'], self::$const_cols)) {
					$field_def = $spec;
					$const_col = true;
				} else if($this->base_model) {
					if($spec['name'] == '_display' || $spec['name'] == '_track_date')
						$field_def = $spec;
					else
						$field_def = $this->getFieldDefinition($spec, $errmsg);
					if(! $field_def) {
						$warn[] = array('union field not defined', null, $spec['name']);
						$added[] = null;
						continue;
					}
				} else {
					$field_def = $spec;
				}
				if(! isset($field_def['source']))
					$field_def['source']['type'] = 'db';
				$usrc = array();
				foreach($this->union_queries as $qid => &$q) {
					if(isset($field_def['union_source'][$qid]))
						$qspec = $field_def['union_source'][$qid];
					else
						$qspec = ['name' => $field_def['name'], 'type' => array_get_default($field_def, 'type', 'varchar')];
					if(! is_array($qspec))
						$qspec = array('name' => $qspec);
					if(isset($field_def['source']) && $field_def['source']['type'] != 'db' && empty($qspec['source']))
						$qspec['source'] = $field_def['source'];
					if(isset($field_def['dynamic_module']) && empty($qspec['dynamic_module']))
						$qspec['dynamic_module'] = $field_def['dynamic_module'];
					if (isset($field_def['type']) && $field_def['type'] == 'widget' && empty($qspec['type']))
						$qspec['type'] = 'widget';
					if(isset($field_def['id_name']))
						$qspec['id_alias'] = $field_def['id_name'];
					if(! ($u_added = $q->addField($qspec, $alias, $replace, ($no_query && $no_query != 'parent'), $allow_dupe))) {
						$ufail = true;
					}
					else {
						$added_spec = $q->getField($u_added);
						$usrc[$qid] = $added_spec;
						if(! isset($utype)) {
							$utype = array_get_default($added_spec, 'type');
							$uvname = array_get_default($added_spec, 'vname');
							$uvname_mod = array_get_default($added_spec, 'vname_module');
						}
					}
				}
				if(! $ufail) {
					$field_def['union_source'] = $usrc;
					if(! isset($field_def['type'])) $field_def['type'] = $utype;
					if(! isset($field_def['vname'])) {
						$field_def['vname'] = $uvname;
						$field_def['vname_module'] = $uvname_mod;
					}
					if(! ($u_added = $this->_addUnionField($field_def, $alias, !! $no_query)) ) {
						$ufail = true;
					}
				}
				if($ufail) {
					$warn[] = array('missing field for one or more union sources', null, $alias);
					$added[] = null;
					continue;
				}
				$added[] = $u_added;
			}
			else {
				$src = array_get_default($spec, 'source');
				if ($src && in_array($src['type'], self::$const_cols)) {
					if($src['type'] == 'module_name')
						$cval = $this->base_model->getModuleDir();
					else if($src['type'] == 'bean_name')
						$cval = $this->base_model->getModelName();
					else if($src['type'] == 'source_name')
						$cval = $this->base_model->getTableAlias(true);
					$field_def = $spec;
					$field_def['source'] = array(
						'type' => 'static',
						'value' => $cval,
					);
					if(! isset($field_def['type']))
						$field_def['type'] = $src['type'];
				} elseif ($src && $src['type'] == 'literal') {
                    if ($this->disable_literal_types) throw new IAHListQueryError('Literal types forbidden');
					$field_def = $spec;
					if(! array_key_exists('value', $src))
						$field_def['source']['value'] = null;
				} else {
					$field_def = $this->getFieldDefinition($spec, $errmsg);
					if (!$field_def) {
						if (($dp = strrpos($spec['name'], '.')) !== false) {
							$link_name = substr($spec['name'], 0, $dp);
							$link_member = substr($spec['name'], $dp + 1);
							if(! empty($spec['link']) && $spec['link'] != $link_name)
								$spec['link'] .= '.'.$link_name;
							else
								$spec['link'] = $link_name;
							$spec['name'] = $link_member;
						} else
							$spec['link'] = null;

						if($spec['name'] == '_display') {
							$added[] = $this->addDisplayName($alias, $spec);
							continue;
						}
						else if($spec['name'] == '_account') {
							$added[] = $this->addPrimaryAccount($alias);
							continue;
						}
						else if($spec['name'] == '_account_id') {
							if(empty($this->added_field_names[$alias])) {
								$main_f = preg_replace('~_id$~', '', $alias);
								if($main_f != $alias) {
									$added[] = $this->addPrimaryAccount($main_f);
									continue;
								}
							}
						}
						else if($spec['name'] == '_status' && ! isset($spec['source'])) {
							$added[] = $this->addStatusField($alias, $spec['link']);
							continue;
						}

						$warn[] = $errmsg;
						$added[] = null;
						continue;
					}
				}

				if ($field_def['type'] == 'ref') {
					$this->preCheckRefField($field_def);
				}

				if (!$this->primary_added && empty($field_def['link'])) {
					if (!isset($pK)) $pK = $this->base_model->getPrimaryKey();
					if ($field_def['name'] == $pK)
						$this->primary_added = $alias;
				}

				if ($no_query)
					$field_def['no_query'] = true;
				$this->fields[$alias] = $field_def;
				$fid = $this->makeFieldId($field_def, true);
				$this->added_field_names[$fid] = $alias;

				// quick fix to prevent linking to ref tables when only the ID is needed (mainly assigned_user)
				if ($field_def['type'] == 'id' && !empty($field_def['inferred'])) {
					$link = array_get_default($field_def, 'link', '');
					if($link) $link .= '.';
					$this->added_field_names[$link . array_get_default($field_def, 'for_ref', '') . '.id'] = $alias;
				}

				$added[] = $alias;
			}

			if ($field_def['type'] == 'ref')
				$check_refs[] = $alias;
			else if ($field_def['type'] == 'name')
				$check_names[] = $alias;
			else if ($field_def['type'] == 'currency' || $field_def['type'] == 'raw_currency' || $field_def['type'] == 'base_currency')
				$check_currency[] = $alias;
			else if ($field_def['type'] == 'duration')
				$check_duration[] = $alias;
			else if ($field_def['type'] == 'image')
				$check_image[] = $alias;
			if ($field_def['source']['type'] != 'db' || isset($field_def['source']['fields']))
				$check_sourced[] = $alias;
            if (isset($field_def['file_mime_type']))
				$addLater[] = $field_def['file_mime_type'];
			if (!empty($field_def['link']))
				$check_links[] = $alias;
            if (isset($field_def['source']['calc_formula']) && !isset($field_def['link'])) {
				$addLater = array_merge($addLater, $this->getFieldsFromFormula($field_def['source']['calc_formula']));
			} elseif (isset($field_def['source']['calc_formula'])) {
				// calculated field from a linked record - remember it for populateSecondary()
				$this->secondary_calculated[$alias] = true;
			}
		}

		foreach ($check_sourced as $alias) {
			$this->checkSourcedField($alias);
		}
		foreach ($check_refs as $alias) {
			$this->checkRefField($alias);
		}
		foreach ($check_names as $alias) {
			$this->checkNameField($alias);
		}

		if(!$this->is_union_child) {
			foreach ($check_links as $alias) {
				$this->checkLinkField($alias);
			}
		}
		foreach ($check_currency as $alias) {
			$this->checkCurrencyField($alias);
		}
		foreach ($check_duration as $alias) {
			$this->checkDurationField($alias);
		}
		foreach ($check_image as $alias) {
			$this->checkImageField($alias);
		}

		if((! $this->is_union || $this->base_model) && ! $no_query && $this->auto_add_fields) {
			$this->addPrimaryKey();
		}

		if (!empty($addLater)) {
			$this->addFields($addLater);
		}

		return $added;
	}

	function addFieldLiteral($name, $sql, $type='varchar') {
        if ($this->disable_literal_types) throw new IAHListQueryError('Literal types forbidden');
		$spec = array(
			'name' => $name,
			'type' => $type,
			'source' => array(
				'type' => 'literal',
				'value' => $sql,
			),
		);
		return $this->addField($spec);
	}

	function addFieldStatic($name, $value, $type='varchar') {
		$spec = array(
			'name' => $name,
			'type' => $type,
			'source' => array(
				'type' => 'static',
				'value' => $value,
			),
		);
		return $this->addField($spec);
	}

	function addColumnFields() {
		if(! $this->base_model) return false;
		$fs = array_merge(
			$this->base_model->getColumnFields(),
			$this->base_model->getCustomFields()
		);
		return $this->addFields($fs);
	}

    function getFieldsFromFormula($formula) {
        require_once('include/layout/FormulaParser.php');
        $parser = new FormulaParser($formula);
        $ret = $parser->getExpressionFields();
        return $ret;
    }

	function addTotal($total, $alias = null, $replace = false) {
		if (!isset($alias)) $alias = 0;
		$added = $this->addTotals(array($alias => $total), $replace);
		if ($added)
			return $added[0];
	}

	function addTotals($totals, $replace = false) {
		$warn =& $this->field_warnings;
		$added = array();
		$ntotals = array();

		foreach ($totals as $alias => $spec) {
			if (is_string($spec)) {
				if (is_int($alias))
					$alias = $spec;
				$spec = array('name' => $spec);
			} else {
				if (!isset($spec['name'])) {
					$warn[] = array('missing field name for total', null, $alias);
					$added[] = null;
					continue;
				}
				if (is_int($alias))
					$alias = array_get_default($spec, 'alias', $spec['name']);
			}

			$added_as = $this->addField($spec, false, true);
			if(! $added_as) {
				$added[] = null;
				continue;
			}
			$fdef = $this->fields[$added_as];
			if(empty($fdef['type']) || empty(self::$valid_totals[$fdef['type']])) {
				$added[] = null;
				continue;
			}
			if(! isset($spec['total']))
				$spec['total'] = 'sum';
			if(! in_array($spec['total'], self::$valid_totals[$fdef['type']])) {
				$warn = array('unsupported total type: '.$spec['total'], null, $alias);
				$added[] = null;
				continue;
			}

			/*
			$fmt = isset($def['format']) ? '!'.$def['format'] : '';
			$alias = $spec['name'] . $fmt . ':' . $spec['total']; // ignoring custom value
			$this->totals[$alias] = $def;
			 */

			$def = array_merge($fdef, $spec);
			$this->totals[$added_as] = $def;
			$added[] = $added_as;
		}

		return $added;
	}


	function _addUnionField($def, $alias=null, $no_query=false) {
		if(is_string($def))
			$def = array('name' => $def);
		if(empty($def['source']) || in_array($def['source']['type'], self::$const_cols))
			$def['source']['type'] = 'db';
		if($no_query)
			$def['no_query'] = true;
		if(isset($def['type']) && $def['type'] == 'ref') {
			$this->preCheckRefField($def);
		}
		if(isset($alias) && $alias != $def['name']) {
			$this->added_field_names[$def['name']] = $alias;
			$def['name'] = $alias;
		} else
			$alias = $def['name'];
		$this->fields[$alias] = $def;
		return $alias;
	}

	function addPrimaryKey($alias = null, $link = null) {
		if($link) {
			$ret = false;
			if(isset($this->links[$link])) {
				$pK = $this->links[$link]->getPrimaryKey();
				if($pK) {
					if (!is_array($pK))
						$ret = $this->addField($link.'.'.$pK, $alias);
					else {
						$added = array();
						foreach($pK as $ppk) {
							$pka = $this->addField($link.'.'.$ppk);
							if($pka) $added[] = $pka;
						}
						$ret = $added;
					}
				}
			}
			return $ret;
		}
		else if (!$this->primary_added) {
			if($this->is_union) {
				if(! isset($alias)) $alias = 'id';
				$usrc = array();
				foreach($this->union_queries as $uk => &$u) {
					if(! ($idk = $u->addPrimaryKey($alias)) ) {
						$fail = true;
						continue;
					}
					$usrc[$uk] = $u->getField($idk);
				}
				if(empty($fail)) {
					$fdef = array('vname' => 'LBL_ID', 'name' => 'id', 'type' => 'id', 'source' => array('type' => 'db'));
					$fdef['union_source'] = $usrc;
					$this->primary_added = $this->_addUnionField($fdef, $alias);
				}
			} else if($this->base_model) {
				$pK = $this->base_model->getPrimaryKey();
				$this->primary_added = $pK;
				if($pK) {
					if (!is_array($pK))
						$this->primary_added = $this->addField($pK, $alias);
					else {
						$added = array();
						foreach($pK as $ppk) {
							$pka = $this->addField($ppk);
							if($pka) $added[] = $pka;
						}
						$this->primary_added = $added;
					}
				}
			}
		}
		return $this->primary_added;
	}

	function getDisplayNameDefinition(&$errmsg) {
		if($this->is_union) return null; // not handled currently
		$dname = $this->getModelDisplayName();
		if ($dname) {
			if (array_get_default($dname, 'type') == 'field' && !empty($dname['field'])) {
				$def = $this->getFieldDefinition($dname['field'], $errmsg);
				if($def && ! empty($def['source']) && is_array($def['source']) && $def['source']['type'] == 'user_display_name') {
					$def['source']['type'] = 'field';
					if($this->checkUseRealNames())
						$def['source']['field'] = 'full_name';
					else {
						$def['source']['field'] = 'user_name';
						if($this->is_union_child)
							// make sure fields are added to the query (for union query with contacts/leads)
							$def['source']['fields'] = ['first_name', 'last_name', 'salutation'];
					}
				}
			} else {
				if(! empty($dname['source']['type']) && $dname['source']['type'] == 'non-db') {
					$dname['name'] = '_display';
					if(! isset($dname['type'])) $dname['type'] = 'varchar';
					return $dname;
				}
				$def = array(
					'name' => '_display',
					'type' => 'name',
					'detail_link' => 1,
					'source' => $dname,
				);
			}
			return $def;
		}
	}

	function addDisplayName($alias = null, $ext_spec=null) {
		if($this->is_union) {
			if($this->getModuleDir() == 'Invitees') {
				// quick hack - auto-added fields mess up the field order for unions
				$this->addFields(array('first_name', 'last_name', 'salutation'));
			}

			if(! isset($alias)) $alias = '_display';
			$usrc = array();
			foreach($this->union_queries as $uk => &$u) {
				if(! ($dnk = $u->addDisplayName($alias))) {
					$fail = true;
					continue;
				}
				$usrc[$uk] = $u->getField($dnk);
			}
			if(empty($fail)) {
				$fdef = array('vname' => 'LBL_DISPLAY_NAME', 'name' => '_display',
					'type' => 'name', 'source' => array('type' => 'db'));
				$fdef['union_source'] = $usrc;
				return $this->_addUnionField($fdef, $alias);
			}
			return false;
		}
		else if(! $this->base_model)
			return false;
		$def = $this->getDisplayNameDefinition($errmsg);
		if ($ext_spec) {
			foreach(array('dynamic_module') as $f)
				if(! empty($ext_spec[$f]) && empty($def[$f])) $def[$f] = $ext_spec[$f];
		}
		if ($def)
			return $this->addField($def, $alias);
	}

	function addPrimaryAccount($alias = null) {
		if(! isset($alias)) $alias = '_account';
		if($this->is_union) {
			$usrc = array();
			foreach($this->union_queries as $uk => &$u) {
				if(! ($dnk = $u->addPrimaryAccount($alias))) {
					$fail = true;
					continue;
				}
				$usrc[$uk] = $u->getField($dnk);
			}
			if(empty($fail)) {
				$fdef = array('vname' => 'LBL_ACCOUNT', 'name' => '_account', 'type' => 'ref', 'source' => array('type' => 'db'));
				$fdef['union_source'] = $usrc;
				return $this->_addUnionField($fdef, $alias);
			}
			return false;
		}
		else if(! $this->base_model)
			return false;
		$dname = $this->getModelPrimaryAccount();
		if ($dname) {
			if($this->addField($dname['id'], $alias . '_id'))
				return $this->addField($dname['field'], $alias);
		}
		$this->addFieldStatic($alias . '_id', '', 'id');
		$this->addFieldStatic($alias, '');
		return true;
	}

	function addStatusField($alias = null, $link_name = null) {
		if(! isset($alias)) $alias = '_status';
		if($this->is_union) {
			$usrc = array();
			foreach($this->union_queries as $uk => &$u) {
				if(! ($dnk = $u->addStatusField($alias, $link_name))) {
					$fail = true;
					continue;
				}
				$usrc[$uk] = $u->getField($dnk);
			}
			if(empty($fail)) {
				$fdef = array('vname' => 'LBL_STATUS', 'name' => '_status', 'type' => 'status', 'source' => array('type' => 'db'));
				$fdef['union_source'] = $usrc;
				return $this->_addUnionField($fdef, $alias);
			}
			return false;
		}
		else if(! $this->base_model)
			return false;
		$desc = $this->getModelStatus($link_name);
		if ($desc) {
			$fdef = array('vname' => 'LBL_STATUS', 'name' => '_status', 'type' => 'status');
			foreach(array('options', 'status_options', 'source') as $k)
				if(isset($desc[$k])) $fdef[$k] = $desc[$k];
			if(isset($fdef['source']['type']) && $fdef['source']['type'] == 'db') {
				$fdef['source']['type'] = 'field';
				$fdef['source']['field'] = $desc['name'];
				if(! empty($desc['link']))
					$fdef['source']['field'] = $desc['link'] .'.'. $fdef['source']['field'];
			}
			return $this->addField($fdef, $alias);
		} else
			return $this->addFieldStatic($alias, '');
	}

	function addModuleNameField($alias = null) {
		$def = array(
			'name' => 'query_module',
			'vname' => 'LBL_MODULE',
			'type' => 'module_name',
			'source' => array(
				'type' => 'module_name',
			),
		);
		return $this->addField($def, $alias);
	}

	function addModelNameField($alias = null) {
		$def = array(
			'name' => 'query_model',
			'vname' => 'LBL_MODEL_NAME',
			'type' => 'bean_name',
			'source' => array(
				'type' => 'bean_name',
			),
		);
		return $this->addField($def, $alias);
	}

	function addSourceNameField($alias = null) {
		$def = array(
			'name' => 'query_source',
			'vname' => 'LBL_SOURCE_NAME',
			'type' => 'source_name',
			'source' => array(
				'type' => 'source_name',
			),
		);
		return $this->addField($def, $alias);
	}

	function preCheckRefField(&$field_def) {
		if (isset($field_def['link']))
			$parent = $field_def['link'] . '.';
		else
			$parent = '';
		if (!empty($field_def['bean_name'])) {
			$bean = $field_def['bean_name'];
			$field_def['module_dir'] = AppConfig::module_for_model($bean);
			$src_type = $field_def['source']['type'];
			if ($src_type == 'non-db' || $src_type == 'custom_field') {
				// return name of referenced object, using method described in bean model
				$dname = $this->getModelDisplayName($bean, $field_def['name'], $field_def);
				if ($dname) {
					if($src_type == 'custom_field' && isset($dname['fields']))
						foreach($dname['fields'] as $k => $v)
							$dname['fields'][$k] = '_cstm.'.$v;
					elseif($src_type == 'custom_field' && isset($dname['field']))
						$dname['field'] = '_cstm.'.$dname['field'];
					$field_def['source'] = $dname;
					ModelManager::fix_field_source($field_def);
					if (AppConfig::db_is_auto_fetched($bean) || $bean == 'User')
						$field_def['source']['prefetched'] = true;
					else if($this->is_union)
						$field_def['source'] = array('type' => 'db');
				}
			}
		}
		if (!empty($field_def['dynamic_module']) && $this->auto_add_fields) {
			$type_f = $parent . $field_def['dynamic_module'];
			$field_def['dynamic_module'] = $parent . $field_def['dynamic_module'];
			$this->addField($type_f);
		}
	}

	function getModelDisplayName($model = null, $link_name = null, $spec = null) {
		if (!isset($model))
			$model = $this->base_model->name;
		if (isset($link_name))
			$link_name .= '.';
		else
			$link_name = '';
		if(isset($spec) && isset($spec['ref_display_name']))
			$display = $spec['ref_display_name'];
		if(empty($display) || $display === '_display')
			$display = AppConfig::setting("model.detail.$model.display_name");
		$fname = 'name';
		if (is_array($display)) {
			if (isset($display['field'])) {
				$display['field'] = $link_name . $display['field'];
				if (empty($display['type']))
					$display['type'] = 'field';
			}
			if (isset($display['fields']) && $link_name) {
				foreach ($display['fields'] as $k => $v)
					$display['fields'][$k] = $link_name . $v;
			}
			//if(! isset($display['type']))
			//	$display['type'] = 'non-db';
			return $display;
		} else {
			if (is_string($display))
				$fname = $display;
			$fdef = AppConfig::setting("model.fields.$model.$fname");
			if ($fdef) {
				$display = array(
					'type' => 'field',
					'field' => $link_name . $fname,
				);
				return $display;
			}
			if(! $link_name) {
				// for scheduler in particular
				$ddef = AppConfig::setting("display.fields.$model.$fname");
				if ($ddef) {
					if(! isset($ddef['source']['type']))
						$ddef['source']['type'] = 'non-db';
					return $ddef;
				}
			}
		}
		return null;
	}

	function getModelPrimaryAccount($model = null) {
		if (!isset($model))
			$model = $this->base_model->name;
		$acc_f = AppConfig::setting("model.detail.$model.primary_account");
		if ($acc_f) {
			$fdef = AppConfig::setting("model.fields.$model.$acc_f");
			if ($fdef && ! empty($fdef['id_name'])) return array(
				'type' => 'field',
				'field' => $fdef['name'],
				'id' => $fdef['id_name'],
			);
		}
	}

	function getModelStatus($link_name = null) {
		if($link_name) {
			$fdef = $this->getFieldDefinition($link_name, $err_msg);
			if (! isset($this->links[$link_name]) && ! $this->addLink($link_name))
				return false;
			$model = $this->links[$link_name];
		}
		else
			$model = $this->base_model;
		$stat_f = AppConfig::setting("model.detail.{$model->name}.record_status");
		if(is_string($stat_f))
			$stat_f = $model->getDisplayField($stat_f, true);
		else if(! is_array($stat_f))
			return;
		if($link_name)
			$stat_f['link'] = $link_name;
		return $stat_f;
	}

	function checkRefField($alias) {
		$field_def =& $this->fields[$alias];
		if (isset($field_def['link']))
			$parent = $field_def['link'] . '.';
		else
			$parent = '';

		// pull in related ID field for refs
		$id_name = $alias . '_id';
		if (isset($this->added_field_names[$id_name])) {
			$id_name = $this->added_field_names[$id_name];
			$field_def['id_name'] = $id_name;
		}
		else if (!isset($this->fields[$id_name]) && $this->auto_add_fields) {
			$field_def['model_id_name'] = $field_def['id_name']; // store defined value in case it's needed later
			$ref_id_name = $parent . $field_def['id_name'];
			$id_alias = array_get_default($field_def, 'id_alias');
			$added_as = $this->addField($ref_id_name, $id_alias);
			$field_def['id_name'] = $added_as;
			if ($id_name != $field_def['id_name'])
				$this->added_field_names[$id_name] = $added_as;
			$id_name = $added_as;
		}

		// for Currency ref fields, pull in related exchange_rate as well
		if (array_get_default($field_def, 'bean_name') == 'Currency' && !empty($field_def['exchange_rate'])) {
			$exid = $parent . $field_def['exchange_rate'];
			if (isset($this->added_field_names[$exid]))
				$exid = $this->added_field_names[$exid];
			else if($this->auto_add_fields)
				$exid = $this->addField($exid);
			$field_def['exchange_rate'] = $exid;
		}
	}

	function checkNameField($alias) {
		$fdef =& $this->fields[$alias];
		if(empty($fdef['no_query'])) {
			$link = array_get_default($fdef, 'link');
			if(! empty($link) && $this->auto_add_fields)
				$this->addPrimaryKey(null, $link);
		}
	}

	function checkLinkField($alias) {
		static $depth = 0;
		if ($depth)
			return;
		$depth = 1;
		$fdef =& $this->fields[$alias];
		if(empty($fdef['no_query'])) {
			$link = array_get_default($fdef, 'link');
			$type = array_get_default($fdef, 'type');
			if(! empty($link) && $type != 'ref' && $type != 'id' && $type != 'ref_id' && $this->auto_add_fields)
				$this->addPrimaryKey(null, $link);
		}
		$depth = 0;
	}

	function checkCurrencyField($alias) {
		// for currency value fields, pull in the related currency
		$field_def =& $this->fields[$alias];
		$link = array_get_default($field_def, 'link');
		$parent = $link ? $link . '.' : '';
		if($field_def['type'] == 'base_currency') {
			$cid = array_get_default($field_def, 'display_currency_id');
			if(! empty($cid) && $this->auto_add_fields) {
				$cid = $parent.$cid;
				if (isset($this->added_field_names[$cid]))
					$cid = $this->added_field_names[$cid];
				else if($this->auto_add_fields)
					$cid = $this->addField($cid);
				$field_def['display_currency_id'] = $cid;
			}
			return;
		}
		$base = array_get_default($field_def, 'base_field');
		if($base) {
			$base = $parent.$base;
			if (isset($this->added_field_names[$base]))
				$base = $this->added_field_names[$base];
			else if($this->auto_add_fields)
				$base = $this->addField($base);
			$field_def['base_field'] = $base;
		}
		$cid = array_get_default($field_def, 'currency');
		if(! $cid) {
			$cm = $this->getModelName($link);
			$cid = AppConfig::setting("model.detail.$cm.display_currency");
			if(! $cid) $cid = 'currency';
		}
		$cid = $parent.$cid;
		if (isset($this->added_field_names[$cid]))
			$cid = $this->added_field_names[$cid];
		else if($this->auto_add_fields)
			$cid = $this->addField($cid);
		$field_def['currency'] = $cid;
		if ($cid) {
			$field_def['currency_id'] = array_get_default($this->fields[$cid], 'id_name');
			$field_def['exchange_rate'] = array_get_default($this->fields[$cid], 'exchange_rate');
		} else {
			$field_def['currency_id'] = null;
			$field_def['exchange_rate'] = null;
		}
	}

	function checkDurationField($alias) {
		// for duration fields, pull in the all_day_flag field, if any
		$field_def =& $this->fields[$alias];
		$allday = array_get_default($field_def, 'all_day_flag');
		if(! $allday)
			return;
		if (isset($this->added_field_names[$allday]))
			$allday = $this->added_field_names[$allday];
		else if($this->auto_add_fields)
			$allday = $this->addField($allday);
		$field_def['all_day_flag'] = $allday;
	}

	function checkUseRealNames() {
		if(isset($this->use_real_names))
			$use_real_names = $this->use_real_names;
		else
			$use_real_names = $GLOBALS['locale']->getUseRealNames();
		return $use_real_names;
	}

	function checkSourcedField($alias) {
		$spec =& $this->fields[$alias];
		$src =& $spec['source'];
		$parent = array_get_default($spec, 'link', '');
		if ($parent) $parent .= '.';

		if ($src['type'] == 'user_display_name') {
			$src['type'] = 'field';
			if($this->checkUseRealNames())
				$src['field'] = 'full_name';
			else
				$src['field'] = 'user_name';
		}

		if (!empty($src['field']) && ! $this->is_union) {
			$add_spec = array('name' => $parent . $src['field']);
			if(! empty($spec['ref_alt_value']) && ! empty($spec['id_name'])) {
				$add_spec['ifnull_check'] = $spec['id_name'];
				$add_spec['ifnull_value'] = $spec['ref_alt_value'];
			}
			$src['alias_field'] = $this->addField($add_spec, null, false, true);
			if($src['alias_field']) {
				$aspec =& $this->fields[$src['alias_field']];
				foreach(array('editable', 'updateable', 'id_name') as $k)
					if(! isset($spec[$k]) && isset($aspec[$k]))
						$spec[$k] = $aspec[$k];
			}
		}
		if (!empty($src['fields'])) {
			if ($alias == 'category') // FIXME
				$src['alias_fields'] = array();
			foreach ($src['fields'] as $f) {
				$no_q = false;
				if ($src['type'] == 'sequence' || $src['type'] == 'concat' || $src['type'] == 'prefixed' || $src['type'] == 'person_name' || ! empty($spec['no_query']))
					$no_q = true;
				if ($src['type'] == 'person_name' && $this->auto_add_fields)
					$no_q = false;
				$src['alias_fields'][$f] = $this->addField($parent . $f, null, false, $no_q);
			}
		}
		if ($src['type'] == 'address')
			$this->checkAddressField($alias);
	}

	function checkAddressField($alias) {
		$field_def = $this->fields[$alias];
		$src =& $this->fields[$alias]['source'];
		$parent = array_get_default($field_def, 'link', '');
		if ($parent) $parent .= '.';

		if (isset($src['prefix'])) {
			$pfx = $src['prefix'];
			$fs = array(
				'address_street', 'address_city',
				'address_state', 'address_country',
				'address_statecode', 'address_countrycode',
				'address_postalcode',
				'address_email', 'address_phone',
				'account', 'account_id',
				'contact', 'contact_id', 'contact.salutation',
			);
			$add_fs = array();
			foreach ($fs as $idx => $f) {
				if (($f == 'account' || $f == 'account_id') && !empty($src['no_account']))
					continue;
				if (($f == 'contact' || $f == 'contact_id' || $idx == 'contact.salutation') && !empty($src['no_contact']))
					continue;
				$fname = $parent . $pfx . $f;
				$add_fs[$parent.$pfx.strtr($f,'.','_')] = $parent . $pfx . $f;
			}
			$added = $this->addFields($add_fs);
			foreach ($fs as $idx => $f) {
				if (isset($added[$idx]))
					$src['alias_fields'][strtr($f,'.','_')] = $added[$idx];
			}
		}
	}


	function checkImageField($alias) {
		// for image fields, pull in the filename, url, and thumbnail fields, if any
		$field_def =& $this->fields[$alias];
		$name = $field_def['name'];
		$subf = array(
			'filename' => $name . '_filename',
			'thumbnail_name' => $name . '_thumb',
		);
		if(! empty($field_def['alt_image'])) {
			$alt = $this->getFieldDefinition($field_def['alt_image'], $errmsg);
			if($alt) {
				$palt = array_get_default($alt, 'link');
				$palt = $palt ? $palt . '.' : '';
				$subf['alt_filename'] = $palt. array_get_default($alt, 'filename', $alt['name'] . '_filename');
				$subf['alt_thumb'] = $palt . array_get_default($alt, 'thumbnail_name', $alt['name'] . '_thumb');
				$field_def['alt_upload_dir'] = empty($alt['store_full_path']) ? array_get_default($alt, 'upload_dir') : null;
			}
		}
		$parent = array_get_default($field_def, 'link', '');
		if ($parent) $parent .= '.';
		foreach($subf as $prop => $default) {
			$f = $parent . array_get_default($field_def, $prop, $default);
			if (isset($this->added_field_names[$f]))
				$f = $this->added_field_names[$f];
			else if($this->auto_add_fields)
				$f = $this->addField($f);
			$field_def[$prop] = $f;
		}
	}


	function getFieldDefinition($spec, &$err_msg) {
		if (is_array($spec)) {
			if (empty($spec['name']))
				return false;
			$field_name = $spec['name'];
			unset($spec['name']);
		} else {
			$field_name = $spec;
			$spec = array();
		}

		if (($dp = strrpos($field_name ?? '', '.')) !== false) {
			$link_name = substr($field_name, 0, $dp);
			$link_member = substr($field_name, $dp + 1);
			if(! empty($spec['link']) && $spec['link'] != $link_name)
				$spec['link'] .= '.'.$link_name;
			else
				$spec['link'] = $link_name;
			$field_name = $link_member;
		}

		if(! empty($spec['link'])) {
			$link_name = $spec['link'];
			if (!$this->addLink($link_name)) {
				$err_msg = array('unknown link', $link_name, $field_name);
				return false;
			}
			$field_def = $this->links[$link_name]->getDisplayField($field_name, true);
			if (!$field_def) {
				$err_msg = array('unknown link field', $link_name, $field_name);
				return false;
			}
		} else if($this->base_model) {
			$field_def = $this->base_model->getDisplayField($field_name, true);
			if (!$field_def) {
				if (!empty($spec['source'])) {
					$check_missing_sourced = true;
					$field_def = array('name' => $field_name);
				} else {
					$err_msg = array('unknown field', $field_name, $field_name);
					return false;
				}
			}
		} else
			return false;

		$field_def = array_merge($field_def, $spec);

		if (!empty($field_def['multi_select_group']) && $field_def['type'] == 'multienum') {
			$grp = $field_def['multi_select_group'];
			$cnt = array_get_default($field_def, 'multi_select_count', 10);
			for ($c = 1; $c <= $cnt; $c++)
				$fnames[] = $grp . ($c == 1 ? '' : $c);
			$field_def['source'] = array(
				'type' => 'concat',
				'fields' => $fnames,
				'separator' => '^,^',
			);
		}

		if (!empty($check_missing_sourced)) {
			if ($field_def['source']['type'] == 'db') {
				$err_msg = array('unknown field', $field_name, $field_name);
				return false;
			}
		}

		return $field_def;
	}

	function getAvailableJoinFields($link_name) {
		if (isset($this->links[$link_name.'~join']))
			return $this->links[$link_name.'~join']->getAllDisplayFields();
		return false;
	}

	function getAvailableFields($link_name=null) {
		if($link_name) {
			if (!$this->addLink($link_name)) {
				//$err_msg = array('unknown link', $link_name, $field_name);
				return false;
			}
			$fields = $this->links[$link_name]->getAllDisplayFields();
		} else if($this->base_model) {
			$fields = $this->base_model->getAllDisplayFields();
		} else
			return false;
		return $fields;
	}

	function getAvailableLinks($link_name=null) {
		if($link_name) {
			if (!$this->addLink($link_name)) {
				//$err_msg = array('unknown link', $link_name, $field_name);
				return false;
			}
			$fields = $this->links[$link_name]->getLinks();
		} else if($this->base_model) {
			$fields = $this->base_model->getLinks();
		} else
			return false;
		return $fields;
	}


	function fieldAdded($f) {
		return !empty($this->fields[$f]);
	}

	function getField($alias) {
		return array_get_default($this->fields, $alias);
	}

	function getFieldAlias($name) {
		if (isset($this->added_field_names[$name]))
			return $this->added_field_names[$name];
		if (isset($this->fields[$name]))
			return $name;
	}

	function makeFieldId(array $spec, $add_link=false) {
		$fmt = ! empty($spec['format']) ? '!'.$spec['format'] : '';
		$tot = ! empty($spec['total']) ? ':'.$spec['total'] : '';
		$link = $add_link ? (! empty($spec['link']) ? $spec['link'] . '.' : '') : '';
		return array_get_default($spec, 'alias', $link.$spec['name'].$fmt.$tot);
	}

	function getFields($inc_no_query = false, $inc_totals = false) {
		$ret = array();
		foreach ($this->fields as $k => $v) {
			if (($inc_no_query || empty($v['no_query'])) && ($inc_totals || empty($v['total'])))
				$ret[$k] = $v;
		}
		return $ret;
	}

	function getFieldNames() {
		return array_keys($this->fields);
	}

	function getTotal($alias) {
		return array_get_default($this->totals, $alias);
	}

	function haveTotals() {
		return is_countable($this->totals) ? count($this->totals) : 0;
	}

	function setFormatted($format='html', $context=null) {
		$this->formatted = $format;
		$this->formatted_context = $context;
	}


	function addLink($full_name, $alias=null, $required=null, $link_type=null) {
		if(substr($full_name, -5) == '~join') {
			$parent = substr($full_name, 0, -5);
			if(! isset($this->links[$parent]))
				$this->addLink($parent, null, $required);
		}
		if(! isset($alias)) $alias = $full_name;
		if (isset($this->links[$alias]))
			return $alias;
		if (($dp = strrpos($full_name, '.')) !== false) {
			$parent = substr($full_name, 0, $dp);
			if (!$this->addLink($parent, null, $required))
				return false;
			$link_name = substr($full_name, $dp + 1);
			$link = $this->links[$parent]->getLinkTargetModel($link_name, $alias, null, $required, $link_type);
		} else {
			$parent = null;
			$link = $this->base_model->getLinkTargetModel($full_name, $alias, null, $required, $link_type);
		}
		if (!$link) {
			$def = $this->getFieldDefinition($full_name, $msg);
			if($def && $def['source']['type'] == 'custom_field') {
				// automatically resolve links to related value custom fields
				return $this->addLink('_cstm.'.$full_name, $alias, $required, $link_type);
			}
			return false;
		}
		$link->link_spec['parent'] = $parent;
		if (($join = $link->getJoinModel())) {
			$link->link_spec = $join->link_spec;
			$this->_addJoinLink($alias, '~parent', $join);
		}
		$this->links[$alias] = $link;
		return $alias;
	}

	function haveLink($name) {
		return ! empty($this->links[$name]);
	}

	function getLink($full_name, $alias=null, $required=null) {
		$name = $this->addLink($full_name, $alias, $required);
		if($name)
			return $this->links[$name];
	}

	function _addJoinLink($parent, $target, ModelDef &$link) {
		$alias = 'jt' . ($this->join_index++);
		$link->setTableAlias($alias);
		$this->links[$parent . '~join'] = $link;
		$this->join_link[$target] = $parent . '~join';
	}


	function addCustomLink(ModelDef $link, $name=null, $link_spec=null, $required=null) {
		if(! $name) {
			$name = $link->getTableAlias();
			if(! $name)
				$name = 'jt' . ($this->join_index++);
		}
		$link->setTableAlias($name);
		if($link_spec)
			$link->link_spec = $link_spec;
		$this->links[$name] = $link;
	}

	function loadSources($spec, &$sources) {
		$first = null;
		$primary = null;
		$add_link = array();

		// find primary source
		foreach ($spec as $idx => &$src) {
			if(empty($src['name']))
				continue;
			$t = array_get_default($src, 'display');
			$t_old = array_get_default($src, 'type'); // older method for specifying display
			if($t_old == 'primary') $t = 'primary';
			$src['display'] = $t;
			if($t == 'primary') {
				$primary = $src;
			} else if($t == 'nested' || $t == 'joined') {
				$add_link[$idx] = $src;
			}
			if(! isset($first)) $first = $idx;
		}

		if(! isset($primary)) {
			if(isset($first)) {
				$primary = $spec[$first];
				unset($add_link[$first]);
			} else
				return false;
		}

		$this->setTableAlias($primary['name']);
		$primary['link_name'] = null;
		$sources[$primary['name']] = $primary;

		// add supporting links
		foreach ($add_link as $idx => $src) {
			if(empty($src['parent']) || empty($src['field_name']))
				continue;
			if(! isset($src['display']))
				$src['display'] = 'joined';
			$parent = $src['parent'];
			$root = false;
			$path = '';
			while(1) {
				if($parent == $primary['name']) {
					$root = true; break;
				}
				else if(isset($sources[$parent]) && isset($sources[$parent]['parent'])) {
					$path = $parent . '.' . $path;
					$parent = $sources[$parent]['parent'];
				}
				else break;
			}
			if(! $root) continue;
			$required = array_get_default($src, 'required');
			$added = $this->addLink($path.$src['field_name'], $path.$src['name'], $required, $src['display']);
			if(! $added)
				continue;
			$src['link_name'] = $added;
			$sources[$src['name']] = $src;
		}
		return true;
	}


	// accessors

	function getTableName($link = null) {
		if (empty($link) && $this->base_model)
			return $this->base_model->getTableName();
		if (isset($this->links[$link]))
			return $this->links[$link]->getTableName();
	}

	function getTableAlias($link = null, $non_empty = false) {
		if (empty($link) && $this->base_model)
			return $this->base_model->getTableAlias($non_empty);
		if (isset($this->links[$link]))
			return $this->links[$link]->getTableAlias($non_empty);
	}

	function &getBaseModel() {
		return $this->base_model;
	}

	function getModuleDir($link = null) {
		if (empty($link) && $this->base_model)
			return $this->base_model->getModuleDir();
		if (isset($this->links[$link]))
			return $this->links[$link]->getModuleDir();
	}

	function getModelName($link = null) {
		if (empty($link) && $this->base_model)
			return $this->base_model->getModelName();
		if (isset($this->links[$link]))
			return $this->links[$link]->getModelName();
	}

	function getModuleDirs() {
		$dirs = array();
		if($this->is_union) {
			$dirs[0] = $this->getModuleDir();
			foreach($this->union_queries as $k => &$q) {
				if(is_int($k)) $k ++;
				$dirs[$k] = $q->getModuleDir();
			}
			return $dirs;
		}
		$dirs[0] = $this->getModuleDir();
		foreach ($this->getLinkNames() as $l) {
			if (isset($this->links[$l]) && $this->links[$l] instanceof ModelDef)
				$dirs[$l] = $this->links[$l]->getModuleDir();
		}
		return $dirs;
	}

	function getPrimaryKey($link = null) {
		if (empty($link)) {
			if($this->primary_added)
				return $this->primary_added;
			else if($this->base_model)
				return $this->base_model->getPrimaryKey();
		}
		if (isset($this->links[$link]))
			return $this->links[$link]->getPrimaryKey();
	}

	function getLinkNames() {
		return array_keys($this->links);
	}


	// query building

	function getAllFieldsSelect($have_links = true, $as_array = false, $inc_totals = false) {
		global $db;
		$fs = array();
		if($this->fields) {
			foreach ($this->fields as $alias => $spec) {
				if (empty($spec['no_query']) && ($inc_totals || empty($spec['total']))) {
					$f = $this->getFieldSelect($spec, $alias, $have_links, 'query');
					if (isset($f))
						$fs[$alias] = $f;
				}
			}
			if ($fs && $this->hasPersonalData() && $this->join_personal) {
				$fs[] = 'erase_personal.erased_fields AS ___erased_fields___';
			}
		}
		if(! $fs) {
			$sel = '*';
			if(! $this->is_union)
				$sel = $db->quoteField($this->getTableAlias(null, true)) . '.' . $sel;
			if ($this->hasPersonalData() && $this->join_personal) {
				$sel .= ', erase_personal.erased_fields AS ___erased_fields___';
			}
			return $sel;
		}
		if ($as_array)
			return $fs;
		return implode(', ', $fs);
	}


	function getFiscalYearOffset() {
		if(empty($this->fy_start_month)) {
			$this->fy_start_month = AppConfig::setting('company.fiscal_year_start', 1);
		}
		$fy_offset = '';
		if($this->fy_start_month > 1)
			$fy_offset = ' - INTERVAL '.($this->fy_start_month - 1).' MONTH';
		return $fy_offset;
	}


	function getTimeZone() {
		if($this->timezone)
			return $this->timezone;
		return AppConfig::system_timezone();
	}

	function setTimeZone($tz) {
		$this->timezone = $tz;
	}

	function loadUserTimeZone($user=null) {
		global $current_user;
		if($user)
			$this->setTimeZone($user->getPreference('timezone'));
		else if(! empty($current_user))
			$this->setTimeZone($current_user->getPreference('timezone'));
	}

	function localOffsetDateTime($sel) {
		static $known = [];
		global $timedate;
		global $db;
		$tz = $this->getTimeZone();
		if (!isset($known[$tz])) {
			$result = $db->query("SELECT @@SESSION.time_zone");
			if (!$db->getLastError()) {
				if ($row = $db->fetchByRow($result)) {
					$sessionTZ = $row[0];
					$known[$tz] = false;
					try {
						$db->query("SET time_zone = '$tz'", false, '', true);
						$known[$tz] = !$db->getLastError();
					} catch (mysqli_sql_exception $e) {}
					$db->query("SET time_zone = '$sessionTZ'", false, '', true);
				}
			}
		}
		if (empty($known[$tz])) {
			$offs = $timedate->getTimeZoneOffset($tz, true) / 60;
			if(! $offs)
				return $sel;
			$hrs = floor($offs / 60);
			$mins = $offs - $hrs * 60;
			$tzs = sprintf("%+d:%02d", $hrs, $mins);
			return "CONVERT_TZ($sel, '+0:00', '$tzs')";
		}
		return "CONVERT_TZ($sel, '+0:00', '$tz')";
	}


	function getFieldSelect($spec, $alias = null, $have_links = true, $context = 'expr', $format = null, $params = null) {
		/** @var $db DBManager */
		global $db;

		if (!is_array($spec)) {
			if (isset($this->fields[$spec]))
				$spec = $this->fields[$spec];
			else if( ($spec = $this->getFieldDefinition($spec, $loaderr)) )
				;
			else
				return null;
		}
		if (!isset($alias)) {
			$alias = $spec['name'];
		}
		if ($have_links)
			$table = $this->getTableAlias(array_get_default($spec, 'link'), true);
		else
			$table = null;
		$type = array_get_default($spec, 'type');
		if (isset($spec['source']) && $spec['source']['type'] != 'db') {
			$sel = $this->getFieldFormula($spec, $table, $have_links, $context, $params);
			$use_alias = true;
		}
		else if (isset($spec['source']['calc_formula']) && array_get_default($spec, 'link') == '_cstm') {
			$sel = $this->getCalculatedFormula($spec, $have_links, $context, $params);
			$alias = $spec['name'];
			$use_alias = true;
		}
		else {
			// will need to call out to custom field manager classes
		}

		if (!isset($sel)) {
			// will need to call out to custom field manager classes
			$fname = $spec['name'];

			// switch to the raw phone number when filtering by numeric values
			if($context == 'where' && $type == 'phone' && isset($params['filter_value']) && is_numeric($params['filter_value']))
				$fname .= '__raw';

			$sel = $db->quoteField($fname, $table);
		}
		if(! isset($format))
			$format = array_get_default($spec, 'format');

		if($format) {
			switch($format) {
				case 'rounded':
					$sel = "ROUND($sel)"; break;
				case 'thousands':
					$sel = "ROUND(($sel)/1000)"; break;
				case 'millions':
					$sel = "ROUND(($sel)/1000000)"; break;
				case 'hours':
					$sel = "(($sel)/60)"; break;
				case 'minutes':
					break;
				case 'date_only':
				case 'time_only':
				case 'month':
				case 'year':
				case 'yearmonth':
				case 'qtr':
				case 'fqtr':
				case 'fy':
					if($spec['type'] == 'datetime') {
						$sel = $this->localOffsetDateTime($sel);
					}
					switch($format) {
						case 'date_only':
							$num_sel = $dat_sel = "DATE($sel)";
							$use_alias = true;
							break;
						case 'time_only':
							$num_sel = $dat_sel = "TIME($sel)";
							$use_alias = true;
							break;
						case 'month':
							$num_sel = $dat_sel = "MONTH($sel)";
							$use_alias = true;
							break;
						case 'year':
							$num_sel = $dat_sel = "YEAR($sel)";
							$use_alias = true;
							break;
						case 'yearmonth':
							$num_sel = "DATE_FORMAT($sel, '%Y%m')";
							if($context == 'where') // uses normal date filters
								$dat_sel = "DATE($sel)";
							else
								$dat_sel = "DATE_FORMAT($sel, '%Y-%m')";
							$use_alias = true;
							break;
						case 'qtr':
							$num_sel = "CONCAT(YEAR($sel), QUARTER($sel))";
							$dat_sel = "CONCAT(YEAR($sel), '-Q', QUARTER($sel))"; break;
							$use_alias = true;
							break;
						case 'fqtr':
							$fy_offset = $this->getFiscalYearOffset();
							$num_sel = "CONCAT(YEAR($sel $fy_offset), QUARTER($sel $fy_offset))";
							$dat_sel = "CONCAT('FY', YEAR($sel $fy_offset), '-Q', QUARTER($sel $fy_offset))";
							$use_alias = true;
							break;
						case 'fy':
							$fy_offset = $this->getFiscalYearOffset();
							$num_sel = "YEAR($sel $fy_offset)";
							$dat_sel = "CONCAT('FY', $num_sel)";
							$use_alias = true;
							break;
					}
					if($context == 'order_by')
						$sel = $num_sel;
					else
						$sel = $dat_sel;
					break;
				case 'daysfuture':
				case 'dayspast':
				case 'monthsfuture':
				case 'monthspast':
				default:
					break;
			}
		}

		if($context != 'order_by' && ! empty($spec['min_length'])) {
			$len = (int)$spec['min_length'];
			$sel = "LPAD($sel, $len, '0')";
		}
		if(! empty($spec['ifnull_value'])) {
			$alt_sel = $this->getFieldSelect($spec['ifnull_value'], '', $have_links, $context, null);
			if($alt_sel) {
				$ifnchk = array_get_default($spec, 'ifnull_check');
				if($ifnchk) {
					$ifn_sel = $this->getFieldSelect($ifnchk, '', $have_links, $context, null);
					if($ifn_sel)
						$sel = "IF($ifn_sel IS NULL, $alt_sel, $sel)";
				}
				else
					$sel = "IFNULL($sel, $alt_sel)";
			}
		}

		if ($context == 'query' && isset($spec['total'])) {
			$sel = strtoupper($spec['total']) . "($sel)";
		}
		if ($context == 'sum_expr' && isset($spec['total'])) {
			if ($spec['type'] == 'bool') {
				$sel = "IF($sel,1,NULL)";
			}
		}
		if ($context == 'query' && strlen($alias) && ($alias != $spec['name'] || !empty($use_alias)))
			$sel .= ' ' . $db->quoteField($alias);
		return $sel;
	}


	static function get_name_format($as_array=false) {
		$format = $GLOBALS['locale']->getLocaleFormatMacro();
		if(! $as_array) return $format;
		$parts = array();
		for($i = 0; $i < strlen($format); $i++) {
			$chr = substr($format, $i, 1);
			if($chr == 's' || $chr == 'f' || $chr == 'l')
				$parts[] = $chr;
		}
		return $parts;
	}


	function getFieldFormula($spec, $table, $have_links, $context, $params=null) {
		/** @var $db DBManager */
		global $db;

		$type = array_get_default($spec, 'type');
		$source = $spec['source'];
		$src_type = $source['type'];
		$sel_ctx = ($context == 'order_by' ? $context : 'expr');

		if ($context == 'query' && !empty($source['prefetched']) && !$this->disable_db_cache)
			// if value is not being used to sort or in a subexpression, fill it in from memory cache
			return '""';
		if ($type == 'ref' && $context == 'where' && is_array($params) && ! empty($params['filter_module'])) {
			// filter on ref field with dynamic module set to specific value
			$ref_bean = AppConfig::module_primary_bean($params['filter_module']);
			if($ref_bean) {
				$lq = new ListQuery($ref_bean, array('_display'), array('auto_add_fields' => false, 'no_default_order_by' => true));
				$qf = $db->quoteField($spec['id_name'], $this->getTableAlias(null, true));
				$lq->addFilterClause("id=$qf");
				$q = $lq->getSql();
				if($q) return "($q)";
			}
		}

		$lnk = array_get_default($spec, 'link');
		if($lnk) $lnk .= '.';
		if(! empty($source['field']) && empty($source['alias_field'])) {
			$source['alias_field'] = $lnk.$source['field'];
		}
		if(! empty($source['fields']) && empty($source['alias_fields'])) {
			// using a field definition that hasn't been 'loaded'
			// FIXME? may need to map these fields to ones that have already been added
			foreach($source['fields'] as $f)
				$source['alias_fields'][$f] = $lnk.$f;
		}

		if ($src_type == 'person_name') {
			if ($context == 'order_by' && empty($spec['ifnull_value'])) {
				$fs = array();
				$src_fs = array_get_default($source, 'alias_fields', $source['fields']);
				$src_fs2 = array_values($src_fs);
				$ord = array_get_default($source, 'order');

				if($ord == 'display') {
					// sort according to display format
					$nfmt = self::get_name_format(true);
					foreach($nfmt as $np) {
						if($np == 'f' && isset($src_fs2[0]))
							$fs[] = $this->getFieldSelect($src_fs2[0], null, $have_links, $sel_ctx);
						else if($np == 'l' && isset($src_fs2[1]))
							$fs[] = $this->getFieldSelect($src_fs2[1], null, $have_links, $sel_ctx);
					}
				} else {
					$order = AppConfig::setting('locale.defaults.name_order_by');
					if($order == 'first_name')
						$ord = array(0,1);
					else
						$ord = array(1,0);
					foreach($ord as $oi) {
						if(isset($src_fs2[$oi]))
							$fs[] = $this->getFieldSelect($src_fs2[$oi], null, $have_links, $sel_ctx);
					}
				}
				return $fs;
			}
			$src_type = 'concat';
			$source['space'] = true;
			$alias_fields = $source['alias_fields'];
			$source['alias_fields'] = array();
			$parts = self::get_name_format(true);
			foreach ($parts as $part) {
				if (!strlen($part))
					continue;
				if ($part == 's')
					continue; //$source['alias_fields'][] = $alias_fields[$source['fields'][2]];
				elseif ($part == 'f')
					$source['alias_fields'][] = $alias_fields[$source['fields'][0]];
				elseif ($part == 'l')
					$source['alias_fields'][] = $alias_fields[$source['fields'][1]];
			}
		}
		else if ($src_type == 'sequence') {
			$src_type = 'concat';
		}
		else if ($src_type == 'prefixed') {
			$src_type = 'concat';
			if (!isset($source['separator']))
				$source['separator'] = ': ';
		}
		else if ($src_type == 'name_rate') {
			$name = array_get_default($source['alias_fields'], 'name');
			/*$rt = array_get_default($source, 'alias_fields', 'rate');
			$dt = array_get_default($source, 'alias_fields', 'discount_type');
			$fixed = array_get_default($source, 'alias_fields', 'fixed_amount');*/
			return $this->getFieldSelect($name, null, $have_links, $sel_ctx);
		}

		else if ($src_type == 'static') {
			$val = array_get_default($source, 'value', '');
			return '"'.$db->quote($val).'"';
		}
		else if ($src_type == 'literal') {
            if ($this->disable_literal_types) throw new IAHListQueryError('Literal types forbidden');
			$val = array_get_default($source, 'value', "''");
			return $val;
		}

		if ($src_type == 'concat') {
			$sel = array();
			if (empty($source['alias_fields'])) {
				return "''";
			}
			foreach ($source['alias_fields'] as $fn) {
				if (isset($fn))
					$sel[] = $this->getFieldSelect($fn, null, $have_links, $sel_ctx);
			}
			if ($context == 'order_by' && empty($spec['ifnull_value']))
				return $sel;
			if (!empty($source['separator']))
				$sep = $source['separator'];
			else if (!empty($source['space']))
				$sep = ' ';
			else
				$sep = '';
			return $db->helper->concatExprs($sel, $sep);
		}
		else if ($src_type == 'age') {
			$sel = $this->getFieldSelect($source['alias_field'], null, $have_links, $sel_ctx);
			if ($context == 'order_by') {
				// reverse order
				return "- UNIX_TIMESTAMP($sel)";
			}
			return $sel;
		}
		else if ($src_type == 'subselect') {
			$fn = $source['function'];
			if ( (isset($source['file']) && is_file($source['file']))
				&& (isset($source['class']) && !class_exists($source['class'])) ){
				require_once($source['file']);
			}
			if(! is_array($fn) && isset($source['class']))
				$fn = array($source['class'], $fn);
			if(is_array($fn))
				require_bean($fn[0]);
			$q = call_user_func($fn, $spec, $table, $context, $this);
			if (is_string($q)) {
				// pass $q to query reformatter?
				return "($q)";
			} else
				return 'NULL';
		}
		else if ($src_type == 'function') {
			if (isset($source['alias_field']))
				return $this->getFieldSelect($source['alias_field'], null, $have_links, $sel_ctx);
			else
				return '""';
		}
		else if ($src_type == 'field' || $src_type == 'custom_field') {
			if(empty($source['alias_field']))
				return '""';
			return $this->getFieldSelect($source['alias_field'], null, $have_links, $sel_ctx);
		}
		else {
			return '""';
		}
	}

   function getCalculatedFormula($spec, $have_links, $context, $params=null) {
       require_once('include/layout/FormulaParser.php');
       $parser = new FormulaParser($spec['source']['calc_formula']);
       $fmap = array();
       $fields = $parser->getExpressionFields();
       foreach($fields as $f)
		   $fmap[$f] = $this->getFieldSelect($f, null, $have_links, $context);
       $sql = $parser->getSql($fmap);
       return $sql;
   }

	function getFieldType($field_name) {
		if (isset($this->fields[$field_name]))
			return array_get_default($this->fields[$field_name], 'type');
		if (isset($this->added_field_names[$field_name])) {
			$added_as = $this->added_field_names[$field_name];
			if ($added_as != $field_name)
				return $this->getFieldType($added_as);
		}
	}


	function getJoinInfo($link_name, $parent_name=null) {
		if (isset($this->links[$link_name])) {

			if(isset($this->links[$link_name.'~join']) && $parent_name != $link_name.'~join')
				// join must proceed through link table first
				$link_name = $link_name.'~join';

			if(isset($parent_name))
				$parent = $this->links[$parent_name];
			else if (($dp = strrpos($link_name, '.')) !== false)
				$parent = $this->links[substr($link_name, 0, $dp)];
			else {
				$lpar = array_get_default($this->links[$link_name]->link_spec, 'parent');
				if($lpar == '_cstm')
					$parent = $this->links['_cstm'];
				else
					$parent = $this->base_model;
			}
			if($link_name == '~join') {
				return $this->links[$link_name]->getJoinInfo($parent);
			}
			else if($this->links[$link_name]->is_join) {
				return $this->links[$link_name]->getJoinInfo($parent, true);
			}
			else if ($link_name == '~parent') {
				if (isset($this->links['~join']))
					return $this->links['~join']->getJoinInfo($this->links[$link_name]);
				else
					return $this->links[$link_name]->getJoinInfo($parent, true);
			}
			if($parent)
				return $parent->getJoinInfo($this->links[$link_name]);
		}
	}


	function getJoin($join_info) {
		/** @var $db DBManager */
		global $db;

		$multiple = ! empty($join_info['many']);

		if (! empty($join_info['required']))
			$q = 'INNER JOIN ';
		else
			$q = 'LEFT JOIN ';
		$table = $join_info['link_table'];
		$q .= $db->quoteField($table);
		$alias = $join_info['link_alias'];
		if (!empty($alias) && $alias != $table) {
			$q .= ' ' . $db->quoteField($alias);
		}
		$q .= ' ON ' . $join_info['clause'];
		if ($join_info['filter_deleted'] && $this->filter_deleted)
			$q .= ' AND ' . $join_info['filter_deleted'];
		return $q;
	}


	// find which links are referenced by added fields
	function getActiveLinks($context = 'default', $order_by=null) {
		$links = array();
		if (!empty($this->links['~join']))
			$links['~join'] = 1;
		if ($this->join_favorites && $context != 'count')
			$links['favorites'] = 1;
		$links = array_merge($links, $this->getWhereClauseLinks());
		if ($context == 'sum_select' || $context == 'count') {
			foreach ($this->fields as $k => &$def) {
				if(empty($def['grouped']) || !empty($def['no_query']))
					continue;
				$links = array_merge($links, $this->getFieldLinks($k, $context, $lookup_err));
			}
			foreach ($this->totals as $k => &$def) {
				if (!empty($def['no_query']))
					continue;
				if(isset($def['link']))
					$links[$def['link']] = 1;
				if($def['source']['type'] == 'custom_field') {
					$fld = $def['source']['field'];
					$p = strrpos($fld, '.');
					$links[substr($fld, 0, $p)] = 1;
				}
			}
		}
		else if ($context != 'count') {
			foreach ($this->fields as $k => &$def) {
				if (! empty($def['source']['prefetched']) || !empty($def['no_query']))
					continue;
				$links = array_merge($links, $this->getFieldLinks($k, $context, $lookup_err));
			}
			$links = array_merge($links, $this->getOrderByLinks($order_by, true));
		}
		// add any parent links of active links
		$l2 = $links;
		foreach($links as $l => $v) {
			if($v && strpos($l, '.') !== false) {
				$s = explode('.', $l);
				$pfx = array_shift($s);
				foreach($s as $part) {
					$l2[$pfx] = 1;
					$pfx .= '.' . $part;
				}
			}
		}
		return array_keys($l2);
	}


	function getFieldLinks($name, $context, &$lookup_err) {
		$ret = array();
		if (isset($this->fields[$name]))
			$fdef =& $this->fields[$name];
		else
			$fdef = $this->getFieldDefinition($name, $lookup_err);
		if ($fdef) {
			if (!empty($fdef['link'])) {
				$ret[$fdef['link']] = 1;
			}
			$ext = array();
			if(empty($fdef['source']['subselect_filter'])) {
				// pre-added field definitions will have alias_* defined, but ones we just looked up will not
				if (isset($fdef['source']['alias_fields']))
					$ext = $fdef['source']['alias_fields'];
				else if (isset($fdef['source']['fields']))
					$ext = $fdef['source']['fields'];
				if (isset($fdef['source']['alias_field']))
					$ext[] = $fdef['source']['alias_field'];
				else if (isset($fdef['source']['field']))
					$ext[] = $fdef['source']['field'];
			}
			foreach($ext as $fsub) {
				if($fsub != $name)
					$ret = array_merge($ret, $this->getFieldLinks($fsub, $context, $sub_err));
			}
			// add custom fields link for custom related value fields
			$s = explode('.', $name, 2);
			if(count($s) == 2) {
				$plink = array_get_default($this->links, $s[0]);
				if($plink) {
					$parent = array_get_default($plink->link_spec, 'parent');
					if ($parent)
						$ret[$parent] = 1;
				}
			}
		}
		return $ret;
	}


	// get join expressions for linked tables
	function getQueryJoins($as_array = false, $limit_links = null) {
		$js = array();

		foreach (array_keys($this->links) as $link_name) {
			if (isset($limit_links) && !in_array($link_name, $limit_links))
				continue;
			if(isset($this->links[$link_name.'~join'])) {
				$info = $this->getJoinInfo($link_name.'~join');
				if ($info)
					$js[$link_name.'~join'] = $this->getJoin($info);
				else
					continue;
				$info = $this->getJoinInfo($link_name, $link_name.'~join');
			} else {
				$info = $this->getJoinInfo($link_name);
			}
			if ($info)
				$js[$link_name] = $this->getJoin($info);
		}
		if ($as_array)
			return $js;
		return implode(' ', $js);
	}


	function getParentKey() {
		if($this->is_union)
			foreach($this->union_queries as $q)
				return $q->getParentKey();
		return $this->parent_key;
	}

	function setParentKey($key) {
		if($this->is_union)
			foreach($this->union_queries as &$q)
				$q->setParentKey($key);
		else
			$this->parent_key = $key;
	}


	function getUnifiedSearchFields($link_name=null) {
		/*if($this->is_union) {
			return array();
		}*/
		if(! $link_name && isset($this->unified_search_fields)) {
			return $this->unified_search_fields;
		}
		else {
			if($link_name) {
				$lnk = $this->addLink($link_name);
				if(! $lnk)
					return array();
				$defs = $this->links[$lnk]->getExtendedFieldDefinitions();
			}
			else if($this->base_model)
				$defs = $this->base_model->getExtendedFieldDefinitions();
			else
				$defs = array();
			$fs = array();
            $fulltext_search = false;
			foreach ($defs as $k => $def) {
				if (!empty($def['unified_search']) || !empty($def['fulltext_search'])) {
					$fs[$k] = $def;
                	if (!empty($def['fulltext_search']))
						$fulltext_search = true;
                }
			}
            $this->fulltext_search = $fulltext_search;
			if(! $link_name)
				$this->unified_search_fields = $fs;
			return $fs;
		}
	}

	function getUnifiedSearchFilter($value, $match = null, $link_name=null, $search_what = 'all') {
		$fs = $this->getUnifiedSearchFields($link_name);
		if (!$fs)
			return false;
		$ret = array();
		$is_ascii = ! preg_match('~[^\x00-\x7F]~u', $value);
		foreach ($fs as $k => $def) {
			if(! $is_ascii && array_get_default($def, 'charset') === 'ascii')
				continue;
            $fulltext_search = array_get_default($def, 'fulltext_search', 0);
            $search_mode = array_get_default($def, 'unified_search');
            if ($search_what === 'subject' && $fulltext_search)
                continue;
            elseif ($search_what === 'text' && ! $fulltext_search)
                continue;
            if($def['type'] == 'email' && strpos($value, '@') === false)
            	continue;
			$clause = array('field' => $k, 'value' => trim($value));
			if($search_mode === 'exact') $clause['operator'] = 'eq';
			switch ($def['type']) {
				case 'id':
					$ret[$k] = $clause;
					break;
				case 'int':
				case 'float':
				case 'double':
				case 'currency':
				case 'raw_currency':
				case 'base_currency':
				case 'sequence_number':
					if (is_numeric($value)) // FIXME - unformat number?
						$ret[$k] = $clause;
					break;
				case 'phone':
				case 'fax':
					$original = $clause['value'];
					$value = self::unformat_phone_number($original);
					if (empty($value) || preg_match('/\pL/', $original))
						$value = $original;
					$clause['value'] = $value;
				default:
					if(! isset($clause['operator']))
						$clause['operator'] = 'like';
					if (isset($match))
						$clause['match'] = $match;
					if(AppConfig::setting('site.features.extended_text_filter') && preg_match('~\s~', $clause['value'])
					&& (in_array($def['type'], self::$ext_search_types) || ! empty($def['extended_search'])))
						$clause['match'] = 'extended';
					$ret[$k] = $clause;
					break;
			}
		}
		return $ret;
	}

	static function unformat_phone_number ($phone) {
		static $extPattern = '/([^0-9+\-.()*#\s][^0-9+\-()*#\s]*)\s*(\d+)\s*$/';
		$phone = trim($phone);
		$stripped = $phone;
		$ext = null;
		$match = array();
		if (preg_match($extPattern, $phone, $match)) {
			$ext_prefix = trim($match[1]);
			$ext = $match[2];
			$stripped = preg_replace($extPattern, '', $stripped);
		}
		$ret = trim(preg_replace('~[^\d]~', '', $stripped));
		if (isset($ext)) {
			$ret .= ' ' . $ext_prefix . ' ' . $ext;
		}
		return $ret;
	}

	function setListFilter(&$lf) {
		$this->list_filter =& $lf;
	}

	function &getListFilter() {
		if (!$this->list_filter)
			$this->list_filter = new ListFilter($this);
		return $this->list_filter;
	}

	function getFilterClauses($context=null) {
		$wc = null;
		if ($this->list_filter)
			$wc = $this->list_filter->getFilterClauses();
		if (!is_array($wc))
			$wc = array();
		if($this->acl_filters && $context != 'links')
		foreach($this->acl_filters as $mode => $chk) {
			if(! $chk) continue;
			$acl = $this->getAclOwnerClauses($mode);
			if($acl) array_extend($wc, $acl);
			if (!$this->skip_acl_groups_filter) {
				$acl = $this->getAclGroupClauses($mode);
				if($acl) array_extend($wc, $acl);
			}
		}
		return $wc;
	}

	function addSimpleFilter($field, $value, $operator = null, $key = null, $match = null) {
		$clause = compact('field', 'value', 'operator');
		if(isset($match)) $clause['match'] = $match;
		$this->addFilterClause($clause, $key);
	}

	function addFilterClause($clause, $key = null) {
		if($this->is_union) {
			foreach($this->union_queries as &$u)
				$u->addFilterClause($clause, $key);
			return true;
		}
		return $this->getListFilter()->addFilterClause($clause, $key);
	}

	function addFilterClauses($clauses) {
		if($this->is_union) {
			foreach($this->union_queries as &$u)
				$u->addFilterClauses($clauses);
			return true;
		}
		return $this->getListFilter()->addFilterClauses($clauses);
	}

	function addFilterPrimaryKey($value) {
		return $this->getListFilter()->addFilterPrimaryKey($value);
	}

	function removeFilter($key) {
		if($this->is_union) {
			foreach($this->union_queries as &$u)
				$u->removeFilter($key);
			return true;
		}
		return $this->getListFilter()->removeFilter($key);
	}

	// construct where clause from pre-set parameters
	function getWhereClause($as_array = false, $have_links = true) {
		global $db;

		$wc = $this->getFilterClauses('where');
		if (isset($this->where) && $this->where !== '')
			$wc['where'] = $this->where;

		// should init this when model is loaded:
		$is_join = isset($this->links['~join']);
		if($this->base_model) {
			$wh = $this->base_model->getStandardWhere($have_links);
			foreach ($wh as $k => $w) {
				if ($k == 'deleted' && (!$this->filter_deleted || $this->is_union))
					continue;
				if (is_int($k))
					$wc[] = $w;
				else
					$wc[$k] = $w;
			}
		}
		if (isset($this->parent_key)) {
			if ($is_join) {
				$wc += $this->links['~join']->getParentKeyWhere($this->links['~parent'], $this->parent_key, $have_links);
				$jspec = $this->links['~join']->getLinkSpec();
				if(isset($jspec['relationship_spec']['join']['role_field']) && empty($this->disable_join_filter)) {
					$jf = $jspec['relationship_spec']['join']['role_field'];
					$jf_enc = $db->quoteField($jf, $this->links['~join']->getTableAlias(true));
					$jv = $db->quote($jspec['relationship_spec']['join']['role_value']);
					$wc['~join-filter'] = "$jf_enc='$jv'";
				}
			} else {
				$wc += $this->base_model->getParentKeyWhere($this->links['~parent'], $this->parent_key, $have_links);
				$lspec = $this->base_model->getLinkSpec();
				if ($lspec) {
					if(isset($lspec['relationship_spec']['right']['role_field'])) {
						$f = $lspec['relationship_spec']['right']['role_field'];
						$v = $db->quote($lspec['relationship_spec']['right']['role_value']);
						$wc['role-filter'] = "$f='$v'";
					}
				}
			}
		}
		if ($is_join) {
			$wh = $this->links['~join']->getStandardWhere($have_links);
			foreach ($wh as $k => $w) {
				if ($k == 'deleted' && !$this->filter_deleted)
					continue;
				if (is_int($k))
					$wc[] = $w;
				else
					$wc['~join.'.$k] = $w;
			}
		}
		$wc = $this->expandWhereClauses($wc, $have_links);
		if ($as_array)
			return $wc;
		return implode(' AND ', $wc);
	}

	function replaceWhereClausesTemplates(&$wc) {
		foreach ($wc as $k => &$v) {
			if (strpos($k, '_template_') === 0 && is_string($v)) {
				$val = $v;
				if ($val == '{CURRENT_USER}')
					$val = AppConfig::current_user_id();
				unset($wc[$k]);
				$wc[substr($k, 10)] = $val;
			} elseif (is_array($v)) {
				$this->replaceWhereClausesTemplates($v);
			}
		}
	}


	// take clauses in array format and convert to string format, doing appropriate escaping
	function expandWhereClauses($wc, $have_links, $depth=0) {
		/** @var $db DBManager */
		global $db;

		$this->replaceWhereClausesTemplates($wc);

		$ret = array();
		foreach ($wc as $k => $clause) {
			$clause = $this->getUnionWhereClause($clause);
			if (is_array($clause)) {
				if (isset($clause['multiple']) && is_array($clause['multiple'])) {
					$operator = strtoupper(array_get_default($clause, 'operator', 'AND'));
					$neg_suball = ($operator == 'NOT' || $operator == 'NOT_AND');
					if($operator == 'NOT') $operator = 'OR';
					else if($operator == 'NOT_AND') $operator = 'AND';
					$exp_clause = $this->expandWhereClauses($clause['multiple'], $have_links, $depth+1);
					if (count($exp_clause)) {
						$ret[$k] = implode(" $operator ", $exp_clause);
						if ($neg_suball || (($depth > 0 || $operator != 'AND') && count($exp_clause) > 1))
							$ret[$k] = ($neg_suball ? 'NOT (' : '(') . $ret[$k] . ')';
					}
				}
				else {
					$op = array_get_default($clause, 'operator', '=');
					if ($op === true) $op = 'true';
					else if ($op === false) $op = 'false';
					else $op = strtolower($op);

					$op_alias = array(
						'eq' => '=', 'not_eq' => '!=',
                        'any_of' => 'in', 'not_any_of' => 'not_in',
						'true' => '', 'not_empty' => '',
						'false' => 'not', 'empty' => 'not',
						'gt' => '>', 'gte' => '>=', 'lt' => '<', 'lte' => '<=',
					);

					if(isset($op_alias[$op])) $op = $op_alias[$op];

					if (substr($op, 0, 4) == 'not_' || $op === 'not') {
						$op = substr($op, 4);
						$negate = ' NOT';
					} else
						$negate = '';

					if ($op == 'literal') {
                        if ($this->disable_literal_types) throw new IAHListQueryError('Literal types forbidden');
						$ret[$k] = "$negate {$clause['value']}";
					}
					else if ($op == 'team') {
						$val = array_get_default($clause, 'value');
						if ((is_array($val) && !count($val)) || !$val) {
							// filtering by no teams
							$ret[$k] = "0";
						}
						else {
							if (is_array($val) && count($val) == 1) {
								reset($val);
								$val = current($val);
							}
							if (is_array($val)) {
								foreach ($val as $vali)
									$v[] = "'" . $db->quote($vali) . "'";
								$m = 'IN (' . implode(',', $v) . ')';
							}
							else {
								$val = $db->quote($val);
								if($negate) { $m = " != '$val'"; $negate = ''; }
								else $m = " = '$val'";
							}
							$mod = $this->getModuleDir();
							$pk = $this->getFieldSelect($this->getPrimaryKey(), null, true, 'where');
							if($mod == 'Users') {
								$team_t = 'securitygroups_users';
								$team_f = "team_t.user_id=$pk";
							} else {
								$team_t = 'securitygroups_records';
								$team_f = "team_t.module='$mod' AND team_t.record_id=$pk";
							}
							$ret[$k] = "$negate (SELECT COUNT(team_t.securitygroup_id) FROM $team_t team_t " .
									"WHERE $team_f AND team_t.deleted=0 " .
									"AND team_t.securitygroup_id $m)";
						}
					}
					else if($op == 'exists') {
						$sub_link = array_get_default($clause, 'link_name');
						$have_link = $sub_link ? $this->getLink($sub_link) : null;
						if($have_link) {
							$join_info = $this->getJoinInfo($sub_link);
							$lqs = new ListQuery($this->base_model, array('_primary'), array('link_name' => $sub_link, 'no_default_order_by' => true));
							$lqs->addFilterClause($join_info['clause']);
							if(isset($clause['clauses'])) {
								$lqs->addFilterClauses($clause['clauses']);
							} else if(isset($clause['where'])) {
								$lqs->addFilterClause($clause['where']);
							}
							$qq = $lqs->getSql();
							if($qq) {
								$ret[$k] = "$negate EXISTS ($qq)";
							}
						}
					}
					else if (!empty($clause['field'])) {
						$f = $clause['field'];
						if (isset($this->added_field_names[$f]))
							$f = $this->added_field_names[$f];
						if (!isset($this->fields[$f]))
							$f = $this->addField($f, null, false, true);
						if (!isset($f))
							continue;
						if(! empty($this->fields[$f]['source']['subselect_filter'])) {
							$filt = $this->getSubselectFilter($f, $clause);
							if(! empty($filt))
								$ret[$k] = $filt;
							continue;
						}

						$format = array_get_default($clause, 'format');
						$val = array_get_default($clause, 'value');
						$params = array();
						if(isset($clause['filter_module']))
							$params['filter_module'] = $clause['filter_module'];
						$params['filter_operator'] = $op;
						$params['filter_value'] = $val;
						$sel = $this->getFieldSelect($f, null, $have_links, 'where', $format, $params);
						$ftype = $this->getFieldType($f);

						unset($m);
						unset($complex);

						if ($op == 'null') {
							if ($negate) {
								$m = ' IS NOT NULL ';
								$negate = '';
							} else {
								$m = ' IS NULL ';
							}
						}
						else if ($op == 'not null') {
							$m = ' IS NOT NULL ';
							$negate = '';
						}
						else if ($op == 'length') {
							$ret[$k] = "$negate LENGTH($sel)";
						}
						else if ($op == 'fulltext') {
							$v = $db->quote($val);
							$sel = "MATCH ($sel) AGAINST('$v')";
							$negate = '';
							$m = '';
						} else if ($op == 'like') {
							$v = $db->quote($val);
							$pos = array_get_default($clause, 'match', 'infix');
							if ($pos == 'prefix')
								$m = " LIKE '$v%'";
							else if ($pos == 'suffix')
								$m = " LIKE '%$v'";
							else if ($pos == 'custom')
								$m = " LIKE '$v'";
							else if ($pos == 'extended') {
								$vs = preg_split('~,?\s+~', $val);
								$ms = array();
								foreach($vs as $v) {
									$v = $db->quote($v);
									$ms[] = "$sel LIKE '%$v%'";
								}
								$ret[$k] = '(' . implode(' AND ', $ms) . ')';
							}
							else
								$m = " LIKE '%$v%'";
						}
						else if ($op == 'in' || $op == '=' || $op == '!=') {
							if($ftype == 'multienum') {
								$cs = array();
								if(! is_array($val))
									$val = isset($val) ? array($val) : array();
								if(count($val)) {
									if ($op == '!=')
										$negate = $negate ? '' : 'NOT';
									foreach($val as $v) {
										$qv = $db->quote($v);
										$cs[] = "$sel $negate RLIKE '(^|\\\\^,\\\\^)$qv(\\\\^,\\\\^|\$)'";
									}
								} else {
									if ($op == '!=') {
										$op = $negate ? '=' : '!=';
									} else {
										$op = $negate ? '!=' : '=';
									}
									$cs[] = "IFNULL($sel, '') $op ''";
								}
								if($op == 'in') $negate = !$negate;
								$m = implode($negate ? ' OR ' : ' AND ', $cs);
								$sel = $negate = '';
								$complex = true;
							}
							else if (is_array($val)) {
								if(! count($val)) {
									$ret[$k] = $negate ? 1 : 0; // empty array - all or nothing. better than a syntax error
								} else if(count($val) == 1) {
									$v = $db->quote(current($val));
									if ($op == '!=') {
										$op = $negate ? '=' : '!=';
									} else {
										$op = $negate ? '!=' : '=';
									}
									$negate = '';
									$m = " $op '$v'";
								} else {
									$v = array();
									foreach ($val as $vali)
										$v[] = "'" . $db->quote($vali) . "'";
									$m = ' IN (' . implode(',', $v) . ')';
									if ($op == '!=')
										$m = ' NOT' . $m;
								}
							}
							else {
								$v = $db->quote($val);
								if ($op == '!=') {
									$op = $negate ? '=' : '!=';
								} else {
									$op = $negate ? '!=' : '=';
								}
								$negate = '';
								$m = " $op '$v'";
							}
						}
						else if ($op == 'non_zero') {
							$sel = "ABS($sel)";
							$m = ' >= 0.005';
						}
						else if ($op == 'between') {
							$start = $val;
							$end = array_get_default($clause, 'end');
							if(isset($start) && isset($end)) {
								if ($negate) {
									$m = " < '$start' OR $sel > '$end'";
									$negate = '';
									$complex = true;
								} else
									$m = " >= '$start' AND $sel <= '$end'";
							}
							else if(isset($start)) {
								$m = ($negate ? ' < ' : ' >= ') . "'$start'";
								$negate = '';
							}
							else if(isset($end)) {
								$m = ($negate ? ' > ' : ' <= ') . "'$end'";
								$negate = '';
							}
						}
						else if ($op == 'before_date' || $op == 'after_date' || $op == 'on_date' || $op == 'between_dates') {
							$start_ts = $start_dt = $end_ts = $end_dt = null;
							if ($ftype == 'datetime') {
								if (preg_match('~^(\d{4})-(\d{2})-(\d{2})$~', $val, $md)) {
									$start_ts = $this->calcDateStartTime((int) $md[1], (int) $md[2], (int) $md[3], true);
									$start_dt = gmdate('Y-m-d H:i:s', $start_ts);
								}
								if ($op == 'on_date' && ! empty($start_dt)) {
									$end_ts = $start_ts + 24 * 3600;
									$end_dt = gmdate('Y-m-d H:i:s', $end_ts);
								}
								else if ($op == 'between_dates') {
									$end_f = array_get_default($clause, 'end');
									if (preg_match('~^(\d{4})-(\d{2})-(\d{2})$~', $end_f, $md)) {
										$end_ts = $this->calcDateStartTime((int) $md[1], (int) $md[2], (int) $md[3], true);
										$end_dt = gmdate('Y-m-d H:i:s', $end_ts + 24 * 3600);
									}
								}
							}
							else {
								if ($ftype != 'date')
									$sel = "LEFT($sel, 10)";
								$start_dt = $val;
								if ($op == 'between_dates')
									$end_dt = array_get_default($clause, 'end');
							}

							if($op == 'before_date' && ! empty($start_dt)) {
								$op = 'after_date';
								$end_dt = $start_dt;
								$start_dt = null;
							}
							if(! empty($start_dt) && ! empty($end_dt)) {
								if ($negate) {
									$m = " < '$start_dt' OR $sel > '$end_dt'";
									$negate = '';
									$complex = true;
								} else
									$m = " >= '$start_dt' AND $sel <= '$end_dt'";
							}
							else if(! empty($start_dt)) {
								if($ftype != 'datetime' && $op == 'on_date')
									$m = ($negate ? ' != ' : ' = ') . "'$start_dt'";
								else
									$m = ($negate ? ' < ' : ' >= ') . "'$start_dt'";
								$negate = '';
							}
							else if(! empty($end_dt)) {
								$m = ($negate ? ' > ' : ' <= ') . "'$end_dt'";
								$negate = '';
							}
						}
						else if ($op == 'yesterday' || $op == 'today' || $op == 'tomorrow') {
							$start_offs = ($op == 'yesterday' ? -1 : ($op == 'tomorrow' ? +1 : 0));
							$start_ts = $this->relativeDayStartTime($start_offs, true);
							if ($ftype == 'datetime') {
								$end_ts = $start_ts + 24 * 3600;
								$start_dt = gmdate('Y-m-d H:i:s', $start_ts);
								$end_dt = gmdate('Y-m-d H:i:s', $end_ts);
								if ($negate) {
									$m = " < '$start_dt' OR $sel >= '$end_dt'";
									$negate = '';
									$complex = true;
								} else
									$m = " >= '$start_dt' AND $sel < '$end_dt'";
							}
							else {
								if ($ftype != 'date')
									$sel = "LEFT($sel, 10)";
								$start_dt = date('Y-m-d', $start_ts);
								$op = $negate ? '!=' : '=';
								$m = " $op '$start_dt'";
								$negate = '';
							}
						}
						else if ($op == 'period_current' || $op == 'period_next' || $op == 'period_prev') {
							$period = array_get_default($clause, 'period', 'week');
							$pd_offset = ($op == 'period_next' ? 1 : ($op == 'period_prev' ? -1 : 0));
							list($start_dt, $end_dt) = $this->getPeriodStartEnd($period, $pd_offset, 1, $ftype == 'date');
							if ($negate) {
								$m = " < '$start_dt' OR $sel >= '$end_dt'";
								$negate = '';
								$complex = true;
							} else
								$m = " >= '$start_dt' AND $sel < '$end_dt'";
						}
						else if (!$op) {
							if(in_array($ftype, self::$numeric_types)) {
								$sel = "IFNULL($sel, 0)";
								if($negate)
									$sel = "NOT $sel";
							} else {
								if ($negate)
									$sel = "TRIM(IFNULL($sel, '')) = ''";
								else
									$sel = "TRIM(IFNULL($sel, '')) > ''";
							}
							$m = '';
							$negate = '';
						}
						else {
							$v = $db->quote($val);
							$m = " {$op} '$v'";
							if(is_numeric($val) && in_array($ftype, self::$numeric_types))
								$m = " {$op} $v";
						}

						if (isset($m)) {
							$ret[$k] = $sel . $negate . $m;
							if (isset($complex)) $ret[$k] = '(' . $ret[$k] . ')';
						}
					}
				}
			}
			else if(strlen($clause))
				$ret[$k] = $clause;
		}
		return $ret;
	}

	static function get_filter_clause_fields(array $clauses) {
		$ret = array();
		foreach($clauses as $clause) {
			if(isset($clause['multiple'])) {
				$fs = self::get_filter_clause_fields($clause['multiple']);
				foreach($fs as $f)
					$ret[$f] = 1;
			}
			else {
				if(isset($clause['field']))
					$ret[$clause['field']] = 1;
			}
		}
		return array_keys($ret);
	}

	// non-sql implementation of filters used primarily by workflow
	static function eval_filter_clauses(array &$clauses, $result, $operator='AND', $context=null) {
		$operator = strtoupper($operator);
		$checks = array();
		foreach($clauses as &$clause) {
			if(isset($clause['multiple'])) {
				$m_op = array_get_default($clause, 'operator', 'AND');
				$checks[] = $clause['check'] = self::eval_filter_clauses($clause['multiple'], $result, $m_op, $context);
				continue;
			}

			$op = array_get_default($clause, 'operator');

			if ($op === true) $op = 'true';
			else if ($op === false) $op = 'false';
			else $op = strtolower($op);

			$op_alias = array(
				'eq' => '=', 'not_eq' => '!=',
				'any_of' => 'in', 'not_any_of' => 'not_in',
				'true' => '', 'not_empty' => 'not',
				'false' => 'not', 'empty' => '',
				'gt' => '>', 'gte' => '>=', 'lt' => '<', 'lte' => '<=',
			);

			if(isset($op_alias[$op])) $op = $op_alias[$op];

			if (substr($op, 0, 4) == 'not_' || $op === 'not') {
				$op = substr($op, 4);
				$negate = true;
			} else
				$negate = false;
			if(! empty($clause['negate']))
				$negate = ! $negate;
			$when = array_get_default($clause, 'when');
			$unsupp = false;
			$check = false;
			if (empty($op))
				$op = '=';

			if($when == 'changed_to') {
				$c1 = $c2 = $clause;
				$c1['when'] = 'previous';
				$c1['operator'] = $op;
				$c1['negate'] = !$negate;
				$c2['when'] = 'current';
				$c2['operator'] = $op;
				$c2['negate'] = $negate;
				$paired = [$c1, $c2];
				$check = self::eval_filter_clauses($paired, $result, 'AND', $context);
				$negate = false;
			}
			else if($when == 'updated' || $when == 'not_updated') {
				if(empty($clause['field'])) {
					$unsupp = true;
				} else {
					$f = $clause['field'];
					$fprev = $result->getField($f, null, true);
					$fcurr = $result->getField($f);
					if($when == 'updated')
						$check = ($fprev != $fcurr);
					else
						$check = ($fprev == $fcurr);
				}
			}
			else if ($op == 'literal')
				$unsupp = true;
			else if ($op == 'team')
				$unsupp = true;
			else if($op == 'exists')
				$unsupp = true;
			else if (!empty($clause['field'])) {
				$f = $clause['field'];
				$fval = $result->getField($f, null, ($when == 'previous'));
				$ftype = array_get_default($clause, 'type');
				$format = array_get_default($clause, 'format');
				$val = array_get_default($clause, 'value');
				$params = array();

				if ($op == 'null') {
					$check = $fval === null;
				}
				else if ($op == 'not null') {
					$check = $fval !== null;
				}
				else if ($op == 'length') {
					$check = strlen($fval);
				}
				else if ($op == 'like') {
					$pos = array_get_default($clause, 'match', 'infix');
					if ($pos == 'prefix')
						$check = substr($fval, 0, strlen($val)) == $val;
					else if ($pos == 'suffix')
						$check = substr($fval, - strlen($val)) == $val;
					else if ($pos == 'custom')
						$unsupp = true;
					else if ($pos == 'extended') {
						$vs = preg_split('~,?\s+~', $val);
						$check = true;
						foreach($vs as $v) {
							if(strpos($fval, $v) === false) $check = false;
						}
					}
					else {
						$check = strpos($fval, $val) !== false;
					}
				}
				else if ($op == 'in' || $op == '=' || $op == '!=') {
					if($ftype == 'multienum') {
						$cs = array();
						if(! is_array($val))
							$val = isset($val) ? array($val) : array();
						sort($val);
						if(strlen($fval)) {
							$fc = explode('^,^', $fval);
							sort($fc);
						}
						else {
							$fc = array();
						}
						if($op == '=') {
							$check = $fc == $val;
						} else if($op == '!=') {
							$check = $fc != $val;
						} else {
							// in / any of
							$check = false;
							if(count($val)) {
								foreach($val as $vali) {
									if(in_array($vali, $fc)) $check = true;
								}
							} else
								$check = count($fc);
						}
					}
					else if (is_array($val)) {
						if(! count($val)) {
							$check = false;
						} else if(count($val) == 1) {
							if($op == '!=')
								$check = $fval != $val[0];
							else
								$check = $fval == $val[0];
						} else {
							$check = in_array($fval, $val);
						}
					}
					else {
						$check = ($fval == $val) ^ ($op == '!=');
					}
				}
				else if ($op == 'non_zero') {
					$check = abs($fval) >= 0.005;
				}
				else if ($op == 'between') {
					$start = $val;
					$end = array_get_default($clause, 'end');
					if(isset($start) && isset($end)) {
						$check = $fval >= $start && $fval <= $end;
					}
					else if(isset($start)) {
						$check = $fval >= $start;
					}
					else if(isset($end)) {
						$check = $fval <= $end;
					}
				}
				else if($op == 'date_range') {
					require_once('include/database/DateRange.php');
					$rng = new IAHDateRange($result->getModel(), $context);
					$check = $rng->evalFilter(array_get_default($clause, 'value'), $result, $f, $when);
				}

				// classic date filter operations - don't work
				/*else if ($op == 'before_date' || $op == 'after_date' || $op == 'on_date' || $op == 'between_dates') {
					$start_ts = $start_dt = $end_ts = $end_dt = null;
					if ($ftype == 'datetime') {
						if (preg_match('~^(\d{4})-(\d{2})-(\d{2})$~', $val, $md)) {
							$start_ts = $this->calcDateStartTime((int) $md[1], (int) $md[2], (int) $md[3], true);
							$start_dt = gmdate('Y-m-d H:i:s', $start_ts);
						}
						if ($op == 'on_date' && ! empty($start_dt)) {
							$end_ts = $start_ts + 24 * 3600;
							$end_dt = gmdate('Y-m-d H:i:s', $end_ts);
						}
						else if ($op == 'between_dates') {
							$end_f = array_get_default($clause, 'end');
							if (preg_match('~^(\d{4})-(\d{2})-(\d{2})$~', $end_f, $md)) {
								$end_ts = $this->calcDateStartTime((int) $md[1], (int) $md[2], (int) $md[3], true);
								$end_dt = gmdate('Y-m-d H:i:s', $end_ts + 24 * 3600);
							}
						}
					}
					else {
						if ($ftype != 'date')
							$fval = substr($fval, 0, 10);
						$start_dt = $val;
						if ($op == 'between_dates')
							$end_dt = array_get_default($clause, 'end');
					}

					if($op == 'before_date' && ! empty($start_dt)) {
						$op = 'after_date';
						$end_dt = $start_dt;
						$start_dt = null;
					}
					if(! empty($start_dt) && ! empty($end_dt)) {
						$check = $fval >= $start_dt && $fval <= $end_dt;
					}
					else if(! empty($start_dt)) {
						if($ftype != 'datetime' && $op == 'on_date')
							$check = ($fval == $start_dt);
						else
							$check = ($fval >= $start_dt);
					}
					else if(! empty($end_dt)) {
						$check = ($fval <= $end_dt);
					}
				}
				else if ($op == 'yesterday' || $op == 'today' || $op == 'tomorrow') {
					$start_offs = ($op == 'yesterday' ? -1 : ($op == 'tomorrow' ? +1 : 0));
					$start_ts = $this->relativeDayStartTime($start_offs, true);
					if ($ftype == 'datetime') {
						$end_ts = $start_ts + 24 * 3600;
						$start_dt = gmdate('Y-m-d H:i:s', $start_ts);
						$end_dt = gmdate('Y-m-d H:i:s', $end_ts);
						$check = ($fval >= $start_dt && $fval < $end_dt);
					}
					else {
						if ($ftype != 'date')
							$fval = substr($fval, 0, 10);
						$start_dt = date('Y-m-d', $start_ts);
						$check = ($fval == $start_dt);
					}
				}
				else if ($op == 'period_current' || $op == 'period_next' || $op == 'period_prev') {
					$period = array_get_default($clause, 'period', 'week');
					$pd_offset = ($op == 'period_next' ? 1 : ($op == 'period_prev' ? -1 : 0));
					list($start_dt, $end_dt) = $this->getPeriodStartEnd($period, $pd_offset);
					$check = ($fval >= $start_dt && $fval < $end_dt);
				}*/

				else if (!$op) {
					$check = !! $fval;
				}
				else {
					switch($op) {
						case '>': $check = $fval > $val; break;
						case '<': $check = $fval < $val; break;
						case '>=': $check = $fval >= $val; break;
						case '<=': $check = $fval <= $val; break;
						case '=': $check = $fval == $val; break;
						case '!=': $check = $fval != $val; break;
						default: $unsupp = true; break;
					}
				}
			}

			if($negate) $check = ! $check;
			if($unsupp) {
				$GLOBALS['log']->error("unsupported filter {$result->base_model} $f $op");
				$clause['unsupp'] = true;
			}
			$checks[] = $clause['check'] = $check;
		}

		$any = array_reduce($checks, function($a, $b) { return $a || $b;}, false);
		$all = array_reduce($checks, function($a, $b) { return $a && $b;}, true);
		if($operator == 'AND') $ret = $all;
		else if($operator == 'OR') $ret = $any;
		else if($operator == 'NOT') $ret = ! $any;
		else if($operator == 'NOT_AND') $ret = ! $all;
		else $ret = false;
		return $ret;
	}


	function getSubselectFilter($f, $clause) {
		if(empty($this->fields[$f]))
			return false;
		$fdef = $this->fields[$f];
		if(isset($fdef['source']['alias_field']))
			$src = $fdef['source']['alias_field'];
		else if(isset($fdef['source']['field']))
			$src = $fdef['source']['field'];
		else {
			$link = array_get_default($fdef, 'link');
			$src = $f;
		}
		if(! isset($link)) {
			if(($p = strrpos($src, '.')) !== false) {
				$parent = $this->getLink(substr($src, 0, $p));
				$clause['field'] = substr($src, $p+1);
			}
		} else
			$parent = $this->getLink($link);
		if(empty($parent))
			return false;
		$join_info = $parent->getJoinInfo($this->base_model);
		$lq = new ListQuery($parent, null, array('no_default_order_by' => true));
		$lq->addPrimaryKey();
		$lq->where = $join_info['clause'];
		$lq->addFilterClause($clause);
		return 'EXISTS (' . $lq->getSql() . ')';
	}


	function getPeriodStartEnd($period, $period_offset = 0, $period_count = 1, $as_date=false) {
		global $timedate, $locale;

		if ($period_count < 0) {
			$period_offset += $period_count;
			$period_count *= -1;
		}
		if(substr($period, 0, 5) == 'days_') {
			$days = (int)substr($period, 5);
			$period = 'days_';
		}
		if ($period != 'day' && $period != 'days_')
			$base = $timedate->user_local_time();
		$back_days = $forw_days = 0;
		switch ($period) {
			case 'day':
				$start_dt = $this->relativeDayStartTime($period_offset);
				$end_dt = $this->relativeDayStartTime($period_offset + $period_count);
				break;
			case 'week':
				$first_dow = $locale->getWeekStartDay();
				$dow = idate('w', $base);
				$back_days = $first_dow - $dow;
				if ($back_days > 0) $back_days -= 7;
				$back_days += $period_offset * 7;
				$forw_days = $back_days + 7 * $period_count;
				$start_dt = $this->relativeDayStartTime($back_days);
				$end_dt = $this->relativeDayStartTime($forw_days);
				break;
			case 'month':
				list($cur_year, $cur_month) = explode(' ', date('Y n', $base));
				if ($period_offset) {
					// find beginning of another month (using zero-based cur_month)
					$this->adjustMonthYear($cur_month, $cur_year, $period_offset);
				}
				$start_dt = $this->calcDateStartTime($cur_year, $cur_month, 1);
				$this->adjustMonthYear($cur_month, $cur_year, $period_count);
				$end_dt = $this->calcDateStartTime($cur_year, $cur_month, 1);
				break;
			case 'quarter':
			case 'fiscal_quarter':
				list($cur_year, $cur_month) = explode(' ', date('Y n', $base));
				if ($period == 'fiscal_quarter') {
					$fy_start = (int) $locale->getFiscalYearStart();
					if ($fy_start) $fy_start--;
					$fiscal_offset = $fy_start % 3;
					$months_offset = ($cur_month - 1) % 3;
					$months_offset = ($months_offset - $fiscal_offset + 3) % 3;
				} else {
					$cur_month = floor(($cur_month - 1) / 3) * 3 + 1;
					$months_offset = 0;
				}
				if ($period_offset) {
					// find beginning of another month (using zero-based cur_month)
					$this->adjustMonthYear($cur_month, $cur_year, $period_offset * 3 - $months_offset);
				}
				$start_dt = $this->calcDateStartTime($cur_year, $cur_month, 1);
				$this->adjustMonthYear($cur_month, $cur_year, $period_count * 3);
				$end_dt = $this->calcDateStartTime($cur_year, $cur_month, 1);
				break;
			case 'year':
			case 'fiscal_year':
				if ($period == 'fiscal_year') {
					$fy_start = (int) $locale->getFiscalYearStart();
					if ($fy_start) $start_month = $fy_start;
					list($cur_year, $cur_month) = explode(' ', date('Y n', $base));
					if ($cur_month < $start_month)
						$cur_year--; // fiscal year adjustment
				}
				else {
					$cur_year = idate('Y', $base);
					$start_month = 1;
				}
				$cur_year += $period_offset;
				$start_dt = $this->calcDateStartTime($cur_year, $start_month, 1);
				$end_dt = $this->calcDateStartTime($cur_year + $period_count, $start_month, 1);
				break;
			case 'days_':
				if ($period_offset) {
					if ($period_offset > 0)
						$back_days = $days * ($period_offset - 1);
					else
						$back_days += $days * $period_offset;
				} else {
					$back_days = -floor($days / 2);
				}
				$start_dt = $this->relativeDayStartTime($back_days);
				$end_dt = $this->relativeDayStartTime($back_days + $days * $period_count);
				break;
			default:
				// throw error?
		}
		if($as_date) {
			list($start_dt, $_) = explode(' ', $timedate->handle_offset($start_dt, $timedate->get_db_date_time_format(), true), 2);
			list($end_dt, $_) = explode(' ', $timedate->handle_offset($end_dt, $timedate->get_db_date_time_format(), true), 2);
		}
		return array($start_dt, $end_dt);
	}


	function adjustMonthYear(&$month, &$year, $offset) {
		if (!$offset)
			return;
		// calculate with a zero-based month:
		$month--;
		$month += $offset;
		if ($month >= 12) {
			$year += ceil($month / 12);
			$month %= 12;
		}
		if ($month < 0) {
			$year -= ceil(abs($month) / 12);
			$month %= 12;
			if ($month < 0) $month += 12;
		}
		// return to one-based
		$month++;
	}

	// calculate the start of the day in the current user's timezone - as seen by MySQL
	// based on TimeDate->to_relative_date_time
	function relativeDayStartTime($day_offset = 0, $in_seconds = false) {
		global $timedate;
		$base_time = $timedate->user_local_time();
		$base_info = getdate($base_time);
		return $this->calcDateStartTime($base_info['year'], $base_info['mon'], $base_info['mday'], $in_seconds, $day_offset);
	}

	function calcDateStartTime($year, $month, $day, $in_seconds = false, $day_offset = 0) {
		global $timedate;
		$base_midnight = gmmktime(0, 0, 0, $month, $day, $year);
		$base_midnight += $timedate->get_seconds_offset($base_midnight, false);
		if ($day_offset) $base_midnight += $day_offset * 24 * 3600;
		if ($in_seconds) return $base_midnight;
		return gmdate('Y-m-d H:i:s', $base_midnight);
	}


	function getUnionWhereClause($clause) {
		if (is_array($clause) && isset($clause['union_source'])) {
			if($this->is_union_child && $this->base_model) {
				$alias = $this->base_model->getTableAlias();
				foreach($clause['union_source'] as $k => $v) {
					if($alias == 'union_q_' . $k) {
						return $v;
					}
				}
			}
			return null;
		}
		return $clause;
	}


	// determine which links are involved in the where clause (for count query)
	function getWhereClauseLinks($clauses = null) {
		if (!isset($clauses))
			$clauses = $this->getFilterClauses('links');
		$ret = array();
		if($this->acl_filters && ($parent = $this->getAclParent()) )
			$ret[$parent] = 1;
		if (is_array($clauses)) {
			if(! $this->fields)
				$no_fields_added = true;
			foreach ($clauses as $k => $w) {
				$w = $this->getUnionWhereClause($w);
				if (is_array($w)) {
					if (!empty($w['multiple'])) {
						$l = $this->getWhereClauseLinks($w['multiple']);
						$ret = array_merge($ret, $l);
					}
					else if (!empty($w['field'])) {
						$f = $w['field'];
						if (isset($this->added_field_names[$f]))
							$f = $this->added_field_names[$f];
						if (!isset($this->fields[$f]))
							$f = $this->addField($f, null, false, true);
						$ret = array_merge($ret, $this->getFieldLinks($f, 'where', $err));
					}
				}
			}
			// ensure any newly added fields are marked no_query
			// to preserve the default behaviour of querying '*' when no fields are added
			if(! empty($no_fields_added) && $this->fields) {
				foreach($this->fields as &$f)
					$f['no_query'] = true;
			}
		}
		return $ret;
	}

	function getOrderByLinks($order = null, $inc_grouped=false) {
		$fields = $this->getOrderByFields($order);

		if($inc_grouped && ($grouped = $this->getGroupedFields())) {
			$fields = array_merge($grouped, $fields);
		}

		$ret = array();
		foreach($fields as $f) {
			$ret = array_merge($ret, $this->getFieldLinks($f, 'order_by', $err));
		}
		return $ret;
	}

	function getOrderByFields($order = null) {
		if (!isset($order))
			$order = $this->getOrderBy();
		$ret = array();
		if (is_array($order)) {
			foreach ($order as $k => $ord) {
				if (is_array($ord)) {
					if (!empty($ord['field'])) {
						$ret[] = $ord['field'];
					}
				}
			}
		}
		return $ret;
	}


	function setOrderBy($order) {
		$this->order_by = $order;
	}

	function setGroupBy($group)
	{
		$this->group_by = $group;
	}

	function getGroupedFields() {
		$ret = array();
		foreach($this->fields as $k => $def) {
			if(! empty($def['grouped'])) {
				$ret[] = $k;
			}
		}
		return $ret;
	}

	function getOrderBy($value = null) {
		$ord = null;
		if($value) {
			$ord = self::parse_order_by($value);
		} else if (isset($this->order_by)) {
			$ord = self::parse_order_by($this->order_by);
		} else if(! $this->no_default_order_by && $this->base_model) {
			$ord = self::parse_order_by($this->base_model->getDefaultOrderBy());
		}
		return $ord;
	}

	// this value is meant for display or persistence, not to use in sql
	function getOrderByString($value = null) {
		$ord = $this->getOrderBy($value);
		return self::combine_order_by($ord);
	}

	static function parse_order_by($order_by) {
		if(is_array($order_by))
			return $order_by;
		if(! $order_by)
			return null;
		$order_by = explode(',', $order_by);
		$expr = array();

		foreach($order_by as $part_ord) {
			$part_sort = '';
			if (($p = strrpos($part_ord, ' ')) !== false) {
				$part_sort = strtolower(trim(substr($part_ord, $p+1)));
				if ($part_sort === 'desc' || $part_sort === 'asc') {
					$part_ord = substr($part_ord, 0, $p);
					$part_sort = $part_sort;
				}
				else $part_sort = '';
			}
			$part_ord = trim($part_ord);
			if(! $part_sort) $part_sort = 'asc';

			if(preg_match('#^[~\w][\w\d\.]*$#', $part_ord)) {
				$expr[] = array('field' => $part_ord, 'order' => $part_sort);
			}
		}
		return $expr;
	}

	static function combine_order_by($order_by) {
		$expr = array();
		if(is_array($order_by)) {
			foreach($order_by as $part) {
				if(! is_array($part) || ! isset($part['field']))
					continue;
				$ord = ' ' . strtoupper(array_get_default($part, 'order', 'asc'));
				$f = $part['field'];
				if(isset($part['source']))
					$f = $part['source'] . '.' . $f;
				$expr[] = $f . $ord;
			}
		}
		return implode(', ', $expr);
	}


	// generate order by clause
	function getOrderByClause($have_links = true, $order_by=null) {
		global $db;
		$order_by = $this->getOrderBy($order_by);
		$expr = null;

		$grouped = $this->getGroupedFields();
		if($grouped) {
			if(! is_array($order_by)) $order_by = array();
			foreach(array_reverse($grouped) as $f)
				array_unshift($order_by, array('field' => $f));
		}

		if (count($this->totals)) {
			if(! is_array($order_by)) $order_by = array();
			foreach (array_reverse($this->totals) as $alias => $spec) {
				$sort = array_get_default($spec, 'sort');
				if ($sort) {
					$sort = (strtolower($sort) === 'desc') ? 'desc' : 'asc';
					array_unshift($order_by, array('order_alias' => $alias, 'sort' => $sort));
				}
			}
		}

		if(is_array($order_by)) {
			$expr = array();
			foreach($order_by as $part) {
				if(isset($part['literal'])) {
                    if ($this->disable_literal_types) throw new IAHListQueryError('Literal types forbidden');
					$expr[] = $part['literal'];
					continue;
                } elseif (isset($part['order_alias'])) {
                    $expr[] = $db->quoteField($part['order_alias']).' '.$part['sort'];
                    continue;
				}
				if(! isset($part['field'])) continue;
				$order = ' ' . strtoupper(array_get_default($part, 'order', 'asc'));
				$ptype = array_get_default($part, 'type');
				if (isset($this->fields[$part['field']])) {
					$fdef = $this->fields[$part['field']];
				} else if($part['field'] == '_display') {
					$fdef = $this->getDisplayNameDefinition($errmsg);
				} else {
					$fdef = $this->getFieldDefinition($part['field'], $errmsg);
				}

				if ($fdef) {
					$qdef = $fdef;
					if (isset($qdef['sort_on'])) {
						$link = (!empty($qdef['link'])) ? $qdef['link'] . '.' : '';
						$qdef = $this->getFieldDefinition($link . $qdef['sort_on'], $errmsg);
					}
					if ($qdef) {
						$fields = $this->getFieldSelect($qdef, null, $have_links, 'order_by');
						if ($fields) {
							if(! is_array($fields))
								$fields = array($fields);
							$fields = array_flatten($fields);
							if ($ptype == 'quicksearch') {
								$unif = $this->getUnifiedSearchFields();
								$qlen = strlen($part['prefix']);
								$qpfx = $db->quote($part['prefix']);
								if($unif) {
									$pnam = array();
									foreach($unif as $uf) {
										// prefer exact match on item number
										if($uf['type'] == 'item_number') {
											$psel = $this->getFieldSelect($uf, null, $have_links, 'where');
											$pnam[$uf['name']] = "CAST($psel AS CHAR)='$qpfx'";
										}
									}
									if($pnam)
										$expr[] = 'IF(' . implode(' OR ', $pnam) . ', 0, 1)';
								}
								$fq = $this->getFieldSelect($fdef, null, $have_links, 'where');
								// using CAST to convert non-utf8 fields before performing comparison
								$expr[] = "IF(SUBSTR(CAST($fq AS CHAR),1,$qlen)='$qpfx', 0, 1)";
							}
							else
								$expr[] = implode($order.', ', $fields) . $order;
						}
					}
				}
			}
			$expr = implode(', ', $expr);
		}

		return $expr;
	}


	function addUnionQuery(ListQuery $query, $key=null) {
		$this->is_union = true;
		$query->is_union_child = true;
		if(isset($key)) {
			$query->setTableAlias($key);
			$this->union_queries[$key] = $query;
		} else
			$this->union_queries[] = $query;
	}

	function removeUnionQuery($key) {
		if (isset($this->union_queries[$key]))
			unset($this->union_queries[$key]);
	}

	function getUnionQuery($key) {
		if(isset($this->union_queries[$key]))
			return $this->union_queries[$key];
	}

	function getUnionQueries() {
		return $this->union_queries;
	}

	function isUnionQuery() {
		return $this->is_union;
	}

	// generate union SQL query
	function getUnionSql($as_array = false, $count_only = false, $alias=null, $order_by=null, $distinct = false) {
		global $db;

		if(! $this->union_queries) {
			throw new IAHListQueryError("missing sources for union query");
		}
		$qs =& $this->union_queries;
		if(count($qs) == 1) {
			reset($qs);
			$k = key($qs);
			return $qs[$k]->getSql($as_array, $count_only, null, $order_by, $distinct);
		}

		$sel = array('select' => 'SELECT');
		if ($count_only) {
			if ($distinct)
				$sel['fields'] = "COUNT(DISTINCT $distinct) as c";
			else
				$sel['fields'] = 'COUNT(*) as c';
		} else
			$sel['fields'] = $this->getAllFieldsSelect(false);
		$sel['from'] = 'FROM';

		$sel['union_start'] = '(';
		$idx = 0;
		$ord_fields = $this->getOrderByFields($order_by);
		if($ord_fields)
			$this->addFields($ord_fields, false, 'parent'); // must add sorted fields to use them in order by clause
		foreach($qs as $qk => &$query) {
			$uid = 'union_q_'.$qk;
			$query->no_default_order_by = true;
			$union_sql = $query->getSql(true, false, $uid);
			if($idx)
				$sel['union_op_'.$qk] = 'UNION ALL';
			if($as_array) {
				$sel['union_qstart_'.$qk] = '(';
				$sel[$uid] = $union_sql;
				$sel['union_qend_'.$qk] = ')';
			} else {
				$sel[$uid] = '(' . implode(' ', $union_sql) . ')';
			}
			$idx ++;
		}
		$sel['union_end'] = ')';
		if(! isset($alias))
			$alias = 'union_all';
		$sel['union_alias'] = $db->quoteField($alias);

		$wh = $this->getWhereClause(false, false);
		if ($wh) $sel['where'] = "WHERE $wh";

		if (!$count_only) {
			$ord = $this->getOrderByClause(false, $order_by);
			if ($ord) $sel['order'] = "ORDER BY $ord";
		}
		if ($as_array)
			return $sel;
		return implode(' ', $sel);
	}


	/**
	 * A simple method to construct a DELETE query. Does not support joins.
	 * Uses just the main table name, order and WHERE clause.
	 */
	public function getDeleteSql($as_array = false, $limit = null) {
		$clauses = $this->getSql(true);
		$delete_clauses = [
			'delete' => 'DELETE',
			'from' => $clauses['from'],
			'where' => $clauses['where'],
			'order' => $clauses['order'] ?? '',
			'limit' => (is_numeric($limit) && $limit > 0) ? "LIMIT $limit" : '',
		];
		return $as_array ? $delete_clauses : join(' ', $delete_clauses);
	}


	// generate SQL query
	function getSql($as_array = false, $count_only = false, $alias=null, $order_by=null, $distinct = false) {
		if($this->is_union) {
			return $this->getUnionSql($as_array, $count_only, $alias, $order_by, $distinct);
		}

		/** @var $db DBManager */
		global $db;

		if (isset($this->links['~join']))
			$pmodel =& $this->links['~join'];
		else
			$pmodel =& $this->base_model;
		$table = $pmodel->getTableName();
		$std_alias = $pmodel->getTableAlias();
		if(! isset($alias))
			$alias = $std_alias;
		else
			// temporarily override alias
			$pmodel->setTableAlias($alias);

		$sel_from = $db->quoteField($table);
		if ($alias && $alias != $table)
			$sel_from .= ' ' . $db->quoteField($alias);

		$links_context = $count_only ? 'count' : 'select';
		$qlinks = $this->getActiveLinks($links_context, $order_by);
		$have_links = !!count($qlinks);
		$have_personal = $this->hasPersonalData();
		foreach($this->fields as $f) {
			if(empty($f['no_query']) && empty($f['hidden']) && @$f['source']['type'] == 'subselect') {
				$have_links = true;
				break;
			}
		}

		$sel = array('select' => 'SELECT');
		if ($count_only) {
			if ($distinct !== false)
				$sel['fields'] = "COUNT(DISTINCT $distinct) as c";
			else
				$sel['fields'] = 'COUNT(*) as c';
		} else
			$sel['fields'] = $this->getAllFieldsSelect($have_links || $have_personal);
		$sel['from'] = "FROM $sel_from";

		$js = $this->getQueryJoins(false, $qlinks);
        if ($this->join_favorites && (! $count_only || in_array('favorites', $qlinks)))
            $js .= $this->getFavoritesJoin();
        if (! $count_only && $have_personal && $this->join_personal)
			$js .= $this->getErasedJoin();
		if ($js) $sel['join'] = $js;

		$wh = $this->getWhereClause(false, $have_links || $have_personal);
		if ($wh) $sel['where'] = "WHERE $wh";

		if ($this->group_by) {
			$sel['group_by'] = "GROUP BY $this->group_by";
		} /*else {
			// only in summation queries currently
			$grouped = $this->getGroupedFields();
			if (!empty($grouped)) {
				$gb = array();
				foreach ($grouped as $galias) {
					$gb[] = $db->quoteField($galias);
				}
				$sel['group_by'] = 'GROUP BY ' . implode(', ', $gb);
			}
		}*/

		if (!$count_only) {
			$ord = $this->getOrderByClause($have_links || $have_personal, $order_by);
			if ($ord) $sel['order'] = "ORDER BY $ord";
		}

		// restore previous value
		$pmodel->setTableAlias($std_alias);
		if ($as_array)
			return $sel;
		return implode(' ', $sel);
	}

	function getFavoritesJoin() {
		global $db;
		$k = $this->getPrimaryKey();
		if(isset($this->fields[$k])) $k = $this->fields[$k]['name']; // un-alias if renamed
		$rec_fld = $db->quoteField($k, $this->getTableAlias(null, true));
		$fav_tbl = $db->quoteField(AppConfig::setting("model.detail.Favorite.table_name"));
		$join = " LEFT JOIN ".$fav_tbl." favorites ON favorites.created_by='".AppConfig::current_user_id()."' AND favorites.module='".$this->getModuleDir()."' AND favorites.module_record_id=".$rec_fld;
		return $join;
	}

	function hasPersonalData() {
		if (isset($this->links['~join']))
			$model =& $this->links['~join'];
		else
			$model =& $this->base_model;
		if (!$model) return false;
		$model_name = $model->getModelName();
		return AppConfig::setting("model.detail.{$model_name}.personal_data");
	}

	function getErasedJoin() {
		global $db;
		$k = $this->getPrimaryKey();
		if(isset($this->fields[$k])) $k = $this->fields[$k]['name']; // un-alias if renamed
		$rec_fld = $db->quoteField($k, $this->getTableAlias(null, true));
		$join = " LEFT JOIN erase_personal ON erase_personal.parent_type='".$this->getModuleDir()."' AND erase_personal.parent_id=".$rec_fld;
		return $join;
	}

	function getCountQuery($as_array = false, $distinct = false) {
		return $this->getSql($as_array, true, null, null, $distinct);
	}

	function getSummationQuery($as_array = false) {
		global $db;

		if (isset($this->links['~join']))
			$pmodel =& $this->links['~join'];
		else
			$pmodel =& $this->base_model;
		$table = $pmodel->getTableName();
		$std_alias = $pmodel->getTableAlias();
		if(! isset($alias))
			$alias = $std_alias;
		else
			// temporarily override alias
			$pmodel->setTableAlias($alias);

		$sel_from = $db->quoteField($table);
		if ($alias && $alias != $table)
			$sel_from .= ' ' . $db->quoteField($alias);

		$qlinks = $this->getActiveLinks('sum_select');
		$have_links = !!count($qlinks);
		$have_personal = $this->hasPersonalData();
		$joins = $this->getQueryJoins(false, $qlinks);
		$wh = $this->getWhereClause(false, $have_links || $have_personal);

		$sums = $this->getSummationSelect($qlinks);
		if(! $sums)
			return false;
		// restore previous value
		$pmodel->setTableAlias($std_alias);

		$qs = array();

		foreach($sums['queries'] as $q) {
			$sel = array(
				'select' => 'SELECT',
				'fields' => $q['select'],
				'from' => "FROM $sel_from",
			);
			if($joins) $sel['join'] = $joins;
			if ($wh) $sel['where'] = "WHERE $wh";
			if($q['group'])
				$sel['group_by'] = "GROUP BY {$q['group']}";

			if(! $as_array) {
				$sel = implode(' ', $sel);
			}
			$qs[] = $sel;
		}

		$ret = array();
		if(count($qs) > 1)
			$ret['union_start'] = '(';
		foreach($qs as $idx => $q) {
			if(count($ret) > 1) $ret['union_all_'.$idx] = ') UNION ALL (';
			$ret['union_'.$idx] = $q;
		}
		if(count($qs) > 1)
			$ret['union_end'] = ')';
		if($sums['order_by'])
			$ret['order_by'] = "ORDER BY ".implode(', ', $sums['order_by']);
		else
			$ret['order_by'] = '';

		if ($as_array)
			return $ret;
		return implode(' ', $ret);
	}


	function getSummationGroups() {
		if(! isset($this->summation_groups)) {
			$qlinks = $this->getActiveLinks('sum_select');
			$this->getSummationSelect($qlinks);
		}
		return $this->summation_groups;
	}


	function getSummationSelect($active_links) {
		global $db;

		$src_totals = array(0 => array());
		$all_totals = $this->totals;
		if(! is_array($all_totals)) $all_totals = array();
		$have_links = is_array($active_links) ? count($active_links) : 0;
		$have_personal = $this->hasPersonalData();

		$root_name = $this->base_model->getTableAlias();
		$active_links[] = $root_name;
		$grouping = [];
		$sum_select = array();

		foreach($all_totals as $alias => $spec) {
			if($spec['total'] == 'count' && $spec['type'] != 'bool' && empty($spec['distinct'])) {
				$spec['name'] = 'id';
			}
			$sel = $this->getFieldSelect($spec, null, $have_links || $have_personal, 'sum_expr');
			if($spec['total'] == 'count') {
				if ($spec['type'] == 'bool')
					$sel = "COUNT($sel)";
				else
					$sel = "COUNT(DISTINCT $sel)";
			}
			else
				$sel = strtoupper($spec['total'])."($sel)";
			$ftype = array_get_default($spec, 'format', array_get_default($spec, 'type'));
			if(in_array($ftype, self::$numeric_types))
				$sel = "IFNULL($sel, 0)";
			$source = array_get_default($spec, 'link', $root_name);
			if(($have_links || $have_personal) && $source != $root_name) ;
				//$alias = $source.'.'.$alias;
			$src_totals[$source][$alias] = $sel;
			$sum_select[0][$alias] = $sel;
		}

		foreach($this->fields as $alias => $spec) {
			$source = array_get_default($spec, 'link', $root_name);
			if(in_array($source, $active_links)) {
				if($source != $root_name && empty($grouping[$source]) && ! empty($this->links[$source])) {
					$ltype = array_get_default($this->links[$source]->link_spec, 'link_type');
					$lparent = array_get_default($this->links[$source]->link_spec, 'parent');
					if($ltype == 'nested') {
						$gkey = ($lparent ? $lparent.'.id' : 'id');
						$grouping[$source][$gkey] = array('type' => 'nested', 'field' => $gkey, 'field_type' => 'varchar', 'order' => 'ASC');
					}
				}
				if(! empty($spec['grouped'])) {
					$gkey = ($have_links || $have_personal ? $source.'.'.$alias : $alias);
					$ftype = array_get_default($spec, 'format', array_get_default($spec, 'type'));
					$grouping[$source][$gkey] = array(
						'type' => 'grouped', 'field' => $alias, 'field_type' => $ftype, 'order' => 'ASC');
				}
			}
		}
		if (!isset($grouping[$root_name])) $grouping[$root_name] = [];

		$grp_fields = array();
		$grp_select = array();
		$order_by = array();
		$data_groups = array();
		$depth = 1;

		foreach($grouping as $src => $lvls) {
			$src_sums = array();
			$seen = false;
			foreach($src_totals as $sum_src => $totals) {
				if(! $seen && ($sum_src === $src || $sum_src === 0))
					$seen = true;
				foreach($totals as $alias => $sum) {
					if(! $seen)
						$sum = "'#SKIP#'";
					$src_sums[$alias] = $sum;
				}
			}
			foreach($lvls as $alias => $level) {
				$val = $this->getFieldSelect($level['field'], null, $have_links || $have_personal, 'sum_expr');
				$grp_fields[$depth] = $val;
				$grp_select['field'.$depth] = "'" . $db->quote($alias) . "'";
				$grp_select['value'.$depth] = $val;
				$t = in_array($level['field_type'], self::$numeric_types) ? '+0 ' : ' ';
				$order_by[] = "value$depth".$t.$level['order'];
				$data_groups[$depth] = array('type' => $level['type'], 'source' => $src, 'field' => $level['field']);
				$sum_select[$depth] = $src_sums;
				$t = ' ';
				$depth ++;
			}
		}
		$this->summation_groups = $data_groups;
		if(! $all_totals)
			return false;

		for($dep = $depth-1; $dep >= 0; $dep--) {
			if($dep < $depth-1)
				$grp_select['value'.($dep+1)] = "''";
			$grp_by = '';
			if(count($grp_fields)) {
				$grp_by = implode(', ', $grp_fields);
				array_pop($grp_fields);
			}
			$sum_queries[$dep] = array(
				'select' => "'$dep' as depth, "
						. make_sql_select_expr(array_merge($grp_select, $sum_select[$dep]), false),
				'group' => $grp_by,
			);
		}

		return array(
			'queries' => $sum_queries,
			'order_by' => $order_by,
		);
	}


	/**
	 * @param RowResult|ListResult $result
	 * @param bool $new_record
	 * @return void
	 */
	function initResult(&$result, $new_record = false, $order_by=null) {
		$result->base_model = $this->base_model ? $this->base_model->name : null;

		$order = $this->getOrderBy($order_by);
		if ($new_record)
			$result->new_record = true;
		else {
			$result->query = $this->getSql(false, false, null, $order);
		}

		$result->fields = $this->getFields();
		if(! $result->fields && $this->base_model) {
			$fs = $this->base_model->getFieldDefinitions();
			$keys = array_merge(
				$this->base_model->getColumnFields(),
				$this->base_model->getCustomFields()
			);
			$result->fields = array_intersect_key($fs, array_flip($keys));
		}
		$result->primary_key = $this->getPrimaryKey();
		$result->use_key_index = $this->default_use_key_index;
		if(! is_string($result->primary_key)) {
			$result->primary_key = null;
			$result->use_key_index = false;
		}
		$result->module_dirs = $this->getModuleDirs();
		$result->is_union = $this->is_union;
		$result->order_by = $this->getOrderByString($order);
		if($this->is_union) {
			$src = array();
			foreach($this->union_queries as $uk => &$u) {
				$src[$uk] = array(
					'module' => $u->getModuleDir(),
					'base_model' => $u->getModelName(),
				);
			}
			$result->sources = $src;
		}
		$result->acl_checks = $this->acl_checks;
		$result->acl_owner_field = $this->acl_owner_field;
		$result->acl = array();
	}

	function getBlankResult() {
		$rowResult = new RowResult;
		$this->initResult($rowResult, true);
		$rowResult->row = array();
		$rowResult->acl_checks = array();
		$rowResult->acl_owner_field = $this->acl_owner_field;
		$rowResult->acl = array();
		$currency = null;
		foreach ($rowResult->fields as $f => $def) {
			if (isset($def['default'])) {
				$rowResult->row[$f] = $def['default'];
				if ($def['type'] == 'date') {
					$rowResult->row[$f] = self::date_default_value($rowResult->row[$f]);
				}
			}
			else if($currency !== false && isset($def['type']) && $def['type'] == 'ref' && array_get_default($def, 'bean_name') == 'Currency' && empty($def['no_default'])) {
				if(! isset($currency)) {
					if ($rowResult->base_model && $rowResult->base_model != 'Currency' && ! AppConfig::in_installer()) {
						$cid = UserPreference::getPreference('currency');
						$currency = AppConfig::db_row('Currency', $cid);
					}
					if(! $currency) $currency = false;
				}
				if($currency) {
					$rowResult->row[$def['id_name']] = $currency['id'];
					if(isset($def['exchange_rate']))
						$rowResult->row[$def['exchange_rate']] = $currency['conversion_rate'];
				}
			}
		}
		if ($this->process_results || $this->formatted)
			self::process_result($rowResult);
		if ($this->formatted)
			$this->formatResult($rowResult);
		return $rowResult;
	}

	static function date_default_value($value)
	{
		$defaults = AppConfig::setting('date_defaults');
		if (isset($defaults[$value])) {
			return date('Y-m-d', strtotime($defaults[$value]));
		}
		return null;
	}



	/**
	 * Perform secondary queries for related record names
	 *
	 * @param RowResult|ListResult $result
	 * @param bool $dynamic_only
	 * @return
	 */
	function populateSecondary(&$result, $dynamic_only = true) {
		$fetch_refs = $fetch_queries = array();
			

		if (!$result->getResultCount())
			return;

		$have_refs = false;
		foreach ($result->fields as $k => $fdef) {
			$calculated = false;
			$afield = $fdef['source']['alias_field'] ?? null;
			if ($afield && !empty($this->secondary_calculated[$afield])) {
				$orig = array_get_default($fdef, 'related_value_orig');
				if ($orig) {
					$link = $orig['ref'];
					$link_field = $orig['field'];
					$ref_def = $this->getFieldDefinition($link, $msg);
					if ($ref_def) {
						$calculated = true;
						$fdef = $ref_def;
						$fdef['ref_display_name'] = $link_field;
					}
				}
			}

			if (array_get_default($fdef, 'type') == 'ref') {
				$have_refs = true;
				$prefetch = array_get_default($fdef['source'], 'prefetched');
				$dname = array_get_default($fdef, 'ref_display_name', '_display');
				if (!empty($fdef['dynamic_module'])) {
					$fetch_refs[$k] = array(1, $fdef['dynamic_module'], $fdef['id_name'], $prefetch, $dname);
				}
				else if (!empty($fdef['bean_name']) && ($calculated || !$dynamic_only || ($prefetch && !$this->disable_db_cache))) {
					$fetch_refs[$k] = array(0, $fdef['bean_name'], $fdef['id_name'], $prefetch, $dname);
				}
			}
		}
		if ($have_refs) {
			// this represents whether queries might need to be done (meaning there are ref fields),
			// still true even if ID columns are currently blank
			$result->secondary_queries = true;
		}
		if (!$fetch_refs)
			return;

		if ($result instanceof RowResult)
			$rows = array(& $result->row);
		else {
			$rows =& $result->rows;
		}

		$bean_name = null;
		foreach ($rows as $rowidx => &$row) {
            foreach ($fetch_refs as $f => $refinfo) {
				if ($refinfo[0]) {
					$ref_mod = array_get_default($row, $refinfo[1]);
					if ($ref_mod)
						$bean_name = AppConfig::module_primary_bean($ref_mod);
				} else
					$bean_name = $refinfo[1];
				$ref_id = array_get_default($row, $refinfo[2]);
				if ($bean_name && $ref_id) {
					if ($bean_name == 'User') {
						// FIXME should be handled by next block (DBCache), but user caching needs to be consolidated
						if($this->checkUseRealNames()) {
							$row[$f] = get_assigned_user_name($ref_id);
						} else {
							$row[$f] = get_user_name($ref_id);
						}
					}
					else if (!empty($refinfo[3]) && !$this->disable_db_cache) {
						// if module is pre-fetched, don't do another query for the element names, grab from the cache
						$ref_row = AppConfig::db_row($bean_name, $ref_id);
						$row[$f] = array_get_default($ref_row, $refinfo[4]);
					} else
						$fetch_queries[$bean_name][$refinfo[4]][$ref_id][] = array($rowidx, $f);
				}
			}
		}

		if ($fetch_queries) {
			// create and run secondary queries
			foreach ($fetch_queries as $bean_name => $by_field) {
				foreach($by_field as $field => $ids_row_idx) {
					$lq = new ListQuery($bean_name);
					$lq->no_default_order_by = true;
					$lq->filter_deleted = $this->filter_deleted_secondary;
					if($field == '_display')
						$dn = $lq->addDisplayName($field);
					else
						$dn = $lq->addField($field);
					$pk = $lq->addPrimaryKey();
					if ($pk && $dn) {
						$fetch_ids = array_keys($ids_row_idx);
						for ($i = 0; $i < count($fetch_ids); $i += LQ_BATCH_REQUEST_SIZE) {
							$batch = array_slice($fetch_ids, $i, LQ_BATCH_REQUEST_SIZE);
							$ref_result = $lq->queryRecordSet($batch);
							if (!$ref_result->failed) {
								foreach ($ref_result->rows as $ref_row) {
									$rid = array_get_default($ref_row, $pk);
									if ($rid && isset($ids_row_idx[$rid]))
										foreach ($ids_row_idx[$rid] as $idx_f)
											$rows[$idx_f[0]][$idx_f[1]] = $ref_row[$dn];
								}
							}
						}
					}
				}
			}
		}
	}

	/* Too low-level for FieldFormatter - used in display names
	*/
	static function format_rate_or_amount($rate, $dtype=null, $fixed=null, $fixed_usd=null, $cur_id=null, $exrt=null) {
		if((! $dtype || $dtype == 'percentage') && isset($rate)) {
			return format_number($rate, 3, -1, array('percentage' => true));
		}
		else if($dtype == 'fixed' && isset($fixed)) {
			$fmt = array(
				'convert' => false,
				'entered_currency_id' => $cur_id,
				'currency_id' => $cur_id,
				'exchange_rate' => $exrt,
				'symbol_space' => '',
			);
			return currency_format_number($fixed, $fmt);
		}
		return '';
	}

	/* Too low-level for FieldFormatter - used in display names
	*/
	static function format_name_rate($name, $rate, $dtype=null, $fixed=null, $fixed_usd=null, $cur_id=null, $exrt=null) {
		$value = $name;
		$rate = self::format_rate_or_amount($rate, $dtype, $fixed, $fixed_usd, $cur_id, $exrt);
		if($rate !== '') {
			$value .= " ($rate)";
		}
		return $value;
	}

	/* Too low-level for FieldFormatter - used in display names
	*/
	static function format_person_name($first='', $last='', $salutation='') {
		$format = self::get_name_format();
		$result = $GLOBALS['locale']->getLocaleFormattedName($first, $last, $salutation, $format);
		return $result;
	}

	/**
	 * Populate calculated fields and default values
	 *
	 * @static
	 * @param RowResult|ListResult $result
	 * @param ListQuery $self
	 * @return
	 */
	static function process_result(&$result) {
		if ($result->failed)
			return;

		$proc = array();
		$defaults = array();

		foreach ($result->fields as $k => &$fdef) {
			$args = null;
			$value_fn = null;

			if(isset($fdef['source'])) {
				$sourceType = isset($fdef['source']['type']) ? $fdef['source']['type'] : '';
				if ($sourceType == 'name_rate')
					$value_fn = array('ListQuery', 'format_name_rate');
				else if ($sourceType == 'rate_or_amount')
					$value_fn = array('ListQuery', 'format_rate_or_amount');
				else if ($sourceType == 'person_name')
					$value_fn = array('ListQuery', 'format_person_name');
				else if (isset($fdef['source']['process'])) {
					$value_fn = $fdef['source']['process'];
					$args = array($k);
				}
			}

			// may be a (class, function) pair array
			$value_fn = array_get_default($fdef['source'] ?? [], 'value_function', $value_fn);
			if ($value_fn) {
				if(! isset($args))
					$args = array_get_default($fdef['source'], 'alias_fields', array());
				$proc[$k] = array($value_fn, $args, array_get_default($fdef['source'], 'assoc'));
			}
		}

		$hooks = array();
		$model_name = $result->getModelName();
		if($model_name)
			$hooks = AppConfig::setting("model.hooks.$model_name.process_list_result", array());

		if (count($proc) || count($defaults) || count($hooks)) {
			if ($result instanceof RowResult)
				$rows = array(& $result->row);
			else {
				$rows =& $result->rows;
			}

			foreach ($rows as $rowidx => &$row) {
				foreach($hooks as $hook) {
					call_hook_function($hook, $model_name, array(&$result, &$row, $rowidx));
				}

				foreach ($proc as $field => $spec) {
					$argvals = array();
					foreach ($spec[1] as $f)
						$argvals[$f] = array_get_default($row, $f);

					// try loading beans classes automatically on the first request
					// triggered for situations like value_function: [Forum, get_latest_thread]
					if (is_array($spec[0]) && count($spec[0]) == 2 && !class_exists($spec[0][0])) {
						$bean_details = AppConfig::setting("model.detail." . $spec[0][0]);
						if (!empty($bean_details) && !empty($bean_details['bean_file'])) {
							require_once($bean_details['bean_file']);
						}
					}

					if (empty($spec[2])) {
						$row[$field] = call_user_func_array($spec[0], array_values($argvals));
					} else {
						$row[$field] = call_user_func($spec[0], $argvals);
					}
				}
			}
		}
	}

	/**
	 * @param int $offset
	 * @param int $limit
	 * @param bool $do_count
	 * @param string|array $order_by
	 * @return ListResult
	 */
	function &runQuery($offset = 0, $limit = null, $do_count = false, $order_by = null, $page_size=null, $use_key_index=null) {
		global $db;

		$listResult = new ListResult;
		$listResult->result_source = 'query';
		$this->initResult($listResult, false, $order_by);
		$listResult->total_counted = false;
		if (! $offset) $offset = 0;
		$listResult->offset = $offset;
		$listResult->limit = $limit;
		$listResult->page_size = $page_size;
		$listResult->page_index = 0;
		if(isset($use_key_index))
			$listResult->use_key_index = $use_key_index;
		$listResult->unbuffered = $this->unbuffered;

		// FIXME query one more row if not counting results, so we know when we're at the end
		$listResult->page_result = $db->limitQuery($listResult->query, $offset, $limit, false, '', $this->unbuffered);
		if(! $this->pageResult($listResult))
			return $listResult;

		if(! isset($listResult->page_size)) {
			if (!isset($limit)) {
				$listResult->total_counted = true;
				$listResult->total_count = $listResult->result_count;
			} else {
				if ($do_count && $limit == $listResult->result_count) {
					$count_query = $this->getCountQuery();
					$listResult->count_query = $count_query;
					// TODO add caching here
					$r = $db->query($count_query, false, false, false, $this->unbuffered);
						if (!$r) {
							$listResult->fail($db->getLastError());
							if(AppConfig::setting('site.allow_debug'))
								pr2($count_query . "\n\n" . $db->getLastError(), 'count query error', true);
							return $listResult;
						}
						if ($row = $db->fetchByRow($r, -1, false)) {
						$listResult->total_counted = true;
						$listResult->total_count = (int) $row[0];
					}
				} else {
					$listResult->total_counted = !! $do_count;
					$listResult->total_count = $listResult->result_count + $listResult->offset;
				}
			}
		}

		return $listResult;
	}

	function fetchResultCount($distinct = false) {
		global $db;
		$count_query = $this->getCountQuery(false, $distinct);
		if($count_query) {
			$r = $db->query($count_query, false);
			if ($row = $db->fetchByRow($r, -1, false)) {
				return (int) $row[0];
			}
		}
	}


	function &runSummationQuery($page_size=null) {
		global $db;

		$q = $this->getSummationQuery();
		if(! $q) {$ret = null; return $ret;}

		$listResult = new ListResult;
		$listResult->result_source = 'sum_query';
		$listResult->query = $q;
		$listResult->fields = array();
		$listResult->grouping = $this->summation_groups;
		$listResult->total_counted = false;
		$listResult->offset = 0;
		$listResult->limit = null;
		$listResult->page_size = $page_size;
		$listResult->page_index = 0;

		$listResult->page_result = $db->query($listResult->query, false);
		$this->pageResult($listResult);
		return $listResult;
	}


	function pageResult(ListResult &$listResult) {
		global $db;

		if (! $listResult->page_result) {
			$listResult->fail($db->getLastError());
			return false;
		}
		if($listResult->use_key_index)
			$key = $listResult->primary_key;
		else
			$key = null;
		$psize = $listResult->page_size;
		if (is_array($key)) $key = false;
		else if ($key && ! isset($listResult->fields[$key])) $key = false;
		$idx = 0;
		$listResult->rows = array();
		while (! isset($psize) || $idx < $psize) {
			$row = $db->fetchByAssoc($listResult->page_result, -1, false);
			if($row === false) {
				$listResult->page_result = null;
				$listResult->total_counted = true;
				$listResult->page_finished = true;
				break;
			}
			if ($key)
				$listResult->rows[$row[$key]] = $row;
			else
				$listResult->rows[$idx] = $row;
			$idx++;
		}

		$listResult->result_count = $idx;
		$listResult->total_count += $idx;
		$listResult->page_index ++;

		if ($this->acl_checks)
			$this->populateAcl($listResult);
		if ($this->run_secondary_queries)
			$this->populateSecondary($listResult);
		if ($this->process_results || $this->formatted)
			self::process_result($listResult);
		if ($this->formatted)
			$this->formatResult($listResult);

		return true;
	}

	function freeResult(ListResult $listResult) {
		global $db;
		if($listResult && $listResult->page_result && ! is_resource($listResult->page_result)) {
			$db->freeResult($listResult->page_result);
			$listResult->page_result = null;
		}
	}


	/**
	 * @param string $order_by
	 * @return ListResult
	 */
	function fetchAll($order_by = null) {
		return $this->runQuery(0, null, false, $order_by);
	}

	function fetchAllRows($order_by = null) {
		$result = $this->runQuery(0, null, false, $order_by);
		if($result && ! $result->failed)
			return $result->rows;
		return array();
	}

	// for compatibility with pre-7.0 code only
	function fetchAllObjects($order_by = null, $encode = true) {
		$all = $this->fetchAll($order_by);
		if (!$all || $all->failed)
			throw new IAHListQueryError("Error fetching all objects: model ".$this->getModelName());
		$ret = array();
		foreach($all->getRowIndexes() as $idx) {
			$bean = self::create_bean_object($all->getRowResult($idx), $encode);
			if($bean) $ret[$bean->id] = $bean;
		}
		return $ret;
	}

	function &runQuerySingle($order_by=null, $for_duplicate=false) {
		/** @var DBManager $db  */
		global $db;

		$rowResult = new RowResult;
		$this->no_default_order_by = true;
		$this->initResult($rowResult, false, $order_by);
		if($for_duplicate)
			$rowResult->setForDuplicate();
		$r = $db->limitQuery($rowResult->query, 0, 1, false);
		if (!$r) {
			$rowResult->fail($db->getLastError());
			if(AppConfig::setting('site.allow_debug'))
				pr2($rowResult->query . "\n\n" . $db->getLastError(), 'query error', true);
		}
		else {
			$rowResult->row = $db->fetchByAssoc($r, -1, false);
			if ($rowResult->row === false)
				$rowResult->fail('record not found');
			else {
				if ($for_duplicate)
					$this->clearCalculatedFields($rowResult);
				if ($this->acl_checks)
					$this->populateAcl($rowResult);
				if ($this->run_secondary_queries)
					$this->populateSecondary($rowResult);
				if ($this->process_results || $this->formatted)
					self::process_result($rowResult);
				if ($this->formatted)
					$this->formatResult($rowResult);
			}
		}

		return $rowResult;
	}


	function &queryRecord($id) {
		$primary = $this->addFilterPrimaryKey($id);
		$ret = $this->runQuerySingle();
		$this->removeFilter($primary);
		return $ret;
	}

	function &queryRecordKey($key, $id) {
		$primary = $this->addSimpleFilter($key, $id, '=');
		$ret = $this->runQuerySingle();
		$this->removeFilter($primary);
		return $ret;
	}

	function &queryRecordSet($ids, $offset = 0, $limit = null, $do_count = false) {
		$primary = $this->addFilterPrimaryKey($ids);
		$ret = $this->runQuery($offset, $limit, $do_count);
		$this->removeFilter($primary);
		return $ret;
	}


	static function quick_fetch($model, $id, $fields = true, $params = null) {
		if(! $model) return null;
		$lq = new ListQuery($model, $fields, $params);
		$ret = $lq->queryRecord($id);
		if (!$ret || $ret->failed)
			return null;
		return $ret;
	}

	static function quick_fetch_row($model, $id, $fields = true, $params = null) {
		if(! $model) return null;
		if (!is_array($params)) $params = array();
		if(! isset($params['process_results']))
			$params['process_results'] = false;
		$result = self::quick_fetch($model, $id, $fields, $params);
		if ($result)
			return $result->row;
		return null;
	}

	static function quick_fetch_blank($model, $params=null) {
		if(! $model) return null;
		$lq = new ListQuery($model, null, $params);
		return $lq->getBlankResult();
	}

	static function quick_fetch_all($model, $fields=true, $params=null, $rows=false) {
		if(! $model) return null;
		$lq = new ListQuery($model, $fields, $params);
		$ret = $lq->fetchAll();
		if (!$ret || $ret->failed)
			return null;
		if($rows)
			return $ret->rows;
		return $ret;
	}

	static function quick_fetch_key($model, $key, $id, $fields=true, $params=null) {
		if(! $model) return null;
		$lq = new ListQuery($model, $fields, $params);
		$ret = $lq->queryRecordKey($key, $id);
		if (!$ret || $ret->failed)
			return null;
		return $ret;
	}

	static function quick_fetch_linked($model, $id, $link_name, $fields=true, $params=null, $rows=false) {
		if(! $model) return null;
		if(! $params) $params = array();
		if(! isset($id)) $id = ''; // prevent accidental retrieve with no parent key
		array_extend($params, array('link_name' => $link_name, 'parent_key' => $id));
		$lq = new ListQuery($model, $fields, $params);
		$ret = $lq->fetchAll();
		if (!$ret || $ret->failed)
			return null;
		if($rows)
			return $ret->rows;
		return $ret;
	}

	// for compatibility with pre-7.0 code only
	static function create_bean_object(RowResult &$result, $encode = true) {
		if($result && ! $result->failed) {
			require_bean($result->base_model);
			$bean = new $result->base_model();
			if ($encode)
				$row = array_map('to_html', $result->row);
			else
				$row = $result->row;
			$bean->populateFromRow($row);
			$bean->processed_dates_times = array();
			$bean->check_date_relationships_load();
			return $bean;
		}
	}

	// for compatibility with pre-7.0 code only
	static function quick_fetch_object($model, $id, $params=null, $encode = true) {
		$result = self::quick_fetch($model, $id, null, $params);
		return self::create_bean_object($result, $encode);
	}

	// for compatibility with pre-7.0 code only
	static function quick_fetch_all_objects($model, $params = null) {
		$lq = new ListQuery($model, true, $params);
		return $lq->fetchAllObjects();
	}

	/* Add ACL actions to be checked for the resulting records - does not affect
	   which rows are returned, but may add additional fields to the query
	*/
	function addAclChecks($checks) {
		if(is_string($checks)) $checks = array_map('trim', explode(',', $checks));
		if(is_array($checks)) {
			$uid = $this->getAclUserId();
			$parent = $this->getAclParent();
			$cuid = AppConfig::current_user_id();
			if($parent)
				$aclmod = $this->links[$parent]->getModuleDir();
			else
				$aclmod = $this->base_model->getModuleDir();
			foreach($checks as $chk) {
				if(! $chk || ! empty($this->acl_checks[$chk]))
					continue;

				$c = $chk;

				// FIXME - move into ACLAction
				if($c == 'admin') {
					$spec = array('aclaccess' => ACL_ALLOW_ADMIN);
				} else if($c == 'owner') {
					$spec = array('aclaccess' => ACL_ALLOW_OWNER);
				} else if($c == 'team' || $c == 'group') {
					$spec = array('aclaccess' => ACL_ALLOW_GROUP);
				} else if($c == 'role') {
					$spec = array('aclaccess' => ACL_ALLOW_ROLE);
				} else {
					$spec = ACLAction::getUserActions($uid, false, $aclmod, 'module', $c);
				}

				if(! $spec || $spec['aclaccess'] == ACL_ALLOW_DISABLED) {
					$check = false;
					$allow = false;
					$level = ACL_ALLOW_DISABLED;
				} else {
					$level = $spec['aclaccess'];
					if($level == ACL_ALLOW_ALL) {
						$check = false;
						$allow = true;
					}
					else if($level == ACL_ALLOW_ADMIN) {
						$check = false;
						$allow = ($cuid == $uid && is_admin());
					}
					else if($level == ACL_ALLOW_SYSTEM) {
						$check = false;
						$allow = false;
					}
					else {
						if(! $this->acl_owner_field && (ACLAction::accessLevelRequiresOwnership($level) || ACLAction::accessLevelRequiresTeam($level))) {
							$of = $this->getAclOwnerField($parent);
							if($of)
								$this->acl_owner_field = $this->addField($of);
						}
						if(! $this->acl_query_teams && ACLAction::accessLevelRequiresTeam($level)) {
							$this->acl_query_teams = true;
						}
						$check = true;
						$allow = null;
					}
				}
				$this->acl_checks[$chk] = compact('check', 'allow', 'level');
			}
		}
	}

	function addAclFilter($mode = 'list') {
		if($this->is_union) {
			foreach($this->union_queries as $qid => &$q) {
				$q->acl_user_id = $this->acl_user_id;
				$q->addAclFilter($mode);
			}
			return;
		}
		$this->acl_filters[$mode] = 1;
	}

	function addAclFilters($fs) {
		if(is_string($fs)) $fs = array_map('trim', explode(',', $fs));
		if(is_array($fs)) {
			foreach($fs as $mode => $v) {
				if(is_int($mode)) { $mode = $v; $v = 1; }
				if($v) $this->addAclFilter($mode);
			}
		}
	}

	function getAclOwnerField($link=null) {
		if(! $this->base_model)
			return false;
		if(empty($link)) {
			$model = $this->base_model;
			$base = '';
		}
		else {
			$model = $this->links[$link];
			$base = $link.'.';
		}
		$modelName = $model->getModelName();
		if($modelName == 'User')
			return $base.'id';
		else if ( ($nm = AppConfig::setting("model.detail.$modelName.assigned_user_id")) )
			return $base.$nm;
		else if (AppConfig::setting("model.fields.$modelName.assigned_user_id"))
			return $base.'assigned_user_id';
		elseif (AppConfig::setting("model.fields.$modelName.created_by"))
			return $base.'created_by';
		elseif (AppConfig::setting("model.fields.$modelName.user_id"))
			return $base.'user_id';
		else
			return null;
	}

	function getAclUserId() {
		if($this->acl_user_id) return $this->acl_user_id;
		return AppConfig::current_user_id();
	}

	function setAclUserId($user_id) {
		$this->acl_user_id = $user_id;
	}

	function getAclParent($link=null) {
		if(empty($link)) {
			$model = $this->base_model;
			$base = '';
		}
		else {
			$model = $this->links[$link];
			$base = $link.'.';
		}
		$bean_name = $model->getModelName();
		$aclmod = $model->getModuleDir();
		if($bean_name == AppConfig::module_primary_bean($aclmod)) {
			$parent = AppConfig::setting("modinfo.by_name.$aclmod.acl.parent");
			if($parent) {
				$def = $model->getFieldDefinition($parent);
				if($def) {
					return $this->addLink($base.$def['name']);
				}
			}
		}
	}

	function getAclOwnerClauses($action, $link=null) {
		$uid = $this->getAclUserId();
		$ret = array();
		if( ($p = strpos($action, '.')) !== false) {
			$link = substr($action, 0, $p);
			$action = substr($action, $p+1);
		}
		if(! $this->base_model || $this->is_union)
			return false;
		if( ($parent = $this->getAclParent($link)) )
			$link = $parent;
		if(empty($link)) {
			$model = $this->base_model;
			$base = '';
		}
		else {
			$model = $this->links[$link];
			$base = $link.'.';
		}
		$aclmod = $model->getModuleDir();
		$bean_name = $model->getModelName();
		$level = ACLAction::getUserActions($uid, false, $aclmod, 'module', $action);
		if($aclmod && ACLController::requireOwner($aclmod, $action, $uid, $this->acl_admin_except)
		&& !ACLController::requireSecurityGroup($aclmod, $action, $uid, $this->acl_admin_except)) {
			$field = $this->getAclOwnerField($link);
			if (!empty($field)) {
				$users = ACLAction::getAllowedUsers($uid, $level['aclaccess']);
				if (empty($users)) {
					$own = array(
						'field' => $base.'id',
						'value' => '',
					);
				} else {
					$own = array(
						'field' => $field,
						'value' => $users,
					);
				}
				if($action == 'view' || $action == 'list') {
					// allow users to see group inboxes
					$grp = AppConfig::setting("modinfo.by_name.$aclmod.acl.group_owner");
					if($grp) {
						$own = array(
							'operator' => 'or',
							'multiple' => array(
								$own,
								array('field' => $field, 'value' => $grp),
							),
						);
					}
				}
				$ret[] = $own;
			}
		}
		else {
			if(! $level || (isset($level['aclaccess']) && ($level['aclaccess'] == ACL_ALLOW_NONE || $level['aclaccess'] == ACL_ALLOW_DISABLED || $level['aclaccess'] == ACL_ALLOW_SYSTEM)))
				$ret[] = '0';
			else if(isset($level['aclaccess']) && $level['aclaccess'] == ACL_ALLOW_ADMIN) {
				if(AppConfig::current_user_id() != $uid || ! is_admin())
					$ret[] = '0';
			}
		}
		$filters = AppConfig::setting("model.detail.$bean_name.force_acl.owner.$action");
		if (is_array($filters)) {
			foreach ($filters as $f)
				$ret[] = $f;
		}
		return $ret;
	}

	function getAclGroupClauses($action, $link=null) {
		global $db;
		$uid = $db->quote($this->getAclUserId());
		$ret = array();
		if( ($p = strpos($action, '.')) !== false) {
			$link = substr($action, 0, $p);
			$action = substr($action, $p+1);
		}
		if(! $this->base_model || $this->is_union)
			return false;
		if( ($parent = $this->getAclParent($link)) )
			$link = $parent;
		if(empty($link)) {
			$model = $this->base_model;
			$base = '';
		}
		else {
			$model = $this->links[$link];
			$base = $link.'.';
		}
		$aclmod = $model->getModuleDir();
		$bean_name = $model->getModelName();
		$tableName = $db->quoteField($model->getTableAlias(true));
		$linked = $model->haveLinkDefinition('securitygroups');
		$shared = $model->haveFieldDefinition('shared_with');
		if($linked && ACLController::requireSecurityGroup($aclmod, $action, $uid, $this->acl_admin_except)) {
			$clause = "
			$tableName.id in (
				select sgr.record_id from securitygroups sg
				inner join securitygroups_users sgu on sg.id = sgu.securitygroup_id and sgu.deleted = 0
					and sgu.user_id = '{$uid}'
				inner join securitygroups_records sgr on sg.id = sgr.securitygroup_id and sgr.deleted = 0
					and sgr.module = '$aclmod'
				where sg.deleted = 0
			)";
			if($shared) {
				$clause = array(
					'operator' => 'AND',
					'multiple' => array(
						array(
							'field' => $base.'shared_with',
							'value' => 'teams',
						),
						$clause,
					),
				);
			}
			$owner = $this->getAclOwnerField($link);
			if($owner) {
				$clause = array(
					'operator' => 'OR',
					'multiple' => array(
						array(
							'field' => $owner,
							'value' => $uid,
						),
						$clause,
					),
				);
			}
			if($shared) {
				$clause = array(
					'operator' => 'OR',
					'multiple' => array(
						array(
							'field' => $base.'shared_with',
							'value' => 'all',
						),
						$clause,
					),
				);
			}
			$ret[] = $clause;
		}
		return $ret;
	}

	function getResultIds($result) {
		$idk = $result->primary_key;
		$ids = [];
		if($result instanceof RowResult) {
			$idv = array_get_default($result->row, $idk);
			if($idv) $ids[] = $idv;
		}
		else if($result instanceof ListResult) {
			foreach($result->rows as $idx => $row) {
				$ids[] = array_get_default($row, $idk);
			}
		}
		return $ids;
	}

	function populateAcl($result) {
		global $db;
		if($result->acl_checks) {
			if($this->acl_query_teams) {
				$ids = $this->getResultIds($result);
				if($ids) {
					$parent = $this->getAclParent();
					if($parent)
						$aclmod = $this->links[$parent]->getModuleDir();
					else
						$aclmod = $this->base_model->getModuleDir();
					if($parent) {
						$mod = $this->base_model->getModuleDir();
						$t = $db->quote($this->base_model->getTableName());
						$pt = $db->quote($this->links[$parent]->getTableName());
						$pm = $this->links[$parent]->getModuleDir();
						$idn = $db->quote($this->links[$parent]->link_spec['id_name']);
						$idq = "('".implode("','", $ids)."')";
						$q = "SELECT sgr.securitygroup_id securitygroup_id, '$mod' module, base.id record_id "
							. "FROM $t base "
							. "LEFT JOIN $pt parent ON parent.id=base.$idn AND parent.deleted=0 "
							. "INNER JOIN securitygroups_records sgr ON sgr.record_id=parent.id "
								. "AND sgr.module='$pm' AND sgr.deleted=0 "
							. "WHERE base.id IN $idq";
						$r = $db->query($q);
						if($r) {
							$teams = $db->fetchRows($r, true, false);
							if($teams) {
								$result->acl_teams = array(
									'rows' => $teams,
								);
							}
						}
					}
					else {
						$lq = new ListQuery('securitygroups_records', array('securitygroup_id', 'record_id'));
						$lq->addSimpleFilter('module', $aclmod);
						$lq->addSimpleFilter('record_id', $ids);
						$teams = $lq->fetchAll();
						if($teams) {
							$result->acl_teams = array(
								'rows' => $teams->rows,
							);
						}
					}
				}
				$uid = $this->getAclUserId();
				$user_teams = AppConfig::setting("acl.user_teams.$uid");
			} else
				$user_teams = null;
			if($result instanceof RowResult) {
				$this->populateAclRow($result->acl, $result->row, $result, $user_teams);
			}
			else if($result instanceof ListResult) {
				foreach($result->rows as $idx => $row) {
					$result->acl[$idx] = array();
					$this->populateAclRow($result->acl[$idx], $row, $result, $user_teams);
				}
			}
		}
	}

	function populateAclRow(&$acl, $row, $result, $user_teams) {
		$owner_f = $result->acl_owner_field;
		$pk = $result->primary_key;
		$mod = $this->base_model->getModelName();
		$uid = $this->getAclUserId();
		$in_group = null;
		foreach($result->acl_checks as $action => $info) {
			if(! $info['check'])
				$acl[$action] = $info['allow'];
			else {
				$owner = ($owner_f ? array_get_default($row, $owner_f) : null);
				$pkv = array_get_default($row, $pk);
				if(! isset($in_group)) {
					$in_group = false;
					if($user_teams && $result->acl_teams) {
						foreach($result->acl_teams['rows'] as $row_team) {
							if($row_team['record_id'] == $pkv && isset($user_teams[$row_team['securitygroup_id']])) {
								$in_group = true;
								break;
							}
						}
					}
				}
				$acl[$action] = ACLAction::has_access($uid, $owner, $in_group, $info['level']);
			}
		}
	}

	/** Used to set up records in the process of being duplicated */
	function clearCalculatedFields(RowResult $result) {
		foreach ($result->fields as $k => &$fdef) {
			if ($fdef['type'] == 'sequence_prefix' || $fdef['type'] == 'sequence_number' || $fdef['source']['type'] == 'sequence')
				$result->assign($k, '');
		}
	}

	function formatResult(&$result) {
		require_once('include/layout/FieldFormatter.php');
		$fmt = new FieldFormatter($this->formatted, $this->formatted_context);
		if($result instanceof RowResult)
			$fmt->formatRowResult($result);
		else if($result instanceof ListResult)
			$fmt->formatListResult($result);
	}

	function getIterator($pageSize = 300): Traversable {
		return new ListQueryIterator($this, $pageSize);
	}

	public function disable_literal(){
	    $this->disable_literal_types = true;
    }
}
