<?php
if(!defined('sugarEntry') || !sugarEntry) die('Not A Valid Entry Point');
/*********************************************************************************
 * The contents of this file are subject to the SugarCRM Public License Version
 * 1.1.3 ("License"); You may not use this file except in compliance with the
 * License. You may obtain a copy of the License at http://www.sugarcrm.com/SPL
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * All copies of the Covered Code must include on each user interface screen:
 *    (i) the "Powered by SugarCRM" logo and
 *    (ii) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.  See full license for
 * requirements.
 *
 * The Original Code is: SugarCRM Open Source
 * The Initial Developer of the Original Code is SugarCRM, Inc.
 * Portions created by SugarCRM are Copyright (C) 2004-2006 SugarCRM, Inc.;
 * All Rights Reserved.
 * Contributor(s): ______________________________________.
********************************************************************************/


require_once('include/database/MysqlManager.php');

class MysqliManager extends MysqlManager
{
    var $dbType = 'mysql';
    var $dbLib = 'mysqli';
    var $extraConns = [];
    
    function checkRequirements() {
    	if(! class_exists('mysqli'))
    		throw new IAHDatabaseError("MySQLi extension not loaded");
    }
    
    /**
     * checks for errors and will either log or die depending on the dieOnError
     *
     * @param STRING $msg - message to log
     * @param BOOLEAN $dieOnError - should die on error
     * @return BOOLEAN - error occured returns true
     */
    function checkError($msg='', $dieOnError=false, $conn=null){
        if(DBManager::checkError($msg, $dieOnError, $conn)){
            return true;
        }
        if(! $conn) $conn = $this->database;
        if ($conn->errno){
        	$errtxt = "MySQL error ".$this->getLastError($conn);
            if($this->dieOnError || $dieOnError){
                $GLOBALS['log']->fatal($msg.$errtxt);
                throw new IAHQueryError($msg.$errtxt);
            }else{
                $this->last_error = $msg.$errtxt;
                $GLOBALS['log']->error($msg.$errtxt);
            }
            return true;
        }
        return false;
    }
    
	function getLastError($conn=null) {
		$e = DBManager::getLastError($conn);
		if($e) return $e;
		if(! $conn) $conn = $this->database;
		if (!$conn)
			return "Unknown error";
		if($conn->errno)
			return $conn->errno.": ".$conn->error;
		return false;
	}

    function internalQuery($sql, $conn=null, $unbuffered=false) {
    	if(! $conn) {
			if($unbuffered && $this->database && $this->connection_spec) {
				if(empty($this->extraConns['unbuffered'])) {
					list($_spec, $this->extraConns['unbuffered']) = $this->open_connection($this->connection_spec);
				}
				$conn = $this->extraConns['unbuffered'];
			} else {
				$conn = $this->database;
			}
    	}
    	return $conn->query($sql, $unbuffered ? MYSQLI_USE_RESULT : MYSQLI_STORE_RESULT);
    }

    function freeResult($result=false){
        if(!$result && $this->lastResult){
            $result = current($this->lastResult);
            while($result){
                mysqli_free_result($result);
                $result = next($this->lastResult);
            }
            $this->lastResult = array();
        }
        if($result){
        	$result->close();
            mysqli_free_result($result);
        }
    }


    function getOne($sql, $dieOnError=false, $msg=''){
        $GLOBALS['log']->info("Get One: . |$sql|");
        $queryresult = $this->query($sql, $dieOnError, $msg);
        if (! $queryresult || ! $queryresult->num_rows) $result = false;
        else $result = $queryresult->fetch_row();
        $this->checkError($msg.' Get One Failed:' . $sql . '::', $dieOnError);
        return $result[0];
    }

    /**
     * Returns the description of fields based on the result
     *
     * @param RESULT RESOURCE $result
     * @param boolean $make_lower_case
     * @return ARRAY - field array
     */
    function getFieldsArray($result, $make_lower_case=false)
    {
        $field_array = array();

        if(! isset($result) || empty($result))
        {
            return 0;
        }
        $i = 0;
        $fields = mysqli_fetch_fields($result);
        if($fields)
        	foreach($fields as $f)
        		$field_array[$i++] = $f->name;

        return $field_array;

    }

    /**
     * Returns the number of rows returned by the result
     *
     * @param RESULT RESOURCE $result
     * @return int
     */
    function getRowCount(&$result){
        if(isset($result) && !empty($result)){
            return $result->num_rows;
        }
        return 0;
	}

    /**
     * Returns the number of rows affected
     *
     * @return INT
     */
    function getAffectedRowCount(){
        return $this->database->affected_rows;
    }
    
	function getFieldCount(&$result) {
		return $result->field_count;
	}
	
	function fetchField(&$result, $index) {
		return $result->fetch_field_direct($index);
	}


    /**
     * fetchs the associative array from a database result
     *
     * @param RESULT RESOURCE $result - database result
     * @param ROW NUMBER $rowNum - row number
     * @param BOOLEAN $encode - convert everything for html display
     * @return ARRAY - associative array
     */
    function fetchByAssoc(&$result, $rowNum = -1, $encode=true){
        if(!$result)return false;
        if($result && $rowNum > -1){
            if($this->getRowCount($result) > $rowNum){
                $result->data_seek($rowNum);
            }
            $this->last_row_index = $rowNum;
        }

        $row = $result->fetch_assoc();
        if(! isset($row)) $row = false; // for compatibility

        if($encode && $this->encode && is_array($row))
        	return array_map('to_html', $row);
        return $row;
    }
    
   function fetchByRow(&$result, $rowNum = -1, $encode=true) {
        if(!$result)return false;
        if($result && $rowNum > -1){
            if($this->getRowCount($result) > $rowNum){
                $result->data_seek($rowNum);
            }
            $this->last_row_index = $rowNum;
        }

        $row = $result->fetch_row();
        if(! isset($row)) $row = false; // for compatibility

        if($encode && $this->encode && is_array($row))
        	return array_map('to_html', $row);
        return $row;
    }
    
    function fetchRows(&$result, $assoc=true, $encode=true) {
    	if(!$result) return false;
    	$ret = array();
    	while(1) {
    		if($assoc) $row = $result->fetch_assoc();
    		else $row = $result->fetch_row();
    		if(! isset($row))
    			break;
    		$ret[] = $row;
    	}
    	if($encode && $this->encode) {
    		foreach($ret as $k => $row)
    			$ret[$k] = array_map('to_html', $row);
    	}
    	return $ret;
    }
    

    /**
     * Encodes a string for storing in the database
     *
     * @param STRING $string
     * @param unknown_type $isLike
     * @return STRING
     */
	function quote($string, $isLike=true) {
		switch (gettype($string)) {
		case 'string':
			break;
		case 'integer':
		case 'float':
		case 'double':
		case 'boolean':
			$string = (string)$string;
			break;
		default:
			$string = '';
		}
		return $this->database->real_escape_string($string);
    }
    
    function open_connection($spec, $asAdmin=false) {
		$persistent = array_get_default($spec, 'persistent');
		if(! isset($persistent)) {
			$persistent = AppConfig::setting('site.db_defaults.persistent', false);
			$spec['persistent'] = $persistent;
		}
		$host_name = array_get_default($spec, 'host_name', 'localhost');
		$spec['host_name'] = $host_name;
		if($asAdmin && isset($spec['admin_user'])) {
			$user_name = $spec['admin_user'];
			$password = array_get_default($spec, 'admin_password', '');
		} else {
			$user_name = $spec['user_name'];
			$password = array_get_default($spec, 'password', '');
		}
		$port = array_get_default($spec, 'port'); // null is fine
		$socket = array_get_default($spec, 'socket'); // null is fine
		if(! isset($spec['name']))
			$spec['name'] = '';
		
		if(substr($host_name, 0, 2) == 'p:') {
			$host_name = $spec['host_name'] = substr($host_name, 2);
			$persistent = $spec['persistent'] = true;
		}

		$conn = null;
		if($persistent && version_compare(phpversion(), '5.3.0', '>=')) {
			try {
				$conn = @mysqli_connect("p:$host_name", $user_name, $password, $spec['name'], $port, $socket);
			} catch (mysqli_sql_exception $e) {}
		}

		if(! $conn) {
			try {
				$conn = @mysqli_connect($host_name, $user_name, $password, $spec['name'], $port, $socket);
			} catch (mysqli_sql_exception $e) {}

			if(! $conn)
				throw new IAHDatabaseError("Could not connect to server $host_name: " . mysqli_connect_error());
            if($persistent)
            	$GLOBALS['log']->error("Could not open a persistent database connection, as specified in config");
        }
        $spec['client_version'] = $this->fetchClientVersion($conn);
		$spec['server_version'] = $this->fetchServerVersion($conn);
        
		$cs_info = $this->getDefaultCharset($spec);
		$charset = $spec['charset'] = $cs_info['charset'];
		$collation = $spec['collation'] = $cs_info['collation'];

		$conn->set_charset($charset);
		// note - above statement is preferred by the library, but doesn't support collation. using both methods:
		@$this->internalQuery("SET NAMES '$charset' COLLATE '$collation'", $conn);
		@$this->internalQuery("SET time_zone = '+0:00'", $conn);
		$engine = strtoupper(array_get_default($spec, 'engine', ''));
		if($engine == 'MYISAM' || $engine == 'INNODB')
			@$this->internalQuery("SET storage_engine=$engine", $conn); // no quotes around engine
		else $engine = 'AUTO';
		$spec['engine'] = $engine;
		
		$sql_mode = $this->getSqlMode($spec);
		$spec['sql_mode'] = $sql_mode;
		if($sql_mode)
			@$this->internalQuery("SET SESSION sql_mode='" . join(',', $sql_mode) . "'", $conn);
		
		$this->connection_spec = $spec;
        
        if($conn->errno)
        	throw new IAHDatabaseError('Could not connect: ' . $this->getLastError($conn));
        return [$spec, $conn];
    }

    /**
	 * Frees Results and disconnects the database
	 *
	 */
    function disconnect() {
        if(isset($this->database) && $this->database) {
            $this->freeResult();
            $this->database->close();
            unset($this->database);
        }
        foreach($this->extraConns as $conn) {
        	if($conn) $conn->close();
        }
        $this->extraConns = [];
        $this->cleanup();
    }

	function getClientInfo($conn=null) {
		if(! $conn) $conn = $this->database;
		return $conn->client_info;
	}

	function getServerInfo($conn=null) {
		if(! $conn) $conn = $this->database;
		return $conn->server_info;
	}

    function getHostInfo($conn=null) {
		if(! $conn) $conn = $this->database;
        return $conn->host_info;
    }

    function getStats($conn=null) {
		if(! $conn) $conn = $this->database;
		return $conn->stat();
	}

}

?>
