JSRD = function(input) {
	this.load(input);
};
JSRD.version = '0.9';
JSRD.known_field_types = [
	'any', 'string', 'integer', 'float', 'date', 'time', 'datetime',
	'boolean', 'record', 'sequence', 'map'
];
JSRD.use_date_objects = true;

if(! window.isset) {
	function isset(val) {
		return (typeof(val) != 'undefined' && val !== null);
	}
}
if(! window.is_array) {
	function is_array(val) {
		return isset(val) && typeof(val) == 'object' && val.pop === Array.prototype.pop;
	}
}
if(! window.is_date) {
	function is_date(val) {
		return isset(val) && typeof(val) == 'object' && val.getYear === Date.prototype.getYear;
	}
}
if(! Array.prototype.indexOf) {
	Array.prototype.indexOf = function(val) {
		for(var i = 0; i < this.length; i++)
			if(this[i] == val) return i;
		return -1;
	}
}
if(! Array.prototype.removeAll) {
	Array.prototype.removeAll = function(val) {
		var c = 0;
		for(var idx = 0; idx < this.length; ) {
			if(this[idx] == val) {
				this.splice(idx, 1);
				c++;
			}
			else
				idx++;
		}
		return c;
	}
}
if(! Object.clone) {
	Object.clone = function(obj, deep) {
		var clone = {};
		if(is_array(obj))
			clone = [];
		for(var property in obj) {
			if(deep && typeof obj[property] == 'object' && obj[property] !== null)
				clone[property] = Object.clone(obj[property], deep);
			else
				clone[property] = obj[property];
		}
		return clone;
	}
}
function is_empty_object(obj) {
	if(! isset(obj) || typeof(obj) != 'object')
		return false;
	for(var k in obj)
		return false;
	return true;
}
function zero_pad(val) {
	if(! val) return '00';
	val = ''+val;
	if(val.length == 1) return '0'+val;
	return val;
}
function format_date(val, type) {
	var d = val.getUTCFullYear() + '-' + zero_pad(val.getUTCMonth()+1) + '-' + zero_pad(val.getUTCDate());
	if(type == 'datetime')
		d += ' ' + zero_pad(val.getUTCHours()) + ':' + zero_pad(val.getUTCMinutes()) + ':' + zero_pad(val.getUTCSeconds());
	return d;
}

JSRD.get_boolean = function(val, defval) {
	if(! isset(val)) return defval;
	if(val == 'true' || val == 1 || val === true)
		return true;
	if(val == 'false' || val == 0 || val === false)
		return false;
	return undefined;
}
JSRD.get_integer = function(val, defval) {
	if(! isset(val)) return defval;
	if(typeof(val) == 'string') {
		val = parseInt(val, 10);
		return isNaN(val) ? undefined : val;
	}
	if(typeof(val) == 'number')
		return Math.round(val);
	return undefined;
}
JSRD.get_float = function(val, defval) {
	if(! isset(val)) return defval;
	if(typeof(val) == 'string') {
		val = parseFloat(val);
		return isNaN(val) ? undefined : val;
	}
	if(typeof(val) == 'number')
		return val;
	return undefined;
}
JSRD.get_string = function(val, defval) {
	if(! isset(val)) return defval;
	if(typeof(val) == 'string')
		return val;
	if(typeof(val) == 'number')
		return val.toString();
	return undefined;
}

JSRD.load = function(input) {
	var J = new JSRD();
	JSRD.last_load_error = null;
	if(J.load(input))
		return J;
	JSRD.last_load_error = J.last_load_error;
	return false;
}
JSRD.get_load_error = function() {
	return JSRD.last_load_error;
}

JSRD.prototype.init = function() {
	this.version = JSRD.version;
	this.schema = {};
	this.root_schema = null;
	this.data = null;
	this.loaded = false;
	this.load_load_error = null;
};

JSRD.prototype.error_on_load = function(msg) {
	this.last_load_error = 'JSRD load error: ' + msg;
	return false;
};

JSRD.prototype.load = function(input) {
	this.init();
	if(! isset(input))
		return (this.loaded=true);
	if(! is_array(input) || input.length != 2)
		return this.error_on_load('Expected array of length 2');
	
	var head = input[0];
	if(! isset(head.JSRD))
		return this.error_on_load('Missing JSRD attribute');
	if(head.JSRD != this.version)
		return this.error_on_load('Incompatible version');
	this.version = head.JSRD;
		
	if(! isset(head.schema))
		return this.error_on_load('Missing schema attribute');
	if(! this.load_schemata(head.schema))
		return false;
	if(isset(head.custom) && typeof(head.custom) == 'object')
		this.custom = Object.clone(head.custom, true);
	else
		this.custom = {};
	
	this.data = this.load_data(input[1], this.schema[this.root_schema]);
	if(this.last_load_error)
		return false;
	
	this.loaded = true;
	return true;
};

JSRD.prototype.load_schemata = function(spec) {
	this.schema = {};
	var schema = new JSRD_Schema();
	if(! schema.load(spec, true))
		return this.error_on_load(schema.get_error());
	this.schema[schema.name] = schema;
	this.root_schema = schema.name;

	if(isset(schema.record_types)) {
		for(var i = 0; i < schema.record_types.length; i++) {
			var sub = schema.record_types[i];
			if(isset(this.schema[sub.name]))
				return this.error_on_load('Duplicate schema name: '+sub.name);
			this.schema[sub.name] = sub;
		}
		// FIXME - check here that all referenced schemas are in fact defined
	}
	return true;
};

JSRD.prototype.load_sequence = function(input, schema) {
	if(! is_array(input))
		return this.error_on_load('Expected array (data)');
	var elt_type = schema.get_element_type();
	if(elt_type == 'record') {
		return this.load_records(input, schema);
	} else {
		var ret = [];
		var v;
		for(var idx = 0; idx < input.length; idx++) {
			v = this.load_simple(input[idx], schema);
			if(typeof(v) == 'undefined')
				return false;
			ret.push(v);
		}
		return ret;
	}
};

JSRD.prototype.load_map = function(input, schema) {
	if(typeof(input) != 'object' || is_array(input))
		return this.error_on_load('Expected object (data)');
	var elt_type = schema.get_element_type();
	if(elt_type == 'record') {
		var keys = [];
		var vals = [];
		for(var idx in input) {
			keys.push(idx);
			vals.push(input[idx]);
		}
		var newvals = this.load_records(vals, schema);
		var ret = {};
		for(var i = 0; i < keys.length; i++)
			ret[keys[i]] = newvals[i];
		return ret;
	} else {
		var ret = {};
		var v;
		for(var idx in input) {
			v = this.load_simple(input[idx], schema);
			if(typeof(v) == 'undefined')
				return false;
			ret[idx] = v;
		}
		return ret;
	}
};

JSRD.prototype.load_records = function(input, schema) {
	if(! is_array(input))
		return this.error_on_load('Expected array (record data)');
	var rec_type = schema.get_record_type();
	if(rec_type) {
		if(! isset(this.schema[rec_type]))
			return this.error_on_load("Unknown record type: "+rec_type);
		schema = this.schema[rec_type];
	}

	var fields = schema.fields;
	var stdnms = schema.standard;
	var stdlen = stdnms.length;
	var extnms = schema.extra;
	var extlen = extnms.length;
	var nd = input.length;
	var ret = [];

	for(var idx = 0; idx < input.length; idx++) {
		var row = input[idx];
		var nrow = {};
		var cr = row.length;
		if(cr != stdlen && cr != stdlen + 1)
			return this.error_on_load('Invalid number of field values ('+schema.name+')');
		for(var j = 0; j < stdlen; j++) {
			var v;
			if(row[j] !== null) {
				v = this.load_data(row[j], fields[stdnms[j]]);
				if(typeof(v) == 'undefined')
					return false;
			}
			if(v === null && ! fields[stdnms[j]].nullable)
				return this.error_on_load('Null value is not allowed for "'+stdnms[j]+'"');
			nrow[stdnms[j]] = v;
		}
		if(stdlen < cr) {
			var opts = row[stdlen];
			if(isset(opts) && typeof(opts) != 'object')
				return this.error_on_load('Expected object for extra field values');
			for(var j = 0; j < extnms.length; j++) {
				if(isset(opts[extnms[j]])) {
					var v = this.load_data(opts[extnms[j]], fields[extnms[j]]);
					if(typeof(v) == 'undefined')
						return false;
					nrow[extnms[j]] = v;
				}
			}
		}
		ret.push(nrow);
	}
	return ret;
};

JSRD.prototype.load_data = function(input, schema) {
	var type = schema.get_type();
	if(type == 'sequence') {
		var ret = this.load_sequence(input, schema);
		if(this.last_load_error)
			return false;
		return ret;
	}
	if(type == 'map') {
		var ret = this.load_map(input, schema);
		if(this.last_load_error)
			return false;
		return ret;
	}
	if(type == 'record') {
		var dat = [input];
		var ret = this.load_records(dat, schema);
		if(this.last_load_error)
			return false;
		return ret[0];
	}
	return this.load_simple(input, schema);
};

JSRD.prototype.load_simple = function(data, schema) {
	var type = schema.get_type();
	if(type == 'sequence' || type == 'map')
		type = schema.get_element_type();
	if(type == 'any') {
		return data;
	}
	if(type == 'string') {
		return JSRD.get_string(data);
	}
	if(type == 'date') {
		var d = JSRD.get_string(data);
		if(! d || d == '0000-00-00')
			return null;
		var m;
		if(! (m = d.match(/^(\d{4})-(\d{1,2})-(\d{1,2})$/))) {
			this.last_load_error = "Invalid date value: "+d;
			return undefined;
		}
		if(JSRD.use_date_objects) {
			d = new Date(Date.UTC(m[1], m[2]-1, m[3]));
			if(isNaN(d.getYear())) {
				this.last_load_error = "Invalid date value: "+d;
				return undefined;
			}
		}
		else
			d = m[1] + '-' + zero_pad(m[2]) + '-' + zero_pad(m[3]);
		return d;
	}
	if(type == 'time') {
		var d = JSRD.get_string(data);
		if(! d || d == '00:00:00')
			return null;
		var tm = JSRD.parse_time(d);
		if(tm)
			return zero_pad(tm[0]) + ':' + zero_pad(tm[1]) + ':' + zero_pad(tm[2]);
		else {
			this.last_load_error = "Invalid time value: "+d;
			return undefined;
		}
	}
	if(type == 'datetime') {
		var d = JSRD.get_string(data);
		if(! d || d == '0000-00-00 00:00:00')
			return null;
		var m;
		if(! (m = d.match(/^(\d{4})-(\d{1,2})-(\d{1,2})[ T](.*)$/))) {
			this.last_load_error = "Invalid datetime value: "+d;
			return undefined;
		}
		var tm = JSRD.parse_time(m[4]);
		if(! tm) {
			this.last_load_error = "Invalid datetime value: "+d;
			return undefined;
		}
		if(JSRD.use_date_objects) {
			d = new Date(Date.UTC(m[1], m[2]-1, m[3], tm[0], tm[1], tm[2]));
			if(isNaN(d.getYear())) {
				this.last_load_error = "Invalid datetime value: "+d;
				return undefined;
			}
		}
		else
			d = m[1] + '-' + zero_pad(m[2]) + '-' + zero_pad(m[3])
			  + ' ' + zero_pad(tm[0]) + ':' + zero_pad(tm[1]) + ':' + zero_pad(tm[2]);
		return d;
	}
	if(type == 'boolean') {
		return JSRD.get_boolean(data);
	}
	if(type == 'integer') {
		return JSRD.get_integer(data);
	}
	if(type == 'float') {
		return JSRD.get_float(data);
	}
	var t = 'coerce_'+type;
	if(typeof(this[t]) == 'function')
		return this[t](data, schema);
	this.last_load_error = 'Error loading field value of type '+type;
	return undefined;
};

JSRD.parse_time = function(t) {
	if(! t || ! (m = t.match(/(\d{1,2}):(\d{1,2})(:(\d{1,2}))?(Z|([+-]?)(\d{2}):?(\d{2})?)?/)))
		// should raise an error for invalid times
		return undefined;
	var hr = parseInt(m[1]);
	var min = parseInt(m[2]);
	if(m[5] && m[5] != 'Z') {
		var sgn = (m[6] == '-' ? -1 : 1);
		hr += parseInt(m[7]) * sgn;
		if(m[8])
			min += parseInt(m[8]) * sgn;
		if(min < 0) { min += 60; hr --; }
		else if(min > 60) { min -= 60; hr ++; }
		if(hr < 0) hr += 24;
		else hr %= 24;
	}
	return [hr, min, parseInt(m[4])];
};

JSRD.prototype.encode_schemata = function(minimal) {
	var ret = {};
	var nested = [];
	for(var k in this.schema) {
		var o = this.schema[k].encode(minimal);
		if(k == this.root_schema)
			ret = o;
		else
			nested.push(o);
	}
	if(nested.length)
		ret.record_types = nested;
	return ret;
};

JSRD.prototype.encode_sequence = function(schema, data) {
	var elt_type = schema.get_element_type();
	if(elt_type == 'record') {
		return this.encode_records(schema, data);
	} else {
		var ret = [];
		for(var i = 0; i < data.length; i++)
			ret.push(this.encode_simple(schema, data[i]));
		return ret;
	}
};

JSRD.prototype.encode_map = function(schema, data) {
	var elt_type = schema.get_element_type();
	if(elt_type == 'record') {
		var keys = [];
		var vals = [];
		for(var idx in data) {
			keys.push(idx);
			vals.push(data[idx]);
		}
		var newvals = this.encode_records(schema, vals);
		var ret = {};
		for(var i = 0; i < keys.length; i++)
			ret[keys[i]] = newvals[i];
		return ret;
	} else {
		var ret = {};
		for(var idx in data)
			ret[idx] = this.encode_simple(schema, data[idx]);
		return ret;
	}
};

JSRD.prototype.encode_records = function(schema, data) {
	var rec_type = schema.get_record_type();
	if(isset(rec_type))
		schema = this.schema[rec_type];

	var stdnms = schema.standard;
	var stdlen = stdnms.length;
	var extnms = schema.extra;
	var extlen = extnms.length;
	var ret = [];
	for(var i = 0; i < data.length; i++) {
		var row = data[i];
		var arr = [];
		var entry;
		for(var j = 0; j < stdnms.length; j++) {
			entry = this.encode_data(schema.fields[stdnms[j]], row[stdnms[j]]);
			arr.push(entry);
		}
		if(extnms.length) {
			var opt = {};
			for(var j = 0; j < extnms.length; j++) {
				if(isset(row[extnms[j]]))
					opt[extnms[j]] = this.encode_data(schema.fields[extnms[j]], row[extnms[j]]);
			}
			if(! is_empty_object(opt))
				arr.push(opt);
		}
		ret.push(arr);
	}
	return ret;
};

JSRD.prototype.encode_data = function(schema, data) {
	var ret;
	var type = schema.get_type();
	if(type == 'sequence')
		ret = this.encode_sequence(schema, data);
	else if(type == 'map')
		ret = this.encode_map(schema, data);
	else if(type == 'record') {
		var dat = [data];
		var enc = this.encode_records(schema, dat);
		ret = enc[0];
	}
	else
		ret = this.encode_simple(schema, data);
	return ret;
};

JSRD.prototype.encode_simple = function(schema, data) {
	var ret;
	if(is_date(data)) {
		ret = format_date(data, schema.type);
	}
	else {
		ret = data;
	}
	return ret;
};

JSRD.prototype.encode_to_object = function(minimal) {
	var head = {JSRD: this.version};
	head.schema = this.encode_schemata(minimal);
	if(! is_empty_object(this.custom))
		head.custom = this.custom;
	var body = this.encode_data(this.schema[this.root_schema], this.data);
	return [head, body];
};

JSRD.prototype.encode = function() {
	var obj = this.encode_to_object();
	return JSON.stringify(obj);
};


// -- JSRD_Schema

JSRD_Schema = function() {
};

JSRD_Schema.prototype.load = function(spec, is_root) {
	if(typeof(spec) == 'string')
		spec = {'name': spec};
	if(typeof(spec) != 'object' || is_array(spec))
		return this.error('Expected string or object for schema');
	if(! isset(spec.name))
		return this.error('Missing schema name');
	if(typeof(spec.name) != 'string')
		return this.error('Expected string (schema name)');
	this.name = spec.name;
	
	var default_type, default_elt_type;
	if(isset(spec.record_type) || isset(spec.fields))
		default_elt_type = 'record';
	else
		default_elt_type = 'any';
	if(isset(spec.element_type))
		default_type = 'sequence';
	else
		default_type = default_elt_type;
	
	if(! isset(spec.type)) {
		spec.type = default_type;
	}
	else if(typeof(spec.type) != 'string')
		return this.error('Expected string (schema type)');
	if(JSRD.known_field_types.indexOf(spec.type) < 0)
		return this.error('Unknown schema type: '+spec.type);
	this.type = spec.type;
	
	this.label = JSRD.get_string(spec.label, null);
	if(isset(spec.custom)) {
		if(typeof(spec.custom) != 'object' || is_array(spec.custom))
			return this.error('Expected object (custom)');
		this.custom = Object.clone(spec.custom, true);
	}
	
	if(this.is_collection()) {
		if(! isset(spec.element_type))
			spec.element_type = default_elt_type;
		if(JSRD.known_field_types.indexOf(spec.element_type) < 0)
			return this.error('Unknown schema element type: '+spec.element_type);
		this.element_type = spec.element_type;
	}
	if(this.type == 'record' || (this.is_collection() && this.element_type == 'record')) {
		if(isset(spec.fields)) {
			if(isset(spec.record_type))
				return this.error('Schema specifies both fields and record_type');
			if(! this.load_fields(spec)) {
				return false;
			}
		}
		else {
			if(! isset(spec.record_type) || typeof(spec.record_type) != 'string')
				return this.error('Expected string (schema record type)');
			this.record_type = spec.record_type;
		}
	}
	
	//if(! this.is_collection()) {
		var maxlen = JSRD.get_integer(spec.maxlen, null);
		if(isset(maxlen)) {
			if(maxlen <= 0)
				return this.error('Expected positive integer (maxlen)');
			this.maxlen = maxlen;
		}
		this.nullable = JSRD.get_boolean(spec.nullable, true);
		this.format = JSRD.get_string(spec.format, null);
	//}
	
	if(isset(spec.record_types)) {
		if(! is_root)
			return this.error('Field record_types defined for non-root schema');
		if(! is_array(spec.record_types))
			return this.error('Expected array (record_types)');
		this.record_types = [];
		for(var i = 0; i < spec.record_types.length; i++) {
			var sub = new JSRD_Schema();
			if(! sub.load(spec.record_types[i], true))
				return this.error(sub.get_error());
			this.record_types.push(sub);
		}
	}
	
	return true;
};

JSRD_Schema.prototype.load_fields = function(spec) {
	this.fields = {};
	this.standard = [];
	this.extra = [];
	var fields = spec.fields;
	if(! is_array(fields))
		return this.error('Expected: array (fields)');
	var extra = spec.extra_fields;
	if(! is_array(extra))
		extra = [];
	var ok = false;
	for(var i = 0; i < fields.length; i++) {
		var schema = new JSRD_Schema();
		if(! schema.load(fields[i]))
			return this.error(schema.get_error());
		if(isset(this.fields[schema.name]))
			return this.error('Duplicate field name: '+schema.name);
		this.fields[schema.name] = schema;
		this.standard.push(schema.name);
		ok = true;
	}
	for(var i = 0; i < extra.length; i++) {
		var schema = new JSRD_Schema();
		if(! schema.load(extra[i]))
			return this.error(schema.get_error());
		if(isset(this.fields[schema.name]))
			return this.error('Duplicate field name: '+schema.name);
		this.fields[schema.name] = schema;
		this.extra.push(schema.name);
	}
	if(! ok)
		return this.error('No fields defined');
	return true;
};

JSRD_Schema.prototype.get_type = function() {
	return this.type;
};
JSRD_Schema.prototype.get_element_type = function() {
	return this.element_type;
};
JSRD_Schema.prototype.get_record_type = function() {
	return this.record_type;
};
JSRD_Schema.prototype.is_collection = function() {
	return (this.type == 'sequence' || this.type == 'map');
};
JSRD_Schema.prototype.get_custom = function() {
	return this.custom;
};
JSRD_Schema.prototype.set_custom = function(custom) {
	this.custom = custom;
};

JSRD_Schema.prototype.error = function(msg) {
	this.error_text = msg;
	return false;
};
JSRD_Schema.prototype.get_error = function() {
	return this.error_text;
};
	
JSRD_Schema.prototype.encode = function(minimal) {
	var field_keys = [
		'name', 'label', 'type', 'element_type', 'record_type',
		'maxlen', 'nullable', 'format', 'custom', 'fields', 'extra'
	];
	var ret = {};
	var simple = true;
	for(var idx = 0; idx < field_keys.length; idx++) {
		var key = field_keys[idx];
		if(isset(this[key])) {
			if(key == 'custom') {
				if(! is_empty_object(this[key])) {
					ret[key] = Object.clone(this[key], true);
					simple = false;
				}
			}
			else if(key == 'type' && this[key] == 'any')
				; // pass
			else if(key == 'nullable' && this[key])
				; // pass
			else if(key == 'fields' || key == 'extra') {
				if(this.type == 'record' || this.element_type == 'record') {
					var target = (key == 'fields') ? key : 'extra_fields';
					var flds = this.encode_fields(key == 'extra', minimal);
					if(flds.length)
						ret[target] = flds;
					simple = false;
				}
			}
			else {
				if(key != 'name')
					simple = false;
				ret[key] = this[key];
			}
		}
	}
	if(simple)
		ret = ret.name;
	return ret;
}
	
JSRD_Schema.prototype.encode_fields = function(extra, minimal) {
	var ret = [];
	var target = (extra ? 'extra' : 'standard');
	var nms = this[target];
	for(var idx = 0; idx < nms.length; idx++) {
		ret.push(this.fields[nms[idx]].encode(minimal));
	}
	return ret;
}
	
JSRD_Schema.prototype.add_field = function(spec, extra) {
	if(! this.fields) {
		this.fields = {};
		this.standard = [];
		this.extra = [];
	}
	
	var fld = new JSRD_Schema();
	if(! fld.load(spec))
		return this.error(fld.get_error());
	if(isset(this.fields[fld.name]))
		this.remove_field(fld.name);
	
	var set_f = extra ? 'extra' : 'standard';
	this.fields[fld.name] = fld;
	this[set_f].push(fld.name);
	return true;
}

JSRD_Schema.prototype.remove_field = function(name) {
	var ok = false;
	if(isset(this.fields[name])) {
		delete this.fields[name];
		this.standard.removeAll(name);
		this.extra.removeAll(name);
		ok = true;
	}
	return ok;
}
