<?php

require_once('include/JSON.php');

if(! function_exists('array_get_default')) {
	function array_get_default($arr, $key, $default=null) {
		if(! isset($arr[$key])) return $default;
		return $arr[$key];
	}
}

class JSRD {

	var $schema;
	var $root_schema;
	var $data;
	var $loaded;
	var $last_load_error;
	
	static $version = '0.9';
	

	function __construct() {
		$this->init();
	}
	
	function init() {
		$this->schema = array();
		$this->root_schema = null;
		$this->data = null;
		$this->custom = array();
		$this->loaded = false;
	}
	

	static function get_boolean($val, $defval=null) {
		if(is_null($val))
			return $defval;
		if($val == 'true' || $val == 1 || $val === true)
			return true;
		if($val == 'false' || $val == 0 || $val === false)
			return false;
		return null;
	}
	static function get_integer($val, $defval=null) {
		if(is_integer($val))
			return $val;
		if(is_float($val))
			return (int)$val;
		if(is_string($val)) {
			$val = sscanf("%d", $val);
			//return isNaN(val) ? undefined : val;
		}
		return null;
	}
	static function get_float($val, $defval=null) {
		if(is_integer($val))
			return (float)$val;
		if(is_float($val))
			return $val;
		if(is_string($val)) {
			$val = sscanf("%f", $val);
			//return isNaN(val) ? undefined : val;
		}
		return null;
	}
	static function get_string($val, $defval=null) {
		if(is_string($val))
			return $val;
		if(is_integer($val) || is_float($val))
			return ''.$val;
		return null;
	}


	function load($str) {
		static $json;
		if(! isset($json))
			$json = new JSON(JSON_LOOSE_TYPE);
		if(is_array($str))
			return JSRD::load_array($str);
		$dat = $json->decode($str);
		return JSRD::load_array($dat);
	}
	
	function load_array($input) {
		// can be called statically or as an object method
		if(! isset($this) || !($this instanceof JSRD)) {
			$obj = new JSRD();
			$ret = $obj->load_array($input);
			if($ret)
				return $obj;
			return $ret;
		}
		$this->init();
		
		if(! is_array($input) || count($input) != 2)
			return $this->error_on_load('Expected array of length 2');
		$head = $input[0];
		if(! isset($head['JSRD']))
			return $this->error_on_load('Missing JSRD attribute');
		if($head['JSRD'] != JSRD::$version)
			return $this->error_on_load('Incompatible version');
		$this->file_version = $head['JSRD'];

		if(! isset($head['schema']))
			return $this->error_on_load('Missing schema attribute');
		if(! $this->load_schemata($head['schema'], true))
			return false;
		if(isset($head['custom']) && is_array($head['custom']))
			$this->custom = $head['custom'];
		
		$undef = false;
		$this->data =& $this->load_data($input[1], $this->schema[$this->root_schema], $undef);
		if($undef)
			return false;
		
		$this->loaded = true;
		return true;
	}
	
	function error_on_load($msg) {
		$this->last_load_error = 'JSRD load error: ' . $msg;
		return false;
	}
	function last_load_error() {
		return $this->last_load_error;
	}
	
	function load_schemata($spec) {
		$this->schema = array();
		$schema = new JSRD_Schema();
		if(! $schema->load($spec, true))
			return $this->error_on_load($schema->get_error());
		$this->schema[$schema->name] =& $schema;
		$this->root_schema = $schema->name;

		if(isset($schema->record_types)) {
			foreach($schema->record_types as $sub) {
				if(isset($this->schema[$sub->name]))
					return $this->error_on_load('Duplicate schema name: '.$sub->name);
				$this->schema[$sub->name] = sub;
			}
			// FIXME - check here that all referenced schemas are in fact defined
		}
		return true;
	}
	
	function load_sequence($input, &$schema) {
		if(! is_array($input))
			return $this->error_on_load('Expected array (data)');
		$elt_type = $schema->get_element_type();
		if($elt_type == 'record') {
			return $this->load_records($input, $schema);
		} else {
			$ret = array();
			foreach($input as $v) {
				$undef = false;
				$v = $this->load_simple($v, $schema, $undef);
				if($undef)
					return false;
				$ret[] = $v;
			}
			return $ret;
		}
	}
	
	function load_map($input, &$schema) {
		if(! is_array($input))
			return $this->error_on_load('Expected array (data)');
		$elt_type = $schema->get_element_type();
		if($elt_type == 'record') {
			$keys = array_keys($input);
			$vals = array_values($input);
			$newvals = $this->load_records($vals, $schema);
			return array_combine($keys, $newvals);
		} else {
			$ret = array();
			foreach($input as $k => $v) {
				$undef = false;
				$v = $this->load_simple($v, $schema, $undef);
				if($undef)
					return false;
				$ret[$k] = $v;
			}
			return $ret;
		}
	}

	function load_records($input, &$for_schema) {
		if(! is_array($input))
			return $this->error_on_load('Expected array (record data)');
		$rec_type = $for_schema->get_record_type();
		if($rec_type) {
			if(! isset($this->schema[$rec_type]))
				return $this->error_on_load("Unknown record type: ".$rec_type);
			$schema =& $this->schema[$rec_type];
		}
		else
			$schema =& $for_schema;

		$ret = array();
		$fields =& $schema->fields;
		$stdnms =& $schema->standard;
		$stdlen = is_countable($stdnms) ? count($stdnms) : 0;
		$extnms =& $schema->extra;
		$extlen = is_countable($extnms) ? count($extnms) : 0; 
		$nd = is_countable($input) ? count($input) : 0; 
		foreach($input as $row) {
			$nrow = array();
			$cr = is_countable($row) ? count($row) : 0;
			if($cr != $stdlen && $cr != $stdlen + 1)
				return $this->error_on_load('Invalid number of field values');
			for($j = 0; $j < $stdlen; $j++) {
				unset($v);
				$v = null;
				if($row[$j] !== null) {
					$undef = false;
					$v =& $this->load_data($row[$j], $fields[$stdnms[$j]], $undef);
					if($undef)
						return false;
				}
				if($v === null && ! $fields[$stdnms[$j]]->nullable)
					return $this->error_on_load('Null value is not allowed for "'.$stdnms[$j].'"');
				$nrow[$stdnms[$j]] = $v;
			}
			if($stdlen < $cr) {
				$opts = $row[$stdlen];
				if(isset($opts) && ! is_array($opts))
					return $this->error_on_load('Expected array for extra field values');
				for($j = 0; $j < $extlen; $j++) {
					unset($v);
					if(isset($opts[$extnms[$j]])) {
						$undef = false;
						$v =& $this->load_data($opts[$extnms[$j]], $fields[$extnms[$j]], $undef);
						if($undef)
							return false;
						$nrow[$extnms[$j]] = $v;
					}
				}
			}
			$ret[] = $nrow;
		}
		return $ret;
	}

	function &load_data($input, &$schema, &$undef) {
		$type = $schema->get_type();
		if($type == 'sequence') {
			$ret = $this->load_sequence($input, $schema);
			if($this->last_load_error) {
				$undef = true;
				$ret = false;
			}
		}
		else if($type == 'map') {
			$ret = $this->load_map($input, $schema);
			if($this->last_load_error) {
				$undef = true;
				$ret = false;
			}
		}
		else if($type == 'record') {
			$dat = array();
			$dat[0] =& $input;
			$ret = $this->load_records($dat, $schema);
			if(! is_array($ret) || $this->last_load_error) {
				$undef = true;
				$ret = false;
			}
			else
				$ret = $ret[0];
		}
		else
			$ret = $this->load_simple($input, $schema, $undef);
		return $ret;
	}

	function load_simple($data, &$schema, &$undef) {
		static $date_pat = '~^(\d{4})-(\d{1,2})-(\d{1,2})$~';
		static $datetime_pat = '~^(\d{4})-(\d{1,2})-(\d{1,2})[ T](.*)$~';
		$type = $schema->get_type();
		if($type == 'sequence' || $type == 'map')
			$type = $schema->get_element_type();
		if($type == 'any') {
			return $data;
		}
		if($type == 'string') {
			return JSRD::get_string($data);
		}
		if($type == 'date') {
			$d = JSRD::get_string($data);
			if(empty($d) || $d == '0000-00-00')
				return null;
			if(! preg_match($date_pat, $d, $m)) {
				$this->last_load_error = "Invalid date value: ".$d;
				$undef = true;
				return null;
			}
			$d = sprintf('%d-%02d-%02d', $m[1], $m[2], $m[3]);
			return $d;
		}
		if($type == 'time') {
			$d = JSRD::get_string($data);
			if(empty($d) || $d == '00:00:00')
				return null;
			$tm = JSRD::parse_time($d);
			if(! $tm) {
				$this->last_load_error = "Invalid time value: ".$d;
				$undef = true;
				return null;
			}
			return sprintf('%02d:%02d:%02d', $tm[0], $tm[1], $tm[2]);
		}
		if($type == 'datetime') {
			$d = JSRD::get_string($data);
			if(empty($d) || $d == '0000-00-00 00:00:00')
				return null;
			if(! preg_match($datetime_pat, $d, $m)) {
				$this->last_load_error = "Invalid datetime value: ".$d;
				$undef = true;
				return null;
			}
			$tm = JSRD::parse_time($m[4]);
			if(! $tm) {
				$this->last_load_error = "Invalid datetime value: ".$d;
				$undef = true;
				return null;
			}
			$d = sprintf('%d-%02d-%02d %02d:%02d:%02d',
				$m[1], $m[2], $m[3], $tm[0], $tm[1], $tm[2]);
			return $d;
		}
		if($type == 'boolean') {
			return JSRD::get_boolean($data);
		}
		if($type == 'integer') {
			return JSRD::get_integer($data);
		}
		if($type == 'float') {
			return JSRD::get_float($data);
		}
		$t = 'coerce_'.$type;
		if(method_exists($this, $t))
			return $this->$t($data, $schema);
		$this->last_load_error = 'Error loading field value of type '.$type;
		$undef = true;
		return null;
	}
	
	function parse_time($t) {
		static $pat = '~^(\d{1,2}):(\d{1,2})(:(\d{1,2}))?(Z|([+-]?)(\d{2}):?(\d{2})?)?$~';
		if(! $t || ! preg_match($pat, $t, $m))
			// should raise an error for invalid times
			return null;
		$hr = (int)$m[1];
		$min = (int)$m[2];
		if(! empty($m[5]) && $m[5] != 'Z') {
			$sgn = ($m[6] == '-' ? -1 : 1);
			$hr += ((int)$m[7]) * $sgn;
			if($m[8])
				$min += ((int)$m[8]) * $sgn;
			if($min < 0) { $min += 60; $hr --; }
			else if($min > 60) { $min -= 60; $hr ++; }
			if($hr < 0) $hr += 24;
			else $hr %= 24;
		}
		return array($hr, $min, (int)array_get_default($m, 4, '0'));
	}
	
	function &encode_schemata($minimal=false) {
		$ret = array();
		$nested = array();
		foreach(array_keys($this->schema) as $k) {
			$o = $this->schema[$k]->encode($minimal);
			if($k == $this->root_schema)
				$ret = $o;
			else
				$nested[] = $o;
		}
		if($nested)
			$ret['record_types'] = $nested;
		return $ret;
	}

	function &encode_sequence(&$schema, &$data) {
		$elt_type = $schema->get_element_type();
		if($elt_type == 'record') {
			return $this->encode_records($schema, $data);
		} else {
			$ret = array();
			foreach($data as $entry)
				$ret[] = $this->encode_simple($schema, $entry);
			return $ret;
		}
	}
	
	function &encode_map(&$schema, &$data) {
		$elt_type = $schema->get_element_type();
		if($elt_type == 'record') {
			$keys = array_keys($data);
			$vals = array_values($data);
			$newvals =& $this->encode_records($schema, $vals);
			return array_combine($keys, $newvals);
		} else {
			$ret = array();
			foreach($data as $k => $entry)
				$ret[$k] = $this->encode_simple($schema, $entry);
			return $ret;
		}
	}
	
	function &encode_records(&$for_schema, &$data) {
		$rec_type = $for_schema->get_record_type();
		if(isset($rec_type))
			$schema =& $this->schema[$rec_type];
		else
			$schema =& $for_schema;

		$ret = array();
		$stdnms =& $schema->standard;
		$stdlen = is_countable($stdnms) ? count($stdnms) : 0;
		$extnms =& $schema->extra;
		$extlen = is_countable($extnms) ? count($extnms) : 0;
		$cr = is_countable($data) ? count($data) : 0;
		for($i = 0; $i < $cr; $i++) {
			$row =& $data[$i];
			unset($arr);
			$arr = array();
			for($j = 0; $j < $stdlen; $j++) {
				$arr[] =& $this->encode_data($schema->fields[$stdnms[$j]], $row[$stdnms[$j]]);
			}
			if($extlen) {
				$opt = array();
				for($j = 0; $j < $extlen; $j++) {
					if(isset($row[$extnms[$j]]))
						$opt[$extnms[$j]] =& $this->encode_data($schema->fields[$extnms[$j]], $row[$extnms[$j]]);
				}
				if(count($opt))
					$arr[] = $opt;
			}
			$ret[] =& $arr;
		}
		return $ret;
	}
	
	function &encode_data(&$schema, &$data) {
		$type = $schema->get_type();
		if($type == 'sequence')
			$ret =& $this->encode_sequence($schema, $data);
		else if($type == 'map')
			$ret =& $this->encode_map($schema, $data);
		else if($type == 'record') {
			$dat = array();
			$dat[0] =& $data;
			$enc =& $this->encode_records($schema, $dat);
			$ret =& $enc[0];
		}
		else
			$ret = $this->encode_simple($schema, $data);
		return $ret;
	}
	
	function encode_simple(&$schema, &$data) {
		// a bit more complex in the javascript version
		return $data;
	}
		
	function encode_to_array($minimal=false) {
		$head = array('JSRD' => JSRD::$version);
		$head['schema'] = $this->encode_schemata($minimal);
		if(! $head['schema'])
			return false;
		if(is_countable($this->custom) && count($this->custom))
			$head['custom'] = $this->custom;
		$body =& $this->encode_data($this->schema[$this->root_schema], $this->data);
		return array($head, $body);
	}
	
	function encode($minimal=false) {
		static $json;
		if(! isset($json))
			$json = new JSON(JSON_LOOSE_TYPE);
		$arr = $this->encode_to_array($minimal);
		if($arr === false)
			$dat = $arr;
		else
			$dat = $json->encode($arr);
		return $dat;
	}
	
	function has_schema($name=null) {
		if(! isset($name) && isset($this->root_schema))
			$name = $this->root_schema;
		return isset($this->schema[$name]);
	}
	
	function add_schema($spec) {
		$schema = new JSRD_Schema();
		if(! $schema->load($spec))
			return $this->error_on_load($schema->get_error());
		$this->schema[$schema->name] =& $schema;
		if(! $this->root_schema)
			$this->root_schema = $schema->name;
		return $schema->name;
	}
	
	function remove_schema($name) {
		if(isset($this->schema[$name])) {
			unset($this->schema[$name]);
			if($this->root_schema == $name)
				$this->root_schema = null;
			return true;
		}
		return false;
	}
	
	function add_rows(&$rows) {
		$sch =& $this->schema[$this->root_schema];
		//$flds =& $sch->fields;
		foreach($rows as $arr) {
			/*$row = array();
			$ext = array();
			foreach($flds['standard'] as $f) {
				$row[] = array_get_default($arr, $f);
			}
			foreach($flds['extra'] as $f) {
				if(isset($arr[$f]))
					$ext[$f] = $arr[$f];
			}
			if(count($ext))
				$row[] = $ext;
			$this->data[] = $row;*/
			
			// verify row?
			$this->data[] = $arr;
		}
	}
	
	function get_custom() {
		return $this->custom;
	}
	function set_custom($custom) {
		$this->custom = $custom;
	}
	
	function set_root_schema($root_name) {
		if(isset($this->schema[$root_name])) {
			$this->root_schema = $root_name;
			return true;
		}
		return false;
	}
	
	function get_simple_type(&$value) {
		if(is_array($value)) {
			reset($value);
			if(count($value) && is_integer(key($value)))
				return 'sequence';
			return 'record';
		}
		else if(is_bool($value))
			return 'boolean';
		else if(is_string($value))
			return 'string';
		else if(is_integer($value))
			return 'integer';
		else if(is_float($value))
			return 'float';
		else if(is_object($value))
			return false; // could encode as a record? or check for a custom method?
		else
			return false;
	}
	
	function load_simple_value($schema_name, $val) {
		$spec = array('name' => $schema_name);
		$t = $spec['type'] = $this->get_simple_type($val);
		if(! $spec['type'])
			return false;
		if($t == 'sequence') {
			$t = $spec['element_type'] = $this->get_simple_type($val[0]);
			if($t == 'record')
				$recval = $val[0];
		}
		else if($spec['type'] == 'record') {
			$recval = $val;
		}
		$this->init();
		if($t == 'record') {
			$spec['fields'] = array();
			foreach(array_keys($recval) as $k) {
				$t = $this->get_simple_type($recval[$k]);
				if(! $t)
					return false;
				$fld = array('name' => $k, 'type' => $t);
				$spec['fields'][] = $fld;
			}
		}
		if(! $this->load_schemata($spec, true)) {
			return false;
		}
		$this->data = $val;
		return true;
	}
	
	function &get_data() {
		return $this->data;
	}
	
}
// end class JSRD


class JSRD_Schema {
	var $type;
	var $label;
	var $element_type;
	var $record_type;
	var $maxlen;
	var $format;
	var $fields;
	var $standard;
	var $extra;
	var $custom;
	var $error_text;
	var $record_types;
	
	static $known_field_types = array(
		'any', 'string', 'integer', 'float', 'date', 'time', 'datetime',
		'boolean', 'record', 'sequence', 'map',
	);
	
	function load($spec, $is_root = false) {
		if(is_string($spec)) {
			$spec = array('name' => $spec);
		}
		if(! is_array($spec))
			return $this->error('Expected string or array for schema');
		if(! isset($spec['name']))
			return $this->error('Missing schema name');
		if(! is_string($spec['name']))
			return $this->error('Expected string (schema name)');
		$this->name = $spec['name'];
		
		if(isset($spec['record_type']) || isset($spec['fields']))
			$default_elt_type = 'record';
		else
			$default_elt_type = 'any';
		if(isset($spec['element_type']))
			$default_type = 'sequence';
		else
			$default_type = $default_elt_type;
	
		if(! isset($spec['type']))
			$spec['type'] = $default_type;
		else if(! is_string($spec['type']))
			return $this->error('Expected string (schema type)');
		if(! in_array($spec['type'], JSRD_Schema::$known_field_types))
			return $this->error('Unknown schema type: '.$spec['type']);
		$this->type = $spec['type'];
		
		$this->label = JSRD::get_string(array_get_default($spec, 'label'));
		if(isset($spec['custom'])) {
			if(! is_array($spec['custom']))
				return $this->error('Expected array (custom)');
			$this->custom = $spec['custom'];
		}
		
		if($this->is_collection()) {
			if(! isset($spec['element_type']))
				$spec['element_type'] = $default_elt_type;
			if(! in_array($spec['element_type'], JSRD_Schema::$known_field_types))
				return $this->error('Unknown schema element type: '.$spec['element_type']);
			$this->element_type = $spec['element_type'];
		}
		if($this->type == 'record' || ($this->is_collection() && $this->element_type == 'record')) {
			if(isset($spec['fields'])) {
				if(isset($spec['record_type']))
					return $this->error('Schema specifies both fields and record_type');
				if(! $this->load_fields($spec)) {
					return false;
				}
			}
			else {
				if(! isset($spec['record_type']) || ! is_string($spec['record_type']))
					return $this->error('Expected string (schema record type)');
				$this->record_type = $spec['record_type'];
			}
		}

		//if(! $this->is_collection()) {
		// these could be applied to the collection values, although they aren't currently
			$maxlen = JSRD::get_integer(array_get_default($spec, 'maxlen'));
			if(isset($maxlen)) {
				if($maxlen <= 0)
					return $this->error('Expected positive integer (maxlen)');
				$this->maxlen = $maxlen;
			}
			$this->nullable = JSRD::get_boolean(array_get_default($spec, 'nullable'), true); /* TODO: raise an error on invalid boolean value */
			$this->format = JSRD::get_string(array_get_default($spec, 'format'));
		//}
		
		if(isset($spec['record_types'])) {
			if(! $is_root)
				return $this->error('Field record_types defined for non-root schema');
			if(! is_array($spec['record_types']))
				return $this->error('Expected array (record_types)');
			$this->record_types = array();
			foreach($spec['record_types'] as $sub_spec) {
				$sub = new JSRD_Schema();
				if(! $sub->load($sub_spec, true))
					return $this->error($sub->get_error());
				$this->record_types[] = $sub;
			}
		}
		
		return true;
	}
	
	function load_fields($spec) {
		$this->fields = array();
		$this->standard = array();
		$this->extra = array();
		$fields = array_get_default($spec, 'fields');
		if(! is_array($fields))
			return $this->error_on_load('Expected array (fields)');
		$extra = array_get_default($spec, 'extra_fields');
		if(! is_array($extra))
			$extra = array();
		$ok = false;
		foreach($fields as $f_def) {
			$schema = new JSRD_Schema();
			if(! $schema->load($f_def))
				return $this->error($schema->get_error());
			if(isset($this->fields[$schema->name]))
				return $this->error('Duplicate field name: '.$schema->name);
			$this->fields[$schema->name] = $schema;
			$this->standard[] = $schema->name;
			$ok = true;
		}
		foreach($extra as $f_def) {
			$schema = new JSRD_Schema();
			if(! $schema->load($f_def))
				return $this->error($schema->get_error());
			if(isset($this->fields[$schema->name]))
				return $this->error('Duplicate field name: '.$schema->name);
			$this->fields[$schema->name] = $schema;
			$this->extra[] = $schema->name;
		}
		if(! $ok)
			return this.error('No fields defined');
		return true;
	}
	
	function get_type() {
		return $this->type;
	}
	function get_element_type() {
		return $this->element_type;
	}
	function get_record_type() {
		return $this->record_type;
	}
	function is_collection() {
		return ($this->type == 'sequence' || $this->type == 'map');
	}
	function get_custom() {
		return $this->custom;
	}
	function set_custom($custom) {
		$this->custom = $custom;
	}
	function error($msg) {
		$this->error_text = $msg;
		return false;
	}
	function get_error() {
		return $this->error_text;
	}
	
	function encode($minimal=false) {
		static $field_keys = array(
			'name', 'label', 'type', 'element_type', 'record_type',
			'maxlen', 'nullable', 'format', 'custom', 'fields', 'extra',
		);
		$ret = array();
		$simple = true;
		foreach($field_keys as $key) {
			if(isset($this->$key)) {
				if($key == 'custom') {
					if(! empty($this->$key)) {
						$ret[$key] = $this->$key;
						$simple = false;
					}
				}
				else if($key == 'type' && $this->$key == 'any')
					; // pass
				else if($key == 'nullable' && $this->$key)
					; // pass
				else if($key == 'fields' || $key == 'extra') {
					if($this->type == 'record' || $this->element_type == 'record') {
						$target = ($key == 'fields') ? $key : 'extra_fields';
						unset($flds);
						$flds =& $this->encode_fields(($key == 'extra'), $minimal);
						if(is_countable($flds) && count($flds))
							$ret[$target] = $flds;
						$simple = false;
					}
				}
				else {
					if($key != 'name')
						$simple = false;
					$ret[$key] = $this->$key;
				}
			}
		}
		if($simple)
			$ret = $ret['name'];
		return $ret;
	}

	function &encode_fields($extra=false, $minimal=false) {
		$ret = array();
		$target = ($extra ? 'extra' : 'standard');
		$nms =& $this->$target;
		foreach($nms as $k2) {
			$ret[] = $this->fields[$k2]->encode($minimal);
		}
		return $ret;
	}
	
	function add_field($spec, $extra=false) {
		if(! isset($this->fields)) {
			$this->fields = array();
			$this->standard = array();
			$this->extra = array();
		}
		
		$fld = new JSRD_Schema();
		if(! $fld->load($spec))
			return $this->error($spec->get_error());
		if(isset($this->fields[$fld->name]))
			$this->remove_field($fld->name);
		
		$set_f = $extra ? 'extra' : 'standard';
		$this->fields[$fld->name] =& $fld;
		$this->{$set_f}[] = $fld->name;
		return true;
	}

	function remove_field($name) {
		$ok = false;
		if(isset($this->fields[$name])) {
			unset($this->fields[$name]);
			foreach($this->standard as $i=>$f)
				if($f == $name) {
					unset($this->standard[$i]);
					break;
				}
			foreach($this->extra as $i=>$f)
				if($f == $name) {
					unset($this->extra[$i]);
					break;
				}
			$ok = true;
		}
		return $ok;
	}
}
// end class JSRD_Schema

?>
