<?php
require_once 'include/ListView/ListViewManager.php';

class DuplicateManager {

    /**
     * @var RowUpdate
     */
    public $new_record;

    /**
     * @var array
     */
    public $input;

    /**
     * @var bool
     */
    public $has_related = false;

    /**
     * @var bool
     */
    public $in_popup = false;

    const LAYOUT_NAME = 'Duplicates';
    const MAX_DUPLICATE_RECORDS = 5;

    /**
     * @param RowUpdate $new_record
     * @param array $input - $_REQUEST
     */
    public function __construct(RowUpdate $new_record, $input = array()) {
        $this->new_record = $new_record;
        $this->input = $input;

        if (! empty($input['in_popup']))
            $this->in_popup = true;

        if (! empty($input['newaccount']) || ! empty($input['newopp'])) {
            $this->has_related = true;

            require_once('modules/Contacts/widgets/RelatedRecordsWidget.php');
            require_once('include/layout/forms/EditableForm.php');
            require_once('include/layout/forms/FormGenerator.php');
        }
    }

    /**
     * Try to find duplicated records by where clauses
     * If found some then redirect to duplicates list
     *
     * @param array $where_clauses
     * @param bool $redirect - redirect to duplicates list
     * @return ListResult|null
     */
    public function check($where_clauses, $redirect = true) {
        $duplicates = $this->getDuplicates($this->new_record->model, $where_clauses);

        if ($duplicates && $redirect) {
            $key = $this->storeData($duplicates);
            $this->goToList($key);
        }

        return $duplicates;
    }

    /**
     * Show list of potential duplicates
     *
     * @param ListResult $duplicates
     */
    public function showList(ListResult $duplicates) {
        if (! $this->has_related) {
            $this->showStandardList($duplicates);
        } else {
            $this->showListWithRelates($duplicates);
        }
    }

    /**
     * Get new record and duplicates from storage
     *
     * @static
     * @param string $key
     * @return null|array
     */
    public static function get_stored_data($key) {
        if (isset($_SESSION['duplicate_' . $key])) {
            $data = $_SESSION['duplicate_' . $key];
            unset($_SESSION['duplicate_' . $key]);
            return $data;
        } else {
            return null;
        }
    }

    /**
     * @param ListResult $duplicates
     */
    private function showStandardList(ListResult $duplicates) {
        $list_manager = $this->getListManager($this->new_record->model, $duplicates);

        echo $this->openListForm();
        $this->showListTitle();
        $list_manager->render();
        echo $this->closeListForm();
    }

    /**
     * @param ListResult $duplicates
     */
    private function showListWithRelates(ListResult $duplicates) {
        $form = new EditableForm('editview', 'Save');

        echo $this->openListForm();
        $this->showListTitle();
        echo $this->getFormButtons();

        $main_manager = $this->getListManager($this->new_record->model, $duplicates);
        $main_manager->render();

        echo $this->getCreateCheck($form, strtolower($this->new_record->model_name), $duplicates);

        $rel_duplicates = $this->getAllRelDuplicates();

        if (sizeof($rel_duplicates) > 0) {
            foreach ($rel_duplicates as $rel_name => $rel_dups) {
                if ($rel_dups) {
                    $manager = $this->getListManager($rel_dups['model'], $rel_dups['duplicates']);
                    $manager->render();

                    echo $this->getCreateCheck($form, $rel_name, $rel_dups['duplicates']);
                }
            }
        }

        $form->exportIncludes();
        echo $this->getFormButtons();
        echo $this->closeListForm();
    }

    /**
     *
     * @param ModelDef $model
     * @param ListResult $duplicates
     * @return ListViewManager
     */
    private function getListManager(ModelDef $model, ListResult $duplicates) {
        $lv = new ListViewManager('listview', true);
        $lv->disable_request_filter = true;
        $lv->show_create_button = false;
        $lv->show_mass_update = false;
        $lv->show_mass_form = false;
        $lv->show_tabs = false;
        $lv->show_checks = false;
        $lv->show_filter = false;
        $lv->show_help = false;
        $lv->show_title = false;
        $lv->hide_navigation_controls = true;
        $lv->no_sort = true;
        unset($_REQUEST['list_limit']);
        $lv->default_list_limit = -1;
        $lv->loadRequest();

        if(! $lv->initModuleView($this->getModule($model), self::LAYOUT_NAME))
            ACLController::displayNoAccess();

        if (! $this->has_related)
            $lv->getFormatter()->addButtons($this->getListButtons());
        $q = $this->getLoadDuplicatesQuery($model, $duplicates);
        $lv->getFormatter()->setQuery($q);

        return $lv;
    }

    /**
     * @return string
     */
    private function openListForm() {
        $form = <<<EOQ
            <form name="Save" id="Save" method="POST">
EOQ;

        $copy_params = array(
        	'layout', 'mgr_type', 'format', 'go_to_view', 'for_ref',
        	'target_name', 'target_id', 'target_layout', 'primary_name',
        	'return_module', 'return_action', 'return_record', 'return_layout', 'return_panel',
        	'rel_lead_id',
        );
        $params = array(
        	'module' => $this->getModule(),
        	'action' => 'EditView',
        	'dup_checked' => 1,
        );
        if ($this->in_popup)
        	$params['in_popup'] = 1;
        if(empty($params['return_module'])) {
        	$params['return_module'] = $this->getModule();
        	$params['return_action'] = 'index';
        }

        foreach($copy_params as $f) {
        	if(isset($this->input[$f]))
        		$params[$f] = $this->input[$f];
        }
        if(! empty($params['primary_name'])) {
        	$params[$params['primary_name']] = array_get_default($this->input, $params['primary_name']);
        }

        $input = $this->new_record->getInput();
        if (empty($input))
            $input = $this->new_record->updates;

        foreach (array_merge($params, $input) as $field_name => $value) {
            $form .= '<input type="hidden" name="' .to_html($field_name). '" value="' .to_html($value). '"/>';
        }

        if ($this->has_related)
            $form .= $this->addRelateHiddenInputs();

        return $form;
    }

    /**
     * @return string
     */
    private function closeListForm() {
        return '</form>';
    }

    /**
     * Add additional relates hidden inputs
     *
     * @return string
     */
    private function addRelateHiddenInputs() {
        $inputs = '<input type="hidden" value="' .$this->new_record->model_name. '" name="base_model" />';

        $relates = array(
            'newaccount' => array('model' => 'Account', 'rel_name' => 'account'),
            'newopp' => array('model' => 'Opportunity', 'rel_name' => 'opportunity'),
            'newappointment' => array('model' => 'Appointment', 'rel_name' => 'appointment'),
        );


        foreach ($relates as $check => $details) {
            if (! empty($this->input[$check]))  {
                $inputs .= '<input type="hidden" value="1" name="' .$check. '" />';
                $rel_data = RelatedRecordsWidget::fill_data($details['rel_name'], $this->input);

                foreach($rel_data as $name => $value)  {
                    $inputs .= '<input type="hidden" value="' .$value. '" name="' .$details['rel_name'] .'_'. $name. '" />';
                }

                if (isset($this->input[$details['rel_name']]))
                    $inputs .= '<input type="hidden" value="' .$this->input[$details['rel_name']]. '" name="' .$details['rel_name']. '" />';

                $rel_notes = RelatedRecordsWidget::fill_notes($details['rel_name'], $this->input);

                foreach($rel_notes as $name => $value)  {
                    $inputs .= '<input type="hidden" value="' .$value. '" name="' .$details['rel_name'] .'_notes_'. $name. '" />';
                }

            }
        }

        return $inputs;
    }

    /**
     * Show duplicates list title
     */
    private function showListTitle() {
        global $mod_strings, $app_strings;
        echo get_module_title($mod_strings['LBL_MODULE_NAME'], $mod_strings['LBL_MODULE_NAME'] .translate('LBL_SEPARATOR', 'app'). $mod_strings['LBL_SAVE_MODULE'], true);

        $message = $mod_strings['MSG_SHOW_DUPLICATES'];
        if ($this->has_related)
            $message = $app_strings['MSG_DUPLICATE'];
        echo '<table width="100%"><tr><td>' .$message. '</td></tr><tr><td height="20"></td></tr></table>';
    }

    /**
     * For form with Related Records
     *
     * @return string
     */
    private function getFormButtons() {
        global $app_strings;
        $in_popup = (int)$this->in_popup;

        $result = <<<EOQ
            <table width="100%" cellpadding="0" cellspacing="0" border="0">
            <tr>
                <td align="left" style="padding-bottom: 2px;">
                <button class="input-button" onclick="try {  var ret = SUGAR.ui.sendForm(document.forms.Save, {'module': 'Contacts', 'action': 'SaveWithRelates'}, {'closePopup': {$in_popup},'resultDiv':'content-main'}, false, {$in_popup});  return ret; } catch(e) { console.error(e); return false; }" tabindex="" style="" title="{$app_strings['LBL_SAVE_BUTTON_TITLE']}" accessKey="{$app_strings['LBL_SAVE_BUTTON_KEY']}">
                    <div class="input-icon left icon-accept"></div>
                    <span class="input-label">{$app_strings['LBL_SAVE_BUTTON_LABEL']}</span>
                </button>
                <button onclick="return SUGAR.ui.cancelEdit(document.forms.Save);" tabindex="" style="" class="input-button" type="button" title="{$app_strings['LBL_CANCEL_BUTTON_TITLE']}" accessKey="{$app_strings['LBL_CANCEL_BUTTON_KEY']}">
                    <div class="input-icon left icon-cancel"></div>
                    <span class="input-label">{$app_strings['LBL_CANCEL_BUTTON_LABEL']}</span>
                </button>
                </td>
            </tr>
            </table>
EOQ;

        return $result;
    }

    /**
     * @return array
     */
    private function getListButtons() {
    	$mgr_type = array_get_default($this->input, 'mgr_type');
    	$go_to_view = (int)array_get_default($this->input, 'go_to_view');
    	$reload_after = (int)($this->in_popup && $mgr_type != 'quick_create');
        $buttons = array(
            'save' => array(
                'label' => translate('LBL_SAVE_BUTTON_LABEL'),
                'icon' => 'icon-accept',
                'type' => 'button',
                'perform' => 'return SUGAR.ui.sendForm(document.forms.Save, {"record_perform":"save", "close_popup":'.(int)$this->in_popup.', "reload_after":'.$reload_after.'}, null, false, '.(int)$this->in_popup.');',
                'params' => array('toplevel' => 1)
            ),
            'edit' => array(
                'label' => translate('LBL_EDIT_BUTTON_LABEL'),
                'icon' => 'icon-edit',
                'type' => 'button',
                'perform' => 'return SUGAR.ui.sendForm(document.forms.Save, {"record_perform":"edit"}, null, false, '.(int)$this->in_popup.');',
                'params' => array('toplevel' => 1)
            ),
            'cancel' => array(
                'label' => translate('LBL_CANCEL_BUTTON_LABEL'),
                'icon' => 'icon-cancel',
                'type' => 'button',
                'perform' => 'return SUGAR.ui.cancelEdit("Save");',
                'params' => array('toplevel' => 1)
            )
        );
        if($mgr_type == 'quick_create') {
			$params = array();
			foreach(array('target_name', 'target_id', 'target_layout', 'primary_name') as $f)
				$params[$f] = array_get_default($this->input, $f);
			if(! empty($params['primary_name']))
				$params['primary_value'] = array_get_default($this->input, $params['primary_name']);
			$json = getJSONobj();
			$params = $json->encode($params);
        	$buttons['save']['perform'] = 'return SUGAR.ui.saveQuickCreate(document.forms.Save, '.$go_to_view.', '.$params.');';
        }
		return $buttons;
    }

    /**
     * Get Create checkbox
     *
     * @param EditableForm $form
     * @param string $name
     * @param ListResult $duplicates
     * @return string
     */
    private function getCreateCheck(EditableForm &$form, $name, ListResult $duplicates) {
        global $mod_strings, $app_strings;

        $spec = array('name' => 'create_' . $name);
        $div_id = 'create_' .$name. 'div';
        $spec['onchange'] = "toggleDisplay('".$div_id."');";
        $check = $form->renderCheck($spec, 1);

        $options = array();
        $duplicate_rows = $duplicates->getRows();
        $selected = '';
        $count = 0;

        foreach ($duplicate_rows as $id => $row) {
            $value = $row['name'];
            if (isset($row['email1']))
                $value .= ' - ' . $row['email1'];
            $options[$id] = $value;
            if ($count == 0)
                $selected = $id;
            $count ++;
        }

        $spec = array('name' => 'selected_' .$name. '_id', 'options' => $options, 'options_add_blank' => false);
        $select = "<div id='create_{$name}div' style='display: none;'><h5 class='dataLabel'>{$app_strings['LBL_OR_SELECT']}</h5>&nbsp;&nbsp;{$form->renderSelect($spec, $selected)}</div>";

        $result = '<span style="padding: 10px;"><h5 class="dataLabel">' .$check. '&nbsp;&nbsp;' .$mod_strings['LNK_NEW_'. strtoupper($name)] .'</h5>'. $select .'</span>';

        return $result;
    }

    /**
     * Store new record and duplicates
     *
     * @param ListResult $duplicates
     * @return String
     */
    private function storeData(ListResult $duplicates) {
        $key = create_guid();
        $_SESSION['duplicate_'.$key] = array('new_record' => $this->new_record, 'duplicates' => $duplicates, 'input' => $this->input);

        return $key;
    }

    /**
     * Redirect to duplicates list (ShowDuplicates)
     *
     * @param string $store_key
     */
    private function goToList($store_key) {
        $redirect_params = array('module' => $this->getModule(), 'action' => 'ShowDuplicates', 'key' => $store_key);
        $path = 'async.php?' . encode_query_string($redirect_params);
		if(isset($this->input['layout']) && $this->input['layout'] == 'QuickCreate') {
            header('Location: '.$path);
            die();
        }
        else
            indexRedirect($redirect_params);

    }

    /**
     * Get record module
     *
     * @param ModelDef|null $model
     * @return string
     */
    private function getModule($model = null) {
        if ($model) {
            $module = $model->getModuleDir();
        } else {
            $module = $this->new_record->model->getModuleDir();
        }
        if (is_array($module))
            $module = $module[0];

        return $module;
    }

    /**
     * @param ModelDef $model
     * @param array $where_clauses
     * @return ListResult|null
     */
    private function getDuplicates($model, $where_clauses) {
    	if(! $where_clauses) return null;

        $lq = new ListQuery($model, null, array('acl_filters' => 'list'));
        $lq->addColumnFields();
        if (! isset($lq->fields['name']))
            $lq->addField('name');

        if (isset($where_clauses['string']) && $where_clauses['string'] == true) {
            $lq->where = $where_clauses['query'];
        } else {
            $lq->addFilterClauses($where_clauses);
        }

        $result = $lq->runQuery(0, self::MAX_DUPLICATE_RECORDS);

        if (! $result->failed && $result->getResultCount()) {
            return $result;
        } else {
            return null;
        }
    }

    /**
     * @return array
     */
    private function getAllRelDuplicates() {
        $rel_duplicates = array();

        switch ($this->new_record->model_name) {
            case "Contact":
                if (! empty($this->input['newaccount']))
                    $rel_duplicates['account'] = $this->getRelDuplicates('Account');
                if (! empty($this->input['newopp']))
                    $rel_duplicates['opportunity'] = $this->getRelDuplicates('Opportunity');
                break;
            case "Account":
                if (! empty($this->input['newopp']))
                    $rel_duplicates['opportunity'] = $this->getRelDuplicates('Opportunity');
                break;
        }

        return $rel_duplicates;
    }

    /**
     * @param $model_name
     * @return array
     */
    private function getRelDuplicates($model_name) {
        require_bean($model_name);
        $update = RowUpdate::for_model($model_name);
        $update->loadInput(RelatedRecordsWidget::fill_data(strtolower($model_name), $this->input), false);

        $duplicates = $this->getDuplicates($update->model, $model_name::get_duplicates_where($update));
        $result = array();

        if ($duplicates)
            $result = array('model' => $update->getModel(), 'duplicates' => $duplicates);

        return $result;
    }

    /**
     * @param ModelDef $model
     * @param ListResult $duplicates
     * @return ListQuery
     */
    private function getLoadDuplicatesQuery(ModelDef $model, ListResult $duplicates) {
        $rows = $duplicates->getRows();
        $duplicate_ids = array();

        foreach($rows as $id => $details) {
            array_push($duplicate_ids, $id);
        }

        $lq = new ListQuery($model);

        $clauses = array('primary_filter' =>
        array(
            'field' => $lq->getPrimaryKey(),
            'value' => $duplicate_ids,
            'operator' => 'in'
        )
        );
        $lq->addFilterClauses($clauses);

        return $lq;
    }
}
