<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo, 
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */
require_once('include/utils/supported_functions.php');

class FormulaParseError extends IAHInternalError {}

class FormulaParser {

    const VARIABLES_REGEXP = '/(\$\w+)/';

    /**
     * @var string
     */
    private $expression;
    private $tree;
    private $remain;
    private $parse_error;
    private $variables;
    
    static $allowed_constants = array('null', 'true', 'false');
    static $seen = array();


	public function getExpression() {
		return $this->expression;
	}

    public function getTree() {
    	return $this->tree;
    }
    
    public function getVariables() {
    	return $this->variables;
    }

    public function getRemain() {
    	return $this->remain;
    }

    public function getParseError() {
    	return $this->parse_error;
    }
    
    
    /**
     * @param $expression - formula expression string
     */
    public function __construct($expression) {
		$this->expression = $expression;
    	if(isset(self::$seen[$expression])) {
    		$old =& self::$seen[$expression];
    		$this->tree = $old->tree;
    		$this->parse_error = $old->parse_error;
			$this->remain = $old->remain;
			$this->variables = $old->variables;
		}
		else {
			try {
				$this->tree = self::parse_expression($expression);
			} catch(FormulaParseError $e) {
				$this->parse_error = $e->getMessage();
			}
			if($this->tree) {
				$this->remain = $this->tree['remain'];
				unset($this->tree['remain']);
			}
			$this->extractVariables($this->tree);
			self::$seen[$expression] =& $this;
		}
    }
    
    public static function parse_expression($expr, $depth=0) {
    	$val = null;
    	if(preg_match('~^\s*(\$?[a-zA-Z][a-zA-Z0-9_]*)(.*)$~u', $expr, $m)) {
    		$token = $m[1];
    		if(substr($token, 0, 1) == '$') {
    			$val = array(
					'type' => 'value',
					'subtype' => 'variable',
					'value' => $token,
					'remain' => $m[2],
				);
			}
    		else if(in_array($token, self::$allowed_constants)) {
    			$val = array(
					'type' => 'value',
					'subtype' => 'special',
					'value' => $token,
					'remain' => $m[2],
				);
			} else {
				// expect function call
				$arg = self::parse_expression($m[2], $depth+1);
				if(! $arg || $arg['type'] != 'operator' || $arg['value'] != '(')
					throw new FormulaParseError("Expected: '('");
				$val = array(
					'type' => 'function',
					'value' => $token,
					'args' => array(),
				);
				$pendArg = null;
				$afterComma = false;
				do {
					$arg = self::parse_expression($arg['remain'], $depth+1);
					if(! $arg)
						throw new FormulaParseError("Expected: ')'");
					if($arg['type'] != 'operator') {
						if(isset($pendArg))
							throw new FormulaParseError("Expected: ')'");
						$pendArg = $arg;
						unset($pendArg['remain']);
						continue;
					}
					if($arg['type'] == 'operator') {
						if($arg['value'] != ')' && $arg['value'] != ',')
							throw new FormulaParseError("Unexpected token: '{$arg['value']}'");
						if(! isset($pendArg) && $afterComma) {
							$pendArg = array(
								'type' => 'value',
								'subtype' => 'special',
								'value' => 'null',
							);
						}
						if(isset($pendArg)) {
							$val['args'][] = $pendArg;
							$pendArg = null;
						}
						if($arg['value'] == ')') {
							$val['remain'] = $arg['remain'];
							break;
						}
						else
							$afterComma = true;
					}
				} while(1);
			}
    	}
    	else if(preg_match('~^\s*([+\-]?\.[0-9]+|[+\-]?[0-9]+(?:\.[0-9]+)?)(.*)$~u', $expr, $m)) {
    		$val = array(
    			'type' => 'value',
    			'subtype' => 'number',
    			'value' => $m[1],
    			'remain' => $m[2],
    		);
    	}
    	else if(preg_match('~^\s*((?:\".*?(?<!\\\\)\")|(?:\'.*?(?<!\\\\)\'))(.*)$~u', $expr, $m)) {
    		$val = array(
    			'type' => 'value',
    			'subtype' => 'string',
    			'value' => $m[1],
    			'remain' => $m[2],
    		);
    	}
    	else {
    		$expr = ltrim($expr);
    		$c = substr($expr, 0, 1);
    		if($c == ',' || $c == ')' || $c == '(') {
    			$val = array(
    				'type' => 'operator',
    				'value' => $c,
    				'remain' => substr($expr, 1),
    			);
    		}
    		else
    			throw new FormulaParseError("Unexpected token: '$c'");
    	}
    	return $val;
    }

    /**
     * Calculate the value by formula expression
     *
     * TODO: don't use eval, move to Expression Engine
     * @param array $row - model $row data
     * @return string
     */
    public function calculate(array $row) {
        $expr = $this->replaceVariables($row, $this->tree);

        $result = '';
        eval("\$result = $expr;");

        return $result;
    }

    /**
     * Get the list of the fields used in formula expression
     *
     * @return array
     */
    public function getExpressionFields() {
        $vars = $this->variables;
        $fields = array();

        if (is_countable($vars) && sizeof($vars) > 0) {
            for ($i = 0; $i < sizeof($vars); $i++) {
                array_push($fields, substr($vars[$i], 1));
            }
        }

        return $fields;
    }

    /**
     * Extract variables from formula expression
     *
     * @return array
     */
    private function extractVariables($tree, $depth=0) {
    	if(! $depth) $this->variables = array();
        if(is_array($tree)) {
        	if($tree['type'] == 'function') {
        		foreach($tree['args'] as $a)
        			$this->extractVariables($a, 1);
        	}
        	else if($tree['type'] == 'value' && $tree['subtype'] == 'variable')
        		$this->variables[] = $tree['value'];
        }
        if(! $depth)
        	$this->variables = array_unique($this->variables);
    }

    /**
     * Replace variables with values in formula expression
     *
     * @param array $row
     * @return string
     */
    private function replaceVariables($row, $tree) {
    	if(! $tree) return 'null';
    	if($tree['type'] == 'function') {
    		$allowed = self::get_allowed_functions();
    		if(! isset($allowed[$tree['value']])) {
    			if(is_countable($tree['args']) && count($tree['args']))
    				return $this->replaceVariables($row, $tree['args'][0]);
    			return 'null';
    		}
    		$val = 'calcf_' . $tree['value'] . '(';
    		$first = 1;
    		foreach($tree['args'] as $a) {
    			if(! $first) $val .= ', ';
    			$first = 0;
    			$val .= $this->replaceVariables($row, $a);
    		}
    		return $val . ')';
    	}
    	else if($tree['type'] == 'value') {
    		if($tree['subtype'] == 'variable') {
    			$field = substr($tree['value'], 1);
                if (isset($row[$field])) {
                    $val = $row[$field];
                    if(! is_numeric($val))
                    	$val = "'" . addcslashes($val, "\\'\n\r") . "'";
                } else
                	$val = 'null';
                return $val;
    		}
    		else if($tree['subtype'] == 'special') {
    			if($tree['value'] == 'true')
    				return 1;
    			else if($tree['value'] == 'false')
    				return 0;
    		}
    		return $tree['value'];
    	}
    }
    
    public function getSql(array $fieldmap) {
    	if($this->tree) {
			return self::get_tree_sql($fieldmap, $this->tree);
    	}
    }
    

	private static function cast_null(&$args, $value) {
		foreach ($args as &$arg) {
			$arg = "IFNULL($arg, $value)";
		}
	}

    private static function get_tree_sql(array $fieldmap, array $tree) {
    	global $db;
    	
    	if($tree['type'] == 'function') {
    		$args = array();
    		foreach($tree['args'] as $a)
				$args[] = self::get_tree_sql($fieldmap, $a);
			$join_args = null;
    		switch($tree['value']) {
    			case 'equal':
    				if(count($args) >= 2)
    					return '(' . $args[0] . ' <=> ' . $args[1] . ')';
    				return 'NULL';
				case 'add':
					self::cast_null($args, 0);
    				return '(' . implode(' + ', $args) . ')';
    			case 'subtract':
					self::cast_null($args, 0);
    				return '(' . implode(' - ', $args) . ')';
    			case 'multiply':
					self::cast_null($args, 0);
    				return '(' . implode(' * ', $args) . ')';
    			case 'divide':
					self::cast_null($args, 0);
    				if(count($args) >= 2)
    					return '(' . $args[0] . ' / ' . $args[1] . ')';
    				return 'NULL';
    			case 'pow':
					self::cast_null($args, 0);
    				if(count($args) >= 2)
    					return "POW({$args[0]}, {$args[1]})";
    				return 'NULL';
    			case 'max':
					self::cast_null($args, 0);
					if (count($args) < 1) {
						return '0';
					}
					if (count($args) < 2) {
						return "(" . $args[0] . ")";
					}
    				$fn = 'GREATEST';
					$join_args = ', ';
    				break;
    			case 'min':
					self::cast_null($args, 0);
					if (count($args) < 1) {
						return '0';
					}
					if (count($args) < 2) {
						return "(" . $args[0] . ")";
					}
					$fn = 'LEAST';
					$join_args = ', ';
    				break;
    			case 'floor':
    			case 'ceil':
    			case 'abs':
    			case 'exp':
					self::cast_null($args, 0);
    				$fn = strtoupper($tree['value']);
    				break;
    			case 'log':
					self::cast_null($args, 0);
    				if(count($args) < 1)
    					return '0';
    				if(count($args) < 2)
    					return "LOG($args[0])";
    				return "LOG($args[0], $args[1])";
    			case 'negate':
					self::cast_null($args, 0);
    				return "-({$args[0]})";
    			case 'ifElse':
    				if(count($args) < 2)
    					return 'NULL';
    				$arg2 = (count($args) < 3) ? 'NULL' : $args[2];
    				return "IF($args[0], $args[1], $arg2)";
    			case 'ifNull':
    				if(count($args) < 1)
    					return 'NULL';
    				if(count($args) < 2)
    					return $args[0];
    				return "IFNULL($args[0], $args[1])";
    			case 'strlen':
					self::cast_null($args, "''");
    				$fn = 'LENGTH';
    				break;
    			case 'concat':
					self::cast_null($args, "''");
    				if(count($args) < 1)
    					return '""';
    				return 'CONCAT(' . implode(', ', $args) . ')';
    			case 'stringToUpper':
					self::cast_null($args, "''");
    				if(count($args) < 1)
    					return '""';
    				$fn = 'UPPER';
    				break;
    			case 'stringToLower':
					self::cast_null($args, "''");
    				if(count($args) < 1)
    					return '""';
    				$fn = 'LOWER';
    				break;
    			case 'substring':
					self::cast_null($args, "''");
    				if(count($args) < 2)
    					return 'NULL';
    				if(count($args) < 3)
    					return "SUBSTRING({$args[0]}, {$args[1]})";
					return "SUBSTRING({$args[0]}, {$args[1]}, {$args[2]})";
    			case 'round':
					self::cast_null($args, 0);
    				$fn = 'ROUND';
    				if(count($args) > 1)
    					return "ROUND({$args[0]}, {$args[1]})";
    				break;
    			default:
    				if(count($args))
    					return $args[0]; // pass through first argument
    				return 'NULL';
    		}
			if(count($args) < 1)
				return '0';
			if ($join_args) {
				return $fn . "(" . join($join_args, $args) . ")";
			} else {
				return "$fn({$args[0]})";
			}
		}
		else if($tree['type'] == 'value') {
			$val = $tree['value'];
			if($tree['subtype'] == 'special') {
				if($val == 'true')
					return 1;
				if($val == 'false')
					return 0;
			}
			else if($tree['subtype'] == 'variable') {
				$name = substr($val, 1);
				return array_get_default($fieldmap, $name, 'NULL');
			}
			else if($tree['subtype'] == 'string') {
				$c = substr($val, 0, 1);
				$val = stripslashes(substr($val, 1, -1));
				return '"' . $db->quote($val) . '"';
			}
			else if($tree['subtype'] == 'number') {
				return $val;
			}
		}
		return 'NULL';
    }


    /**
     * Get supported functions list
     *
     * @static
     * @return array
     */
    static function get_allowed_functions() {
        $functions = AppConfig::setting('field_formulas.operations', array());
        return $functions;
    }
}
?>
