<?php

namespace Psr\Log;

/**
 * Describes log levels
 */
class LogLevel
{
	const EMERGENCY = 'emergency';
	const ALERT	 = 'alert';
	const CRITICAL  = 'critical';
	const ERROR	 = 'error';
	const WARNING   = 'warning';
	const NOTICE	= 'notice';
	const INFO	  = 'info';
	const DEBUG	 = 'debug';
}

/**
 * Describes a logger instance
 *
 * The message MUST be a string or object implementing __toString().
 *
 * The message MAY contain placeholders in the form: {foo} where foo
 * will be replaced by the context data in key "foo".
 *
 * The context array can contain arbitrary data, the only assumption that
 * can be made by implementors is that if an Exception instance is given
 * to produce a stack trace, it MUST be in a key named "exception".
 *
 * See https://github.com/php-fig/fig-standards/blob/master/accepted/PSR-3-logger-interface.md
 * for the full interface specification.
 */
interface LoggerInterface
{
	/**
	 * System is unusable.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function emergency($message, array $context = array());

	/**
	 * Action must be taken immediately.
	 *
	 * Example: Entire website down, database unavailable, etc. This should
	 * trigger the SMS alerts and wake you up.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function alert($message, array $context = array());

	/**
	 * Critical conditions.
	 *
	 * Example: Application component unavailable, unexpected exception.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function critical($message, array $context = array());

	/**
	 * Runtime errors that do not require immediate action but should typically
	 * be logged and monitored.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function error($message, array $context = array());

	/**
	 * Exceptional occurrences that are not errors.
	 *
	 * Example: Use of deprecated APIs, poor use of an API, undesirable things
	 * that are not necessarily wrong.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function warning($message, array $context = array());

	/**
	 * Normal but significant events.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function notice($message, array $context = array());

	/**
	 * Interesting events.
	 *
	 * Example: User logs in, SQL logs.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function info($message, array $context = array());

	/**
	 * Detailed debug information.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function debug($message, array $context = array());

	/**
	 * Logs with an arbitrary level.
	 *
	 * @param mixed $level
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function log($level, $message, array $context = array());
}

/**
 * This is a simple Logger implementation that other Loggers can inherit from.
 *
 * It simply delegates all log-level-specific methods to the `log` method to
 * reduce boilerplate code that a simple Logger that does the same thing with
 * messages regardless of the error level has to implement.
 */
abstract class AbstractLogger implements LoggerInterface
{
	/**
	 * System is unusable.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function emergency($message, array $context = array())
	{
		$this->log(LogLevel::EMERGENCY, $message, $context);
	}

	/**
	 * Action must be taken immediately.
	 *
	 * Example: Entire website down, database unavailable, etc. This should
	 * trigger the SMS alerts and wake you up.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function alert($message, array $context = array())
	{
		$this->log(LogLevel::ALERT, $message, $context);
	}

	/**
	 * Critical conditions.
	 *
	 * Example: Application component unavailable, unexpected exception.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function critical($message, array $context = array())
	{
		$this->log(LogLevel::CRITICAL, $message, $context);
	}

	/**
	 * Runtime errors that do not require immediate action but should typically
	 * be logged and monitored.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function error($message, array $context = array())
	{
		$this->log(LogLevel::ERROR, $message, $context);
	}

	/**
	 * Exceptional occurrences that are not errors.
	 *
	 * Example: Use of deprecated APIs, poor use of an API, undesirable things
	 * that are not necessarily wrong.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function warning($message, array $context = array())
	{
		$this->log(LogLevel::WARNING, $message, $context);
	}

	/**
	 * Normal but significant events.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function notice($message, array $context = array())
	{
		$this->log(LogLevel::NOTICE, $message, $context);
	}

	/**
	 * Interesting events.
	 *
	 * Example: User logs in, SQL logs.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function info($message, array $context = array())
	{
		$this->log(LogLevel::INFO, $message, $context);
	}

	/**
	 * Detailed debug information.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function debug($message, array $context = array())
	{
		$this->log(LogLevel::DEBUG, $message, $context);
	}
	
	// generate a backtrace
	public function trace($message='', $level=null) {
		if(! isset($level)) $level = LogLevel::CRITICAL;
		$context = format_backtrace(debug_backtrace());
		$this->log($level, $message, $context);
	}

}


namespace Katzgrau\KLogger;

use DateTime;
use RuntimeException;
use Psr\Log\AbstractLogger;
use Psr\Log\LogLevel;

/**
 * Finally, a light, permissions-checking logging class.
 *
 * Originally written for use with wpSearch
 *
 * Usage:
 * $log = new Katzgrau\KLogger\Logger('/var/log/', Psr\Log\LogLevel::INFO);
 * $log->info('Returned a million search results'); //Prints to the log file
 * $log->error('Oh dear.'); //Prints to the log file
 * $log->debug('x = 5'); //Prints nothing due to current severity threshhold
 *
 * @author  Kenny Katzgrau <katzgrau@gmail.com>
 * @since   July 26, 2008 - Last update July 1, 2012
 * @link	http://codefury.net
 * @version 0.2.0
 */

/**
 * Class documentation
 */
class Logger extends AbstractLogger
{

	const MAX_LOG_SIZE = 128 * 1024 * 1024;
	const DEFAULT_LOG_SIZE = 128 * 1024 * 1024;
	const MAX_LOG_COUNT = 20;
	const DEFAULT_LOG_COUNT = 5;

	/**
	 * Path to the log file
	 * @var string
	 */
	private $logFilePath = null;
	private $logFileName = null;
	private $logDirectory = null;
	private $logHeader = null;
	private $maxFileSize = self::DEFAULT_LOG_SIZE;
	private $backupCount = self::DEFAULT_LOG_COUNT;
	private $context;

	/**
	 * Current minimum logging threshold
	 * @var integer
	 */
	private $logLevelThreshold = LogLevel::DEBUG;

	private $logLevels = array(
		LogLevel::EMERGENCY => 1,
		LogLevel::ALERT	 => 2,
		LogLevel::CRITICAL  => 3,
		LogLevel::ERROR	 => 4,
		LogLevel::WARNING   => 5,
		LogLevel::NOTICE	=> 6,
		LogLevel::INFO	  => 7,
		LogLevel::DEBUG	 => 8,
	);

	/**
	 * This holds the file handle for this instance's log file
	 * @var resource
	 */
	private $fileHandle = null;
	private $fileSize = 0;

	/**
	 * Valid PHP date() format string for log timestamps
	 * @var string
	 */
	private $dateFormat = 'Y-m-d G:i:s,u';

	/**
	 * Octal notation for default permissions of the log file
	 * @var integer
	 */
	private $defaultPermissions = 0666;
	
	private $lastError;

	/**
	 * Class constructor
	 *
	 * @param string  $context       Context tag to prefix log messages
	 * @param array   $params        Override default parameters from config
	 * @return void
	 */
	public function __construct($context = null, $params = null)
	{
		$this->setContext($context);
		$this->init($params);
	}
	
	public function init($params = null) {
		if(defined('NULL_LOGGER') || $params === false)
			$params = [];
		else if(! is_array($params))
			$params = \AppConfig::setting('site.log');
		if(isset($params['file']) && is_string($params['file'])) {
			$filename = trim($params['file']);
			if (substr(strtolower($filename), -4, 4) !== ".log") {
				$filename .= ".log";
			}
			$log_dir = isset($params['dir']) ? $params['dir'] : \AppConfig::setting('site.paths.logs_dir');
			$this->setLogFilePath($filename, $log_dir);
		} else
			$this->setLogFilePath(null);
		$this->setLogHeader(isset($params['header']) ? $params['header'] : null);
		$this->setMaxFileSize(isset($params['file_size']) ? $params['file_size'] : '10MB');
    	$this->setBackupCount(isset($params['backup_count']) ? $params['backup_count'] : 5);
		$this->setLogLevelThreshold(isset($params['level']) ? $params['level'] : 'warn');
		$this->openLog();
	}

	/**
	 * Class destructor
	 */
	public function __destruct()
	{
		$this->closeLog();
	}
	
	public function setContext($c)
	{
		$this->context = $c;
	}
	
	public function setLogHeader($head)
	{
		$this->logHeader = $head;
	}
	
	public function setLogFilePath($path, $dir=null)
	{
		if($path && strlen($dir))
			$path = $dir . '/' . $path;
		$path = str_replace(DIRECTORY_SEPARATOR, '/', $path);
		$this->logFilePath = $path;
		if(($p = strrpos($path, '/')) !== null) {
			$this->logDirectory = substr($path, 0, $p);
			$this->logFileName = substr($path, $p+1);
		} else {
			$this->logDirectory = '.';
			$this->logFileName = $path;
		}
	}
	
	public function setMaxFileSize($size)
	{
		$setSize = self::DEFAULT_LOG_SIZE;
		if(is_string($size) && preg_match('~^(\d*\.?\d+)\s*(([KMG])B?)?$~', $size, $m)) {
			$setSize = $m[1];
			if($m[3] == 'K')
				$setSize *= 1024;
			else if($m[3] == 'M')
				$setSize *= 1024 * 1024;
			else if($m[3] == 'G')
				$setSize *= 1024 * 1024 * 1024;
			$setSize = round($setSize);
		}
		else if (is_numeric($size)) $setSize = $size;
		if ($setSize < 0) $setSize = self::DEFAULT_LOG_SIZE;
		else if ($setSize > self::MAX_LOG_SIZE) $setSize = self::MAX_LOG_SIZE;
		$this->maxFileSize = $setSize;
	}
	
	public function setBackupCount($count)
	{
		$setCount = self::DEFAULT_LOG_COUNT;
		if (is_numeric($count)) $setCount = $count;
		if ($setCount < 1) $setCount = 1;
		elseif ($setCount > self::MAX_LOG_COUNT) $setCount = self::MAX_LOG_COUNT;
		$this->backupCount = $setCount;
	}
	
	public function checkPermissions()
	{
		if(! $this->logFilePath)
			return false;
		if(! file_exists($this->logDirectory)) {
			if(! mkdir($this->logDirectory, $this->defaultPermissions, true)) {
				$this->lastError = "Error creating log directory ($this->logDirectory). Check permissions.";
				return false;
			}
		}
		if(! is_dir($this->logDirectory)) {
			$this->lastError = "Error creating log directory ($this->logDirectory). Already exists as a file.";
			return false;
		}
		if(file_exists($this->logFilePath) && ! is_writable($this->logFilePath)) {
			$this->lastError = "The system log file ($this->logFilePath) could not be written to. Check that appropriate permissions have been set.";
			return false;
		}
		return true;
	}
	
	public function openLog()
	{
		$this->closeLog();
		if(! $this->checkPermissions())
			return false;
		$this->fileHandle = @fopen($this->logFilePath, 'a');
		if(! $this->fileHandle) {
			$this->lastError = "The system log file ($this->logFilePath) could not be opened. Check permissions.";
			return false;
		}
		$this->fileSize = filesize($this->logFilePath);
		if($this->logHeader && ! $this->fileSize)
			$this->write($this->logHeader.PHP_EOL);
		return true;
	}
	
	public function closeLog()
	{
		$ret = false;
		if($this->fileHandle) {
			fclose($this->fileHandle);
			$ret = true;
		}
		$this->fileHandle = null;
		return $ret;
	}
	
	public function getLastError()
	{
		return $this->lastError;
	}

	/**
	 * Sets the date format used by all instances of KLogger
	 * 
	 * @param string $dateFormat Valid format string for date()
	 */
	public function setDateFormat($dateFormat)
	{
		$this->dateFormat = $dateFormat;
	}

	/**
	 * Sets the Log Level Threshold
	 * 
	 * @param string $dateFormat Valid format string for date()
	 */
	public function setLogLevelThreshold($logLevelThreshold)
	{
		$lvl = $this->getLogLevel($logLevelThreshold);
		if($lvl) $this->logLevelThreshold = $lvl;
	}

	/**
	 * Logs with an arbitrary level.
	 *
	 * @param mixed $level
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function log($level, $message, array $context = array())
	{
		$level = $this->fixLogLevel($level);
		$levelNum = $this->getLogLevel($level, true);
		if(! $levelNum || $this->logLevelThreshold < $levelNum) {
			return;
		}
		$message = $this->formatMessage($level, $message, $context);		
		$this->write($message);
	}

	/**
	 * Writes a line to the log without prepending a status or timestamp
	 *
	 * @param string $line Line to write to the log
	 * @return void
	 */
	public function write($message)
	{
		if($this->fileHandle) {
			if (@fwrite($this->fileHandle, $message) === false) {
				throw new RuntimeException('The system log file could not be written to. Check that appropriate permissions have been set.');
			}
			if (isset($this->maxFileSize) && ftell($this->fileHandle) + $this->fileSize > $this->maxFileSize)    
				$this->rollOverLogs();
		}
	}
	
	public function writeLine($message)
	{
		return $this->write($message . PHP_EOL);
	}
	
	function rollOverLogs()
	{
		if($this->backupCount <= 0 || ! $this->logFilePath)
			return false;
		$fileName = $this->logFilePath;
		$lastFile = "$fileName.{$this->backupCount}";
		if(is_writable($lastFile))
			unlink($lastFile);
		for($i = $this->backupCount - 1; $i >= 1; $i--) {
			$nextFile = "$fileName.$i";
			if(is_readable($nextFile)) {
				$lastFile = "$fileName." . ($i + 1);
				rename($nextFile, $lastFile);
			}
		}
		$nextFile = "$fileName.1";
		if($this->closeLog())
			$reopen = 1;
		rename($fileName, $nextFile);
		if(! empty($reopen))
			$this->openLog();
	}

	/**
	 * Formats the message for logging.
	 *
	 * @param  string $level   The Log Level of the message
	 * @param  string $message The message to log
	 * @param  array  $context The context
	 * @return string
	 */
	private function formatMessage($level, $message, $context)
	{
		$level = strtoupper($level);
		if(! is_string($message))
			$message = $this->stringFormatMessage($message);
		if (! empty($context)) {
			$message .= PHP_EOL.$this->indent($this->contextToString($context));
		}
		$pid = getmypid();
		$ctx = isset($this->context) ? ' ['.$this->context.']' : '';
		return "{$this->getTimestamp()} [$pid] {$level}{$ctx} - {$message}".PHP_EOL;
	}
	
	private function stringFormatMessage($m) {
		return var_export($m, true);
	}

	/**
	 * Gets the correctly formatted Date/Time for the log entry.
	 * 
	 * PHP DateTime is dump, and you have to resort to trickery to get microseconds
	 * to work correctly, so here it is.
	 * 
	 * @return string
	 */
	private function getTimestamp()
	{
		$originalTime = microtime(true);
		$micro = sprintf("%06d", ($originalTime - floor($originalTime)) * 1000000);
		$date = new DateTime(date('Y-m-d H:i:s.'.$micro, round($originalTime)));

		return $date->format($this->dateFormat);
	}

	/**
	 * Takes the given context and coverts it to a string.
	 * 
	 * @param  array $context The Context
	 * @return string
	 */
	private function contextToString($context)
	{
		$export = '';
		foreach ($context as $key => $value) {
			$export .= "{$key}: ";
			$export .= preg_replace(array(
				'/=>\s+([a-zA-Z])/im',
				'/array\(\s+\)/im',
				'/^  |\G  /m',
			), array(
				'=> $1',
				'array()',
				'	',
			), str_replace('array (', 'array(', var_export($value, true)));
			$export .= PHP_EOL;
		}
		return str_replace(array('\\\\', '\\\''), array('\\', '\''), rtrim($export));
	}

	/**
	 * Indents the given string with the given indent.
	 * 
	 * @param  string $string The string to indent
	 * @param  string $indent What to use as the indent.
	 * @return string
	 */
	private function indent($string, $indent = '	')
	{
		return $indent.str_replace("\n", "\n".$indent, $string);
	}
	
	public function fixLogLevel($level)
	{
		$level = strtolower($level);
		if($level == 'all') $level = LogLevel::DEBUG;
		else if($level == 'warn') $level = LogLevel::WARNING;
		else if($level == 'fatal') $level = LogLevel::CRITICAL;
		if(isset($this->logLevels[$level]))
			return $level;
		return null;
	}
	
	public function getLogLevel($level, $fixed=false)
	{
		if(is_int($level)) return $level;
		if(! $fixed)
			$level = $this->fixLogLevel($level);
		if($level)
			return $this->logLevels[$level];
		return 0;
	}
	
	public function getLogFilePath()
	{
		return $this->logFilePath;
	}

	public function getLogFileName()
	{
		return $this->logFileName;
	}
	
	// aliases for old methods
	public function fatal($message, array $context = array())
	{
		$this->log(LogLevel::CRITICAL, $message, $context);
	}
	public function warn($message, array $context = array())
	{
		$this->log(LogLevel::WARNING, $message, $context);
	}	
}

?>
