<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo, 
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */
if(!defined('sugarEntry') || !sugarEntry) die('Not A Valid Entry Point');


/*
 * class to make a single HTTP request using either cURL or native methods.
 * use instead of file_get_contents for remote URLs.
 */
 
class RemoteRequest {
	var $url;
	var $effective_url;
	var $postdata;
	var $request_headers;
	var $method = 'GET';
	var $raise_exceptions = false;
	var $ignore_error_status = false;
	var $timeout = 30;
	var $connect_timeout = null;
	var $default_method = 'curl';
	var $auto_proxy = true;
	var $redirects = 0;
	var $max_redirects = 20;
	var $status_line;
	var $status_ver;
	var $status_code;
	var $response_headers;
	var $response_length;
	var $error;
	var $progress_callback;

	/**
	 *
	 * Convenience factory method
	 *
	 */
	static function create($url, $method = 'GET', $postdata = null, $headers = null, $options = []) {
		$timeout = array_get_default($options, 'timeout');
		$connect_timeout = array_get_default($options, 'connect_timeout');
		$exceptions = array_get_default($options, 'exceptions');
		$auth = array_get_default($options, 'auth');
		if (is_array($auth)) {
			$username = array_get_default($auth, 'user');
			$password = array_get_default($auth, 'password');
			if (strlen($username) && strlen($password)) {
				if (!is_array($headers)) {
					$headers = [];
				}
				$headers[] = 'Authorization: Basic ' . base64_encode($username . ':' . $password);
			}
		}
		$req = new self($url, $postdata, $headers, $exceptions, $timeout, $connect_timeout);
		$method = strtoupper($method);
		if ($postdata !== null && $method != 'POST' && $method != 'PATCH' && $method != 'PUT') {
			$method = 'POST';
		}
		$req->method = $method;
		return $req;
	}
	
	function __construct($url, $postdata=null, $headers=null, $except=false, $timeout=null, $connect_timeout=null) {
		$this->url = $url;
		$this->postdata = $postdata;
		if ($postdata !== null) {
			$this->method = 'POST';
		}
		$this->request_headers = $headers;
		$this->raise_exceptions = $except;
		if(isset($timeout))
			$this->timeout = $timeout;
		if(isset($connect_timeout))
			$this->connect_timeout = $timeout;
		else
			$this->connect_timeout = ini_get('default_socket_timeout');
	}
	
	function setProgressCallback(Callable $cb) {
		$this->progress_callback = $cb;
	}
	
	function fetch() {
		if($this->default_method == 'curl' && function_exists('curl_init'))
			return $this->fetchCurl();
		return $this->fetchNative();
	}
	
	function parseStatus($status) {
		$this->status_line = $status;
		if(! preg_match('~^(HTTP/[\d\.]+) (\d+) ~', $status, $m))
			return false; // bad http status
		$this->status_ver = $m[1];
		$this->status_code = $m[2];
		return true;
	}
	
	function curlProgress($resource, $download_size, $downloaded, $upload_size, $uploaded) {
		$this->logProgress($download_size, $downloaded);
	}
	
	function logProgress($download_size, $downloaded, $connected=true) {
		if(isset($this->progress_callback)) {
			$perc = ($download_size ? $downloaded / $download_size * 80.0 : 0);
			if($connected) $perc = 20.0 + $perc * 0.8;
			$pgc = $this->progress_callback;
			$pgc($perc, $download_size, $downloaded);
		}
	}
	
	function fetchCurl() {
		$ch = curl_init();
		
		curl_setopt($ch, CURLOPT_URL, $this->url);
		if($this->request_headers) {
			curl_setopt($ch, CURLOPT_HTTPHEADER, $this->request_headers);
		}
		curl_setopt($ch, CURLOPT_HEADER, 1);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
		curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, $this->connect_timeout);
		curl_setopt($ch, CURLOPT_TIMEOUT, $this->timeout);
		if(! ini_get('open_basedir')) {
			curl_setopt($ch, CURLOPT_FOLLOWLOCATION, 1);
			curl_setopt($ch, CURLOPT_MAXREDIRS, $this->max_redirects);
		}
		if(isset($this->postdata)) {
			curl_setopt($ch, CURLOPT_POSTFIELDS, $this->postdata);
		}
		curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $this->method);
		if($this->auto_proxy && ($proxy = AppConfig::get_proxy_info())) {
			curl_setopt($ch, CURLOPT_PROXY, $proxy['host']);
			curl_setopt($ch, CURLOPT_PROXYPORT, $proxy['port']);
			if($proxy['auth'])
				curl_setopt($ch, CURLOPT_PROXYUSERPWD, $proxy['username'].':'.$proxy['password']);
		}
		curl_setopt($ch, CURLOPT_PROGRESSFUNCTION, [$this, 'curlProgress']);
		curl_setopt($ch, CURLOPT_NOPROGRESS, false);
		
		$data = curl_exec($ch);
		$this->error = curl_error($ch);

		$hsize = curl_getinfo($ch, CURLINFO_HEADER_SIZE);
		$this->response_length = curl_getinfo($ch, CURLINFO_CONTENT_LENGTH_DOWNLOAD);
		$this->redirects = curl_getinfo($ch, CURLINFO_REDIRECT_COUNT);
		$this->effective_url = curl_getinfo($ch, CURLINFO_EFFECTIVE_URL);

		curl_close($ch);

		if($data !== false) {
			if($hsize) {
				$head = substr($data, 0, $hsize);
				$data = substr($data, $hsize);
				$hlines = explode("\r\n", $head);
				// cURL may return multiple HTTP status lines depending on redirects, also 100 Continue
				// so we remove those extra lines
				$last = null;
				for($i = count($hlines)-1; $i > 0; $i--) {
					if(substr($hlines[$i], 0, 5) == 'HTTP/') {
						$last = $i;
						break;
					}
				}
				if($last) $hlines = array_slice($hlines, $last);
				$this->response_headers = [];
				if(! $this->parseStatus(array_shift($hlines)))
					$data = false;
				else {
					foreach($hlines as $line) {
						if( ($p = strpos($line, ':')) !== false)
							$this->response_headers[strtolower(trim(substr($line, 0, $p)))] = trim(substr($line, $p+1));
					}
				}
			} else
				$data = false;
		}

		return $data;
	}
	
	
	/**
	  * More complicated than file_get_contents in order to avoid long delays
	  * when fetching remote data over https.
	  * Issue happens on at least php 5.6.23 and 7.0.4 with our server, where
	  * requests are not closed after finishing and stay pending for 60+ seconds.
	  * Maybe related to https://bugs.php.net/bug.php?id=61285
	  * or https://bugs.php.net/bug.php?id=68329
	  * This issue is present in each of file_get_contents, normal fsockopen/fread,
	  * and http stream wrappers.
	  */
	function fetchNative() {
		$url = $this->url;
		$this->redirects = 0;
		
		while($this->redirects < $this->max_redirects) {
			$this->effective_url = $url;
			$parse = parse_url($url);
			if(! $parse) {
				$this->error = "Error parsing URL";
				return false;
			}
			$host = array_get_default($parse, 'host');
			if(! $host) {
				$this->error = "Missing URL hostname";
				return false;
			}
		
			$fetch = $fetch_host = $host;
			$port = array_get_default($parse, 'port');
			$scheme = array_get_default($parse, 'scheme', 'http');
			if($scheme == 'https' || $scheme == 'ssl') {
				$fetch = 'ssl://'.$fetch;
				if(! $port) $port = 443;
			}
			else if($scheme == 'tls')
				$fetch = 'tls://'.$fetch;
			if($port) {
				$fetch_port = $port;
				$port = ':'.$port;
			} else {
				$fetch_port = 80;
			}
			$path = $fetch_path = array_get_default($parse, 'path', '/');
			$query = array_get_default($parse, 'query');
			if(isset($query)) $fetch_path .= '?'.$query;
			$req_headers = $this->request_headers ? $this->request_headers : [];
			
			if($this->auto_proxy && ($proxy = AppConfig::get_proxy_info())) {
				$fetch = $proxy['host'];
				$fetch_host = $proxy['host'];
				$fetch_port = $proxy['port'];
				$fetch_path = $url;
				if($proxy['auth']) {
					$req_headers['Proxy-Authorization'] = 'Basic ' . base64_encode($proxy['username'].':'.$proxy['password']);
				}
			}

			$sock = fsockopen($fetch, $fetch_port, $errno, $errstr, $this->connect_timeout);
			if(! $sock) {
				$this->error = "Socket creation error";
				return false; // socket creation error
			}
			$method = 'GET';
			if(isset($this->postdata)) $method = 'POST';

			$request  = "$method $fetch_path HTTP/1.1\r\n";
			$request .= "Host: $fetch_host\r\n";
			$sent_headers = array();
			foreach($req_headers as $k => $v) {
				if(is_int($k)) {
					if( ($p = strpos($v, ':')) !== false)
						$sent_headers[strtolower(substr($v, 0, $p))] = ltrim(substr($v, $p+1));
					$request .= "$v\r\n";
				} else {
					$sent_headers[strtolower($k)] = $v;
					$request .= "$k: $v\r\n";
				}
			}
			if(isset($this->postdata)) {
				if(! isset($sent_headers['content-type']))
					$request .= "Content-Type: application/x-www-form-urlencoded\r\n";
				$request .= 'Content-Length: ' . mb_strlen($this->postdata, '8bit') . "\r\n";
			}
			$request .= "Connection: close\r\n\r\n";
			if(isset($this->postdata))
				$request .= $this->postdata;

			$this->status_line = null;
			$this->status_ver = null;
			$this->status_code = null;
			$this->response_length = 0;
			$this->response_headers = [];
			$headers = [];
			$response = '';
			$chunked = false;
			$read_length = 0;

			if(! fwrite($sock, $request))
				return false; // write error

			stream_set_timeout($sock, $this->timeout);
			while(! feof($sock)) {
				$line = stream_get_line($sock, 1024, "\r\n");
				if($line === false) {
					$this->error = "Socket read error";
					return false;
				}
				if(! strlen($line))
					// blank line - done headers
					break;
				if(! $this->status_line) {
					if(! $this->parseStatus($line)) {
						$this->error = "Error parsing HTTP status";
						return false;
					}
				}
				else {
					if( ($p = strpos($line, ':')) !== false)
						$headers[strtolower(trim(substr($line, 0, $p)))] = trim(substr($line, $p+1));
				}
			}
			$this->response_headers = $headers;
		
			if(($this->status_code == 301 || $this->status_code == 302) && ! empty($headers['location'])) {
				$loc = $headers['location'];
				if(! preg_match('~^[a-zA-Z]+://~', $loc)) {
					// relative url
					if(substr($loc, 0, 1) !== '/') {
						$pfx = $path;
						if( ($p = strrpos($pfx, '/')) !== false)
							$pfx = substr($pfx, 0, $p+1);
						$loc = $pfx . $loc;
					}
					$loc = "$scheme://$host:$port".$loc;
				}
				$url = $loc;
				fclose($sock);
				continue;
			}
			if(! $this->ignore_error_status && ($this->status_code < 200 || $this->status_code > 299)) {
				$this->error = "Error status from server: {$this->status_line}";
				return false; // not found, server error, etc
			}
		
			if(strtolower(array_get_default($headers, 'transfer-encoding')) == 'chunked')
				$chunked = true;
			if(! isset($headers['content-length']) && ! $chunked) {
				$response = stream_get_contents($sock);
				$len = mb_strlen($response, '8bit');
				$this->logProgress($len, $len);
				return $response;
			}
			$read_length = array_get_default($headers, 'content-length');
		
			while($chunked || $read_length) {
				if($chunked) {
					$read_length = stream_get_line($sock, 1024, "\r\n");
					if($read_length === false && ! feof($sock)) {
						$this->error = "Error reading from socket";
						return false;
					}
					$read_length = base_convert($read_length, 16, 10);
					if(! $read_length)
						break;
				}

				while($read_length && ! feof($sock)) {
					$ask_len = min($read_length, 1024);
					$chunk = fread($sock, $ask_len);
					if($chunk === false) {
						$this->error = "Error reading from socket";
						return false;
					}
					$response .= $chunk;
					$length = mb_strlen($chunk, '8bit');
					$read_length -= $length;
					$this->response_length += $length;
					$this->logProgress($chunked ? 0 : $read_length, $this->response_length);
					if($read_length) {
						$info = stream_get_meta_data($sock);
						if($info['timed_out']) {
							$this->error = "Connection timed out";
							return false; // timed out
						}
					}
				}
			
				if($chunked) {
					$blank = stream_get_line($sock, 1024, "\r\n");
					if($blank !== '' && ! feof($sock)) {
						$this->error = "Error reading chunked data";
						return false; // expected blank line after chunk
					}
				}
			}
		
			fclose($sock);
			$this->logProgress($this->response_length, $this->response_length);
			return $response;
		}
		
		$this->error = "Too many redirects";
		return false;
	}
	
	function getError() {
		return $this->error;
	}
}


if(class_exists('SoapClient')) {
	class IAHSoapClient extends SoapClient {
		var $conn;
		var $progress_callback;
		
		function setProgressCallback(Callable $cb) {
			$this->progress_callback = $cb;
		}
		
		public function __doRequest($request, $location, $action, $version, $one_way = false) : ?string {
			$headers = ['SoapAction' => $action];
			if($version == SOAP_1_2)
				$headers['Content-Type'] = 'application/soap+xml; charset=utf-8'; // + ; action=""?
			else
				$headers['Content-Type'] = 'text/xml; charset=utf-8';
			$this->conn = new RemoteRequest($location, $request, $headers);
			if($this->progress_callback)
				$this->conn->setProgressCallback($this->progress_callback);
			$this->conn->ignore_error_status = true; // soapfault uses 500 error
			$ret = $this->conn->fetch();
			if(! $ret) $ret = '';
			return $ret;
		}
		
		function decode($data) {
			$ret = array();
			foreach($data as $k => $v) {
				if(is_object($v) || is_array($v))
					$v = $this->decode($v);
				$ret[$k] = $v;
			}
			return $ret;
		}
	}
}

class SoapCompat {
	var $disable_native = false;
	var $url = null;
	var $wsdl = null;
	var $instance;
	var $instance_type;
	var $faultcode;
	var $faultstring;
	var $error;
	var $progress_callback;
	
	
	function __construct($url) {
		$this->url = $url;
		if(AppConfig::setting('site.performance.disable_native_soap'))
			$this->disable_native = true;
	}
	
	function setProgressCallback(Callable $cb) {
		$this->progress_callback = $cb;
	}
	
	function create() {
		if($this->instance) return true;
		if(class_exists('IAHSoapClient') && ! $this->disable_native) {
			try {
				$wsdl = $this->fetchWsdl();
				$params = array(
					'encoding' => 'utf-8',
					'exceptions' => true,
					//'trace' => true,
					'cache_wsdl' => WSDL_CACHE_NONE,
					'features' => SOAP_USE_XSI_ARRAY_TYPE,
				);
				$proxy = AppConfig::get_proxy_info();
				if($proxy) {
					$params['proxy_host'] = $proxy['host'];
					$params['proxy_port'] = $proxy['port'];
					if($proxy['auth']) {
						$params['proxy_login'] = $proxy['username'];
						$params['proxy_password'] = $proxy['password'];
					}
				}
				$this->instance = @new IAHSoapClient($wsdl, $params);
				if($this->progress_callback)
					$this->instance->setProgressCallback($this->progress_callback);
				$this->instance_type = 'soapclient';
				return true;
			} catch (SoapFault $e) {
				$this->faultcode = $e->faultcode;
				$this->error = $this->faultstring = $e->faultstring;
				// create nusoap instead
			}
		}
		require_once('include/nusoap/nusoap.php');
		$this->instance = new nusoap_client($this->url, false, false, false, false, false, 30, 30);
		$this->instance_type = 'nusoap';
		return true;
	}
	
	function fetchWsdl() {
		if($this->wsdl) {
			$wsdl = $this->wsdl;
		} else {
			$wsdl = $this->url . '?wsdl';
		}
		if(strpos($wsdl, '://') === false)
			return $wsdl;
		$req = new RemoteRequest($wsdl);
		$data = $req->fetch();
		if($data) {
			$tmp = tempnam(sys_get_temp_dir(), 'wsdl');
			if($tmp && ($fp = fopen($tmp, 'w'))) {
				fwrite($fp, $data);
				fclose($fp);
				return $tmp;
			} else {
				$this->error = "Error writing temp file";
			}
		} else {
			$this->error = "Error fetching wsdl: ".$req->getError();
		}
	}
	
	function exec($action, $args=null) {
		if(! $this->create()) return;
		if(! $args) $args = array();
		if($this->instance_type == 'soapclient') {
			try {
				$result = call_user_func_array([$this->instance, $action], $args);
				if(! $result)
					$this->error = "No response from server";
				$result = $this->instance->decode($result); // convert to array format
			} catch (SoapFault $e) {
				$this->faultcode = $e->faultcode;
				$this->faultstring = $e->faultstring;
				$this->error = $e->faultcode . ': ' . $e->faultstring;
				$result = null;
			}
		}
		else {
			$result = $this->instance->call($action, $args);
			$this->error = $this->instance->getError();
			$this->faultcode = $this->instance->faultcode;
			$this->faultstring = $this->instance->faultstring;
		}
		return $result;
	}
	
	function getError() {
		return $this->error;
	}
}

