<?php

if(!defined('sugarEntry') || !sugarEntry) die('Not A Valid Entry Point');

use League\OAuth2\Client\Grant\RefreshToken;
use OneCRM\OAuth\OAuthProvider;
use OneCRM\OAuth\OAuthTokenOwner;

require_once 'include/utils/dl_utils.php';
require_once 'include/OAuth/OAuthProvider.php';
require_once 'include/OAuth/OAuthTokenOwner.php';


function oauth2Encrypt($data, $key, &$out, $isPrivate) {
	$password = openssl_random_pseudo_bytes(128);
	$ok = $isPrivate
		? openssl_private_encrypt($password, $encryptedPassword, $key, OPENSSL_PKCS1_PADDING)
		: openssl_public_encrypt($password, $encryptedPassword, $key, OPENSSL_PKCS1_PADDING)
		;
	if (!$ok) return $ok;
	$iv = openssl_random_pseudo_bytes(16);
	$encrypted = openssl_encrypt($data, 'aes-256-ctr', $password, OPENSSL_RAW_DATA, $iv);
	if ($encrypted === false) return false;
	$out = base64_encode($iv . $encryptedPassword . $encrypted);
	return true;
}

function oauth2Decrypt($data, $key, &$out, $isPrivate) {
	$data = base64_decode($data);
	if (strlen($data) < 272) return false;
	$iv = substr($data, 0, 16);
	$encryptedPassword = substr($data, 16, 256);
	$ok = $isPrivate
		? openssl_private_decrypt($encryptedPassword, $password, $key, OPENSSL_PKCS1_PADDING)
		: openssl_public_decrypt($encryptedPassword, $password, $key, OPENSSL_PKCS1_PADDING)
		;
	if (!$ok) return $ok;
	$decrypted = openssl_decrypt(substr($data, 272), 'aes-256-ctr', $password, OPENSSL_RAW_DATA, $iv);
	if ($decrypted === false) return false;
	$out = $decrypted;
	return true;
}

function getOAuth2Keys() {
	$config_key = "site.oauth2_server_keys";
	$keys = AppConfig::setting($config_key);
	if (!$keys) {
		$private_key = openssl_pkey_new([
			'private_key_type' => OPENSSL_KEYTYPE_RSA,
		    "private_key_bits" => 2048,
		]);
		$public_key_pem = openssl_pkey_get_details($private_key)['key'];
		openssl_pkey_export($private_key, $exported);
		$keys = ['public' => $public_key_pem, 'private' => $exported];
		AppConfig::set_local($config_key, $keys);
		AppConfig::save_local();
	}
	return $keys;
}

function getOAuth2ProxyKey($service = null) {
	$server = AppConfig::setting('company.oauth_server');
	$force = $service == 'gmail' || $service == 'google_sync';
	if ($server == 'custom' || $force) {
		return getOAuth2Keys()['public'];
	}
	return AppConfig::setting("company.oauth2_crypto_keys.$server");
}


function buildOAuth2RedirectUri($service) {
	$server = str_replace('_', '.', AppConfig::setting('company.oauth_server'));
	$force = $service == 'gmail' || $service == 'google_sync';
	if ($server == 'custom' || $force) {	
		$uri =  AppConfig::site_url() . 'oauth_redirect.php';
	} else {
		$uri =  "https://$server/";
	}
	return $uri;
}

function buildOAuth2Uri(string $service, array $extra = []) {
	$keys = getOAuth2Keys();
	$params = array_merge($extra, [
		'user_id' => AppConfig::current_user_id(),
		'service' => $service,
		'nonce' => uniqid("_"),
	]);
	$encoded = json_encode($params);

	oauth2Encrypt($encoded, $keys['public'], $state, false);
	$redirect = AppConfig::site_url() . 'oauth_redirect.php';
	$req = json_encode([
		'state' => $state,
		'redirect' => $redirect,
		'service' => $service,
		'signature' => sha1(AppConfig::unique_key() . '|' . $redirect),
	]);

	$key = getOAuth2ProxyKey($service);
	oauth2Encrypt($req, $key, $data, false);

	$uri = buildOAuth2RedirectUri($service);
	return "$uri?_a=" . urlencode($data) . '&_s=' . urlencode($service);
}


function refreshOauth2TokenCustom($service, $owner, $refreshToken) {
	$provider = new OAuthProvider([
		'tokenOwner' => $owner,
		'service' => $service,
	]);
	$token = $provider->getAccessToken(new RefreshToken(), ['refresh_token' => $refreshToken], true);
	return $token;
}

function refershOauth2Token($service, $owner, $refreshToken) {
	$result = false;
	$server = str_replace('_', '.', AppConfig::setting('company.oauth_server'));
	if ($server == 'custom' || $service == 'gmail' || $service == 'google_sync') {
		$result = refreshOauth2TokenCustom($service, $owner, $refreshToken);
	} else {
		$key = getOAuth2ProxyKey($service);
		$nonce = uniqid("_");
		$params = ['service' => $service, 'refreshToken' => $refreshToken, 'nonce' => $nonce];
		$encoded = json_encode($params);
		oauth2Encrypt($encoded, $key, $encrypted, false);
		$url = "https://$server/?_r=" . urlencode($encrypted) ;
		$req = new RemoteRequest($url);
		$req->max_redirects = 2;
		$data = $req->fetch();
		if ($data !== false) {
			$reply = @json_decode($data, true);
			if ($reply && isset($reply['token'])) {
				$result = $reply['token'];
			}
		}
	}
	if ($result) {
		if (!isset($result['refresh_token'])) {
			$result['refresh_token'] = $refreshToken;
		}
		$tokenOwner = new OAuthTokenOwner($owner[0], $owner[1] ?? "");
		$tokenOwner->persist($result);
	}
	return $result;
}


function processOAuthAuthenticationRequest($request) {
	$keys = getOAuth2Keys();
	$ok = oauth2Decrypt($request, $keys['private'], $json, true);
	if (!$ok) die('Bad request');
	$request = json_decode($json, true);
	$auth = new OAuthProvider(['service' => $request['service']]);
	$authUrl = $auth->getAuthorizationUrl(['state' => $request['state']]);
	header('Location: ' . $authUrl);
	exit;
}


function processOAuthTokenRequest($request) {
	$key = getOAuth2ProxyKey();
	$keys = getOAuth2Keys();

	$ok = oauth2Decrypt($request, $key, $data, false);
	if (!$ok) die('Bad request');
	$data = json_decode($data, true);
	if (!$data) die('Bad request');

	$error = null;
	$ok = oauth2Decrypt($data['state'], $keys['private'], $stateJson, true);
	if (!$ok) die('Bad request');
	$state = @json_decode($stateJson, true);

	$success_message = translate('LBL_O365_AUTH_SuCCESS', 'Users');
	if (!is_array($state) || array_get_default($state, 'user_id') !== AppConfig::current_user_id()) {
		$error = translate('LBL_O365_AUTH_UNEXPECTED_ERROR', 'Users');
	}
	if (isset($data['error']) && !isset($error)) {
		$error = $data['error'];
	}
	formatOauth2Reply($error, $success_message, $data['token'], $state['input']);
}

function processOAuthCodeRequest($request) {
	$error_message =  $success_message = null;
	
	$state_encrypted = $request['state'];

	$keys = getOAuth2Keys();
	
	$ok = oauth2Decrypt($state_encrypted, $keys['private'], $state_json, true);
	$state = @json_decode($state_json, true);

	if (!is_array($state) || array_get_default($state, 'user_id') !== AppConfig::current_user_id()) {
		$error_message = translate('LBL_O365_AUTH_UNEXPECTED_ERROR', 'Users');
	}

	if (!$error_message) {
		if (!empty($_GET['code'])) {
			try {
				$auth = new OAuthProvider(['service' => $state['service']]);
				$token = $auth->getAccessToken('authorization_code', ['code' => $_GET['code']], false);
				$success_message = translate('LBL_O365_AUTH_SuCCESS', 'Users');
			} catch (Exception $e) {
				$error_message = $e->getMessage();
			}
		} elseif (!empty($_GET['error_description'])) {
			$error_message = $_GET['error_description'];
		} else {
			$error_message = translate('LBL_O365_AUTH_UNEXPECTED_ERROR', 'Users');
		}
	}
	formatOauth2Reply($error_message, $success_message, $token, $state['input']);
}

function formatOauth2Reply($error_message, $success_message, $token, $inputId) {
	$reply = $error_message 
		? ['success' => false, 'message' => $error_message]
		: ['success' => true, 'message' => $success_message, 'token' => $token]
		;

	$result = json_encode($reply);

	echo <<<SCRIPT
	<script>window.opener.o365_auth_result($result, '{$inputId}');
	window.close();
	</script>
	SCRIPT;
	exit;
}
