<?php

require_once 'include/SabreDAV/autoload.php';
require_once 'dav_token.php';

use Sabre\VObject\Reader;
use Sabre\VObject\Component\VCalendar;
use Sabre\VObject\Node;
use Sabre\VObject\Property\ICalendar\DateTime as SabreDateTime;
use Sabre\VObject\DateTimeParser;
use Sabre\VObject\Component\VTodo;
use Sabre\VObject\Component\VEvent;

class vCalUtils {

	/**
	 * Update vCal
	 * If $orig_vcal is not specified, it will be taken from $row,
	 * of fetched from database if not present in $row.
	 *
	 * @param RowResult|RowUpdate $row
	 * @param string $orig_vcal
	 *
	 * @return updated vCal text
	 */

	static private $EMPTY;

	static public function vcal_update($row, $orig_vcal = null, $update_rev = false) {
		self::$EMPTY = new StdClass;
		if (!$orig_vcal) {
			$orig_vcal = $row->getFieldValue('vcal', self::$EMPTY);
			if ($orig_vcal === self::$EMPTY) {
				$id = $row->getFieldValue('id');
				if (!$id) {
					$orig_vcal = '';
				} else {
					$vcal_row = ListQuery::quick_fetch_row($row->getModelName(), $id, array('vcal'));
					if ($vcal_row)
						$orig_vcal = @$vcal_row['vcal'];
					else
						$orig_vcal = '';
				}
			}
		}
		if($orig_vcal) {
			$vcal = self::read_vcal($orig_vcal);
		} else {
			$vcal = new VCalendar;
		}
		self::repair($row, $vcal);
		$vcal->VERSION = '2.0';


		$data_row = $row->getModelName() == 'RecurrenceRule' ? @json_decode($row->getField('template'), true) : $row->getRow();
		if (!is_array($data_row)) return;
		self::update_basic($data_row, $row->getModelName(), $vcal);
		self::update_dates($data_row, $row->getModelName(), $vcal);
		if ($row->getModelName() == 'Task')
			self::update_task($row, $vcal);
		else
			self::update_event($data_row, $vcal);
		self::update_recur($row, $vcal);

		//Try to Repair
		$vcal->validate(Node::REPAIR);
		//Check CalDAV errors
		$validate_res = $vcal->validate(Node::PROFILE_CALDAV);

		if (sizeof($validate_res) > 0) {
			foreach ($validate_res as $error) {
				$GLOBALS['log']->fatal("Can't create vCal Card for the {$row->getModelName()} {$row->getField('id')}: {$error['message']}");
			}

			return '';
		}

		$tokenType = $row->getModelName();
		if ($tokenType == 'RecurrenceRule')
			$tokenType = 'Meeting';
		self::update_sync_token($tokenType . 's');

		return $vcal->serialize();
	}

	public static function vcal_import($vcal_text, $row) {
		$data_updates = self::vcal_import_to_arrays($vcal_text, $row);
		if (!$data_updates) {
			return false;
		}
		list($data, $updates) = $data_updates;
		$row->set($updates);
		return true;
	}

	public static function vcal_import_to_arrays($vcal_text, $row) {
		self::$EMPTY = new StdClass;
		$vcal = self::read_vcal($vcal_text, false);
		if (!$vcal) {
			return false;
		}
		$data = [];
		$updates = [];
		if (!self::import_basic($vcal, $row, $data))
			return false;
		if (!self::import_dates($vcal, $row, $data)) 
			return false;
		if ($row->getModelName() == 'RecurrenceRule') {
			if (!self::import_recurrence($vcal, $row, $data))
				return false;
			$data['assigned_user_id'] = AppConfig::current_user_id();
			$updates['template'] = json_encode($data, JSON_UNESCAPED_SLASHES);
			$updates['parent_type'] = 'Meetings';
		} else {
			$updates = $data;
		}
		return [$data, $updates];
	}

	/**
	 * Returns current CalDAV sync-token
	 * (if token doesn't isset the method will create new one)
	 *
	 * @param string $type: calls, meetings, tasks
	 * @return int|null - timestamp
	 */
	public static function get_sync_token($type) {
		$davToken = new DavToken('caldav_' . strtolower($type));
		return $davToken->getCurrent();
	}

	/**
	 * Update sync-token with current timestamp value
	 *
	 * @param string $type: calls, meetings, tasks
	 */
	public static function update_sync_token($type) {
		$davToken = new DavToken('caldav_' . strtolower($type));
		$davToken->update();
	}

	/**
	 * @param string|int $token - timestamp
	 *
	 * @return false|string
	 */
	public static function convert_sync_token_to_date($token) {
		$davToken = new DavToken('caldav');
		return $davToken->convertToDate($token);
	}

	/**
	 * Resync events next time or not
	 *
	 * @param bool $val
	 * @param User $user
	 */
	public static function set_resync_value($val, $user) {
		if ($val) {
			$status = "resync";
		} else {
			$status = "synced";
		}
		UserPreference::setPreference(
			"caldav_resync",
			$status,
			0,
			'global',
			$user
		);
		UserPreference::savePreferencesToDB($user);
	}

	/**
	 * Return resync value
	 *
	 * @return bool
	 */
	public static function get_resync_value() {
		return UserPreference::getPreference("caldav_resync");
	}

	private static function import_dates($vcal, $row, array &$data) {
		global $timedate;
		if ($row->getModelName() == 'Task') {
			$ev = $vcal->select('VTODO');
			$ev = current($ev);
		} else {
			$ev = $vcal->select('VEVENT');
			$ev = current($ev);
		}

		if (@$ev->DTSTART instanceof SabreDateTime) {
			if ($ev->DTSTART->getValueType() == 'DATE') {
				$data['is_daylong'] = 1;
				if($row->getModelName() != 'Task') {
					$date_start_gmt = gmdate('Y-m-d H:i:s', $ev->DTSTART->getDateTime()->getTimestamp());
					$date_start_user = $timedate->handle_offset($date_start_gmt, 'Y-m-d 08:00:00', true);
					$date_start_gmt = $timedate->handle_offset($date_start_user, 'Y-m-d H:i:s', false);
					$data['date_start'] = $date_start_gmt;;
				}
			} else {
				$data['is_daylong'] =  0;
				if($row->getModelName() != 'Task')
					$data['date_start'] = gmdate('Y-m-d H:i:s', $ev->DTSTART->getDateTime()->getTimestamp());
				if (isset($ev->DURATION)) {
					$range = DateTimeParser::parseDuration($ev->DURATION, true);

					$seconds = strtotime($ev->DTSTART->getDateTime()->format('Y-m-d H:i:s e') . ' ' . $range) - $ev->DTSTART->getDateTime()->getTimestamp();
					$data['duration'] = (int)($seconds / 60);
				} else if (@$ev->DTEND instanceof SabreDateTime) {
					$diff = $ev->DTEND->getDateTime()->diff($ev->DTSTART->getDateTime());
					$data['duration'] = $diff->format('%a') * 24*60 + $diff->format('%h') * 60  + $diff->format('%i');
				}
			}
		}
		if (@$ev->DUE instanceof SabreDateTime) {
			$data['date_due'] = gmdate('Y-m-d H:i:s', $ev->DUE->getDateTime()->getTimestamp());
		}
		return true;
	}

	private static function import_recurrence($vcal, $row, array &$data) {
		global $timedate;
		if ($row->getModelName() != 'RecurrenceRule')
			return;
		$ev = $vcal->select('VEVENT');
		$ev = current($ev);
		if (empty($ev->RRULE))
			return;
		$rrule = (string)$ev->RRULE;
		$parts = explode(';', $rrule);
		$ruleValues = [];
		$row->set([
			'freq_interval' => 1,
			'freq' => '',
			'until' => '',
		]);
		foreach ($parts as $part) {
			@list($k, $v) = explode('=', $part);
			switch (strtolower($k)) {
			case 'until':
				if (strpos($v, 'T') !== false)
					$until = gmdate('Y-m-d H:i:s', strtotime(str_replace("Z", "", $v) . ' GMT'));
				else {
					$untilDate = gmdate('Y-m-d', strtotime($v . ' GMT'));
					$until = $timedate->handle_offset($untilDate . " 23:59:59", "Y-m-d H:i:s", false);
				}
				$row->set('until',  $until);
				break;
			case 'freq':
				$row->set('freq', $v);
				break;
			case 'interval':
				$row->set('freq_interval',  $v);
				break;
			default:
				$ruleValues[] = "$k=$v";
			}
		}
		$row->set('rule', join(';', $ruleValues));
		return true;
	}

	private static function import_basic($vcal, $row, array &$data) {
		if ( ($model = $row->getModelName()) == 'Task')
			$ev = $vcal->select('VTODO');
		else
			$ev = $vcal->select('VEVENT');
		if (empty($ev)) {
			return false;
		}
		$ev = current($ev);
		$name = (string)$ev->SUMMARY;
		if (!strlen($name))
			$name = translate('LBL_EMPTY_MEETING_TITLE', 'Calendar');
		$data['name'] = $name;
		$data['description'] = (string)$ev->DESCRIPTION;
		if (isset($ev->CLASS))
			$data['is_private'] = (string)$ev->CLASS !== 'PUBLIC' ? 1 : 0;
		if (isset($ev->LOCATION))
			$data['location'] = (string)$ev->LOCATION;
		if (isset($ev->STATUS)) {
			if ($model == 'Task') {
				$status_map = ['In Progress' => 'IN-PROCESS', 'Completed' => 'COMPLETED', 'Pending Input' => 'NEEDS-ACTION', 'Deferred' => 'CANCELLED'];
			} else {
				$status_map = ['Planned' => 'TENTATIVE', 'Held' => 'CONFIRMED', 'Not Held' => 'CANCELLED'];
			}
			$status_map = array_flip($status_map);
			$s = (string)$ev->STATUS;
			if (isset($status_map[$s]))
				$data['status'] = $status_map[$s];
		}
		if (isset($ev->PRIORITY)) {
			$p = (int)(string)$ev->PRIORITY;
			if ($p > 6)
				$data['priority'] = 'P0';
			else if ($p > 3)
				$data['priority'] = 'P1';
			else
				$data['priority'] = 'P2';

		} else {
			$data['priority'] = '';
		}
		return true;
	}

	private static function update_recur($row, $vcal) {
		global $timedate;

		if ($row->getModelName() != 'RecurrenceRule') 
			return;

		$template = @json_decode($row->getField('template'), true);
		if (!is_array($template)) return;
		$allDay = !empty($template['is_allday']);

		$rule = [
			'FREQ=' . $row->getField('freq'),
			'INTERVAL=' . $row->getField('freq_interval')
		];
		$until = $row->getField('until');
		$rule_opts = $row->getField('rule');
		if ($until)
			$rule[] = 'UNTIL=' . self::db_date_to_ical($until, $allDay);
		if ($rule_opts)
			$rule[] = $rule_opts;
		$ev = $vcal->select('VEVENT');
		$ev = current($ev);
		$ev->RRULE = join(';', $rule);;
	}

	private static function update_dates($row, $modelName, $vcal) {
		$now = gmdate('Y-m-d H:i:s');
		if ($modelName == 'Task') {
			$ev = $vcal->select('VTODO');
			$ev = current($ev);
			$start = array_get_default($row, 'date_start');
			$due = array_get_default($row, 'date_due');
			if ($start)
				$ev->DTSTART = self::makeDateProp($vcal, 'DTSTART', $start);
			if ($due) {
				$ev->DUE = self::makeDateProp($vcal, 'DUE', $due);
				unset($ev->DURATION);
			}
			$ev->remove('VALARM');
		} else {
			$ev = $vcal->select('VEVENT');
			$ev = current($ev);
			$allday = array_get_default($row, 'is_daylong');
			$ev->DTSTART = self::makeDateProp($vcal, 'DTSTART', array_get_default($row, 'date_start'), !$allday );
			if (!$allday) {
				$ev->DURATION = 'PT' . ((int)array_get_default($row, 'duration')) . 'M';
				unset($ev->DTEND);
			} else {
				unset($ev->DURATION);
				$ev->DTEND = self::makeDateProp($vcal, 'DTEND', array_get_default($row, 'date_start'), false );
			}
		}
		$ev->DTSTAMP = self::makeDateProp($vcal, 'DTSTAMP', array_get_default($row, 'date_modified', $now));

	}

	private static function makeDateProp($doc, $name, $value, $with_time = true) {
		//$type = $with_time ? \Sabre\VObject\Property\ICalendar\DateTime::UTC : \Sabre\VObject\Property\ICalendar\DateTime::DATE;
		$d = $doc->createProperty($name);
		if ($with_time) {
			$d->setDateTime(new \DateTime($value, new \DateTimeZone('UTC'))/*, $type*/);
		} else {
			global $timedate;
			$value = $timedate->handle_offset($value, 'Y-m-d');
			$dt = new \DateTime($value, new \DateTimeZone('UTC'));
			//$dt->add(new \DateInterval('P1D'));
			$d['VALUE'] = 'DATE';
			$d->setDateTime($dt/*, $type*/);
		}
		return $d;
	}

	private static function update_basic($row, $modelName, $vcal) {
		if ($modelName == 'Task')
			$ev = $vcal->select('VTODO');
		else
			$ev = $vcal->select('VEVENT');
		$ev = current($ev);

		$private = array_get_default($row, 'is_private', self::$EMPTY);
		if ($private !== self::$EMPTY)
			$ev->CLASS = $private ? 'PRIVATE' : 'PUBLIC';

		$loc = array_get_default($row, 'location', self::$EMPTY);
		if ($loc === self::$EMPTY || ! strlen($loc))
			unset($ev->LOCATION);
		else
			$ev->LOCATION = $loc;

		$ev->SUMMARY = array_get_default($row, 'name');
		$ev->DESCRIPTION = array_get_default($row, 'description', '');
		if (!isset($ev->UID))
			$ev->UID = array_get_default($row, 'id');
	}

	private static function update_task($row, $vcal) {
		$ev = $vcal->select('VTODO');
		$ev = current($ev);
		$prio = $row->getFieldValue('priority', self::$EMPTY);
		if ($prio !== self::$EMPTY) {
			switch ($prio) {
			case 'P0':
				$ev->PRIORITY = '9';
				break;
			case 'P1':
				$ev->PRIORITY = '5';
				break;
			case 'P2':
				$ev->PRIORITY = '1';
				break;
			}
		}
		$status = $row->getFieldValue('status', self::$EMPTY);
		if ($status !== self::$EMPTY) {
			$status_map = ['In Progress' => 'IN-PROCESS', 'Completed' => 'COMPLETED', 'Pending Input' => 'NEEDS-ACTION', 'Deferred' => 'CANCELLED'];
			if (isset($status_map[$status]))
				$ev->STATUS = $status_map[$status];
		}
	}

	private static function update_event($row, $vcal) {
		$ev = $vcal->select('VEVENT');
		$ev = current($ev);
		$status = array_get_default($row, 'status', self::$EMPTY);
		if ($status !== self::$EMPTY) {
			$status_map = ['Planned' => 'TENTATIVE', 'Held' => 'CONFIRMED', 'Not Held' => 'CANCELLED'];
			if (isset($status_map[$status]))
				$ev->STATUS = $status_map[$status];
		}
	}

	private static function repair($row, $vcal) {
		global $sugar_version;
		$prod_id = $vcal->select('PRODID');
		if (empty($prod_id))
			$vcal->PRODID = '-//1CRM Corp.//1CRBM ' . $sugar_version . '//EN';
		if ($row->getModelName() == 'Task') {
			$todo = $vcal->select('VTODO');
			if (empty($todo)) {
				$todo = new VTodo($vcal, 'VTODO');
				$todo->UID = $row->getField('id');
				$vcal->add($todo);
			}
		} else {
			$event = $vcal->select('VEVENT');
			if (empty($event)) {
				$event = new VEvent($vcal, 'VEVENT');
				$event->UID = $row->getField('id');
				$vcal->add($event);
			}
		}
	}

	private static function read_vcal($vcal, $createEmpty = true) {
		try {
			return Reader::read($vcal, Reader::OPTION_FORGIVING | Reader::OPTION_IGNORE_INVALID_LINES);
		} catch (Exception $e) {
			if ($createEmpty) return new VCalendar;
			return null;
		}
	}

	// collects VEVENTS and EXDATEs from exception instances and updates
	// RecurrenceRule's VCAL
	public static function rebuild_recur_vcal($recurRuleId) {
		if (!$recurRuleId) return;
		$rule = ListQuery::quick_fetch('RecurrenceRule', $recurRuleId);
		if (!$rule) return;
		$vcal = self::read_vcal($rule->getField('vcal'), false);
		if (!$vcal) return;
		$template = @json_decode($rule->getField('template'), true);
		if (!is_array($template)) return;
		$allday = !empty($template['is_allday']);
		$events = $vcal->select('VEVENT');
		if (empty($events)) return;
		if (empty($events[0]->RRULE)) return;
		$recurEvent = $events[0];
		$vcal->remove('VEVENT');
		$vcal->remove('EXDATE');
		$vcal->add($recurEvent);
		$lq = new ListQuery('Meeting', null, ['filter_deleted' => false]);

		$lq->addSimpleFilter('recurrence_rule_id', $recurRuleId);
		$lq->addSimpleFilter('recurrence_exception', 1);
		$listResult = $lq->runQuery(0, null, false, $order_by = 'recurrence_start_time', 100);
		foreach ($listResult->getIterator() as $m) {
			$mVCal = self::read_vcal($m->getField('vcal'), false);
			if (!$mVCal) continue;
			$events = $mVCal->select('VEVENT');
			if (empty($events[0])) continue;
			$mEvent = $events[0];
			if ($m->getField('deleted')) {
				$recurEvent->add('EXDATE', (string)$mEvent->DTSTART);
			} else {
				$mEvent->remove('RECURRENCE-ID');
				$mEvent->remove('UID');
				$mEvent->add('RECURRENCE-ID', self::db_date_to_ical($m->getField('recurrence_start_time'), $allday));
				$mEvent->add('UID', (string)$recurEvent->UID);
				$vcal->add($mEvent);
			}
		}
		$ruleUpd = RowUpdate::for_result($rule);
		$vcalData = $vcal->serialize();
		$ruleUpd->set('vcal', $vcalData);
		$ruleUpd->set('vcal_size', strlen($vcalData));
		$ruleUpd->save();
		return $vcalData;
	}

	// for current user
	private static function db_date_to_ical($db_date, $allday) {
		global $timedate;
		if (!$allday) return gmdate('Ymd\\THis', strtotime($db_date . ' GMT')) . 'Z';
		return $timedate->handle_offset($db_date, 'Ymd', true);
	}
}
