<?php

require_once 'include/SabreDAV/autoload.php';
require_once 'modules/SocialAccounts/SocialAccount.php';
require_once 'modules/SocialAccounts/SocialUrlParser.php';
require_once 'dav_token.php';

class vCardUtils {

	/**
	 * Update vCard
	 * If $orig_vcard is not specified, it will be taken from $row,
	 * of fetched from database if not present in $row.
	 *
	 * @param RowResult|RowUpdate $row
	 * @param string $orig_vcard
	 *
	 * @return updated vCard text
	 */

	static private $EMPTY;

	public static function vcard_update($row, $orig_vcard = null, $update_rev = false, &$errors = []) {
		self::$EMPTY = new StdClass;
		if (!$orig_vcard) {
			$orig_vcard = $row->getFieldValue('vcard') ?? self::$EMPTY;
			if ($orig_vcard === self::$EMPTY) {
				$id = $row->getFieldValue('id');
				if (!$id) {
					$orig_vcard = '';
				} else {
					$vcard_row = ListQuery::quick_fetch_row($row->getModelName(), $id, array('vcard'));
					if ($vcard_row)
						$orig_vcard = @$vcard_row['vcard'];
					else
						$orig_vcard = '';
				}
			}
		}
		if($orig_vcard) {
			try {
				$vcard = \Sabre\VObject\Reader::read($orig_vcard,  \Sabre\VObject\Reader::OPTION_FORGIVING | \Sabre\VObject\Reader::OPTION_IGNORE_INVALID_LINES);
			} catch (Exception $e) {
				$vcard = new \Sabre\VObject\Component\VCard;
			}
		} else
			$vcard = null;
		if (!($vcard instanceof \Sabre\VObject\Component\VCard))
			$vcard = new \Sabre\VObject\Component\VCard;
		self::update_name($vcard, $row);
		self::update_emails($vcard, $row);
		self::update_url($vcard, $row);
		self::update_addresses($vcard, $row);
		self::update_phones($vcard, $row);
		self::update_skype($vcard, $row);
		self::update_social_accounts($vcard, $row);
		if (!isset($vcard->UID))
			$vcard->UID = $row->getField('id');
		$vcard->VERSION = '3.0';
		if ($update_rev)
			$vcard->REV = gmdate('Y-m-d\\TH:i:s\\Z');

		//Try to Repair
		$vcard->validate(\Sabre\VObject\Node::REPAIR);
		//Check CardDAV errors
		$validate_res = $vcard->validate(\Sabre\VObject\Node::PROFILE_CARDDAV);

		if (sizeof($validate_res) > 0) {
			$name = trim($row->getField('first_name') ." ". $row->getField("last_name"));
			if (empty($name)) $name = $row->getField("id");

			foreach ($validate_res as $error) {
				$errors[$row->getField('id')] = [
					'name' => $name,
					'message' => $error['message']
				];
				$GLOBALS['log']->fatal("Can't create vCard for the Contact {$row->getField('id')}: {$error['message']}");
			}

			return '';
		}
		self::update_sync_token();

		return $vcard->serialize();
	}

	public static function vcard_import($vcard_text, $row) {
		self::$EMPTY = new StdClass;
		try {
			$vcard = \Sabre\VObject\Reader::read($vcard_text,  \Sabre\VObject\Reader::OPTION_FORGIVING | \Sabre\VObject\Reader::OPTION_IGNORE_INVALID_LINES);
		} catch (Exception $e) {
			return false;
		}
		$ret = self::import_name($vcard, $row);
		if (!$ret)
			return false;
		self::import_emails($vcard, $row);
		self::import_url($vcard, $row);
		self::import_phones($vcard, $row);
		self::import_addresses($vcard, $row);
		self::import_skype($vcard, $row);
		self::import_social_accounts($vcard, $row);

		return $ret;
	}

	/**
	 * Returns current CardDAV sync-token
	 * (if token doesn't isset the method will create new one)
	 *
	 * @return int|null - timestamp
	 */
	public static function get_sync_token() {
		$davToken = new DavToken('carddav');
		return $davToken->getCurrent();
	}

	/**
	 * Update sync-token with current timestamp value
	 */
	public static function update_sync_token() {
		$davToken = new DavToken('carddav');
		$davToken->update();
	}

	/**
	 * Resync vCards next time or not
	 *
	 * @param bool $val
	 * @param User $user
	 */
	public static function set_resync_value($val, $user) {
		if ($val) {
			$status = "resync";
		} else {
			$status = "synced";
		}
		UserPreference::setPreference(
			"carddav_resync",
			$status,
			0,
			'global',
			$user
		);
		UserPreference::savePreferencesToDB($user);
	}

	/**
	 * Return resync value
	 *
	 * @return bool
	 */
	public static function get_resync_value() {
		$status = UserPreference::getPreference("carddav_resync");

		return $status === "resync";
	}

	/**
	 * @param string|int $token - timestamp
	 *
	 * @return false|string
	 */
	public static function convert_sync_token_to_date($token) {
		$davToken = new DavToken('carddav');
		return $davToken->convertToDate($token);
	}

	private static function import_skype($vcard, $row) {
		$impp = self::find_impp($vcard, 'skype');
		if ($impp) {
			$skype = explode(':', $impp->getValue(), 2);
			$skype = count($skype) == 2 ? $skype[1] : $skype[0];
			$row->set('skype_id', $skype);
		} else {
			$row->set('skype_id', '');
		}
	}

	private static function update_social_accounts($vcard, $row) {
		if (! ($row instanceof \RowUpdate) )
			return;
		$map = ['youtube' => 'YouTube', 'instagram' => 'Instagram', 'xing' => 'Xing', 'facebook' => 'facebook', 'twitter' => 'twitter', 'linkedin' => 'linkedin'];
		$existing = SocialAccount::getList($row->getModuleDir(), $row->getField('id'));
		$accounts = self::find_social_accounts($vcard);
		foreach ($existing as $type => $details) {
			if(! isset($map[$type]))
				continue;
			$stype = $map[$type];
			if(isset($accounts[$stype])) {
				$prop = $accounts[$stype];
				unset($accounts[$stype]);
				$prop->setValue($details['url']);
			}
			else {
				$prop = $vcard->createProperty('X-SOCIALPROFILE', $details['url']);
				$prop->add('TYPE', $stype);
				$vcard->add($prop);
			}
		}
		foreach ($accounts as $acc) {
			self::remove_property($vcard, $acc);
		}
	}

	private static function import_social_accounts($vcard, $row) {
		$map = ['Google+' => 'google_plus', 'Xing' => 'xing'];
		$accounts = self::find_social_accounts($vcard);
		$existing = SocialAccount::getList($row->getModuleDir(), $row->getField('id'));

		foreach($existing as $type => $details) {
			$rec = ListQuery::quick_fetch('social_accounts', $details['id']);
			if ($rec) {
				$upd = \RowUpdate::for_result($rec);
				$upd->deleteRow();
			}

		}

		foreach ($accounts as $type => $sa) {
			if (!$sa)
				continue;

			$ins_row = \SocialAccount::quick_create_row($sa->getValue(), $row->getModuleDir(), $row->getField('id'));
			if ($ins_row) {
				$ins_row->save();
			}
		}

	}

	private static function update_skype($vcard, $row) {
		$impp = self::find_impp($vcard, 'skype', true);
		$skype = $row->getFieldValue('skype_id') ?? self::$EMPTY;
		if ($skype === self::$EMPTY || !$skype)
			self::remove_property($vcard, $impp);
		else
			$impp->setValue('skype:' . $skype);
	}

	private static function import_name($vcard, $row) {
		$prop = self::find_property($vcard, 'N');
		if (!($prop instanceof \Sabre\VObject\Property\Text)) {
			return false;
		}
		@list($lname, $fname, $addname, $salutation) = $prop->getParts();
		if (!strlen(trim($lname))) {
			return false;
		}
		$row->set([
			'first_name' => $fname,
			'last_name' => $lname,
			'salutation' => self::untranslateSalutation($salutation),
		]);

		$ret = true;
		$org = self::find_property($vcard, 'ORG');

		if ($org instanceof \Sabre\VObject\Property\Text) {

			@list($org_name, $unit1, $unit2) = $org->getParts();

			if (strlen($org_name) && UserPreference::getPreference('carddav_create_accounts')) {

				$account = ListQuery::quick_fetch_key('Account', 'name', $org_name);

				if (! $account) {
					$account = RowUpdate::blank_for_model('Account');
					$aid = create_guid();
					$account->set(array(
						'name' => $org_name,
						'id' => $aid,
						'assigned_user_id' => AppConfig::current_user_id(),
					));
				}

				$ret = $account;
			}

			$row->set('department', strlen($unit1) ? $unit1 : $unit2);
		}

		self::import_direct($vcard, $row, 'TITLE', 'title');
		self::import_direct($vcard, $row, 'NOTE', 'description');
		return $ret;
	}

	private static function update_direct($vcard, $row, $name, $fname) {
		$prop = self::find_property($vcard, $name);
		$value = $row->getFieldValue($fname) ?? self::$EMPTY;
		if ($value !== self::$EMPTY) {
			if (strlen($value ?? '')) {
				if (!$prop) {
					$prop = $vcard->createProperty($name, '');
				}
				$prop->setValue($value);
				$vcard->$name = $prop;
			} else {
				if ($prop)
					self::remove_property($vcard, $prop);
			}
		}
	}

	private static function import_direct($vcard, $row, $name, $fname) {
		$prop = self::find_property($vcard, $name);
		if ($prop)
			$row->set($fname, $prop->getValue());
		else
			$row->set($fname, '');
	}

	private static function find_url($vcard) {
		$url = null;
		foreach ($vcard->select('URL') as $u) {
			if (self::has_param($u, 'TYPE', 'PREF')) {
				return '' . $u;
			}
			if (!$url)
				$url = '' . $u;
		}
		return $url;
	}

	private static function find_emails($vcard) {
		$email1 = $email2 = null;
		foreach ($vcard->select('EMAIL') as $e) {
			if (self::has_param($e, 'TYPE', 'PREF')) {
				if ($email1 && !$email2) {
					$email2 = $email1;
				} else {
					$email1 = $e;
				}
			} else {
				if (!$email1)
					$email1 = $e;
				elseif (!$email2)
					$email2 = $e;
			}
		}
		return [$email1, $email2];
	}

	private static function find_phones($vcard) {
		$work = $home = $mobile = $fax = $other = null;
		foreach ($vcard->select('TEL') as $t) {
			if (self::has_param($t, 'TYPE', 'CELL') && !$mobile) {
				$mobile = $t;
			} else if (self::has_param($t, 'TYPE', 'FAX') && !$fax) {
				$fax = $t;
			} else if (self::has_param($t, 'TYPE', 'HOME') && !$home) {
				$home = $t;
			} else if (self::has_param($t, 'TYPE', 'WORK') && !$work) {
				$work = $t;
			} else if (!$other) {
				$other = $t;
			}
		}
		return [$work, $home, $mobile, $fax, $other];
	}

	private static function import_phones($vcard, $row) {
		list ($work, $home, $mobile, $fax, $other) = self::find_phones($vcard);
		foreach (['work', 'home', 'mobile', 'fax', 'other'] as $v) {
			if (${$v}) {
				$row->set('phone_' . $v, ${$v}->getValue());
			} else {
				$row->set('phone_' . $v, '');
			}
		}
	}

	private static function update_phones($vcard, $row) {
		list ($work, $home, $mobile, $fax, $other) = self::find_phones($vcard);
		$types = ['WORK', 'HOME', 'CELL', 'FAX', 'OTHER'];
		foreach (['work', 'home', 'mobile', 'fax', 'other'] as $i => $v) {
			$value = $row->getFieldValue('phone_' . $v, self::$EMPTY);
			if ($value !== self::$EMPTY) {
				if (strlen($value ?? '')) {
					if (!${$v}) {
						${$v} = $vcard->createProperty('TEL', '');
						$vcard->add('TEL', ${$v}, array('TYPE' => $types[$i]));
					}
					${$v}->setValue($value);
				} else {
					if (${$v})
						self::remove_property($vcard, ${$v});
				}
			}
		}
	}

	private static function import_url($vcard, $row) {
		$url = self::find_url($vcard);
		if ($url) {
			$row->set('website', $url);
		}
	}

	private static function import_emails($vcard, $row) {
		list($email1, $email2) = self::find_emails($vcard);
		if ($email1) {
			$row->set('email1', $email1->getValue());
		} else {
			$row->set('email1', '');
		}
		if ($email2) {
			$row->set('email2', $email2->getValue());
		} else {
			$row->set('email2', '');
		}
	}

	private static function import_addresses($vcard, $row) {
		list($primary, $alt) = self::find_addresses($vcard);
		self::import_address($vcard, $row, $primary, true);
		self::import_address($vcard, $row, $alt, false);
	}

	private static function import_address($vcard, $row, $addr, $primary) {
		$map = array('', '', 'street', 'city', 'state', 'postalcode', 'country');
		$prefix = $primary ? 'primary_address_' : 'alt_address_';
		if ($addr)
			$parts = $addr->getParts();
		foreach ($map as $idx => $f) {
			if (!$f)
				continue;
			if ($addr)
				$val = $parts[$idx];
			else
				$val = '';
			$row->set($prefix . $f, $val);
		}
	}

	private static function update_name($vcard, $row) {
		$prop = self::find_property($vcard, 'N');
		if (!($prop instanceof \Sabre\VObject\Property\FlatText) && !($prop instanceof \Sabre\VObject\Property\Text)) {
			self::remove_property($vcard, $prop);
			$prop = null;
		}
		if (!$prop) {
			$prop = $vcard->createProperty('N', '');
		}
		@list($lname, $fname, , $salutation) = $prop->getParts();
		$val = $row->getFieldValue('last_name', self::$EMPTY);
		if ($val !== self::$EMPTY) $lname = $val ?? '';
		$val = $row->getFieldValue('first_name', self::$EMPTY);
		if ($val !== self::$EMPTY) $fname = $val ?? '';
		$val = $row->getFieldValue('salutation', self::$EMPTY);
		if ($val !== self::$EMPTY) $salutation = $val ?? '';
		$prop->setParts([
				$lname ?? '',
				$fname ?? '',
				'',
				self::translateSalutation($salutation ?? ''),
				'',
				]);
		$vcard->N = $prop;

		$prop = self::find_property($vcard, 'FN');
		if (!$prop) {
			$prop = $vcard->createProperty('FN', '');
		}
		$prop->setValue($fname . ' ' . $lname);
		$vcard->FN = $prop;

		$org = self::find_property($vcard, 'ORG');
		$department = $row->getFieldValue('department', self::$EMPTY);
		$account = $row->getFieldValue('primary_account_id', self::$EMPTY);
		if ($department !== self::$EMPTY || $account !== self::$EMPTY) {
			if (!($org instanceof \Sabre\VObject\Property\Text)) {
				$org = $vcard->createProperty('ORG', '');
			}
			$parts = $org->getParts();
			$remove = true;
			if ($department !== self::$EMPTY &&  strlen($department ?? '')) {
				$remove = false;
				$parts[1] = $department ?? '';
			} else if ($department !== self::$EMPTY) {
				$parts[1] = '';
			} else {
				$remove = false;
			}
			if ($account !== self::$EMPTY &&  strlen($account ?? '')) {
				$acc_row = \ListQuery::quick_fetch_row('Account', $account);
				if ($acc_row)
					$account = $acc_row['name'];
			}
			if ($account !== self::$EMPTY &&  strlen($account ?? '')) {
				$remove = false;
				$parts[0] = $account ?? '';
			} else if ($account !== self::$EMPTY) {
				$parts[0] = '';
			} else {
				$remove = false;
			}
			if (!$remove) {
				$org->setParts([$parts[0] ?? '', @$parts[1] ?? '']);
				$vcard->ORG  = $org;
			} else {
				self::remove_property($vcard, $org);
			}
		}
		self::update_direct($vcard, $row, 'TITLE', 'title');
		self::update_direct($vcard, $row, 'NOTE', 'description');
	}

	private static function update_addresses($vcard, $row) {
		list ($primary, $alt) = self::find_addresses($vcard);
		$primary = self::update_address($vcard, $row, $primary, true);
		$alt = self::update_address($vcard, $row, $alt, false);
	}

	private static function translateSalutation($sal) {
		$lang = AppConfig::setting('locale.defaults.language');
		$strings = AppConfig::setting("lang.lists.$lang.app.salutation_dom");
		if (isset($strings[$sal]))
			return $strings[$sal];
		return $sal;
	}
	private static function untranslateSalutation($sal) {
		$lang = AppConfig::setting('locale.defaults.language');
		$strings = AppConfig::setting("lang.lists.$lang.app.salutation_dom");
		$idx = array_search($sal, $strings);
		if ($idx !== false)
			return $idx;
		return $sal;
	}

	private static function update_address($vcard, $row, $prop, $primary) {
		$prefix = $primary ? 'primary_address' : 'alt_address';
		$values = [];
		if ($prop)
			$values = $prop->getParts();
		$map = array('', '', 'street', 'city', 'state', 'postalcode', 'country');
		$set = false;

		foreach ($map as $idx => $f) {
			$value = $row->getFieldValue($prefix . '_' . $f, self::$EMPTY);
			if ($value !== self::$EMPTY)
				$values[$idx] = $value ?? '';
			if (!array_key_exists($idx, $values))
				$values[$idx] = '';
			if (strlen($values[$idx]))
				$set = true;
		}
		ksort($values, SORT_NUMERIC);
		if (!$set) {
			if ($prop)
				self::remove_property($vcard, $prop);
			return null;
		}
		if($prop && self::has_param($prop, 'WORK')) { // cleanup
			self::remove_property($vcard, $prop);
			$prop = null;
		}
		if (!$prop) {
			$prop = $vcard->createProperty('ADR', '');
			$prop->add('TYPE', 'WORK');
			if ($primary)
				$prop->add('TYPE', 'PREF');
			$vcard->add($prop);
		}
		$prop->setParts($values);
		return $prop;
	}

	private static function remove_property($vcard, $prop) {
		if ($prop) {
			if (is_object($prop) && ! property_exists($prop, 'name'))
				return false;

			if (is_object($prop) && ! $vcard->select($prop->name))
				return false;

			$vcard->remove($prop);
		}

		return true;
	}

	private static function find_addresses($vcard) {
		$primary = $alt = null;
		$props = $vcard->select('ADR');
		foreach ($props as $prop) {
			if (!($prop instanceof \Sabre\VObject\Property\Text))
				continue;
			if (self::has_param($prop, 'TYPE', 'PREF')) {
				if ($primary && !$alt) {
					$alt = $primary;
					$primary = $prop;
				} else {
					$primary = $prop;
				}
			} else {
				if (!$primary)
					$primary = $prop;
				else if (!$alt)
					$alt = $prop;
			}
		}
		return [$primary, $alt];
	}

	private static function update_url($vcard, $row) {
		$props = $vcard->select('URL');
		// clean up invalid url properties
		foreach($props as $p) {
			if(self::has_param($p, 'PREF'))
				self::remove_property($vcard, $p);
		}
		$prop = self::find_property($vcard, 'URL', ['TYPE' => 'PREF']);
		if (!$prop) {
			$prop = $vcard->createProperty('URL', $row->getFieldValue('website') ?? '');
			$prop->add('TYPE', 'PREF');
			$vcard->add($prop);
		}
		$val = $row->getFieldValue('website', self::$EMPTY);
		if ($val !== self::$EMPTY) {
			if (strlen($val ?? ''))
				$prop->setValue($val);
			else
				self::remove_property($vcard, $prop);
		}
	}

	private static function update_emails($vcard, $row) {
		$props = $vcard->select('EMAIL');
		// clean up invalid email properties
		foreach($props as $p) {
			if(self::has_param($p, 'INTERNET'))
				self::remove_property($vcard, $p);
		}
		$prop = self::find_property($vcard, 'EMAIL', ['TYPE' => 'WORK']);
		if (!$prop) {
			$prop = $vcard->createProperty('EMAIL', $row->getFieldValue('email1'));
			//$prop->add('TYPE', 'INTERNET');
			$prop->add('TYPE', 'WORK');
			$vcard->add($prop);
		}
		$val = $row->getFieldValue('email1', self::$EMPTY);

		if ($val !== self::$EMPTY) {
			if (strlen($val ?? ''))
				$prop->setValue($val);
			else
				self::remove_property($vcard, $prop);
		}

		$prop = self::find_property($vcard, 'EMAIL', ['TYPE' => 'HOME']);
		if (!$prop) {
			$prop = $vcard->createProperty('EMAIL', $row->getFieldValue('email2'));
			//$prop->add('TYPE', 'INTERNET');
			$prop->add('TYPE', 'HOME');
			$vcard->add($prop);
		}
		$val = $row->getFieldValue('email2', self::$EMPTY);
		if ($val !== self::$EMPTY) {
			if (strlen($val ?? ''))
				$prop->setValue($val);
			else
				self::remove_property($vcard, $prop);
		}
	}

	static private function find_property($vcard, $name, $params = array(), $case = false) {
		$props = $vcard->select($name);
		if (!$props)
			return null;
		foreach ($props as $prop) {
			if (!$params)
				return $prop;
			foreach ($params as $pn => $pv) {
				if (!self::has_param($prop, $pn, $pv, $case))
					continue 2;
			}
			return $prop;
		}
		return null;
	}

	static private function has_param($elt, $name, $value = null, $case = false) {
		foreach ($elt->parameters as $prop) {
			if ($prop->name != strtoupper($name))
				continue;
			if (is_null($value))
				return true;
			$val = $prop->getValue();
			if (!$case) {
				$val = strtoupper($val);
				$value = strtoupper($value);
			}
			$vals = explode(',', $val);
			foreach ($vals as $val)
				if ($val === $value)
					return true;
		}
		return false;
	}

	static private function find_impp($vcard, $type, $create = false) {
		$props = $vcard->select('IMPP');
		if ($props) {
			$type = strtoupper($type);
			foreach ($props as $prop) {
				if(self::has_param($prop, 'WORK')) { // cleanup
					self::remove_property($vcard, $prop);
					continue;
				}
				$parts = explode(':', $prop->getValue(), 2);
				if (count($parts) == 2) {
					if (strtoupper($parts[0]) == $type)
						return $prop;
				}
				if (self::has_param($prop, 'X-SERVICE-TYPE', $type))
					return $prop;
			}
		}
		if (!$create)
			return null;
		$prop = $vcard->createProperty('IMPP', '');
		$prop->add('TYPE', 'WORK');
		$prop->add('X-SERVICE-TYPE', $type);
		$vcard->add($prop);
		return $prop;
	}

	private static function find_social_accounts($vcard) {
		$types = ['twitter', 'facebook', 'google+' => 'Google+', 'linkedin', 'xing' => 'Xing', 'plus.google.com', 'xing.com'];
		$ret = [];
		$props = $vcard->select('X-SOCIALPROFILE');
		foreach($props as $prop) {
			foreach($prop->parameters as $p) {
				if(strtoupper($p->name) == 'TYPE') {
					$val = explode(',', $prop->getValue());
					foreach($val as $v) {
						$lv = strtolower($v);
						if(isset($types[$lv]))
							$v = $types[$lv];
						if(isset($ret[$v]))
							$ret[] = $prop;
						else
							$ret[$v] = $prop;
						break;
					}
					break;
				}
			}
		}
		return $ret;
	}
}
