<?php

class PhoneNumbersNormalizer {
	static $noise_pattern = '~([^a-z0-9]+)~i';
	public static $all_rules = array(
		'UK' => array(
			'country' => 'United Kingdom',
			'patterns' => array(
				array(
					'pattern' => '00(\d+)',
					'replacement' => '\1',
					'name' => 'International numbers starting with 00',
					'description' => 'Initial 2 zeros will be removed, the rest of the number must start with country code',
				),
				array(
					'pattern' => '0(\d+)',
					'replacement' => '44\1',
					'name' => 'UK numbers starting with 0',
					'description' => 'Initial zero will be replaced by 44 - UK international country code',
				),
			),
		),
		'US' => array(
			'country' => 'United States/Canada',
			'patterns' => array(
				array(
					'pattern' => '011(\d+)',
					'replacement' => '\1',
					'name' => 'International numbers starting with 011',
					'description' => 'Initial 011 will be removed, the rest of the number must start with country code',
				),
				array(
					'pattern' => '\d{10}',
					'replacement' => '1\0',
					'name' => 'US numbers consisting of 10 digits',
					'description' => '1 will be added in front of the number, indicating US international country code',
				),
			),
		),
		'FR' => array(
			'country' => 'France',
			'patterns' => array(
				array(
					'pattern' => '00(\d+)',
					'replacement' => '\1',
					'name' => 'International numbers starting with 00',
					'description' => 'Initial 2 zeros will be removed, the rest of the number must start with country code',
				),
				array(
					'pattern' => '0(\d+)',
					'replacement' => '33\1',
					'name' => 'FR numbers starting with 0',
					'description' => 'Initial zero will be replaced by 33 - FR international country code',
				),
				array(
					'pattern' => '33(0)(\d+)',
					'replacement' => '33\2',
					'name' => 'FR numbers starting with 33 (0)',
					'description' => 'Zero in brackets will be removed ("+" symbol also will be removed)',
				),
			),
		),
	);

	private $group;
	private $rules;
	private $bad_patterns;
	private $rules_matched;

	public function __construct($rules_id) {
		$this->reset();
		$this->rules = isset(self::$all_rules[$rules_id]) ? self::$all_rules[$rules_id] : array();
	}

	public function getRules() {
		return $this->rules;
	}

	public function reset() {
		$this->group = 'global';
		$this->bad_patterns = array();
		$this->rules_matched = array();
	}

	public function startGroup($g) {
		$this->group = $g;
	}

	public function getBadPatterns() {
		return $this->bad_patterns;
	}

	public function getRulesMatched() {
		return $this->rules_matched;
	}

	public function nines($d) { return str_repeat('9', strlen($d[0])); }

	public function normalize($number, &$changed) {
		$changed = false;
		$orig = trim($number);
		$number = preg_replace(self::$noise_pattern, '', strtolower($orig));
		$modified = null;
		foreach ($this->rules['patterns'] as $ridx => $r) {
			$pattern = '~^' . $r['pattern'] . '$~';
			if (preg_match($pattern, $number)) {
				$modified = preg_replace($pattern, $r['replacement'], $number);
				$changed = true;
				@$this->rules_matched[$ridx][$this->group]++;
				break;	
			}
		}
		if (!$modified && !preg_match('~^\d+$~', $number)) {
			$pattern= preg_replace_callback('~\d+~', array(get_class($this), 'nines'), $number);
			$pattern= preg_replace('~[^\d]+~', ' \0 ', $pattern);
			@$this->bad_patterns[$pattern][$this->group]++;
			$modified = $orig;
		} elseif (!$modified) {
			$changed = $orig !== $number;
			$modified = $number;
		}
		return $modified;
	}
}

global $mod_strings;

$def_country = 'UK';
$addr = ListQuery::quick_fetch_key('CompanyAddress', 'main', 1);
if($addr && $addr_country = $addr->getField('address_country')) {
	if(in_array($addr_country, array('CA', 'Canada', 'US', 'USA', 'United States')))
		$def_country = 'US';
}
$country = array_get_default($_REQUEST, 'country', $def_country);
$do_run = array_get_default($_REQUEST, 'do_run');
$test_run = array_get_default($_REQUEST, 'test_run', $do_run ? 0 : 1) ? 'checked="checked"' : '';

$norm = new PhoneNumbersNormalizer($country);

echo get_module_title('Repair', $mod_strings['LBL_NORMALIZE_PHONES'], true);

echo '<div style="font-size: 120%; max-width: 700px;">';
echo $mod_strings['LBL_NORMALIZE_PHONES_DESC_LONG'];

$opts = array();
foreach (PhoneNumbersNormalizer::$all_rules as $k => $v) {
	$opts[$k] = $v['country'];
}
$options = get_select_options_with_id($opts, $country);

$rules = $norm->getRules();
$modules = array('Contacts' => 'Contact', 'Leads' => 'Lead', 'Accounts' => 'Account');

if ($do_run) {
	$phone_fields = array('phone_office', 'phone_alternate', 'phone_work', 'phone_home', 'phone_mobile', 'phone_other');
	foreach ($modules as $module => $model) {
		$norm->startGroup($module);
		$lq = new ListQuery($model);
		$lq->addFields($phone_fields);
		$res = $lq->runQuery(0, null, false, null, 50);
		while (true) {
			foreach ($res->rows as $row_idx => $row) {
				$updates = array();
				foreach ($phone_fields as $f) {
					if (isset($row[$f]) && strlen($row[$f])) {
						$changed = false;
						$modified = $norm->normalize($row[$f], $changed);
						if ($changed && !$test_run) {
							$updates[$f] = $modified;
						}
					}
				}
				if (!empty($updates)) {
					$row_result = $res->getRowResult($row_idx);
					$upd = RowUpdate::for_result($row_result);
					$upd->update_date_modified = false;
					$upd->prohibit_workflow = true;
					$upd->set($updates);
					$upd->save();
				}
			}
			if ($res->page_finished)
				break;
			$lq->pageResult($res);
		}
	}
}

$csrf_input = create_csrf_input();
echo <<<EOF
<form method="POST" action="index.php">
	{$csrf_input}
	<input type="hidden" name="module" value="Administration">
	<input type="hidden" name="action" value="NormalizePhoneNumbers">
	<table class="tabForm" cellpadding="0" cellspacing="0" width="100%">
		<tr>
			<td class="dataLabel">
				{$mod_strings['LBL_NORMALIZE_PHONES_COUNTRY']}
			</td>
			<td class="dataLabel">
				<select name="country" onchange="this.form.submit();">
					{$options}
				</select>
			</th>
		</tr>
		<tr>
			<th class="dataLabel" colspan="2">
				<h4>{$mod_strings['LBL_NORMALIZE_PHONES_RULES_LIST']}</h4>
			</th>
		</tr>
		<tr>
			<td class="dataLabel" colspan="2">
				<dl>
EOF;
foreach ($rules['patterns'] as $r) {
	echo '<dt>' . $r['name'] . '</dt><dd>' . $r['description'] . '</dd>';
}
echo <<<EOF
				</dl>
			</td>
		</tr>
		<tr>
			<td class="dataLabel" colspan="2">
				<input type="checkbox" name="test_run" value="1" {$test_run} > {$mod_strings['LBL_NORMALIZE_PHONES_TEST']}
			</td>
		</tr>
		<tr>
			<td class="dataLabel" colspan="2">
				<input class="input-button input-outer" type="submit" name="do_run" value="{$mod_strings['LBL_NORMALIZE_PHONES_RUN']}">
			</td>
		</tr>
EOF;
if ($do_run) {
	ini_set( "memory_limit", "-1" );
	ini_set( "max_execution_time", "0" );	
	$matched = $norm->getRulesMatched();
	$bad = $norm->getBadPatterns();
	$label = $test_run ? $mod_strings['LBL_NORMALIZE_PHONES_RESULTS_TEST'] : $mod_strings['LBL_NORMALIZE_PHONES_RESULTS'];
	echo <<<EOF
		<tr>
			<th class="dataLabel" colspan="2">
				<h2 style="text-align:center"><b>{$label}</b></h2>
			</td>
		</tr>
		<tr>
			<th class="dataLabel" colspan="2" style="text-align:center">
				<h4>{$mod_strings['LBL_NORMALIZE_PHONES_RULES_MATCHED']}</h4>
			</th>
		</tr>
		<tr>
			<td class="dataLabel" colspan="2">
				<table cellpadding="5" cellspacing="1" border="1" width="100%">
					<tr><td>Rule</td>
EOF;
	foreach ($modules as $k => $v)
		echo '<td>' . $k . '</td>';
	echo '</tr>';
	foreach ($rules['patterns'] as $pidx => $p) {
		echo '<tr><td>' . $p['name'] . '</td>';
		foreach ($modules as $k => $v) {
			$val = array_get_default($matched, $pidx, array());
			echo '<td>' . array_get_default($val, $k, '-') . '</td>';
		}
		echo '</tr>';
	}
	echo <<<EOF
				</table>
			</td>
		</tr>
		<tr>
			<th class="dataLabel" colspan="2" style="text-align:center">
				<h4>{$mod_strings['LBL_NORMALIZE_PHONES_BAD_PATTERN']}</h4>
			</th>
		</tr>
		<tr>
			<td class="dataLabel" colspan="2">
				<table cellpadding="5" cellspacing="1" border="1" width="100%">
					<tr><td>Pattern</td>
EOF;
	foreach ($modules as $k => $v)
		echo '<td>' . $k . '</td>';
	echo '</tr>';
	foreach ($bad as $p => $m) {
		echo '<tr><td>' . $p . '</td>';
		foreach ($modules as $k => $v)
			echo '<td>' . array_get_default($m, $k, '-') . '</td>';
		echo '</tr>';
	}
	echo <<<EOF
				</table>
			</td>
		</tr>
EOF;
}
echo <<<EOF
	</table>
</form>
EOF;
echo '</div>';

