<?php

abstract class ZipHelper {

	protected $cwd;

	protected function __construct() {}

	public static function create($encrypt = false) {
		return self::can_use_zip() ? new ZipHelperCmd($encrypt) : new ZipHelperPhp();
	}

	public static function can_use_zip() {
		$descriptorspec = [
			0 => ["pipe", "r"],
			1 => ["pipe", "w"],
			2 => ["pipe", "w"]
		];

		$process = proc_open('zip -v', $descriptorspec, $pipes);
		if (is_resource($process)) {
			fclose($pipes[0]);

			stream_get_contents($pipes[1]);
			fclose($pipes[1]);

			stream_get_contents($pipes[2]);
			fclose($pipes[2]);

			$return_value = proc_close($process);
			return !$return_value;
		}
		return false;
	}

	/*
	 *
	 *	XXX $zipFile is supposed to be an absolute path!!
	 * 
	 */
	abstract public function open($zipFile);
	
	abstract public function close();
	
	abstract public function removeFile($zipPath);

	public function setCwd($path) {
		$this->cwd = $path;
	}

	abstract public function addFile($path, $localName = null);

	public function addDir($rootPath, $structureOnly, $exclude = null) {
		$exceptions = array(".", "..", ".svn", ".git", ".gitattributes", ".gitignore", ".idea");
		if (!is_array($exclude)) $exclude = [];
		if(! strlen($rootPath)) $rootPath = '.';
		$stack = array();
		$prefix = $rootPath == '.' ? '' : $rootPath . '/';
		$h = opendir($rootPath);
		
		while($h !== false) {
			while ($file = readdir($h)) {
				if (in_array($file, $exceptions)) {
					continue;
				}
				$path = $prefix . $file;
				if (in_array($path, $exclude)) {
					continue;
				}
				if(is_dir($path)) {
					if ($structureOnly) {
						$indexFileName = $path . '/index.html';
						if (file_exists($indexFileName) || touch($indexFileName)) {
							$this->addFile($indexFileName);
						}
					} 
					$stack[] = array($h, $prefix);
					$h = opendir($path);
					$prefix .= $file . '/';
					continue;
				} elseif (!$structureOnly) {
					$this->addFile($path);
				}
			}
			closedir($h);
			if($stack) {
				list($h, $prefix) = array_pop($stack);
			} else
				$h = false;
		}
	}
	
	public function flush() {
		return true;
	}

}

class ZipHelperCmd extends ZipHelper {
	private $zip_process;
	private $encrypt;
	private $ossl_process;

	private $zip_input_stream;
	private $ossl_input_stream;
	private $file_output_stream;
	
	protected function __construct($encrypt = false) {
		$this->encrypt = $encrypt;
	}

	public function open($zipFile) {

		if ($this->encrypt) {
			$encFileName = $zipFile . '.enc';
			$cert = AppConfig::setting('backup.crypto_cert');
			if (!$cert) {
				return false;
			}
			$cert_fname = tempnam ( sys_get_temp_dir(), "crt" );
			register_shutdown_function(function() use ($cert_fname) {
				unlink($cert_fname);
			});
			if (!file_put_contents($cert_fname, $cert)) {
				return false;
			}
			$encFile = fopen($encFileName, 'w');
			if (!$encFile) return false;
			$this->file_output_stream = $encFile;
			$ossl_desc = [
				0 => ['pipe', 'r'],
				1 => $encFile,
				2 => ['file', 'php://temp', 'w'],
			];
			$process = proc_open(
				'openssl smime -binary -outform DER -encrypt -aes-256-cbc ' . escapeshellarg($cert_fname), 
				$ossl_desc, $ossl_pipes, $this->cwd
			);
			if (!is_resource($process)) {
				unlink($encFileName);
				return false;
			}
			$this->ossl_process = $process;
			$this->ossl_input_stream = $ossl_pipes[0];
		}

		if (!$this->ossl_process) {
			$this->file_output_stream = fopen($zipFile, 'w');
			if (!$this->file_output_stream) return false;
		}
		$descriptorspec = [
			0 => ["pipe", "r"],
			1 => $this->ossl_process ? $this->ossl_input_stream :  $this->file_output_stream,
			2 => ['file', 'php://temp', 'w'],
		];

		$process = proc_open('zip -@ -r - ', $descriptorspec, $pipes, $this->cwd);
		if (is_resource($process)) {
			$this->zip_process = $process;
			$this->zip_input_stream = $pipes[0];
			return true;
		}
		$this->closeOsslProcess();
		return false;
	}
	
	public function close() {
		$ret = $this->closeZipProcess();
		$ret = $this->closeOsslProcess() && $ret;
		if ($this->file_output_stream) {
			$ret = fclose($this->file_output_stream) && $ret;
			$this->file_output_stream = null;
		}
		return $ret;
	}

	private function closeZipProcess() {
		if (!$this->zip_process) return true;
		fclose($this->zip_input_stream);
		$retcode = proc_close($this->zip_process);
		$this->zip_process = $this->zip_input_stream = null;
		return !$retcode;
	}

	private function closeOsslProcess() {
		if (!$this->ossl_process) return true;
		$retcode = proc_close($this->ossl_process);
		$this->ossl_process = $this->ossl_input_stream = null;
		return !$retcode;
	}

	public function addFile($path, $localName = null) {
		return fwrite($this->zip_input_stream, $path . "\n") !== false;
	}

	public function removeFile($zipPath) {
		if ($this->encrypt) $zipPath .= '.enc';
		return unlink($zipPath);
	}
}

class ZipHelperPhp extends ZipHelper {
	private $zip;
	protected $zipFile;

	public function open($zipFile) {
		$this->zipFile = $zipFile;
    	$zip = new ZipArchive();
    	if (!$zip->open($zipFile, ZipArchive::CREATE)) {
			return false;
		}
		$this->zip = $zip;
		return true;
	}
	
	public function addFile($path, $localName = null) {
		if (!$this->zip) return false;
		return $this->zip->addFile($path, $localName);
	}

	public function close() {
		if (!$this->zip) return false;
		$result = $this->zip->close();
		$this->zip = null;
		return $result;
	}
	
	public function flush() {
		if (!$this->close()) return false;
		return $this->open($this->zipFile);
	}

	public function removeFile($zipPath) {
		return unlink($zipPath);
	}
}

