<?php

class FormCaptcha {

	const SETTING_INDEX = 'web2lead_captcha';

	const API_URL = 'https://www.google.com/recaptcha/api.js';

	const VERIFICATION_URL = 'https://www.google.com/recaptcha/api/siteverify';

	/**
	 * @return string
	 */
	public function getApiUrl() {
		return self::API_URL .'?render='. $this->getKey();
	}

	/**
	 * @return string
	 */
	public function getKey() {
		return AppConfig::setting(self::SETTING_INDEX . '.key');
	}

	/**
	 * @return string
	 */
	public function getSecret() {
		return AppConfig::setting(self::SETTING_INDEX . '.secret');
	}

	/**
	 * Is CAPTCHA valid or not
	 *
	 * @param string $token
	 * @param string $action
	 *
	 * @return bool
	 */
	public function isValid($token, $action) {
		$result = $this->requestVerification($token);

		if (is_array($result) && $result['success'] && $result['score'] > 0 && $result['action'] == $action) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Refresh CAPTCHA key and secret with data from input
	 *
	 * @param array $input
	 */
	public function refreshKeys($input) {
		$captcha_key = $this->getKey();
		$captcha_secret = $this->getSecret();

		$input_key = $input['captcha_key'] ?? '';
		$input_secret = $input['captcha_secret'] ?? '';
		$update = false;

		if ($input_key != $captcha_key) {
			AppConfig::set_local(self::SETTING_INDEX . '.key', $input_key);
			$update = true;
		}

		if ($input_secret != $captcha_secret) {
			AppConfig::set_local(self::SETTING_INDEX . '.secret', $input_secret);
			$update = true;
		}

		if ($update)
			AppConfig::save_local();
	}

	/**
	 * Get CAPTCHA page code
	 *
	 * @return string
	 */
	public function getPageCode() {
		if (empty($this->getKey()) && empty($this->getSecret()))
			return '';

		$result = "<script src='".$this->getApiUrl()."'></script>
            <input type=\"hidden\" name=\"token\" id=\"token\" value=\"\">
            <input type=\"hidden\" name=\"c_action\" id=\"c_action\" value=\"1crm_web2lead\">
			<script>
			    var captchaSubmit = function() {
			        grecaptcha.ready(function() {
	                    grecaptcha.execute('" . $this->getKey() . "', {action: '1crm_web2lead'}).then(function(token) {
	                        document.getElementById('token').value = token;
	                        document.WebToLeadForm.submit();
	                    });
                    });
				};
			</script>";

		return $result;
	}

	/**
	 * Get CAPTCHA page code
	 *
	 * @return string
	 */
	public function getJqueryPageCode() {
		if (empty($this->getKey()) && empty($this->getSecret()))
			return '';

		$result = "<script src='".$this->getApiUrl()."'></script>
			<script>
				jQuery(document).ready(function() {
					jQuery('.ocrmlf-form').on('submit', function(e) {
						var frm = this;
						e.preventDefault();
				        grecaptcha.ready(function() {
		                    grecaptcha.execute('" . $this->getKey() . "', {action: '1crm_web2lead'}).then(function(token) {
		                        jQuery('#token').val(token);
		                        frm.submit();
		                    });
	                    });
					});    
			    });
			</script>";

		return $result;
	}

	/**
	 * @return array
	 */
	public function getFormFields() {
		return [
			'token' => '',
			'c_action' => '1crm_web2lead',
			'captcha_script' => $this->getJqueryPageCode()
		];
	}

	/**
	 * Get CAPTCHA key and secret rows HTML
	 *
	 * @param HtmlFormGenerator $gen
	 * @param array $mod_strings
	 *
	 * @return string
	 */
	public function getSettingsHtml(HtmlFormGenerator $gen, $mod_strings) {
		$captcha_key = $this->getKey();
		$captcha_secret = $this->getSecret();

		$spec = array('name' => 'add_captcha', 'onchange' => 'addCaptcha();');
		$add_captcha_check = $gen->form_obj->renderCheck($spec, 0);

		$html = <<<EOQ
			<div class="row form-row">
				<div class="column form-cell sm-12 vertical">
					<div class="form-label">{$mod_strings['LBL_CAPTCHA']}</div>
					<div class="form-entry">
                        {$add_captcha_check}&nbsp;{$mod_strings['LBL_ADD_CAPTCHA']}
					</div>
				</div>
			</div>		
			<div class="row form-row" id="captcha_description_container" style="display: none;">
				<div class="column form-cell sm-6 vertical">
					<div class="form-entry">
					{$mod_strings['LBL_CAPTCHA_DESC1']} <a href="http://www.google.com/recaptcha/admin" target="_blank">{$mod_strings['LBL_CAPTCHA_DESC_URL']}</a>.
					{$mod_strings['LBL_CAPTCHA_DESC2']}
					{$mod_strings['LBL_CAPTCHA_DESC3']} <a href="https://developers.google.com/recaptcha/intro" target="_blank">{$mod_strings['LBL_CAPTCHA_DESC4']}</a>.
					{$mod_strings['LBL_CAPTCHA_DESC5']} <a href="https://developers.google.com/recaptcha/docs/domain_validation" target="_blank">{$mod_strings['LBL_CAPTCHA_DESC6']}</a>.					
					</div>
				</div>
			</div>		
			<div class="row form-row" id="captcha_key_container" style="display: none;">
				<div class="column form-cell sm-12 vertical">
					<div class="form-label"><span id="captcha_key_label">{$mod_strings['LBL_CAPTCHA_KEY']}</span><span class="required">*</span></div>
					<div class="form-entry">
                        <input class="input-text" id="captcha_key" name="captcha_key" size="60" value="{$captcha_key}" type="text">
					</div>
				</div>
			</div>
			<div class="row form-row" id="captcha_secret_container" style="display: none;">
				<div class="column form-cell sm-12 vertical">
					<div class="form-label"><span id="captcha_secret_label">{$mod_strings['LBL_CAPTCHA_SECRET_KEY']}</span><span class="required">*</span></div>
					<div class="form-entry">
                        <input class="input-text" id="captcha_secret" name="captcha_secret" size="60" value="{$captcha_secret}" type="text">
					</div>
				</div>
			</div>
EOQ;

		$layout = $gen->getLayout();
		/** @var FormLayout $layout */
		$layout->addScriptInclude('modules/Campaigns/captcha.js');

		return $html;
	}

	/**
	 * Request CAPTCHA verification
	 *
	 * @param string $token
	 *
	 * @return bool|array
	 */
	private function requestVerification($token) {
		$body = http_build_query([
			'secret' => $this->getSecret(),
			'response' => $token
		]);

		$options = [
			CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
			CURLOPT_HEADER => 0,
			CURLOPT_RETURNTRANSFER => 1,
			CURLOPT_FOLLOWLOCATION => 0,
			CURLOPT_URL =>  self::VERIFICATION_URL,
			CURLOPT_CUSTOMREQUEST => 'POST',
			CURLOPT_POSTFIELDS => $body
		];

		$ch = curl_init();

		foreach ($options as $name => $value) {
			curl_setopt($ch, $name, $value);
		}

		$result = curl_exec($ch);
		curl_close($ch);

		if ($result === false)
			return false;

		return json_decode($result, true);
	}
}