<?php

require_once 'include/layout/forms/FormGenerator.php';
require_once 'include/layout/FieldFormatter.php';
require_once 'include/database/ListQuery.php';
require_once 'include/config/ModelDef.php';

class PersonalInfoViewer {

	private $module;

	private $model_name;

	private $record_id;

	public function __construct($module, $id) {
		$this->module = $module;
		$this->model_name = AppConfig::module_primary_bean($module);
		$this->record_id = $id;

		if (!$this->model_name || !AppConfig::setting("model.detail.{$this->model_name}.personal_data")) {
			ACLController::displayNoAccess();
			return;
		}

		if (!ACLController::checkAccess($this->module, 'manage_personal')) {
			ACLController::displayNoAccess();
			return;
		}
	}

	/**
	 * @param $format
	 *
	 * @throws IAHError
	 * @throws IAHPermissionError
	 */
	public function processRequest($format) {
		if (empty($format))
			$format = 'html';

		try {
			$gen = $this->generateForm($format);
		} catch (Exception $e) {
			ACLController::displayNoAccess();
			return;
		}

		echo '<div class="detailOuter">' . $gen->getResult() . '</div>';
		$gen->exportIncludes();
	}

	public function generatePdfFile() {
		try {
			$gen = $this->generateForm('pdf');
		} catch (Exception $e) {
			return null;
		}

		$file = tempnam(sys_get_temp_dir(), 'PDF');
		$gen->pdf->Output($file, 'F');

		return $file;
	}

	/**
	 * @param $format
	 *
	 * @return HtmlFormGenerator|PdfFormGenerator
	 * @throws IAHError
	 * @throws IAHPermissionError
	 */
	private function generateForm($format) {
		global $app_list_strings;
		$result = ListQuery::quick_fetch($this->model_name, $this->record_id);

		if (! $result) {
			ACLController::displayNoAccess();
			return;
		}

		list($layout, $extra_fields) = $this->createLayout($result);
		$fmt = new FieldFormatter($format);
		$fmt->formatRowResult($result);
		$model = new ModelDef($this->model_name);
		$model->addDisplayFieldDefinitions($extra_fields);

		$title = translate('LBL_PERSONAL_DATA_TITLE', 'DataPrivacy');
		if ($format == 'pdf')
			$title .= ' : ' . array_get_default($app_list_strings['moduleList'], $this->module, $this->module);

		$gen = FormGenerator::new_form($format, $model, $layout, 'ViewPersonal');
		$gen->layout->setTitle($title);
		$gen->setPerform('view');
		//$hidden = array('module' => $_REQUEST['module'], 'action' => $_REQUEST['action'], 'record' => $_REQUEST['record'], 'perform_erase' => 1);
		//$gen->getFormObject()->addHiddenFields($hidden);
		$gen->renderForm($result);

		return $gen;
	}

	protected function addField(array &$fields, $f) {
		if (is_array($f))
			$f = array_get_default($f, 'name');
		if (!in_array($f, $fields))
			$fields[] = $f;
	}

	protected function createLayout($result) {
		$extra_fields = [];
		$fields = $this->getPreferredFieldsOrder();
		$sources = $this->fetchPersonalInfoSources();
		foreach ($sources as $k => $v) {
			$fname = '____pi_' . $k;
			$fdef = [
				'type' => 'varchar',
				'name' => '____pi_' . $k,
				'vname' => 'LBL_PERSONAL_INFO_SOURCE_LABEL',
				'vname_module' => 'DataPrivacy',
			];
			$extra_fields[$fname] = $fdef;
			$result-> fields[$fname] = $fdef;
			$result->row['____pi_' . $k] = $v;
		}
		$layoutFields = [];
		foreach ($fields as $f) {
			$layoutFields[] = $f;
			if (is_array($f))
				continue;
			$layoutFields[] = '____pi_' . $f;
		}
		$layout = [
			'type' => 'view',
			'form_buttons' => [
				'print' => [
					'icon' => 'icon-print',
					'vname' => 'LBL_PDF_PDF',
					'perform' => "window.open('index.php?module=" . $this->module . "&format=pdf&action=ViewPersonal&record=" . $this->record_id . "'); return false;",
				],
				'cancel' => [
					'icon' => 'icon-cancel',
					'vname' => 'LBL_CANCEL_BUTTON_LABEL',
					'url' => 'index.php?module=' . $this->module . '&action=DetailView&record=' . $this->record_id,
				],
			],
			'sections' => [
				[
					'id' => 'main',
					'label_position' => 'top',
					'elements' => $layoutFields,
				],
			],
		];
		return [$layout, $extra_fields];
	}

	protected function fetchPersonalInfoSources() {
		$ret = [];
		$model = new ModelDef($this->model_name);
		if (!empty($model->audit_table_name)) {
			$lq = new ListQuery('audit');
			$lq->base_model->table_name = $model->audit_table_name;
			$lq->addSimpleFilter('parent_id', $this->record_id);
			$lq->setOrderBy('date_created asc');
			$existing = $lq->fetchAllRows();
			foreach ($existing as $row) {
				$ret[$row['field_name']] = $row['source'];
			}
		}
		return $ret;
	}

	protected function getPreferredFieldsOrder() {
		$layout = AppConfig::setting("views.layout.{$this->module}.record.Standard");
		if (!is_array($layout)) return [];
		$fields = [];
		if (!empty($layout['summary']['header']['fields'])) {
			foreach ($layout['summary']['header']['fields'] as $f) {
				$this->addField($fields, $f);
			}
		}
		if (!empty($layout['summary']['subheader']['fields'])) {
			foreach ($layout['summary']['subheader']['fields'] as $f) {
				$this->addField($fields, $f);
			}
		}
		if (!empty($layout['summary']['meta']['fields'])) {
			foreach ($layout['summary']['meta']['fields'] as $f) {
				$this->addField($fields, $f);
			}
		}
		if (!empty($layout['summary']['actions'])) {
			foreach ($layout['summary']['actions'] as $a) {
				if (!empty($a['fields'])) {
					foreach ($a['fields'] as $f) {
						$this->addField($fields, $f);
					}
				}
			}
		}
		if (!empty($layout['sections'])) {
			foreach ($layout['sections'] as $s) {
				if (!empty($s['elements'])) {
					foreach ($s['elements'] as $f) {
						$this->addField($fields, $f);
					}
				}
			}
		}
		$modelDef = new ModelDef($this->model_name);
		$all_fields = $modelDef->getFieldDefinitions();

		foreach($all_fields as $f => $_) {
			$this->addField($fields, $f);
		}

		$fields = array_values(array_filter($fields, function($f) {
			return AppConfig::setting("model.fields.{$this->model_name}.$f.personal_info");
		}));
		$fields = array_merge([ ['name' => 'id', 'colspan' => 2] ], $fields);
		return $fields;
	}

}
