<?php

require_once 'include/GoogleSync.php';
require_once 'include/upload_file.php';
require_once 'include/database/ListQuery.php';
require_once 'include/database/RowUpdate.php';
require_once 'modules/Documents/Document.php';
require_once 'modules/DocumentRevisions/DocumentRevision.php';

class GoogleDocsSync extends GoogleSync
{
	
	protected function getItems($query) {

		$optParams = [
			'pageSize' => $query['maxResults'],
			'q' =>  $query['q'],
			'orderBy' => $query['orderBy'],
			'pageToken' => array_get_default($query, 'pageToken'),
			'fields' => '*'
		];

		try {
			$response = $this->service->files->listFiles($optParams);
		} catch (\Google\Exception $e) {
			$this->logException($e);
		}

		return array_get_default($response, 'files', []);
	}

	protected function getService()
	{
		$service = new Google_Service_Drive($this->client);
		return $service;
	}
	
	protected function createIAHRevision($doc, $iahDoc)
	{
		global $db;

		// Note that the exported content is limited to 10MB.
		// https://developers.google.cn/drive/api/v3/reference/files/export
		$maxExportSize = 10 * 1024 * 1024;
		
		$tmpfname = tempnam(sys_get_temp_dir(), "gds_");
		$f = @fopen($tmpfname, 'w');
		if (!$f) {
			@unlink($tmpfname);
			return false;
		}
		
		try {
			$type = $this->getExportedMimeType($doc);
			$fileName = $doc->getName();
			$fileSize = $doc->getSize();
			
			if (!is_null($type) && $maxExportSize > $fileSize) {
				$content = $this->service->files->export($doc->getId(), $type['type'], ["alt" => "media"]);
				$contentType = $type['type'];
				$fileName .= '.' . $type['ext'];
				while (!$content->getBody()->eof()) {
					fwrite($f, $content->getBody()->read(1024));
				}
			} else {
				$http = $this->client->authorize();
				$contentType = $doc->getMimeType();
				$chunkSizeBytes = 1 * 1024 * 1024;
				$chunkStart = 0;
				while ($chunkStart < $fileSize) {
					$chunkEnd = $chunkStart + $chunkSizeBytes;
					$response = $http->request(
						'GET',
						sprintf('/drive/v3/files/%s', $doc->getId()),
						[
							'query' => ['alt' => 'media'],
							'headers' => [
								'Range' => sprintf('bytes=%s-%s', $chunkStart, $chunkEnd)
							]
						]
					);
					$chunkStart = $chunkEnd + 1;
					fwrite($f, $response->getBody()->getContents());
				}
				
			}
		} catch (\Google\Exception $e) {
			$this->logException($e);
		}

		fclose($f);
		
		$fl = new UploadFile('');
		$fl->use_soap = true;
		$fl->stored_file_name = $fileName;
		$fl->create_stored_filename();
		$upload_dir = AppConfig::upload_dir();
		$dst = $fl->stored_file_name;
		rename($tmpfname, $upload_dir . $dst);

		$revNo = 1;
		if ($iahDoc->new_record) {
			$iahDoc->set('id', create_guid());
		} else {
			$query = "SELECT MAX(revision) as rev FROM document_revisions WHERE document_id = '" . $db->quote($iahDoc->getField('id')) . "'";
			$res = $db->query($query);
			$row = $db->fetchByAssoc($res);
			$revNo = ((int)$row['rev']) + 1;
		}
		$revision = RowUpdate::blank_for_model('DocumentRevision');
		$revision->set('revision', $revNo);
		$revision->set('filename', $fl->get_stored_file_name());
		$revision->set('file_mime_type', $contentType);
		$revision->set('file_ext', $fl->file_ext);
		$revision->set('document_id', $iahDoc->getField('id'));
		$revision->set('date_modified', gmdate('Y-m-d H:i:s', strtotime($this->getUpdated($doc))));
		$revision->set('created_by', $this->user->id);
		$revision->set('date_entered', gmdate('Y-m-d H:i:s', strtotime($this->getUpdated($doc))));
		$revision->save();

		$iahDoc->set('document_revision_id', $revision->getPrimaryKeyValue());
	}

	protected function getExportedMimeType($doc)
	{
		$type = null;
		switch ($doc->getMimeType()) {
			case 'application/vnd.google-apps.spreadsheet':
				$type = $this->getSpreadsheetMimeType($doc);
				break;
			case 'application/vnd.google-apps.document':
				$type = $this->getDocMimeType($doc);
				break;
			case 'application/vnd.google-apps.presentation':
				$type = $this->getPresentationMimeType($doc);
				break;
		}
		return $type;
	}

	protected function getDocMimeType($doc)
	{
		$format = $this->user->getPreference('google_docs_doc');
		
		if ($format == 'odt') {
			return ['type' => 'application/vnd.oasis.opendocument.text', 'ext' => 'odt'];
		} elseif ($format == 'rtf') {
			return ['type' => 'application/rtf', 'ext' => 'rtf'];
		} 
		return ['type' => 'application/vnd.openxmlformats-officedocument.wordprocessingml.document', 'ext' => 'docx'];
	}

	protected function getSpreadsheetMimeType($doc)
	{
		$format = $this->user->getPreference('google_docs_spreadsheet');
		if ($format == 'csv') {
			return ['type' => 'text/csv', 'ext' => 'csv'];
		}
		return ['type' => 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet', 'ext' => 'xlsx'];
	}

	protected function getPresentationMimeType($doc)
	{
		$format = $this->user->getPreference('google_docs_presentation');
		if ($format != 'pdf') {
			return ['type' => 'application/vnd.openxmlformats-officedocument.presentationml.presentation', 'ext' => 'pdf'];
		}
		return ['type' => 'application/pdf', 'ext' => 'pdf'];
	}

	protected function getConfigKey()
	{
		return 'google_docs';
	}

	protected function getUpdated($item) {
		return $item->getModifiedTime() ? $item->getModifiedTime() : $item->getCreatedTime();
	}

	protected function deleteItem($item) {
  		$this->service->files->delete($item->getId());
	}
	
	protected function isItemDeleted($item) {
		$del =  $item->explicitlyTrashed;
		return $del;
	}

	protected function getQuery()
	{
		$query = array(
			'q' => "(mimeType != 'application/vnd.google-apps.folder' and 'me' in owners)",
			'orderBy' => 'modifiedTime',
		);
		return $query;
	}

	protected function createIAHItem($sql_row = null)
	{
		$id = null;
		if (is_array($sql_row))
			$id = array_get_default($sql_row, 'id');
		if (!$id)
			return RowUpdate::blank_for_model('Document');
		$row = ListQuery::quick_fetch('Document', $id, true, array('filter_deleted' => false));
		if ($row)
			return RowUpdate::for_result($row);
		
		return RowUpdate::blank_for_model('Document');
	}

	protected function setIAHItemFields($iahItem, $item)
	{
		$iahItem->set('document_name', $item->getName());
		$iahItem->set('status_id', 'Archived');
		$this->createIAHRevision($item, $iahItem);
	}

	protected function getSQLQuery($lastSyncTime)
	{
		$query = "SELECT d.id, s.* "
			." FROM documents d "
			." LEFT JOIN google_sync s ON s.related_id=d.id AND s.user_id='".$this->user->id . "' AND s.related_type='Documents'"
			." WHERE (IFNULL(s.google_id,'')!='' OR NOT d.deleted) AND IFNULL(d.section, '') != 'hr' ";
		$query .= " AND (d.date_modified > s.google_last_sync OR s.google_last_sync IS NULL)";
		$query .= ' ORDER BY d.date_modified ';
		$query .= ' LIMIT ' . (int)array_get_default($this->params, 'max_items', 50);
		return $query;
	}


	protected function uploadFile($revision, $file, $insert) {
			$filename = strlen($revision->getField('full_filename')) ? $revision->getField('full_filename') : $revision->getField('filename');
			$upload_dir = AppConfig::upload_dir();
			$filenameParts = explode('.', $filename);
			$fileExtension = end($filenameParts);
			$mimeType = $revision->getField('file_mime_type');
			if (empty($mimeType)) {
				return true;
			}
			$full_fname = $upload_dir . $filename;
			$chunkSizeBytes = 1 * 1024 * 1024;
			// Call the API with the media upload, defer so it doesn't immediately return.
			$this->client->setDefer(true);
			if ($insert) {
				$request = $this->service->files->create($file);
			} else {
				$emptyFile = new Google_Service_Drive_DriveFile();
				$request = $this->service->files->update($file->getId(), $emptyFile);
			}
			// Create a media file upload to represent our upload process.
			$media = new Google_Http_MediaFileUpload(
				$this->client,
				$request,
				$mimeType,
				null,
				true,
				$chunkSizeBytes
			);
			$media->setFileSize(filesize($full_fname));

			$status = false;
			$handle = fopen($full_fname, "rb");
			while (!$status && !feof($handle)) {
				$chunk = fread($handle, $chunkSizeBytes);
				$status = $media->nextChunk($chunk);
			}

			// The final value of $status will be the data from the API for the object
			// that has been uploaded.
			$result = false;
			if ($status != false) {
				$result = $status;
			}
			fclose($handle);
			$this->client->setDefer(false);
			return $result;
	}

	protected function realCreateGoogleItem($iahItem)
	{

		$revision = ListQuery::quick_fetch('DocumentRevision', $iahItem->getField('document_revision_id'), true);
		if ($revision) {
			$file = new Google_Service_Drive_DriveFile();
			$file->setName($iahItem->getField('document_name') . '.' . $revision->getField('file_ext'));
			return $this->uploadFile($revision, $file, true);
		}
		return true;
	}

	protected function realUpdateGoogleItem($item, $iahItem)
	{
		$item->setName($iahItem->getField('document_name'));
		$revision = ListQuery::quick_fetch('DocumentRevision', $iahItem->getField('document_revision_id'));
		if ($revision) {
			return $this->uploadFile($revision, $item, false);
		}
		return $item;
	}

	protected function getListEntry($googleId)
	{
		$this->client->setDefer(false);
		return $this->service->files->get($googleId);
	}

	protected function mapLegacyGoogleId($googleId) {
		if (!preg_match('~https://~', $googleId))
			return $googleId;
		$parts = parse_url($googleId);
		$id = urldecode(basename($parts['path']));
		if (strpos($id, ':') !== false)
			list(, $id) = explode(':', $id);
		return $id;
	}

	protected function saveItem($item, $iahItem)
	{
		try {
			//$item->save();
		} catch (Zend_Gdata_App_IOException $e) {
			$iahItem->_google_fields['google_sync_error'] = 1;
			$iahItem->_google_fields['google_last_sync'] = '1970-01-01 00:00:00';
			return true;
		}
	}

	protected function supportsSyncToken() {
		return false;
	}

	protected function setQueryUpdatedMin(&$query, $min) {
		if ($min) {
			if (empty($query['q']))
				$query['q'] = "modifiedTime >= '$min'";
			else 
				$query['q'] = "({$query['q']}) AND modifiedTime >= '$min'";
		}
	}
	
	protected function getGoogleIdClause($googleId) {
		$clause = array('multiple' => array(), 'operator' => 'or');
		$clause['multiple'][] = array(
			'field' => 'google_id',
			'value' => $googleId,
		);
		$clause['multiple'][] = array(
			'field' => 'google_id',
			'value' => 'http%://%:' . $googleId,
			'operator' => 'like',
			'match' => 'custom',
		);
		$clause['multiple'][] = array(
			'field' => 'google_id',
			'value' => 'http%://%%3A' . $googleId,
			'operator' => 'like',
			'match' => 'custom',
		);
		return $clause;
	}
}

