<?php

require_once 'include/config/format/ConfigParser.php';
require_once 'include/config/format/ConfigWriter.php';
require_once 'include/database/RowUpdate.php';
require_once 'modules/EmailFolders/EmailFolder.php';

class EmailDistributor
{


	private $user_ids;
	private $method;

	private static $VALID_METHODS = array(
		'direct',
		'roundRobin',
		'leastBusy',
	);

	const VALID = 0;
	const INVALID_METHOD = 1;
	const INVALID_NO_USERS = 2;
	const INVALID_DIRECT_USERS = 3;

	private $rr_users;

	private $unread_counts;

	public function __construct($method, array $user_ids)
	{
		$this->method = $method;
		$this->user_ids = $user_ids;
	}

	public static function error_text($code)
	{
		switch ($code) {
		case self::INVALID_METHOD:
			return translate('ERR_INVALID_DISTRIBUTION_METHOD', 'Emails');
		case self::INVALID_NO_USERS:
			return translate('ERR_INVALID_DISTRIBUTION_NO_USERS', 'Emails');
		case self::INVALID_DIRECT_USERS:
			return translate('ERR_INVALID_DISTRIBUTION_DIRECT_USERS', 'Emails');
		}
	}

	public function validate()
	{
		if (!in_array($this->method, self::$VALID_METHODS))
			return self::INVALID_METHOD;

		$idsCount = is_countable($this->user_ids) ? count($this->user_ids) : 0;
		if ($idsCount < 1)
			return self::INVALID_NO_USERS;

		if ($idsCount > 1 && $this->method == 'direct')
			return self::INVALID_DIRECT_USERS;

		return self::VALID;

	}

	public function initialize()
	{
		$method = 'initialize_' . $this->method;
		if (method_exists($this, $method)) {
			$this->$method();
		}
	}


	public function finalize()
	{
		$method = 'finalize_' . $this->method;
		if (method_exists($this, $method)) {
			$this->$method();
		}
	}

	public function distribute(RowResult $email)
	{
		$method = 'distribute_' . $this->method;
		$this->$method($email);
	}


	/*
	 * *******************************************
	 *           Round-robin 
	 * *******************************************
	 */

	private function initialize_roundRobin()
	{
		$this->rr_users = array();
		foreach ($this->user_ids as $uid) {
			if (!in_array($uid, $this->rr_users)) {
				$this->rr_users[] = $uid;
			}
		}
	}

	private function finalize_roundRobin()
	{
		return true;
	}

	private function distribute_roundRobin(RowResult $email)
	{
		$uid = array_shift($this->rr_users);
		if($uid) {
			// rotate the list
			$this->rr_users[] = $uid;

			$upd = RowUpdate::for_result($email);
			$upd->set(array(
				'folder' => EmailFolder::get_std_folder_id($uid, STD_FOLDER_INBOX),
				'assigned_user_id' => $uid,
				'isread' => 0,
			));
			$upd->save();
		}
	}


	/*
	 * *******************************************
	 *           Direct assignment 
	 * *******************************************
	 */

	private function distribute_direct(RowResult $email)
	{
		$uid = $this->user_ids[0];
		$upd = RowUpdate::for_result($email);
		$upd->set(array(
			'folder' => EmailFolder::get_std_folder_id($uid, STD_FOLDER_INBOX),
			'assigned_user_id' => $uid,
			'isread' => 0,
		));
		$upd->save();
	}

	/*
	 * *******************************************
	 *           Direct assignment 
	 * *******************************************
	 */


	private function initialize_leastBusy()
	{
		$lq = new ListQuery('Email');
		$lq->addSimpleFilter('assigned_user_id', $this->user_ids);
		$lq->addSimpleFilter('isread', 0);
		$lq->addField(array('name' => 'assigned_user_id', 'grouped' => true));
		$lq->addTotal(array('name' => 'id', 'total' => 'count'));
		$res = $lq->runSummationQuery();
		$counts = array();
		foreach ($res->rows as $row) {
			if ($row['depth'] != 1)
				continue;
			$counts[$row['value1']] = $row['id:count'];
		}
		foreach ($this->user_ids as $uid) {
			if (!isset($counts[$uid]))
				$counts[$uid] = 0;
		}
		$this->unread_counts = $counts;
	}

	private function distribute_leastBusy(RowResult $email)
	{
		// first shuffle the counts array, so users with equal
		// unread counts will be picked up randomly
		$keys = array_keys($this->unread_counts);
		shuffle($keys);
		$randomized = array();
		foreach($keys as $key) {
			$randomized[$key] = $this->unread_counts[$key];
		}

		// now sort the counts array
		asort($randomized, SORT_NUMERIC);
		$this->unread_counts = $randomized;

		$keys = array_keys($this->unread_counts);
		$uid = $keys[0];

		$upd = RowUpdate::for_result($email);
		$upd->set(array(
			'folder' => EmailFolder::get_std_folder_id($uid, STD_FOLDER_INBOX),
			'assigned_user_id' => $uid,
			'isread' => 0,
		));
		$upd->save();

		$was = $email->getField('assigned_user_id');
		if (!$email->getField('isread') && $was != $uid) {
			$this->unread_counts[$was]--;
			$this->unread_counts[$uid]++;
		}

	}
}
