<?php

class FinalMilestone {

	/**
	 * @var array
	 */
	private $invoices;

	/**
	 * FinalMilestone constructor.
	 *
	 * @param array $invoices: milestone invoices
	 */
	public function __construct($invoices) {
		$this->invoices = $invoices;
	}

	/**
	 * Create milestone invoice group for Final Milestone Invoice
	 *
	 * @param string $group_id
	 *
	 * @return array
	 */
	public function createInvoicesGroup($group_id) {
		global $mod_strings;

		$group = [
			$group_id => [
				'id' => $group_id,
				'status' => 'Closed Accepted',
				'name' => $mod_strings['LBL_FINAL_MILESTONE_INVOICE_GROUP_NAME'],
				'lines' => [],
				'lines_order' => [],
				'adjusts' => [],
				'group_type' => 'products',
				'tax' => 0,
				'shipping' => 0
			]
		];

		foreach($this->invoices as $row) {
			if ($row['cancelled'])
				continue;

			if ($row['tax_exempt']) {
				$this->addInvoiceSimpleLine($row, $group, $group_id);
			}  else {


				$this->addInvoiceAssemblyLine($row, $group, $group_id);
			}
		}

		return $group;
	}

	/**
	 * Return true if group is milestone invoices group
	 * and false otherwise
	 *
	 * @param array $group - invoices line items group
	 *
	 * @return bool
	 */
	static public function is_milestone_invoices_group($group) {
		$result = false;

		//If at least one line in the group is Invoice so it's Milestone Invoices group
		foreach ($group['lines'] as $line) {
			if (isset($line['related_type']) && $line['related_type'] == 'Invoice')
				$result = true;
			break;
		}

		return $result;
	}

	/**
	 * @param array $group - invoices line items group
	 *
	 * @return array
	 */
	static public function calculate_invoices_group_totals($group) {
		$totals = [
			'total' => $group['total'],
			'tax' => 0,
			'shipping' => 0
		];

		foreach ($group['lines'] as $line) {
			if ($line['related_type'] == 'ShippingProviders') {
				$totals['shipping'] += $line['unit_price'];
			} elseif ($line['related_type'] == 'TaxRates') {
				$totals['tax'] += $line['unit_price'];
			}
		}

		return $totals;
	}

	/**
	 * Return additional Grand Totals for Final Milestone Invoice
	 *
	 * @return array
	 */
	static public function get_grand_totals() {
		return [
			'tax' => [
				'milestone_tax' => 'LBL_MILESTONE_INVOICES_TAX',
				'milestone_tax_remaining' => 'LBL_MILESTONE_INVOICES_TAX_REMAINING',
			],
			'shipping' =>[
				'milestone_shipping' => 'LBL_MILESTONE_INVOICES_SHIPPING',
				'milestone_shipping_remaining' => 'LBL_MILESTONE_INVOICES_SHIPPING_REMAINING',
			],
			'total' => [
				'milestone_total' => 'LBL_MILESTONE_INVOICES'
			]
		];
	}

	/**
	 * Add invoice to group line items (for tax exempt invoices)
	 *
	 * @param array $row - invoice row
	 * @param array $group
	 * @param string $group_id
	 */
	private function addInvoiceSimpleLine($row, &$group, $group_id) {
		$line = ['related_type' => 'Invoice'];
		$line['id'] = 'temp~' . rand(1000, 2000);
		$line['related_id'] = $row['id'];
		$line['currency_id'] = $row['currency_id'];
		$line['exchange_rate'] = $row['exchange_rate'];
		$line['quantity'] = 1;
		$line['name'] = $row['prefix'] . $row['invoice_number'] .': '. $row['name'];
		$line['unit_price'] = $row['amount'] * -1;
		$line['unit_price_usd'] = $row['amount_usdollar'] * -1;

		$group[$group_id]['lines'][$line['id']] = $line;
		$group[$group_id]['lines_order'][] = $line['id'];
	}

	/**
	 * Convert invoice data array to line item and add to group
	 *
	 * @param array $row: invoice row
	 * @param array $group: new invoice line group
	 * @param string $group_id
	 */
	private function addInvoiceAssemblyLine($row, &$group, $group_id) {
		$main_line_id = 'temp~' . rand(1000, 2000);

		$line = ['related_type' => 'Invoice'];
		$line['id'] = $main_line_id;
		$line['related_id'] = $row['id'];
		$line['currency_id'] = $row['currency_id'];
		$line['exchange_rate'] = $row['exchange_rate'];
		$line['quantity'] = 1;
		$line['name'] = $row['prefix'] . $row['invoice_number'] .': '. $row['name'];
		$line['sum_of_components'] = 1;

		$group[$group_id]['lines'][$line['id']] = $line;
		$group[$group_id]['lines_order'][] = $line['id'];

		$adjusts = $this->getInvoiceAdjusts($row['id']);

		//Add pretax line
		$pretax_row = $row;
		$pretax_row['name'] = translate('LBL_PRETAX', 'Quotes');
		$pretax_row['invoice_id'] = $row['id'];
		$pretax_row['amount'] = $row['pretax'];
		$pretax_row['amount_usd'] = $row['pretax_usd'];

		$line = $this->fillAssemblyItem($pretax_row, $main_line_id);

		$group[$group_id]['lines'][$line['id']] = $line;
		$group[$group_id]['lines_order'][] = $line['id'];

		//Add taxes and shipping lines
		foreach ($adjusts as $item) {
			$line = $this->fillAssemblyItem($item, $main_line_id);

			$group[$group_id]['lines'][$line['id']] = $line;
			$group[$group_id]['lines_order'][] = $line['id'];
		}
	}

	/**
	 * Fill assembly component item for Final milestone invoice
	 *
	 * @param array $row: data row
	 * @param string $parent_line_id
	 *
	 * @return array
	 */
	private function fillAssemblyItem($row, $parent_line_id) {
		$name = $row['name'];

		if (! empty($row['rate']))
			$name .= " ({$row['rate']}%)";

		$line = [];

		if (! empty($row['invoice_id'])) {
			$line['related_type'] = 'Invoice';
			$line['related_id'] = $row['invoice_id'];
		} else {
			$line['related_type'] = $row['type'];
			$line['related_id'] = '';
		}

		$line['id'] = 'temp~' . rand(1000, 2000);
		//$line['type'] = isset($row['type']) ? $row['type'] : 'Pretax';
		$line['parent_id'] = $parent_line_id;
		$line['currency_id'] = $row['currency_id'];
		$line['exchange_rate'] = $row['exchange_rate'];
		$line['quantity'] = 1;
		$line['depth'] = 1;
		$line['name'] = $name;
		$line['unit_price'] = $row['amount'] * -1;
		$line['unit_price_usd'] = $row['amount_usd'] * -1;

		return $line;
	}

	/**
	 * Get invoice's adjust lines (only taxes and shipping)
	 *
	 * @param string $id
	 *
	 * @return array
	 */
	private function getInvoiceAdjusts($id) {
		$invoice = ListQuery::quick_fetch('Invoice', $id);

		/** @var TallyUpdate $tally */
		$tally = RowUpdate::for_result($invoice);

		$groups = $tally->getGroups();
		$result = [];

		foreach ($groups as $item) {
			if (! isset($item['adjusts']))
				continue;

			foreach ($item['adjusts'] as $row) {
				if ($row['related_type'] == 'Discounts')
					continue;

				$result[] = [
					//'invoice_id' => $row['invoice_id'],
					'currency_id' => $invoice->getField('currency_id'),
					'exchange_rate' => $invoice->getField('exchange_rate'),
					'name' => $row['name'] ? $row['name'] : $row['type'],
					'amount' => $row['amount'],
					'amount_usd' => $row['amount_usd'],
					'rate' => $row['rate'],
					'type' => $row['related_type']
				];
			}
		}

		return $result;
	}
}