<?php

require_once 'modules/LiveChatInc/APIClient.php';

class LiveChatIncSetup {

	var $mode;
	var $client;
	var $userInfo;
	var $missing_hooks;
	var $invalid_hooks;
	var $valid_hooks;
	var $hooks;

	static private $expectedHooks = [
		[
			'event_type' => 'chat_started',
			'data_types' => ['chat', 'visitor', 'pre_chat_survey'],
		],
		[
			'event_type' => 'chat_ended',
			'data_types' => ['chat', 'visitor', 'pre_chat_survey'],
		],
		[
			'event_type' => 'chat_changed',
			'data_types' => ['chat', 'visitor'],
		],
		[
			'event_type' => 'visitor_queued',
			'data_types' => ['visitor'],
		],
		[
			'event_type' => 'ticket_created',
			'data_types' => ['ticket'],
		],
	];


	function __construct() {
		$this->mode = array_get_default($_POST, 'mode');
		$this->client = new LiveChatIncAPIClient($_POST['username'], $_POST['api_key']);
	}

	function processRequest() {
		$this->userInfo = $this->checkErrors($this->client->send('GET', '/agents/' . $_POST['username']), 'Cannot retrieve Live Chat user info');
		if ($this->userInfo['permission'] != 'owner' && $this->userInfo['permission'] != 'administrator') {
			$msg = is_array($this->userInfo) ? '' : $this->userInfo;
			$this->dieWithErrors('This user is not Live Chat account owner or administrator', [$msg]);
		}
		$this->checkHooks();

		switch ($this->mode) {
		case 'fix':
			$this->fixWebhooks();
			break;
		case 'remove':
			$this->removeWebhooks();
			break;
		}

		$this->render();
	}

	function render() {
		echo '<div style="margin:5px">';
		$this->renderCheckResults();
		echo '</div>';
	}

	function renderCheckResults() {
		$quoted_username = to_html($_POST['username']);
		$quoted_key = to_html($_POST['api_key']);
		$quoted_avatar = to_html($this->userInfo['avatar']);
		$quoted_name = to_html($this->userInfo['name']);
		$csrf_input = create_csrf_input();
		echo <<<HTML
		<form action="async.php" method="POST">
			{$csrf_input}
			<input type="hidden" name="module" value="LiveChatInc">
			<input type="hidden" name="action" value="Setup">
			<input type="hidden" name="username" value="{$quoted_username}">
			<input type="hidden" name="api_key" value="{$quoted_key}">
		<div class="card-body summary-body">
			<div class="row form-row summary-row">
				<div class="column summary-image">
					<div class="image-wrap" style="border: none">
						<img style="border-radius: 50%" src="//{$quoted_avatar}" border="0"  width="90" height="90" >
					</div>
				</div>
				<div class="column summary-main">
					<div class="row">
						<div class="column summary-left">
							<h4>{$quoted_name}</h4>
						</div>
					</div>
				</div>
			</div>
		</div>
HTML;
		echo '<h3>Check results</h3>';
		if (!empty($this->invalid_hooks) || !empty($this->missing_hooks)) {
			if (!empty($this->invalid_hooks) && is_countable($this->invalid_hooks)) {
				printf ("<p>Found %d invalid webhooks</p>", count($this->invalid_hooks));
			}
			if (!empty($this->missing_hooks) && is_countable($this->missing_hooks)) {
				printf ("<p>%d webhooks are missing </p>", count($this->missing_hooks));
			}
			echo '<button class="input-button" onclick="fixLiveChatIncWebhooks(this.form); return false;" ><span class="input-label" >Fix 1CRM webhooks</span></button>';
		} else {
			echo '<p>All webhooks are set up correctly</p>';
		}
		if (is_countable($this->valid_hooks) && count($this->valid_hooks)) {
			echo ' <button class="input-button" onclick="removeLiveChatIncWebhooks(this.form); return false;" ><span class="input-label" >Remove all 1CRM webhooks</span></button>';
		}

		echo '</form>';
	}

	function removeWebhooks() {
		foreach ($this->valid_hooks as $idx) {
			$hook = $this->hooks[$idx];
			$r = $this->checkErrors($this->client->send('DELETE', '/webhooks/' . $hook['id']), 'Failed to remove webhook');
		}
		$this->checkHooks();
	}

	function fixWebhooks() {
		$url = $this->getHookUrl();
		foreach ($this->invalid_hooks as $info) {
			list($expectedIndex, $actualIndex) = $info;
			$hook = $this->hooks[$actualIndex];
			$r = $this->checkErrors($this->client->send('DELETE', '/webhooks/' . $hook['id']), 'Failed to remove invalid webhook');
			$this->missing_hooks[] = $expectedIndex;
		}
		foreach ($this->missing_hooks as $idx) {
			$hook = self::$expectedHooks[$idx];
			$hook['url'] = $url;
			$r = $this->checkErrors($this->client->send('POST', '/webhooks', $hook), 'Failed to create webhook');
		}
		$this->checkHooks();
	}

	function checkHooks() {
		$hooks = $this->checkErrors($this->client->send('GET', '/webhooks'), 'Cannot retrieve Live Chat webhooks configuration');
		$this->hooks = $hooks['events'];
		$this->missing_hooks = [];
		$this->invalid_hooks = [];
		$this->valid_hooks = [];
		$url = $this->getHookUrl();
		foreach (self::$expectedHooks as $index => $hookInfo) {
			$existing = array_reduce($this->hooks, function($carry, $hook) use ($hookInfo, $url) {
				if ($carry > 0) {
					return $carry;
				} elseif ($hook['event_type'] == $hookInfo['event_type'] && $hook['url'] == $url) {
					return -$carry;
				} else {
					return $carry - 1;
				}
			}, -1);
			if ($existing > 0) {
				$hook = $this->hooks[$existing - 1];
				$expected = $hookInfo['data_types'];
				$actual = $hook['data_types'];
				sort($expected);
				sort($actual);
				if ($expected != $actual) {
					$this->invalid_hooks[] = [$index, $existing - 1];
				} else {
					$this->valid_hooks[] = $existing - 1;
				}
			} else {
				$this->missing_hooks[] = $index;
			}
		}
	}

	function getHookUrl() {
		$url = preg_replace('~^http:~', 'https:', AppConfig::site_url());
		return $url . 'chatapi.php/private/livecahtinc/webhook/' . AppConfig::unique_key();
	}

	function checkErrors($reply, $message) {
		$errors = null;
		if (!$reply) {
			$errors = ['Live Chat server returned incorrect reply'];
		} elseif (is_string($reply)) {
			$errors = [$reply];
		} else {
			if (!empty($reply['error'])) {
				$errors = [$reply['error']];
			} elseif (!empty($reply['errors'])) {
				$errors = $reply['errors'];
			}
		}
		if (!empty($errors)) {
			$this->dieWithErrors($message, $errors);
		}
		return $reply;
	}

	function dieWithErrors($message, $errors) {
		echo '<div style="margin: 10px">';
		echo '<p>'. $message . '</p>';
		echo '<p><b>' . join('<br>', $errors) . '</b></p>';
		echo '</div>';
		exit;
	}
}

$setup = new LiveChatIncSetup;
$setup->processRequest();

