<?php
require_once('include/OnlineMeetings/OnlineMeetingsFactory.php');

class RefreshOMToken {

	private  $lock_handle = null;

	/**
	 * @var OnlineMeeting
	 */
	private $onlineMeeting;

	/**
	 * @return bool
	 */
	static function do_refresh() {
		try {
			$inst = new RefreshOMToken();
		} catch (Exception $e) {
			//$GLOBALS['log']->fatal('Refreshing online meeting token error: ' . $e->getMessage());
			return false;
		}

		$inst->doRefresh();
		return true;
	}

	/**
	 * @param array|RowUpdate $updates
	 *
	 * @return bool
	 */
	public static function authorize(&$updates) {
		if (is_array($updates) && isset($updates['company.online_meeting_app'])) {
			$meeting_app = $updates['company.online_meeting_app'];
			$for_company = true;
		} else {
			$meeting_app = AppConfig::setting('company.online_meeting_app');
			$for_company = false;
		}


		try {
			$inst = new RefreshOMToken($meeting_app);
		} catch (Exception $e) {
			return false;
		}

		if (is_array($updates)) {
			$params = $updates;
		} else {
			/** @var RowUpdate $params */
			$params = $updates;
			$params = $params->getRow();
			$params = array_merge($params, $_REQUEST);
			$params['user_id'] = $params['id'];
		}

		$params['for_company'] = $for_company;

		try {
			$inst->onlineMeeting->onSaveSettingsEvent($params);
		} catch (OnlineMeetingException $e) {
			add_flash_message($e->getMessage(), 'warning');
		}

		return true;
	}

	/**
	 * RefreshOMToken constructor.
	 *
	 * @param string $meeting_app: gotomeeting, ms_teams, zoom
	 * @throws Exception
	 */
	public function __construct($meeting_app = null) {
		try {
			$this->onlineMeeting = OnlineMeetingsFactory::getInstance($meeting_app);
		} catch (OnlineMeetingException $e) {
			throw new Exception($e->getMessage());
		}
	}

	public function doRefresh() {
		//GoToMeeting uses separate class
		if ($this->onlineMeeting->getAppName() == 'gtm')
			return false;

		$status = $this->checkStatus();

		if(! $status['any_refresh'])
			return false;

		if(! $this->lock())
			return false;
		
		// check again now that we have a lock
		$status = $this->checkStatus();
		if(! $status['any_refresh'])
			return false;

		if ($status['system_refresh']) {
			try {
				$this->onlineMeeting->setAccessToken($status['system_token']);
				$newToken = $this->onlineMeeting->refreshToken();
				$this->onlineMeeting->saveAccessToken($newToken, true);
			} catch (Exception $e) {
				$GLOBALS['log']->fatal("Error refreshing system-wide {$this->onlineMeeting->getAppName()} token  " . AppConfig::current_user_id() . ": " . $e->getMessage());
			}
		}

		if ($status['user_refresh']) {
			try {
				$this->onlineMeeting->setAccessToken($status['user_token']);
				$newToken = $this->onlineMeeting->refreshToken();
				$this->onlineMeeting->saveAccessToken($newToken);
			} catch (Exception $e) {
				$GLOBALS['log']->fatal("Error refreshing {$this->onlineMeeting->getAppName()} token for user " . AppConfig::current_user_id() . ": " . $e->getMessage());
			}
		}
		
		$this->unlock();
	}

	/**
	 * @return array
	 */
	private function checkStatus() {
		$system_token = $this->onlineMeeting->loadStoredAccessToken(true);
		$user_token = $this->onlineMeeting->loadStoredAccessToken();
		$system_refresh = false;
		$user_refresh = false;
		$now = time();

		if ($system_token) {
			if (! $system_token) {
				$GLOBALS['log']->debug("System-wide {$this->onlineMeeting->getAppName()} token could not be decoded, will not be refreshed");
			} else if ($system_token['refresh_after'] >= $now) {
				$GLOBALS['log']->debug("System-wide {$this->onlineMeeting->getAppName()} token is new enough, no need to refresh");
			} else if ($system_token['refresh_before'] < $now) {
				$GLOBALS['log']->debug("System-wide {$this->onlineMeeting->getAppName()} token expired, no refresh attempt needed");
			} else {
				$system_refresh = true;
			}
		}

		if ($user_token) {
			if ($user_token['refresh_after'] >= $now) {
				$GLOBALS['log']->debug("{$this->onlineMeeting->getAppName()} token for user " . AppConfig::current_user_id() . " is new enough, no need to refresh");
			} else if ($user_token['refresh_before'] < $now) {
				$GLOBALS['log']->debug("{$this->onlineMeeting->getAppName()} token for user " . AppConfig::current_user_id() . " expired, no refresh attempt needed");
			} else {
				$user_refresh = true;
			}
		}

		$any_refresh = $system_refresh || $user_refresh;

		return compact(
			'system_refresh', 'system_token',
			'user_refresh', 'user_token',
			'any_refresh'
		);
	}

	/**
	 * @return string
	 */
	private function getLockFile() {
		return "cache/.{$this->onlineMeeting->getAppName()}.lock";
	}

	/**
	 * @return bool|null|resource
	 */
	private function lock() {
		$this->lock_handle = @fopen($this->getLockFile(), 'w');
		if($this->lock_handle)
			flock($this->lock_handle, LOCK_EX);
		return $this->lock_handle;
	}

	private function unlock() {
		if ($this->lock_handle) {
			flock($this->lock_handle, LOCK_UN);
			@fclose($this->lock_handle);
			$this->lock_handle = null;
		}
	}
}

