<?php

require_once 'include/config/format/ConfigWriter.php';
require_once 'include/config/format/ConfigParser.php';
require_once 'include/upload_file.php';

class ModuleDesigner
{
	var $input;
	var $module_name;
	var $bean_name;
	var $errors;

	public function __construct($module_name, $input)
	{
		$this->input = $input;
		$this->module_name = $module_name;
		$this->bean_name = AppConfig::setting("modinfo.by_name.{$module_name}.detail.primary_bean");
		if (!$this->bean_name)
			$this->bean_name = uniqid($module_name);
		$this->errors = [];
	}

	public function save()
	{
		if (empty($this->input['__save']))
			return false;
		if ($this->module_name == '') {
			$this->errors[] = 'LBL_MISSING_MODULE_NAME';
			return false;
		}
		
		$icon = new UploadFile('icon');
		$icon_path = null;
		if ($icon->confirm_upload(true)) {
			$ext = strtolower($icon->file_ext);
			if ($ext == 'gif') {
				$icon_path = "modules/{$this->module_name}/icon.$ext";
				$icon->final_move($icon_path, true);
				$this->input['icon_path'] = $icon_path;
			} else {
				$icon->clean_incoming();
			}
		}
		
		$info = $this->generateModInfo();
		$this->generateLanguage($info['by_designer']);
		$dir = 'modules/' . $this->module_name . '/';
		mkdir_recursive($dir . 'metadata');
		mkdir_recursive($dir . 'models');
		mkdir_recursive($dir . 'views');


		$cw = new ConfigWriter;
		if ($info['bean'] || $info['by_designer']) {
			$cw->writeFile($dir . 'metadata/module_info.php', $info['info']);
			if($info['bean'] || $info['update_bean']) {
				$b = !empty($info['bean']) ? $info['bean'] : $info['update_bean'];
				$cw->writeFile($dir . "models/bean.{$this->bean_name}.php", $b);
			}
			if($info['bean']) {
				$bean_path = $info['bean']['detail']['bean_file'];
				if(! file_exists($bean_path)) {
					$f = fopen($bean_path, 'w');
					fwrite($f, <<<TEXT
<?php
require_once 'data/SugarBean.php';
class {$this->bean_name} extends SugarBean {
	var \$new_schema = true;
	var \$object_name = "{$this->bean_name}";
	var \$module_dir = "{$this->module_name}";
}
TEXT
					);
					fclose($f);
				}
			}
			$this->createDefaultViews();
			$repair = true;
			AppConfig::invalidate_cache('display');
			AppConfig::invalidate_cache('views');
		} else {
			$cw->writeDiff(AppConfig::custom_dir() . '/' . $dir . 'metadata/module_info.php', $info['info'], $info['orig_info']);
			$b = array_get_default($info, 'update_bean');
			if ($b) {
				$repair = true;
				$cw->writeFile(AppConfig::custom_dir() . '/' . $dir . "models/bean.{$this->bean_name}.php", $b);
			}
		}
		
		AppConfig::invalidate_cache('modinfo');
		AppConfig::invalidate_cache('model');
		if(! empty($repair)) {
			require_once 'include/database/DBChecker.php';
			$checker = new DBChecker();
			try {
				$checker->checkRepairModel($this->bean_name, DB_CHECK_STANDARD | DB_CHECK_AUDIT, true, false);
			} catch(Exception $e) {
				$this->errors[] = $e->getMessage();
			}
		}
		AppConfig::invalidate_cache('lang');
		AppConfig::invalidate_cache('acl');
		return true;
	}

	public function getErrors()
	{
		return $this->errors;
	}

	public function isCustom()
	{
		return !is_dir('modules/' . $this->module_name);
	}

	private function generateModInfo()
	{
		$primary_bean = AppConfig::setting("modinfo.by_name.{$this->module_name}.detail.primary_bean");
		$by_designer = AppConfig::setting("modinfo.by_name.{$this->module_name}.detail.created_by_module_designer");

		$bean = null;
		$update_bean = null;
		$ret = $orig = [];

		if (!$primary_bean) {
			$ret['detail']['created_by_module_designer'] = $by_designer = true;
			$bean = array(
				'detail' => array(
					'type' => 'bean',
					'bean_file' => "modules/{$this->module_name}/{$this->bean_name}.php",
					'optimistic_locking' => true,
					'table_name' => strtolower($this->module_name) . uniqid('_'),
					'primary_key' => 'id',
					'default_order_by' =>  'name',
					'personal_data' => empty($this->input['personal_data']) ? false : true,
				),
				'fields' => array(
					'app.id' => array(),
					'app.date_entered' => array(),
					'app.date_modified' => array(),
					'app.modified_user' => array(),
					'app.assigned_user' => array(),
					'app.created_by_user' => array(),
					'app.deleted' => array(),
					'name' => array(
						'type' =>  'name',
						'vname' => 'LBL_NAME',
						'required' => true,
						'default' => "",
					),
				),
			);
			$primary_bean = $this->bean_name;
			$ret['basic_pdf_template'] = true;
		} else {
			$fname = 'modules/' . $this->module_name . '/metadata/module_info.php';
			$ret = $orig = ConfigParser::load_file($fname);
			$update_bean = [];
			if ($by_designer) {
				$dir = 'modules/' . $this->module_name . '/';
				$update_bean = ConfigParser::load_file($dir . "models/bean.{$this->bean_name}.php");
				$update_bean['detail']['reportable'] = empty($this->input['reportable']) ? false : true;
			}
			$update_bean['detail']['personal_data'] = empty($this->input['personal_data']) ? false : true;
		}

		$ret['detail']['default_group'] = $this->input['tab_group'];
		$ret['detail']['primary_bean'] = $primary_bean;
		if(! empty($this->input['icon_path']))
			$ret['detail']['icon'] = $this->input['icon_path'];

		$ret['acl']['editable'] = $this->input['acl_level'] == 'editable';
		if (!$ret['acl']['editable']) {
			foreach(array('list', 'view', 'edit', 'delete') as $action) {
				$ret['acl']['fixed'][$action] = ACLAction::levelIntToString($this->input['acl_' . $action]);
			}
		}
		return array(
			'by_designer' => $by_designer,
			'info' => $ret,
			'orig_info' => $orig,
			'bean' => $bean,
			'update_bean' => $update_bean,
		);

	}
	
	private function generateLanguage($native)
	{
		$base = AppConfig::setting('lang.base');
		$meta_p = "modules/{$this->module_name}/language/lang.$base.meta.php";
		$orig = null;
		$cw = new ConfigWriter;
		if(file_exists($meta_p)) {
			$meta = $orig = ConfigParser::load_file($meta_p);
		} else
			$meta = array();
		$meta['detail']['label'] = $this->input['mod_title'];
		if(isset($orig) && ! $native) {
			$cw->writeDiff(AppConfig::custom_dir() . $meta_p, $meta, $orig);
		} else {
			$cw->writeFile($meta_p, $meta);
		}
		
		$strings_p = "modules/{$this->module_name}/language/lang.$base.strings.php";
		if($native && ! file_exists($strings_p)) {
			$strings = array();
			$cw->writeFile($strings_p, $strings);
		}
	}

	private function createDefaultViews()
	{
		$views = array(
			'record' => array(
				'detail' => array(
					'type' => 'record',
					'title' => 'LBL_MODULE_TITLE',
				),
				'layout' => array(
					'summary' => array(),
					'sections' => array(
						array(
							'elements' => array(
							),
						),
					),
				),
			),
			'edit' => array(
				'detail' => array(
					'type' => 'editview',
					'title' => 'LBL_MODULE_TITLE',
				),
				'layout' => array(
					'sections' => array(
						array(
							'elements' => array(
								'name',
							),
						),
					),
				),
			),
			'view' => array(
				'detail' => array(
					'type' => 'view',
					'title' => 'LBL_MODULE_TITLE',
				),
				'layout' => array(
					'sections' => array(
						array(
							'elements' => array(
								'name',
							),
						),
					),
				),
			),
			'list' => array(
				'detail' => array(
					'type' => 'list',
					'title' => 'LBL_MODULE_TITLE',
				),
				'layout' => array(
					'columns' => array(
						'name',
						'assigned_user',
					),
				),
			),
		);

		$cw = new ConfigWriter;
		$dir = 'modules/' . $this->module_name . '/';
		foreach ($views as $view => $data) {
			$path = $dir . 'views/' . $view .  '.Standard.php';
			if(! file_exists($path))
				$cw->writeFile($path, $data);
		}

	}

}


