<?php

use OneCRM\NotariusClient\Document;
use OneCRM\NotariusClient\FormField;
use OneCRM\NotariusClient\Workflow;
use OneCRM\NotariusClient\WorkflowAction;
use OneCRM\NotariusClient\ClientOptions;
use OneCRM\NotariusClient\Client;

require_once'include/upload_file.php';

class PDFSigner {

	private $record;
	private $template;
	private $signers;
	private $address_id;
	private $parallel;
	private $inputs = [];

	public function loadInput($input) {
		$errors =  [];
		if (!$this->loadRecord($input)) {
			return ['Record not found'];
		}
		$template = ListQuery::quick_fetch('PDFTemplate', $input['pdf_template']);
		if (!$template || !$template->getField('signing_enabled')) {
			return ['Invalid template'];
		}
		$this->loadInputs($input);
		$this->template = $template;
		$this->parallel = !empty($input['parallel']);
		$res = $this->loadSigners($input);
		if (!empty($res)) {
			$errors = array_merge($errors, $res);
		}
		$this->address_id = $input['use_address'] ?? null;
		return empty($errors) ? true : $errors;
	}

	public function generateProject() {
		$printer = new PDFTemplatePrinter($this->template->getField('id'), $this->record->getField('id'));
		$printer->setInputs($this->inputs);
		$printer->useCompanyAddress($this->address_id);
		$pdf = $printer->render();
		$blocks = $printer->getSignatureBlocks();
		
		$filename = filename_safe_string($printer->getFileName());
		$title = $printer->getProvider()->getSignedDocumentTitle();

		$fl = new UploadFile('');
		$fl->use_soap = true;
		$fl->stored_file_name = $filename;
		$fl->create_stored_filename();
		$upload_dir = AppConfig::upload_dir();
		$dst = $fl->stored_file_name;
		$pdf->Output($upload_dir . $dst, 'F');

		$opts = new ClientOptions;
		$sandbox = AppConfig::setting("site.notarius.sandbox_uri");
		if ($sandbox)
			$opts = $opts->withSandbox($sandbox);

		$opts = $opts
			->withAuthKey(AppConfig::setting('site.notarius.auth_key', ''))
			->withSecret(AppConfig::setting('site.notarius.auth_secret', ''))
		;

		$opts = $opts->build();

		$docID = create_guid();
		$projID = create_guid();

		$client = new Client($opts);
		$client->authenticate();

		$meta = $client->uploadDocument($filename, fopen($upload_dir . $dst, "r"));

		$fields = array_map(function($b) {
			return [
				'x' => $b['x'],
				'y' => $b['y'],
				'page' => $b['page'],
				'assignedTo' => $b['signer_ref'],
			];
		}, $blocks);

		$actions = array_map(
			function($s) {
				return $this->generateSignerAction($s['record'], $s['ref'], $s['is_approver']);
			}, 
			$this->signers
		);

		$projectData = [
			'name' => $title,
			'expiresOn' => gmdate("Y-m-d", strtotime("now + 1 month")),
			'status' => 1,
			'documents' => [
				[
					'documentId' => $meta->documentId,
					'fields' => $fields,
				],
			],
			'actions' => $actions,
			'webhooks' => [
				[
					'url' => $this->getWebhookURI($projID),
					'insecure' => true,
				],
			],
		];
		$w = new Workflow($projectData);
		$w = $client->createWorkflow($w);

		$project  = RowUpdate::blank_for_model('SigningProject');
		$project->set([
			'id' => $projID,
			'linked_document_id' => $docID,
			'name' => $title,
			'workflow_id' => $w->id,
			'assigned_user_id' => AppConfig::current_user_id(),
		]);
		$project->save();
			
		$doc = RowUpdate::blank_for_model('Document');
		$doc->set([
			'id' => $project->getField('linked_document_id'),
			'document_name' => $title,
			'signing_project_id' => $project->getField('id'),
			'parent_id' => $this->record->getField('id'),
			'parent_type' => $this->record->getModuleDir(),
		]);

		$revision = RowUpdate::blank_for_model('DocumentRevision');
		$revision->set([
			'id' => create_guid(),
			'revision' => 1,
			'filename' => $fl->get_stored_file_name(),
			'file_mime_type' => 'application/pdf',
			'file_ext' => $fl->file_ext,
			'document_id' => $doc->getField('id'),
			'change_log' => translate('LBL_CHANGE_LOG_INITIAL', 'SigningProjects'),
		]);
		$revision->save();

		$doc->set('document_revision_id', $revision->getField('id'));
		$doc->save();	


	}

	private function loadInputs($input) {
		$inputs = @json_decode($input['inputs'], true);
		if (is_array($inputs)) {
			$this->inputs = $inputs;
		}
	}

	private function generateSignerAction($signer, $ref, $is_approver) {
		$phone = $this->getSignerPhone($signer);
		$sig_type = $signer->getField('notarius_signature_type');
		$subject = $signer->getField('notarius_subject_dn');
		$secondFactor = $signer->getField('notarius_second_factor');
		$language = $signer->getField('notarius_language', 'en');
		if ($language !== 'en' && $language !== 'fr') $language = 'en';
		switch ($secondFactor) {
		case 'call':
		case'sms':
			break;
		case'question':
			$secondFactor = 'secret';
			break;
		default:
			$secondFactor = 'sms';
		}
		$action = [
			'mode' => 'remote',
			'ref' => $ref,
			'signer' => [
				'type' => $is_approver ? 'approver' : ($sig_type == 'certifio' ? 'certifio' :'esig'),
				'firstName' => $signer->getField('first_name'),
				"lastName" => $signer->getField('last_name'),
				'email' => $signer->getField('email1'),
				'phone' => $phone,
				'amr' => ['link', $secondFactor],
				'lang' => $language,
			]
		];
		if ($sig_type == 'certifio')
			$action['signer']['subjectDN'] = $subject;
		if ($this->parallel)
			$action['step'] = 0;
		if ($secondFactor == 'secret') {
			$action['signer']['secretQuestion'] = $signer->getField('notarius_question');
			$action['signer']['secretAnswer'] = $signer->getField('notarius_answer');
			$action['signer']['isSecretAnswerChanged'] = true;
		}
		return $action;
	}

	private function getSignerPhone($signer) {
		$phone = null;
		$phone_fields = ['notarius_phone', 'phone_mobile', 'phone_wordk'];
		foreach ($phone_fields as $f) {
			$phone = $signer->getField($f);
			if (!empty($phone)) break;
		}
		return $phone;
	}

	private function loadRecord($input) {
		$model = AppConfig::module_primary_bean($input['module']);
		if (!$model) return false;
		$record = ListQuery::quick_fetch($model, $input['record']);
		if (!$record) return false;
		$this->record = $record;
		return true;
	}

	private function loadSigners($input) {
		$errors = [];
		$signers_def = json_decode($this->template->getField('signers'), true);
		$signers = json_decode($input['signers'], true);
		$loaded_signers = [];

		foreach ($signers as $signer_info) {
			$signer_errors = [];
			$model = '';
			$is_approver = ($signer_info['type'] ?? 'signer') == 'approver' ;
			$type = $signer_info['signer']['type'] ?? '';
			$id = $signer_info['signer']['id'] ?? '';
			switch ($type) {
			case 'Users':
			case 'Contacts':
				$model = $type == 'Users' ? 'User' : 'Contact';
				break;
			case 'module_field':
				$field_def = $this->record->getModel()->getFieldDefinition($id);
				$id = $this->record->getField($field_def['id_name']);
				$model = $field_def['bean_name'];
				break;
			}
			$rec = null;
			if ($model && $id) {
				$rec = ListQuery::quick_fetch($model, $id);
			}
			if ($rec) {
				$this->validateSigner($rec, $signer_errors);
				foreach ($signer_errors as $e) {
					$errors[] = $signer_info['role'] . ': ' . translate($e, 'PDFTemplates');
				}
				$loaded_signers[] = [
					'is_approver' => $is_approver,
					'ref' => $signer_info['ref'],
					'record' => $rec,
				];
			} else {
				$errors[] = $signer_info['role'] . ': ' . translate('LBL_CANT_LOAD_SIGNER', 'PDFTemplates');
			}
		}
		$this->signers = $loaded_signers;
		return $errors;
	}

	protected function getWebhookURI($projID) {
		$base = AppConfig::setting('site.notarius.webhook_base_url', AppConfig::site_url());
		$url = $base .'notarius.php?' . http_build_query(['p' => $projID, 's' => sha1($projID . '_' . AppConfig::unique_key())]);
		return $url;
	}

	protected function validateSigner($signer, array &$errors) {
		$secondFactor = $signer->getField('notarius_second_factor');
		$email = $signer->getField('email1');
		if (empty($email))
			$errors[] = 'LBL_MISSING_EMAIL';
		switch ($secondFactor) {
		case 'sms':
		case 'call':
			$phone = $this->getSignerPhone($signer);
			if (empty($phone))
				$errors[] = 'LBL_MISSING_PHONE';
			break;
		case 'question':
			$question = $signer->getField('notarius_question');
			$answer = $signer->getField('notarius_answer');
			if (!strlen($question)) 
				$errors[] = 'LBL_MISSING_QUESTION';
			if (!strlen($answer)) 
				$errors[] = 'LBL_MISSING_ANSWER';
			break;
		}
	}

}


