<?php

use setasign\Fpdi\Fpdi;

require_once 'modules/PDFTemplates/PDFTemplate.php';
require_once 'modules/PDFTemplates/utils.php';
require_once 'modules/PDFTemplates/PDFTemplatePDFManager.php';
require_once 'modules/PDFTemplates/PDFTemplateProviderBase.php';
require_once 'modules/PDFTemplates/PDFSigner.php';

class PDFTemplatePrinter {

	protected $focus;
	protected $tpl;
	protected $layout;
	protected $override_layout;
	protected $address_id;
	protected $in_header_footer = 0;
	protected $orientation = 'P';
	protected $headerSection, $footerSection;
	protected $headerSize = 0, $footerSize = 0;
	protected $sectionY, $sectionPage;
	protected $pageHeight;
	protected $marginTop, $marginBottom, $marginLeft, $marginRight;
	protected $marginPadding = 10;
	protected $set_margins = [];
	protected $signatureBlocks = [];
	protected $inputs = [];
	protected $provider;
	protected $pdf;
	protected $marginHeader = 0;
	protected $marginFooter = 0;

	public function __construct($template_id, $focus_id, $override_layout = null, $orientation=null, $margin_left=null, $margin_top=null)
	{
		$this->tpl = ListQuery::quick_fetch('PDFTemplate', $template_id);
		$module = $this->tpl->getField('template_module');
		$type = $this->tpl->getField('template_type');
		$this->orientation = $this->tpl->getField('page_orientation');
		if ($this->tpl) {
			$this->set_margins = [
				'left' => $this->tpl->getField('margin_left'),
				'top' => $this->tpl->getField('margin_top'),
			];
		}
		$this->set_margins = [
			'left' => array_get_default($this->set_margins, 'left', AppConfig::setting('pdf.margins.left')),
			'top' => array_get_default($this->set_margins, 'top', AppConfig::setting('pdf.margins.top')),
		];

		if(isset($margin_left) && isset($margin_top) && $margin_top != '' && $margin_left != '') {
			$this->set_margins = [
				'left' => $margin_left,
				'top' => $margin_top,
			];
		} else if(isset($margin_left) && $margin_left != '') {
			$this->set_margins = [
				'left' => $margin_left,
				'top' => array_get_default($this->set_margins, 'top', AppConfig::setting('pdf.margins.top')),
			];
		} else if(isset($margin_top) && $margin_top != '') {
			$this->set_margins = [
				'left' => array_get_default($this->set_margins, 'left', AppConfig::setting('pdf.margins.left')),
				'top' => $margin_top,
			];
		}

		$model = AppConfig::module_primary_bean($module);

		$this->override_layout = $override_layout;
		if(isset($orientation))
			$this->orientation = ($orientation == 'landscape' || $orientation == 'L' ? 'L' : 'P');

		$json = getJSONObj();
		$this->layout = $json->decode($this->tpl->getField('template'));
		PDFTemplate::translate_template($this->layout, $module);

		$this->provider = PDFTemplateProviderBase::create_provider($module, $type);
		$lq = new ListQuery($model, true, array('formatted' => 'pdf', 'use_real_names' => true));
		$lq->addFilterPrimaryKey($focus_id);
		$this->provider->initialize($this->getLayout(), $lq);
	}

	function setInputs($inputs) {
		$this->inputs = $inputs;
	}

	function getProvider() {
		return $this->provider;
	}

	public function getSignatureBlocks() {
		return $this->signatureBlocks;
	}

	public function recalcPageHeight()
	{
		$this->pageHeight = $this->pdf->getPageHeight();
		$m = $this->pdf->getMargins();
		$this->marginTop = $m['top'];
		$this->marginLeft = $m['left'];
		$this->marginRight = $m['right'];
		$this->marginBottom = $m['bottom'];
	}

	public function setHeaderSection($sect) {
		$this->headerSection = $sect;
		$this->headerSize = $this->scaleSize($sect['height']) + $this->marginPadding;
	}

	public function setFooterSection($sect) {
		$this->footerSection = $sect;
		$this->footerSize = $this->scaleSize($sect['height']) + $this->marginPadding;
	}

	public function renderHeaderFooter($footer=false) {
		$section = $footer ? $this->footerSection : $this->headerSection;
		if ($section) {
			$params = array(
				'address_id' => $this->pdf->getCompanyAddressId(),
			);
			$data = PDFTemplateProviderBase::get_page_header_footer_data($params);
			foreach ($data as $k => $v) {
				$data[$k] = $v['value'];
			}
			$data = array_merge($data, $this->provider->getData('header'));

			$data['__page_number'] = $this->pdf->getPage();
			$data['__total_pages'] = '{{nb}}';

			$this->renderSection($section, $data, array('context' => $footer ? 'footer' : 'header'));
			return true;
		}
	}

	public function useCompanyAddress($id)
	{
		$this->address_id = $id;
	}

	public function getLayout()
	{
		$layout = $this->override_layout ? $this->override_layout : $this->layout;
		initialize_page_header_footer($layout);
		return $layout;
	}

	public function getOrientation()
	{
		return $this->orientation;
	}

	public function render($pdf = null, $startPageGroup = false)
	{
		if (!$pdf) {
			$this->pdf = new PDFTemplatePDFManager('', $this->getOrientation());
			$this->pdf->SetMargins($this->set_margins['left'], $this->set_margins['top']);
			$this->pdf->loadLayout($this->getLayout(), $this);
			$this->pdf->useCompanyAddress($this->address_id);
		} else {
			$this->pdf = $pdf;
		}

		if ($startPageGroup)
			$this->pdf->startPageGroup();

		$this->recalcPageHeight();
		$this->pdf->new_page();
		$this->sectionPage = $this->pdf->getPage();
		$this->sectionY = $this->marginTop + $this->headerSize;

		$this->provider->begin();
		$this->provider->iterate($this);
		$this->provider->end();
		return $this->pdf;
	}

	private function splitSection($section) {
		$ret = array();
		$nextSection = $section;
		if(! isset($section['elements']))
			$section['elements'] = array();
		$nextSection['elements'] = array();
		$adjust = 0;
		foreach($section['elements'] as $el) {
			if (!empty($el['page_break'])) {
				if (!empty($nextSection['elements'])) {
					$ret[] = $nextSection;
				}
				$adjust = $el['y'];
				$nextSection = $section;
				$nextSection['elements'] = array();
				if (isset($nextSection['height']))
					$nextSection['height'] -= $adjust;
			}
			$el['y'] -= $adjust;
			$nextSection['elements'][] = $el;
		}
		if (!empty($nextSection['elements']))
			$ret[] = $nextSection;
		return $ret;
	}

	public function renderSection($section, $data, $params=null)
	{
		$type = array_get_default($section, 'type', 'freeform');
		if(! is_array($params)) $params = array();

		if ($type == 'freeform' && empty($params['split_section'])) {
			$params['split_section'] = true;
			$subsections = $this->splitSection($section);
			foreach ($subsections as $s) {
				$this->renderSection($s, $data, $params);
			}
			return;
		}

		$context = array_get_default($params, 'context');
		$in_header_footer = ($context == 'header' ? 1 : ($context == 'footer' ? 2 : 0));

		if ($in_header_footer && $this->in_header_footer)
			return;

		if ($in_header_footer) {
			$savePage = $this->sectionPage;
			$saveY = $this->sectionY;
			$this->sectionPage = $this->pdf->getPage();
			if ($in_header_footer == 1)
				$this->sectionY = $this->marginTop;
			else
				$this->sectionY = $this->pageHeight - $this->marginBottom - $this->footerSize + $this->marginPadding;
			$this->in_header_footer = $in_header_footer;
		}

		$this->pdf->setPage($this->sectionPage);
		$this->pdf->setY($this->sectionY);
		$this->initMargins();

		if ($type == 'freeform') {
			$this->renderFreeForm($section, $data, $params);
		} else if ($type == 'table') {
			$this->renderTable($section, $data);
		}

		if ($in_header_footer) {
			$this->in_header_footer = 0;
			$this->sectionPage = $savePage;
			$this->sectionY = $saveY;
			$this->initMargins();
		} else {
			$h = ! empty($section['height']) ? $this->scaleSize($section['height']) : null;
			$this->updateSectionPosition($h);
		}
	}

	public function initMargins()
	{
		if($this->in_header_footer) {
			$this->pdf->SetMargins($this->marginLeft, $this->marginTop);
			$this->pdf->SetAutoPageBreak(false, $this->marginBottom);
		} else {
			$this->pdf->SetMargins($this->marginLeft, $this->marginTop + $this->headerSize);
			$this->pdf->SetAutoPageBreak(true, $this->marginBottom + $this->footerSize);
		}
	}

	public function updateSectionPosition($prev_height=null)
	{
		$oldPage = $this->sectionPage;
		$y = ($this->sectionPage-1) * $this->pageHeight + $this->sectionY + $prev_height;
		$curPage = $this->pdf->getPage();
		if(! isset($prev_height) || $curPage != $this->sectionPage) {
			$y2 = ($curPage-1) * $this->pageHeight + $this->pdf->getY();
			$y = max($y, $y2);
		}
		$this->sectionPage = (int)($y / $this->pageHeight) + 1;
		$this->sectionY = $y - ($this->sectionPage-1) * $this->pageHeight;
		while($this->pdf->getNumPages() < $this->sectionPage)
			$this->pdf->new_page();
	}

	public function checkIfWrapped($y=0, $h=null)
	{
		$p = $this->pdf->getPage();
		if($p > $this->sectionPage) {
			$oldp = $this->sectionPage;
			$this->sectionPage = $p;
			$dy = $this->marginTop + $this->headerSize - ($this->pageHeight - $this->sectionY - $this->pdf->getBreakMargin());
			$dy -= ($this->pageHeight - $this->pdf->getBreakMargin() - $this->marginTop - $this->headerSize) * ($p - $oldp - 1);
			if(isset($h)) {
				$dy2 = $this->pdf->getY() - $y - $h;
				$dy = max($dy, $dy2);
			}
			$this->sectionY = $dy;
		}
	}

	public function testWrapSection($need_height, $y=null)
	{
		if(! isset($y))
			$y = $this->pdf->getY();
		$sp = $this->pdf->getPage();
		if($y + $need_height > $this->pageHeight - $this->pdf->getBreakMargin()) {
			$this->sectionPage = $sp + 1;
			$this->sectionY = $this->marginTop + $this->headerSize;
			while($this->pdf->getNumPages() < $this->sectionPage)
				$this->pdf->new_page();
			$this->pdf->setPage($this->sectionPage);
			$this->pdf->setY($this->sectionY);
		}
	}

	public function adjustSectionHeight(&$section, $data)
	{
		if(!is_array($section['elements'])) {
			return;
		}

		usort($section['elements'], array('PDFTemplatePrinter', 'sortByY'));
		if (! isset($section['height']))
			return;
		$maxY = 0;
		$maxYA = 0;
		$adjusts = array();
		foreach ($section['elements'] as $idx => $el) {
			$y = $this->scaleSize($el['y']);
			if(isset($el['height'])) {
				$h = $this->scaleSize($el['height']);
				$maxY = max($maxY, $y + $h);
			} else
				$h = 0;
			foreach($adjusts as $adj) {
				if($y >= $adj[0])
					$section['elements'][$idx]['y'] = $this->unscaleSize($y += $adj[2]);
			}
			if(! empty($el['adjust']) && ($el['type'] == 'text' || $el['type'] == 'field')) {
				$params = $this->prepareTextParams($el, $data);
				unset($params[1]['fill-height']);
				$newH = $this->pdf->CalcCellHeight($params[0], $params[1]);
				if($newH > $h) {
					$adjH = $newH - $h;
					$adjusts[] = array($y+$h, $h, $adjH);
					$h = $newH;
					$section['elements'][$idx]['height'] = $this->unscaleSize($h);
				}
			}
			$maxYA = max($maxYA, $y + $h);
		}
		if($maxYA > $maxY) {
			$sH = $this->scaleSize($section['height']);
			$newH = $sH + $maxYA - $maxY;
			$section['height'] = $this->unscaleSize($newH);
		}
	}

	protected function renderFreeForm(&$section, $data, $params=null)
	{
		if(empty($section['elements'])) return;
		$this->adjustSectionHeight($section, $data);

		if(! is_array($params)) $params = array();
		$no_split = ! empty($params['no_split']);
		if($this->in_header_footer)
			$no_split = false;

		$h = $this->scaleSize($section['height']);
		if($no_split)
			$this->testWrapSection($h);
		foreach ($section['elements'] as $el) {
			$this->renderElement($el, $data);
		}
	}

	protected function renderElement($el, $data)
	{
		switch (array_get_default($el, 'type')) {
			case 'image':
				$this->renderImage($el, $data);
				break;
			case 'text':
			case 'field':
				$this->renderText($el, $data);
				break;
			case 'signature':
				$this->renderSignature($el);
				break;
			case 'html':
				$this->renderHtml($el, $data);
				break;
			case 'line':
				$this->renderLine($el);
				break;
		}
	}

	protected function renderLine($line)
	{
		$x = $this->scaleSize($line['x']);
		$line_width = $this->scaleSize(array_get_default($line, 'line_width', $this->pdf->default_line_width));
		$y = $this->scaleSize($line['y']);
		$w = $this->scaleSize($line['width']);
		$color = $this->parseColor(array_get_default($line, 'color', '000'));
		$this->pdf->Rect($x + $this->marginLeft, $y + $this->sectionY, $w, $line_width, 'F', array(), $color);
		$this->checkIfWrapped($y, $line_width);
	}

	protected function parseColor($hex, $fore = true)
	{
		if (!is_string($hex) || !preg_match('/^\W*([0-9A-F]{3}([0-9A-F]{3})?)\W*$/i', $hex, $m))
			return $fore ? array(0, 0, 0) : null;
		if (strlen($m[1]) == 6) {
			return array(
				hexdec(substr($m[1], 0, 2)),
				hexdec(substr($m[1], 2, 2)),
				hexdec(substr($m[1], 4, 2)),
			);
		} else {
			return array(
				hexdec(substr($m[1], 0, 1)) * 17,
				hexdec(substr($m[1], 1, 1)) * 17,
				hexdec(substr($m[1], 2, 1)) * 17,
			);
		}
	}

	protected function renderImage($img, $data)
	{
		$x = $this->scaleSize($img['x']);
		$y = $this->scaleSize($img['y']);
		$w = $this->scaleSize($img['width']);
		$h = $this->scaleSize($img['height']);
		$this->testWrapSection($this->scaleSize($img['height']));
		if (!empty($img['pdf_logo'])) {
			$info = $this->pdf->get_company_logo_info();
			$src = $info['path'];
		//print signature images
		} elseif (!empty($img['signature_image'])) {
			$fieldname = (!empty($img['field'])) ? $img['field'] : false;
			$src = $this->provider->get_signature_image_info($fieldname);
		} else {
			$src = $img['src'];
		}
		if (!empty($src)) $this->pdf->Image($src, $x + $this->marginLeft, $y + $this->sectionY, $w, $h);
		$this->checkIfWrapped($y);
	}

	protected function prepareTextParams($txt, $data)
	{
		$color = $this->parseColor(array_get_default($txt, 'color'));
		$background = $this->parseColor(array_get_default($txt, 'background'), false);
		$borderColor = $this->parseColor(array_get_default($txt, 'border_color'), false);
		$borderWidth = $this->scaleSize(array_get_default($txt, 'border_width', 0));

		$w = $this->scaleSize($txt['width']);
		$h = $this->scaleSize($txt['height']);
		if (isset($txt['field'])) {
			$text = array_get_default($data, $txt['field'], '');
		} else {
			$text = array_get_default($txt, 'text', '');
			$text = preg_replace_callback("~<\{field:(.*)\}>~U", function($match) use ($data) {
				return array_get_default($data, $match[1], '');
			}, $text);
			$text = preg_replace_callback("~<\{input(!?):([^\\|]+)(\\|(.*))?\}>~U", function($match) {
				$key = $match[2];
				$value = [];
				if (is_array($this->inputs)) {
					foreach ($this->inputs as $input) {
						if (array_get_default($input, 'id') == $key) {
							$value = array_get_default($input, 'value', '');
						}
					}
				}
				return $value;
			}, $text);
		}
		$align = array_get_default($txt, 'align');
		if (is_array($text)) {
			list($text, $type) = $text;
			if ($type == 'currency' || $type == 'float')
				$align = array_get_default($txt, 'number_align', $align);
		}
		if(! $align) $align = 'left';
		$align = strtoupper(substr($align , 0, 1));

		return array(
			array(
				'text' => $text,
				'width' => $w,
			),
			array (
				'color' => $color,
				'background-color' => $background,
				'fill-height' => $h,
				'font-face' => self::map_font(array_get_default($txt, 'font', 'sans')),
				'font-size' => $this->fontSize(array_get_default($txt, 'size', 10)),
				'line-height' => array_get_default($txt, 'line_height'),
				'padding' => $this->pdf->lv('2pt'),
				'font-weight' => array_get_default($txt, 'bold') ? 'bold' : '',
				'font-style' => array_get_default($txt, 'italic') ? 'italic' : '',
				'border-width' => $borderWidth,
				'border-color' => $borderColor,
				'text-align' => $align,
				'link-url' => array_get_default($txt, 'url'),
				'border' => $borderWidth,
				'allow-page-break' => true,
			),
		);
	}

	protected function renderSignature($sig) {
		$x = $this->scaleSize($sig['x']) + $this->marginLeft;
		$y = $this->scaleSize($sig['y']);
		$w = $this->scaleSize($sig['width']);
		$h = $this->scaleSize($sig['height']);
		$this->testWrapSection($h);
		$pageNum = $this->pdf->getPage();

		$this->signatureBlocks[] = [
			'x' => $x,
			'y' => $y + $this->sectionY,
			'page' => $pageNum,
			'signer_ref' => $sig['signer_ref'],
		];
	}

	protected function renderText($txt, $data)
	{
		if (!empty($txt['page_break'])) {
			$sp = $this->pdf->getPage();
			$this->sectionPage = $sp + 1;
			$this->sectionY = $this->marginTop + $this->headerSize;
			while($this->pdf->getNumPages() < $this->sectionPage)
				$this->pdf->new_page();
			$this->pdf->setPage($this->sectionPage);
			$this->pdf->setY($this->sectionY);
		}

		$params = $this->prepareTextParams($txt, $data);
		$x = $this->scaleSize($txt['x']);
		$y = $this->scaleSize($txt['y']);
		$h = $this->scaleSize($txt['height']);
		$this->pdf->setXY($x + $this->marginLeft, $y + $this->sectionY);
		$p1 = $this->pdf->getPage();

		$this->pdf->CellRow(
			array(
				$params[0],
			),
			$params[1]
		);

		//$this->pdf->MultiCell($w, $h, $text, 0, 'L', $background ? 1 : 0, 2);
		$this->checkIfWrapped($y, $h);
	}

	protected function renderHtml($html, $data)
	{
		if (!empty($html['page_break'])) {
			$sp = $this->pdf->getPage();
			$this->sectionPage = $sp + 1;
			$this->sectionY = $this->marginTop + $this->headerSize;
			while($this->pdf->getNumPages() < $this->sectionPage)
				$this->pdf->new_page();
			$this->pdf->setPage($this->sectionPage);
			$this->pdf->setY($this->sectionY);
		}

		$x = $this->scaleSize($html['x']);
		$y = $this->scaleSize($html['y']);
		$h = $this->scaleSize($html['height']);
		$this->pdf->setXY($x + $this->marginLeft, $y + $this->sectionY);

		if (isset($html['field'])) {
			$value = array_get_default($data, $html['field'], '');
		} else {
			$value = '';
		}

		$this->pdf->writeHTML($value, true, false, true, false, '');
		$this->checkIfWrapped($y, $h);
	}

	function renderTable(&$section, $data)
	{
		$cols = array();
		$empty = array();
		if (empty($section['columns']))
			return;
		$offsetOrig = array_get_default($section, 'offset', 0);
		$offset = $this->scaleSize($offsetOrig);
		$this->pdf->SetMargins($this->marginLeft + $offset, $this->marginTop + $this->headerSize, $this->marginLeft);
		$colwidths = $this->rescaleColumns($section['columns'], $offsetOrig);
		foreach ($section['columns'] as $idx => $col) {
			$align = array_get_default($col, 'align', 'left');
			$header_align = array_get_default($col, 'header_align', $align);
			$cols[$col['field']] = array(
				'title' => $col['label'],
				'text-align' => strtoupper(substr($align, 0, 1)),
				'header-text-align' => strtoupper(substr($header_align, 0, 1)),
				'width' => $this->scaleSize($colwidths[$idx])+2,
			);
		}
		$bwidth = $this->scaleSize(array_get_default($section, 'cell_border_width', 0));
		$opts = array(
			'border' => $bwidth ? 1 : 0,
			'border-width' => $bwidth,
			'border-color' => array_get_default($section, 'cell_border_color', '000000'),
			'alt-background-color' => $this->parseColor(array_get_default($section, 'even_background', 'ECECEC')),
			'background-color' => $this->parseColor(array_get_default($section, 'odd_background', 'FFFFFF')),
			'color' => $this->parseColor(array_get_default($section, 'color', '000000'), true),
			'font-face' => self::map_font(array_get_default($section, 'cell_font', 'sans')),
			'font-size' => $this->fontSize(array_get_default($section, 'cell_font_size', 11)),
			'padding' => 2,
		);

		$this->pdf->setXY($this->marginLeft + $offset, $this->sectionY);

		$bwidth = $this->scaleSize(array_get_default($section, 'header_border_width', $this->pdf->default_line_width));
		$header = array(
			'border' => $bwidth ? 1 : 0,
			'border-width' => $bwidth,
			'border-color' => array_get_default($section, 'header_border_color', '000000'),
			'font-face' => self::map_font(array_get_default($section, 'header_font', 'sans')),
			'font-size' => $this->fontSize(array_get_default($section, 'header_font_size', 11)),
			'font-weight' => array_get_default($section, 'header_bold', true) ? 'bold' : 'normal',
			'background-color' => $this->parseColor(array_get_default($section, 'header_background', 'C8C8C8')),
			'color' => $this->parseColor(array_get_default($section, 'header_color', '000000'), true),
			'padding' => 2,
		);

		$this->pdf->DrawTable($empty, $cols, '', $header, $opts);
		$bwidth = $this->pdf->lv($bwidth);
		$this->pdf->setY($this->pdf->getY() + $bwidth / 2);
		$this->pdf->DrawTable($data, $cols, '', null, $opts);
		$this->pdf->SetMargins($this->marginLeft, $this->marginTop + $this->headerSize, $this->marginLeft);
	}

	static function map_font($f) {
		if ($f == 'sans')
			return 'freesans';
		if ($f == 'serif')
			return 'freeserif';
		return $f;
	}

	function rescaleColumns($cols, $offset)
	{
		$fullWidth = 800;
		$target = $fullWidth - $offset;
		$scale = $fullWidth / $target;
		$tot = 0;
		$ws = array();
		foreach($cols as $idx => $c) {
			$tot += $ws[$idx] = $c['width'];
		}
		if($tot > $target) {
			foreach($cols as $idx => $c) {
				$ws[$idx] = $ws[$idx] * $target / $tot;
			}
		}
		foreach($cols as $idx => $c) {
			$ws[$idx] = $ws[$idx] * $scale;
		}
		return $ws;
	}


	function scaleSize($s)
	{
		$s = (float)$s / 8.0;
		return $this->pdf->lv($s . '%');
	}

	function fontSize($s) {
		return $this->scaleSize($s * 800 / 720 * 0.9);
	}

	function unscaleSize($s)
	{
		$m = $this->pdf->getMargins();
		$maxw = $this->pdf->getPageWidth() - $m['right'] - $m['left'];
		$k = $this->pdf->getScaleFactor();
		return $s * $k * 800.0 / $maxw;
	}

	function adjustPage($y, $h)
	{
		$pageStart = floor($y / $this->pageHeight) + 1;
		$hh = $this->in_header_footer ? 0 : $this->marginTop;
		$ypos = $y - ($pageStart * $this->pageHeight) + $this->pageHeight + $hh;
		if($this->in_header_footer)
			$pageEnd = $pageStart;
		else {
			$pageEnd = floor(($y + $h) / $this->pageHeight) + 1;
			if($pageEnd > $pageStart) {
				$pageEnd = $pageStart + 1;
				$ypos = $hh;
			}
		}
		if ($this->in_header_footer != 2) {
			while ($this->pdf->getNumPages() < $pageEnd) {
				$this->pdf->new_page();
			}
		}
		$ret = array(
			$pageEnd,
			$ypos,
			$pageStart * $this->pageHeight,
		);
		if ($this->in_header_footer == 2) {
			$ret[1] += $this->pageHeight + $this->marginHeader + $this->marginFooter;
			$ret[0] --;
		}
		return $ret;

	}

	public static function sortByY($a, $b)
	{
		$d = $a['y'] - $b['y'];
		if ($d == 0) {
			if (!empty($a['adjust']))
				$d = 1;
			elseif (!empty($b['adjust']))
				$d = -1;
		}
		return $d == 0 ? 0 : ($d < 0 ? -1 : 1);
	}

	public function getFileName() {
		return $this->provider->getFileName();
	}

	public function getMassFileName() {
		return $this->provider->getMassFileName();
	}

	public static function start_signing() {
		$signer = new PDFSigner;
		$errors = $signer->loadInput($_REQUEST);
		if ($errors === true) {
			try {
				$signer->generateProject();
				echo '<script>window.AppBase.ui.PopupManager.currentModal().close();</script>';
				exit;
			} catch (Exception $e) {
				$errors = [$e->getMessage()];
			}
		}
		$_REQUEST['config'] = 1;
		return self::handle_request(true, $errors);
	}

	public static function handle_request($attach_note=true, $errors = [])
	{
		if (!empty($_REQUEST['config'])) {
			$GLOBALS['pageInstance']->set_partial_page();
			self::render_config($errors);
			return;
		}
		$tpl_id = array_get_default($_REQUEST, 'pdf_template');
		if ($tpl_id == '_') {
			$url = 'async.php?module=' . $_REQUEST['module'] . '&action=DetailView&format=pdf&record=' . $_REQUEST['record'];
			if (isset($_REQUEST['layout']))
				$url .= '&layout=' . $_REQUEST['layout'];
			header('Location: ' . $url);
			exit;
		}
		if ($tpl_id == '~') {
			$module = $_REQUEST['module'];
			$model = AppConfig::module_primary_bean($module);
			$mgr = AppConfig::setting("display.view.$model.tally_pdf_manager");
			if(! $mgr)
				$mgr = AppConfig::setting("display.view.$model.pdf_manager");
			if ($mgr) {
				require_once('include/Tally/TallyPDF.php');
				TallyPDFManager::handle_request();
				exit;
			}
		}
		$module = $_REQUEST['module'];
		$model = AppConfig::module_primary_bean($module);
		if (!$model) exit;
		if (!empty($_REQUEST['list_uids'])) {
			$records = explode(';', $_REQUEST['list_uids']);
		} else {
			$records = [ $_REQUEST['record'] ];
		}
		$pdf = null;
		if ($records === ['all']) {
			//$lq = ListQuery::quick_fetch_row
		}
		$mass = count($records) > 1;
		foreach ($records as $record) {
			$printer = new PDFTemplatePrinter($tpl_id, $record);
			$printer->useCompanyAddress(array_get_default($_REQUEST, 'use_address'));
			$pdf = $printer->render($pdf, $mass);
		}
		$pdf->filename = filename_safe_string($mass ? $printer->getMassFileName() : $printer->getFileName());

		$filename = $pdf->filename;
		$mergePDF = $printer->handle_merge($pdf);
		ob_clean();
		if ($attach_note) {
			$printer->create_note($mergePDF, $filename);
		}
		else {
			session_write_close(); // prevent session locking while file is generated
			$mergePDF->Output($pdf->filename, 'I');
			sugar_cleanup(true);
		}
	}

	public static function render_config($errors = [])
	{
		$formId = uniqid("printForm-");
		if (!empty($_REQUEST['list_uids'])) {
			$records = explode(';', $_REQUEST['list_uids']);
		} else {
			$records = [ $_REQUEST['record'] ];
		}
		$mass = $records == ['all'] || count($records) > 1;

		$fields = get_config_fields(!$mass, !$mass);
		$layout = self::get_config_layout($fields, $formId);
		$result = new RowResult;
		$result->row = array();
		foreach($fields as $k => $f) {
			if(isset($f['default']))
				$result->row[$k] = $f['default'];
			if(isset($f['default_id']))
				$result->row[$f['id_name']] = $f['default_id'];
		}
		$result->fields = $fields;
		if (!empty($errors)) {
			$messages = array_map(function($e) {
				return [$e, 'error'];
			}, $errors);
			$msgs =  display_flash_messages_default($messages, false);
			echo '<div class="card-outer form-outer default-form"><div class="card-body form-body"><div class="form-section">'
				. $msgs
				. '</div></div></div>'
			;
		}

		require_once('include/layout/forms/FormGenerator.php');
		$model = new ModelDef(AppConfig::module_primary_bean($_REQUEST['module']));
		$model->addDisplayFieldDefinitions($result->fields);
		$gen = FormGenerator::html_form($model, $layout, $formId);
		$gen->getFormObject()->max_ref_width = 50;
		$gen->getFormObject()->addHiddenFields(array('layout' => array_get_default($_REQUEST, 'layout')));
		$gen->renderForm($result);
		echo $gen->getResult();
		$gen->exportIncludes();
		$t = javascript_escape(translate('LBL_SELECT_PDF_FORM', 'app'));
		$GLOBALS['pageInstance']->add_js_literal("SUGAR.popups.initContent({title: '$t'});", null, LOAD_PRIORITY_FOOT);
		$limit = !empty($_REQUEST['list_uids']) ? 1 : 2;
		if (is_countable($fields['pdf_template']['options']) && count($fields['pdf_template']['options']) < $limit) {
			$GLOBALS['pageInstance']->add_js_literal(
				'var prepend = SUGAR.ui.getFormInput("' . $formId . '", "prepend_file"); var append = SUGAR.ui.getFormInput("' . $formId . '", "append_file"); var btn = document.getElementById( "printForm_email"); if (btn) btn.style.display =  "none"; if (prepend) prepend.setDisabled(true); if (append) append.setDisabled(true);',
				null, LOAD_PRIORITY_FOOT
			);
		}
		$GLOBALS['pageInstance']->add_js_include("modules/PDFTemplates/print_dialog.js");
		$GLOBALS['pageInstance']->add_js_literal(
			'AppBase.module("pdf-print-dialog").initialize("' . $formId . '")'
		);

	}

	public static function default_company_address() {
		$rs = AppConfig::db_all_rows('CompanyAddress');
		$val = null;
		foreach($rs as $a) {
			if(! isset($val) || $a['main'])
				$val = $a;
		}
		return $val;
	}

	protected static function get_config_layout($fields, $formId) {
		$layout = array(
			'type' => 'config',
			'editable' => true,
			'buttons_position' => 'bottom',
			'form_hidden_fields' => self::get_config_hidden(),
			'form_buttons' => self::get_config_buttons($formId),
			'current_tab' => 'general',
			'base_module' => 'PDFTemplates',
			'tabs' => array(
				array(
					'name' =>  'general',
					'label' => 'LBL_TAB_GENERAL',
					'icon' => 'icon-action',
				),
				array(
					'name' =>  'signers',
					'label' => 'LBL_TAB_SIGNERS',
					'icon' => 'icon-user',
				),
				array(
					'name' =>  'inputs',
					'label' => 'LBL_TAB_INPUTS',
					'icon' => 'icon-edit',
				),
			),
			'sections' => array(
				array(
					'tab' => 'general',
					'label_position' => 'top',
					'columns' => 1,
					'elements' => self::tab_fields($fields, 'general'),
				),
				array(
					'tab' => 'signers',
					'label_position' => 'top',
					'columns' => 1,
					'elements' => self::tab_fields($fields, 'signers'),
				),
				array(
					'tab' => 'inputs',
					'label_position' => 'top',
					'columns' => 1,
					'elements' => self::tab_fields($fields, 'inputs'),
				),
			),
		);
		return $layout;
	}

	protected static function tab_fields($fields, $tab) {
		$fields = array_filter(array_keys($fields), function($name) use ($fields, $tab) {
			return array_get_default($fields[$name], 'tab') === $tab;
		});
		return $fields;
	}

	protected static function get_config_hidden() {
		return array(
			'module' => $_REQUEST['module'],
			'action' => 'PrintTemplate',
			'record' => array_get_default($_REQUEST, 'record'),
			'list_uids' => array_get_default($_REQUEST, 'list_uids'),
		);
	}

	protected static function get_config_buttons($formId) {
		$buttons = array(
			'print' => array(
				'icon' => 'icon-print',
				'label' => translate('LBL_OPEN_AS_PDF', 'app'),
				'params' => array(),
				'async' => false,
				'align' => 'right',
				'target' => '_blank',
				'post_perform' => "SUGAR.popups.close()",
			),
			'email' => array(
				'icon' => 'theme-icon module-Emails',
				'label' => translate('LBL_EMAIL_AS_PDF', 'app'),
				'params' => array(
					'attach_to_email' => true,
				),
				'async' => false,
				'align' => 'right',
			),
			'sign' => array(
				'icon' => 'theme-icon module-SigningProjects',
				'label' => translate('LBL_START_SIGNING', 'app'),
				'params' => array(
					'start_signing' => true,
				),
				'perform' => 'SUGAR.ui.beforeSubmitForm(this.form); ; SUGAR.ui.PopupManager.lastOpened().fetchFormContent(this.form, {start_signing: true}, {url: "async.php"}, function() { debugger; setTimeout(function() {AppBase.module("pdf-print-dialog").initialize("' . $formId . '"); }, 100)}); return false;',
				'async' => false,
				'align' => 'right',
			)
		);
		if(!empty($_COOKIE['IOS_APP'])) {
			unset($buttons['print']['target']);
		}
		return $buttons;
	}

	function handle_merge($pdf)
	{
		$prepend = array();
		$append  = array();
		$path = AppConfig::upload_dir();

		$custom = $pdf->getProvider()->get_custom_merge();

		if(is_array($custom)) {
			if(isset($custom['prepend']))
				$prepend[] = $custom['prepend'];
			if(isset($custom['append']))
				$append[] = $custom['append'];
		}

		$meta = array(
			array('prepend_file_id', 'prepend'),
			array('append_file_id', 'append'),
		);

		foreach ($meta as $x) {
			list($f, $v) = $x;
			if (!empty($_REQUEST[$f])) {
				$note = ListQuery::quick_fetch_row('Note', $_REQUEST[$f]);
				if ($note) {
					if (substr(strtolower($note['filename']), -4) == '.pdf') {
						$fname = $path . '/' . $note['filename'];
						if (is_file($fname) && is_readable($fname)) {
							${$v}[] = $fname;
						}
					}
				}
			}
		}

		if (!$prepend && !$append)
			return $pdf;

		$files = array();
		if ($prepend)
			array_extend($files, $prepend);
		$temp = tempnam(sys_get_temp_dir(), 'PDF');
		$pdf->Output($temp, 'F');
		$files[] = $temp;
		if ($append)
			array_extend($files, $append);

		$newPDF = new Fpdi;
		foreach($files as $file) {
			try {
				$pagecount = $newPDF->setSourceFile($file);
			} catch (Exception $e) {
				die('<b>' . $file . '</b>: ' . $e->getMessage());
			}
            for ($i = 1; $i <= $pagecount; $i++) {
                 $tplidx = $newPDF->importPage($i);
                 $s = $newPDF->getTemplatesize($tplidx);
                 $newPDF->AddPage($s['width'] > $s['height'] ? 'L' : 'P', array($s['width'], $s['height']));
                 $newPDF->useTemplate($tplidx);
            }
        }

		return $newPDF;

	}

	function create_note($pdf, $filename, $redirect = true, $props=null)
	{
		$focus = $this->provider->getRow();
		$module = $focus->getModuleDir();
		$content = $pdf->Output($filename, 'S');

		require_once('include/upload_file.php');
		$upload_file = new UploadFile();
		$upload_file->set_for_soap($filename, $content);

		$note = RowUpdate::blank_for_model('Note');
		$note_data = array(
			'name' => $this->provider->getNoteTitle(),
			'filename' => $upload_file->create_stored_filename(),
			'assigned_user_id' => AppConfig::current_user_id(),
			'parent_type' => $module,
			'parent_id' => $focus->getField('id'),
			'file_mime_type' => 'application/pdf',
		);
		$cid = $focus->getField('billing_contact_id');
		if(! $cid) $cid = $focus->getField('supplier_contact_id');
		$note_data['contact_id'] = $cid;
		$aid = $focus->getField('billing_account_id');
		if(! $aid) $aid = $focus->getField('supplier_id');
		if($module == 'Payments' || $module == 'PaymentsOut') {
			$aid = $focus->getField('account_id');
		}
		$note_data['account_id'] = $aid;
		if($props) array_extend($note_data, $props);
		$note->set($note_data);
		$note->save();
		$upload_file->final_move($note->getPrimaryKeyValue());

		if (!$redirect)
			return $note->getPrimaryKeyValue();

		if(empty($_REQUEST['attach_to_email']) || $_REQUEST['attach_to_email'] == 'false') {
			$path = 'download.php?type=Notes&id=' . $note->getPrimaryKeyValue();
			header('Location: ' . $path);
		} else {
			$rel_mod = $module;
			$rel_id = $focus->getField('id');
			$rel_name = $focus->getField('name');
			$cbillto = null;
			if ($note->getField('contact_id')) {
				require_bean('Contact');
				$cbillto = new Contact;
				$cbillto->retrieve($note->getField('contact_id'), false);
			} elseif ($note->getField('account_id')) {
				require_bean('Account');
				$cbillto = new Account;
				$cbillto->retrieve($note->getField('account_id'), false);
			}
			if($rel_mod == 'Payments' || $rel_mod == 'PaymentsOut') {
				$rel_mod = 'Accounts';
				$rel_id = $note->getField('account_id');
				$cbillto = new Account;
				if($cbillto->retrieve($rel_id))
					$rel_name = $cbillto->name;
			}
			$note_id = $note->getPrimaryKeyValue();
			$label = urlencode($this->provider->getEmailTitle());
			$note = "reattach_note_id={$note_id}&relabel_note={$label}";
			$parent = "parent_type=$rel_mod&parent_id={$rel_id}&parent_name=" . urlencode($rel_name);
			$id = $focus->getField('id');
			$return = "return_module=$module&return_action=DetailView&return_record={$id}";
			if ($cbillto) {
				$contact = "&contact_name=". urlencode($cbillto->name) . "&to_email_addrs={$cbillto->email1}";
				if ($cbillto->module_dir == 'Contacts') {
					$contact .= "&contact_id={$cbillto->id}";
				}
			} else {
				$contact = '';
			}
			header("Location: index.php?module=Emails&action=EditView&type=out&{$note}&{$parent}{$contact}&{$return}");
		}
		sugar_cleanup(true);
	}

	static function listupdate_perform($mu, $handler_name, $fmt, $listResult, $uids, $no_edit) {
		global $pageInstance;
		$_REQUEST['list_uids'] = join(';', $uids);
		self::handle_request();
		echo $pageInstance->format_includes('body');
		echo $pageInstance->format_includes('foot');
		echo $pageInstance->format_includes('end');
		exit;
	}
}

