<?php

require_once 'modules/PDFTemplates/PDFTemplateProvider.php';
require_once 'modules/PDFTemplates/utils.php';

abstract class PDFTemplateProviderBase implements PDFTemplateProvider
{
	static protected $field_types = array(
		'varchar', 'text',
		'int', 'double', 'bool', 'email', 'enum', 'multienum',
		'date', 'url', 'ref', 'currency', 'name',
		'phone', 'fax', 'datetime', 'item_number', 'status', 'sequence_number'
	);

	protected $header_data = [];
	public $layout;
	public $row;
	public $groups;
	public $gtotal;
	public $data = [];
	public $display_name;


	protected function loadHeaderFields(ListQuery $lq)
	{
		require_once 'include/layout/FieldFormatter.php';
		$fmt = new FieldFormatter('pdf');
		$fields = is_array($this->row->fields) ? $this->row->fields : [];
		$row = is_array($this->row->row) ? $this->row->row : [];
		$formatted = $fmt->formatRow($fields, $row);


		$this->header_data = [];
		$fs = $lq->base_model->getFieldDefinitions();
		foreach ($fs as $fk => $f) {
			if (isset($f['vname']) && in_array($f['type'], self::$field_types)) {
				if (isset($formatted[$fk]))
					$this->header_data['_' . $f['name']] = $formatted[$fk];
				elseif (isset($formatted['_' . $fk]))
					$this->header_data['_' . $f['name']] = $formatted['_' . $fk];
			}
		}

		$this->header_data['__page_number'] = 1;
		$this->header_data['__total_pages'] = '{{nb}}';

		require_once('include/Tally/TallyUpdate.php');
		$tally = TallyUpdate::for_result($this->row);
		if ($tally instanceof TallyUpdate) {
			$this->groups = $tally->getGroups();
			$this->gtotal = $tally->getTotals();
			$this->gtotal['id'] = 'GRANDTOTAL';
		}

		$cls = get_class($this);
		foreach ($cls::get_extra_sources() as $s) {
			$model = new ModelDef($s['model']);
			$fs = $model->getFieldDefinitions();
			foreach ($fs as $fk => $f) {
				$data_field = $s['prefix'] . '.' . $fk;
				if (isset($f['vname']) && in_array($f['type'], self::$field_types) && $f['type'] != 'ref') {
					if (isset($formatted[$data_field])) {
						$this->header_data[$data_field] = $formatted[$data_field];
					}
				}
			}
		}
	}

	public static function get_extra_sources() {
		return array();
	}

	protected function injectProductInfo()
	{
		require_once 'include/layout/FieldFormatter.php';
		$fmt = new FieldFormatter('pdf');
		$lq = new ListQuery('Product');
		$lq->addColumnFields();
		$lq->addField('weight');
		foreach ($this->groups as &$grp) {
			foreach ($grp['lines'] as &$line) {
				if ($line['related_type'] == 'ProductCatalog' && $line['related_id']) {
					$prod = $lq->queryRecord($line['related_id']);
					$formatted = $fmt->formatRow($lq->getFields(), $prod->row);
					$line['vendor_part_no'] = $formatted['vendor_part_no'];
					$line['weight'] = $formatted['weight'];
					foreach ($lq->getFields() as $k => $f) {
						if (isset($f['vname']) && in_array($f['type'], self::$field_types)) {
							if (!empty($f['created_by_module_designer'])
								|| array_get_default($f['source'], 'type') == 'custom_field') {
									$line[$k] = array_get_default($prod->row, $k);
							}
						}
					}
				}
			}
		}
	}

	public function initialize($layout, ListQuery $lq)
	{
		$this->layout = $layout;
		$cls = get_class($this);
		$sources = $cls::get_extra_sources();
		foreach ($sources as $s) {
			$model = new ModelDef($s['model']);
			$fs = $model->getFieldDefinitions();
			foreach ($fs as $fk => $f) {
				$data_field = $s['prefix'] . '.' . $fk;
				if (isset($f['vname']) && in_array($f['type'], self::$field_types) && $f['type'] != 'ref') {
					$lq->addField($data_field, $data_field, false, false, true);
				}
			}
		}
		$this->row = $lq->runQuerySingle();
		$lq->populateSecondary($this->row, false);
		$this->loadHeaderFields($lq);
	}

	function get_field($name) {
		if($this->row) return $this->row->getField($name);
	}

	function get_custom_merge() {
	}

	public static function get_template_skeleton($mod, $type)
	{
		if ($type == 'basic') {
			return self::get_basic_template_skeleton($mod);
		}
		$info = AppConfig::setting("modinfo.by_name.$mod.pdf_templates.provider");
		$file = $info['file'];
		$cls = $info['class'];
		require_once $file;
		$layout = $cls::template_skeleton($type);
		initialize_page_header_footer($layout);
		return $layout;
	}

	public static function get_predefined_signers_list($mod, $type, $record_id = null) {
		$signers = [];
		$modelName = AppConfig::module_primary_bean($mod);
		if (!$modelName) {
			return $signers;
		}
		$model = new ModelDef($modelName);
		$fields = $model->getFieldDefinitions();
		$fields = array_filter($fields, function($f) {
			if ($f['type'] === 'ref') {
				if (isset($f['from_app_field']) && $f['from_app_field'] !== 'app.assigned_user') {
					return false;
				}
				$ref_model = array_get_default($f, 'bean_name');
				if ($ref_model == 'User' || $ref_model == 'Contact') {
					return true;
				}
			}
		});
		$record = null;
		if ($record_id) {
			$record = ListQuery::quick_fetch_row($modelName, $record_id, array_keys($fields));
		}
		foreach ($fields as $f) {
			$ref_model = array_get_default($f, 'bean_name');
			$value =
				$record
				? ' (' . array_get_default($record, $f['name'], '') . ')'
				: ''
			;
			$signers[] = [
				'key' => $f['name'],
				'value' => [
					'label' => translate($f['vname'], array_get_default($f, 'vname_module', $mod)) . $value,
					'icon' => $ref_model === 'User' ? 'theme-icon module-Users' : 'theme-icon module-Contacts',
					'type' => 'module_field',
				]
			];
		}
		if ($type == 'basic') {
			return $signers;
		}
		$info = AppConfig::setting("modinfo.by_name.$mod.pdf_templates.provider");
		if (!$info) {
			return $signers;
		}
		$file = $info['file'];
		$cls = $info['class'];
		require_once $file;
		if (!method_exists($cls, 'get_predefined_signers')) {
			return $signers;
		}
		$signers = array_merge($signers, $cls::get_predefined_signers($mod, $type));
		return $signers;
	}

	public static function get_template_types($mod)
	{
		$ret = array();
		$info = AppConfig::setting("modinfo.by_name.$mod.pdf_templates.provider");
		if ($info) {
			$file = $info['file'];
			$cls = $info['class'];
			require_once $file;
			$ret = $cls::template_types();
		}
		if (AppConfig::setting("modinfo.by_name.$mod.basic_pdf_template")) {
			$ret['basic'] = translate('LBL_BASIC_TEMPLATE', 'PDFTemplates');
		}
		return $ret;
	}

	public static function create_provider($mod, $type)
	{
		if (AppConfig::setting("modinfo.by_name.$mod.basic_pdf_template") && $type == 'basic') {
			return new PDFTemplateBasicProvider($mod);
		}
		$info = AppConfig::setting("modinfo.by_name.$mod.pdf_templates.provider");
		$file = $info['file'];
		$cls = $info['class'];
		require_once $file;
		return $cls::create($type);
	}

	protected function findSection($name)
	{
		foreach ($this->layout['sections'] as $s) {
			if ($s['name'] == $name)
				return $s;
		}
		return array();
	}

	public function getSignedDocumentTitle() {
		return $this->getNoteTitle();
	}

	public static function get_sample_data($mod, $type)
	{
		if (AppConfig::setting("modinfo.by_name.$mod.basic_pdf_template") && $type == 'basic') {
			$data = self::get_basic_sample_data($mod);
			$header_data = self::get_page_header_footer_data();
			foreach ($header_data as $k => $v) {
				if (strpos($k, '__') === 0) {
					$data['header'][$k] = $v;
				}
			}
			return $data;
		}
		$info = AppConfig::setting("modinfo.by_name.$mod.pdf_templates.provider");
		$file = $info['file'];
		$cls = $info['class'];
		require_once $file;
		$data =  $cls::template_sample_data($type);

		foreach ($cls::get_extra_sources() as $s) {
			$model = new ModelDef($s['model']);
			$fs = $model->getFieldDefinitions();
			foreach ($fs as $fn => $f) {
				if (isset($f['vname']) && in_array($f['type'], self::$field_types) && $f['type'] != 'ref') {
					$data['header'][$s['prefix'] . '.' . $fn] = array(
						'label' => translate($f['vname'], $s['module']),
						'value' => translate($f['vname'], $s['module']),
					);
				}
			}
		}
		$header_data = array_get_default($data, 'header', array());
		$header_data = array_merge(self::get_page_header_footer_data(), $header_data);
		$data['page_header'] = $header_data;
		foreach ($header_data as $k => $v) {
			if (strpos($k, '__') === 0) {
				$data['header'][$k] = $v;
			}
		}
		return $data;
	}

	public static function get_basic_sample_data($mod)
	{
		return self::get_page_header_footer_data();
	}

	public static function get_page_header_footer_data($params=null)
	{
		if(! is_array($params)) $params = array();

		global $locale;
		require_bean('CompanyAddress');
		$data = array();
		$addr_id = array_get_default($params, 'address_id');
		$address_array = CompanyAddress::get_address_array($addr_id);
		$address_array['account_name'] = $address_array['name'];
		$tpl = $locale->getLocaleAddressTemplate('pdf', $address_array['address_format']);
		$address = $locale->getLocaleFormattedAddress($address_array, '', 'pdf', $tpl);

		$address = preg_replace('/\r\n|\n/', ' • ', $address);
		$phones = array();
		if (!empty($address_array['phone'])) {
			$phones[] = translate('LBL_PDF_PHONE', 'app') . ': ' . $address_array['phone'];
		}
		if (!empty($address_array['fax'])) {
			$phones[] = translate('LBL_PDF_FAX', 'app') . ': ' . $address_array['fax'];
		}
		if (!empty($phones)) {
			$address .= "\n" . implode(" • ", $phones);
		}
		$web = array();
		if (!empty($address_array['address_url']) && AppConfig::setting('company.pdf_show_header_url')) {
			$web[] = $address_array['address_url'];
		}
		if($web)
			$address .= "\n" . implode(" • ", $web);
		$data['__formatted_address'] = array(
			'label' => translate('LBL_FORMATTED_ADDRESS', 'PDFTemplates'),
			'value' => $address,
		);
		$data['__page_number'] = array(
			'label' => translate('LBL_PAGE_NUMBER', 'PDFTemplates'),
			'value' => '1',
		);
		$data['__total_pages'] = array(
			'label' => translate('LBL_TOTAL_PAGES', 'PDFTemplates'),
			'value' => '3',
		);
		return $data;
	}

	public static function get_basic_template_skeleton($mod)
	{
		return array(
			'sections' => array(
				array(
					'name' => 'header',
					'type' => 'freeform',
					'title' => translate('LBL_FORM_LAYOUT', 'PDFTemplates'),
					'height' => 400,
					'elements' => array(),
				),
			)
		);
	}

	public static function get_field_sources($mod, $type) {
		if ($type == 'basic')
			return array();
		$info = AppConfig::setting("modinfo.by_name.$mod.pdf_templates.provider");
		$file = $info['file'];
		$cls = $info['class'];
		require_once $file;

		$ret = array();
		foreach ($cls::get_extra_sources() as $s) {
			$ret[$s['prefix']] = translate($s['label'], $mod);
		}
		return $ret;
	}

	public function get_signature_image_info($fieldname)
	{
		global $current_language;
		$row =& $this->row;
		//multiple signature images are possible with additional user relations
		if (!empty($fieldname))
			$user_id = $row->getField($fieldname . '_id');
		else
			$user_id = $row->getField('assigned_user_id');
		$user = ListQuery::quick_fetch('User', $user_id, array('signature_image_filename'));
		//return false when no user is selected //TODO or checkbox is not checked - would need to be added to modules
		if (empty($user)) // || !$row->getField('use_signature_pictures'))
			return false;
		$user_signature = $user->getField('signature_image_filename');
		$signatures_folder = str_replace('//', '/', str_replace('{FILES}', AppConfig::files_dir(), AppConfig::setting("model.supported_types.image.upload_dir")));
		$default_image = "include/language/images/$current_language/default_signature.png";
		if (! file_exists($default_image))
			$default_image = "include/language/images/en_us/default_signature.png";
		$signature_info = !empty($user_signature) ? $signatures_folder . $user_signature : $default_image;
		return $signature_info;
	}

	public function getMassFileName()
	{
		$model = $this->row->getModelName();
		$module = AppConfig::module_for_model($model);
		$filename = translate('LBL_MODULE_TITLE', $module);
		$filename = preg_replace('~\.+$~', '', trim($filename));
		$filename .= '.pdf';
		return $filename;
	}

}

require_once 'modules/PDFTemplates/PDFTemplateBasicProvider.php';


