<?php

require_once 'modules/PaymentProcessors/PaymentProcessor.php';
require_once "modules/PaymentProcessors/inc/BaseRecord.php";
require_once __DIR__ . '/vendor/autoload.php';

// NOTE: there's autoloader sugar in composer.json for ./inc

use Psr\Http\Message\ServerRequestInterface as Request;
use Psr\Http\Message\ResponseInterface as Response;
use OneCRM\API\Errors\BadRequest;
use OneCRM\API\Errors\InternalError;

/*
 * Intent of this class is to provide the following Payment Processor-specific methods:
 *
 * 1 Manage Subscriptions, Contacts, Accounts, Plans, Addons, Coupons, ? Credits ?
 *   Other records are queried as needed (Payment Methods, Cards, Balances)
 * 2 Act on Payment Processor Webhooks for real-time updates
 * 3 Regular bulk sync/verification between PP and 1CRM
 *
 * Any methods which can be shared should be in PaymentProcessor class instead.
 *
 * ChargeBee is a Subscription Management provider so they're the Canonical Source of all records
 *
 * Instead of registering API endpoints to be serviced by this Module here, events are delegated
 * from the Subscriptions Module /subscription/events endpoint.
 *
 * For this class to function a PaymentProcessor with name matching this Class MUST be created
 * and configured to the correct API credentials for both the Webhooks and the PP's HTTP API
 *
 * Debugging Chargebee Webhooks:
 *   Stylus Host: *-test.chargebee.com
 *   #webhooks_update > div:nth-child(8) > div{ position: fixed; top:4em; left:2em;z-index:1;}
 *   #responseContent {max-height:initial;}
 *
 * NOTE: This information is only recorded for debugging / forensics reasons:
 *   Chargebee internal ids are like '3Nl7oGgQqGFprj9'
 *   '3Nl7oGgQqG' portion is common to all records for the same customer
 *   Subscription id and Customer id are identical
 *   Invoice ids are serial but lineItems are of format `li_3Nl7oGgQqGXXXXXX`
 */


class PaymentProcessorChargeBee extends PaymentProcessor implements PaymentProcessorInterface
{
    public static $pp_sync_records = [
        'Customer' => true,
        'Subscription' => true,
        //'Plan' => true,
        //"Addon" => true,
        //'Coupon' => true,
    ];
    public $name = 'ChargeBee';
    /**
     * API Endpoint / Event objects:
     */
    public $params = null;
    public $response = null;
    public $request = null;
    public $payload = null;
    public $content = null;
    /**
     * Shared / Cached objects:
     */
    public $subscription = null;
    /** @var SubscriptionRecord $subscription */
    public $customer = null;
    /** @var CustomerRecord $customer */
    public $plan = null;
    /** @var PlanRecord $plan */
    public $addon = null;
    /** @var AddonRecord $addon */
    public $coupon = null;
    /** @var CouponRecord $coupon */
    public $optionSet = null;
    /** @var OptionSetRecord $optionSet */
    public $invoice = null;
    /** @var InvoiceRecord $invoice */
    public $transaction = null;
    /** @var TransactionRecord $transaction */


    /**
     * Process an event from the Payment Processor - API or Webhook.
     * @param array $params the request parameters
     * @param Response $response
     * @param Request $request
     * @return bool             always return true unless it throws
     * @throws BadRequest       if handler returns an error processing the request
     * @throws InternalError    for a webhook without a defined handler
     */
    public function processEvent(array $params, Response $response, Request $request)
	{
        // simplify the call model a ton since these are available via the singleton:
        $this->params = $params;
        $this->response = $response;
        $this->request = $request;

        $payload = $this->payload = $request->getParsedBody();
		if (PaymentProcessor::isEventProcessed($this, $payload['id'])) {
			return true;
		}
		$content = $this->content = $payload['content'];

        $handler = 'handle_' . $payload['event_type'];
        if (is_callable([$this, $handler])) {
            if ($this->$handler($content)) {
				PaymentProcessor::markEventProcessed($this, $payload['id']);
				$hooks = AppConfig::setting('subscriptions.hooks.event');
				if (!empty($hooks)) {
					$eventData = $this->extractEventData($payload['event_type'], $content);
					foreach ($hooks as $hook) {
						call_hook_function($hook, null, [$this->name, $payload['event_type'], $eventData]);
					}
				}
                return true;
            } else {
                throw new BadRequest('Bad data');
            }
        } else {
            return true;
        }
    }

    /*****************************************************************
     *
     * remote reporting event handlers initiated by Payment Processor:
     *
     ****************************************************************/

    /**
     * Customer/Account notifications:
     * try to update a matching Account first.
     * if no match, create a new Contact.
     * see Subscription::find_account() for matching rules
     *
     * @param array $content
     * @return bool
     */
    public function handle_customer_created($content)
    {
        $this->customer = new \OneCRM\PaymentProcessors\CustomerRecord($content['customer'], 'native');

        return $this->customer->onecrm_save(function (RowUpdate $ru) {
            if (!$ru->new_record) {
                $ru->set('personal_info_source', $this->name);
            }
        });
    }

    public function handle_customer_changed($content)
    {
        return $this->handle_customer_created($content);
    }

    public function handle_customer_deleted($content)
    {
        return $this->handle_customer_created($content);
    }


    /**
     * plan management notifications:
     */

    public function handle_plan_created($content)
    {
        $this->plan = new \OneCRM\PaymentProcessors\PlanRecord($content['plan'], 'native');
        return $this->plan->onecrm_save();
    }

    public function handle_plan_updated($content)
    {
        return $this->handle_plan_created($content);
    }

    public function handle_plan_deleted($content)
    {
        return $this->handle_plan_created($content);
    }

    /**
     * addon management notifications:
     */
    public function handle_addon_created($content)
    {
        $this->addon = new \OneCRM\PaymentProcessors\AddonRecord($content['addon'], 'native');
        return $this->addon->onecrm_save();
    }

    public function handle_addon_updated($content)
    {
        return $this->handle_addon_created($content);
    }

    public function handle_addon_deleted($content)
    {
        return $this->handle_addon_created($content);
    }

    /**
     * coupon management notifications:
     */

    public function handle_coupon_created($content)
    {
        $this->coupon = new \OneCRM\PaymentProcessors\CouponRecord($content['coupon'], 'native');
        return $this->coupon->onecrm_save();
    }

    public function handle_coupon_updated($content)
    {
        return $this->handle_coupon_created($content);
    }

    public function handle_coupon_deleted($content)
    {
        return $this->handle_coupon_created($content);
    }

    /**
     * subscription management notifications:
     * pre-depends the Contact, Plan, Addons, Coupons being created
     * during creation both Subscription and Account must be created and linked to the Contact
     *
     * @param array $content
     * @return bool
     */
    public function handle_subscription_created($content)
    {
        if (isset($content['customer'])) $success=$this->handle_customer_created($content);

        $this->subscription = new \OneCRM\PaymentProcessors\SubscriptionRecord($content['subscription'], 'native');

        return $success && $this->subscription->onecrm_save();
    }

    /**
     * NOTE: shipping address is stored in the Billing Account linked to the Subscription.
     */

    public function handle_subscription_shipping_address_updated($content)
    {
        return true;  // handled by subscription/customer updates.
    }

    public function handle_subscription_started($content)
    {
        return $this->handle_subscription_created($content);
    }

    /* delegate to updated after triggering activation hook */
    public function handle_subscription_activated($content)
    {
        return $this->handle_subscription_created($content);
    }

    /* delegate the update notices here */
    public function handle_subscription_changed($content)
    {
        return $this->handle_subscription_created($content);
    }

    /* delegate to the activated */
    public function handle_subscription_reactivated($content)
    {
        return $this->handle_subscription_created($content);
    }

    /* state sanity check would be good here */
    public function handle_subscription_renewed($content)
    {
        return $this->handle_subscription_created($content);
    }

    public function handle_subscription_renewal_reminder($content)
    {
        return $this->handle_subscription_created($content);
    }

    /* delegate to cancellation_reminder */
    public function handle_subscription_cancellation_scheduled($content)
    {
        return $this->handle_subscription_created($content);
    }

    /* trigger cancellation hook then delegate to update */
    public function handle_subscription_cancelled($content)
    {
        return $this->handle_subscription_created($content);
    }

    /* set non-renewing if scheduled_date <= renewal_date */
    public function handle_subscription_cancellation_reminder($content)
    {
        return $this->handle_subscription_created($content);
    }

    /* sanity check then delegate to update */
    public function handle_subscription_deleted($content)
    {
        return $this->handle_subscription_created($content);
    }

    /* trigger hook then delegate to update */
    public function handle_subscription_paused($content)
    {
        return $this->handle_subscription_created($content);
    }

    /* delegate to started */
    public function handle_subscription_resumed($content)
    {
        return $this->handle_subscription_created($content);
    }


    /**
     * invoice management notifications:
     *
     * NOTE: Draft invoices are prepared as 'pending' invoices
     * NOTE: Invoices have sequential id but pending are like '3Nl7oGgQqGFprj9'
     * @throws InternalError
     */

    public function handle_invoice_created($content)
	{
		$sid = $content['invoice']['subscription_id'];
		if (!\Subscription::find_subscription(['pp_subs_id' => $sid, 'payment_processor_id' => $this->id])) {
				return true;
		}
		$this->invoice = new \OneCRM\PaymentProcessors\InvoiceRecord($content['invoice'], 'native');
		return $this->invoice->onecrm_save();
    }

    public function handle_invoice_updated($content)
    {
        return $this->handle_invoice_created($content);
    }

    public function handle_invoice_deleted($content)
    {
        return $this->handle_invoice_created($content);
    }

    public function handle_pending_invoice_created($content)
    {
        return $this->handle_invoice_created($content);
    }

    public function handle_pending_invoice_updated($content)
    {
        return $this->handle_invoice_created($content);
    }

    public function handle_invoice_generated($content)
    {
        return $this->handle_invoice_created($content);
    }


    /**
     * credit_note management notifications:
     */

    public function handle_credit_note_created($content)
    {
		$sid = $content['credit_note']['subscription_id'];
		if (!\Subscription::find_subscription(['pp_subs_id' => $sid, 'payment_processor_id' => $this->id])) {
				return true;
		}
        $this->creditnote = new \OneCRM\PaymentProcessors\CreditNoteRecord($content['credit_note'], 'native');
        return $this->creditnote->onecrm_save();
    }

    public function handle_credit_note_updated($content)
    {
        return $this->handle_credit_note_created($content);
    }

    public function handle_credit_note_deleted($content)
    {
        return $this->handle_credit_note_created($content);
    }


    /**
     * payment_source management notifications:
     */

    public function handle_payment_source_created($content)
    {
        return 'Not Implemented';
    }

    public function handle_payment_source_updated($content)
    {
        return 'Not Implemented';
    }

    public function handle_payment_source_deleted($content)
    {
        return 'Not Implemented';
    }


    /**
     * card management notifications:
     */

    public function handle_card_created($content)
    {
        return 'Not Implemented';
    }

    public function handle_card_updated($content)
    {
        return 'Not Implemented';
    }

    public function handle_card_expiry_reminder($content)
    {
        return 'Not Implemented';
    }

    public function handle_card_expired($content)
    {
        return 'Not Implemented';
    }

    public function handle_card_deleted($content)
    {
        return 'Not Implemented';
    }


    /**
     * transaction management notifications:
     * @throws InternalError
     */

    public function handle_transaction_created($content)
    {
        // insert/update the transaction:
        $this->transaction = new \OneCRM\PaymentProcessors\TransactionRecord($content['transaction'], 'native');
        return $this->transaction->onecrm_save();
    }

    public function handle_transaction_updated($content)
    {
        return $this->handle_transaction_created($content);
    }

    public function handle_transaction_deleted($content)
    {
        return $this->handle_transaction_created($content);
    }


    /**
     * payment / refund / credit management notifications:
     * @throws InternalError
     */

    public function handle_payment_initiated($content)
    {
        // sync related customer, subscription, invoice first if present:
        if (@$content['customer']) {
            $this->handle_customer_changed($content);
        }
        if (@$content['subscription']) {
            $this->handle_subscription_changed($content);
        }
        if (@$content['invoice']) {
            $this->handle_invoice_updated($content);
        }
        if (@$content['credit_note']) {
            $this->handle_credit_note_created($content);
        }
        // sync transaction, link invoice:
        if (@$content['transaction']) {
            $this->handle_transaction_created($content);
        }
        return true;
    }

    public function handle_payment_succeeded($content)
    {
        return $this->handle_payment_initiated($content);
    }

    public function handle_payment_failed($content)
	{
        if (!empty($content['subscription'])) {
            return $this->handle_subscription_changed($content);
        }
    }

    public function handle_refunded_initiated($content)
    {
        return $this->handle_payment_initiated($content);
    }

    public function handle_payment_refunded($content)
    {
        return $this->handle_payment_initiated($content);
    }

    public function handle_promotional_credits_added($content)
    {
        return "Not Implemented";
    }

    public function handle_promotional_credits_deducted($content)
    {
        return "Not Implemented";
    }

	public function getPaymentMethods($customerId) {
		$methods = \ChargeBee_PaymentSource::all(['customer_id[is]' => $customerId]);
		$ret = [];
		foreach ($methods as $method) {
			$method = $method->paymentSource()->getValues();
			$ret[] =  [
				'id' => $method['id'],
				'name' => 'Card ending with ' . $method['card']['last4'],
			];
		}
		return $ret;
	}

	public function getPaymentMethodsURL($customerId, $returnURL) {
		$result = \ChargeBee_HostedPage::managePaymentSources([
			'customer[id]' => $customerId,
			'redirect_url' => $returnURL,
			'cancel_url' => $returnURL,
		]);
		return $result->hostedPage()->url;
	}

    /**
     * order management notifications:
     */

    public function handle_order_created($content)
    {
        return 'Not Implemented';
    }

    public function handle_order_updated($content)
    {
        return 'Not Implemented';
    }

    public function handle_order_cancelled($content)
    {
        return 'Not Implemented';
    }

    public function handle_order_delivered($content)
    {
        return 'Not Implemented';
    }

    public function handle_order_returned($content)
    {
        return 'Not Implemented';
    }

    public function handle_order_ready_to_process($content)
    {
        return 'Not Implemented';
    }

    public function handle_order_ready_to_ship($content)
    {
        return 'Not Implemented';
    }




    /**************************************************************
     *
     * remote API methods initiated by 1CRM / Subscriptions API / Portal
     *
     **************************************************************/


    /**
     * Create a new customer at the Payment Processor.
     * input is 1CRM normal
     *
     * @param $customerData : accepts union of Account and Contact model fields and adds a new customer
     * @return bool
     * @throws BadRequest
     *
     * returns [ 'pp_customer_id' => $pp_customer_id, 'contact'=> $pp_customer_data ]
     * or false on error.*
     */
    public function addCustomer(array $customerData)
    {
        return $this->sync_record('Customer', $customerData);
    }

    /**
     * Update existing customwe at the Payment Processor
     * input is 1CRM normal
     *
     * @param $customerData : array 1CRM Customer model field format data
     * @return bool
     * @throws BadRequest
     * returns [ 'pp_customer_id' => $pp_customer_id, 'contact'=> $pp_customer_data ]
     * or false on error.
     */
    public function updateCustomer($customerData)
    {
        return $this->sync_record('Customer', $customerData);
    }

    /*
     * GDPR compliance - if customer requests to be deleted entirely.
     * use upstream GDPR compliance supression call
     * or replace all sensitive data with pseudorandom bytes
     * and remove all registered payment methods, if any
     *
     * @param $customerId : string pp_customer_id
     *
     * fail if there are active subscriptions or if there is a balance due.
     * WARN: do not rely on this failure mode, it is not guaranteed.
     *
     * on failure dispatch a system error event (with blame backtrace) and return false
     * otherwise return true.
     */

    public function suppressCustomer($customerId)
    {

    }

    /**
     * Create a new Subscription for a customer
     * Called when a Subscription transtions from Draft state
     *
     * @param $subscriptionData : array 1CRM Subscription model data
     * @return bool
     * @throws BadRequest
     */

    public function createSubscription($subscriptionData)
    {
        return $this->sync_record('Subscription', $subscriptionData);
    }

	public function convertAddons($addons) {
		$json = getJSONobj();
		if (is_string($addons)) {
			$addons = $json->decode($addons);
		}
		$pp_ids = array_keys($addons);
		$lq = new \ListQuery('Addon');
		$lq->addSimpleFilter('pp_addon_id', $pp_ids);
		$all_addons = $lq->fetchAllRows();
		$ret = [];
		foreach ($addons as $pp_id => $qty) {
			$addon_rows = array_filter($all_addons, function($row) use ($pp_id) {
				return $row['pp_addon_id'] == $pp_id;
			});
			$addon_row = current($addon_rows);
			if ($addon_row) {
				$ret[$addon_row['id']] = [
					'qty' => $qty,
				];
			}
		}
		return $ret;
	}


    /*
     * Update an existing Subscription for a customer
     *
     * @param $customerId : string pp_customer_id
     * @param $subscriptionData : array 1CRM Subscription model data
     * returns [ 'pp_customer_id' => $pp_customer_id, 'subscription' => $pp_subscription_data ]
     * NOTE: subscription is in 1CRM Subscription model format
     *
     * WARN: can throw if data is in unexpected format or invalid operation
     */

    public function updateSubscription($subscriptionData)
    {
        return $this->sync_record('Subscription', $subscriptionData);
    }

    /*
     * Cancel a Subscription for a customer
     *
	 * @param $subscriptionId : string pp_subscription_id
	 * @param $options : array with options
	 *
     */

    public function cancelSubscription($subscriptionId, $options)
	{
		$params = [];
		if (empty($options['end_of_term']))  {
			switch ($options['credits'] ?? null) {
			case 'none':
				$params['creditOptionForCurrentTermCharges'] = 'none';
				break;
			case 'prorate':
				$params['creditOptionForCurrentTermCharges'] = 'prorate';
				break;
			case 'full':
				$params['creditOptionForCurrentTermCharges'] = 'full';
				break;
			}
		} else {
			$params['endOfTerm'] = true;
		}
		try {
			ChargeBee_Subscription::cancel($subscriptionId, $params);
		} catch (Exception $e) {
			$GLOBALS['log']->error($e->getMessage());
			throw $e;
		}
		return true;
    }

    /*
     * Get a list of the contact's subscriptions
     *
     * @param $customerId : string pp_customer_id
     * @param $filter : array of desired states from Subscription->state
     */

    public function getSubscriptions($customerId, $state)
    {

    }

    /*
     * returns (string) unique Payment Processor ID
     */
    public function id()
    {
        return $this->id;
    }

    /***************************
     *
     * singleton implementation:
     *
     ***************************/


    /**
     * @return ChargeBee instance
     */
    public static final function instance()
    {
        static $single = null;

        if (is_null($single)) $single = NEW self();
        return $single;
    }

    private function __clone()
    {

    }

    public function __sleep()
    {

    }

    public function __wakeup()
    {

    }

    /**
     * ChargeBee constructor.
     * sets up ChargeBee API connector and shared processor_id field of related Model records
     * @param $name : Unique Name of a PaymentProcessor Model record
     */
    public function __construct()
    {
        parent::__construct($this->name);
        OneCRM\PaymentProcessors\ChargeBeeRecord::$processor_id = $this->id;
        if ($this->api_user) {
            // prepare ChargeBee API connection:
            ChargeBee_Environment::configure($this->api_user, $this->api_pass);
        }
	}


	public static function savePPConfig(&$updates) {
		if (!isset($updates['subscriptions.processor_type'])) {
			return; // saving was not initiated from "Subscriptions Settings" configurator screen
		}
		$pp = ListQuery::quick_fetch_key('PaymentProcessor', 'name', 'ChargeBee');

		$password = array_get_default($updates, 'subscriptions.chargebee.webhook_password');
		$user = array_get_default($updates, 'subscriptions.chargebee.webhook_user');
		$api_password = array_get_default($updates, 'subscriptions.chargebee.api_password');
		$api_user = array_get_default($updates, 'subscriptions.chargebee.api_user');
		$ssl = array_get_default($updates, 'subscriptions.chargebee.webhook_ssl');
		unset($updates['subscriptions.chargebee.api_password']);
		unset($updates['subscriptions.chargebee.api_user']);
		unset($updates['subscriptions.chargebee.webhook_password']);
		unset($updates['subscriptions.chargebee.webhook_user']);
		unset($updates['subscriptions.chargebee.webhook_ssl']);

		if (array_get_default($updates, 'subscriptions.processor_type') != 'ChargeBee') {
			// Payment processor was not set to "ChargeBee" – mark ChargeBee as non-default
			if ($pp) {
				$upd = RowUpdate::for_result($pp);
				$upd->set([
					'is_default' => 0,
					'active' => 'Inactive',
				]);
				$upd->save();
			}
			return;
		}
		if ($pp) {
			$upd = RowUpdate::for_result($pp);
		} else {
			$upd = RowUpdate::blank_for_model('PaymentProcessor');
			$upd->set([
				'name' => 'ChargeBee',
			]);
		}
		$upd->set([
			'is_default' => 1,
			'active' => 'Active',
			'api_user' => $api_user,
			'api_pass' => $api_password,
			'webhook_user' => $user,
			'webhook_ssl' => $ssl ? 1 : 0,
		]);
		if (strlen($password)) {
			$upd->set('webhook_pass', sha1($password));
		}
		$upd->save();
	}

	public static function loadPPConfig(&$cfg) {
		$pp = ListQuery::quick_fetch_key('PaymentProcessor', 'name', 'ChargeBee');
		if (!$pp) return;
		$cfg['subscriptions_chargebee_api_user'] = $pp->getField('api_user');
		$cfg['subscriptions_chargebee_api_password'] = $pp->getField('api_pass');
		$cfg['subscriptions_chargebee_webhook_user'] = $pp->getField('webhook_user');
		$cfg['subscriptions_chargebee_webhook_ssl'] = $pp->getField('webhook_ssl');
	}

	protected function extractEventDataCustomer($data) {
		return $data;
	}

	protected function extractEventDataSubscription($data) {
		return $data;
	}

	protected function extractEventDataTransaction($data) {
		return $data;
	}

	public function extractEventData($event, $data) {
		$resources = [
			'customer' => 'extractEventDataCustomer',
			'subscription' => 'extractEventDataSubscription',
			'transaction' => 'extractEventDataTransaction',
		];
		$ret = [];
		foreach ($resources as $r => $method) {
			if (isset($data[$r])) {
				$ret[$r] = $this->$method($data[$r]);
			}
		}
		return $ret;
	}

	public function formatForDisplay(array $native) {
		$format = AppConfig::setting('locale.defaults.date_format');
		$plan = ListQuery::quick_fetch_key('Plan', 'pp_plan_id', $native['plan_id']);
		$addons = [];
		if (!empty($native['addons'])) {
			foreach ($native['addons'] as $a) {
				$addon = ListQuery::quick_fetch_key('Addon', 'pp_addon_id', $a['id']);
				$addon_name = $addon ? $addon->getField('name') : '?????';
				$addons[] = [
					'name' => $addon_name,
					'quantity' => array_get_default($a, 'quantity', 1),
				];
			}
		}
		return [
			'next_billing_at' => date($format, $native['next_billing_at']),
			'current_term_start' => date($format, $native['current_term_start']),
			'current_term_end' => date($format, $native['current_term_end']),
			'due_invoices_count' => $native['due_invoices_count'],
			'mrr' => sprintf("%.2f", $native['mrr'] / 100),
			'addons' => $addons,
			'qty_label' => $plan ? $plan->getField('qty_label') : '',
			'quantity' => array_get_default($native, 'plan_quantity', 1),
		];
	}

}
