<?php

namespace OneCRM\PaymentProcessors;

/*
 * Framework for handling normalization, read/write, and synchronization of
 * individual Payment Processor records with 1CRM db Models
 *
 * BaseRecord fleshes out the Processor-specific shared methods.
 * Each Payment Processor data object should be extended from this class.
 *
 * WARN: we only support week, month, year as billing intervals
 *       See to_native() method if this needs to change.
 */

// FIXME: to_onccrm() assumption breaks Yen and other non-float currencies

use OneCRM\API\Errors\BadRequest;

class ChargeBeeRecord extends BaseRecord
{
    /** @var string $processor_id : MUST be defined here to share with child *Record classes */
    public static $processor_id = '';   // REQUIRED, shared with child classes.
    public $map_meta_data = [];         // ChargeBee 'metadata' fields for Plan, Addon, Coupon, Customer

    /**
     * looks up one or more objects at the Payment Processor
     * stores the result in $this->list and returns self
     *
     * @param mixed $filter : if string is passed, search for $this->name_id == $this->pp_id
     * @param boolean $to_onecrm : true to normalize and copy the fields to $this->onecrm
     * @return BaseRecord $this
     */

    public function pp_getter($filter, $to_onecrm = true)
    {
        if (!is_array($filter)) $filter = ['limit' => 1, $this->ppid_field.'[is]' => $filter];
        /** @var \ChargeBee_Addon $chargebeeModel */
        $chargebeeModel = '\\ChargeBee_' . $this->name;
        if (!method_exists($chargebeeModel,'all')) return $this;
        $all = $chargebeeModel::all($filter);
        $result = json_decode($all->toJson(), true);
        $this->list = array_column($result, strtolower($this->name));
        $this->reset($to_onecrm);
        return $this;
    }

    /**
     * Write the object back to the Payment Processor
     * @return array|bool PaymentProcessor's unique ID for the record
     * @throws BadRequest
     */
	public function pp_save(){
		$self = $this;
		$runHooks = function($type, &$native, $result = null) use($self) {
			$hooks = \AppConfig::setting("subscriptions.hooks.{$type}.{$self->name}");
			if (!empty($hooks)) {
				foreach ($hooks as $hook) {
					call_hook_function($hook, null, [$self, &$native, $result]);
				}
			}
		};

		/** @var \ChargeBee_Customer $chargebeeModel */
		$chargebeeModel = '\\ChargeBee_' . $this->name;

        $native =& $this->native;

		list ($create, $update) = $this->getSaveMethods($chargebeeModel, $native);

		if (! (\PaymentProcessorChargeBee::$pp_sync_records[$this->name] && is_callable($create))) {
            throw new BadRequest('Save ' . $this->name . ': Not Implemented');
        }

		if ($native[$this->ppid_field] ?? null){
			$runHooks('before_pp_save', $native);
			try {
				$res = $update($native[$this->ppid_field],$native);
			} catch (\Exception $e) {
				$runHooks('after_pp_save', $native, $e);
				throw $e;
			}
			$runHooks('after_pp_save', $native, $res);
        } else {
            unset($native[$this->ppid_field]);
			$runHooks('before_pp_save', $native);
			try {
				$res = $create($native);
			} catch (\Exception $e) { // could be Throwable, but we cannot guarantee we are running PHP 7
				$runHooks('after_pp_save', $native, $e);
				throw $e;
			}
			$runHooks('after_pp_save', $native, $res);
		}
        if (! $res) throw new BadRequest('Refused by Processor');
        $name=strtolower($this->name);
        if (is_callable([$res,$name]) && $res = $res->{$name}()){
            /** @var \ChargeBee_Customer $res */
            try {
                return [
                    'model' => $this->name,
                    'id' => $res->param($this->ppid_field),
                    'key' => $this->model_ppid_field,
                ];
            } catch (\Exception $e){
                throw new BadRequest($e->getMessage());
            }
        }
        return false;
    }

    /**
     * translates $this->native to $this->onecrm
     * @param $data : if not null then $data is translated to $this->onecrm instead
     * @return BaseRecord $this
     * @throws BadRequest
     */
    public function to_onecrm($data = null)
    {

        parent::to_onecrm($data);
        $onecrm = &$this->onecrm;
        $native = &$this->native;

        // bring in meta_data fields:
        if (!$data) $data = $this->native;
        foreach ($this->map_meta_data as $in => $out) {
            if (isset($data['meta_data'][$in]))
                $onecrm[$out] = $data['meta_data'][$in];
        }

        // date fields from UNIX timestamp to MySQL GMT timestamp:
		$dates = ['date_entered', 'date_modified', 'start_date', 'current_period_end',
			'current_period_start', 'next_billing_date',
            'due_date', 'payment_date', 'date_from', 'date_to'];
        foreach ($dates as $date){
            if (!isset($onecrm[$date])) continue;
            $onecrm[$date] = gmdate('Y-m-d H:i:s', $onecrm[$date]);
            if (!$onecrm[$date]) unset($onecrm[$date]);
        }
        $onecrm['deleted'] = 0;
        if (array_get_default($data, 'deleted') || array_get_default($data, 'status') === 'deleted') {
            $onecrm['deleted'] = 1;
        }
        // convert financial data fields as required:
        if ($native['currency_code']??null) {
            $currency_id = \PaymentProcessor::currencyByIso4217($native['currency_code']);
            if (!$currency_id) throw new BadRequest('Unknown ISO4217 code: ' . $native['currency_code']);
            $this->onecrm['currency_id'] = $currency_id;
        }

        // convert these price fields:
        $prices = ['amount', 'amount_due', 'discount_amount', 'total', 'subtotal',
            'setup', 'price', 'unit_price', 'ext_price', 'mrr'];
        foreach ($prices as $price) {
            if (! isset($onecrm[$price])) continue;
            $value = $onecrm[$price] / 100;
            $onecrm[$price] = (string) $value;
        }

        // assume tiers are already in base currency:
        if (isset($onecrm['pricing_tiers'])) {
            $tmp = $onecrm['pricing_tiers'];
            foreach ($tmp as &$tier) {
                // prevent json_encode from turning 0.15 into 0.14999999999999
                $tier['price'] = (string)($tier['price'] / 100);
            }
            unset($tier);
            $tmp = array_column($tmp, 'price', 'starting_unit');
            $onecrm['pricing_tiers'] = json_encode($tmp);
        }

        return $this;
    }

    /**
     * add one or more meta fields to $this->native and optionally mapped into $this->onecrm
     * @param array|string $values
     * @param null|string $key
     * @param bool $to_onecrm
     * @return $this
     */
    public function meta_set($values, $key=null, $to_onecrm=true){
        if (! is_array($values)) $values=[$key=>$values];
        if ($this->native['meta_data'] ?? null){
            $meta = json_decode($this->native['meta_data']) + $values;
        } else $meta = $values;
        $this->native['meta_data'] = json_encode($meta);
        if ($to_onecrm) foreach ($values as $key => $value) {
            if (!($this->map_meta_data[$key] ?? null)) continue;
            $this->onecrm[$this->map_meta_data[$key]] = $value;
        }
        return $this;
    }

    public function meta_get($key){
        if ($this->native['meta_data']){
            $meta=json_decode($this->native['meta_data']);
        } else $meta=[];
        return $meta[$key] ?? null;
    }

    /**
     * translates $this->onecrm to $this->native
     * @param $data : if not null then $data is translated to $this->native instead
     * @return BaseRecord $this
     */
    public function to_native($data=null)
    {
        // copy the fields defined by the map:
        parent::to_native($data);
        $native =& $this->native;
        $onecrm =& $this->onecrm;

        if (is_countable($this->map_meta_data) && count($this->map_meta_data)){
            $meta=[];
            foreach ($this->map_meta_data as $native_key => $onecrm_key){
                if (isset($onecrm[$onecrm_key])) $meta[$native_key] = $onecrm[$onecrm_key];
			}
			if (!empty($meta))
	            $this->meta_set($meta);
        }

        // date fields from UNIX timestamp to MySQL GMT timestamp:
        $dates=['started_at', 'trial_start', 'trial_end',
            'date_from', 'date_to', 'due_date', 'payment_date'];
        foreach($dates as $date) {
            if (isset($onecrm[$date])) {
                $native[$date] = strtotime($native[$date] . ' UTC');
            }
		}
		unset($native['created_at']);
		unset($native['updated_at']);
        if (isset($native['billing_period_unit'])) {
            $intervals = ["monthly" => "month", "yearly" => "year", "weekly" => "week", 'single'=>''];
            $native['billing_period_unit'] = $intervals[strtolower($native['billing_period_unit'])];
            $native['billing_period']=1;
            if (!$native['billing_period_unit']) $native['type']='on_off';
        }

        // TODO: amount fields, tiers


        return $this;
	}

	protected function getSaveMethods($model, $native) {
		return [$model . '::create', $model . '::update'];
	}

}
