<?php

namespace OneCRM\PaymentProcessors;

use OneCRM\API\Errors\BadRequest;
use ListQuery;

class CouponRecord extends ChargeBeeRecord
{

    public $name = 'Coupon';                    // name of object at Payment Processor
    public $ppid_field = 'id';                  // PP unique id field name
    public $model = 'Coupon';                   // name of matching 1CRM Model
    public $model_ppid_field = 'pp_coupon_id';  // 1CRM Model field name holding PP unique id

    public $map = [
        "id" => "pp_coupon_id",
        "name" => "name",
        "description" => "description",
        'discount_type' => 'discount_type',
        'discount_amount' => 'discount_amount',
        'discount_percentage' => 'discount_percentage',
        'apply_on' => 'discount_mode',
        'status' => 'active',
        'plan_constraint' => 'plan_mode',
        'addon_constraint' => 'addon_mode',
        "updated_at" => "date_modified",
        "currency_code" => "currency",      // ISO
    ];
    public $map_types = [                     // native => onecrm
        'discount_types' => [
            'fixed_amount' => 'Fixed Amount',
            'percentage' => 'Percentage',
        ],
        'status' => [
            'active' => 'Active',
            'archived' => 'Inactive',
        ],
        'apply_on' => [
            'each_specified_item' => 'Specific Item',
            'invoice_amount' => 'Invoice Amount',
        ],
        'plan_constraint' => [
			'not_applicable' => 'Not Applicable',
            'specific' => 'Selected Plans',
            'all' => 'All Plans',
            "none" => 'None',
        ],
		'addon_constraint' => [
			'not_applicable' => 'Not Applicable',
            'specific' => 'Selected Addons',
            'all' => 'All Addons',
            "none" => 'None',
        ],
    ];
    public $map_meta_data = ['description' => 'description'];


    /**
     * normalization of records:
     */

    /**
     * translates $this->native to $this->onecrm
     * @param array|null $data
     * @return BaseRecord $this
     * @throws BadRequest
     */
    public function to_onecrm($data = null)
    {
        // copy the fields defined by the map:
        parent::to_onecrm($data);

        return $this;
    }

    /**
     * translates $this->onecrm to $this->native
     * @param array|null $data
     * @return BaseRecord $this
     */
    public function to_native($data = null)
    {
        // copy the fields defined by the map:
        parent::to_native($data);

        // coupon-specific fields:
        die('TODO: coupon to_native()');

        return $this;
    }

    public function onecrm_save($before_update = null)
    {
        $onecrm =& $this->onecrm;
        if (!($onecrm['id'] ?? null)) {
            $onecrm['id'] = \Subscription::find_coupon($onecrm);
        }

        // this has to happen here so we're guaranteed to have a Coupon id to work with
        $res = parent::onecrm_save($before_update);

        // TODO: migrate shared portions of this code to BaseRecord / Traits

        if (!$res) return false;
        
        // Resolve the Plan and Addon links:
        
        $payment_processor_id = static::$processor_id;
        $rowUpdate =& $this->rowUpdate;

        // Resolve Plan links first (no special reason)

        $pp_plans = array_flip((array)($this->native['plan_ids']??null));

        $current_plans = (array)ListQuery::quick_fetch_linked('Coupon', $onecrm['id'],
            'plans', ['id', 'pp_plan_id'], null, true);
        $current_plans = array_column($current_plans,'id','pp_plan_id');

        $both = $current_plans + $pp_plans;
        $unlink_plans=array_diff_key($both, $pp_plans);
        if (count($unlink_plans))
            $res = $res && ($rowUpdate->removeLinks('plans', $unlink_plans));

        $link_pp_plans=array_diff_key($both, $current_plans);
        $link_plans=[];
        foreach ($link_pp_plans as $pp_plan_id => $junk) {
            if (!isset($native_lut[$pp_plan_id])) $add_plans[] = $pp_plan_id;
            $id = \Subscription::find_plan(compact('pp_plan_id', 'payment_processor_id'));
            if (!$id) {
                // sync a missing plan:
                $plan = new PlanRecord($pp_plan_id);
                $res = $res && ($plan->onecrm_save());
                $id = $plan->onecrm['id'];
            }
            $link_plans[] = $id;
        }
        if (count($link_plans))
            $res = $res && ($rowUpdate->addUpdateLinks('plans', $link_plans));

        // Resolve the Addons links second (no special reason)

        $pp_addons = array_flip((array)($this->native['addon_ids']??null));

        $current_addons = (array)ListQuery::quick_fetch_linked('Coupon', $onecrm['id'],
            'addons', ['id', 'pp_addon_id'], null, true);
        $current_addons = array_column($current_addons,'id','pp_addon_id');

        $both = $current_addons + $pp_addons;
        $unlink_addons=array_diff_key($both, $pp_addons);
        if (count($unlink_addons))
            $res = $res && ($rowUpdate->removeLinks('addons', $unlink_addons));

        $link_pp_addons=array_diff_key($both, $current_addons);
        $link_addons=[];
        foreach ($link_pp_addons as $pp_addon_id => $junk) {
            if (!isset($native_lut[$pp_addon_id])) $add_addons[] = $pp_addon_id;
            $id = \Subscription::find_addon(compact('pp_addon_id', 'payment_processor_id'));
            if (!$id) {
                // sync a missing addon:
                $addon = new AddonRecord($pp_addon_id);
                $res = $res && ($addon->onecrm_save());
                $id = $addon->onecrm['id'];
            }
            $link_addons[] = $id;
        }
        if (count($link_addons))
            $res = $res && ($rowUpdate->addUpdateLinks('addons', $link_addons));

        return $res;

    }


}
