<?php

namespace OneCRM\PaymentProcessors;

class CustomerRecord extends ChargeBeeRecord
{

    public $name = 'Customer';                      // name of object at Payment Processor
    public $ppid_field = 'id';                      // PP unique id field name
    public $model = 'Account';                      // name of matching 1CRM Model
    public $model_ppid_field = 'pp_customer_id';    // 1CRM Model field name holding PP unique id

    // filter for Account data:
    public $map = [                         // map of field names between CB.Customer and 1CRM.Contact
        'id' => 'pp_customer_id',
        'email' => 'email1',
        'created_at' => 'date_entered',
        'updated_at' => 'date_modified',
        'deleted' => 'deleted',
        'preferred_currency_code' => 'currency',
        'phone' => 'phone_home',
        'line1' => 'billing_address_street',
        'city' => 'billing_address_city',
        'state' => 'billing_address_state',
        'state_code' => 'billing_address_statecode',
        'country' => 'billing_address_countrycode',
        'zip' => 'billing_address_postalcode',
        'company' => 'name',
    ];

    public $map_meta_data=[
        'website' => 'website', // example pushing custom fields to Account
        'contact_id' => 'primary_contact_id',
        'account_id' => 'id',
    ];

    // separate filter for Contact data:
    public $onecrm_contact = [];
    public $contact_map = [
        'first_name' => 'first_name',
        'last_name' => 'last_name',
        'company' => 'name',
        'email' => 'email1',
        'phone' => 'phone_home',
        'line1' => 'primary_address_street',
        'city' => 'primary_address_city',
        'state' => 'primary_address_state',
        'state_code' => 'primary_address_statecode',
        'country' => 'primary_address_countrycode',
        'zip' => 'primary_address_postalcode',
    ];

    // used for to_native() for sync to the ChargeBee API only
    public $billing_address_map = [
        'billing_address[first_name]' => 'first_name',
        'billing_address[last_name]' => 'last_name',
        'billing_address[company]' => 'name',
        'billing_address[email]' => 'email1',
        'billing_address[phone]' => 'phone_office',
        'billing_address[line1]' => 'billing_address_street',
        'billing_address[city]' => 'billing_address_city',
        'billing_address[state]' => 'billing_address_state',
        'billing_address[state_code]' => 'billing_address_statecode',
        'billing_address[country]' => 'billing_address_countrycode',
        'billing_address[zip]' => 'billing_address_postalcode',
    ];

    public function onecrm_save($before_update = null)
	{
		$this->onecrm['id'] = \Subscription::find_account($this->onecrm);
		if ($this->onecrm['id'] && \AppConfig::setting('subscriptions.dont_update_customer_info')) {
			return true;
		}
		$res = parent::onecrm_save(function (\RowUpdate &$ru) use ($before_update) {
            if (is_callable($before_update)) $before_update($ru);
            // prevent overwrite of email1 with email2:
            if ($ru->getField('email2') === $this->onecrm['email1'])
                unset($ru->updates['email1']);
            // set the account type only if it has not already been set:
            if (!$ru->getField('account_type'))
                $ru->set(['account_type' => $this->account_type]);
            if (!$ru->getField('id')) {
                // a new Account is about to be created
                if (!$ru->getField('primary_contact_id')) {
                    // No primary contact is bound. Try to find a match:
                    $match = \ListQuery::quick_fetch_all('Contact', ['id'], ['filters' => [[
                        'operator' => 'OR', 'multiple' => [
                            ['field' => 'email1', 'value' => $this->onecrm['email1']],
                            ['field' => 'email2', 'value' => $this->onecrm['email1']],
                        ]
                    ]]])->getRowIndexes();

                    if (is_countable($match) && count($match)) {
                        $this->onecrm_contact['id'] = $match[0];
                        $ru->set('primary_contact_id', $match[0]);
                    } else {
                        // create a new Contact:
                        $this->onecrm_contact['assigned_user_id'] = $GLOBALS['current_user']->id;
                        // sourced from the Account which was sourced from...
                        $this->onecrm_contact['personal_info_source'] = $ru->getField('personal_info_source');
                        $this->onecrm_contact['lead_source'] = $this->lead_source;
                        $contact = \RowUpdate::blank_for_model('Contact');
                        $contact->set($this->onecrm_contact);
                        $contact->save('insert');
                        $this->onecrm_contact['id'] = $contact->getField('id');
                        $ru->set('primary_contact_id', $contact->getField('id'));
                    }
                    // keep local Account data up to date:
                    $this->onecrm['primary_contact_id'] = $this->onecrm_contact['id'];
                }
            }
            // TODO: evaluate moving this into the new Account block above.
        });
        $this->rowUpdate->addUpdateLink('contacts',$this->rowUpdate->getField('primary_contact_id'));
        return $res;
    }

    /**
     * translates $this->native to $this->onecrm
     * @param null $data
     * @return BaseRecord $this
     * @throws \OneCRM\API\Errors\BadRequest
     */
    public function to_onecrm($data = null)
    {
        if (is_null($data)) $data = $this->native;
        if (isset($data['billing_address'])) $data += $data['billing_address'];

        parent::to_onecrm($data);
        foreach ($this->contact_map as $native => $onecrm) {
            $this->onecrm_contact[$onecrm] = $data[$native] ?? null;
        }
        $onecrm = &$this->onecrm;
        // keep the Account name in sync with Processor's Customer data:
        $onecrm['name'] = $this->native['company'];
        // FIXME: work around bug in RowUpdate where it won't save countrycode if country is empty:
        if (!($onecrm['billing_address_country'] ?? null))
            $onecrm['billing_address_country'] = $onecrm['billing_address_countrycode'] ?? null;

        return $this;
    }

    public function to_native($data = null){
        parent::to_native($data);
        // merge Contact with Account, and create Billing Address subfield:
        foreach ($this->contact_map+$this->billing_address_map as $native => $onecrm) {
            $this->native[$native] = $this->onecrm[$onecrm] ?? null;
        }
        return $this;
    }
}

