<?php

namespace OneCRM\PaymentProcessors;

use OneCRM\API\Errors\BadRequest;
use OneCRM\API\Errors\InternalError;
use RowUpdate;

class InvoiceRecord extends ChargeBeeRecord
{
	use TallyTrait;

	protected $tally_line_class = '\OneCRM\PaymentProcessors\InvoiceLineRecord';
	protected $tally_adjust_class = '\OneCRM\PaymentProcessors\InvoiceAdjustmentRecord';

	public $name = 'Invoice';                   // name of object at Payment Processor
	public $ppid_field = 'id';                  // PP unique id field name
	public $model = 'Invoice';                  // name of matching 1CRM Model
	public $model_ppid_field = 'pp_invoice_id'; // 1CRM Model field name holding PP unique id
	public $terms = 'Due on Receipt';
	public $map = [
		"id" => "pp_invoice_id",
		'customer_id' => 'pp_customer_id',
		"updated_at" => "date_modified",
		'due_date' => 'due_date',
		"total" => "amount",                    // Amount Billable
		"amount_to_collect" => "amount_due",
	];


	/**
	 * normalization of records:
	 */

	/**
	 * translates $this->native to $this->onecrm
	 * @return BaseRecord $this
	 * @throws BadRequest
	 */
	public function to_onecrm($data = null)
	{
		parent::to_onecrm($data);
		$onecrm =& $this->onecrm;
		$onecrm['subtotal'] = $onecrm['amount'];
		$onecrm['total'] = $onecrm['amount'];
		$onecrm['terms'] = $this->terms;
		$onecrm['cancelled'] = ($data['status'] == 'voided') ? 1 : 0;
		// take Invoice name from first Product
		$onecrm['name'] = $this->native['line_items'][0]['description'];

		return $this;
	}

	/**
	 * @param null $before_update
	 * @return bool
	 * @throws InternalError
	 */
	public function onecrm_save($before_update = null)
	{
		$onecrm =& $this->onecrm;

		// TODO: abstract out the use of $native so onecrm_save() can move to InvoiceTrait
		$native =& $this->native;


		if (!$onecrm['id'] = \Subscription::find_invoice($onecrm)) {

			// creating a new Invoice - need billing account
			if (!$onecrm['billing_account_id'] = \Subscription::find_account($onecrm)) {
				// need to sync the customer first
				$customer = new CustomerRecord($onecrm['pp_customer_id'], 'pp_id');
				$customer->onecrm_save();
				$onecrm['billing_account_id'] = $customer->onecrm['id'];
			}
			$onecrm['shipping_account_id'] = $onecrm['billing_account_id'];
		}

		if (!parent::onecrm_save($before_update)) throw new InternalError();

		$new_groups = $this->generateTallyGroups();

		// update the invoice:
		$this->rowUpdate->replaceGroups($new_groups, true);
		$this->rowUpdate->save();

		// link / sync subscription if one is present in the Invoice:
		if ($pp_subs_id = $native['subscription_id'] ?? null) {
			$filter = ['pp_subs_id' => $pp_subs_id, "payment_processor_id" => self::$processor_id];
			if (!$subscription_id = \Subscription::find_subscription($filter)) {
				$subscription = new SubscriptionRecord($pp_subs_id);
				$subscription->onecrm_save();
				$rowUpdate = $subscription->rowUpdate;
			} else {
				$rowUpdate = RowUpdate::blank_for_model('Subscription');
				$rowUpdate->retrieveRecord($subscription_id);
			}
			$rowUpdate->addUpdateLink('invoices', $onecrm['id']);
		}

		$desc = [
			'account_id' => $onecrm['billing_account_id'],
			'processor_id' => self::$processor_id,
			'currency_id' => $onecrm['currency_id'],
			'exchange_rate' => $onecrm['exchange_rate'],
			'for_invoice' => $native['id'],
			'allocations' => [],
		];
		if ($credits = $native['applied_credits'] ?? null) {
			foreach ($credits as $cr) {
				$alloc = [
					'cn_id' => $cr['cn_id'],
					'inv_id' => $native['id'],
					'applied_at' => $cr['applied_at'],
					'amount' => $cr['applied_amount'],
				];
				$desc['allocations'][] = $alloc;
			}
		}
		CreditNotePayments::allocate_payments($desc);
		return true;
	}

	function get_native_tally_lines()
	{
		$native =& $this->native;
		$invoice_lines = [];

		foreach ($native['line_items'] as $position => $line_item) {
			$tiered = false;
			$line = $line_item + [
			    'position' => $position,
			    'base_currency_code' => $native['base_currency_code'],
			    'currency_code' => $native['currency_code'],
			    'exchange_rate' => $native['exchange_rate'],
			];
			if (!empty($native['line_item_tiers'])) {
				foreach ($native['line_item_tiers'] as $t) {
					if ($t['line_item_id'] === $line['id']) {
						$tiered = true;
						$tline = $line;
						$tline['quantity'] = $t['quantity_used'];
						$tline['unit_amount'] = $t['unit_amount'];
						$tline['amount'] = $tline['quantity'] * $tline['unit_amount'];
						$invoice_lines[] = $tline;
					}
				}
			}
			if (!$tiered) {
				$invoice_lines[] = $line;
			}
		}
		return $invoice_lines;
	}

	/**
	 * @return array
	 */
	public function get_native_adjustments()
	{
		$native =& $this->native;
		$out = [];
		$descriptions = array_column($native['discounts'], 'description', 'entity_id');
		foreach ($native['line_item_discounts'] as $position => $line) {
			$out[$position] = $line + [
				    'description' => $descriptions[$line['coupon_id']],
				    'position' => $position,
				    'base_currency_code' => $native['base_currency_code'],
				    'currency_code' => $native['currency_code'],
				    'exchange_rate' => $native['exchange_rate'],
				];
		}
		return $out;
	}

}
