<?php

namespace OneCRM\PaymentProcessors;

use OneCRM\API\Errors\BadRequest;
use ListQuery;

class PlanRecord extends ChargeBeeRecord
{

    public $name = 'Plan';                      // name of object at Payment Processor
    public $ppid_field = 'id';                  // PP unique id field name
    public $model = 'Plan';                     // name of matching 1CRM Model
    public $model_ppid_field = 'pp_plan_id';    // 1CRM Model field name holding PP unique id
    public $map = [
        'id' => 'pp_plan_id',
        'name' => 'name',                   // taken from product
        'description' => 'description',
        'period_unit' => 'billing_cycle',        // needs to be translated
        'period' => 'billing_cycle_mult',        // needs to be translated
        'status' => 'active',
        'updated_at' => 'date_modified',
        'currency_code' => 'currency',      // ISO
        'pricing_model' => 'pricing_model',
        'tiers' => 'pricing_tiers',         // needs to be translated
        'price' => 'amount',                // needs to be translated
        'setup_cost' => 'setup',            // needs to be translated
    ];
    public $map_types = [                   // native => onecrm
        'status' => [
            'active' => 'Active',
            'archived' => 'Inactive',
        ],
        'pricing_model' => [
            'flat_fee' => 'Flat Fee',
            'per_unit' => 'Per Unit',
            'volume' => 'Volume',
            'stairstep' => 'Stairstep',
            'tiered' => 'Tiered',
        ],
        'period_unit' => [
            '' => 'Single',
            'week' => 'Weekly',
            'month' => 'Monthly',
            'year' => 'Yearly',
        ],
    ];
    public $map_meta_data = [                 // custom records
        'product_id' => 'product_id',
        'qty_label' => 'qty_label',
    ];

    /**
     * normalization of records:
     */

    /**
     * translates $this->native to $this->onecrm
     * @param array|null $data
     * @return BaseRecord $this
     * @throws BadRequest
     */
    public function to_onecrm($data = null)
    {
        // copy the fields defined by the map:
        return parent::to_onecrm($data);
    }

    /**
     * translates $this->onecrm to $this->native
     * @param array|null $data
     * @return BaseRecord $this
     */
    public function to_native($data = null)
    {
        // copy the fields defined by the map:
        parent::to_native($data);

        // plan-specific fields:
        die('TODO: plan to_native()');

        return $this;
    }

    public function onecrm_save($before_update = null)
    {
        $onecrm = &$this->onecrm;
        if (!$onecrm['id'] = \Subscription::find_plan($onecrm)) {
            if (!($onecrm['product_id'] ?? null)) {
                $res = \Subscription::query_model('Product', ['name' => $onecrm['name']]);
                if ($res) {
                    $onecrm['product_id'] = $res['id'];
                } else {
                    // proceed without a product id. Maybe a warning email?
                }
            }
        }

        // this has to happen here so we're guaranteed to have a Plan id to work with
        $res = parent::onecrm_save($before_update);

        // TODO: ? Can we migrate this to Plan model post-update hook instead ?

        if (!$res) return false;

        if (isset($this->native['applicable_addons'])) {
            $pp_addons = array_column($this->native['applicable_addons'], 'id', 'id');
        } else $pp_addons=[];
        $addon_list = json_encode(array_values($pp_addons));

        // shortcut - check the encoded list against the incoming one:
        if ($addon_list === ($onecrm['addons'] ?? null)) return $res;

        $unlink_addons = $link_addons = [];
        $payment_processor_id = static::$processor_id;
        $rowUpdate =& $this->rowUpdate;

        // get the currently linked addons:
        $current_addons = ListQuery::quick_fetch_linked('Plan', $onecrm['id'],
            'addons', ['id', 'pp_addon_id'], null, true);

        // unlink unwanted addons:
        foreach ($current_addons as $addon_id => $addon_row) {
            if (!isset($pp_addons[$addon_row['pp_addon_id']])) $unlink_addons[] = $addon_row['id'];
        }
        if (count($unlink_addons))
            $res = $res && ($rowUpdate->removeLinks('addons', $unlink_addons));

        // link missing addons:
        $native_lut = array_column($current_addons, 'id', 'pp_addon_id');
        foreach ($pp_addons as $pp_addon_id) {
            if (!isset($native_lut[$pp_addon_id])) $add_addons[] = $pp_addon_id;
            $id = \Subscription::find_addon(compact('pp_addon_id', 'payment_processor_id'));
            if (!$id) {
                // sync a missing addon:
                $addon = new AddonRecord($pp_addon_id);
                $res = $res && ($addon->onecrm_save());
                $id = $addon->onecrm['id'];
            }
            $link_addons[] = $id;
        }
        if (count($link_addons))
            $res = $res && ($rowUpdate->addUpdateLinks('addons', $link_addons));

        // save the new encoded list of addons:
        $onecrm['addons'] = $addon_list;
        $rowUpdate->set('addons', $addon_list);
        $res = $res && ($rowUpdate->save());

        return $res;
    }

}
