<?php

namespace OneCRM\PaymentProcessors;

use OneCRM\API\Errors\BadRequest;
use ListQuery;

require_once "modules/PaymentProcessorChargeBee/inc/ChargeBeeRecord.php";


// FIXME: to_onccrm() assumption breaks Yen and other non-float currencies

class SubscriptionRecord extends ChargeBeeRecord
{

    public $name = 'Subscription';              // name of object at Payment Processor
    public $ppid_field = 'id';                  // PP unique id field name
    public $model = 'Subscription';             // name of mathching 1CRM Model
    public $model_ppid_field = 'pp_subs_id';    // 1CRM Model field name holding PP unique id
    public $map_meta_data = [                 // custom records
        'onecrm_options' => 'onecrm_options',
    ];
    public $map = [
        'id' => 'pp_subs_id',
        "customer_id" => 'pp_customer_id',
        "plan_id" => 'pp_plan_id',
        "plan_quantity" => 'quantity',
        "billing_period" => 'billing_interval',     // ie 1
        "billing_period_unit" => 'billing_cycle',   // ie monthly
        "status" => 'state',                        // ie 'future' -> Draft
        "started_at" => 'start_date',
        "current_term_end" => 'current_period_end',
		"current_term_start" => 'current_period_start',
		'next_billing_at' => 'next_billing_date',
		'mrr' => 'mrr',
        "created_at" => 'date_entered',
        "updated_at" => 'date_modified',
        "deleted" => 'deleted',
        "currency_code" => 'currency',              // ISO Code
		"addons" => 'addons', // [0 => ['id' => 'addon_sku', 'quantity' => 1, 'unit_price' => 95]]
		"coupon_ids" => 'coupon_ids',
		"auto_collection" => "auto_collection",
		'payment_source_id' => 'payment_source_id',
    ];


    /**
     * normalization of records:
     */

    /**
     * translates $this->native to $this->onecrm
     * @param array|null $data : if not null then $data is translated to $this->onecrm instead
     * @return BaseRecord $this
     * @throws BadRequest
     */
    public function to_onecrm($data = null)
    {
        // copy the fields defined by the map:
		parent::to_onecrm($data);
		$onecrm =& $this->onecrm;
        if (is_null($data)) $data = $this->native;
        // create hash maps for Addons and Coupons for this Subscription:
        $onecrm['addons'] = json_encode(
            (object)array_column((array)($data['addons']??null), 'quantity', 'id'));
        $onecrm['coupons'] = json_encode(
            (object)array_column((array)($data['coupons']??null), 'applied_count', 'coupon_id'));
        $onecrm['options'] = json_encode($this->mapCustomFieldsToOptions($onecrm, $data));
		$onecrm['has_unpaid_invoices'] = empty($data['due_invoices_count']) ? 0 : 1;

        return $this;
	}

	protected function mapCustomFieldsToOptions($onecrm, $data) {
		$options = $data['meta_data']['onecrm_options'] ?? [];
		$planId = \Subscription::find_plan($onecrm);
		if ($planId) {
			$opts = \ListQuery::quick_fetch_linked('Plan', $planId, 'options');
			foreach ($opts->rows as $o) {
				if ($o['custom_field_name'] && !array_key_exists($o['input'], $options)) {
					$k = 'cf_' . $o['custom_field_name'];
					if (array_key_exists($k, $data)) {
						$options[$o['input']] = $data[$k];
					}
				}
			}
		}
		return $options;
	}

    /**
     * translates $this->onecrm to $this->native
     * @param $data : if not null then $data is translated to $this->native instead
     * @return BaseRecord $this
     */
    public function to_native($data = null)
    {
        // copy the fields defined by the map:
		parent::to_native($data);
		if (is_null($data)) {
			$data =& $this->onecrm;
		}
		$native =& $this->native;
		if (!empty($native['addons']['id'])) {
			$addons = $native['addons'];
			unset($native['addons']);
			foreach ($addons['id'] as $k => $id) {
				$native["addons[id][$k]"] = $id;
				if (!empty($addons['quantity'][$k])) {
					$native["addons[quantity][$k]"] = $addons['quantity'][$k];
				}
			}
		}
		$native['replace_addon_list'] = true;
		if (!empty($native['coupon_ids'])) {
			$coupons = $native['coupon_ids'];
			unset($native['coupon_ids']);
			$k = 0;
			foreach ($coupons as $cid) {
				$native["coupon_ids[$k]"] = $cid;
				$k++;
			}
		}
		if (isset($data['invoice_notes'])) {
			$native['invoice_notes'] = $data['invoice_notes'];
		}
        return $this;
    }

    /**
     * @param callable|null $before_update
     * @return bool
     */
    public function onecrm_save($before_update = null)
    {
        $onecrm = &$this->onecrm;
        $native = $this->native;

        // find an existing Subscription:
        if (!($onecrm['id']??null && $onecrm['account_id']??null && $onecrm['contact_id']??null)) {
            if ( $res = \Subscription::find_subscription($onecrm, ['id', 'account_id', 'contact_id'])) {
                $onecrm = $res + $onecrm;
            }
            // link / sync Customer - this code should never run since Customer is always synced before Subscription.
            if (! ($onecrm['contact_id']??null && $onecrm['account_id']??null)) {
                if (! $res = \Subscription::find_account($onecrm, ['id', 'primary_contact_id'])) {
                    $customer = new CustomerRecord($onecrm['pp_customer_id']);
                    $customer->onecrm_save();
                    $res = $customer->onecrm;
                }
                $onecrm['account_id'] = $res['id'];
                $onecrm['contact_id'] = $res['primary_contact_id'];
            }
        }

        // link / sync Plan:
        if (!($onecrm['plan_id']??null)) {
            if (!$plan_id = \Subscription::find_plan($onecrm)) {
                $plan = new PlanRecord($onecrm['pp_plan_id']);
                $plan->onecrm_save();
                $plan_id = $plan->onecrm['id'];
            }
            $onecrm['plan_id'] = $plan_id;
        }

        // sync any missing Addons:
        foreach ($native['addons'] ?? [] as $addon) {
            $where = ['pp_addon_id' => $addon['id'], 'payment_processor_id' => self::$processor_id];
            if (!\Subscription::find_addon($where)) {
                $addonRecord = new AddonRecord($addon['id']);
                $addonRecord->onecrm_save();
            }
        }

        // sync any missing Coupons:
        foreach ($native['coupons'] ?? [] as $coupon) {
            $where = ['pp_coupon_id' => $coupon['coupon_id'], 'payment_processor_id' => self::$processor_id];
            if (!\Subscription::find_coupon($where)) {
                $couponRecord = new CouponRecord($coupon['coupon_id']);
                $couponRecord->onecrm_save();
            }
        }

        if (! ($onecrm['name']??null)){
            $plan = ListQuery::quick_fetch_row('Plan', $onecrm['plan_id'], ['name']);
            $contact = ListQuery::quick_fetch_row('Contact', $onecrm['contact_id'], ['name']);
            $onecrm['name']=$contact['name'].' - '.$plan['name'];
        }
        return parent::onecrm_save($before_update);
    }

	protected function getSaveMethods($model, $native) {
		$ret = parent::getSaveMethods($model, $native);
		if (!empty($native['customer_id'])) {
			$cid = $native['customer_id'];
			$ret[0] = function ($params, $env = null, $headers = []) use ($model, $cid) {
				return $model::createForCustomer($cid, $params, $env, $headers);
			};
		}
		return $ret;
	}
}
