<?php

namespace OneCRM\PaymentProcessors;

use OneCRM\API\Errors\BadRequest;
use RowUpdate;

class TransactionRecord extends ChargeBeeRecord
{

    /* Bi-Directional class, supports ::all() */

    public $name = 'Transaction';                   // name of object at Payment Processor
    public $ppid_field = 'id';                      // PP unique id field name
    public $model = 'Payment';                      // name of matching 1CRM Model
    public $model_ppid_field = 'pp_transaction_id'; // 1CRM Model field name holding PP unique id
    public $prefix = 'PID-';
    public $payment_type = 'ChargeBee';
    public $allocated = false;
    public $map = [
        'id' => 'pp_transaction_id',
        'customer_id' => 'pp_customer_id',
        'id_at_gateway' => 'customer_reference',
        'subscription_id' => 'pp_subs_id',
        'payment_method' => 'payment_method',
        'gateway' => 'gateway',
        'type' => 'direction',
        'date' => 'payment_date',
        'exchange_rate' => 'exchange_rate',
        'amount' => 'amount',
        'status' => 'success',
        'updated_at' => 'date_modified',
        'deleted' => 'deleted',
        'masked_card_number' => 'card_ending', // use last 4 digits
        'currency_code' => 'currency',
    ];
    public $map_types = [                   // native => onecrm
        'type' => [
            'payment' => 'incoming',
            'refund' => 'refund',
        ],
        'pricing_model' => [
            'flat_fee' => 'Flat Fee',
            'per_unit' => 'Per Unit',
            'volume' => 'Volume',
            'stairstep' => 'Stairstep',
            'tiered' => 'Tiered',
        ],
        'period_unit' => [
            '' => 'Single',
            'week' => 'Weekly',
            'month' => 'Monthly',
            'year' => 'Yearly',
        ],
    ];

    /**
     * normalization of records:
     */

    /**
     * translates $this->native to $this->onecrm
     * @param array|null $data
     * @return BaseRecord $this
     * @throws BadRequest
     */
    public function to_onecrm($data = null)
    {
        // copy the fields defined by the map:
        parent::to_onecrm($data);
        $onecrm =& $this->onecrm;

        if ($onecrm['type'] == 'payment') {
            $onecrm['direction'] = 'incoming';
            $onecrm['total_amount'] = $onecrm['amount'];
        }

        return $this;
    }

    /**
     * @param null $before_update
     * @return bool
     * @throws BadRequest
     */
    public function onecrm_save($before_update = null)
    {
        $onecrm =& $this->onecrm;
        $native =& $this->native;

        if (!($onecrm['id'] ?? null)) $onecrm['id'] = \Subscription::find_payment($onecrm);

		$isRefund = false;
		$success = parent::onecrm_save(function (RowUpdate &$ru) use ($before_update, $onecrm, $native, &$isRefund) {
			$isRefund = $ru->getField('direction') == 'refund';
            if (!$ru->getField('account_id')) {
                if (!$account_id = \Subscription::find_account($onecrm)) {
                    $account = new CustomerRecord($onecrm['pp_customer_id']);
                    $account->onecrm_save();
                    $account_id = $account->onecrm['id'];
                }
                $ru->set('account_id', $account_id);
            }

			if (!$ru->getField('related_invoice_id') && !$isRefund) {
                $onecrm['pp_invoice_id'] = $native['linked_invoices'][0]['invoice_id'];
                if (!$invoice_id = \Subscription::find_invoice($onecrm)) {
                    $invoice = new InvoiceRecord($onecrm['pp_invoice_id']);
                    $invoice->onecrm_save();
                    $invoice_id = $invoice->onecrm['id'];
                }
                $ru->set('related_invoice_id', $invoice_id);
            }

            if ($ru->new_record) {
                $ru->set([
                    'payment_type' => $this->payment_type,
                    'prefix' => $this->prefix,
                ]);
            } else {
                $this->allocated = true;
            }

            if (is_callable($before_update)) $before_update($ru);
        });


        if ($success && !$isRefund) {
            $success = $this->rowUpdate->addUpdateLink('invoices', array(
                'id' => $onecrm['related_invoice_id'],
                'amount' => $onecrm['amount'],
                'currency_id' => $onecrm['currency_id'],
                'exchange_rate' => $onecrm['exchange_rate'],
            ));
            if (!$success) throw new BadRequest("Payment cannot be linked to Invoice");
		} else if ($success && $isRefund) {
			$cn_query = [
				'pp_credit_id' => $native['linked_credit_notes'][0]['cn_id'],
			];
			if ($credit_id = \Subscription::find_creditnote($cn_query)) {
				$success = $this->rowUpdate->addUpdateLink('credits', array(
					'id' => $credit_id,
					'amount' => $onecrm['amount'],
					'currency_id' => $onecrm['currency_id'],
					'exchange_rate' => $onecrm['exchange_rate'],
				));
				if (!$success) throw new BadRequest("Credit note cannot be linked to Invoice");
			}
		}

		$params = [
			'payment' => $this->rowUpdate,
			'subscription_id' => \Subscription::find_subscription($onecrm),
		];
		if (!empty($onecrm['related_invoice_id'])) {
			$invoice = \ListQuery::quick_fetch('Invoice', $onecrm['related_invoice_id']);
			if ($invoice) {
				$params['invoice'] = \RowUpdate::for_result($invoice);
			}
		}

        return $success;
    }

}
