<?php

require_once 'data/SugarBean.php';

use OneCRM\API\Errors\BadRequest;
use Psr\Http\Message\ServerRequestInterface as Request;
use Psr\Http\Message\ResponseInterface as Response;


/*
 * Defines the state and options of a PaymentProcessor
 * NOTE: several SugarBean functions can corrupt state.  If in doubt avoid parent methods.
 *
 * Evaluate 1CRM Dev Guide 2.3 Integration Using Workflow Actions for suitability
 *
 * PaymentProcessorInterface specifies Payment-Processor-Specific methods
 * PaymentProcessor Class provides reusable base methods for the Model
 *
 * Each Payment Processor Module should extend PaymentProcessor and implement PaymentProcessorInterface
 * recommended Module name is PaymentProcessor$Name so they're grouped in tree
 * recommended db name for PP Module bean is processor$name (if required)
 */

interface PaymentProcessorInterface
{
	/*
	 * returns singleton instance
	 */
	public static function instance();

	/*
	 * returns (string) unique payment processor ID from the processor
	 */
	public function id();

	/*
	 * Process an event from the Payment Processor - API or Webhook.
	 *
	 * @param $eventData : data from Payment Processor
	 *
	 * WARN: can throw if the data is in unexpected format
	 */
	public function processEvent(array $params, Response $response, Request $request);

	/*
	 * Create a new customer at the Payment Processor
	 *
	 * @param $customerData : accepts Contact model fields and adds a new customer
	 *
	 * returns [ 'pp_customer_id' => $pp_customer_id, 'contact'=> $pp_customer_data ]
	 * or false on error.
	 *
	 * NOTE: $pp_customer_data must be normalized to 1CRM Contact model format.
	 */
	public function addCustomer(array $customerData);

	/*
	 * Update existing customwe at the Payment Processor
	 *
	 * @param $customerData : array 1CRM Customer model field format data
	 * returns [ 'pp_customer_id' => $pp_customer_id, 'contact'=> $pp_customer_data ]
	 * or false on error.
	 *
	 * NOTE: $pp_customer_data must be normalized to 1CRM Contact model format.
	 * TODO: consult re: wisdom of linking customer with PP right here instead of in Subscription.
	 */
	public function updateCustomer($customerData);

	/*
	 * GDPR compliance - if customer requests to be deleted entirely.
	 * use upstream GDPR compliance supression call
	 * or replace all sensitive data with pseudorandom bytes
	 * and remove all registered payment methods, if any
	 *
	 * @param $customerId : string pp_customer_id
	 *
	 * fail if there are active subscriptions or if there is a balance due.
	 * WARN: do not rely on this failure mode, it is not guaranteed.
	 *
	 * on failure dispatch a system error event (with blame backtrace) and return false
	 * otherwise return true.
	 */
	public function suppressCustomer($customerId);

	/*
	 * Create a new Subscription for a customer
	 * Called when a Subscription transtions from Draft state
	 *
	 * @param $subscriptionData : array 1CRM Subscription model data
	 *
	 * WARN: can throw if data is in unexpected format
	 */

	public function createSubscription($subscriptionData);

	/*
	 * Update an existing Subscription for a customer
	 *
	 * @param $subscriptionData : array 1CRM Subscription model data
	 * returns [ 'pp_customer_id' => $pp_customer_id, 'subscription' => $pp_subscription_data ]
	 * NOTE: subscription is in 1CRM Subscription model format
	 */
	public function updateSubscription($subscriptionData);

	/*
	 * Cancel a Subscription for a customer
	 *
	 * @param $subscriptionId : string pp_subscription_id
	 * @param $options : array with options
	 *
	 * $options accepts the following options in general, but not all may be supported
	 *  end_of_term - if true, subscription will be calcelled at the end of current term, immediately otherwise
	 *  credits - one of none, prorate, full
	 */
	public function cancelSubscription($subscriptionId, $options);

	/*
	 * Get a list of the contact's subscriptions
	 *
	 * @param $customerId : string pp_customer_id
	 * @param $filter : array of desired states from Subscription->state
	 */
	public function getSubscriptions($customerId, $state);

	public function getPaymentMethods($customerId);

	public function getPaymentMethodsURL($customerId, $returnURL);

	public function convertAddons($addons);

	public function formatForDisplay(array $native);

	/*
	 * Extract event data to pass to hooks
	 *
	 * @param $event : string event type
	 * @param $data : array event data
	 *
	 * @return array event data for hook
	 **/
	public function extractEventData($event, $data);

}


class PaymentProcessor extends SugarBean
{
	// model meta:
	var $table_name = "paymentprocessors";
	var $object_name = "PaymentProcessor";
	var $new_schema = true;

	// system fields:
	var $id;                    // LBL_PROCESSOR_ID
	var $deleted;
	var $date_entered;
	var $date_modified;
	var $name;

	var $active;                // LBL_ACTIVE
	var $features;              // LBL_PROCESSOR_FEATURES
	var $api_endpoint;          // LBL_API_ENDPOINT
	var $auth_mech;             // LBL_PROCESSOR_AUTH_MECH
	var $api_user;              // LBL_API_USER
	var $api_pass;              // LBL_API_PASS
	var $api_ssl;               // LBL_API_SSL
	var $webhook_user;          // LBL_WEBHOOK_USER
	var $webhook_pass;          // LBL_WEBHOOK_PASS
	var $webhook_ssl;           // LBL_REQUIRE_SSL
	var $pay_gw;                // LBL_PAY_GW
	var $auth_gw;               // LBL_AUTH_GW

	// relationships
	var $payment_processor;     // LBL_PAYMENT_PROCESSOR
	var $product;               // LBL_PRODUCT

	public static $pp_sync_records = []; // record name as key, toggles sync available


	// additional form fields
	var $additional_column_fields = Array();

	// key => model remationships
	var $relationship_fields = Array();


	// FIXME:  HERE WE ARE.

	/*
	 * initialize with properties from model row of same name
	 *
	 * @param string $name
	*/
	public function __construct($name)
	{
		parent::__construct();
		if ($name) {
			$this->retrieve_by_string_fields(['name' => $name], false);
		}
		return $this;
	}

	/**
	 * @param string $iso4217 code ie 'USD'
	 * @param bool $row whether to return the row or just the id field value.
	 * @return bool|string|array id or Currency model row matching the ISO code, false on no match.
	 */
	public static function currencyByIso4217($iso4217=null, $row=false){
		static $currencyByIso4217 = null;
		if (is_null($currencyByIso4217)) {
			$currencyByIso4217 = array_column(AppConfig::db_all_rows('Currency'), 'id', 'iso4217');
		}
		$id = $currencyByIso4217[$iso4217] ?? null;
		if ($id && $row) {
			return AppConfig::db_all_rows('Currency')[$id] ?? null;
		} elseif ($id) {
			return $id;
		}
		return false;
	}

	/**
	 * @param string $record Name of the record type ie Customer
	 * @param array $data 1CRM normal for the backing Model
	 * @return array|bool [
	 * @throws BadRequest
	 */
	public function sync_record($record, $data)
	{
		if (! (static::$pp_sync_records[$record] ?? null) ) {
			throw new BadRequest("Not Implemented: Sync $record to {$this->name}");
		}
		$className = "\\OneCRM\\PaymentProcessors\\{$record}Record";
		$record = new $className($data, 'onecrm');
		/** @var CustomerRecord $record */
		return $record->pp_save();
	}


	function get_summary_text()
	{
		return $this->name;
	}


	public function getPaymentMethods($customer_id) {
		throw new BadRequest("Not Implemented: {$this->name}::getPaymentMethods()");
	}

	public static function getDefault() {
		$lq = new \ListQuery('PaymentProcessor');
		$lq->addSimpleFilter('is_default', true);
		$lq->addSimpleFilter('active', 'Active');
		$pp = $lq->runQuerySingle();
		if (!$pp || $pp->failed)
		   	return;
		$name = $pp->getField('name');
		$filename = "modules/PaymentProcessor$name/$name.php";
		$className = "PaymentProcessor$name";
		if (is_file($filename)) {
			require_once($filename);
			if (class_exists($className)) {
				$pp = new $className();
				return $pp;
			}
		}
	}

	/**
	 *
	 * Checks if event has been processed already.
	 *
	 * @param PaymentProcessorInterface $pp Payment processor
	 * @param string $eventId event ID as specified by the payment processor
	 * @return bool
	 */
	public static function isEventProcessed(PaymentProcessorInterface $pp, $eventId)
	{
	    if ($_SERVER["HTTP_WEBHOOK_DEDUP"] ?? null === "1") return false;
		require_once 'include/database/ListQuery.php';
		$lq = new ListQuery('subscription_events');
		$lq->addSimpleFilter('processor_id', $pp->id());
		$lq->addSimpleFilter('event_id', $eventId);
		return (bool)$lq->fetchResultCount();
	}

	/**
	 *
	 * Marks an event as processed.
	 *
	 * @param PaymentProcessorInterface $pp Payment processor
	 * @param string $eventId event ID as specified by the payment processor
	 */
	public static function markEventProcessed(PaymentProcessorInterface $pp, $eventId)
	{
		$mgr = DBManager::getInstance();
		$query = sprintf(
			"INSERT IGNORE INTO subscription_events (processor_id, event_id, date_modified) VALUES ('%s', '%s', '%s')",
			$mgr->quote($pp->id()),
			$mgr->quote($eventId),
			gmdate('Y-m-d H:i:s')
		);
		$mgr->query($query);
	}

	static function after_save(RowUpdate &$update) {
		$db = DBManager::getInstance();
		if (!$update->getField('is_default')) return;
		$query = sprintf(
			"UPDATE paymentprocessors SET is_default = 0 WHERE id != '%s'",
			$db->quote($update->getPrimaryKeyValue())
		);
		$db->query($query);
	}



}
