<?php

namespace OneCRM\PaymentProcessors;
require_once 'modules/Subscriptions/Subscription.php';

use OneCRM\API\Errors\InternalError;
use RowUpdate;
use TallyUpdate;

/*
 * Framework for handling normalization, read/write, and synchronization of
 * individual Payment Processor records with 1CRM db Models.
 *
 * usage should be as simple as:
 *      $dr = new DerivedRecord();
 *      $dr->pp_getter($pp_record_id)->onecrm_save();       // sync pp->1crm
 *      $dr->onecrm_getter($onecrm_record_id)->pp_save();   // sync 1crm->pp
 *
 * TODO: optimisations around record differencing to reduce unnecessary writebacks
 * TODO: compare the dates and refuse to overwrite a more recent record.
 * TODO: admin email for failure events
 */


/* methods and properties shared between PaymentProcessors for each *Record class. */
trait SubscriptionTrait
{

}

trait PlanTrait
{
}

trait AddonTrait
{

}

trait CouponTrait
{

}

trait TallyTrait {

    /**
     * @return array $new_groups
     * @throws InternalError
     */
    function generateTallyGroups()
    {
        $new_groups = [
            'group~0' => [
                'status' => 'Closed Accepted',
                'lines' => [],
                'adjusts' => [],
            ],
        ];

		$native_lines = $this->get_native_tally_lines();
		foreach ($native_lines as $position => $line_item) {
            $tallyLineRecord = new $this->tally_line_class($line_item, 'native');
			unset($tallyLineRecord->onecrm['payment_processor_id']);
			$row = $tallyLineRecord->onecrm;
			$rounding_error = $row['quantity'] * $row['unit_price'] != $row['ext_price'];
			if ($rounding_error) {
				$orig_price = $row['ext_price'] / $row['quantity'];
				$orig_qty = $row['quantity'];
				$row['unit_price'] = $row['ext_price'];
				$row['quantity'] = 1;
				$row['name'] .= "$orig_qty x $orig_price";
			}

			$new_groups['group~0']['lines']['line~' . $position] = $row;
            // TODO: Taxes - requires tax_class_id set on each line
        }

        $native_adjustments = $this->get_native_adjustments();
        foreach ($native_adjustments as $position => $adjustment) {
            $tallyAdjustRecord = new $this->tally_adjust_class($adjustment, 'native');
            unset($tallyAdjustRecord->onecrm['payment_processor_id']);
            $new_groups['group~0']['adjusts']['adjdisc~' . $position] = $tallyAdjustRecord->onecrm;
        }
        /* TODO: Taxes - look up 1CRM tax classes against CB's ?
         *       see install/asyncDemo.php line 2030 '$taxes = array(' ...
         *
         * TODO: Shipping?  Not for our Subscriptions but maybe for customers?
         */
        return $new_groups;
    }

}

abstract class BaseRecord
{
    public $name = null;        // ie 'Customer' name of object at Payment Processor
    public $ppid_field = null;  // PP unique id field name
    public $model = null;       // ie 'Contact' name of mathching 1CRM Model
    public $model_ppid_field = null;    // 1CRM Model field name holding PP unique id
    public $map = [];           // ['native_field'=>'onecrm_field']
    public $map_types=[];       // ['native_field'=>['native_value'=>'onecrm_value']
    public $list = [];          // rows of pp-native data from the getter
    public $native = [];        // data formatted in Payment Processor format
    public $onecrm = [];        // data formatted in 1CRM Model format
    public $account_type = 'Customer';          // the default account type to create
    public $lead_source = 'Subscription';       // may differ per Payment Processor

    /** @var TallyUpdate $rowUpdate */
    public $rowUpdate = null;   //  - valid after onecrm_save()
    public $pp_save_success = null;
    public $onecrm_save_success = null;

    /*
     * NOTE: each Processor's derived base record MUST define its own
     * static $processor_id for use by the shared methods defined here
     */
    public static $processor_id = 'see '.__CLASS__.': line '.(__LINE__ - 4);

    /**
     * PaymentProcessorObject constructor.
     * can be optionally initialized from 1CRM DB, PP API,
     * or by passing in array structured data
     *
     * @param mixed $data
     * @param string $format : one of 'pp_id','onecrm_id','native','onecrm'
     * @param string $processor_id : stored statically at shared parent class
     */
    public function __construct($data = null, $format = 'pp_id', $processor_id=null)
    {
        if ($processor_id) {
            // don't share statically below specific Payment Processor:
            if (count(class_parents($this))>1) static::$processor_id = $processor_id;
        } else {
            $processor_id = static::$processor_id;
        }
        $this->rowUpdate=RowUpdate::blank_for_model($this->model);
        if (is_null($data)) return $this;

        if ($format=='onecrm' && $processor_id)
            $data['payment_processor_id'] = $processor_id;

        if ($data) switch ($format) {
            case 'pp_id':
                $this->pp_getter($data, true);
                break;
            case 'onecrm_id':
                $this->onecrm_getter($data, true);
                break;
            case 'native':
                $this->native = $data;
                $this->to_onecrm();
                break;
            case 'onecrm':
                $this->onecrm = $data;
                $this->to_native();
                break;
            DEFAULT:
                break;
        }

        if (!isset($this->onecrm['payment_processor_id']))
            $this->onecrm["payment_processor_id"]=$processor_id;

        return $this;
    }

    /**
     * adapts PaymentProcessor::onecrm_getter() :
     * pre-fills the model, looks up a matching record, stores it and returns $this
     *
     * @param mixed $filter
     * @param boolean $to_native : true to normalize and copy the fields to $this->native
     * @param mixed $fields : [field,names] or true for all fields.
     * @return BaseRecord $this
     */
    public function onecrm_getter($filter, $to_native = true, $fields = true)
    {
        $res = \Subscription::onecrm_getter($this->model, $filter, $fields);
        $this->onecrm = array_get_default($res, 0, []);
        if ($to_native) $this->to_native();
        return $this;
    }

    /**
     * looks up one or more objects at the Payment Processor
     * stores the result in $this->list and returns self
     *
     * @param mixed $filter : if string is passed, search for $this->pp_id
     * @param boolean $to_onecrm : true to normalize and copy the fields to $this->onecrm
     * @return BaseRecord $this
     */
    abstract public function pp_getter($filter, $to_onecrm = true);

    /**
     * increments the list pointer and returns the current item or false
     * sets $native to the result, or empty on error
     *
     * @param boolean $to_onecrm : true to normalize and copy the fields to $this->onecrm
     * @return mixed
     */
    public function next($to_onecrm = true)
    {
        $res = next($this->list);
        if ($res) {
            $this->native = $res;
            if ($to_onecrm) $this->to_onecrm();
        } else {
            $this->native = $this->onecrm = [];
        }
        return $res;
    }

    /**
     * decrements the list pointer and returns the current item or false
     * sets $native to the result, or empty on error
     *
     * @param boolean $to_onecrm : true to normalize and copy the fields to $this->onecrm
     * @return mixed
     */
    public function prev($to_onecrm = true)
    {
        $res = prev($this->list);
        if ($res) {
            $this->native = $res;
            if ($to_onecrm) $this->to_onecrm();
        } else {
            $this->native = $this->onecrm = [];
        }
        return $res;
    }

    /**
     * increments the list pointer and returns the current item or false
     * sets $native to the result, or empty on error
     *
     * @param boolean $to_onecrm : true to normalize and copy the fields to $this->onecrm
     * @return mixed
     */
    public function reset($to_onecrm = true)
    {
        $res = reset($this->list);
        if ($res) {
            $this->native = $res;
            if ($to_onecrm) $this->to_onecrm();
        } else {
            $this->native = $this->onecrm = [];
        }
        return $res;
    }

    /**
     * write the object back to the Payment Processor
     * @return boolean
     */
    abstract public function pp_save();

    /**
     * write the object back to 1CRM DB
     * if id field is not set it will create a new record and set the 'id' field of $this->onecrm
     * @param callable $before_update (&RowUpdate) called after row is read, before it is written
     * @return boolean
     */
    public function onecrm_save($before_update = null)
    {
		$rowUpdate =& $this->rowUpdate;
		$self = $this;
		$runHooks = function($type) use($self) {
			$hooks = \AppConfig::setting("subscriptions.hooks.{$self->name}.{$type}");
			if (!empty($hooks)) {
				foreach ($hooks as $hook) {
					call_hook_function($hook, null, [$self]);
				}
			}
		};

        if ($this->onecrm['id'] ?? null) {
            // update an existing record:
            $rowUpdate->retrieveRecord($this->onecrm['id']);
            $rowUpdate->set($this->onecrm);
			if (is_callable($before_update)) $before_update($rowUpdate);
			$runHooks('before_onecrm_save');
            $rowUpdate->save('update');
			$runHooks('after_onecrm_save');
		} else {
            // create a new record:
            if ($rowUpdate->getFieldType('assigned_user_id')) {
                $this->onecrm['assigned_user_id'] = $GLOBALS['current_user']->id;
            }
			$rowUpdate->set($this->onecrm);
            if (is_callable($before_update)) $before_update($rowUpdate);
			$runHooks('before_onecrm_save');
			$rowUpdate->save('insert');
			$runHooks('after_onecrm_save');
            //$this->onecrm['id'] = $rowUpdate->getField('id');
        }
        $this->onecrm = $rowUpdate->orig_row;
        return ($rowUpdate->errors) ? false : true;
    }


    /**
     * translates $this->native to $this->onecrm
     * @param $data : if not null then $data is translated to $this->onecrm instead
     * @return BaseRecord $this
     */
    public function to_onecrm($data = null)
	{
		$json = getJSONobj();
        $this->onecrm = [];
        if (is_null($data)) $data = $this->native;
        foreach ($this->map as $native => $onecrm) {
            if (isset($this->map_types[$native])){
                $this->onecrm[$onecrm] = $this->map_types[$native][$data[$native]] ?? null;
            } else {
                $this->onecrm[$onecrm] = $data[$native] ?? null;
            }
        }
        // Policy: only store 4 digits of PAN
        if (isset($this->onecrm['card_ending']) && strlen($this->onecrm['card_ending'])>4){
            $this->onecrm['card_ending'] = substr($this->onecrm['card_ending'],-4);
        }

        $this->onecrm['payment_processor_id'] = static::$processor_id;
        $this->onecrm['native'] = $json->encode($this->native);

        return $this;
    }

    /**
     * translates $this->onecrm to $this->native
     * @param $data : if not null then $data is translated to $this->native instead
     * @return BaseRecord $this
     */
    public function to_native($data = null)
    {
        $this->native = [];
        if (is_null($data)) $data = $this->onecrm;
        foreach ($this->map as $native => $onecrm) {
            if (isset($this->map_types[$native])){
                $types=array_flip($this->map_types[$native]);
                $this->native[$native] = $types[$data[$onecrm]] ?? null;
            } else {
                $this->native[$native] = $data[$onecrm] ?? null;
            }
        }
        return $this;
    }

    /**
     * translates $this->native to $this->onecrm
     * @param array $data : assoc array of fields to update
     * @param boolean $to_native : true to normalize and copy the fields to $this->to_native
     * @return BaseRecord $this
     */
    public function update_onecrm($data, $to_native = true)
    {
        $this->onecrm = $data + $this->onecrm;
        if ($to_native) $this->to_native();
        return $this;
    }

    /**
     * translates $this->onecrm to $this->native
     * @param array $data : assoc array of fields to update
     * @param boolean $to_onecrm : true to normalize and copy the fields to $this->onecrm
     * @return BaseRecord $this
     */
    public function update_native($data, $to_onecrm = true)
    {
        $this->native = $data + $this->native;
        if ($to_onecrm) $this->to_onecrm();
        return $this;
    }

}
