<?php

require_once 'modules/PriceBooks/utils.php';

class PriceBookFormulaParser {

	private static $allowed_vars = array(
		'cost',
		'list_price',
		'purchase_price',
	);


	public static function execute($operations, $vars, $round=null)
	{
		$stack = array();
		foreach ($operations as $o) {
			list($op, $pos) = $o;
			if (is_numeric($op)) {
				$stack[] = $op;
			} else if ($op === '+u') {
			} else if ($op === '-u') {
				$stack[] = -array_pop($stack);
			} else if ($op === '-') {
				$stack[] = -array_pop($stack) + array_pop($stack);
			} else if ($op === '+') {
				$stack[] = array_pop($stack) + array_pop($stack);
			} else if ($op === '*') {
				$stack[] = array_pop($stack) * array_pop($stack);
			} else if ($op === '/') {
				$x = array_pop($stack);
				$stack[] = array_pop($stack) / $x;
			} else if ($op === '^') {
				$x = array_pop($stack);
				$stack[] = pow(array_pop($stack), $x);
			} else {
				$stack[] = (double)array_get_default($vars, $op, 0);
			}
		}
		$val = array_pop($stack);
		$val = self::apply_discount($val, $vars);
		$val = self::round_to_base($val, $round);
		return $val;
	}
	
	public static function apply_discount($val, $vars) {
		if(! empty($vars['discount_type'])) {
			if($vars['discount_type'] == 'percentage' && ! empty($vars['discount_value'])) {
				$val = (100 - $vars['discount_value']) * $val / 100;
			}
			else if($vars['discount_type'] == 'fixed') {
				$val -= $vars['discount_value'];
			}
		}
		return $val;
	}
	
	public static function round_to_base($val, $round)
	{
		if(isset($val) && isset($round)) {
			$round = abs($round);
			if($round > 0)
				$val = round($val / $round) * $round;
		}
		return $val;
	}

	public static function parse($formula)
	{
		$re = array(
			'\\^',
			'\\(',
			'\\)',
			'\\+',
			'\\*',
			'/',
			'-',
		);

		$pattern = '~(' . join('|', $re) . ')~i';
		$tokens = preg_split($pattern, $formula, -1, PREG_SPLIT_DELIM_CAPTURE | PREG_SPLIT_OFFSET_CAPTURE);
		try {
			return self::parse_to_rpn($tokens);
		} catch (PriceBookFormulaParserException $e) {
			$GLOBALS['log']->error($e->getMessage());
			throw $e;
		}
	}

	private static function parse_to_rpn($tokens) {

		$output = array();
		$stack = array();

		$last_was_operator = true;
		$token = "\x01";

		foreach ($tokens as $tokenEl) {
			$tokenEl[0] = trim($tokenEl[0]);
			if (strlen($tokenEl[0])) {
				$last_was_operator = $token === "\x01" || self::is_operator($token) || $token === '(';
				if ($last_was_operator) {
					if (self::allowed_unary($tokenEl[0])) {
						$tokenEl[0] .= 'u';
					} elseif (self::is_operator($tokenEl[0])) {
						throw new PriceBookFormulaParserException($tokenEl[0] . ' cannot be used as unary operator', $tokenEl[1]);
					}
				}
			} else {
				continue;
			}
			list($token, $offset) = $tokenEl;
			if (is_numeric($token) || preg_match('~^[a-z_][a-z0-9_]*$~i', $token)) {
				if (!is_numeric($token))
					if (!in_array($token, self::$allowed_vars))
						throw new PriceBookFormulaParserException('Unknown variable ' . $token, $tokenEl[1]);
				$output[] = $tokenEl;
				continue;
			}
			if (self::is_operator($token)) {
				if (!self::is_unary($token)) {
					while (($t2 =  end($stack)) !== false) {
						list($o2, $unused) = $t2;
						if (!self::is_operator($o2)) break;
						$la = self::isLeftAssociative($token);
						$p1 = self::precedence($token);
						$p2 = self::precedence($o2);
						if (($la && $p1 <= $p2) || ($p1 < $p2)) {
							$output[] = array_pop($stack);
						} else {
							break;
						}
					}
				}
				$stack[] = $tokenEl;
				continue;
			}
			if ($token === '(') {
				$stack[] = $tokenEl;
				continue;
			}
			if ($token === ')') {
				$found = false;
				while (count($stack)) {
					list($t2, $unused) = end($stack);
					if ($t2 === '(') {
						$found = true;
						break;
					} else {
						$output[] = array_pop($stack);
					}
				}
				if (!$found) {
					throw new PriceBookFormulaParserException('Mismatched parethesis', $tokenEl[1]);
				}
				array_pop($stack);
				continue;
			}
			throw new PriceBookFormulaParserException('Invalid token ', $tokenEl[1]);
		}

		while (count($stack)) {
			list($t2, $o) = array_pop($stack);
			if ($t2 === '(' || $t2 === ')')
				throw new PriceBookFormulaParserException('Mismatched parethesis', $o);
			$output[] = array($t2, $o);
		}

		return $output;
	}

	private static function is_operator($token) {
		return $token === '+' 
			|| $token === '-' 
			|| $token === '*' 
			|| $token === '/' 
			|| $token === '^' 
			|| $token === '+u' 
			|| $token === '-u'
			;
	}

	private static function is_binary($token) {
		return $token === '+' 
			|| $token === '-' 
			|| $token === '*' 
			|| $token === '/' 
			|| $token === '^' 
			;
	}

	private static function allowed_unary($token) {
		return $token === '+' 
			|| $token === '-' 
			;
	}

	private static function is_unary($token) {
		return $token === '+u' 
			|| $token === '-u' 
			;
	}

	private static function isLeftAssociative($op) {
		return $op === '+' || $op === '-' || $op === '+u' || $op === '-u';
	}

	private static function precedence($op) {
		switch ($op) {
		case '^' : 
			return 5;
		case '+u' :
		case '-u' :
			return 4;
		case '*' : 
		case '/' : 
			return 3;
		case '+' :
		case '-' :
			return 2;
		}
		return 1;
	}
}

class PriceBookFormulaParserException extends Exception 
{
	public function __construct($message, $pos = null)
	{
		if ($pos !== null) {
			$message .= ' at position ' . ($pos+1);
		}
		parent::__construct($message);
	}
}

