<?php


use OneCRM\NotariusClient\Workflow;
use OneCRM\NotariusClient\ClientOptions;
use OneCRM\NotariusClient\Client;
use GuzzleHttp\Psr7\Utils;

if(!defined('sugarEntry') || !sugarEntry) die('Not A Valid Entry Point');

require_once'include/upload_file.php';

class NotariusWebhookProcessor {
	protected $project;
	protected $document;
	protected $projectUpd;
	protected $documentUpd;
	protected $projectUpdated = false;
	protected $documentUpdated = false;

	public function process() {
		$this->verifyProjectId();
		$notification = json_decode(file_get_contents("php://input"), true);
		if (!is_array($notification))
			$this->badRequest();
		if (empty($notification['event']) || !is_string($notification['event']))
			$this->badRequest();
		$this->loadRecords();
		$method = 'process' . join('', array_map('ucfirst', explode('.', $notification['event'])));
		if (is_callable([$this, $method])) {
			$this->$method($notification);
		}
	}

	protected function badRequest() {
        header('HTTP/1.1 400 Bad request');
		exit;
	}

	protected function notFound() {
        header('HTTP/1.1 404 Not found');
		exit;
	}

	protected function verifyProjectId() {
		if (empty($_GET['p']) || empty($_GET['s']))
			$this->badRequest();
		if (sha1($_GET['p'] . '_' . AppConfig::unique_key()) !== $_GET['s'])
			$this->badRequest();
	}

	protected function loadRecords() {
		$projectID = $_GET['p'];
		$project = ListQuery::quick_fetch('SigningProject', $projectID);
		if (!$project)
			$this->notFound();
		$docId = $project->getField('linked_document_id');
		$lq = new ListQuery('Document');
		$lq->addField('revision_filename');
		$document = $lq->queryRecord($docId);
		if (!$document)
			$this->notFound();
		$this->project = $project;
		$this->document = $document;
		$this->projectUpd = RowUpdate::for_result($project);
		$this->documentUpd = RowUpdate::for_result($document);
	}

	protected function processWorkflowCreated($notification) {
		$this->updateWorkflow($notification['context']['workflowId']);
	}

	protected function processWorkflowSignatureCompleted($notification) {
		$this->updateWorkflow($notification['context']['workflowId']);
	}

	protected function processWorkflowUserDeleted($notification) {
		$this->markWorkflowDeleted();
	}
	protected function processWorkflowLaunched($notification) {
		$this->updateWorkflow($notification['context']['workflowId']);
	}

	protected function processWorkflowUpdated($notification) {
		$this->updateWorkflow($notification['context']['workflowId']);
	}

	protected function processWorkflowCompleted($notification) {
		$this->updateWorkflow($notification['context']['workflowId']);
	}

	protected function processWorkflowDeclined($notification) {
		$this->updateWorkflow($notification['context']['workflowId']);
	}

	protected function processWorkflowSignatureDeclined($notification) {
		$this->updateWorkflow($notification['context']['workflowId']);
	}

	protected function processWorkflowExpired($notification) {
		$this->updateWorkflow($notification['context']['workflowId']);
	}

	protected function processWorkflowDeleted($notification) {
		$this->markWorkflowDeleted();
	}

	protected function markWorkflowDeleted() {
	}

	protected function updateWorkflow($id) {
		$opts = new ClientOptions;
		$sandbox = AppConfig::setting("site.notarius.sandbox_uri");
		if ($sandbox)
			$opts = $opts->withSandbox($sandbox);

		$opts = $opts
			->withAuthKey(AppConfig::setting('site.notarius.auth_key', ''))
			->withSecret(AppConfig::setting('site.notarius.auth_secret', ''))
		;

		$opts = $opts->build();
		$client = new Client($opts);
		$client->authenticate();

		$w = $client->getWorkflow($id);
		$wData = $w->toArray();

		$this->projectUpd->set([
			'workflow_status' => array_get_default($wData, 'status', 0),
			'parallel_actions' => array_get_default($wData, 'parallelActions', 0) ? 1 : 0,
			'expires_on' => gmdate('Y-m-d H:i:s', strtotime(array_get_default($wData, 'expiresOn'))),
		]);

		$this->projectUpd->set('project_source', json_encode($wData));
		$this->projectUpd->set('workflow_id', $id);

		if ($w->status == 2 && !$this->project->getField('audit_trail')) {
			$this->downloadAuditTrail($client, $w->id);
		}

		if ($w->status == 2 && $this->document->getField('status_id') != 'Approved') {
			$this->downloadFinalDocument($client, $w->id, $w->documents[0]->documentId);
		}


		if ($this->documentUpdated) {
			$this->documentUpd->save();
		}
		$this->projectUpd->save();

	}

	protected function downloadFinalDocument($client, $wfId, $docId) {
		$filename = basename($this->document->getField('revision_filename'));
		$stream = $client->downloadSingleDocument($wfId, $docId);
		$revision = RowUpdate::blank_for_model('DocumentRevision');

		$fl = new UploadFile('');
		$fl->use_soap = true;
		$fl->stored_file_name = $filename;
		$fl->create_stored_filename();
		$upload_dir = AppConfig::upload_dir();
		$dst = $fl->stored_file_name;

		$f = fopen($upload_dir . $dst, "w");
		Utils::copyToStream($stream, Utils::streamFor($f));
		// not closing the file, because Psr7 Stream will close it when going out of scope

		$revision->set([
			'id' => create_guid(),
			'revision' => 'Final',
			'filename' => $fl->get_stored_file_name(),
			'file_mime_type' => 'application/pdf',
			'file_ext' => 'pdf',
			'document_id' => $this->document->getField('id'),
			'change_log' => translate('LBL_CHANGE_LOG_FINAL', 'SigningProjects'),
		]);
		$revision->save();

		$this->documentUpd->set('document_revision_id', $revision->getField('id'));
		$this->documentUpd->set('status_id', 'Approved');
		$this->documentUpdated = true;
	}

	protected function downloadAuditTrail($client, $wfId) {
		$stream = $client->downloadAuditTrail($wfId);
		$fl = new UploadFile('');
		$fl->use_soap = true;
		$fl->stored_file_name = 'audit_trail.pdf';
		$fl->create_stored_filename();
		$upload_dir = AppConfig::upload_dir();
		$dst = $fl->stored_file_name;

		$f = fopen($upload_dir . $dst, "w");
		Utils::copyToStream($stream, Utils::streamFor($f));
		// not closing the file, because Psr7 Stream will close it when going out of scope
		
		$this->projectUpd->set([
			'audit_trail' => $fl->get_stored_file_name(),
		]);
	}
}
