<?php

namespace OneCRM\API\Subscription;
require_once('modules/Subscriptions/RingBuffer.php');

use OneCRM\API\Endpoint;
use Psr\Http\Message\ServerRequestInterface as Request;
use Slim\Http\Response;
use ListQuery;

/**
 * Class BaseEndpoint
 * @package OneCRM\API\Subscription
 *
 * this is the place for duplicated code between the /subscription/* Endpoints
 * in particular the authentication, and logging.
 */
abstract class BaseEndpoint extends Endpoint
{

    abstract function descriptorGet();
    abstract function descriptorPost();
    /**
     * @return bool always true to enable our custom login mech
     */
    public function providesCustomAuthMethod()
    {
        return true;
    }

    /**
     * Authenticate using credentials stored in the PaymentProcessor Model
     *
     * https://demo.1crmcloud.com/api.php/subscription/events/{processor}
     * processor must match a PaymentProcessor name field
     *
     * @param  Request  $request    : the request data passed in
     * @return bool|Request         : return Request object on success
     */
    public function checkCustomAuthentication(Request $request)
    {
        $auth = false;

        // try Basic / API Client Secret first:
        if (isset($_SERVER['PHP_AUTH_USER']) && isset($_SERVER['PHP_AUTH_PW'])) {
            global $db, $log;
            unset($GLOBALS['login_error']);
            // NOTE: $this->params holds the path params.  $request does not.
            $rows = ListQuery::quick_fetch_all('PaymentProcessor',
                ['webhook_user', 'webhook_pass', 'webhook_ssl'], ['filters' => [
                    ['field' => 'name', 'value' => $this->params['processor']],
                    ['field' => 'active', 'value' => 'Active'],
                ]],true
            );
            if ($rows){
                $row=reset($rows);
                $user = $row['webhook_user'];
                $pass = $row['webhook_pass'];
                $ssl = $row['webhook_ssl'];
                if ((!$ssl || ($ssl && $_SERVER["HTTPS"] === 'on')) &&
                    $user === $_SERVER['PHP_AUTH_USER'] && $pass === sha1($_SERVER['PHP_AUTH_PW'])) {
                    $auth = true;
                }
            }
        } else {
            // TODO: try OAuth2;
            return false;
        }

        if ($auth === true) {
            // fake admin login with all access:
			global $current_user;
			$sys_user_id = \AppConfig::setting('subscriptions.chargebee.webhook_system_user');
			$current_user = new \User();
			if ($sys_user_id) {
				if (!$current_user->retrieve($sys_user_id)) {
            		$current_user->getSystemUser();
				}
			} else {
            	$current_user->getSystemUser();
			}
            return $request->withAttribute('oauth_scopes', ['read', 'write', 'profile']);
        } else {
            return false;
        }
    }

    public function providesQueryLog(){
        return true;
    }

    /**
     * @param array $params     : includes path params
     * @param Request $request
     * @param null $logEntry    : if not null this is a target row id to update
     * @param Response $result  : if not null this the API response
     * @return bool|int         : if a new entry was created return the row id
     */
    public function queryLog($params,Request $request,$logEntry=null,$result=null){
		// AD: Disabled for now. RingBuffer has a problem with transactions,
		// it never commits it, making other DB connections lock
		return true;

        $RB = new RingBuffer($GLOBALS['db'], 'subscription_ring_buffer', 20000);
        if (is_null($logEntry)) {
            // log the payload and processor:
            return $RB->add_entry($request->getParsedBody(), 'events/' . $params['processor'], '');
        } else {
            return $RB->update_entry($logEntry,null,null,
                $result->getStatusCode().'/'.$result->getReasonPhrase().'/'.$result->getBody()
            );
        }
    }

    /**
     * @return array allowed HTTP access methods
     */
    public function allowedMethods()
    {
        return ['post', 'get',];
    }

    /**
     * @param $method
     * @return array scopes required to access the method for the endpoint
     */
    public function requiredScopes($method)
    {
        return ['read', 'write', 'profile',];
    }

    /**
     * @param string $method routes documentation calls for the given method
     * @return mixed structured documetentation record.
     */
    public function descriptor($method)
    {
        switch ($method) {
            case 'post':
                return $this->descriptorPost();
                break;
            case 'get':
                return $this->descriptorGet();
                break;
        }
        return false;
    }

    /**
     * @return bool always false.
     */
    public function isAuthEndpoint()
    {
        return false;
    }
}
