<?php

namespace OneCRM\API\Subscription;

use OneCRM\API\Endpoint;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use OneCRM\API\Errors;

class CreateSubscription  extends Endpoint {

    public function endpoint() {
        return '/subscription/signup';
    }

	public function allowedMethods() {
        return ['post'];
    }

    public function requiredScopes($method) {
        return ['read', 'write'];
    }

    public function descriptor($method)
    {
        switch ($method) {
            case 'post':
                return $this->descriptorPost();
                break;
        }
        return false;
    }

	public function descriptorPost() {
        static $desc = [
			'hidden' => true,
            'title' => 'Create a new subscription',
            'docs' => 'Creates a subscription (and customer, if required)',
            'params' => [
				'subscription_id' => [
					'type' => 'String',
					'location' => 'body',
				],
                'subscription' => [
                    'type' => 'Object',
                    'location' => 'body',
                    'docs' => 'Subscription data',
                ],
                'contact_id' => [
                    'type' => 'String',
                    'location' => 'body',
                    'docs' => 'Customer\'s contact ID.',
                ],
                'account_id' => [
                    'type' => 'String',
                    'location' => 'body',
                    'docs' => 'Customer\'s account ID.',
                ],
                'customer' => [
					'type' => 'Object',
                    'location' => 'body',
                    'docs' => 'Customer\'s data. Will be used for subscription, and for customer if a new customer is to be created',
                ],
            ],
            'return' => [
            ],
        ];
        return $desc;
	}

	protected function post($input, Response $res, Request $req) {
		require_once 'modules/PaymentProcessors/PaymentProcessor.php';
		$pp = \PaymentProcessor::getDefault();
		if (!$pp)
           	throw new Errors\InternalError("Payment processor not found");
		try {
			$account_id = array_get_default($input, 'account_id');
			$contact_id = array_get_default($input, 'contact_id');
			$customer_id = null;
			if (!$account_id && $contact_id) {
				$contact = \ListQuery::quick_fetch_row('Contact', $contact_id);
				if ($contact) {
					$account_id = array_get_default($contact, 'primary_account_id');
				}
			}
			$account = null;
			if ($account_id) {
				$account = \ListQuery::quick_fetch_row('Account', $account_id);
				if ($account) {
					$customer_id = array_get_default($account, 'pp_customer_id');
				}
			}
			if (!$customer_id) {
				$data = [
					'name' => array_get_default($input['customer'], 'account_name'),
					'first_name' => array_get_default($input['customer'], 'first_name'),
					'last_name' => array_get_default($input['customer'], 'last_name'),
					'email1' => array_get_default($input['customer'], 'email'),
					'id' => array_get_default($input, 'account_id'),
					'primary_contact_id' => array_get_default($input, 'contact_id'),
				];
				if ($result = $pp->sync_record('Customer', $data)){
					$customer_id = array_get_default($result, 'id');
				}
			}
			if ($customer_id) {
				$plan = \ListQuery::quick_fetch('Plan', $input['subscription']['plan']);
				if ($plan) {
					$coupons = [];
					if ($account_id && $account && !empty($account['customer_level_id'])) {
						$level = \ListQuery::quick_fetch_row('CustomerLevel', $account['customer_level_id']);
						if ($level) {
							$customer_level = \CustomerLevel::formatLevel($level);
							if (!empty($customer_level['discounts']['default']))
								$coupons[] = $customer_level['discounts']['default']['pp_coupon_id'];
							foreach ($customer_level['discounts']['per_plan'] as $perplan) {
								if ($perplan['plan_id'] == $plan->getField('id')) {
									if (empty($perplan['keep_default']))
										$coupons = [];
									$coupons[] = $perplan['coupon']['pp_coupon_id'];
								}
							}
						}
					}

					$data = [
						'pp_plan_id' => $plan->getField('pp_plan_id'),
						'pp_customer_id' => $customer_id,
						'quantity' => array_get_default($input['subscription'], 'qty'),
						'payment_source_id' => $input['customer']['payment_method'],
					];

					$coupons = array_unique($coupons);
					foreach ($coupons as $coupon_id)
						$data['coupon_ids'][] = $coupon_id;

					if (!empty($input['subscription']['options']))
						$data['onecrm_options'] = $input['subscription']['options'];
					if (!empty($input['subscription']['addons'])) {
						$i = 0;
						foreach ($input['subscription']['addons'] as $aid => $adata) {
							$addon = \ListQuery::quick_fetch_row('Addon', $aid);
							if ($addon) {
								$data["addons"]['id'][$i] = $addon['pp_addon_id'];
								if (!empty($adata['qty']))
									$data["addons"]['quantity'][$i] = $adata['qty'];
								$i++;
							}
						}
					}
					if (\AppConfig::setting('subscriptions.edit_invoice_notes')) {
						if (isset($input['subscription']['invoice_notes'])) {
							$data['invoice_notes'] = $input['subscription']['invoice_notes'];
						}
					}
					$data['pp_subs_id'] = array_get_default($input, 'subscription_id');
					$result = $pp->sync_record('Subscription', $data);
				}
			}
		} catch (\Exception $e) {
			return $result = ['error' => $e->getMessage()];
		}
		$GLOBALS['log']->fatal($result);

		$res->getBody()->write(json_encode($result));
		return $res->withHeader('Content-Type', 'application/json');
	}

}

