<?php

namespace OneCRM\API\Subscription;
require_once('modules/Subscriptions/Subscription.php');


use OneCRM\API\Endpoints;
use OneCRM\API\Errors;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;


/*
 * NOTE: get /subscription/data/Customer is an alias for /subscription/data/Account
 *      post /subscription/data/Customer creates an Account and Primary Contact
 *
 * For all records, if payment_processor_id is passed then the record will be synced to the provider on post/patch
 */

class Data extends Endpoints\Data\ListEndpoint
{

    static $allowed_modules = [
        'SubscriptionPlans' => ['get'],
        'SubscriptionAddons' => ['get'],
        'Coupons' => ['get'],
        'SubscriptionOptions' => ['get'],
        'PaymentProcessors' => ['get'],
        'Subscriptions' => ['get', 'post', 'patch'],
        'Invoice' => ['get'],
        'PaymentMethods' => ['get'],
        'Accounts' => ['get', 'post', 'patch'],
        'Contacts' => ['get', 'post', 'patch'],
        'Customer' => ['get', 'post', 'patch'],
    ];
    static $allowed_fields = [ // true means all fields
        'Plan' => [
            'id', 'name', 'description', 'pp_plan_id', 'addons', 'coupons', 'options',
            'qty_label', 'qty_min', 'qty_max', 'billing_cycle', 'pricing_model', 'pricing_tiers',
            'currency', 'currency_code', 'setup', 'amount', 'active',
            'product.image_url', 'product.thumbnail_url', 'product.url', 'product.description_long',
        ],
        'Addon' => [
            'id', 'name', 'description', 'pp_addon_id', 'coupons', 'options',
            'qty_label', 'global', 'billing_cycle', 'pricing_model', 'pricing_tiers',
            'currency', 'currency_code', 'amount', 'active', 'global',
            'product.image_url', 'product.thumbnail_url', 'product.url', "product.description_long",
        ],
        'Coupon' => [
            'id', 'name', 'pp_coupon_id', 'discount_amount', 'discount_type', 'discount_percentage',
            'addon_mode', 'plan_mode', 'currency', 'currency_code', 'active', 'plan_mode', 'addon_mode',
        ],
        'Option' => [
            'option_number', 'name', 'label', 'placeholder', 'input', 'description',
            'type', 'meta', 'len', 'required', 'updateable', 'options', 'default',
        ],
        'PaymentProcessor' => ['id', 'name', 'deleted', 'active',],
        'Subscription' => true,
        'Invoice' => true,
        'PaymentMethod' => true,
        'Account' => true,
        'Contact' => true,
        'Customer' => [
            'id', 'name', 'pp_customer_id', 'primary_contact_id', 'date_entered', 'date_modified', 'deleted',
            'currency', 'email1', 'website', 'phone_home', 'first_name','last_name','salutation','title',
            'billing_address_street','billing_address_city','billing_address_state','billing_address_country','billing_address_postalcode',
            'primary_address_street','primary_address_city','primary_address_state','primary_address_country','primary_address_postalcode',
        ],
    ];

    static $dedup_lookup = [ // method, return fields,
        'Customer' => ['find_account', ['id', 'pp_customer_id', 'payment_processor_id']],
    ];

    public function endpoint()
    {
        return '/subscription/data/{object_name}';
    }

    public function descriptor($method)
    {
        if ($method == 'get')
            return $this->descriptorGet();
        elseif ($method == 'post')
            return $this->descriptorPost();
    }


    private function descriptorGet()
    {
        static $desc = [
			'hidden' => true,
            'title' => 'Get list of records',
            'docs' => 'Retrieve a list of records of specified `object`',
            'params' => [
                'object_name' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Object to query',
                    'required' => true,
                ],
                'fields' => [
                    'type' => [
                        'type' => 'Array',
                        'element_type' => [
                            'type' => 'String',
                        ]
                    ],
                    'location' => 'query',
                    'docs' => 'Array of field names to fetch. When omitted, a limited number of fields are returned, depending on object. Record ID is guaranteed to be returned, and for most objects, `name` and/or `_display` fields.',
                ],
                'filter_text' => [
                    'location' => 'query',
                    'type' => 'String',
                    'docs' => 'Generic search string. Fields involved in search depend on object',
                ],
                'filters' => [
                    'location' => 'query',
                    'type' => 'Object',
                    'docs' => 'Filters to apply. To get list of available filters for a object, use [metadata](#group_meta)',
                    'value_type' => [
                        'type' => 'String',
                    ]
                ],
                'order' => [
                    'location' => 'query',
                    'type' => 'String',
                    'docs' => 'Sort order',
                ],
                'offset' => [
                    'location' => 'query',
                    'type' => 'Int',
                    'min' => 0,
                    'default' => 0,
                    'docs' => 'Offset in the list to start retrieval from',
                ],
                'limit' => [
                    'location' => 'query',
                    'type' => 'Int',
                    'min' => 1,
                    'max' => 200,
                    'default' => 20,
                    'docs' => 'Limits number of records returned',
                ],

            ],
            'return' => [
                'records' => [
                    'type' => [
                        'type' => 'Array',
                        'element_type' => [
                            'type' => 'Object',
                        ]
                    ],
                    'docs' => 'Array of retrieved records',
                    'required' => true,
                ],
                'total_results' => [
                    'type' => 'Int',
                    'docs' => 'Total number of results, without taking offset and limit in account',
                    'required' => true,
                ]
            ]
        ];
        return $desc;
    }

    private function descriptorPost()
    {
        static $desc = [
			'hidden' => true,
            'title' => 'Create records',
            'docs' => 'Create a record of specified `object`',
            'params' => [
                'object_name' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Model',
                    'required' => true,
                ],
                'data' => [
                    'location' => 'body',
                    'type' => [
                        'type' => 'Object',
                    ],
                    'docs' => 'An object with keys matching field names to update. Fields not defined in this object are not modified',
                    'required' => true,
                ],
                'sync' => [
                    'location' => 'body',
                    'type' => 'Bool',
                    'docs' => 'Set false to prevent syncing records to the PaymentProcessor',
                    'required' => false,
                ],
            ],
            'return' => [
                'id' => [
                    'type' => 'String',
                    'required' => true,
                    'docs' => 'ID of created record',
                ],
            ],
        ];
        return $desc;
    }

    /**
     * @param array $input
     * @param Response $res
     * @param Request $req
     * @return Response
     * @throws Errors\Forbidden
     * @throws Errors\InternalError
     * @throws Errors\NotFound
     */
    protected function get($input, $res, $req)
    {
        // we do not apply ACL filters to query explicitly,
        // because that is done dy ListFormatter

        $model = $input['object_name'];
        if ($model === 'Customer') $model = 'Account';

        $module = \AppConfig::module_for_model($model);
        if (!$module)
            throw new Errors\NotFound("Model $model not found");
        self::check_module_access($module, 'get');

        require_once('include/ListView/ListFormatter.php');
        $fmt = new \ListFormatter($model);
        $lq =& $fmt->getQuery();

        $filter =& $fmt->getFilterForm();
        $filter->loadFilterLayout('Standard');

        $filter_data = array_get_default($input, 'filters', array());
        if (isset($input['filter_text']))
            $filter_data['filter_text'] = $input['filter_text'];

        if (@$filter_data['_lq']){
            $lq->addFilterClauses($filter_data['_lq']);
            $lq->disable_literal();
            unset($filter_data['_lq']);
        }

        $fmt->loadFilter($filter_data);

        $fieldNames=self::prefilter_fields($model,$input['fields']);
        $lq->addFields($fieldNames);
        $lq->addField('name');
        $lq->addField('_display');

        if (isset($input['order']))
            $lq->setOrderBy($input['order']);
        $this->applyPortalAcl($lq, $req);

        $result = $fmt->getQueryResult($input['offset'], $input['limit']);
        if (!$result || $result->failed)
            throw new Errors\InternalError;

        self::filterExposedFields($result->rows, $model);
        $ret = [
            'records' => array_values($result->rows),
            'total_results' => $result->total_count,
        ];
        if (\AppConfig::module_primary_bean($module) === $model) {
            $ret['module'] = $module;
        }
        $ret['query'] = $result->query;

		$res->getBody()->write(json_encode($ret));
		return $res->withHeader('Content-Type', 'application/json');

    }

    /**
     * @param string $model
     * @param array|true $fields requested fields
     * @return array
     */
    public static function prefilter_fields($model, $fields)
    {
        $allowed_fields = self::$allowed_fields[$model];
        if ($allowed_fields === null) return [];
        if ($allowed_fields === true) {
            $model = new \ModelDef($model);
            return $model->getColumnFields();
        }
        if ($fields==[true]){
            return $allowed_fields;
        }
        return array_intersect($allowed_fields,(array)$fields);
    }

    /**
     * @param $module
     * @param string $method
     * @throws Errors\Forbidden
     */
    protected static function check_module_access($module, $method = null)
    {
        if (array_search($method, (array)@self::$allowed_modules[$module]) === false)
            throw new Errors\Forbidden('Forbidden: ' . $module);
    }

    private function applyOverrideFilters($model, $filter)
    {
        if ($model == 'Note') {
            $filter->override_filters = ['email_attachment' => ['field' => 'email_attachment', 'value' => 0]];
        }
    }

    public static function filterExposedFields(&$rows, $model)
    {
        $modelDef = new \ModelDef($model);
        if (($row = reset($rows)) && isset($row['currency_id'])) {
            $currency=\AppConfig::db_all_objects('Currency')[$row['currency_id']];
            $add['currency_code'] = $currency->iso4217;
        } else $add = [];
        foreach ($rows as &$row) {
            if (!is_array($row)) continue;
            if (@self::$allowed_fields[$model] !== true) {
                $clip = array_flip((array)@self::$allowed_fields[$model]);
                $row = array_intersect_key($row, $clip);
            }
            foreach ($row as $k => $v) {
                $f = $modelDef->getFieldDefinition($k);
                if (!array_get_default($f, 'api_exposed', true))
                    unset($row[$k]);
            }
            $row += $add;
        }
    }

    /**
     * @param array $input
     * @param Response $res
     * @param Request $req
     * @return Response
     * @throws Errors\Forbidden
     * @throws Errors\NotFound
     * @throws Errors\BadRequest
     */
    protected function post($input, $res, $req)
    {
        $model = $input['object_name'];
        $this->dedup($input);
        if ($model === 'Customer') {
            $contact = $input;
            unset($contact['data']['id']);
            $contact = $this->post_model('Contact', $contact, $req);
            $input['data']['primary_contact_id'] = $contact['id'];
            $ret = $this->post_model('Account', $input, $req);
        } else {
            $ret = $this->post_model($model, $input, $req);
        }
        if (@$input['sync'] !== false) {
            $input['id'] = $ret['id'];
            self::syncProcessorRecord($model, $input, $ret['update']);
        }
        unset($ret['update']);

		$res->getBody()->write(json_encode($ret));
		return $res->withHeader('Content-Type', 'application/json');
    }

    /**
     * @param $input
     * @throws Errors\BadRequest
     */
    protected function dedup($input)
    {
        if ($rec = @self::$dedup_lookup[$input['object_name']]) {
            $func = '\\Subscription::' . $rec[0];
            if (is_callable($func) && $res = $func($input['data'], $rec[1], @$rec[2])) {
                $flat = json_encode($res);
                throw new Errors\BadRequest($input['object_name'] . ' exists: ' . $flat);
            }
        }
    }

    /**
     * @param string $model
     * @param array $input
     * @param Request $req
     * @return array
     * @throws Errors\Forbidden
     * @throws Errors\NotFound
     */
    protected function post_model($model, $input, $req)
    {
        $this->disabledForPortal($req);
        $module = \AppConfig::module_for_model($model);
        if (!$module)
            throw new Errors\NotFound("Model $model not found");
        self::check_module_access($module, 'post');
        $allowed = \ACLController::checkAccess($module, 'edit');
        if (!$allowed) {
            throw new Errors\Forbidden("Access denied");
        }

        $update = \RowUpdate::blank_for_model($model);
        $_REQUEST = $input['data'];
        $update->loadRequest('', false);
        foreach ($input['data'] as $k => $v) {
            if (strpos($k, "_social_") === 0) {
                $update->set($k, $v);
            }
        }
        if (isset($input['X_ONECRM_APPLICATION'])) {
            $update->setRelatedData('X_ONECRM_APPLICATION', $input['X_ONECRM_APPLICATION']);
        }
        $update->save();
        $ret = [
            'id' => $update->getPrimaryKeyValue(),
            'update' => $update,
        ];
        return $ret;
    }

    /**
     * @param $model
     * @param $input
     * @return bool
     * @throws Errors\BadRequest
     */
    static function syncProcessorRecord($model, $input, \RowUpdate $update)
    {
        if (!$payment_processor_id = $input['data']['payment_processor_id']) return false;
        $processor = \ListQuery::quick_fetch_row('PaymentProcessor', $payment_processor_id, ['name']);
        $name = $processor['name'];
        $filename = "modules/PaymentProcessor$name/$name.php";
        $className = "\\OneCRM\\PaymentProcessors\\$name";
        if (is_file($filename)) {
            try {
                require_once($filename);
                /** @var \OneCRM\PaymentProcessors\ChargeBee $className */
                if (class_exists($className) && ($className::$pp_sync_records)[$model]) {
                    $pp = new $className();
                    /** @var \OneCRM\PaymentProcessors\ChargeBee $pp */
                    if ($res = $pp->sync_record($model, $input['data'])){
                        // set the payemnt processor unique ID for the model:
                        $update->set($res['key'],$res['id']);
                        $update->save();
                    }
                }
            } catch (\Exception $e) {
                $trace = explode(PHP_EOL, stripslashes($e->getTraceAsString()));
                $trace = implode(PHP_EOL, array_slice($trace, 0, 3));
                $trace = $e->getFile().'('.$e->getLine().'):'.$e->getMessage().PHP_EOL.$trace;
                throw new Errors\BadRequest($trace);
            }
        } else {
            throw new Errors\BadRequest('Missing Processor: '.$processor['name']);
        }
        return false;
    }

}


