<?php

namespace OneCRM\API\Subscription;
use OneCRM\API\Endpoints;
use OneCRM\API\Errors;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;

/*
 * NOTE: /subscription/data/Customer/{id} is derived from Account[id] + Contact[Account.primary_contact]
 */
class DataRecord extends Endpoints\Data\Record {

    public function endpoint() {
        return '/subscription/data/{object_name}/{id}';
    }

    public function descriptor($method) {
        if ($method == 'get')
            return $this->descriptorGet();
        elseif ($method == 'patch')
            return $this->descriptorPatch();
        elseif ($method == 'delete')
            return $this->descriptorDelete();
    }

    public function requiredScopes($method) {
        if ($method == 'get')
            return ['read'];
        return ['write'];
    }

    public function allowedMethods() {
        return ['get', 'patch', 'delete'];
    }

    private function descriptorGet() {
        static $desc = [
			'hidden' => true,
            'title' => 'Get single record',
            'docs' => 'Retrieve single record of specified `object`',
            'params' => [
                'object_name' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Model to query',
                    'required' => true,
                ],
                'id' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Record ID',
                    'required' => true,
                ],
                'sync' => [
                    'location' => 'body',
                    'type' => 'Bool',
                    'docs' => 'Set false to prevent syncing records to the PaymentProcessor',
                    'required' => false,
                ],
                'fields' => [
                    'type' => [
                        'type' => 'Array',
                        'element_type' => [
                            'type' => 'String',
                        ]
                    ],
                    'location' => 'query',
                    'docs' => 'Array of field names to fetch. When omitted, all available fields are returned.',
                ],

            ],
            'return' => [
                'record' => [
                    'type' => [
                        'type' => 'Object',
                    ],
                    'docs' => 'Retrieved record',
                    'required' => true,
                ],
            ]
        ];
        return $desc;
    }

    private function descriptorPatch() {
        static $desc = [
			'hidden' => true,
            'title' => 'Update a record',
            'docs' => 'Update record belonging to specified `object`, identified by `id`',
            'params' => [
                'object_name' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Model to query',
                    'required' => true,
                ],
                'id' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Record ID',
                    'required' => true,
                ],
                'data' => [
                    'location' => 'body',
                    'type' => [
                        'type' => 'Object',
                    ],
                    'docs' => 'An object with keys matching field names to update. Fields not defined in this object are not modified',
                    'required' => true,
                ],
                'create' => [
                    'location' => 'body',
                    'type' => [
                        'type' => 'Bool',
                    ],
                    'docs' => 'If true, record will be created if it does not exist',
                    'required' => false,
                ],
                'sync' => [
                    'location' => 'body',
                    'type' => 'Bool',
                    'docs' => 'Set false to prevent syncing records to the PaymentProcessor',
                    'required' => false,
                ],
            ],
            'return' => [
                'result' => [
                    'type' => 'Bool',
                    'docs' => 'Always true',
                    'required' => true,
                ],
            ]
        ];
        return $desc;
    }

    private function descriptorDelete() {
        static $desc = [
			'hidden' => true,
            'title' => 'Delete a record',
            'docs' => 'Delete record belonging to specified `object`, identified by `id`',
            'params' => [
                'object_name' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Model',
                    'required' => true,
                ],
                'id' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Record ID',
                    'required' => true,
                ],

            ],
            'return' => [
                'result' => [
                    'type' => 'Bool',
                    'docs' => 'True if deleted successfully',
                    'required' => true,
                ],
            ]
        ];
        return $desc;
    }

    /**
     * @param array $input
     * @param Response $res
     * @param Request $req
     * @return Response
     * @throws Errors\Forbidden|Errors\NotFound|Errors\BadRequest
     */
    protected function delete($input, $res, $req) {
        $this->disabledForPortal($req);
        $model = $input['object_name'];
        $id = $input['id'];
        $module = \AppConfig::module_for_model($model);
        if (!$module)
            throw new Errors\NotFound("Object $model not found");
        Data::check_module_access($module, 'delete');
        $record = \ListQuery::quick_fetch($model, $input['id'], true, ['acl_filters' => 'delete']);
        if (!$record)
            throw new Errors\NotFound;
        $update = \RowUpdate::for_result($record);
        $result = $update->markDeleted();
        if (@$input['sync'] !== false) {
            Data::syncProcessorRecord($model, $input, $update);
        }

		$res->getBody()->write(json_encode(['result' => !!$result]));
		return $res->withHeader('Content-Type', 'application/json');
    }
    /**
     * @param array $input
     * @param Response $res
     * @param Request $req
     * @return Response
     * @throws Errors\Forbidden
     * @throws Errors\NotFound
     */

    protected function get($input, $res, $req) {

        $model=$input['object_name'];
        switch($model){
            case 'Customer':
                $account = $this->get_model('Account', $input, $req);
                $contact = $this->get_model('Contact', ['id'=>@$account['record']['primary_contact_id']], $req);
                $account['record'] += $contact['record'];
                $account['module'] = 'Customer';
                $ret = $account;
                break;
            DEFAULT:
                $ret = $this->get_model($model, $input, $req);
                break;
        }

        if (@Data::$allowed_fields[$model] !==true){
            $clip=array_flip((array)@Data::$allowed_fields[$model]);
            $ret['record'] = array_intersect_key($ret['record'], $clip);
        }

		$res->getBody()->write(json_encode($ret));
		return $res->withHeader('Content-Type', 'application/json');
    }

    /**
     * @param $model
     * @param $input
     * @param Request $req
     * @return array
     * @throws Errors\Forbidden
     * @throws Errors\NotFound
     */
    protected function get_model($model, $input, $req) {
        require_once 'include/upload_file.php';
        $module = \AppConfig::module_for_model($model);
        $id = $input['id'];
        if (!$module)
            throw new Errors\NotFound("Object $model not found");
        Data::check_module_access($module, 'get');

        $fieldNames = Data::prefilter_fields($model, $input['fields']);
        $fields = array_combine($fieldNames, $fieldNames);

        $params = ['process_results' => false, 'acl_filters' => 'view'];
        $lq = new \ListQuery($model, $fields, $params);
        $lq->addAclFilter('view');
        $this->applyPortalAcl($lq, $req);
        $result = $lq->queryRecord($id);
        if ($result && !$result->failed) {
            $data = $result->row;
        } else {
            throw new Errors\NotFound;
        }

        foreach ($data as $k => $v) {
            $def = $lq->getField($k);
            if ($def['type'] == 'image') {
                $upload_dir = \UploadFile::replace_placeholders($def['upload_dir']);
                $filename_field = $def['filename'];
                $filename = array_get_default($data, $filename_field);
                if (!$filename && isset($def['alt_filename'])) {
                    $filename = array_get_default($data, $def['alt_filename']);
                }
                if ($filename) {
                    $value = \UploadFile::get_path($filename, '', $upload_dir);
                    $data[$filename_field] = $value;
                }
            }
            if (!array_get_default($def, 'api_exposed', true))
                unset($data[$k]);
        }
        if (is_array($fields) && array_search('_social_accounts', $fields) !== false) {
            require_once 'modules/SocialAccounts/SocialAccount.php';
            $accounts = \SocialAccount::getList($module, $id);
            if (!empty($accounts)) {
                foreach ($accounts as &$acc) {
                    if (isset($acc['params'])) {
                        $params = unserialize($acc['params']);
                        if ($params && isset($params['entered_url'])) {
                            $acc['entered_url'] = $params['entered_url'];
                        }
                    }
                }
                $data['_social_accounts'] = array_values($accounts);
            }
        }

        $ret = [
            'record' => $data
        ];
        Data::filterExposedFields($ret, $model);
        $module = \AppConfig::module_for_model($model);
        if (\AppConfig::module_primary_bean($module) === $model) {
            $ret['module'] = $module;
        }
        return $ret;
    }

    /**
     * @param array $input
     * @param Response $res
     * @param Request $req
     * @return Response
     * @throws Errors\Forbidden|Errors\NotFound|Errors\BadRequest
     */
    protected function patch($input, $res, $req) {
        $this->disabledForPortal($req);
        $model = $input['model'];
        $module = \AppConfig::module_for_model($model);
        if (!$module)
            throw new Errors\NotFound("Model $model not found");
        Data::check_module_access($module, 'patch');
        $record = \ListQuery::quick_fetch($model, $input['id'], true, ['acl_filters' => 'edit']);
        if (!$record && empty($input['create']))
            throw new Errors\NotFound;
        if ($record) {
            $update = \RowUpdate::for_result($record);
        } else {
            $update = \RowUpdate::blank_for_model($model);
            $update->set('id', $input['id']);
        }
        $_REQUEST = $input['data'];
        $update->loadRequest('', false);
        foreach ($input['data'] as $k => $v) {
            if (strpos($k, "_social_") === 0) {
                $update->set($k, $v);
            }
        }
        if (isset($input['X_ONECRM_APPLICATION'])) {
            $update->setRelatedData('X_ONECRM_APPLICATION', $input['X_ONECRM_APPLICATION']);
        }
        $update->save();
        if (@$input['sync'] !== false) {
            Data::syncProcessorRecord($model, $input, $update);
        }
        $ret = [
            'result' => true,
        ];

		$res->getBody()->write(json_encode($ret));
		return $res->withHeader('Content-Type', 'application/json');
    }

}


