<?php

namespace OneCRM\API\Subscription;
use OneCRM\API\Endpoints;
use OneCRM\API\Errors;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;

/*
 * NOTE: /subscription/data/Customer/{id}/{link} is an alias for Account
 *       /subscription/data/Customer/{id}/payment_methods derives from the Payment Processor
 */
class DataRelated extends Endpoints\Data\RelatedList {

    public function endpoint() {
        return '/subscription/data/{object_name}/{id}/{link}';
    }

    public function descriptor($method) {
        if ($method == 'get')
            return $this->descriptorGet();
        elseif ($method == 'post')
            return $this->descriptorPost();
    }

    private function descriptorGet() {
        static $desc = [
			'hidden' => true,
            'title' => 'Get list of related records',
            'docs' => 'Retrieve list of related records belonging to specified `object` and `id` via specific `link`',
            'params' => [
                'object_name' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'object to query',
                    'required' => true,
                ],
                'id' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => 1,
                    'docs' => 'Record ID',
                    'required' => true,
                ],
                'link' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => 1,
                    'docs' => 'Link name',
                    'required' => true,
                ],
                'fields' => [
                    'type' => [
                        'type' => 'Array',
                        'element_type' => [
                            'type' => 'String',
                        ]
                    ],
                    'location' => 'query',
                    'docs' => 'Array of field names to fetch. When omitted, a limited number of fields are returned, depending on object. Record ID is guaranteed to be returned, and for most objects, `name` and/or `_display` fields.',
                ],
                'filter_text' => [
                    'location' => 'query',
                    'type' => 'String',
                    'docs' => 'Generic search string. Fields involved in search depend on model',
                ],
                'filters' => [
                    'location' => 'query',
                    'type' => 'Object',
                    'docs' => 'Filters to apply. To get list of available filters for an object, use [metadata](#group_meta)',
                    'value_type' => [
                        'type' => 'String',
                    ]
                ],
                'order' => [
                    'location' => 'query',
                    'type' => 'String',
                    'docs' => 'Sort order',
                ],
                'offset' => [
                    'location' => 'query',
                    'type' => 'Int',
                    'min' => 0,
                    'default' => 0,
                    'docs' => 'Offset in the list to start retrieval from',
                ],
                'limit' => [
                    'location' => 'query',
                    'type' => 'Int',
                    'min' => 1,
                    'max' => 200,
                    'default' => 20,
                    'docs' => 'Limits number of records returned',
                ],

            ],
            'return' => [
                'records' => [
                    'type' => [
                        'type' => 'Array',
                        'element_type' => [
                            'type' => 'Object',
                        ]
                    ],
                    'docs' => 'Array of retrieved records',
                    'required' => true,
                ],
                'total_results' => [
                    'type' => 'Int',
                    'docs' => 'Total number of results, without taking offset and limit in account',
                    'required' => true,
                ]
            ]
        ];
        return $desc;
    }

    private function descriptorPost() {
        static $desc = [
			'hidden' => true,
            'title' => 'Add related records',
            'docs' => 'Add list of related records belonging to specified `object` and `id` via specific `link`. See [Working with data](#endpoint_dataRelatedList_post)',
            'params' => [
                'object_name' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Parent model',
                    'required' => true,
                ],
                'id' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => 1,
                    'docs' => 'Parent ID',
                    'required' => true,
                ],
                'link' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => 1,
                    'docs' => 'Link name',
                    'required' => true,
                ],
                'records' => [
                    'type' => [
                        'type' => 'Array',
                        'element_type' => [
                            'type' => 'String',
                            'min_len' => 1,
                        ]
                    ],
                    'location' => 'body',
                    'docs' => 'Array of related record IDs to be added to specified link',
                ],
                'records_with_data' => [
                    'type' => [
                        'type' => 'Array',
                        'element_type' => [
                            'type' => 'Object',
                        ],
                    ],
                    'location' => 'body',
                    'docs' => 'Array of objects, each representing additional data to be inserted into join table',
                ],

            ],
        ];
        return $desc;
    }


    /**
     * @param $input
     * @param Response $res
     * @param Request $req
     * @return Response
     * @throws Errors\Forbidden
     * @throws Errors\NotFound
     */
    protected function post($input, $res, $req) {
        $this->disabledForPortal($req);
        $model = $input['object_name'];
        if ($model==='Customer') $model='Account';
        $id = $input['id'];
        $link = $input['link'];
        $records = array_get_default($input, 'records', []);
        $records_with_data = array_get_default($input, 'records_with_data', []);
        $module = \AppConfig::module_for_model($model);
        if (!$module)
            throw new Errors\NotFound("Object $model not found");
        Data::check_module_access($module, 'post');
        $row = \ListQuery::quick_fetch($model, $id, true, ['acl_filters' => 'edit']);
        if (!$row || $row->failed)
            throw new Errors\NotFound("Record not found");
        $upd = \RowUpdate::for_result($row);
        $upd->addUpdateLinks($link, $records);
        $upd->addUpdateLinks($link, $records_with_data);

		$res->getBody()->write(json_encode(['result' => true]));
		return $res->withHeader('Content-Type', 'application/json');
    }

    /**
     * @param $input
     * @param Response $res
     * @param Request $req
     * @return Response
     * @throws Errors\Forbidden
     * @throws Errors\InternalError
     * @throws Errors\NotFound
     */
    protected function get($input, $res, $req) {
        $model = $input['object_name'];
        if ($model==='Customer') $model='Account';
        $module = \AppConfig::module_for_model($model);
        Data::check_module_access($module, 'get');
        if (!$module)
            throw new Errors\NotFound("Object $model not found");

        if ($model='Account' && $input['link']==='payment_methods'){
            // TODO: Poll Payment Processor for payment methods:
            $ret=[
                'records' => ['id' => '1', 'name' => 'Fake method 1'],
                'total_results' => 1,
            ];
        } else {
			require_once('include/ListView/ListFormatter.php');
			$fmt = new \ListFormatter($model, $input['link'], null, 'listview', ['parent_key' => $input['id']]);
			$lq =& $fmt->getQuery();

			$filter =& $fmt->getFilterForm();
			$filter->loadFilterLayout('Standard');

			$filter_data = array_get_default($input, 'filters', array());
			if (isset($input['filter_text']))
				$filter_data['filter_text'] = $input['filter_text'];


			$fmt->loadFilter($filter_data);


			if (!empty($input['fields']))
				$lq->addFields($input['fields']);
			$lq->addField('name');
			$lq->addField('_display');
			if (isset($input['order']))
				$lq->setOrderBy($input['order']);
			$this->applyPortalAcl($lq, $req);
			$result = $fmt->getQueryResult($input['offset'], $input['limit']);
			if (!$result || $result->failed)
				throw new Errors\InternalError;

			Data::filterExposedFields($result->rows, $model);
			$ret = [
				'records' => array_values($result->rows),
				'total_results' => $result->total_count,
			];
		}

		$res->getBody()->write(json_encode($ret));
		return $res->withHeader('Content-Type', 'application/json');

    }

}


