<?php

namespace OneCRM\API\Subscription;
require_once(__DIR__.'/BaseEndpoint.php');

use OneCRM\API\Errors;
use OneCRM\PaymentProcessors\PaymentProcessorInterface;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;

/**
 * Class Events
 * @package OneCRM\API\Subscription
 *
 * Routes the PaymentProcessor-initiated webhooks to Processor-specific handler class
 */
class Events extends BaseEndpoint
{
    /**
     * @return string endpoint URI
     */

    public function endpoint()
    {
        return '/subscription/events/{processor}';
    }

    /**
     * @param  array $params combined post/get parameters
     * @param  Response $response used to process response
     * @param  Request $request the request data passed in
     * @throws Errors\BadRequest if the event can't be processed
     * @return Response
     */
    public function post($params, $response, $request)
    {
        /* comment in to see debugging
        ini_set('display_errors', 1);
        ini_set('display_startup_errors', 1);
        error_reporting(E_ALL);
        ob_end_clean();/**/
        if (@$params["processor"]) {
            // processor name must not contain \ or /
            $processor = str_replace(['/', '\\'], '-', $params["processor"]);
            $classFile = 'modules/PaymentProcessor' . $processor . '/' . $processor . '.php';
			if (is_file($classFile)) {
                require_once($classFile);
                $fn = "PaymentProcessor$processor::instance";
                /** @var PaymentProcessorInterface $pp */
				$pp = $fn(@$params["processor"]);
				$fname = \AppConfig::cache_dir() . '/.subscriptions.lock';
				if (! ($fp = fopen($fname, 'w')) )
					throw new Errors\InternalError('Cannot create lock file');
				if (!flock($fp, LOCK_EX))
					throw new Errors\InternalError('Cannot obtain file lock');
				try {
					$result = $pp->processEvent($params, $response, $request);
				} finally {
					flock($fp, LOCK_UN);
					fclose($fp);
				}
                if ($result) {
					$response->getBody()->write(json_encode(['result' => 'OK']));
					return $response->withHeader('Content-Type', 'application/json');
                }
                throw new Errors\BadRequest("bad data format");
            }
            throw new Errors\BadRequest("processor not found");
        }
        throw new Errors\BadRequest("processor name missing");
    }

    /**
     * @param  array $params combined post/get parameters
     * @param  Response $response used to process response
     * @param  Request $request the request data passed in
     * @throws Errors\BadRequest if the event can't be processed
     * @return Response
     */
    public function get($params, $response, $request)
    {
        return $this->post($params, $response, $request);
    }

    /**
     * @return array Documentation descriptor for the POST access method
     */
    public function descriptorPost()
    {
        return [
			'hidden' => true,
            'title' => 'Subscription Manager POST Method',
            'docs' => '@+modules/Subscriptions/docs/manager.post',
            'params' => [
                'processor' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Payment Processor Name',
                    'required' => true,
                ],
            ],
            'return' => [
                '' => [
                    'type' => 'String',
                    'docs' => 'return code is "OK" if successful',
                    'required' => 'true',
                ],
            ],
        ];
    }

    /**
     * @return bool always false
     *
     * we don't document GET access but it is supported with same args as POST
     * violates REST but not all Processors have proper REST reporting.
     */
    public function descriptorGet()
    {
		return [
			'hidden' => true,
		];
    }

}
