<?php

namespace OneCRM\API\Subscription;
use OneCRM\API\Endpoints;
use OneCRM\API\Errors;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;

/*
 * NOTE: /subscription/meta/fields/Customer is derived from Account + Contact[Account.primary_contact]
 */
class Metadata extends Endpoints\Meta\Fields {

    public function endpoint() {
        return '/subscription/meta/fields/{object_name}';
    }

    public function descriptor($method) {
        if ($method == 'get')
            return $this->descriptorGet();
    }

    public function requiredScopes($method) {
        return ['read'];
    }

    public function allowedMethods() {
        return ['get'];
    }

    private function descriptorGet()
    {
        static $desc = [
			'hidden' => true,
            'title' => 'Get fields definition for a model',
            'docs' => 'documentation for subscription fields meta',
            'params' => [
                'object_name' => [
                    'location' => 'path',
                    'type' => 'String',
                    'min_len' => '1',
                    'docs' => 'Model to query',
                    'required' => true,
                ],
            ],
            'return' => [
                'fields' => [
                    'type' => 'Array',
                    'docs' => 'An array with fields definitions',
                    'required' => true,
                ],
                'filters' => [
                    'type' => 'Array',
                    'docs' => 'An array with filters definitions',
                    'required' => true,
                ]
            ]
        ];
        return $desc;
    }

    /**
     * @param array $input
     * @param Response $res
     * @param Request $req
     * @return Response
     * @throws Errors\NotFound|Errors\Forbidden
     */

    protected function get($input,$res,$req){
        $model = $input['object_name'];
        // Customer is an alias for the union of an Account with its Primary Contact
        if ($model==='Customer'){
            $contact_meta=$this->get_object_meta('Contact');
            $account_meta=$this->get_object_meta('Account');
            $meta =[
                'fields' => $account_meta['fields'] + $contact_meta['fields'],
                'filters' => $account_meta['filters'] + $contact_meta['filters'],
            ];
        } else {
            $meta=$this->get_object_meta($model);
        }
        foreach ($meta as &$target){
            // keep only the relevant fields:
            if (@DATA::$allowed_fields[$model] !==true){
                $clip=array_flip((array)@DATA::$allowed_fields[$model]);
                $target=array_intersect_key($target, $clip);
            }
        } unset($target);

		$res->getBody()->write(json_encode($meta));
		return $res->withHeader('Content-Type', 'application/json');
    }

    /**
     * @param $model
     * @return array
     * @throws Errors\Forbidden
     * @throws Errors\NotFound
     */
    protected function get_object_meta($model) {
        $module = \AppConfig::module_for_model($model);
        Data::check_module_access($module,'get');

        $fields = \AppConfig::setting("model.fields.$model");
        if (!$fields)
            throw new Errors\NotFound("No fields for object $model");
        $return_fields = [];
        foreach ($fields as $k => &$f) {
            if (!array_get_default($f, 'api_exposed', true))
                continue;
            if (array_get_default($f, 'type') == 'reminder') {
                $f['type'] = 'enum';
                $f['options'] = 'reminder_time_options';
            }
            if (isset($f['vname'])) {
                $f['vname'] = translate($f['vname'], array_get_default($f, 'vname_module', $module));
            }
            if (isset($f['vname_list'])) {
                $f['vname_list'] = translate($f['vname_list'], array_get_default($f, 'vname_module', $module));
            }
            if (isset($f['vdesc'])) {
                $f['vdesc'] = translate($f['vdesc'], array_get_default($f, 'vdesc_module', $module));
            }
            if (isset($f['options'])) {
                $dom = $f['options'];
                if(is_array($dom)) {
                    $f['options'] = $dom;
                } else {
                    $opts = \AppConfig::setting('lang.lists.current.' . $module . '.' . $dom);
                    if (!is_array($opts)) {
                        $opts = isset($GLOBALS['app_list_strings'][$dom]) ? $GLOBALS['app_list_strings'][$dom] : array();
                    }
                    $opts_list = array();
                    foreach ($opts as $ok => $ov) {
                        $opts_list[] = ["label" => $ov, "value" => (string)$ok];
                    }
                    $f['options'] = $opts_list;
                }
            }
            $return_fields[$k] = $f;
        }

        require_once('include/ListView/ListFormatter.php');
        $fmt = new \ListFormatter($model);

        $filter =& $fmt->getFilterForm();
        $filter->loadFilterLayout('Standard');

        $options = $filter->getFilterOptions();
        foreach ($options as $k => &$f) {
            if (!array_get_default($f, 'api_exposed', true))
                continue;
            if (isset($f['vname'])) {
                $f['vname'] = translate($f['vname'], array_get_default($f, 'vname_module', $module));
            }
            if (isset($f['vname_list'])) {
                $f['vname_list'] = translate($f['vname_list'], array_get_default($f, 'vname_module', $module));
            }
            if (isset($f['vdesc'])) {
                $f['vdesc'] = translate($f['vdesc'], array_get_default($f, 'vdesc_module', $module));
            }
            if (isset($f['options'])) {
                $dom = $f['options'];
                if(is_array($dom)) {
                    foreach ($dom as &$val) {
                        if (is_array($val)) {
                            $val = translate($val['vname'], array_get_default($val, 'vname_module', $module));
                        }
                    }
                    $f['options'] = $dom;
                } else {
                    $opts = \AppConfig::setting('lang.lists.current.' . $module . '.' . $dom);
                    if (!is_array($opts)) {
                        $opts = isset($GLOBALS['app_list_strings'][$dom]) ? $GLOBALS['app_list_strings'][$dom] : array();
                    }
                    $f['options'] = $opts;
                }
            }
        }

        return ['fields' => $return_fields, 'filters' => $options];
    }

}


