<?php

namespace OneCRM\API\Subscription;

use OneCRM\API\Endpoint;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;

require_once 'modules/CustomerLevels/CustomerLevel.php';

class SignupInfoEndpoint  extends Endpoint {

    public function endpoint() {
        return '/subscription/meta/signup';
    }

	public function allowedMethods() {
        return ['get'];
    }

    public function requiredScopes($method) {
        return ['read'];
    }

    public function descriptor($method)
    {
        switch ($method) {
            case 'get':
                return $this->descriptorGet();
                break;
        }
        return false;
    }

	public function descriptorGet() {
        static $desc = [
			'hidden' => true,
            'docs' => 'Get metadata required to render a signup screen for one or more products',
            'title' => 'Get plans metadata',
            'params' => [
                'products' => [
                    'type' => [
                        'type' => 'Array',
                        'element_type' => [
                            'type' => 'String',
                        ]
                    ],
                    'location' => 'query',
                    'docs' => 'Array of product IDs',
                ],
                'contact_id' => [
                    'type' => 'String',
                    'location' => 'query',
                    'docs' => 'Customer\'s contact ID. Used to retrieve payment procossor\'s customer ID, if any',
                ],
            ],
            'return' => [
                'products' => [
                    'type' => [
                        'type' => 'Array',
                        'element_type' => [
                            'type' => 'Object',
                        ]
                    ],
                    'docs' => 'Array of metadata grouped by product',
                    'required' => true,
                ],
            ]
        ];
        return $desc;
	}

	protected function get($input, Response $res, Request $req) {

		$json = getJSONobj();
		$decode_tiers = function($plan) use ($json) {
			if (!empty($plan['pricing_tiers'])) {
				$descriptions = explode("\n", array_get_default($plan, 'pricing_tiers_desc'));
				$tiers = $json->decode($plan['pricing_tiers']);
				ksort($tiers, SORT_NUMERIC);
				$plan['pricing_tiers'] = [];
				$idx = -1;
				foreach ($tiers as $k => $v) {
					$idx++;
					$plan['pricing_tiers'][] = [
						'qty' => $k,
						'amount' => (double)$v,
						'description' => array_get_default($descriptions, $idx),
					];
				}
			}
			return $plan;
		};
		$lq = new \ListQuery('Product', ['id', 'name', 'description', 'description_portal']);
		$lq->addSimpleFilter('id', $input['products']);
		$products = $lq->fetchAllRows();
		$prod_ids = array_column($products, 'id');

		$lq = new \ListQuery('Plan', [
			'id', 'name', 'description',
			'pricing_tiers', 'product_id',
			'amount', 'pricing_model', 'billing_cycle',
			'billing_cycle_mult',
			'pp_plan_id', 'qty_label', 'qty_comment',
			'long_name', 'popup_info', 'display_order',
			'qty_min_message', 'qty_max_message',
			'qty_min', 'qty_max', 'active',
		]);
		$lq->addSimpleFilter('product_id', $prod_ids);
		$lq->addSimpleFilter('hidden', 0);
		$plans = $lq->fetchAllRows();
		$plans = array_map($decode_tiers, $plans);

		$lq = new \ListQuery('plans_addons');
		$lq->addSimpleFilter('plan_id', array_column($plans, 'id'));
		$addons = $lq->fetchAllRows();

		$lq = new \ListQuery('plans_options');
		$lq->addSimpleFilter('plan_id', array_column($plans, 'id'));
		$options = $lq->fetchAllRows();

		$all_addon_ids = array_unique(array_column($addons, 'addon_id'));
		$all_option_ids = array_unique(array_column($options, 'option_id'));

		$lq = new \ListQuery('Addon');
		$lq->addSimpleFilter('id', $all_addon_ids);
		$all_addons = $lq->fetchAllRows();
		$all_addons = array_map($decode_tiers, $all_addons);

		foreach ($addons as $k => $a) {
			if (!isset($all_addons[$a['addon_id']]))
				unset($addons[$k]);
		}

		$lq = new \ListQuery('Option');
		$lq->addSimpleFilter('id', $all_option_ids);
		$all_options = $lq->fetchAllRows();

		foreach ($options as $k => $o) {
			if (!isset($all_options[$o['option_id']]))
				unset($options[$k]);
		}

		foreach ($addons as $a) {
			$plans[$a['plan_id']]['addons'][] = $a['addon_id'];
		}

		foreach ($options as $o) {
			$plans[$o['plan_id']]['options'][] = $o['option_id'];
		}

		foreach ($plans as $plan) {
			$products[$plan['product_id']]['plans'][$plan['id']] = $plan;
		}

		$edit_invoice_notes = \AppConfig::setting('subscriptions.edit_invoice_notes') ? true : false;
		$levels = array_filter(explode('^,^', \AppConfig::setting('subscriptions.invoice_notes_levels')));

		$customer_data = [];
		$coupons = [
			'default' => null,
			'by_plan' => [
				uniqid('_dummy_') => null,
			],
		];
		if (!empty($input['contact_id'])) {
			$contact = \ListQuery::quick_fetch_row('Contact', $input['contact_id']);
			if ($contact) {
				$customer_data['first_name'] = $contact['first_name'];
				$customer_data['last_name'] = $contact['last_name'];
				$account = \ListQuery::quick_fetch_row('Account', $contact['primary_account_id']);
				if ($account) {
					if ($edit_invoice_notes && !empty($levels)) {
						$level_id = $account['customer_level_id'];
						if (!in_array($level_id, $levels))
							$edit_invoice_notes = false;
					}
					$customer_data['account_name'] = $account['name'];
					$customer_data['customer_id'] = $account['pp_customer_id'];
					if (!empty($account['customer_level_id'])) {
						$level = \ListQuery::quick_fetch_row('CustomerLevel', $account['customer_level_id']);
						if ($level) {
							$customer_data['customer_level'] = \CustomerLevel::formatLevel($level);
							$def_coupon = \ListQuery::quick_fetch_row('Coupon', $level['default_coupon_id']);
							$coupons['default'] = self::formatCouponData($def_coupon);
							$level_coupons = json_decode($level['coupons'], true);
							if (is_array($level_coupons)) {
								foreach ($level_coupons as $lc) {
									$coupons['by_plan'][$lc['plan_id']] = [
										'keep_default' => $lc['keep_default'],
										'coupon' => self::formatCouponData(\ListQuery::quick_fetch_row('Coupon', $lc['coupon_id'])),
									];
								}
							}
						}
					}
					if ($customer_data['customer_id']) {
						require_once 'modules/PaymentProcessors/PaymentProcessor.php';
						$pp = \PaymentProcessor::getDefault();
						if ($pp) {
							$customer_data['payment_methods'] = $pp->getPaymentMethods($customer_data['customer_id']);
						}
					}
				}
			}
		}

		$res->getBody()->write(json_encode([
			'products' => $products,
			'addons' => $all_addons,
			'options' => $all_options,
			'customer_data' => $customer_data,
			'edit_invoice_notes' => $edit_invoice_notes,
			'coupons' => $coupons,
		]));
		return $res->withHeader('Content-Type', 'application/json');
	}


	private static function formatCouponData($coupon) {
		if (!$coupon) return null;
		$copy = [
			'discount_amount', 'name',
			'discount_type', 'discount_mode',
			'discount_percentage', 'plan_mode',
			'addon_mode'
		];
		$ret = [];
		foreach ($copy as $k) $ret[$k] = $coupon[$k];
		return $ret;
	}

}

