<?php

namespace OneCRM\API\Subscription;

/*
 * Implements a Ring Buffer for holding a limited number of recent events
 * see modules/Subscription/models/table.subscription_ring_buffer.php for table definition
 *
 * Note: row with id=-1 holds the current position pointer in 'data' column
 *       rows with id 0 .. max_entries-1 hold recorded events.
 *
 * TODO: use blowfish_encode / blowfish_decode to secure the entries
 * TODO: make this a module so it can be reused by all endpoints for optional logging?
 */

use MysqliManager;

class RingBuffer
{

    /** @var MysqliManager $db */
    public $db;
    public $table = 'subscription_ring_buffer';
    public $max_entries = 20000;

    /**
     * @param \MysqliManager $db
     * @param string $table         : defaults to 'subscriptions_ring_buffer
     * @param $max_entries          : ring buffer loops around after N entries
     */
    public function __construct(MysqliManager $db, $table = null, $max_entries = 20000)
    {
        $this->db = $db;
        if ($table) $this->table = $table;
        if ($max_entries) $this->max_entries = $max_entries;
    }

    /**
     * @param int $id        : the row to update, required
     * @param mixed $data    : any param not null will be updated.
     * @param string $key
     * @param string $value
     * @param $date_modified
     * @return bool : false if $id is out of range
     *
     * updates $data, $key, $value if they are not null
     */
    public function update_entry($id, $data = null, $key = null, $value = null, $date_modified = null)
    {
        $id = (int)$id;
        if ($id < 0 or $id >= $this->max_entries) return false;
        if (is_array($data) OR is_object($data)) {
            $data = json_encode($data, JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES
                | JSON_UNESCAPED_UNICODE | JSON_PARTIAL_OUTPUT_ON_ERROR
            );
        }
        // AFAIK the RTL lacks generic SQL table row ops for non-models.
        foreach (get_defined_vars() as $key => $value) {
            if (is_null($value)) continue;
            $values[] = '`' . $key . '`=\'' . $this->db->quote($value) . '\'';
        }
        $values = implode(', ', $values);
        $this->db->query("UPDATE `$this->table` SET $values WHERE id=$id");
        return true;
    }

    /**
     * @param array $data   : stuff to record (no index)
     * @param string $key   : used for lookups (index)
     * @param string $value : used for lookups (index)
     * @return int|bool     : log entry id, or false on failure.
     */
    public function add_entry($data, $key = '', $value = '')
    {
        $data = $this->db->quote(json_encode(
            $data, JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE | JSON_PARTIAL_OUTPUT_ON_ERROR
        ));
        $key = $this->db->quote($key);
        $value = $this->db->quote($value);
        $max = round($this->max_entries);
        $query = <<<SQL
START TRANSACTION;
SELECT @A:= `data` FROM `$this->table` WHERE `id`=-1;
SELECT @A:= if(CAST(@A AS INT)is null,0,@A+1) MOD $max;
REPLACE INTO `$this->table` (`id`, `date_modified`, `key`, `value`, `data`) 
  VALUES (-1, NOW(),0, 0, @A);
REPLACE INTO `$this->table` (`id`, `date_modified`, `key`, `value`, `data`) 
  VALUES(@A, NOW(), '$key','$value', '$data');
COMMIT;
SQL;
        // use multi-query to perform transaction set:
        if ($this->db->database->multi_query($query)) {
            $query = 1;
            do {
                if ($res = $this->db->database->store_result()) {
                    while ($row = $res->fetch_row()) {
                        $rows[$query][] = $row;
                    }
                    $res->free();
                }
                $query++;
            } while ($this->db->database->next_result());
            // return the ID of the inserted row in the ring buffer:
            return $rows[3][0][0];
		}
        return false;
    }
}
