<?php

require_once 'data/SugarBean.php';

/*
 * Subscription holds the state of a subscription and links to customer / account / plan / payment processor
 * NOTE: several SugarBean functions can corrupt state.  If in doubt avoid parent methods.
 *
 * 
 * Event hooks:
 *
 * INTERNAL  HOOKS    
 * The  Subscriptions  module  will  provide  a  foundation  for  subscription  management,  
 * but  to  be  really  useful  this  module  must  be  accompanied  by  business-specific  
 * custom  extensions.  For  example,  for  1CRM,  we  have  to  issue  licenses,  create  
 * cloud  instances,  move  instances  between  VMs  etc,  all  in  response  to  various  
 * events  generated  during  a  subscription  lifecycle.  Business-specific  extensions  
 * will  be  implemented  as  1CRM  modules.  Each  module  will  provide  a  file  named 
 * modules/_MODULE_NAME_/ext/config/standard/subscriptions.php with  a  list  of  hooks:
 *
 * hooks  
 *	−−    
 *		file: modules/OneCRMModule/OneCRMEventProcessor.php
 *		class: OneCRMEventProcessor
 *		function: process_event
 *	 
 *	 
 * When  an  event  occurs,  Subscriptions  module  will  obtain  a  list  of  hooks  
 * (AppCon-  fig::setting(“subscriptions.hooks”)) and  call  each  hook  using  
 * call_hook_function,  passing  event  data.  The  following  events  should  be  generated:  
 *	
 *	• new  subscription  
 *	• modified  subscription  
 *	• canceled  subscription  
 *	• new  invoice  
 *	• successful  payment  
 *	• failed  payment
 *	• new Contact?
 *
 * REST API:
 *		inbound webhooks notification processing
 *		
 *		Portal CRUD 
 *			1CRM records
 *			Payment Processor Records
 * 
 *		Interval Logging ( webhooks witness and 1CRM API )
 *
 *		Bulk / ATM verification with Payment Processor
 * 
 *		Endpoints:
 *			/plans			<-- provide list for Portal (unauthorized client?) to render Signups
 *			/subscriptions	<-- provide list & management for Portal Profile clients
 *			/accounts		<-- " " "
 *			/contacts		<-- " " "
 *			/invoices		<-- " " " 
 *			/payment_processor <-- " " "
 * 
 *		Subscription State Change:
 *			/subscriptions	<-- hook on alteration of subscription.state
 * 
 *		Access Control : chain of ownership is slow.  To speed that up, implement Capabilities.
 *		Perform chain of ownership one time then pass back a token to the client which can be 
 *		used to prove ownership without re-testing for a short time.
 *		example: sha1sum(NONCE.Resource_ID.Contact_ID.time.API_SECRET)
 *	
 *		Note: Plans need coupons to be Baked In so we don't have to pass a list of Coupons
 *		to the Portal to populate them with.
 * 
 * 
 * 
 *  For coupons/addons connected by pivot instead of hash:
 *		See link.messages_seen.php for a managed link table with arbitrary columns
 *		See table.payment_fees.php for a fully module-managed table
 * 
 */


class Subscription extends \SugarBean
{
    // model meta:
    var $table_name = "subscriptions";
    var $object_name = "Subscription";
    var $new_schema = true;

    // system fields:
    var $id;                    // LBL_SUBSCRIPTION_ID
    var $deleted;
    var $date_entered;
    var $date_modified;
    var $name;

    // subscription-specific state:
    var $start_date;            // LBL_START_DATE
    var $cancel_at_period_end;  // LBL_CANCELLED
    var $billing_cycle;         // LBL_SUBSCRIPTION_CYCLE
    var $current_period_start;  // LBL_SUBSCRIPTION_PERIOD_START
    var $current_period_end;    // LBL_SUBSCRIPTION_PERIOD_START
    var $quantity;              // LBL_QUANTITY
    var $addons;                // LBL_ADDONS
    var $coupons;               // LBL_COUPONS
    var $options;               // LBL_SUBSCRIPTION_OPTIONS
    var $state;                 // LBL_SUBSCRIPTION_STATUS
    var $balance;               // LBL_SUBSCRIPTION_BALANCE

    // relationships
    var $contact;               // LBL_CONTACT
    var $account;               // LBL_ACCOUNT
    var $plan;                  // LBL_PLAN
    var $payment_processor;     // LBL_PAYMENT_PROCESSOR
    var $pp_subs_id;            // LBL_PP_SUBS_ID


    // additional form fields  - only assigned user_id should be editable out of these:
    var $additional_column_fields = Array(
        'assigned_user_name', 'assigned_user_id', 'account_id', 'contact_id',
    );

    // key => model remationships
    // FIXME: need plan_id, option_id, coupon_id, invoice_id, maybe more.
    var $relationship_fields = Array(
        'account_id' => 'accounts', 'assigned_user_id' => 'contacts', 'contact_id' => 'contacts',
    );


    // FIXME:  HERE WE ARE.

    function get_summary_text()
    {
        return $this->name;
    }

    function get_contacts()
    {
        return $this->get_linked_beans('Contact');
    }


    /*
     * some utility methods used when working with subscription-related models:
     */

    /**
     * @param string|array $model : name of the Model type to search
     * @param string|array $data  : matching values to search on
     * @param string|array $match : fields from $data to search on, '*' = use all fields
     * @param array $fields       : fields to return
     * @param array $params       : ListQuery parameters arg
     * @param string $order       : ORDER arg, none by default
     * @param int $limit          : max records to return, 1 by default
     * @param int $offset         : records to skip, 0 by default
     * @return array              : match, list of matches, or empty array.
     *
     * $model can be used to pass args selectively by name
     * if $data is a string it will be used for lookup by id.
     */
    public static function query_model($model, $data, $match = '*', $fields = ['id'], $params = [], $order = '', $limit = 1, $offset = 0)
    {
        if (is_array($model)) extract($model, EXTR_IF_EXISTS);
        if (!$model) return [];
        if (is_string($data)) {
            $data = ['id' => $data];
            $match = ['id'];
        }
        if ($match === '*') $match = array_keys($data);
        $lq = new ListQuery($model, $fields, $params);
        foreach ((array)$match as $key){
            if (!array_key_exists($key,$data)) continue;
            $lq->addSimpleFilter($key, $data[$key]);
        }
        if ($order) $lq->setOrderBy($order);
        $res = $lq->runQuery($offset, $limit);
        if ($res && !$res->failed) {
            $rows = $res->getRows();
            if (is_countable($rows) && count($rows)) return $limit == 1 ? current($rows) : $rows;
        }
        return [];
    }

    /**
     * @param string $model : name of 1CRM Model
     * @param array $filter : see ListQuery::quick_fetch() $filter arg
     *                        if $filter is string it will be use to look up the model by Primary Key
     *                        if $filter['key'] then it will be used as Primary Key with $filter['value']
     * @param mixed $fields : see ListQUery::quick_fetch() id arg
     * @return array        : list of matching objects
     */
    public static function onecrm_getter($model, $filter, $fields = true)
    {
        if (is_string($filter)) {
            $res = ListQuery::quick_fetch($model, $filter, $fields);
            if ($res) $res = $res->getRow();
        } elseif (is_array($filter) && @$filter['key']) {
            $res = ListQuery::quick_fetch_key($model, $filter['key'], $filter['value'], $fields);
            if ($res) $res = $res->getRow();
        } else {
            $res = ListQuery::quick_fetch_all($model, $fields, $filter, true);
        }
        return $res ? $res : [];
    }

    /**
     * @param array|string $account : account data including id, account_id, pp_customer_id, payment_processor_id, email1
     *                              : if $account is string it will be interpreted as the account id.
     * @param array $fields         : if count($fields)==1 it returns the field directly.
     *
     * Precedence: match by account_id then id first
     * Match by payment_processor_id AND pp_customer_id second
     * Match by (email AND payment_processor_id) last
     * @return array|string|bool
     */
    public static function find_account($account, $fields = ['id'])
	{
		global $db;
		$order_by = [];
        $clauses=[];
        $listQuery = new ListQuery('Account', $fields);
        if (is_string($account)) $account = ['id' => $account];
        if ($account['account_id']??null) $account =['id'=>$account['account_id']];
        if ($account['id']??null) $clauses[] = ['field' => 'id', 'value' => $account['id']];
        if ($account['email1']??null) $clauses[] = ['operator'=>'AND',"multiple"=>[
            ['field' => 'payment_processor_id', 'value' => $account['payment_processor_id']??null],
            ['operator'=>'OR', 'multiple'=>[
                ['field' => 'email1', 'value' => $account['email1']],
                ['field' => 'email2', 'value' => $account['email1']],
        ]]]];
		if ($account['pp_customer_id']??null) {
			$clauses[] = ['operator' => 'AND', 'multiple' => [
            	['field' => 'pp_customer_id', 'value' => $account['pp_customer_id']],
	            ['field' => 'payment_processor_id', 'value' => $account['payment_processor_id']??null]
			]];
			$order_by[] = ['literal' => 'IF (pp_customer_id = \'' . $db->quote($account['pp_customer_id']) . '\', 1, 0) DESC'];
		}

        $listQuery->addFilterClause(['operator' => 'OR', 'multiple' => $clauses]);
		$order_by[] = ['field' => 'id', 'order' => 'desc'];
		$order_by[] = ['field' => 'pp_customer_id', 'order' => 'desc'];
		$listQuery->setOrderBy($order_by);
		$res = $listQuery->runQuerySingle();
        if (!$res) return false;
        if (is_array($res->row) && count($res->row) > 1) return $res->row;
        return @reset($res->row);
    }
	
	static function extract_subscription_details($model, &$row) {
		require_once 'modules/PaymentProcessors/PaymentProcessor.php';
		$pp = PaymentProcessor::getDefault();
		if (!$pp) return;
		if (empty($row['native'])) return;
		$json = getJSONobj();
		$native = $json->decode($row['native']);
		$row['_formatted'] = $pp->formatForDisplay($native);
	}

    private static function find_model_row($model, $data, $match, $fields)
    {
        if (is_string($data)) unset($match);
        $res = self::query_model($model, $data, @$match, $fields);
        if (!count($res)) return false;
        if (count($res) == 1 AND $fields == ['id']) return $res['id'];
        return $res;
    }

    /**
     * @param string|array $data    : 1CRM Model id or fields normalized from native data
     * @param array $fields         : field list to return.
     * @return array|bool|string    : if only [id] it will be returned as string.
     */
    public static function find_subscription($data, $fields = ['id'])
    {
        return self::find_model_row('Subscription', $data, ['pp_subs_id', 'payment_processor_id'], $fields);
    }

    public static function find_plan($data, $fields = ['id'])
    {
        return self::find_model_row('Plan', $data, ['pp_plan_id', 'payment_processor_id'], $fields);
    }

    public static function find_addon($data, $fields = ['id'])
    {
        return self::find_model_row('Addon', $data, ['pp_addon_id', 'payment_processor_id'], $fields);
    }

    public static function find_coupon($data, $fields = ['id'])
    {
        return self::find_model_row('Coupon', $data, ['pp_coupon_id', 'payment_processor_id'], $fields);
    }

    public static function find_invoice($data, $fields = ['id'])
    {
        return self::find_model_row('Invoice', $data, ['pp_invoice_id', 'payment_processor_id'], $fields);
    }

    public static function find_creditnote($data, $fields = ['id'])
    {
        return self::find_model_row('CreditNote', $data, ['pp_credit_id', 'payment_processor_id'], $fields);
    }

    public static function find_payment($data, $fields = ['id'])
    {
        return self::find_model_row('Payment', $data, ['pp_transaction_id', 'payment_processor_id'], $fields);
    }

	static function portal_filter($lq, $contact_id, $account_id) {
		if ($account_id) {
			$clause  = ['field' => 'account_id', 'value' => $account_id];
		} else {
			$clause = ['operator' => 'literal', 'value' => '0'];
		}
		$lq->addFilterClause($clause);
	}

    /*
     * TODO:
     * see Dev Guide 4.4.1 pp 44-45 re UX/Render optimsation/extension.
     * after_login / page_init / before_page_render / after_page_render?
     */

	static function get_customer_level_options() {
		$options = [];
		$lq = new ListQuery('CustomerLevel');
		$res = $lq->runQuery();
		$it = $res->getIterator();
		foreach ($it as $l) {
			$options[$l->getField('id')] = $l->getField('name');
		}
		return $options;
	}

}
