<?php
require_once('modules/WooCommerce/automattic_api/vendor/autoload.php');
require_once('exceptions/WooAuthException.php');

use Automattic\WooCommerce\Client;

class WooAuth {

	/**
	 * @var string WooCommerce URL
	 */
	private $url;

	/**
	 * @var string Consumer Key
	 */
	private $key;

	/**
	 * @var string Consumer Secret
	 */
	private $secret;

	/**
	 * @var string WooCommerce API version
	 */
	private $api_version;

	/**
	 * @var Client
	 */
	private $client;

	/**
	 * @var null|string
	 */
	private $error_message;

	/**
	 * @var Client|null
	 */
	private static $client_inst;

	/**
	 * @return null|string
	 */
	public function getErrorMessage() {
		return $this->error_message;
	}

	/**
	 * @return Client
	 * @throws WooAuthException
	 */
	public function getClient() {
		if (! $this->testConnection())
			throw new WooAuthException($this->error_message);

		return $this->client;
	}

	/**
	 * WooAuth constructor.
	 *
	 * @param string $url - WooCommerce URL
	 * @param string $key - Consumer Key
	 * @param string $secret - Consumer Secret
	 * @param string $version - API version
	 */
	public function __construct($url, $key, $secret, $version = 'wc/v3') {
		$this->url = $url;
		$this->key = $key;
		$this->secret = $secret;
		$this->api_version = $version;

		$this->error_message = null;
		$this->initClient();

		self::$client_inst = null;
	}

	/**
	 * @return Client|null
	 */
	public static function get_client() {
		if (self::$client_inst instanceof Client) {
			return self::$client_inst;
		} else {
			$auth = new WooAuth(
				AppConfig::setting('woocommerce.url'),
				AppConfig::setting('woocommerce.consumer_key'),
				AppConfig::setting('woocommerce.consumer_secret')
			);

			try {
				self::$client_inst = $auth->getClient();
			} catch (WooAuthException $e) {
				$GLOBALS['log']->fatal($e->getMessage());
			}
		}

		return self::$client_inst;
	}

	/**
	 * Test WoCommerce API connection
	 * If returns false, check getErrorMessage() method for details
	 *
	 * @return bool
	 */
	public function testConnection() {
		try {
			$this->client->get('customers');
			$result = true;
		} catch (Exception $e) {
			$result = false;
			$this->error_message = $e->getMessage();
		}

		return $result;
	}

	/**
	 * Initialize WooCommerce API Client
	 */
	private function initClient() {
		$this->client = new Client(
			$this->url,
			$this->key,
			$this->secret,
			['version' => $this->api_version]
		);
	}
}