<?php

require_once('WooRequest.php');
require_once('WooIdLocker.php');

require_once('exceptions/WooWrongSignatureException.php');
require_once('exceptions/WooWrongResourceException.php');
require_once('exceptions/WooWrongHookContentException.php');
require_once('exceptions/WooIdLockedException.php');

class WooManager {

	/**
	 * WooCommerce entries: order => ['entry']
	 *
	 * @var array
	 */
	private $entries = [
		1 => 'ProductCategory',
		2 => 'Product',
		3 => 'Customer',
		4 => 'CustomerContact',
		5 => 'Order',
	];

	/**
	 * The list of WooCommerce resources
	 *
	 * @var array
	 */
	private $resources = [
		'product',
		'customer',
		'order'
	];

	/**
	 * @var string: invoice or sales order if empty
	 */
	private $order_type;

	/**
	 * @var WooRequest
	 */
	private $request;

	/**
	 * @var WooIdLocker
	 */
	private $id_locker;

	/**
	 * WooManager constructor.
	 *
	 * @param DBManager $db
	 */
	public function __construct(DBManager $db) {
		$this->order_type = $this->getSetting('order_type');
		$this->request = new WooRequest();
		$this->id_locker = new WooIdLocker($db);
	}

	/**
	 * @return bool
	 */
	public function isEnabled() {
		return (bool)$this->getSetting('enabled');
	}

	/**
	 * @return bool
	 */
	public function isSyncStarted() {
		return (bool)$this->getSetting('sync_started');
	}

	/**
	 * Start sync
	 */
	public function startSync() {
		$this->saveSetting('sync_started', true);
	}

	/**
	 * Stop sync
	 */
	public function stopSync() {
		$this->saveSetting('sync_started', false);
	}

	/**
	 * @return bool
	 */
	public function isInitialSyncCompleted() {
		return (bool)$this->getSetting('initial_sync');
	}

	/**
	 * @param string $setting
	 *
	 * @return mixed
	 */
	public function getSetting($setting) {
		return AppConfig::setting("woocommerce.{$setting}");
	}

	/**
	 * Get number of synced records
	 *
	 * @param string $sync_type: pull or push
	 * @param $entry_name: like SalesOrder, Invoice or Product
	 *
	 * @return mixed
	 */
	public function getSyncedNum($sync_type, $entry_name) {
		return $this->getSetting($sync_type .'_'. strtolower($entry_name));
	}

	/**
	 * @param string $resource
	 *
	 * @return bool
	 */
	public function isSupportedHookResource($resource) {
		if (! in_array($resource, $this->resources)) {
			return false;
		} else {
			return true;
		}
	}

	/**
	 * Handle WooCommerce hook
	 *
	 * @param string $content
	 * @param string $signature
	 * @param string $resource
	 *
	 * @throws WooWrongSignatureException|WooWrongResourceException|WooWrongHookContentException|WooIdLockedException
	 */
	public function hookHandler($content, $signature, $resource) {

		if (! $this->isContentSignedCorrect($content, $signature)) {
			throw new WooWrongSignatureException('Wrong Woo hook signature');
		}

		if (! in_array($resource, $this->resources)) {
			throw new WooWrongResourceException('Unsupported Woo resource');
		}

		$decoded = json_decode($content);

		if (! $decoded) {
			throw new WooWrongHookContentException('Bad content from Woo');
		}

		$woo_id = $decoded->id;
		$entry_instance = $this->initEntryByResource($resource);

		if ($this->id_locker->isLocked($entry_instance->getModel(), $woo_id)) {
			$this->id_locker->unlock($entry_instance->getModel(), $woo_id);
			throw new WooIdLockedException("{$entry_instance->getModel()} with ID {$woo_id} temporary locked for updates");
		}

		try {
			$entry_instance->store($decoded);
			$this->id_locker->lock($entry_instance->getModel(), $woo_id);
		} catch (WooStoreEntryException $e) {
			$GLOBALS['log']->error("Store Woo record in CRM: {$e->getMessage()}");
		}
	}

	/**
	 * Run sync process
	 */
	public function sync() {
		if (! $this->isSyncStarted()) {
			$this->startSync();

			$this->handleError();

			$this->initialSync();
			$this->regularSync();

			$this->restoreErrorHandler();

			$this->stopSync();
			$this->setLastSync();
		}
	}

	/**
	 * Run re-sync process
	 *
	 * @return bool
	 */
	public function reSync() {
		if (! $this->isEnabled())
			return false;

		if (! $this->isSyncStarted()) {
			$this->startSync();
			$this->handleError();

			$entries = $this->getEntries();

			foreach ($entries as $entry) {
				$object = $this->getEntryInstance($entry, $this->order_type);

				if (! $object)
					continue;

				if ($object->isInitFromWoo())
					$this->pullAllFromWoo($object);

				if ($object->isInitToWoo() || $object->isUpdatesToWoo())
					$this->pushAllToWoo($object, true, $this->getResyncCheckDate());
			}

			$this->stopSync();
			$this->restoreErrorHandler();
		}

		$this->setLastSync();

		return true;
	}

	/**
	 * Entries after_save hook: unlock updated entry for push
	 *
	 * @param RowUpdate $upd
	 */
	public static function unlockEntry(RowUpdate &$upd) {
		$upd->set('woo_locked', 0);
	}

	/**
	 * Initial synchronization with WooCommerce
	 * (in both ways but depends on selected settings)
	 *
	 * @return bool
	 */
	private function initialSync() {
		if (! $this->isEnabled())
			return false;

		$entries = $this->getEntries();

		foreach ($entries as $entry) {
			$object = $this->getEntryInstance($entry, $this->order_type);

			if (! $object)
				continue;

			$pull_setting = 'init_' .strtolower($entry). '_pull';

			if ($object->isInitFromWoo() && ! $this->getSetting($pull_setting)) {
				$this->pullAllFromWoo($object);
				$this->saveSetting($pull_setting, true);
			}

			$push_setting = 'init_' .strtolower($entry). '_push';

			if ($object->isInitToWoo() && ! $this->getSetting($push_setting)) {
				$this->pushAllToWoo($object);
				$this->saveSetting($push_setting, true);
			}
		}

		if (! $this->isInitialSyncCompleted()) {
			$this->addWebHooks();
			$this->saveSetting('initial_sync', true);
		}

		return true;
	}

	/**
	 * Push updates/new records to WooCommerce
	 *
	 * @return bool
	 */
	private function regularSync() {
		if (! $this->isEnabled() && ! $this->isInitialSyncCompleted())
			return false;

		$entries = $this->getEntries();
		/** @var WooEntry $entry */

		foreach ($entries as $entry) {

			if ($entry == 'Order') {

				//Update both types of orders - Sales Order and Invoice
				$object = $this->getEntryInstance($entry, 'SalesOrder');

				if ($object->isUpdatesToWoo())
					$this->pushAllToWoo($object, true, $this->getLastSync());

				$object = $this->getEntryInstance($entry, 'Invoice');

				if ($object->isUpdatesToWoo())
					$this->pushAllToWoo($object, true, $this->getLastSync());

	 		} else {

				$object = $this->getEntryInstance($entry, $this->order_type);

				if ($object->isUpdatesToWoo())
					$this->pushAllToWoo($object, true, $this->getLastSync());

			}

		}

		return true;
	}

	/**
	 * @param WooEntry $entry
	 */
	private function pullAllFromWoo(WooEntry $entry) {
		$page = 1;
		$params = $entry->getListSearchParams();

		while (true) {
			$params['page'] = $page;
			$result = $this->request->get($entry->getEndpoint(), $params);

			if (empty($result))
				break;

			foreach ($result as $row) {

				try {
					$entry->store($row);
					$this->id_locker->lock($entry->getModel(), $row->id);
				} catch (WooStoreEntryException $e) {
					$GLOBALS['log']->error("Store Woo record in CRM: {$e->getMessage()}");
				}

			}

			$page ++;
		}
	}

	/**
	 * @param WooEntry $entry
	 * @param bool $updated_only - push only recently updated/created records
	 * @param string|null $last_sync - date string in Y-m-d H:i:s format
	 *
	 * @return bool
	 */
	private function pushAllToWoo(WooEntry $entry, $updated_only = false, $last_sync = null) {
		if ($updated_only && ! $entry->isUpdatesToWoo()) {
			return false;
		} elseif (! $updated_only && ! $entry->isInitToWoo()) {
			return false;
		}

		if ($updated_only) {
			$filters = $entry->getUpdatedRecordsFilters($last_sync);
		} else {
			$filters = $entry->getAllRecordsFilters();
		}

		$records = $this->getRecordsByFilters(
			$entry,
			$filters
		);

		foreach ($records as $record) {
			if (isset($record['woo_id']) && $this->id_locker->isLocked($entry->getModel(), $record['woo_id'])) {
				$this->id_locker->unlock($entry->getModel(), $record['woo_id']);
				continue;
			}

			$id = $entry->storeInWoo($record);

			if ($id)
				$this->id_locker->lock($entry->getModel(), $id);
		}

		return true;
	}

	/**
	 * @param WooEntry $entry
	 * @param array $filters
	 *
	 * @return array
	 */
	private function getRecordsByFilters(WooEntry $entry, $filters) {
		$lq = new ListQuery($entry->getModel());
		$lq->addFilterClauses($filters);
		$result = $lq->runQuery();

		return $result->getRows();
	}

	/**
	 * Add create/update WebHooks to WooCommerce
	 */
	private function addWebHooks() {
		$webHook = $this->getEntryInstance('WebHook');
		$webHook->storeInWoo(['secret' => $this->getWebHookSecret()]);
	}

	/**
	 * @param string $content
	 * @param string $signature
	 *
	 * @return bool
	 */
	private function isContentSignedCorrect($content, $signature) {
		$secret = $this->getSetting('webhook_secret');
		$content_sig = base64_encode(hash_hmac('sha256', $content, $secret, true));

		if ($content_sig != $signature) {
			return false;
		} else {
			return true;
		}
	}

	/**
	 * @return array
	 */
	private function getEntries(): array {
		$entries = $this->entries;
		ksort($entries);

		return $entries;
	}

	/**
	 * @param string $entry - Woo entry
	 * @param null|string $instance_type - type param for instance factory
	 *
	 * @return WooEntry|null
	 */
	private function getEntryInstance($entry, $instance_type = null) {
		$class = "Woo{$entry}";
		$path = "modules/WooCommerce/wrappers/entries/{$class}.php";

		if (is_file($path)) {
			require_once($path);

			if (method_exists($class, 'getInstance')) {
				return $class::getInstance($instance_type, $this->request);
			} else {
				return new $class($this->request);
			}

		} else {
			return null;
		}
	}

	/**
	 * @param string $resource
	 *
	 * @return null|WooEntry
	 */
	private function initEntryByResource($resource) {
		return $this->getEntryInstance(ucfirst($resource), $this->order_type);
	}

	/**
	 * Set last sync date
	 */
	private function setLastSync() {
		//small delay help to avoid excess updates
		sleep(1);
		$this->saveSetting('last_sync', gmdate('Y-m-d H:i:s'));
	}

	/**
	 * @return string: date string in Y-m-d H:i:s format
	 */
	private function getLastSync() {
		$last_sync = $this->getSetting('last_sync');

		if (empty($last_sync))
			$last_sync = gmdate('Y-m-d H:i:s');

		return $last_sync;
	}

	/**
	 * @return string - date string in Y-m-d H:i:s format
	 */
	private function getResyncCheckDate() {
		$last_sync = $this->getLastSync();
		$check_date = new DateTime($last_sync);
		$check_date->modify('-1 week');

		return  $check_date->format('Y-m-d H:i:s');
	}

	/**
	 * @return string
	 */
	private function getWebHookSecret() {
		$secret = $this->getSetting('webhook_secret');

		if (! $secret) {
			$bytes = openssl_random_pseudo_bytes(4);
			$secret = bin2hex($bytes);

			$this->saveSetting('webhook_secret', $secret);
		}

		return $secret;
	}

	/**
	 * @param string $setting
	 * @param mixed $value
	 */
	private function saveSetting($setting, $value) {
		AppConfig::set_local("woocommerce.{$setting}", $value);
		AppConfig::save_local();
	}

	/**
	 * Handle errors (reset running sync status)
	 */
	private function handleError() {
		set_error_handler(function ($errno, $msg, $file, $line) {
			$this->stopSync();

			if (!(error_reporting() & $errno))
				return false;

			$GLOBALS['log']->error("WooCommerce Sync Error: $msg, file: $file, line: $line");
			return true;
		});
	}

	/**
	 * Restore error handler to default
	 */
	private function restoreErrorHandler() {
		restore_error_handler();
	}
}
