<?php
require_once('WooEntry.php');
require_once('WooCustomerContact.php');

class WooCustomer extends WooEntry {

	/**
	 * @var array
	 */
	static $id_cache;

	/**
	 * @var array
	 */
	static $contacts_id_cache;

	/**
	 * @var array
	 */
	static $contacts_by_account_cache;

	/**
	 * @var WooCustomerContact
	 */
	private $contact;

	/**
	 * @var string
	 */
	private $contact_model;

	const GUEST_ID = 0;

	const GUEST_FAKE_ID = 9999999999;

	public function __construct(WooRequest $request) {
		parent::__construct($request);

		$this->model = 'Account';
		$this->contact_model = 'Contact';
		$this->endpoint = 'customers';

		$this->contact = new WooCustomerContact($request);

		self::$id_cache = [];
		self::$contacts_id_cache = [];

		$this->init_from_woo = $this->getSetting('init_woo_customers_sync');
		$this->init_to_woo = false;
		$this->updates_to_woo = $this->getSetting('customer_updates_sync');
	}

	/**
	 * @param string $last_sync
	 *
	 * @return array
	 */
	public function getUpdatedRecordsFilters($last_sync) {
		return [
			['field' => 'woo_id', 'value' => null, 'operator' => 'NOT NULL'],
			['field' => 'date_modified', 'value' => $last_sync, 'operator' => 'gte']
		];
	}

	/**
	 * @param stdClass $record
	 *
	 * @return bool|string: false on failure, record ID otherwise
	 * @throws WooStoreEntryException
	 */
	public function store($record) {
		global $current_user;
		$id = $this->isExists($this->getIdentifyFilter($record));
		$contact_id = null;

		if ($record->id == self::GUEST_FAKE_ID)
			unset($record->id);

		if ($id) {
			$record->primary_account_id = $id;
		} else {
			//If new account create primary contact
			try {
				$contact_id = $this->contact->store($record);
			} catch (WooStoreEntryException $e) {
				$GLOBALS['log']->error("Store Woo Account's Primary Contact in CRM: {$e->getMessage()}");
			}
		}

		$data = $this->map($record);

		if($data['name'] == ' ' || !isset($data['name']) || $data['name'] == '') {
			$data['name'] = $data['username'] ;
		}

		if ($contact_id)
			$data['primary_contact_id'] = $contact_id;

		$user_id = $this->getAssignedUserId();
		$current_user->id = $user_id;
		$add = false;

		if (! $this->isCorrectRole($data['role']))
			return false;

		unset($data['role']);

		$lq = new ListQuery('Account');
		$lq->addSimpleFilter('woo_id', $data['woo_id'], '=');
		$res = $lq->fetchAllRows();

		foreach ($res as $r) {
			$data['id'] = $r['id'];
		}

		if (isset($data['id'])) {
			$result = ListQuery::quick_fetch($this->model, $data['id']);
			$upd = RowUpdate::for_result($result, null, $user_id);

			//Update primary contact
			if ($upd->getField('primary_contact_id')) {
				try {
					$this->contact->store($record);
				} catch (WooStoreEntryException $e) {
					$GLOBALS['log']->error("Update Woo Account's Primary Contact in CRM: {$e->getMessage()}");
				}
			}

			unset($data['id']);
		} else {
			$upd = RowUpdate::blank_for_model($this->model, $user_id);
			$add = true;
		}

		foreach ($data as $k => &$dat) {
			$field_def = $upd->getFieldDefinition($k);
			if(isset($field_def['len'])) {
				if($field_def['len'] < strlen($dat))
				$dat = substr($dat, 0, $field_def['len']);
			}
		}

		$upd->set($data);

		try {
			$result = $upd->save();
			try {
				$record->primary_account_id = $upd->getField('id');
				$this->contact->store($record);
			} catch (WooStoreEntryException $e) {
				$GLOBALS['log']->error("Update Woo Account's Primary Contact in CRM: {$e->getMessage()}");
			}
		} catch (Exception $e) {
			throw new WooStoreEntryException($e->getMessage());
		}

		if ($result) {
			if ($add && isset($record->id) && $record->id != self::GUEST_ID)
				$this->incrementPulledNum();

			return $upd->getField('id');
		} else {
			return false;
		}
	}

	/**
	 * @param array $record
	 *
	 * @return int|null
	 */
	public function storeInWoo($record) {
		$data = $this->mapForWoo($record);

		return $this->push($data, $record['woo_id'], $record['id']);
	}

	/**
	 * Get Account ID by Woo ID
	 * if account doesn't exist in the CRM create it
	 *
	 * @param int $woo_id
	 * @param array $order_data
	 *
	 * @return string|null
	 */
	public function getAccountId($woo_id, $order_data = []) {

		if (! isset(self::$id_cache[$woo_id]) || $woo_id == self::GUEST_ID) {

			$contact_id = null;
			$account_id = null;

			if ($woo_id == self::GUEST_ID) {

				try {
					$account_id = $this->createGuestAccount($order_data);
				} catch (WooStoreEntryException $e) {
					$GLOBALS['log']->error("Store Woo Guest Account in CRM: {$e->getMessage()}");
				}

			} else {
				$record = $this->loadByWooId($woo_id);

				if ($record) {
					$account_id = $record['id'];
					$contact_id = $record['primary_contact_id'];

					if (! empty($contact_id))
						self::$contacts_id_cache[$woo_id] = $contact_id;
				} else {
					$record = $this->loadFromWoo($woo_id);

					try {
						$account_id = $this->store($record);
					} catch (WooStoreEntryException $e) {
						$GLOBALS['log']->error("Store Woo Account in CRM: {$e->getMessage()}");
					}
				}
			}

			if (! $account_id)
				$account_id = null;

			self::$id_cache[$woo_id] = $account_id;
		}

		return self::$id_cache[$woo_id];
	}

	/**
	 * Get Account's Primary Contact ID by Woo ID or Account ID
	 *
	 * @param int $woo_id
	 * @param string $account_id - account ID
	 *
	 * @return string|null
	 */
	public function getContactId($woo_id, $account_id) {

		if (! empty($woo_id)) {

			if (! isset(self::$contacts_id_cache[$woo_id])) {
				$record = $this->loadByWooId($woo_id);
				$contact_id = null;

				if ($record)
					$contact_id = $record['primary_contact_id'];

				if (! empty($contact_id))
					self::$contacts_id_cache[$woo_id] = $contact_id;

			}

			return self::$contacts_id_cache[$woo_id];

		} elseif (! empty($account_id)) {

			if (! isset(self::$contacts_by_account_cache[$account_id])) {
				$result = ListQuery::quick_fetch($this->model, $account_id);

				if ( ! $result->failed)
					$contact_id = $result->getField('primary_contact_id');

				if (! empty($contact_id))
					self::$contacts_by_account_cache[$account_id] = $contact_id;
			}

			return self::$contacts_by_account_cache[$account_id];

		}

		return null;
	}

	/**
	 * @param array $data
	 *
	 */
	protected function adjustDataFromWoo(&$data) {
		$name = '';

		if (! empty($data['company'])) {
			$name = $data['company'];
			unset($data['company']);
		} else {
			if (! empty($data["first_name"]))
				$name = $data["first_name"];

			if (! empty($data["last_name"])) {
				if (! empty($name)) {
					$name .= ', ' . $data["last_name"];
				} else {
					$name = $data["last_name"];
				}
			}
		}

		if (isset($data["first_name"]))
			unset($data['first_name']);

		if (isset($data["last_name"]))
			unset($data['last_name']);

		$data['name'] = trim($name);
		$data['account_type'] = 'Customer';

		if (empty($data['id']))
			$data['assigned_user_id'] = $this->getAssignedUserId();

		$data['modified_user_id'] = $data['assigned_user_id'];
	}

	/**
	 * Adjust data for compatibility with WooCommerce
	 *
	 * @param array $data
	 */
	protected function adjustDataForPush(&$data) {
		if (isset($data['account_id']))
			unset($data['account_id']);

		if (isset($data['name'])) {
			$name = explode(', ', $data['name']);

			if (is_countable($name) && sizeof($name) == 1) {
				$data['billing']['company'] = $name[0];
			} else {
				if (! empty($name[0]))
					$data['first_name'] = $name[0];

				if (! empty($name[1]))
					$data['last_name'] = $name[1];
			}

			unset($data['name']);
		}
	}

	/**
	 * @return array
	 */
	protected function getMap() {
		return [
			'id' => 'woo_id',
			'account_id' => 'id',
			'email' => 'email1',
			'first_name' => 'first_name',
			'last_name' => 'last_name',
			'role' => 'role',
			'name' => 'name',

			'billing' => [
				'company' => 'company',
				'address_1' => 'billing_address_street',
				'city' => 'billing_address_city',
				'state' => 'billing_address_state',
				'postcode' => 'billing_address_postalcode',
				'country' => 'billing_address_country',
				'phone' => 'phone_office'
			],

			'shipping' => [
				'address_1' => 'shipping_address_street',
				'city' => 'shipping_address_city',
				'state' => 'shipping_address_state',
				'postcode' => 'shipping_address_postalcode',
				'country' => 'shipping_address_country',
			],
		];
	}

	/**
	 * @param array $order_data
	 *
	 * @return bool|string: false on failure, record ID otherwise
	 * @throws WooStoreEntryException
	 */
	private function createGuestAccount($order_data) {
		//if exist contact data in the order so create account and contact record using order data
		if (! empty($order_data['customer_first_name']) && ! empty($order_data['customer_last_name'])
			&& (! empty($order_data['customer_email']) || ! empty($order_data['customer_phone']))) {

			$record = new stdClass();

			$record->id = self::GUEST_FAKE_ID;
			$record->email = $order_data['customer_email'];
			$record->first_name = $order_data['customer_first_name'];
			$record->last_name = $order_data['customer_last_name'];
			$record->role = 'customer';
			$record->name = '';

			$billing = new stdClass();
			$billing->company = $order_data['customer_company'];
			$billing->address_1 = $order_data['billing_address_street'];
			$billing->city = $order_data['billing_address_city'];
			$billing->state = $order_data['billing_address_state'];
			$billing->postcode = $order_data['billing_address_postalcode'];
			$billing->country = $order_data['billing_address_country'];
			$billing->phone = $order_data['customer_phone'];

			$record->billing = [$billing];

			$shipping = new stdClass();
			$shipping->address_1 = $order_data['shipping_address_street'];
			$shipping->city = $order_data['shipping_address_city'];
			$shipping->state = $order_data['shipping_address_state'];
			$shipping->postcode = $order_data['shipping_address_postalcode'];
			$shipping->country = $order_data['shipping_address_country'];

			$record->shipping = [$shipping];

			return $this->store($record);

		//if contact data is empty so create empty guest account
		} else {
			$data = [
				'name' => translate('LBL_GUEST_ACCOUNT_NAME', 'WooCommerce'),
				'woo_id' => self::GUEST_ID,
				'assigned_user_id' => $this->getAssignedUserId(),
				'modified_user_id' => $this->getAssignedUserId()
			];

			$upd = RowUpdate::blank_for_model($this->model, $this->getAssignedUserId());
			$upd->set($data);

			try {
				$result = $upd->save();
			} catch (Exception $e) {
				throw new WooStoreEntryException($e->getMessage());
			}

			if ($result) {
				$this->incrementPulledNum();

				return $upd->getField('id');
			} else {
				return false;
			}
		}
	}

	/**
	 * @param stdClass $record
	 *
	 * @return array
	 */
	private function getIdentifyFilter($record) {

		return [
			[   'operator' => 'OR',
				'multiple' => [
					['field' => 'woo_id', 'value' => $record->id],
					[
						'operator' => 'OR',
						'multiple' => [
							['field' => 'email1', 'value' => $record->email],
							['field' => 'email2', 'value' => $record->email],
						]
					]
				]
			]
		];

	}

	/**
	 * @param string $role
	 *
	 * @return bool
	 */
	private function isCorrectRole($role) {
		if(strpos($role, 'administrator') !== false){
			return false;
		} else{
			return true;
		}
	}
}
