<?php
require_once('WooEntry.php');

class WooCustomerContact extends WooEntry {

	public function __construct(WooRequest $request) {
		parent::__construct($request);

		$this->model = 'Contact';
		$this->endpoint = 'customers';

		$this->updates_to_woo = $this->getSetting('customer_updates_sync');
	}

	/**
	 * @param string $last_sync
	 *
	 * @return array
	 */
	public function getUpdatedRecordsFilters($last_sync) {
		return [
			['field' => 'woo_id', 'value' => null, 'operator' => 'NOT NULL'],
			['field' => 'date_modified', 'value' => $last_sync, 'operator' => 'gte']
		];
	}

	/**
	 * @param stdClass $record
	 *
	 * @return bool|string: false on failure, record ID otherwise
	 * @throws WooStoreEntryException
	 */
	public function store($record) {
		$id = $this->isExists($this->getIdentifyFilter($record));

		if ($id)
			$record->contact_id = $id;

			$data = $this->map($record);
			$user_id = $this->getAssignedUserId();

		$lq = new ListQuery('Contact');
		$lq->addSimpleFilter('woo_id', $data['woo_id'], '=');
		$res = $lq->fetchAllRows();

		foreach ($res as $r) {
			$data['id'] = $r['id'];
		}

		if(($data['last_name'] == '' && $data['first_name'] == '') || (!isset($data['first_name']) && !isset($data['last_name']))) {
			$data['last_name'] = $data['name'] ?? $data['username'];
		}

		if (isset($data['id'])) {
			$result = ListQuery::quick_fetch($this->model, $data['id']);
			$upd = RowUpdate::for_result($result, null, $user_id);

			unset($data['id']);
		} else {
			$upd = RowUpdate::blank_for_model($this->model, $user_id);
		}

		foreach ($data as $k => &$dat) {
			$field_def = $upd->getFieldDefinition($k);
			if(isset($field_def['len'])) {
				if($field_def['len'] < strlen($dat))
				$dat = substr($dat, 0, $field_def['len']);
			}
		}

		$upd->set($data);

		try {
			$result = $upd->save();
			$res = ListQuery::quick_fetch('Account', $upd->getField('primary_account_id'));
			if($res) {
				$upd2 = RowUpdate::for_result($res, null, $user_id);
				$upd2->set(['primary_contact_id' => $upd->getPrimaryKeyValue()]);
				$upd2->save();
			}
		} catch (Exception $e) {
			throw new WooStoreEntryException($e->getMessage());
		}

		if ($result) {
			return $upd->getField('id');
		} else {
			return false;
		}
	}

	/**
	 * @param array $record
	 *
	 * @return int|null
	 */
	public function storeInWoo($record) {
		$data = $this->mapForWoo($record);

		return $this->push($data, $record['woo_id'], $record['id']);
	}

	/**
	 * @param array $data
	 *
	 */
	protected function adjustDataFromWoo(&$data) {
		if (empty($data['id']))
			$data['assigned_user_id'] = $this->getAssignedUserId();
	}

	/**
	 * Adjust data for compatibility with WooCommerce
	 *
	 * @param array $data
	 */
	protected function adjustDataForPush(&$data) {
		if (isset($data['contact_id']))
			unset($data['contact_id']);
	}

	/**
	 * @return array
	 */
	protected function getMap() {
		return [
			'id' => 'woo_id',
			'contact_id' => 'id',
			'email' => 'email1',
			'first_name' => 'first_name',
			'last_name' => 'last_name',
			'primary_account_id' => 'primary_account_id',

			'billing' => [
				'address_1' => 'primary_address_street',
				'city' => 'primary_address_city',
				'state' => 'primary_address_state',
				'postcode' => 'primary_address_postalcode',
				'country' => 'primary_address_country',
				'phone' => 'phone_work'
			]
		];
	}

	/**
	 * @param stdClass $record
	 *
	 * @return array
	 */
	private function getIdentifyFilter($record) {

		return [
			[   'operator' => 'OR',
				'multiple' => [
					['field' => 'woo_id', 'value' => $record->id],
					[
						'operator' => 'OR',
						'multiple' => [
							['field' => 'email1', 'value' => $record->email],
							['field' => 'email2', 'value' => $record->email],
						]
					]
				]
			]
		];

	}
}
