<?php
require_once('modules/WooCommerce/wrappers/exceptions/WooStoreEntryException.php');

abstract class WooEntry {

	/**
	 * @var string
	 */
	protected $model;

	/**
	 * WooCommerce API endpoint
	 *
	 * @var string
	 */
	protected $endpoint;

	/**
	 * Pull records from WooCommerce
	 *
	 * @var bool
	 */
	protected $init_from_woo;

	/**
	 * Push record to WooCommerce
	 *
	 * @var bool
	 */
	protected $init_to_woo;

	/**
	 * Push entry updates to WooCommerce
	 *
	 * @var bool
	 */
	protected $updates_to_woo;

	/**
	 * @var WooRequest
	 */
	private $request;

	/**
	 * @return string
	 */
	public function getModel() {
		return $this->model;
	}

	/**
	 * @return string
	 */
	public function getEndpoint() {
		return $this->endpoint;
	}

	/**
	 * @return WooRequest
	 */
	public function getRequest(): WooRequest {
		return $this->request;
	}

	/**
	 * @return bool
	 */
	public function isInitFromWoo() {
		return $this->init_from_woo;
	}

	/**
	 * @return bool
	 */
	public function isInitToWoo() {
		return $this->init_to_woo;
	}

	/**
	 * @return mixed
	 */
	public function isUpdatesToWoo() {
		return $this->updates_to_woo;
	}

	/**
	 * WooEntry constructor.
	 *
	 * @param WooRequest $request
	 */
	public function __construct(WooRequest $request) {
		$this->init_from_woo = false;
		$this->init_to_woo = false;
		$this->updates_to_woo = false;

		$this->request = $request;
	}

	/**
	 * Returns additional search params for getting
	 * the list of records from WooCommerce
	 *
	 * @return array
	 */
	public function getListSearchParams() {
		return [];
	}

	/**
	 * Returns filters for getting the list of all records for push
	 *
	 * @return array
	 */
	public function getAllRecordsFilters() {
		return [
			['field' => 'woo_id', 'value' => null, 'operator' => 'NOT'],
		];
	}

	/**
	 * Returns filters for getting the list of updated records for push
	 *
	 * @param string $last_sync: date in Y-m-d H:i:s format
	 *
	 * @return array
	 */
	public function getUpdatedRecordsFilters($last_sync) {
		return [
			[
				'operator' => 'OR',
				'multiple' => [
					['field' => 'woo_id', 'value' => null, 'operator' => 'NOT'],
					['field' => 'date_modified', 'value' => $last_sync, 'operator' => 'gte']
				]
			]
		];
	}

	/**
	 * Store record from WooCommerce in DB
	 *
	 * @param stdClass $record
	 *
	 * @return bool|string - stored record ID or false otherwise
	 * @throws WooStoreEntryException
	 */
	public abstract function store($record);

	/**
	 * Store record in WooCommerce
	 *
	 * @param array $data
	 *
	 * @return int|null
	 */
	public abstract function storeInWoo($data);

	/**
	 * @param array $search_params
	 *
	 * @return bool|int: false if record doesn't exists and record ID otherwise
	 */
	protected function isExistsInWoo($search_params) {
		$result = $this->request->get($this->endpoint, $search_params);

		if (is_countable($result) && sizeof($result) > 0 && property_exists($result[0], 'id')) {
			return $result[0]->id;
		} else {
			return false;
		}
	}

	/**
	 * @param array $filters
	 *
	 * @return bool|string: false if record doesn't exists and record ID otherwise
	 */
	protected function isExists($filters) {
		$lq = new ListQuery($this->model);
		$lq->addFilterClauses($filters);
		$result = $lq->runQuerySingle();

		if (! empty($result->getField('id'))) {
			return $result->getField('id');
		} else {
			return false;
		}
	}

	/**
	 * Make fields mapping for data from WooCommerce
	 *
	 * @param stdClass $record
	 * @param bool $adjust
	 *
	 * @return array
	 */
	protected function map($record, $adjust = true) {
		$map = $this->getMap();
		$data = [];

		foreach ($map as $wooField => $field) {
			if (! property_exists($record, $wooField))
				continue;

			if (is_array($field)) {

				foreach ($field as $woo_inner => $inner) {
					if (is_array($record->$wooField)) {
						if (! empty($record->$wooField))
							$data[$inner] = $record->$wooField[0]->$woo_inner;
					} else {
						$data[$inner] = $record->$wooField->$woo_inner;
					}
				}

			} else {
				$data[$field] = $record->$wooField;
			}
		}

		if ($adjust)
			$this->adjustDataFromWoo($data);

		return $data;
	}

	/**
	 * Make fields mapping for pushing to WooCommerce
	 *
	 * @param array $record
	 * @param bool $adjust
	 *
	 * @return array
	 */
	protected function mapForWoo($record, $adjust = true) {
		$map = $this->getMap();
		$data = [];

		foreach ($map as $wooField => $origField) {
			if (is_array($origField)) {

				foreach ($origField as $woo_inner => $inner) {
					if (! isset($record[$inner]))
						continue;

					$data[$wooField][$woo_inner] = $record[$inner];
				}

			} else {
				if (! isset($record[$origField]))
					continue;

				$data[$wooField] = $record[$origField];
			}
		}

		if ($adjust)
			$this->adjustDataForPush($data);

		return $data;
	}

	/**
	 * Adjust data from WooCommerce for compatibility with 1CRM
	 *
	 * @param array $data
	 */
	protected function adjustDataFromWoo(&$data) {}

	/**
	 * Adjust data for compatibility with WooCommerce
	 *
	 * @param array $data
	 */
	protected function adjustDataForPush(&$data) {}

	/**
	 * Load record from WooCommerce
	 *
	 * @param int $id - woo record ID
	 *
	 * @return stdClass|null
	 */
	protected function loadFromWoo($id) {
		$result = $this->request->get("$this->endpoint/{$id}", []);

		if ($result && (property_exists($result, 'id') || isset($result['id']))) {
			return $result;
		} else {
			return null;
		}
	}

	/**
	 * Load record from 1CRM by woo_id
	 *
	 * @param int $woo_id - woo_id
	 *
	 * @return array|null
	 */
	protected function loadByWooId($woo_id) {
		$lq = new ListQuery($this->model);
		$lq->addSimpleFilter('woo_id', $woo_id, '=');
		$result = $lq->runQuerySingle();

		if (! empty($result->getField('id'))) {
			return $result->getRow();
		} else {
			return null;
		}
	}

	/**
	 * Add/Update record in WooCommerce
	 *
	 * @param array $data
	 * @param int $woo_id
	 * @param int $orig_record_id
	 *
	 * @return mixed
	 */
	protected function push($data, $woo_id, $orig_record_id) {
		$set_woo_id = true;

		if ($woo_id) {

			if (isset($data['slug']))
				unset($data['slug']);

			$post_result = $this->request->update("{$this->endpoint}/{$woo_id}", $data);
			$add = false;
			$set_woo_id = false;
		} else {
			$post_result = $this->request->add($this->endpoint, $data);
			$add = true;
		}

		if ($post_result && property_exists($post_result, 'id')) {
			$woo_id = $post_result->id;

			if ($add)
				$this->incrementPushedNum();
		}

		if (isset($data['id']))
			$set_woo_id = false;

		if ($set_woo_id)
			$this->setWooId($orig_record_id, $woo_id);

		return $woo_id;
	}

	/**
	 * @param string $record_id
	 * @param int $woo_id
	 */
	protected function setWooId($record_id, $woo_id) {
		$user_id = $this->getAssignedUserId();

		try {
			$upd = RowUpdate::blank_for_model($this->model, $user_id);
			$upd->retrieveRecord($record_id);
			$upd->set('woo_id', $woo_id);
			$upd->save();
		} catch (Exception $e) {
			$GLOBALS['log']->fatal('Error during setting of woo_id ' . $e->getMessage());
		}
	}

	/**
	 * Get fields map: 'woo_field' => 'crm_field'
	 *
	 * @return array
	 */
	protected function getMap() {
		return [];
	}

	/**
	 * @param string $string
	 *
	 * @return string
	 */
	protected function createSlug($string) {
		$string = str_replace(' ', '-', strtolower($string)); // Replaces all spaces with hyphens.
		$string = preg_replace('/[^A-Za-z0-9\-]/', '', $string); // Removes special chars.

		return preg_replace('/-+/', '-', $string); // Replaces multiple hyphens with single one.
	}

	/**
	 * @param string $setting
	 *
	 * @return mixed
	 */
	protected function getSetting($setting) {
		return AppConfig::setting("woocommerce.{$setting}");
	}

	/**
	 * @return string
	 */
	protected function getAssignedUserId() {
		return $this->getSetting('assigned_user_id');
	}

	/**
	 * Increase number of pushed entries
	 */
	protected function incrementPushedNum() {
		$this->incrementSyncedNum('push');
	}

	/**
	 * Increase number of pulled entries
	 */
	protected function incrementPulledNum() {
		$this->incrementSyncedNum('pull');
	}

	/**
	 * @param string $type: push, pull
	 */
	private function incrementSyncedNum($type) {
		$current_value = $this->getSetting("{$type}_" . strtolower($this->model));

		if ($current_value) {
			$new_value = $current_value + 1;
		} else {
			$new_value = 1;
		}

		AppConfig::set_local("woocommerce.{$type}_" . strtolower($this->model), $new_value);
		AppConfig::save_local();
	}
}
