<?php
require_once('WooOrder.php');
require_once('WooPayment.php');

class WooInvoice extends WooOrder {

	/**
	 * WooInvoice constructor.
	 *
	 * @param WooRequest $request
	 */
	public function __construct(WooRequest $request) {
		$this->model = 'Invoice';
		parent::__construct($request);
	}

	/**
	 * @param array $data
	 *
	 * @return RowUpdate
	 */
	protected function init($data) {
		require_bean($this->model);

		$upd = RowUpdate::blank_for_model($this->model, $this->getAssignedUserId());
		Invoice::init_record($upd, $data);

		return $upd;
	}

	/**
	 * @param array $data
	 */
	protected function adjustDataFromWoo(&$data) {
		parent::adjustDataFromWoo($data);
		$data['terms'] = 'Due on Receipt';
	}

	/**
	 * Handle WooCommerce order's status
	 *
	 * @param stdClass $original_invoice_record
	 * @param string $invoice_id
	 */
	protected function statusHandler($original_invoice_record, $invoice_id) {
		$status = $original_invoice_record->status;
		$date_paid = $original_invoice_record->date_paid;

		if($date_paid != '')
			$status = 'completed';

		switch ($status) {
			case "failed":
			case "cancelled":
				$this->cancel($invoice_id);
				break;
			case "completed":
				$this->addPayment($original_invoice_record, $invoice_id);
				break;
			case "refunded":
				$this->refund($invoice_id);
				break;
			default:
				break;
		}
	}

	/**
	 * Cancel invoice
	 *
	 * @param string $id - invoices ID
	 *
	 * @return bool
	 */
	private function cancel($id) {
		$result = ListQuery::quick_fetch($this->model, $id);

		if ($result->getField('cancelled'))
			return false;

		$upd = RowUpdate::for_result($result, null, $this->getAssignedUserId());
		$upd->set('cancelled', 1);

		try {
			$upd->save();
			return true;
		} catch (Exception $e) {
			$GLOBALS['log']->error("Woo Invoice cancellation error (ID: {$id}): {$e->getMessage()}");
			return false;
		}
	}

	/**
	 * Add payment to invoice
	 *
	 * @param stdClass $record
	 * @param string $id
	 *
	 * @return bool;
	 */
	private function addPayment($record, $id) {
		$result = ListQuery::quick_fetch($this->model, $id);
		$amount_due = $result->getField('amount_due');

		if (empty($amount_due))
			return false; //invoice already paid

		$payment = new WooPayment($this->getRequest());
		$payment->setInvoice($result);
		$payment_id = null;

		try {
			$payment_id = $payment->store($record);
		} catch (WooStoreEntryException $e) {
			$GLOBALS['log']->error("Store Woo payment record in CRM (Invoice ID: {$id}): {$e->getMessage()}");
		}

		if ($payment_id) {
			return true;
		} else {
			$GLOBALS['log']->error("Payment for Invoice {$id} was not stored");
			return false;
		}
	}

	/**
	 * Make refund
	 *
	 * @param string $id
	 *
	 * @return bool
	 */
	private function refund($id) {
		$result = ListQuery::quick_fetch($this->model, $id);

		if ($result->getField('cancelled'))
			return false;

		$payment = new WooPayment($this->getRequest());
		$payment->setRefunded($result);

		$upd = RowUpdate::for_result($result, null, $this->getAssignedUserId());
		$upd->set('description', $result->getField('description') .' '. translate('LBL_INVOICE_REFUND_NOTES', 'WooCommerce'));
		$upd->set('cancelled', 1);

		try {
			$upd->save();
		} catch (Exception $e) {
			$GLOBALS['log']->error("Refunded Woo Invoice cancellation error (ID: {$id}): {$e->getMessage()}");
		}

		return true;
	}
}
