<?php
require_once('WooEntry.php');
require_once('WooLineItem.php');
require_once('WooCustomer.php');

class WooOrder extends WooEntry {

	const TAX_EXEMPT_CODE_ID = '-99';

	/**
	 * @var WooLineItem
	 */
	private $line_item;

	/**
	 * @var WooCustomer
	 */
	private $customer;

	/**
	 * WooOrder constructor.
	 *
	 * @param WooRequest $request
	 */
	public function __construct(WooRequest $request) {
		parent::__construct($request);

		$this->endpoint = 'orders';
		$this->line_item = new WooLineItem($request);
		$this->customer = new WooCustomer($request);

		$this->init_from_woo = $this->getSetting('init_woo_orders_sync');
		$this->init_to_woo = false;
		$this->updates_to_woo = $this->getSetting('order_updates_sync');
	}

	/**
	 * @param string $type
	 * @param WooRequest $request
	 *
	 * @return WooOrder
	 */
	public static function getInstance($type, WooRequest $request) {
		if ($type == 'invoice') {
			require_once('WooInvoice.php');
			return new WooInvoice($request);
		} else {
			require_once('WooSalesOrder.php');
			return new WooSalesOrder($request);
		}
	}

	/**
	 * @return array
	 */
	public static function get_order_types_options() {
		return [
			"" => translate('LBL_ORDER_TYPE_SO', 'WooCommerce'),
			"invoice" => translate('LBL_ORDER_TYPE_INVOICE', 'WooCommerce')
		];
	}

	/**
	 * @return array
	 */
	public function getListSearchParams() {
		//return ['status' => ['pending', 'processing', 'on-hold']];
		return [];
	}

	/**
	 * @param string $last_sync
	 *
	 * @return array
	 */
	public function getUpdatedRecordsFilters($last_sync) {
		return [
			['field' => 'woo_id', 'value' => null, 'operator' => 'NOT NULL'],
			['field' => 'date_modified', 'value' => $last_sync, 'operator' => 'gte']
		];
	}

	/**
	 * @param stdClass $record
	 *
	 * @return bool
	 * @throws WooStoreEntryException
	 */
	public function store($record) {
		global $current_user;
		$filters = [['field' => 'woo_id', 'value' => $record->id]];
		$id = $this->isExists($filters);

		if ($id) {
			$record->order_id = $id;
		}

		$data = $this->map($record);
		$user_id = $this->getAssignedUserId();
		$add = false;

		$data['woo_order_stage'] = $record->status;

		if(!$data['billing_account_id']) {
			$data['billing_account_id'] = $this->getBillingAccountId($record->customer_id);
		}

		if (isset($data['id'])) {
			$result = ListQuery::quick_fetch($this->model, $data['id']);
			$current_user->id = $user_id;
			$upd = RowUpdate::for_result($result, null, $user_id);

			unset($data['id']);
			$upd->set($data);
		} else {
			/** @var TallyUpdate $upd */
			$upd = $this->init($data);
			$current_user->id = $user_id;
			$upd->set($data);

			$group = $this->createGroup($record, $upd->getField('currency_id'));
			$upd->replaceGroups($group);

			$add = true;
		}

		try {
			$result = $upd->save();
		} catch (Exception $e) {
			throw new WooStoreEntryException($e->getMessage());
		}

		if ($result) {
			if ($add)
				$this->incrementPulledNum();

			$stored_id = $upd->getField('id');
			$this->statusHandler($record, $stored_id);

			return $stored_id;
		} else {
			return false;
		}
	}

	/**
	 * @param array $record
	 *
	 * @return int|null
	 */
	public function storeInWoo($record) {
		// Woo Status List:
		// pending, processing, on-hold, completed,
		// cancelled, refunded, failed and trash.
		$status = null;
		$result = null;

		if (! empty($record['so_stage'])) {
			$status = $this->convertSOStageToWooStatus($record['so_stage']);
		} elseif (isset($record['cancelled']) && $record['cancelled']) {
			$status = 'cancelled';
		} elseif (! empty($record['shipping_stage'])) {

			if ($this->isInvoiceCompleted($record)) {
				$status = 'completed';
			} else {
				$status = $this->convertShippingStageToWooStatus($record['shipping_stage']);
			}

		}

		if ($status) {
			$default_stages = ['pending', 'processing', 'on-hold', 'completed', 'cancelled'];

			if(!in_array($record['woo_order_stage'], $default_stages)) {
				$status = $record['woo_order_stage'];
			}

			$data = ['status' => $status];
			if($status != '')
				$result = $this->push($data, $record['woo_id'], $record['id']);
		}

		return $result;
	}

	/**
	 * Handle WooCommerce order's status
	 *
	 * @param stdClass $original_invoice_record
	 * @param string $invoice_id
	 */
	protected function statusHandler($original_invoice_record, $invoice_id) {}

	/**
	 * @param array $data
	 */
	protected function adjustDataFromWoo(&$data) {
		if (empty($data['id'])) {
			$data['assigned_user_id'] = $this->getAssignedUserId();
			$data['modified_user_id'] = $this->getAssignedUserId();
			$data['billing_account_id'] = $this->getBillingAccountId($data['customer_id'], $data);
			$data['billing_contact_id'] = $this->getBillingContactId($data['customer_id'], $data['billing_account_id']);
			$data['shipping_contact_id'] = $this->getBillingContactId($data['customer_id'], $data['shipping_account_id']);

			$data['currency_id'] = $this->getCurrencyId($data['currency']);
		} else {
			$account_id = $data['billing_account_id'] ?? null;
			$s_account_id = $data['shipping_account_id'] ?? null;
			$data['billing_contact_id'] = $this->getBillingContactId($data['customer_id'], $account_id);
			$data['shipping_contact_id'] = $this->getBillingContactId($data['customer_id'], $s_account_id);
		}

		unset($data['customer_company']);
		unset($data['customer_email']);
		unset($data['customer_phone']);
		unset($data['customer_first_name']);
		unset($data['customer_last_name']);

		unset($data['customer_id']);
		unset($data['currency']);

		$data['name'] = translate('LBL_ORDER_TITLE', 'WooCommerce') . " #{$data['name']}";
		$data['description'] = $data['name'];
	}

	/**
	 * @param array $data
	 *
	 * @return RowUpdate
	 */
	protected function init(/* @noinspection PhpUnusedParameterInspection */ $data) {
		return null;
	}

	/**
	 * @return array
	 */
	protected function getMap() {
		return [
			'id' => 'woo_id',
			'order_id' => 'id',
			'number' => 'name',
			'currency' => 'currency',
			'customer_id' => 'customer_id',

			//'status' => 'status'
			//'discount_total' => 'discount_total',
			//'discount_tax' => 'discount_tax',
			//'shipping_total' => 'shipping_total',
			//'shipping_tax' => 'shipping_tax',
			//'cart_tax' => 'cart_tax',
			//'total' => 'amount',
			//'total_tax' => 'total_tax',
			//'prices_include_tax' => 'prices_include_tax',

			'billing' => [
				'address_1' => 'billing_address_street',
				'city' => 'billing_address_city',
				'state' => 'billing_address_state',
				'postcode' => 'billing_address_postalcode',
				'country' => 'billing_address_country',

				'company' => 'customer_company',
				'first_name' => 'customer_first_name',
				'last_name' => 'customer_last_name',
				'email' => 'customer_email',
				'phone' => 'customer_phone'
			],
			'shipping' => [
				'address_1' => 'shipping_address_street',
				'city' => 'shipping_address_city',
				'state' => 'shipping_address_state',
				'postcode' => 'shipping_address_postalcode',
				'country' => 'shipping_address_country',
			]
		];
	}

	/**
	 * @param stdClass $record
	 * @param string $currency_id
	 *
	 * @return array
	 */
	private function createGroup($record, $currency_id) {
		$group_id = 'temp~0';

		$group = [
			$group_id => [
				'id' => $group_id,
				'status' => 'Closed Accepted',
				'lines' => [],
				'lines_order' => [],
				'adjusts' => [],
				'adjusts_order' => [],
				'group_type' => 'products',
				'tax' => $record->total_tax,
				'shipping' => $record->shipping_total
			]
		];

		$lines = $this->fillLineItems($record, $currency_id);

		foreach ($lines as $line) {
			$group[$group_id]['lines'][$line['id']] = $line;
			$group[$group_id]['lines_order'][] = $line['id'];
		}

		$adjusts = $this->fillAdjusts($record);

		foreach ($adjusts as $line) {
			$group[$group_id]['adjusts'][$line['id']] = $line;
			$group[$group_id]['adjusts_order'][] = $line['id'];
		}

		return $group;
	}

	/**
	 * @param stdClass $record
	 * @param $currency_id
	 *
	 * @return array
	 */
	private function fillLineItems(stdClass $record, $currency_id) {
		$line_items = [];

		foreach ($record->line_items as $item) {
			$item->currency_id = $currency_id;
			$line_items[] = $this->line_item->map($item);
		}

		foreach($record->fee_lines as $fee_line) {
			$line_items[] = [
				'id' => 'temp~' . rand(1000, 2000),
				'list_price' => $fee_line->total,
				'ext_price' => $fee_line->total,
				'unit_price' => $fee_line->total,
				'name' => $fee_line->name,
				'quantity' => 1,
				'total' => $fee_line->total
			];
		}

		return $line_items;
	}

	/**
	 * @param stdClass $record
	 *
	 * @return array
	 */
	private function fillAdjusts(stdClass $record) {
		$adjusts = [];
		$taxed_shipping = ! empty((float)$record->shipping_tax);

		$shipping = [
			'id' => 'temp~' . rand(1000, 2000),
			'amount' => $record->shipping_total,
			'related_type' => 'ShippingProviders',
			'type' => $taxed_shipping ? 'TaxedShipping' : 'UntaxedShipping',
			'tax_class_id' => self::TAX_EXEMPT_CODE_ID
		];

		if ($taxed_shipping) {
			$tax_rate = $this->line_item->getTaxRate();
			$tax_rate_id = $this->extractShippingTax($record);
			$rate_details = $tax_rate->getByWooId($tax_rate_id);

			if (isset($rate_details['code_id']))
				$shipping['tax_class_id'] = $rate_details['code_id'];

			$adjusts[] = $shipping;
		}

		//If untaxed shipping then add it to the end
		if (! $taxed_shipping)
			$adjusts[] = $shipping;

		return $adjusts;
	}

	/**
	 * @param stdClass $record
	 *
	 * @return int|string
	 */
	private function extractShippingTax(stdClass $record) {
		$result = self::TAX_EXEMPT_CODE_ID;

		if (isset($record->shipping_lines[0])) {
			$line = $record->shipping_lines[0];

			if (isset($line->taxes[0])) {
				$result = $line->taxes[0]->id;
			}
		}

		return $result;
	}

	/**
	 * @param string $stage: SO Stage
	 *
	 * @return string|null
	 */
	private function convertSOStageToWooStatus($stage) {
		$stages = [
			'Ordered' => 'on-hold',
			'In Manufacturing' => 'processing',
			'Partially Shipped and Invoiced' => 'processing',
			'Partially Shipped and not Invoiced' => 'processing',
			'Shipped and not Invoiced' => 'processing',
			'Closed - Shipped and Invoiced' => 'completed',
		];

		$result = null;

		if (isset($stages[$stage]))
			$result = $stages[$stage];

		return $result;
	}

	/**
	 * @param string $stage: Invoice's Shipping Stage
	 *
	 * @return string|null
	 */
	private function convertShippingStageToWooStatus($stage) {
		$stages = [
			'Pending' => 'pending',
			'Partially Shipped' => 'processing',
			'Shipped' => 'processing',
		];

		$result = null;

		if (isset($stages[$stage]))
			$result = $stages[$stage];

		return $result;
	}

	/**
	 * Is Invoice completed or not
	 *
	 * @param array $invoice - invoice's data
	 *
	 * @return bool
	 */
	private function isInvoiceCompleted($invoice) {
		$result = false;

		if ($invoice['amount_due'] == 0) {
			if ($invoice['shipping_stage'] == 'None' || $invoice['shipping_stage'] == 'Delivered') {
				$result = true;
			}
		}

		return $result;
	}

	/**
	 * @param int $customer_woo_id
	 * @param array $order_data
	 *
	 * @return null|string
	 */
	private function getBillingAccountId($customer_woo_id, $order_data = []) {
		return $this->customer->getAccountId($customer_woo_id, $order_data);
	}

	/**
	 * @param int $customer_woo_id
	 * @param null|string $account_id - billing account ID
	 *
	 * @return null|string
	 */
	private function getBillingContactId($customer_woo_id, $account_id = null) {
		return $this->customer->getContactId($customer_woo_id, $account_id);
	}

	/**
	 * @param string $iso4217
	 *
	 * @return string
	 */
	private function getCurrencyId($iso4217) {
		$currency_id = $this->retrieveCurrencyId($iso4217);

		if (! $currency_id) {
			require_bean('Currency');

			$currency = new Currency();
			$currency->loadDefaultCurrency();
			$currency_id = $currency->id;
		}

		return $currency_id;
	}

	/**
	 * @param string $iso4217
	 *
	 * @return string
	 */
	private function retrieveCurrencyId($iso4217) {
		/** @var MysqliManager $db */
		global $db;
		$query = "SELECT `id` FROM `currencies` WHERE `iso4217`='".$db->quote($iso4217)."' AND `deleted` = 0;";
		$result = $db->query($query);

		if($result) {
			$row = $db->fetchByAssoc($result);
			if (isset($row['id']))
				return $row['id'];
		}

		return '';
	}
}
