<?php

class WooPayment extends WooEntry {

	const DEFAULT_PAYMENT_METHOD = "Credit Card";

	/**
	 * @var RowResult
	 */
	private $invoice;

	/**
	 * @param RowResult $invoice
	 */
	public function setInvoice(RowResult $invoice) {
		$this->invoice = $invoice;
	}

	/**
	 * WooTaxRate constructor.
	 *
	 * @param WooRequest $request
	 */
	public function __construct(WooRequest $request) {
		$this->model = 'Payment';
		$this->endpoint = 'orders';
		$this->invoice = null;

		parent::__construct($request);
	}

	/**
	 * @param stdClass $record
	 *
	 * @return bool|string
	 * @throws WooStoreEntryException
	 */
	public function store($record) {
		if (! $this->invoice)
			return false;

		$order_data = $this->map($record);
		$payment_date = new DateTime($order_data['date_paid']);

		$data = [
			'account_id' => $this->invoice->getField('billing_account_id'),
			'direction' => 'incoming',
			'assigned_user_id' => $this->getAssignedUserId(),
			'modified_user_id' => $this->getAssignedUserId(),
			'notes' => translate('LBL_PAYMENT_CREATED_IN_WOO', 'WooCommerce'),

			'payment_date' => $payment_date->format('Y-m-d'),
			'payment_type' => $this->matchPaymentMethod(
				$order_data['payment_method'],
				$order_data['payment_method_title']
			),

			'amount' => $this->invoice->getField('amount_due'),
			'total_amount' => $this->invoice->getField('amount_due'),

			'related_invoice_id' => $this->invoice->getField('id'),
			'applied_amount' => $this->invoice->getField('amount_due'),

			'currency_id' => $this->invoice->getField('currency_id'),
			'exchange_rate' => $this->invoice->getField('exchange_rate'),
			'applied_currency_id' => $this->invoice->getField('currency_id'),
			'applied_rate' => $this->invoice->getField('exchange_rate'),
		];

		$upd = RowUpdate::blank_for_model($this->model, $this->getAssignedUserId());
		$upd->set($data);

		try {
			$result = $upd->save();
		} catch (Exception $e) {
			throw new WooStoreEntryException($e->getMessage());
		}

		if ($result) {
			$upd->addUpdateLinks('invoices',
				[
					[
						'invoice_id' => $this->invoice->getField('id'),
						'amount' => $this->invoice->getField('amount_due'),
						'amount_usdollar' => $this->invoice->getField('amount_due_usdollar'),
						'exchange_rate' => $this->invoice->getField('exchange_rate'),
					]
				]
			);

			return $upd->getField('id');
		} else {
			return false;
		}
	}

	public function storeInWoo($data) {}

	/**
	 * @param RowResult $invoice_result
	 */
	public function setRefunded(RowResult $invoice_result) {
		$list_result = $invoice_result->getLinkResults('payments');

		if(! $list_result->failed) {

			foreach($list_result->getRowIndexes() as $idx) {
				$row = $list_result->getRowResult($idx);

				if ($row->getField('refunded'))
					continue;

				$notes = $row->getField('notes') .' '. translate('LBL_PAYMENT_REFUND_NOTES', 'WooCommerce');

				$upd = RowUpdate::for_result($row, null, $this->getAssignedUserId());
				$upd->set('notes', $notes);
				$upd->set('refunded', 1);

				try {
					$upd->save();
				} catch (Exception $e) {}

			}

		}
	}

	/**
	 * @return array
	 */
	protected function getMap() {
		return [
			'payment_method' => 'payment_method',
			'payment_method_title' => 'payment_method_title',
			'date_paid' => 'date_paid'
		];
	}

	/**
	 * Match Woo Payment method with CRM method
	 *
	 * @param string $woo_method
	 * @param string $method_title
	 *
	 * @return string
	 */
	private function matchPaymentMethod($woo_method, $method_title) {

		switch ($woo_method) {
			case 'bacs':
				$result = 'Wire Transfer';
				break;
			case 'cheque':
				$result = 'Check';
				break;
			case 'cod':
				$result = 'Cash';
				break;
			case 'paypal':
				$result = 'PayPal';
				break;
			case 'chargebee':
				$result = 'ChargeBee';
				break;
			case 'stripe':
				$result = 'Stripe';
				break;
			default:
				$result = self::DEFAULT_PAYMENT_METHOD;
				break;
		}

		if ($result == self::DEFAULT_PAYMENT_METHOD
			&& strpos(strtolower($method_title), 'western union') !== false) {

			$result = 'Western Union';
		}

		return $result;
	}
}
