<?php
require_once('WooEntry.php');
require_once('WooProductCategory.php');
require_once('WooProductAttribute.php');

class WooProduct extends WooEntry {

	/**
	 * @var WooProductCategory
	 */
	private $category;

	/**
	 * @var WooProductAttribute
	 */
	private $attribute;

	/**
	 * WooProduct constructor.
	 *
	 * @param WooRequest $request
	 */
	public function __construct(WooRequest $request) {
		parent::__construct($request);

		$this->model = 'Product';
		$this->endpoint = 'products';
		$this->category = new WooProductCategory($request);
		$this->attribute = new WooProductAttribute($request);

		$this->init_from_woo = $this->getSetting('init_woo_products_sync');
		$this->init_to_woo = $this->getSetting('products_sync');
		$this->updates_to_woo = $this->getSetting('products_sync');
	}

	/**
	 * @return array
	 */
	public function getListSearchParams() {
		return ['status' => 'publish'];
	}

	/**
	 * Returns filters for getting the list of all records for push
	 *
	 * @return array
	 */
	public function getAllRecordsFilters() {
		$filters = parent::getAllRecordsFilters();
		$filters[] = ['field' => 'eshop', 'value' => 1];

		return $filters;
	}

	/**
	 * Returns filters for getting the list of updated records for push
	 *
	 * @param string $last_sync: date in Y-m-d H:i:s format
	 *
	 * @return array
	 */
	public function getUpdatedRecordsFilters($last_sync) {
		$filters = parent::getUpdatedRecordsFilters($last_sync);
		$filters[] = ['field' => 'eshop', 'value' => 1];

		return $filters;
	}

	/**
	 * @param stdClass $record
	 *
	 * @return bool|string
	 * @throws WooStoreEntryException
	 */
	public function store($record) {
		$id = $this->isExists($this->getIdentifyFilter($record));

		if ($id) {
			$record->product_id = $id;
		} else {
			if ($record->status != 'publish')
				return false;
		}

		$data = $this->map($record);
		$attributes = $data['attributes'];
		unset($data['attributes']);

		$user_id = $this->getAssignedUserId();
		$add = false;

		if (isset($data['id'])) {
			$result = ListQuery::quick_fetch($this->model, $data['id']);
			$upd = RowUpdate::for_result($result, null, $user_id);

			unset($data['id']);
		} else {
			$upd = RowUpdate::blank_for_model($this->model, $user_id);
			$add = true;
		}

		$upd->set($data);

		try {
			$result = $upd->save();
		} catch (Exception $e) {
			throw new WooStoreEntryException($e->getMessage());
		}

		if ($result) {
			if ($add)
				$this->incrementPulledNum();

			$stored_id = $upd->getField('id');

			if (! empty($attributes))
				$this->addAttributes($attributes, $stored_id);

			return $stored_id;
		} else {
			return false;
		}
	}

	/**
	 * @param array $record
	 *
	 * @return int|null
	 */
	public function storeInWoo($record) {
		$data = $this->mapForWoo($record);

		if (empty($record['woo_id'])) {
			$woo_id = $this->isExistsInWoo(
				[
					'search' => $data['name'],
					'category' => $data['categories'][0]['id']
				]
			);
		} else {
			$woo_id = $record['woo_id'];
		}

		return $this->push($data, $woo_id, $record['id']);
	}

	/**
	 * @param int $woo_id
	 *
	 * @return array
	 */
	public function getByWooId($woo_id) {
		$record = $this->loadByWooId($woo_id);

		if (! $record) {
			$record = $this->loadFromWoo($woo_id);
			$result = $this->map($record, false);
		} else {
			$result = $record;
		}

		return $result;
	}

	/**
	 * Adjust data from WooCommerce for compatibility with 1CRM
	 *
	 * @param array $data
	 */
	protected function adjustDataFromWoo(&$data) {
		if (empty($data['id'])) {
			$data['assigned_user_id'] = $this->getAssignedUserId();
			$data['modified_user_id'] = $this->getAssignedUserId();
			$data['cost'] = $data['list_price'];

			if (empty($data['cost']))
				$data['cost'] = 0;

			if (empty($data['list_price']))
				$data['list_price'] = 0;

			if (empty($data['purchase_price']))
				$data['purchase_price'] = 0;
		}

		if (empty($data['purchase_price']) && ! empty($data['list_price']))
			$data['purchase_price'] = $data['list_price'];

		if (! empty($data['id'])) {
			unset($data['list_price']);

			if (empty($data['purchase_price']))
				unset($data['purchase_price']);

			if (empty($data['cost']))
				unset($data['cost']);

			if (empty($data['manufacturers_part_no']))
				unset($data['manufacturers_part_no']);
		}

		if ($data['is_available'] == 'instock') {
			$data['is_available'] = 'yes';
		} else {
			$data['is_available'] = 'no';
		}

		$data['name'] = htmlspecialchars_decode($data['name']);

		if (is_array($data['product_category_id'])) {
			if (! empty($data['product_category_id'][0])) {
				if (property_exists($data['product_category_id'][0], 'id')) {
					$data['product_category_id'] = $this->category->getCategoryIdByWooId($data['product_category_id'][0]->id);
				}
			}
		}

		if (is_array($data['image_url'])) {
			if (! empty($data['image_url'][0])) {
				if (property_exists($data['image_url'][0], 'src'))
					$data['image_url'] = $data['image_url'][0]->src;
			}
		}
	}

	/**
	 * @param array $data
	 */
	protected function adjustDataForPush(&$data) {
		if (isset($data['product_id']))
			unset($data['product_id']);

		$data['slug'] = $this->createSlug($data['name']);

		/*$weight = '';
		if (! empty($row['weight_1']))
			$weight .= $row['weight_1'] . translate('LBL_PRODUCT_WEIGHT_UNIT_1', 'ProductCatalog');

		if (! empty($row['weight_2']))
			$weight .= ' ' . $row['weight_2'] . translate('LBL_PRODUCT_WEIGHT_UNIT_2', 'ProductCatalog');

		$row['weight'] = trim($weight);*/

		if ($data['stock_status'] == 'yes') {
			$data['stock_status'] = 'instock';
		} else {
			$data['stock_status'] = 'outofstock';
		}

		if (! empty($data['images']))
			$data['images'] = [['src' => $data['images']]];

		if (! empty($data['categories']))
			$data['categories'] = [['id' => $this->category->getWooId($data['categories'])]];
	}

	/**
	 * @return array
	 */
	protected function getMap() {
		return [
			'id' => 'woo_id',
			'product_id' => 'id',
			'name' => 'name',
			'description' => 'description_long',
			'short_description' => 'description',
			'sku' => 'manufacturers_part_no',
			'price' => 'list_price',
			'regular_price' => 'purchase_price',
			//'sale_price' => 'purchase_price',
			//'tax_class' => 'tax_code',
			'stock_quantity' => 'all_stock',
			'stock_status' => 'is_available',
			//'weight' => 'weight',
			'images' => 'image_url',
			'categories' => 'product_category_id',
			'attributes' => 'attributes',
		];
	}

	/**
	 * @param array $attributes
	 * @param string $product_id
	 *
	 * @return bool
	 * @throws WooStoreEntryException
	 */
	private function addAttributes($attributes, $product_id) {
		if (! is_array($attributes) || sizeof($attributes) === 0)
			return false;

		foreach ($attributes as $attribute) {
			if (! $attribute->visible)
				continue;
			if (empty($attribute->options))
				continue;
			foreach ($attribute->options as $option) {
				$item = clone $attribute;
				$item->value = $option;
				$item->product_id = $product_id;
				$this->attribute->store($item);
			}
		}

		return true;
	}

	/**
	 * @param stdClass $record
	 *
	 * @return array
	 */
	private function getIdentifyFilter($record) {

		return [
			[   'operator' => 'OR',
				'multiple' => [
					['field' => 'woo_id', 'value' => $record->id],
					['field' => 'manufacturers_part_no', 'value' => $record->sku],
				]
			]
		];

	}
}
