<?php
require_once('WooEntry.php');

class WooProductCategory extends WooEntry {

	/**
	 * @var array
	 */
	static $id_cache;

	/**
	 * @var array
	 */
	static $woo_id_cache;

	/**
	 * WooProductCategory constructor.
	 *
	 * @param WooRequest $request
	 */
	public function __construct(WooRequest $request) {
		parent::__construct($request);

		$this->model = 'ProductCategory';
		$this->endpoint = 'products/categories';

		self::$id_cache = [];
		self::$woo_id_cache = [];

		$this->init_from_woo = $this->getSetting('init_woo_products_sync');
		$this->init_to_woo = $this->getSetting('products_sync');
		$this->updates_to_woo = $this->getSetting('products_sync');
	}

	/**
	 * @param stdClass $record
	 *
	 * @return bool|string
	 * @throws WooStoreEntryException
	 */
	public function store($record) {
		$id = $this->isExists($this->getIdentifyFilter($record));

		if ($id)
			$record->cat_id = $id;

		$data = $this->map($record);
		$user_id = $this->getAssignedUserId();
		$add = false;

		if (isset($data['id'])) {
			$result = ListQuery::quick_fetch($this->model, $data['id']);
			$upd = RowUpdate::for_result($result, null, $user_id);

			unset($data['id']);
		} else {
			$upd = RowUpdate::blank_for_model($this->model, $user_id);
			$add = true;
		}

		$upd->set($data);

		try {
			$result = $upd->save();
		} catch (Exception $e) {
			throw new WooStoreEntryException($e->getMessage());
		}

		if ($result) {
			if ($add)
				$this->incrementPulledNum();

			return $upd->getField('id');
		} else {
			return false;
		}
	}

	/**
	 * @param array $record
	 *
	 * @return int|null
	 */
	public function storeInWoo($record) {
		if (empty($record['woo_id'])) {
			$woo_id = $this->isExistsInWoo(['search' => $record['name']]);
		} else {
			$woo_id = $record['woo_id'];
		}

		$data = [
			'name' => $record['name'],
			'slug' => $this->createSlug($record['name']),
			'description' => $record['description'],
		];

		if (isset($record['parent_woo_id']))
			$data['parent'] = $record['parent_woo_id'];

		return $this->push($data, $woo_id, $record['id']);
	}

	/**
	 * Adjust data from WooCommerce for compatibility with 1CRM
	 *
	 * @param array $data
	 */
	protected function adjustDataFromWoo(&$data) {
		if (! empty($data['parent'])) {
			$parent_woo_id = $this->getCategoryIdByWooId($data['parent']);

			if (! empty($parent_woo_id))
				$data['parent_id'] = $parent_woo_id;

			unset($data['parent']);
		}

		$data['name'] = htmlspecialchars_decode($data['name']);
	}

	/**
	 * @param int $woo_id
	 *
	 * @return string
	 */
	public function getCategoryIdByWooId($woo_id) {

		if (! isset(self::$id_cache[$woo_id])) {
			$cat_id = $this->isExists([['field' => 'woo_id', 'value' => $woo_id]]);

			if (! $cat_id) {
				$record = $this->loadFromWoo($woo_id);

				if (property_exists($record, 'parent') && ! empty($record->parent)) {
					$parent = $this->loadFromWoo($record->parent);

					try {
						$record->parent_id = $this->store($parent);
					} catch (WooStoreEntryException $e) {}
				}

				try {
					$cat_id = $this->store($record);
				} catch (WooStoreEntryException $e) {}
			}

			self::$id_cache[$woo_id] = $cat_id;
		}

		return self::$id_cache[$woo_id];

	}

	/**
	 * @param string $category_id
	 *
	 * @return null|int
	 */
	public function getWooId($category_id) {

		if (! isset(self::$woo_id_cache[$category_id])) {
			$woo_id = null;
			$record = $this->loadById($category_id);

			if (! empty($record['woo_id'])) {
				$woo_id = $record['woo_id'];
			} else {
				if (! empty($record['parent_id']))
					$record['parent_woo_id'] = $this->getWooId($record['parent_id']);

				$woo_id = $this->storeInWoo($record);
			}

			self::$woo_id_cache[$category_id] = $woo_id;
		}

		return self::$woo_id_cache[$category_id];

	}

	/**
	 * @return array
	 */
	protected function getMap() {
		return [
			'id' => 'woo_id',
			'cat_id' => 'id',
			'name' => 'name',
			'description' => 'description',
			'parent_id' => 'parent_id',
			'parent' => 'parent'
		];
	}

	/**
	 * @param string $id
	 *
	 * @return null|array
	 */
	private function loadById($id) {
		$record = ListQuery::quick_fetch('ProductCategory', $id);
		$result = null;

		if ($record)
			$result = $record->getRow();

		return $result;
	}

	/**
	 * @param stdClass $record
	 *
	 * @return array
	 */
	private function getIdentifyFilter($record) {
		return [
			[
				'operator' => 'OR',
				'multiple' => [
					['field' => 'woo_id', 'value' => $record->id],
					['field' => 'name', 'value' => $record->name],
				]
			]
		];
	}
}