<?php

class WooTaxRate extends WooEntry {

	/**
	 * @var array
	 */
	static $id_cache;

	/**
	 * WooTaxRate constructor.
	 *
	 * @param WooRequest $request
	 */
	public function __construct(WooRequest $request) {
		$this->model = 'TaxRate';
		$this->endpoint = 'taxes';
		self::$id_cache = [];

		parent::__construct($request);
	}

	/**
	 * @param stdClass $record
	 *
	 * @return bool|string
	 * @throws WooStoreEntryException
	 */
	public function store($record) {
		$id = $this->isExists($this->getIdentifyFilter($record));

		if ($id)
			$record->rate_id = $id;

		$data = $this->map($record);
		$user_id = $this->getAssignedUserId();

		if (isset($data['id'])) {
			$result = ListQuery::quick_fetch($this->model, $data['id']);
			$upd = RowUpdate::for_result($result, null, $user_id);

			$upd->set(['woo_id' => $data['woo_id']]);
			$result = $upd->save();
		} else {
			$tax_code_id = $this->createTaxCode($data);

			$upd = RowUpdate::blank_for_model($this->model, $user_id);
			$upd->set($data);

			try {
				$result = $upd->save();
				$upd->addUpdateLink('taxcodes', $tax_code_id);
			} catch (Exception $e) {
				throw new WooStoreEntryException($e->getMessage());
			}
		}

		if ($result) {
			return $upd->getField('id');
		} else {
			return false;
		}
	}

	public function storeInWoo($data) {}

	/**
	 * Get Tax Rate ID by Woo ID
	 * if rate doesn't exist in the CRM create it
	 *
	 * @param int $woo_id
	 *
	 * @return array
	 */
	public function getByWooId($woo_id) {

		if (! isset(self::$id_cache[$woo_id])) {
			$data = [];
			$rate_id = $this->isExists([['field' => 'woo_id', 'value' => $woo_id]]);

			if (! $rate_id) {
				$record = $this->loadFromWoo($woo_id);

				try {
					$rate_id = $this->store($record);
				} catch (WooStoreEntryException $e) {}
			}

			if ($rate_id) {
				$rate = ListQuery::quick_fetch($this->getModel(), $rate_id);
				$codes = $rate->getLinkResults('taxcodes');
				$rows = $codes->getRows();
				$code_id = null;

				foreach ($rows as $row) {
					$code_id = $row['id'];
					break;
				}

				$data = [
					'code_id' => $code_id,
					'rate_id' => $rate_id,
					'rate' => $rate->getField('rate'),
					'name' => $rate->getField('name')
				];
			}

			self::$id_cache[$woo_id] = $data;
		}

		return self::$id_cache[$woo_id];

	}

	/**
	 * @param array $data
	 */
	protected function adjustDataFromWoo(&$data) {
		if ($data['compounding']) {
			$data['compounding'] = 'compounded';
		}
	}

	/**
	 * @return array
	 */
	protected function getMap() {
		return [
			'id' => 'woo_id',
			'rate_id' => 'id',
			'rate' => 'rate',
			'name' => 'name',
			'compound' => 'compounding'
		];
	}

	/**
	 * Create Tax Code by Tax Rate
	 *
	 * @param array $tax_rate
	 *
	 * @return string
	 */
	private function createTaxCode($tax_rate) {
		$query = new ListQuery('TaxCode');
		$query->addSimpleFilter('code', $tax_rate['name']);
		$result = $query->runQuerySingle();

		if (! $result || $result->failed) {
			$data = [
				'code' => $tax_rate['name'],
				'name' => $tax_rate['name'],
				'position' => 2
			];

			$upd = RowUpdate::blank_for_model('TaxCode', $this->getAssignedUserId());
			$upd->set($data);

			try {
				$upd->save();
			} catch (Exception $e) {}

			return $upd->getField('id');
		} else {
			return $result->getField('id');
		}
	}

	/**
	 * @param stdClass $record
	 *
	 * @return array
	 */
	private function getIdentifyFilter($record) {
		return [
			[
				'operator' => 'OR',
				'multiple' => [
					['field' => 'woo_id', 'value' => $record->id],
					['field' => 'name', 'value' => $record->name],
				]
			]
		];
	}
}