<?php
/*
 *
 * The contents of this file are subject to the info@hand Software License Agreement Version 1.3
 *
 * ("License"); You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://1crm.com/pdf/swlicense.pdf>.
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 * specific language governing rights and limitations under the License,
 *
 * All copies of the Covered Code must include on each user interface screen:
 * (i) the 1CRM copyright notice,
 * (ii) the "Powered by the 1CRM Engine" logo,
 *
 * (iii) the "Powered by SugarCRM" logo, and
 * (iv) the SugarCRM copyright notice
 * in the same form as they appear in the distribution.
 * See full license for requirements.
 *
 * The Original Code is : 1CRM Engine proprietary commercial code.
 * The Initial Developer of this Original Code is 1CRM Corp.
 * and it is Copyright (C) 2004-2012 by 1CRM Corp.
 *
 * All Rights Reserved.
 * Portions created by SugarCRM are Copyright (C) 2004-2008 SugarCRM, Inc.;
 * All Rights Reserved.
 *
 */


class WFProcessingError extends IAHInternalError {}
class WFTemporaryError extends WFProcessingError {}
class WFMissingParamError extends WFProcessingError {}


abstract class WFProcessorBase {

	var $module;
	var $type;
	var $label;
	var $icon;
	var $bean_name;
	var $as_user_id;
	var $params;
	var $allow_when;
	var $perform_when;
	var $source;
	var $action;
	var $workflow_id;
	var $operation_id;
	var $deferred_id;
	var $always_defer = false;
	var $form_name = 'WFActionForm';
	var $related_model;
	var $unique;
	var $alt_cache;

	static $debug;


	function process(RowUpdate $record) {
		$rel = $this->initRelatedRecord($record);
		if($rel) {
			if($rel->prohibit_workflow !== 0)
				$rel->prohibit_workflow = true;
			if($rel->validate()) {
				if($rel->save()) {
					$this->processRelatedRecord($record, $rel);
					return $rel->getPrimaryKeyValue();
				} else {
					$GLOBALS['log']->error("Error saving new record from workflow: ".$rel->getErrors(true));
				}
			} else {
				$GLOBALS['log']->error("Error validating new record from workflow: ".$rel->getErrors(true));
			}
		}
	}

	// for simple workflows that create a single related record
	function initRelatedRecord(RowUpdate $record) {
		if($this->related_model) {
			$user_id = $this->getUserId();
			$upd = RowUpdate::blank_for_model($this->related_model, $user_id);
			$values = $this->getRecordParams($record);

			if($this->module == 'Contacts') {
				$values['contact_id'] = $record->getPrimaryKeyValue();
			}

			if($values['assign_to_parent_user'])
				$values['assigned_user_id'] = $record->getField('assigned_user_id');
			else
				$values['assigned_user_id'] = $user_id;
			$upd->set($values);
			return $upd;
		}
	}

	function getRecordParams(RowUpdate $record) {
		return $this->getParams();
	}

	// any post processing once the related record is saved
	function processRelatedRecord(RowUpdate $record, RowUpdate $related) {
	}

	function __construct($module, $spec) {
		$this->module = $module;
		$this->bean_name = AppConfig::module_primary_bean($module);
		if(! empty($spec['type']))
			$this->type = $spec['type'];
		if(! empty($spec['icon']))
			$this->setIcon($spec['icon']);
		if(! empty($spec['label']))
			$this->setLabel($spec['label']);
		if(isset($spec['allow_when']))
			$this->allow_when = $spec['allow_when'];
	}

	static function get_actions($module=null, $manual=false) {
		static $acts = [];
		$key = ($module ? $module : '') . ($manual ? '-1' : '-0');
		if(! isset($acts[$key])) {
			$load = AppConfig::setting('modinfo.procs', array());
			$ret = array();
			foreach($load as $idx => $act) {
				$act['type'] = $idx;
				if(! array_get_default($act, 'enabled', true)) {
					continue;
				}
				if(empty($act['label'])) {
					$vmod = array_get_default($act, 'vname_module', $act['module']);
					$act['label'] = translate($act['vname'], $vmod);
				}
				if(empty($act['icon'])) {
					$act['icon'] = 'theme-icon module-'.$act['module'];
				}
				$act['module_label'] = translate_dd('moduleList', $act['module']);
				if(isset($act['allow_when']) && ! is_array($allow = $act['allow_when'])) {
					$act['allow_when'] = array_map('trim', explode(',', $allow));
				}
				else if(empty($act['allow_when']))
					$act['allow_when'] = array('after', 'manual');
				if(isset($act['allow_modules']) && ! is_array($allow = $act['allow_modules'])) {
					$act['allow_modules'] = array_map('trim', explode(',', $allow));
				}
				if(isset($act['block_modules']) && ! is_array($block = $act['block_modules'])) {
					$act['block_modules'] = array_map('trim', explode(',', $block));
				}
				if($manual && ! in_array('manual', $act['allow_when'])) {
					continue;
				}
				$ret[$idx] = $act;
			}
			uasort($ret, array('WFProcessorBase', 'cmp_actions'));
			$acts[$key] = $ret;
		}
		if($module) {
			$ret = array();
			foreach($acts[$key] as $idx => $act) {
				if(self::check_module_allowed($module, $act))
					$ret[$idx] = $act;
			}
			$acts[$key] = $ret;
		}
		return $acts[$key];
	}

	static function cmp_actions($a, $b) {
		$oa = array_get_default($a, 'order', 20);
		$ob = array_get_default($b, 'order', 20);
		if($oa != $ob) return ($oa < $ob) ? -1 : 1;
		if($a['module'] == $b['module'])
			return strcmp($a['label'], $b['label']);
		return strcmp($a['module_label'], $b['module_label']);
	}

	static function get_action_options($module=null, $manual=false) {
		$ret = array();
		foreach(self::get_actions($module, $manual) as $idx => $act) {
			// FIXME - custom label for manual mode?
			$ret[$idx] = array('icon' => $act['icon'], 'label' => $act['label']);
		}
		return $ret;
	}

	static function get_action_info($action_type) {
		$acts = self::get_actions();
		if(isset($acts[$action_type]))
			return $acts[$action_type];
	}

	static function create_module_action($module, $action_type, $manual=false) {
		$act = self::get_action_info($action_type);
		if($act) {
			if($manual && ! in_array('manual', $act['allow_when'])) {
				return false;
			}
			if(! self::check_module_allowed($module, $act)) {
				return false;
			}
			$cls = $act['class_name'];
			if(! class_exists($cls)) {
				$cls_path = array_get_default($act, 'file');
				if(! $cls_path)
					$cls_path = "modules/{$act['module']}/procs/{$cls}.php";
				require_once($cls_path);
			}
			return new $cls($module, $act);
		}
	}

	static function check_module_allowed($module, $act) {
		if(! is_array($act)) $act = self::get_action_info($act);
		if(! $act || ! $module) return;
		if(isset($act['allow_modules']) && ! in_array($module, $act['allow_modules']))
			return false;
		if(isset($act['block_modules']) && in_array($module, $act['block_modules']))
			return false;
		if(isset($act['check_module']))
			return call_hook_function($act['check_module'], null, array($module));
		return true;
	}

	function getIcon() {
		return 'module-Workflow';
		return $this->icon;
	}
	function setIcon($icon) {
		$this->icon = $icon;
	}

	function getLabel() {
		return $this->label;
	}
	function setLabel($label) {
		$this->label = $label;
	}

	function getModuleName() {
		return $this->module;
	}

	function getRequiredFields() {
		$ps = array();
		foreach($this->getParams() as $k => $v) {
			if(substr($k, -4) == '_alt' && substr($v, 0, 1) == '=')
				$ps[] = substr($v, 1);
		}
		return $ps;
	}

	// may override for more efficient processing
	function processList(ListResult $records) {
		foreach($records as $rec)
			$this->process($rec->getRowUpdate());
	}

	function loadDefinition(RowResult $act) {
		$this->setPerformWhen($act->getField('perform_when'));
		$this->setUserId($act->getField('assigned_user_id'));
		$this->setParams($act->getField('params'));
	}

	function getUserId() {
		if($this->as_user_id) return $this->as_user_id;
		return AppConfig::current_user_id();
	}

	function setUserId($user_id) {
		$this->as_user_id = $user_id;
	}

	function setSource($source, $action, $rel_id=null, $op_id=null, $unique = null) {
		$this->source = $source;
		$this->action = $action;
		if($source == 'workflow') {
			$this->workflow_id = $rel_id;
			$this->operation_id = $op_id;
			$this->unique = $unique;
		}
	}

	function getParams($json=false) {
		$ret = $this->params;
		if(! $ret) $ret = array();
		if($this->related_model) {
			$refs = array();
			foreach($this->getConfigFields() as $idx => $f) {
				$name = array_get_default($f, 'name', $idx);
				if(is_array($f) && array_get_default($f, 'type') == 'ref') {
					$refs[$name] = $f['id_name'];
				}
			}
			if($refs) {
				$fields = $this->getConfigFields();
				$result = new RowResult();
				$result->fields = $fields;
				$result->assign($ret);
				$lq = new ListQuery($this->related_model, $this->getConfigFields());
				$lq->populateSecondary($result, false);
				if($result->row)
					$ret = $result->row;
			}
		}
		if($json) {
			$js = getJSONobj();
			$ret = $js->encode($ret);
		}
		return $ret;
	}

	function setParams($params) {
		$this->params = array();
		if(is_string($params)) {
			$json = getJSONobj();
			$params = $json->decode($params);
		}
		if(is_array($params))
			$this->params = $params;
	}

	function getParam($name, $default=null) {
		return array_get_default($this->params, $name, $default);
	}

	function setParam($name, $value) {
		$this->params[$name] = $value;
	}

	function getPerformWhen() {
		return $this->perform_when;
	}

	function setPerformWhen($p) {
		$this->perform_when = $p;
	}

	function getConfigLayout($context, $hidden=null) {
		$layout = array(
			'form_hooks' => array('onsubmit' => ''),
			'form_hidden_fields' => $hidden,
			'type' => 'configform',
			'base_module' => 'Workflow',
			//'buttons_position' => 'bottom',
			'class' => 'workflow-action',
			'editable' => true,
			'sections' => $this->getConfigSections(),
			'icon' => $this->getIcon(),
			'title' => $this->getLabel(),
		);
		if($context == 'wf_view') {
			$layout['editable'] = false;
			$layout['type'] = 'view';
		}
		if($context == 'workflow' || $context == 'listview') {
			$layout['form_buttons']['save'] = array(
				'vname' => 'LBL_ACCEPT_BUTTON_LABEL',
				'icon' => 'icon-accept',
				'type' => 'submit',
				'perform' => 'WorkflowEditor.submitConfigForm(this.form); return false;',
			);
			$layout['form_buttons']['cancel'] = array(
				'vname' => 'LBL_CANCEL_BUTTON_LABEL',
				'icon' => 'icon-cancel',
				'perform' => 'SUGAR.popups.close()',
				'type' => 'button',
			);
		}
		if($context == 'listview') {
			$layout['scripts'][] = ['file' => 'modules/Workflow/workflow.js'];
			$layout['scripts'][] = ['language' => 'Workflow'];
		}

		return $layout;
	}

	function getConfigSections() {
		$fields = $this->getConfigFields();
		$elts = array();
		foreach($fields as $f => $def)
			if(empty($def['hidden']))
				$elts[] = $f;
		$sections = array(
			array(
				'id' => 'main',
				'label_position' => 'left',
				'columns' => 1,
				'widths' => array('100%'),
				'elements' => $elts,
			),
		);
		return $sections;
	}

	function getAllowWhen() {
		if($this->allow_when) {
			$allow = $this->allow_when;
			if($this->source == 'listview') {
				$nallow = ['immed'];
				foreach($allow as $a) {
					if($a != 'before' && $a != 'after')
						$nallow[] = $a;
				}
				$allow = $nallow;
			}
			return $allow;
		}
	}

	function getConfigFields() {
		$ret = array();
		$allow = $this->getAllowWhen();
		if($allow) {
			$spec = array(
				'vname' => 'LBL_OPERATIONS_PERFORM',
				'widget' => 'PerformWhenInput',
				'allow_when' => $allow,
				'required' => true,
			);
			if(in_array('after', $allow))
				$spec['default'] = 'after';
			else if(in_array('immed', $allow))
				$spec['default'] = 'immed';
			$ret['perform_when'] = $spec;
			$ret['perform_when_unit'] = array(
				'type' => 'enum',
				'options' => 'OPTIONS_TIME_CONDITION',
				'hidden' => true,
			);
			$ret['unique'] = array(
				'type' => 'bool',
				'hidden' => true,
			);
		}
		return $ret;
	}

	function getConfigDefaults() {
		$ret = array();
		if($this->related_model) {
			$fields = $this->getConfigFields();
			$upd = RowUpdate::blank_for_model($this->related_model);
			foreach($fields as $k => $f) {
				if(array_key_exists('default', $f)) {
					$ret[$k] = $f['default'];
				} else {
					$val = $upd->getField($k);
					if(isset($val))
						$ret[$k] = $val;
				}
				if(isset($f['id_name'])) {
					$idn = $f['id_name'];
					if(! isset($fields[$idn]['default'])) {
						$val = $upd->getField($idn);
						if(isset($val))
							$ret[$idn] = $val;
					}
				}
			}
		}
		return $ret;
	}

	function getRenderFields() {
		return $this->getConfigFields();
	}

	function getConfigParams() {
		$vals = array_merge($this->getConfigDefaults(), $this->getParams());
		return $vals;
	}

	function getRenderParams() {
		return $this->getConfigParams();
	}

	function renderConfigForm($context, $entry_id) {
		require_once('include/layout/FieldFormatter.php');
		require_once('include/layout/forms/FormGenerator.php');
		$result = new RowResult;
		$result->base_model = 'WFOperation';
		$result->module_dirs[] = 'Workflow';
		$this->setSource($context, 'config');

		$hidden = compact('context', 'entry_id');
		$hidden['trigger_module'] = $this->module;
		$hidden['operation_type'] = $this->type;

		$result->fields = $this->getConfigFields();
		$result->row = $this->getConfigParams();
		$result->row['perform_when'] = $this->getPerformWhen();
		foreach($result->fields as $fid => &$f) {
			if(! isset($f['editable'])) $f['editable'] = true;
			if(! isset($result->row[$fid]) && isset($f['default']))
				$result->row[$fid] = $f['default'];
			if(empty($f['editable']) && isset($f['default']))
				$hidden[$fid] = $f['default'];
		}
		$fmt = new FieldFormatter('html', 'editview');
		$result->formatted = $fmt->formatRow($result->fields, $result->row);

		$layout = $this->getConfigLayout($context, $hidden);

		$model = new ModelDef('WFOperation');
		$model->addDisplayFieldDefinitions($result->fields);
		$gen = FormGenerator::html_form($model, $layout, $this->form_name, 'wfaction_'.$entry_id);
		$gen->getLayout()->setInPopup(true);

		$this->beforeRenderConfigForm($gen, $context, $entry_id);

		$gen->renderForm($result);
		$out = $gen->getResult();
		$gen->exportIncludes();

		return $out;
	}

	function beforeRenderConfigForm(HtmlFormGenerator $gen, $context, $entry_id) {
		// for adding JS, CSS includes
	}

	function loadConfigRequest() {
		require_once('include/layout/FieldFormatter.php');

		$fields = $this->getConfigFields();
		$add_f = array();
		$formatted = array();
		foreach($fields as $f => $fdef) {
			if(isset($_REQUEST[$f]))
				$formatted[$f] = $_REQUEST[$f];
			$fields[$f]['source']['type'] = 'db'; // required for unformatter
		}

		$fmt = new FieldFormatter('html', 'editview');
		$row = $fmt->unformatRow($fields, $formatted);
		$when = array_get_default($row, 'perform_when');
		unset($row['perform_when']);
		$ret = array(
			'trigger_module' => $this->module,
			'operation_type' => $this->type,
			'perform_when' => $when,
			'label' => $this->getLabel(),
			'params' => $row,
		);
		return $ret;
	}

	function renderView($context, $entry_id) {
		require_once('include/layout/FieldFormatter.php');
		require_once('include/layout/forms/FormGenerator.php');
		$result = new RowResult;
		$result->base_model = 'WFOperation';
		$result->module_dirs[] = 'Workflow';

		$result->row = $this->getRenderParams();
		$result->row['perform_when'] = $this->getPerformWhen();
		$result->fields = $this->getRenderFields();
		foreach($result->fields as $idx => &$f) {
			if(! isset($f['name'])) $f['name'] = $idx;
			$check_idx = array_get_default($f, 'check_hide', $idx);
			if(empty($f['required']) && empty($result->row[$check_idx])) {
				$f['hidden'] = true;
			}
			if(! empty($f['hide_label']))
				$f['hide_label'] = false;
			if(isset($f['type']) && $f['type'] == 'ref')
				$f['detail_link'] = true;
		}
		$fmt = new FieldFormatter('html', 'view');
		$result->formatted = $fmt->formatRow($result->fields, $result->row);

		$layout = $this->getConfigLayout($context);

		$model = new ModelDef('WFOperation');
		$model->addDisplayFieldDefinitions($result->fields);
		$gen = FormGenerator::html_form($model, $layout, 'WFActionForm', 'wfaction_'.$entry_id);

		$gen->renderBodyOnly($result);
		$out = $gen->getResult();
		$gen->exportIncludes();

		return $out;
	}

	function getRelativeDate($name, $with_time=true) {
		global $timedate;
		$type = $this->getParam($name.'_type');
		$count = $this->getParam($name.'_count');
		$time = $this->getParam($name.'_time', '12:00:00');
		$fmt = 'Y-m-d H:i:s';
		if($type == 'same_day') {
			$start = gmdate($fmt);
		} else if($type == 'next_day') {
			$start = gmdate($fmt, time() + 24*3600);
		} else if($type == 'days_of') {
			$start = gmdate($fmt, time() + 24*3600*$count);
		} else
			return;
		$date_start = $timedate->handle_offset($start, $fmt, true);
		list($dt, $tm) = explode(' ', $date_start, 2);
		$date_start = $dt . ' ' . $time;
		$ret = $timedate->handle_offset($date_start, $fmt, false);
		if(! $with_time) {
			list($dt, $tm) = explode(' ', $date_start, 2);
			$ret = $dt;
		}
		return $ret;
	}

	function getRelatedFields(array $spec) {
		$ftype = array_get_default($spec, 'type');
		$fname = array_get_default($spec, 'name');
		if(! $ftype) return;
		if(! isset($this->alt_cache)) {
			$this->alt_cache = ['ref' => [], 'text' => []];
		}
		if($ftype == 'ref') {
			$rel_type = array_get_default($spec, 'bean_name');
			if(! $rel_type) return;
			if(! in_array($rel_type, $this->alt_cache['ref']))
				$this->alt_cache['ref'][$rel_type] = $this->getRelatedRefFields($rel_type);
			$ret = $this->alt_cache['ref'][$rel_type];
		}
		else {
			if(! in_array($ftype, $this->alt_cache['text']))
				$this->alt_cache['text'][$ftype] = $this->getRelatedTextFields($ftype);
			$ret = $this->alt_cache['text'][$ftype];
		}
		if(isset($ret)) {
			unset($ret['='.$fname]);
			return $ret;
		}
	}

	function getRelatedTextFields($field_type=null) {
		$ban = array(
			'id', 'ref_id', 'blob', 'file_ref', 'image_ref', 'time',
			'enum', 'multienum', // would need to have the same options
		);
		require_once('include/layout/forms/EditableForm.php');
		$number = $exact = false;
		$exact_types = ['bool', 'duration', 'currency', 'email', 'phone', 'url', 'html'];
		if($field_type) {
			if(in_array($field_type, EditableForm::$number_types))
				$number = true;
			if(in_array($field_type, $exact_types))
				$exact = true;
		}
		$ret = array();
		$model = new ModelDef($this->bean_name);
		if(! in_array($field_type, $ban))
			foreach($model->getFieldDefinitions() as $k => $f) {
				$ft = array_get_default($f, 'type');
				$ok = 0;
				if($k == 'deleted' || ! empty($f['protected']) || $ft == 'id' || $ft == 'ref_id')
					continue;
				else if($exact || in_array($ft, $exact_types))
					$ok = ($ft == $field_type);
				else if($number)
					$ok = in_array($ft, EditableForm::$number_types);
				else if($field_type == 'date')
					$ok = ($ft == 'datetime' || $ft == 'date');
				else if($field_type == 'datetime')
					$ok = ($ft == 'datetime');
				else if($ft == 'ref')
					$ok = 1; // copies text from related name
				else if($ft == 'text')
					$ok = ($field_type == 'text');
				else if(! in_array($ft, $ban) && empty($f['widget']) && (! isset($f['source']['type']) || $f['source']['type'] != 'non-db') && ! empty($f['vname']))
					$ok = 1;
				if($ok) {
					$lbl = translate($f['vname'], array_get_default($f, 'vname_module', $this->module));
					$ret['='.$k] = translate('LBL_RELATED_PREFIX', 'Workflow') . $lbl;
				}
			}
		return $ret;
	}

	function getRelatedRefFields($rel_type, $parentField = null) {
		$model = $this->getModelDefinition($this->bean_name);
		$mod = AppConfig::module_for_model($rel_type);
		if(! $mod) return;
		$ret = [];
		foreach($model->getFieldDefinitions() as $k => $f) {
			if($f['type'] == 'ref') {
				if((isset($f['bean_name']) && $f['bean_name'] == $rel_type) || (isset($f['module']) && $f['module'] == $mod)) {
					$lbl = translate($f['vname'], array_get_default($f, 'vname_module', $this->module));
					$ret['='.$k] = translate('LBL_RELATED_PREFIX', 'Workflow') . $lbl;
				}
				if (!$parentField) {
					$lblPrefix = translate($f['vname'], array_get_default($f, 'vname_module', $this->module));
					$parentModelName = array_get_default($f, 'bean_name', AppConfig::module_primary_bean(array_get_default($f, 'module')));
					$parentMod = AppConfig::module_for_model($parentModelName);
					if ($parentModelName && $parentMod) {
						$parentModel = $this->getModelDefinition($parentModelName);
						foreach($parentModel->getFieldDefinitions() as $kk => $ff) {
							if($ff['type'] == 'ref') {
								if((isset($ff['bean_name']) && $ff['bean_name'] == $rel_type) || (isset($ff['module']) && $ff['module'] == $mod)) {
									$lbl = translate($ff['vname'], array_get_default($ff, 'vname_module', $parentMod));
									$ret['='. $k . '.' . $kk ] = translate('LBL_RELATED_PREFIX', 'Workflow') . $lblPrefix . ' : ' . $lbl;
								}
							}
						}
					}
				}
			}
		}
		return $ret;
	}


	function getModelDefinition($modelName) {
		static $models = [];
		$model = isset($models[$modelName]) ? $models[$modelName] : new ModelDef($modelName);
		$models[$modelName] = $model;
		return $model;
	}

	function getNestedFieldDefinition($base, $name) {
		if (strpos($name, '.') === false) {
			return [false, $base->getFieldDefinition($name)];
		}
		$parts = explode('.', $name, 2);
		$def =  $base->getFieldDefinition($parts[0]);
		if (!$def || array_get_default($def, 'type') != 'ref')
			return null;
		$parentModelName = array_get_default($def, 'bean_name', AppConfig::module_primary_bean(array_get_default($def, 'module')));
		if (!$parentModelName)
			return null;
		$parentModel = $this->getModelDefinition($parentModelName);
		if (!$parentModel)
			return null;
		return [$parts[0], $parentModel->getFieldDefinition($parts[1])];
	}

	function combineValuesWithAlt($base, RowUpdate $upd, array $fields, array $values) {
		$ret = [];
		$empty = new StdClass;
		$fieldMap = [];
		foreach($fields as $field) {
			$f = is_array($field) ? array_get_default($field, 'field', array_get_default($field, 'name')) : $field;
			if(! $f || ! is_string($f)) continue;
			$def = $upd->getFieldDefinition($f);
			$ftype = $def ? array_get_default($def, 'type') : null;
			$alt = array_get_default($values, $f.'_alt');
			if($alt && strlen($alt) > 1 && $alt[0] == '=') {
				$altf = substr($alt, 1);
				if($ftype == 'ref') {
					$bdef = $this->getNestedFieldDefinition($base, $altf);
					if ($bdef) {
						list ($prefix, $bdef) = $bdef;
						if($bdef && array_get_default($bdef, 'type') == 'ref') {
							if (!$prefix) {
								$fieldMap[$def['id_name']] = $bdef['id_name'];
							} else {
								$fieldMap[$def['id_name']] = $prefix . '.' . $bdef['id_name'];
							}
						}
					}
				} else {
					$fieldMap[$f] = $altf;
				}
			} else if(array_get_default($def, 'type') == 'ref') {
				$GLOBALS['log']->fatal($def);
				$GLOBALS['log']->fatal($values);
				$id_name = $def['id_name'];
				if (isset($values[$id_name])) {
					$ret[$id_name] = $values[$id_name];
				}
			} else if(array_key_exists($f, $values)) {
				$ret[$f] = $values[$f];
			}

		}
		$missing = [];
		if (!empty($fieldMap)) {
			foreach ($fieldMap as $_ => $sourceField) {
				if ($base->getFieldValue($sourceField, $empty) === $empty) {
					$missing[] = $sourceField;
				}
			}
			if (!empty($missing)) {
				$existing = ListQuery::quick_fetch($base->model_name, $base->getField('id'), $missing);
				if ($existing) {
					foreach ($missing as $fname) {
						$base->set($fname, $existing->getField($fname));
					}
				}
			}
		}
		foreach ($fieldMap as $destField => $sourceField) {
			$ret[$destField] = $base->getField($sourceField);
		}
		return $ret;
	}

	function setValuesWithAlt($base, RowUpdate $upd, array $fields, array $values) {
		$updates = $this->combineValuesWithAlt($base, $upd, $fields, $values);
		if($updates)
			$upd->set($updates);
	}

	function handleRelatedRef($base, $upd, $field) {
		$fdef = $upd->getFieldDefinition($field);
		if($fdef && isset($fdef['id_name'])) {
			$idn = $fdef['id_name'];
			$p = $this->getParam($fdef['name'].'_alt');
			if($p && substr($p, 0, 1) == '=') {
				$def = $base->getFieldDefinition(substr($p, 1));
				if($def && isset($def['id_name'])) {
					$idv = $base->getField($def['id_name']);
					$upd->set($idn, $idv);
				}
			}
		}
	}

	function handleRelatedValue($base, $upd, $field) {
		$fdef = $upd->getFieldDefinition($field);
		$p = $this->getParam($fdef['name'].'_alt');
		if($p && substr($p, 0, 1) == '=') {
			$val = $base->getField(substr($p, 1));
			$upd->set($field, $val);
		}
	}

	static function process_workflow(RowUpdate $rowUpdate, $workflow, $opt, $owner=null) {
		static $admin;
		if(! isset($admin)) {
			$admin = new User;
			$admin->retrieveAdminUser();
		}
		$user_id = null;
		if ($workflow->operation_owner == 'trigger') {
			$user_id = $owner ? $owner : AppConfig::current_user_id();
		} elseif ($workflow->operation_owner == 'owner') {
			$user_id = $rowUpdate->getField('assigned_user_id');
			if (!$user_id)
				$user_id = $rowUpdate->getField('created_by');
			if (!$user_id)
				$user_id = AppConfig::current_user_id();
		} elseif ($workflow->operation_owner != 'admin') {
			if (strlen(get_assigned_user_name($workflow->operation_owner)))
				$user_id = $workflow->operation_owner;
		}
		if (!$user_id) {
			$user_id = $admin->id;
		}

		$GLOBALS['log']->debug('operation type: '.$opt->operation_type);

		$action = WFProcessorBase::create_module_action($workflow->trigger_module, $opt->operation_type);
		if(! $action) {
			$GLOBALS['log']->error("Unknown operation type: {$workflow->trigger_module}/{$opt->operation_type}");
			return;
		}

		$action->setSource('workflow', $workflow->trigger_action, $workflow->id, $opt->id, $opt->unique);
		$action->setUserId($user_id);
		$action->setPerformWhen($opt->perform_when);
		$action->setParams($opt->params);

		self::debug_msg(get_class($action), 'run workflow');
		try {
			$defer = $action->processDeferral($rowUpdate);
			if($defer)
				return $defer;
			$GLOBALS['log']->debug('calling process()');
			return $action->process($rowUpdate);
		} catch(WFProcessingError $e) {
			$GLOBALS['log']->error("Caught WFProcessingError: ".$e->getMessage());
		}
		// on temporary error, create deferral record? only if retries specified in params
	}


	static function process_manual(RowUpdate $rowUpdate, $action_name, $trigger, $when='immed', $params=null, $owner=null) {
		static $admin;
		if(! isset($admin)) {
			$admin = new User;
			$admin->retrieveAdminUser();
		}
		$user_id = $owner ? $owner : AppConfig::current_user_id();
		if (!$user_id)
			$user_id = $admin->id;

		$module = $rowUpdate->getModuleDir();
		$GLOBALS['log']->debug("manual workflow operation: $module/$action_name");

		$action = WFProcessorBase::create_module_action($module, $action_name, true);
		if(! $action) {
			$GLOBALS['log']->error("Unknown operation type: $module/$action_name");
			return;
		}

		$action->setSource('manual', $trigger);
		$action->setUserId($user_id);
		if($when == 'defer') {
			$when = 'immed';
			$action->always_defer = true;
		}
		$action->setPerformWhen($when);
		$action->setParams($params);

		self::debug_msg(get_class($action), 'run workflow');
		try {
			$defer = $action->processDeferral($rowUpdate);
			if($defer)
				return $defer;
			return $action->process($rowUpdate);
		} catch(WFProcessingError $e) {
			$GLOBALS['log']->error("Caught WFProcessingError: ".$e->getMessage());
		}
		// on temporary error, create deferral record? only if retries specified in params
	}


	function getTimeInterval() {
		static $ivals = array('h' => 'hours', 'd' => 'days', 'w' => 'weeks', 'm' => 'months', 'y' => 'years');
		$unit = $this->getParam('perform_when_unit');
		if (preg_match('/^(-?\d+)([a-z])$/', strtolower($unit), $m)) {
			if(isset($ivals[$m[2]])) {
				$offs = $m[1];
				if($offs >= 0) $offs = '+'.$offs;
				$offs .= ' ' . $ivals[$m[2]];
				$date = gmdate('Y-m-d H:i:s', strtotime("NOW {$offs}"));
				return $date;
			}
		}
	}

	function initDeferred(RowUpdate $rowUpdate) {
		$module = $rowUpdate->getModuleDir();
		$rowid = $rowUpdate->getPrimaryKeyValue();
		$ival = $this->getTimeInterval();
		$userid = $this->getUserId();
		$defer = RowUpdate::blank_for_model('RecordAction', $userid);
		$status = ($this->perform_when == 'scheduled' ? 'scheduled' : 'pending');
		if($status != 'scheduled')
			$ival = null;
		$defer->set(array(
			'id' => create_guid(),
			'type' => $this->type,
			'status' => $status,
			'date_start' => $ival,
			'source' => $this->source,
			'action' => $this->action,
			'parent_type' => $module,
			'parent_id' => $rowid,
			'assigned_user_id' => $userid,
		));
		if($this->source == 'workflow') {
			$defer->set(array(
				'workflow_id' => $this->workflow_id,
				'wf_action_id' => $this->operation_id,
			));
			$defer->setRelatedData('unique', $this->unique);
		}
		$json = getJSONobj();
		$this->deferred_id = $defer->getField('id');
		$params = $this->getDeferParams($rowUpdate);
		if(isset($params['description'])) {
			$defer->set('description', $params['description']);
			unset($params['description']);
		}
		$defer->set('params', $json->encode($params));
		return $defer;
	}

	function checkDefer() {
		if($this->always_defer) return true;
	}

	function getDeferParams(RowUpdate $rowUpdate) {
		return $this->getParams();
	}

	function processDeferral(RowUpdate $rowUpdate) {
		$ival = $this->getTimeInterval();
		if( ($this->perform_when == 'scheduled' && $ival) || $this->checkDefer()) {
			if($this->perform_when != 'scheduled')
				$ival = null;
			$defer = $this->initDeferred($rowUpdate, $ival);
			if(! $defer->validate() || ! $defer->save()) {
				$GLOBALS['log']->error("Error deferring {$this->type} action: ".$defer->getErrors(true));
			} else {
				if ($defer->getRelatedData('unique')) {
					$parent_type = $defer->getField('parent_type');
					$parent_id = $defer->getField('parent_id');
					$action_id = $defer->getField('wf_action_id');
					$this_id = $defer->getField('id');
					if ($action_id) {
						global $db;
						$query = "DELETE FROM record_actions WHERE parent_id = '{$parent_id}' AND parent_type='{$parent_type}' AND wf_action_id = '{$action_id}' AND id != '{$this_id}'";
						$db->query($query);
					}
				}
			}
			return $defer;
		}
	}

	static function exec_deferred(RowResult $defer) {
		$user_id = $defer->getField('assigned_user_id');
		$module = $defer->getField('parent_type');
		$rel_id = $defer->getField('parent_id');
		$type = $defer->getField('type');
		$deferUpd = $defer->getRowUpdate($user_id);
		$act = self::create_module_action($module, $type);

		if($type == 'runWorkflow') {
			$model = AppConfig::module_primary_bean($module);
			$lq = new ListQuery($model, true, ['filter_deleted' => false]);
			$rel = $lq->queryRecord($rel_id);
			$rowUpdate = $rel->getRowUpdate($user_id);
			$wf = new Workflow();
			$wf->retrieve($defer->getField('workflow_id'));
			$wf->lazy_load();
			if(!$wf->checkConditions($rowUpdate)) {
				$deferUpd->markDeleted();
				return false;
			}

			$applicable_operations = $wf->is_applicable($rowUpdate, 'after', $wf->trigger_action, true);

			try {
				foreach($applicable_operations as $opt) {
					self::process_workflow($rowUpdate, $wf, $opt, $user_id);
				}
				$deferUpd->markDeleted();
				return true;
			} catch(WFProcessingError $e) {
				$GLOBALS['log']->error("Caught WFProcessingError: ".$e->getMessage());
				return false;
			}
		}

		if($act) {
			$act->setSource($defer->getField('source'), $defer->getField('action'), $defer->getField('workflow_id'), $defer->getField('wf_action_id'));
			$act->setPerformWhen('immed');
			$act->deferred_id = $defer->getField('id');
			$act->setUserId($user_id);
			$act->setParams($defer->getField('params'));
			$model = AppConfig::module_primary_bean($module);
			if($model) {
				$lq = new ListQuery($model, true, ['filter_deleted' => false]);
				$reqf = $act->getRequiredFields();
				if($reqf) $lq->addFields($reqf);
				$rel = $lq->queryRecord($rel_id);
			}
			if(! empty($rel)) {
				$rowUpdate = $rel->getRowUpdate($user_id);
				$deferUpd->set('status', 'pending');
				$deferUpd->set('date_start', gmdate('Y-m-d H:i:s'));
				$deferUpd->save();
				try {
					$ret = $act->process($rowUpdate);
					$deferUpd->set('status', 'complete');
					$deferUpd->save();
					return true;
				} catch(WFTemporaryError $e) {
					$GLOBALS['log']->error('Workflow action temp error: '.$e->getMessage());
					$retry = $act->getParam('retries');
					if($retry > 0) {
						$act->setParam('retries', $retry-1);
						$deferUpd->set('params', $act->getParams(true));
						$deferUpd->set('status', 'scheduled');
					} else {
						$deferUpd->set('status', 'failed');
					}
					$deferUpd->set('errmsg', $e->getMessage());
					$deferUpd->save();
				} catch(WFProcessingError $e) {
					$GLOBALS['log']->error('Workflow action error: '.$e->getMessage());
					$deferUpd->set('status', 'failed');
					$deferUpd->set('errmsg', $e->getMessage());
					$deferUpd->save();
				}
			} else {
				$GLOBALS['log']->error("RecordAction related record not found, record deleted");
				$deferUpd->markDeleted();
			}
		} else {
			$GLOBALS['log']->error("Unsupported {$type} action, record deleted");
			// delete action
			$deferUpd->markDeleted();
		}
		return false;
	}

	static function debug_msg($body, $head=null) {
		if(! isset(self::$debug)) {
			$dbg = false;
			if( ($chk = AppConfig::setting('site.workflow_debug')) ) {
				if($chk !== 'manual' || ! empty($_REQUEST['debug']))
					$dbg = true;
			}
			self::$debug = $dbg;
		}
		if(self::$debug) pr2($body, $head, true);
	}
}
